/* Generated by Cython 3.0.0a11 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/tmp/build-env-ns8d2qk1/lib/python3.10/site-packages/capnp/helpers/asyncHelper.h",
            "/tmp/build-env-ns8d2qk1/lib/python3.10/site-packages/capnp/helpers/asyncIoHelper.h",
            "/tmp/build-env-ns8d2qk1/lib/python3.10/site-packages/capnp/helpers/capabilityHelper.cpp",
            "/tmp/build-env-ns8d2qk1/lib/python3.10/site-packages/capnp/helpers/capabilityHelper.h",
            "/tmp/build-env-ns8d2qk1/lib/python3.10/site-packages/capnp/helpers/checkCompiler.h",
            "/tmp/build-env-ns8d2qk1/lib/python3.10/site-packages/capnp/helpers/rpcHelper.h",
            "pytact/graph_api.capnp.h"
        ],
        "extra_compile_args": [
            "-O3",
            "-std=c++14"
        ],
        "include_dirs": [
            "pytact",
            "/tmp/build-env-ns8d2qk1/lib/python3.10/site-packages"
        ],
        "language": "c++",
        "libraries": [
            "capnpc",
            "capnp",
            "capnp-rpc"
        ],
        "name": "pytact.graph_api_capnp_cython",
        "sources": [
            "pytact/graph_api_capnp_cython.pyx",
            "pytact/graph_api.capnp.cpp"
        ]
    },
    "module_name": "pytact.graph_api_capnp_cython"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#define CYTHON_ABI "3_0_0a11"
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x030000AB
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(CYTHON_LIMITED_API)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS 1
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(p))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE(obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pytact__graph_api_capnp_cython
#define __PYX_HAVE_API__pytact__graph_api_capnp_cython
/* Early includes */
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include "capnp/helpers/checkCompiler.h"
#include "capnp/helpers/capabilityHelper.h"
#include "capnp/helpers/rpcHelper.h"
#include "capnp/helpers/asyncHelper.h"
#include "capnp/dynamic.h"
#include "capnp/orphan.h"
#include "capnp/schema.h"
#include "capnp/any.h"
#include "capnp/blob.h"
#include "capnp/message.h"
#include "capnp/schema.capnp.h"
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include "capnp/common.h"
#include "kj/common.h"
#include "kj/array.h"
#include "capnp/serialize.h"
#include "capnp/serialize-packed.h"
#include "kj/string.h"
#include "kj/exception.h"
#include "kj/memory.h"
#include "kj/async.h"
#include "kj/string-tree.h"
#include "kj/time.h"
#include "kj/async-io.h"
#include "capnp/capability.h"
#include "capnp/rpc-twoparty.h"
#include "capnp/schema-parser.h"
#include "capnp/helpers/asyncIoHelper.h"
#include "capnp/helpers/fixMaybe.h"
#include "capnp/helpers/serialize.h"
#include "capnp/helpers/capabilityHelper.cpp"
#include "graph_api.capnp.h"
#include "capnp/list.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 1
#define __Pyx_PyObject_FromString __Pyx_PyStr_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyStr_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const wchar_t *u)
{
    const wchar_t *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "<stringsource>",
  "pytact/graph_api_capnp_cython.pyx",
  "type.pxd",
  "capnp.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* #### Code section: numeric_typedefs ### */

/* "capnp/includes/types.pxd":3
 * from cpython.ref cimport PyObject, Py_INCREF, Py_DECREF
 * from libc.stdint cimport *
 * ctypedef unsigned int uint             # <<<<<<<<<<<<<<
 * ctypedef uint8_t byte
 * ctypedef uint8_t UInt8
 */
typedef unsigned int __pyx_t_5capnp_8includes_5types_uint;

/* "capnp/includes/types.pxd":4
 * from libc.stdint cimport *
 * ctypedef unsigned int uint
 * ctypedef uint8_t byte             # <<<<<<<<<<<<<<
 * ctypedef uint8_t UInt8
 * ctypedef uint16_t UInt16
 */
typedef uint8_t __pyx_t_5capnp_8includes_5types_byte;

/* "capnp/includes/types.pxd":5
 * ctypedef unsigned int uint
 * ctypedef uint8_t byte
 * ctypedef uint8_t UInt8             # <<<<<<<<<<<<<<
 * ctypedef uint16_t UInt16
 * ctypedef uint32_t UInt32
 */
typedef uint8_t __pyx_t_5capnp_8includes_5types_UInt8;

/* "capnp/includes/types.pxd":6
 * ctypedef uint8_t byte
 * ctypedef uint8_t UInt8
 * ctypedef uint16_t UInt16             # <<<<<<<<<<<<<<
 * ctypedef uint32_t UInt32
 * ctypedef uint64_t UInt64
 */
typedef uint16_t __pyx_t_5capnp_8includes_5types_UInt16;

/* "capnp/includes/types.pxd":7
 * ctypedef uint8_t UInt8
 * ctypedef uint16_t UInt16
 * ctypedef uint32_t UInt32             # <<<<<<<<<<<<<<
 * ctypedef uint64_t UInt64
 * ctypedef int8_t Int8
 */
typedef uint32_t __pyx_t_5capnp_8includes_5types_UInt32;

/* "capnp/includes/types.pxd":8
 * ctypedef uint16_t UInt16
 * ctypedef uint32_t UInt32
 * ctypedef uint64_t UInt64             # <<<<<<<<<<<<<<
 * ctypedef int8_t Int8
 * ctypedef int16_t Int16
 */
typedef uint64_t __pyx_t_5capnp_8includes_5types_UInt64;

/* "capnp/includes/types.pxd":9
 * ctypedef uint32_t UInt32
 * ctypedef uint64_t UInt64
 * ctypedef int8_t Int8             # <<<<<<<<<<<<<<
 * ctypedef int16_t Int16
 * ctypedef int32_t Int32
 */
typedef int8_t __pyx_t_5capnp_8includes_5types_Int8;

/* "capnp/includes/types.pxd":10
 * ctypedef uint64_t UInt64
 * ctypedef int8_t Int8
 * ctypedef int16_t Int16             # <<<<<<<<<<<<<<
 * ctypedef int32_t Int32
 * ctypedef int64_t Int64
 */
typedef int16_t __pyx_t_5capnp_8includes_5types_Int16;

/* "capnp/includes/types.pxd":11
 * ctypedef int8_t Int8
 * ctypedef int16_t Int16
 * ctypedef int32_t Int32             # <<<<<<<<<<<<<<
 * ctypedef int64_t Int64
 * 
 */
typedef int32_t __pyx_t_5capnp_8includes_5types_Int32;

/* "capnp/includes/types.pxd":12
 * ctypedef int16_t Int16
 * ctypedef int32_t Int32
 * ctypedef int64_t Int64             # <<<<<<<<<<<<<<
 * 
 * ctypedef char * Object
 */
typedef int64_t __pyx_t_5capnp_8includes_5types_Int64;

/* "capnp/includes/types.pxd":15
 * 
 * ctypedef char * Object
 * ctypedef bint Bool             # <<<<<<<<<<<<<<
 * ctypedef float Float32
 * ctypedef double Float64
 */
typedef int __pyx_t_5capnp_8includes_5types_Bool;

/* "capnp/includes/types.pxd":16
 * ctypedef char * Object
 * ctypedef bint Bool
 * ctypedef float Float32             # <<<<<<<<<<<<<<
 * ctypedef double Float64
 * from libcpp cimport bool as cbool
 */
typedef float __pyx_t_5capnp_8includes_5types_Float32;

/* "capnp/includes/types.pxd":17
 * ctypedef bint Bool
 * ctypedef float Float32
 * ctypedef double Float64             # <<<<<<<<<<<<<<
 * from libcpp cimport bool as cbool
 */
typedef double __pyx_t_5capnp_8includes_5types_Float64;
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField;
struct __pyx_obj_5capnp_3lib_5capnp__StringArrayPtr;
struct __pyx_obj_5capnp_3lib_5capnp_SchemaParser;
struct __pyx_obj_5capnp_3lib_5capnp__DynamicOrphan;
struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader;
struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder;
struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnumField;
struct __pyx_obj_5capnp_3lib_5capnp__Schema;
struct __pyx_obj_5capnp_3lib_5capnp__InterfaceSchema;
struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnum;
struct __pyx_obj_5capnp_3lib_5capnp__DynamicListBuilder;
struct __pyx_obj_5capnp_3lib_5capnp__MessageBuilder;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List;
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List;
struct __pyx_obj___Pyx_EnumMeta;

/* "capnp/includes/types.pxd":14
 * ctypedef int64_t Int64
 * 
 * ctypedef char * Object             # <<<<<<<<<<<<<<
 * ctypedef bint Bool
 * ctypedef float Float32
 */
typedef char *__pyx_t_5capnp_8includes_5types_Object;

/* "capnp/includes/capnp_cpp.pxd":76
 *         Promise[T] attach(Own[PyRefCounter] &, Own[PyRefCounter] &, Own[PyRefCounter] &, Own[PyRefCounter] &)
 * 
 * ctypedef Promise[PyObject *] PyPromise             # <<<<<<<<<<<<<<
 * ctypedef Promise[void] VoidPromise
 * 
 */
typedef  ::kj::Promise<PyObject *>  __pyx_t_5capnp_8includes_9capnp_cpp_PyPromise;

/* "capnp/includes/capnp_cpp.pxd":77
 * 
 * ctypedef Promise[PyObject *] PyPromise
 * ctypedef Promise[void] VoidPromise             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "kj/string-tree.h" namespace " ::kj":
 */
typedef  ::kj::Promise<void>  __pyx_t_5capnp_8includes_9capnp_cpp_VoidPromise;

/* "capnp/includes/capnp_cpp.pxd":108
 *     ctypedef Array[PyObject *] PyArray' ::kj::Array<PyObject *>'
 * 
 * ctypedef Promise[PyArray] PyPromiseArray             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "kj/time.h" namespace " ::kj":
 */
typedef  ::kj::Promise< ::kj::Array<PyObject *>>  __pyx_t_5capnp_8includes_9capnp_cpp_PyPromiseArray;
struct __pyx_opt_args_5capnp_3lib_5capnp_18_StructSchemaField__init;
struct __pyx_opt_args_5capnp_3lib_5capnp_20_DynamicStructReader__init;
struct __pyx_opt_args_5capnp_3lib_5capnp_20_DynamicStructReader_as_builder;
struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder__init;
struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder_init;
struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder__init_by_field;
struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder_copy;

/* "capnp/lib/capnp.pxd":26
 *     cdef C_StructSchema.Field thisptr
 *     cdef object _parent
 *     cdef _init(self, C_StructSchema.Field other, parent=?)             # <<<<<<<<<<<<<<
 * 
 * cdef class _StringArrayPtr:
 */
struct __pyx_opt_args_5capnp_3lib_5capnp_18_StructSchemaField__init {
  int __pyx_n;
  PyObject *parent;
};

/* "capnp/lib/capnp.pxd":58
 *     cdef object _schema
 * 
 *     cdef _init(self, C_DynamicStruct.Reader other, object parent, bint isRoot=?, bint tryRegistry=?)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _get(self, field)
 */
struct __pyx_opt_args_5capnp_3lib_5capnp_20_DynamicStructReader__init {
  int __pyx_n;
  int isRoot;
  int tryRegistry;
};

/* "capnp/lib/capnp.pxd":67
 *     cpdef _has_by_field(self, _StructSchemaField field)
 * 
 *     cpdef as_builder(self, num_first_segment_words=?)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_5capnp_3lib_5capnp_20_DynamicStructReader_as_builder {
  int __pyx_n;
  PyObject *num_first_segment_words;
};

/* "capnp/lib/capnp.pxd":77
 *     cdef object _schema
 * 
 *     cdef _init(self, DynamicStruct_Builder other, object parent, bint isRoot=?, bint tryRegistry=?)             # <<<<<<<<<<<<<<
 * 
 *     cdef _check_write(self)
 */
struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder__init {
  int __pyx_n;
  int isRoot;
  int tryRegistry;
};

/* "capnp/lib/capnp.pxd":88
 *     cpdef _set(self, field, value)
 *     cpdef _has(self, field)
 *     cpdef init(self, field, size=?)             # <<<<<<<<<<<<<<
 *     cpdef _get_by_field(self, _StructSchemaField field)
 *     cpdef _set_by_field(self, _StructSchemaField field, value)
 */
struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder_init {
  int __pyx_n;
  PyObject *size;
};

/* "capnp/lib/capnp.pxd":92
 *     cpdef _set_by_field(self, _StructSchemaField field, value)
 *     cpdef _has_by_field(self, _StructSchemaField field)
 *     cpdef _init_by_field(self, _StructSchemaField field, size=?)             # <<<<<<<<<<<<<<
 *     cpdef init_resizable_list(self, field)
 *     cpdef _DynamicEnumField _which(self)
 */
struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder__init_by_field {
  int __pyx_n;
  PyObject *size;
};

/* "capnp/lib/capnp.pxd":100
 * 
 *     cpdef as_reader(self)
 *     cpdef copy(self, num_first_segment_words=?)             # <<<<<<<<<<<<<<
 * 
 * cdef class _DynamicEnumField:
 */
struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder_copy {
  int __pyx_n;
  PyObject *num_first_segment_words;
};

/* "capnp/lib/capnp.pxd":23
 * 
 * 
 * cdef class _StructSchemaField:             # <<<<<<<<<<<<<<
 *     cdef C_StructSchema.Field thisptr
 *     cdef object _parent
 */
struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__StructSchemaField *__pyx_vtab;
   ::capnp::StructSchema::Field thisptr;
  PyObject *_parent;
};


/* "capnp/lib/capnp.pxd":28
 *     cdef _init(self, C_StructSchema.Field other, parent=?)
 * 
 * cdef class _StringArrayPtr:             # <<<<<<<<<<<<<<
 *     cdef StringPtr * thisptr
 *     cdef object parent
 */
struct __pyx_obj_5capnp_3lib_5capnp__StringArrayPtr {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__StringArrayPtr *__pyx_vtab;
   ::kj::StringPtr *thisptr;
  PyObject *parent;
  size_t size;
};


/* "capnp/lib/capnp.pxd":34
 *     cdef ArrayPtr[StringPtr] asArrayPtr(self) except +reraise_kj_exception
 * 
 * cdef class SchemaParser:             # <<<<<<<<<<<<<<
 *     cdef C_SchemaParser * thisptr
 *     cdef public dict modules_by_id
 */
struct __pyx_obj_5capnp_3lib_5capnp_SchemaParser {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp_SchemaParser *__pyx_vtab;
   ::capnp::SchemaParser *thisptr;
  PyObject *modules_by_id;
  PyObject *_all_imports;
  struct __pyx_obj_5capnp_3lib_5capnp__StringArrayPtr *_last_import_array;
};


/* "capnp/lib/capnp.pxd":41
 *     cpdef _parse_disk_file(self, displayName, diskPath, imports) except +reraise_kj_exception
 * 
 * cdef class _DynamicOrphan:             # <<<<<<<<<<<<<<
 *     cdef C_DynamicOrphan thisptr
 *     cdef public object _parent
 */
struct __pyx_obj_5capnp_3lib_5capnp__DynamicOrphan {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicOrphan *__pyx_vtab;
   ::capnp::Orphan< ::capnp::DynamicValue> thisptr;
  PyObject *_parent;
};


/* "capnp/lib/capnp.pxd":51
 * 
 * 
 * cdef class _DynamicStructReader:             # <<<<<<<<<<<<<<
 *     cdef C_DynamicStruct.Reader thisptr
 *     cdef public object _parent
 */
struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicStructReader *__pyx_vtab;
   ::capnp::DynamicStruct::Reader thisptr;
  PyObject *_parent;
  int is_root;
  PyObject *_obj_to_pin;
  PyObject *_schema;
};


/* "capnp/lib/capnp.pxd":70
 * 
 * 
 * cdef class _DynamicStructBuilder:             # <<<<<<<<<<<<<<
 *     cdef DynamicStruct_Builder thisptr
 *     cdef public object _parent
 */
struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicStructBuilder *__pyx_vtab;
   ::capnp::DynamicStruct::Builder thisptr;
  PyObject *_parent;
  int is_root;
  int _is_written;
  PyObject *_schema;
};


/* "capnp/lib/capnp.pxd":102
 *     cpdef copy(self, num_first_segment_words=?)
 * 
 * cdef class _DynamicEnumField:             # <<<<<<<<<<<<<<
 *     cdef object thisptr
 * 
 */
struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnumField {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicEnumField *__pyx_vtab;
  PyObject *thisptr;
};


/* "capnp/lib/capnp.pxd":108
 *     cpdef _str(self)
 * 
 * cdef class _Schema:             # <<<<<<<<<<<<<<
 *     cdef C_Schema thisptr
 * 
 */
struct __pyx_obj_5capnp_3lib_5capnp__Schema {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *__pyx_vtab;
   ::capnp::Schema thisptr;
};


/* "capnp/lib/capnp.pxd":119
 *     cpdef get_proto(self)
 * 
 * cdef class _InterfaceSchema:             # <<<<<<<<<<<<<<
 *     cdef C_InterfaceSchema thisptr
 *     cdef object __method_names, __method_names_inherited, __methods, __methods_inherited
 */
struct __pyx_obj_5capnp_3lib_5capnp__InterfaceSchema {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__InterfaceSchema *__pyx_vtab;
   ::capnp::InterfaceSchema thisptr;
  PyObject *_InterfaceSchema__method_names;
  PyObject *_InterfaceSchema__method_names_inherited;
  PyObject *_InterfaceSchema__methods;
  PyObject *_InterfaceSchema__methods_inherited;
};


/* "capnp/lib/capnp.pxd":124
 *     cdef _init(self, C_InterfaceSchema other)
 * 
 * cdef class _DynamicEnum:             # <<<<<<<<<<<<<<
 *     cdef capnp.DynamicEnum thisptr
 *     cdef public object _parent
 */
struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnum {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicEnum *__pyx_vtab;
   ::capnp::DynamicEnum thisptr;
  PyObject *_parent;
};


/* "capnp/lib/capnp.pxd":131
 *     cpdef _as_str(self) except +reraise_kj_exception
 * 
 * cdef class _DynamicListBuilder:             # <<<<<<<<<<<<<<
 *     cdef C_DynamicList.Builder thisptr
 *     cdef public object _parent
 */
struct __pyx_obj_5capnp_3lib_5capnp__DynamicListBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicListBuilder *__pyx_vtab;
   ::capnp::DynamicList::Builder thisptr;
  PyObject *_parent;
};


/* "capnp/lib/capnp.pxd":144
 *     cpdef init(self, index, size)
 * 
 * cdef class _MessageBuilder:             # <<<<<<<<<<<<<<
 *     cdef schema_cpp.MessageBuilder * thisptr
 *     cpdef init_root(self, schema)
 */
struct __pyx_obj_5capnp_3lib_5capnp__MessageBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_5capnp_3lib_5capnp__MessageBuilder *__pyx_vtab;
   ::capnp::MessageBuilder *thisptr;
};


/* "pytact/graph_api_capnp_cython.pxd":561
 *         uint size()
 * 
 * cdef class Graph_Node_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Graph_Node_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_vtab;
  Graph::Node::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":567
 *     cdef init(C_Graph_Node_Reader source, object root)
 * 
 * cdef class Graph_EdgeTarget_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Graph_EdgeTarget_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_vtab;
  Graph::EdgeTarget::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":573
 *     cdef init(C_Graph_EdgeTarget_Reader source, object root)
 * 
 * cdef class Graph_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Graph_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_vtab;
  Graph::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":579
 *     cdef init(C_Graph_Reader source, object root)
 * 
 * cdef class IntP_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_IntP_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_vtab;
  IntP::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":585
 *     cdef init(C_IntP_Reader source, object root)
 * 
 * cdef class Definition_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Definition_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_vtab;
  Definition::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":591
 *     cdef init(C_Definition_Reader source, object root)
 * 
 * cdef class FloatP_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_FloatP_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_vtab;
  FloatP::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":597
 *     cdef init(C_FloatP_Reader source, object root)
 * 
 * cdef class Graph_Node_Label_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Graph_Node_Label_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_vtab;
  Graph::Node::Label::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":603
 *     cdef init(C_Graph_Node_Label_Reader source, object root)
 * 
 * cdef class Definition_Status_Substituted_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Definition_Status_Substituted_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_vtab;
  Definition::Status::Substituted::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":609
 *     cdef init(C_Definition_Status_Substituted_Reader source, object root)
 * 
 * cdef class Definition_Status_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Definition_Status_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_vtab;
  Definition::Status::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":615
 *     cdef init(C_Definition_Status_Reader source, object root)
 * 
 * cdef class ProofStep_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_ProofStep_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_vtab;
  ProofStep::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":621
 *     cdef init(C_ProofStep_Reader source, object root)
 * 
 * cdef class Tactic_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Tactic_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_vtab;
  Tactic::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":627
 *     cdef init(C_Tactic_Reader source, object root)
 * 
 * cdef class ProofStep_Tactic_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_ProofStep_Tactic_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_vtab;
  ProofStep::Tactic::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":633
 *     cdef init(C_ProofStep_Tactic_Reader source, object root)
 * 
 * cdef class Outcome_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Outcome_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_vtab;
  Outcome::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":639
 *     cdef init(C_Outcome_Reader source, object root)
 * 
 * cdef class Outcome_Term_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Outcome_Term_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_vtab;
  Outcome::Term::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":645
 *     cdef init(C_Outcome_Term_Reader source, object root)
 * 
 * cdef class ProofState_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_ProofState_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_vtab;
  ProofState::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":651
 *     cdef init(C_ProofState_Reader source, object root)
 * 
 * cdef class Argument_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Argument_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_vtab;
  Argument::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":657
 *     cdef init(C_Argument_Reader source, object root)
 * 
 * cdef class ProofState_Root_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_ProofState_Root_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_vtab;
  ProofState::Root::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":663
 *     cdef init(C_ProofState_Root_Reader source, object root)
 * 
 * cdef class Node_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Node_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_vtab;
  Node::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":669
 *     cdef init(C_Node_Reader source, object root)
 * 
 * cdef class Argument_Term_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Argument_Term_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_vtab;
  Argument::Term::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":675
 *     cdef init(C_Argument_Term_Reader source, object root)
 * 
 * cdef class Graph_EdgeTarget_Target_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Graph_EdgeTarget_Target_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_vtab;
  Graph::EdgeTarget::Target::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":681
 *     cdef init(C_Graph_EdgeTarget_Target_Reader source, object root)
 * 
 * cdef class AbstractTactic_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_AbstractTactic_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_vtab;
  AbstractTactic::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":687
 *     cdef init(C_AbstractTactic_Reader source, object root)
 * 
 * cdef class DataVersion_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_DataVersion_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_vtab;
  DataVersion::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":693
 *     cdef init(C_DataVersion_Reader source, object root)
 * 
 * cdef class Dataset_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Dataset_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_vtab;
  Dataset::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":699
 *     cdef init(C_Dataset_Reader source, object root)
 * 
 * cdef class Exception_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_Exception_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_vtab;
  Exception::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":705
 *     cdef init(C_Exception_Reader source, object root)
 * 
 * cdef class ExecutionResult_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_ExecutionResult_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_vtab;
  ExecutionResult::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":711
 *     cdef init(C_ExecutionResult_Reader source, object root)
 * 
 * cdef class ExecutionResult_NewState_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_ExecutionResult_NewState_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_vtab;
  ExecutionResult::NewState::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":717
 *     cdef init(C_ExecutionResult_NewState_Reader source, object root)
 * 
 * cdef class PredictionProtocol_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_PredictionProtocol_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_vtab;
  PredictionProtocol::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":723
 *     cdef init(C_PredictionProtocol_Reader source, object root)
 * 
 * cdef class PredictionProtocol_Request_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_PredictionProtocol_Request_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_vtab;
  PredictionProtocol::Request::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":729
 *     cdef init(C_PredictionProtocol_Request_Reader source, object root)
 * 
 * cdef class PredictionProtocol_Request_Initialize_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_PredictionProtocol_Request_Initialize_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_vtab;
  PredictionProtocol::Request::Initialize::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":735
 *     cdef init(C_PredictionProtocol_Request_Initialize_Reader source, object root)
 * 
 * cdef class PredictionProtocol_Request_Predict_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_PredictionProtocol_Request_Predict_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_vtab;
  PredictionProtocol::Request::Predict::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":741
 *     cdef init(C_PredictionProtocol_Request_Predict_Reader source, object root)
 * 
 * cdef class PredictionProtocol_Prediction_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_PredictionProtocol_Prediction_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_vtab;
  PredictionProtocol::Prediction::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":747
 *     cdef init(C_PredictionProtocol_Prediction_Reader source, object root)
 * 
 * cdef class PredictionProtocol_TextPrediction_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_PredictionProtocol_TextPrediction_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_vtab;
  PredictionProtocol::TextPrediction::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":753
 *     cdef init(C_PredictionProtocol_TextPrediction_Reader source, object root)
 * 
 * cdef class PredictionProtocol_Response_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_PredictionProtocol_Response_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_vtab;
  PredictionProtocol::Response::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":759
 *     cdef init(C_PredictionProtocol_Response_Reader source, object root)
 * 
 * cdef class PredictionProtocol_Response_Alignment_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_PredictionProtocol_Response_Alignment_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_vtab;
  PredictionProtocol::Response::Alignment::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":765
 *     cdef init(C_PredictionProtocol_Response_Alignment_Reader source, object root)
 * 
 * cdef class ConflatableEdges_Reader:             # <<<<<<<<<<<<<<
 *     cdef C_ConflatableEdges_Reader source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_vtab;
  ConflatableEdges::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":771
 *     cdef init(C_ConflatableEdges_Reader source, object root)
 * 
 * cdef class Uint16_List:             # <<<<<<<<<<<<<<
 *     cdef C_Uint16_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_vtab;
   ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":777
 *     cdef init(C_Uint16_List source, object root)
 * 
 * cdef class Uint32_List:             # <<<<<<<<<<<<<<
 *     cdef C_Uint32_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_vtab;
   ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":783
 *     cdef init(C_Uint32_List source, object root)
 * 
 * cdef class Uint64_List:             # <<<<<<<<<<<<<<
 *     cdef C_Uint64_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_vtab;
   ::capnp::List<uint64_t, ::capnp::Kind::PRIMITIVE>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":789
 *     cdef init(C_Uint64_List source, object root)
 * 
 * cdef class Int64_List:             # <<<<<<<<<<<<<<
 *     cdef C_Int64_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_vtab;
   ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":795
 *     cdef init(C_Int64_List source, object root)
 * 
 * cdef class String_List:             # <<<<<<<<<<<<<<
 *     cdef C_String_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_String_List *__pyx_vtab;
   ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":800
 *     @staticmethod
 *     cdef init(C_String_List source, object root)
 * cdef class PredictionProtocol_Prediction_Reader_List:             # <<<<<<<<<<<<<<
 *     cdef C_PredictionProtocol_Prediction_Reader_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_vtab;
   ::capnp::List<PredictionProtocol::Prediction, ::capnp::Kind::STRUCT>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":805
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Prediction_Reader_List source, object root)
 * cdef class Graph_Node_Reader_List:             # <<<<<<<<<<<<<<
 *     cdef C_Graph_Node_Reader_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_vtab;
   ::capnp::List<Graph::Node, ::capnp::Kind::STRUCT>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":810
 *     @staticmethod
 *     cdef init(C_Graph_Node_Reader_List source, object root)
 * cdef class AbstractTactic_Reader_List:             # <<<<<<<<<<<<<<
 *     cdef C_AbstractTactic_Reader_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_vtab;
   ::capnp::List<AbstractTactic, ::capnp::Kind::STRUCT>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":815
 *     @staticmethod
 *     cdef init(C_AbstractTactic_Reader_List source, object root)
 * cdef class ProofStep_Reader_List:             # <<<<<<<<<<<<<<
 *     cdef C_ProofStep_Reader_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_vtab;
   ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":820
 *     @staticmethod
 *     cdef init(C_ProofStep_Reader_List source, object root)
 * cdef class PredictionProtocol_TextPrediction_Reader_List:             # <<<<<<<<<<<<<<
 *     cdef C_PredictionProtocol_TextPrediction_Reader_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_vtab;
   ::capnp::List<PredictionProtocol::TextPrediction, ::capnp::Kind::STRUCT>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":825
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_TextPrediction_Reader_List source, object root)
 * cdef class ProofState_Reader_List:             # <<<<<<<<<<<<<<
 *     cdef C_ProofState_Reader_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_vtab;
   ::capnp::List<ProofState, ::capnp::Kind::STRUCT>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":830
 *     @staticmethod
 *     cdef init(C_ProofState_Reader_List source, object root)
 * cdef class Graph_EdgeTarget_Reader_List:             # <<<<<<<<<<<<<<
 *     cdef C_Graph_EdgeTarget_Reader_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_vtab;
   ::capnp::List<Graph::EdgeTarget, ::capnp::Kind::STRUCT>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":835
 *     @staticmethod
 *     cdef init(C_Graph_EdgeTarget_Reader_List source, object root)
 * cdef class Argument_Reader_List:             # <<<<<<<<<<<<<<
 *     cdef C_Argument_Reader_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_vtab;
   ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":840
 *     @staticmethod
 *     cdef init(C_Argument_Reader_List source, object root)
 * cdef class Outcome_Reader_List:             # <<<<<<<<<<<<<<
 *     cdef C_Outcome_Reader_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_vtab;
   ::capnp::List<Outcome, ::capnp::Kind::STRUCT>::Reader source;
  PyObject *root;
};


/* "pytact/graph_api_capnp_cython.pxd":845
 *     @staticmethod
 *     cdef init(C_Outcome_Reader_List source, object root)
 * cdef class Node_Reader_List:             # <<<<<<<<<<<<<<
 *     cdef C_Node_Reader_List source
 *     cdef object root
 */
struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_vtab;
   ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader source;
  PyObject *root;
};


/* "EnumBase":12
 * 
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):             # <<<<<<<<<<<<<<
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)
 */
struct __pyx_obj___Pyx_EnumMeta {
  PyHeapTypeObject __pyx_base;
};



/* "capnp/lib/capnp.pxd":23
 * 
 * 
 * cdef class _StructSchemaField:             # <<<<<<<<<<<<<<
 *     cdef C_StructSchema.Field thisptr
 *     cdef object _parent
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__StructSchemaField {
  PyObject *(*_init)(struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField *,  ::capnp::StructSchema::Field, struct __pyx_opt_args_5capnp_3lib_5capnp_18_StructSchemaField__init *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__StructSchemaField *__pyx_vtabptr_5capnp_3lib_5capnp__StructSchemaField;


/* "capnp/lib/capnp.pxd":28
 *     cdef _init(self, C_StructSchema.Field other, parent=?)
 * 
 * cdef class _StringArrayPtr:             # <<<<<<<<<<<<<<
 *     cdef StringPtr * thisptr
 *     cdef object parent
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__StringArrayPtr {
   ::kj::ArrayPtr< ::kj::StringPtr>  (*asArrayPtr)(struct __pyx_obj_5capnp_3lib_5capnp__StringArrayPtr *);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__StringArrayPtr *__pyx_vtabptr_5capnp_3lib_5capnp__StringArrayPtr;


/* "capnp/lib/capnp.pxd":34
 *     cdef ArrayPtr[StringPtr] asArrayPtr(self) except +reraise_kj_exception
 * 
 * cdef class SchemaParser:             # <<<<<<<<<<<<<<
 *     cdef C_SchemaParser * thisptr
 *     cdef public dict modules_by_id
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp_SchemaParser {
  PyObject *(*_parse_disk_file)(struct __pyx_obj_5capnp_3lib_5capnp_SchemaParser *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp_SchemaParser *__pyx_vtabptr_5capnp_3lib_5capnp_SchemaParser;


/* "capnp/lib/capnp.pxd":41
 *     cpdef _parse_disk_file(self, displayName, diskPath, imports) except +reraise_kj_exception
 * 
 * cdef class _DynamicOrphan:             # <<<<<<<<<<<<<<
 *     cdef C_DynamicOrphan thisptr
 *     cdef public object _parent
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicOrphan {
  PyObject *(*_init)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicOrphan *,  ::capnp::Orphan< ::capnp::DynamicValue>, PyObject *);
   ::capnp::Orphan< ::capnp::DynamicValue> (*move)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicOrphan *);
  PyObject *(*get)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicOrphan *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicOrphan *__pyx_vtabptr_5capnp_3lib_5capnp__DynamicOrphan;


/* "capnp/lib/capnp.pxd":51
 * 
 * 
 * cdef class _DynamicStructReader:             # <<<<<<<<<<<<<<
 *     cdef C_DynamicStruct.Reader thisptr
 *     cdef public object _parent
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicStructReader {
  PyObject *(*_init)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *,  ::capnp::DynamicStruct::Reader, PyObject *, struct __pyx_opt_args_5capnp_3lib_5capnp_20_DynamicStructReader__init *__pyx_optional_args);
  PyObject *(*_get)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_has)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnumField *(*_which)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *, int __pyx_skip_dispatch);
  PyObject *(*_which_str)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *, int __pyx_skip_dispatch);
  PyObject *(*_get_by_field)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *, struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField *, int __pyx_skip_dispatch);
  PyObject *(*_has_by_field)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *, struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField *, int __pyx_skip_dispatch);
  PyObject *(*as_builder)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *, int __pyx_skip_dispatch, struct __pyx_opt_args_5capnp_3lib_5capnp_20_DynamicStructReader_as_builder *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicStructReader *__pyx_vtabptr_5capnp_3lib_5capnp__DynamicStructReader;


/* "capnp/lib/capnp.pxd":70
 * 
 * 
 * cdef class _DynamicStructBuilder:             # <<<<<<<<<<<<<<
 *     cdef DynamicStruct_Builder thisptr
 *     cdef public object _parent
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicStructBuilder {
  PyObject *(*_init)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *,  ::capnp::DynamicStruct::Builder, PyObject *, struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder__init *__pyx_optional_args);
  PyObject *(*_check_write)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *);
  PyObject *(*to_bytes)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, int __pyx_skip_dispatch);
  PyObject *(*to_segments)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_to_bytes_packed_helper)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*to_bytes_packed)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_get)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_set)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_has)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*init)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder_init *__pyx_optional_args);
  PyObject *(*_get_by_field)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField *, int __pyx_skip_dispatch);
  PyObject *(*_set_by_field)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_has_by_field)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField *, int __pyx_skip_dispatch);
  PyObject *(*_init_by_field)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField *, int __pyx_skip_dispatch, struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder__init_by_field *__pyx_optional_args);
  PyObject *(*init_resizable_list)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnumField *(*_which)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_which_str)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, int __pyx_skip_dispatch);
  PyObject *(*adopt)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, PyObject *, struct __pyx_obj_5capnp_3lib_5capnp__DynamicOrphan *, int __pyx_skip_dispatch);
  PyObject *(*disown)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*as_reader)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, int __pyx_skip_dispatch);
  PyObject *(*copy)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder *, int __pyx_skip_dispatch, struct __pyx_opt_args_5capnp_3lib_5capnp_21_DynamicStructBuilder_copy *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicStructBuilder *__pyx_vtabptr_5capnp_3lib_5capnp__DynamicStructBuilder;


/* "capnp/lib/capnp.pxd":102
 *     cpdef copy(self, num_first_segment_words=?)
 * 
 * cdef class _DynamicEnumField:             # <<<<<<<<<<<<<<
 *     cdef object thisptr
 * 
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicEnumField {
  PyObject *(*_init)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnumField *, PyObject *);
  PyObject *(*_str)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnumField *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicEnumField *__pyx_vtabptr_5capnp_3lib_5capnp__DynamicEnumField;


/* "capnp/lib/capnp.pxd":108
 *     cpdef _str(self)
 * 
 * cdef class _Schema:             # <<<<<<<<<<<<<<
 *     cdef C_Schema thisptr
 * 
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema {
  PyObject *(*_init)(struct __pyx_obj_5capnp_3lib_5capnp__Schema *,  ::capnp::Schema);
  PyObject *(*as_const_value)(struct __pyx_obj_5capnp_3lib_5capnp__Schema *, int __pyx_skip_dispatch);
  PyObject *(*as_struct)(struct __pyx_obj_5capnp_3lib_5capnp__Schema *, int __pyx_skip_dispatch);
  PyObject *(*as_interface)(struct __pyx_obj_5capnp_3lib_5capnp__Schema *, int __pyx_skip_dispatch);
  PyObject *(*as_enum)(struct __pyx_obj_5capnp_3lib_5capnp__Schema *, int __pyx_skip_dispatch);
  PyObject *(*get_proto)(struct __pyx_obj_5capnp_3lib_5capnp__Schema *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *__pyx_vtabptr_5capnp_3lib_5capnp__Schema;


/* "capnp/lib/capnp.pxd":119
 *     cpdef get_proto(self)
 * 
 * cdef class _InterfaceSchema:             # <<<<<<<<<<<<<<
 *     cdef C_InterfaceSchema thisptr
 *     cdef object __method_names, __method_names_inherited, __methods, __methods_inherited
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__InterfaceSchema {
  PyObject *(*_init)(struct __pyx_obj_5capnp_3lib_5capnp__InterfaceSchema *,  ::capnp::InterfaceSchema);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__InterfaceSchema *__pyx_vtabptr_5capnp_3lib_5capnp__InterfaceSchema;


/* "capnp/lib/capnp.pxd":124
 *     cdef _init(self, C_InterfaceSchema other)
 * 
 * cdef class _DynamicEnum:             # <<<<<<<<<<<<<<
 *     cdef capnp.DynamicEnum thisptr
 *     cdef public object _parent
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicEnum {
  PyObject *(*_init)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnum *,  ::capnp::DynamicEnum, PyObject *);
  PyObject *(*_as_str)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnum *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicEnum *__pyx_vtabptr_5capnp_3lib_5capnp__DynamicEnum;


/* "capnp/lib/capnp.pxd":131
 *     cpdef _as_str(self) except +reraise_kj_exception
 * 
 * cdef class _DynamicListBuilder:             # <<<<<<<<<<<<<<
 *     cdef C_DynamicList.Builder thisptr
 *     cdef public object _parent
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicListBuilder {
  PyObject *(*_init)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicListBuilder *,  ::capnp::DynamicList::Builder, PyObject *);
  PyObject *(*_get)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicListBuilder *, int64_t, int __pyx_skip_dispatch);
  PyObject *(*_set)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicListBuilder *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*adopt)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicListBuilder *, PyObject *, struct __pyx_obj_5capnp_3lib_5capnp__DynamicOrphan *, int __pyx_skip_dispatch);
  PyObject *(*disown)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicListBuilder *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*init)(struct __pyx_obj_5capnp_3lib_5capnp__DynamicListBuilder *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicListBuilder *__pyx_vtabptr_5capnp_3lib_5capnp__DynamicListBuilder;


/* "capnp/lib/capnp.pxd":144
 *     cpdef init(self, index, size)
 * 
 * cdef class _MessageBuilder:             # <<<<<<<<<<<<<<
 *     cdef schema_cpp.MessageBuilder * thisptr
 *     cpdef init_root(self, schema)
 */

struct __pyx_vtabstruct_5capnp_3lib_5capnp__MessageBuilder {
  PyObject *(*init_root)(struct __pyx_obj_5capnp_3lib_5capnp__MessageBuilder *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_root)(struct __pyx_obj_5capnp_3lib_5capnp__MessageBuilder *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_root_as_any)(struct __pyx_obj_5capnp_3lib_5capnp__MessageBuilder *, int __pyx_skip_dispatch);
  PyObject *(*set_root)(struct __pyx_obj_5capnp_3lib_5capnp__MessageBuilder *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_segments_for_output)(struct __pyx_obj_5capnp_3lib_5capnp__MessageBuilder *, int __pyx_skip_dispatch);
  PyObject *(*new_orphan)(struct __pyx_obj_5capnp_3lib_5capnp__MessageBuilder *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5capnp_3lib_5capnp__MessageBuilder *__pyx_vtabptr_5capnp_3lib_5capnp__MessageBuilder;


/* "pytact/graph_api_capnp_cython.pyx":13
 * import pytact.graph_api_capnp
 * 
 * cdef class Graph_Node_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     Fits exactly in 128 bits.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Reader {
  PyObject *(*init)(Graph::Node::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Reader;


/* "pytact/graph_api_capnp_cython.pyx":96
 * pytact.graph_api_capnp.Graph.Node.schema = _Schema()._init(getGraph_Node_Schema()).as_struct()
 * 
 * cdef class Graph_EdgeTarget_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     Fits exactly in 64 bits. Let's keep it that way.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader {
  PyObject *(*init)(Graph::EdgeTarget::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader;


/* "pytact/graph_api_capnp_cython.pyx":128
 * pytact.graph_api_capnp.Graph.EdgeTarget.schema = _Schema()._init(getGraph_EdgeTarget_Schema()).as_struct()
 * 
 * cdef class Graph_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     A graph is the main data store (the 'heap') that contains the bulk of the data in the dataset and
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Reader {
  PyObject *(*init)(Graph::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Reader;


/* "pytact/graph_api_capnp_cython.pyx":178
 * pytact.graph_api_capnp.Graph.schema = _Schema()._init(getGraph_Schema()).as_struct()
 * 
 * cdef class IntP_Reader:             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_IntP_Reader {
  PyObject *(*init)(IntP::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_IntP_Reader;


/* "pytact/graph_api_capnp_cython.pyx":202
 * pytact.graph_api_capnp.IntP.schema = _Schema()._init(getIntP_Schema()).as_struct()
 * 
 * cdef class Definition_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     A definition of the CIC, which is either an constant, inductive, constructor, projection or section
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Reader {
  PyObject *(*init)(Definition::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Reader;


/* "pytact/graph_api_capnp_cython.pyx":375
 * pytact.graph_api_capnp.Definition.schema = _Schema()._init(getDefinition_Schema()).as_struct()
 * 
 * cdef class FloatP_Reader:             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_FloatP_Reader {
  PyObject *(*init)(FloatP::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_FloatP_Reader;


/* "pytact/graph_api_capnp_cython.pyx":399
 * pytact.graph_api_capnp.FloatP.schema = _Schema()._init(getFloatP_Schema()).as_struct()
 * 
 * cdef class Graph_Node_Label_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     Inlined for efficiency purposes
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader {
  PyObject *(*init)(Graph::Node::Label::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader;


/* "pytact/graph_api_capnp_cython.pyx":631
 *         return self.source.hasPrimitive()
 * 
 * cdef class Definition_Status_Substituted_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     A definition that was obtained by performing some sort of module functor substitution.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader {
  PyObject *(*init)(Definition::Status::Substituted::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader;


/* "pytact/graph_api_capnp_cython.pyx":669
 *         return self.source.getNodeIndex()
 * 
 * cdef class Definition_Status_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     A definition can have different origins, which are encoded here
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Status_Reader {
  PyObject *(*init)(Definition::Status::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Status_Reader;


/* "pytact/graph_api_capnp_cython.pyx":729
 *         return self.source.isSubstituted()
 * 
 * cdef class ProofStep_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     A proof step is the execution of a single tactic on one or more proof states, producing a list of outcomes.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Reader {
  PyObject *(*init)(ProofStep::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Reader;


/* "pytact/graph_api_capnp_cython.pyx":770
 * pytact.graph_api_capnp.ProofStep.schema = _Schema()._init(getProofStep_Schema()).as_struct()
 * 
 * cdef class Tactic_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     A concrete tactic with it's parameters determined. Somewhat strangely, this struct does not actually include
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Tactic_Reader {
  PyObject *(*init)(Tactic::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Tactic_Reader;


/* "pytact/graph_api_capnp_cython.pyx":868
 * pytact.graph_api_capnp.Tactic.schema = _Schema()._init(getTactic_Schema()).as_struct()
 * 
 * cdef class ProofStep_Tactic_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     The tactic that generated the proof step. Note that the arguments of the tactic can be found in the
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader {
  PyObject *(*init)(ProofStep::Tactic::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader;


/* "pytact/graph_api_capnp_cython.pyx":922
 *         return self.source.hasKnown()
 * 
 * cdef class Outcome_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     An outcome is the result of running a tactic on a proof state. A tactic may run on multiple proof states.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Reader {
  PyObject *(*init)(Outcome::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Reader;


/* "pytact/graph_api_capnp_cython.pyx":996
 * pytact.graph_api_capnp.Outcome.schema = _Schema()._init(getOutcome_Schema()).as_struct()
 * 
 * cdef class Outcome_Term_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     The proof term that witnesses the transition from the before state to the after states. It contains a hole
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader {
  PyObject *(*init)(Outcome::Term::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1035
 *         return self.source.getNodeIndex()
 * 
 * cdef class ProofState_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     A proof state represents a particular point in the tactical proof of a constant.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Reader {
  PyObject *(*init)(ProofState::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1135
 * pytact.graph_api_capnp.ProofState.schema = _Schema()._init(getProofState_Schema()).as_struct()
 * 
 * cdef class Argument_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     A concrete argument of a tactic.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Reader {
  PyObject *(*init)(Argument::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1184
 * pytact.graph_api_capnp.Argument.schema = _Schema()._init(getArgument_Schema()).as_struct()
 * 
 * cdef class ProofState_Root_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     The entry-point of the proof state, all nodes that are 'part of' the proof state are reachable from here.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader {
  PyObject *(*init)(ProofState::Root::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1221
 *         return self.source.getNodeIndex()
 * 
 * cdef class Node_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     A node in the CIC graph
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Node_Reader {
  PyObject *(*init)(Node::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Node_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1259
 * pytact.graph_api_capnp.Node.schema = _Schema()._init(getNode_Schema()).as_struct()
 * 
 * cdef class Argument_Term_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     The root of a graph representing an argument that is a term in the calculus of constructions.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Term_Reader {
  PyObject *(*init)(Argument::Term::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Term_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1296
 *         return self.source.getNodeIndex()
 * 
 * cdef class Graph_EdgeTarget_Target_Reader:             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader {
  PyObject *(*init)(Graph::EdgeTarget::Target::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1330
 *         return self.source.getNodeIndex()
 * 
 * cdef class AbstractTactic_Reader:             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader {
  PyObject *(*init)(AbstractTactic::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1364
 * pytact.graph_api_capnp.AbstractTactic.schema = _Schema()._init(getAbstractTactic_Schema()).as_struct()
 * 
 * cdef class DataVersion_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     Version info of a dataset.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_DataVersion_Reader {
  PyObject *(*init)(DataVersion::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_DataVersion_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1398
 * pytact.graph_api_capnp.DataVersion.schema = _Schema()._init(getDataVersion_Schema()).as_struct()
 * 
 * cdef class Dataset_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     Every file in the dataset contains a single message of type `Dataset`. Every file corresponds to
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Dataset_Reader {
  PyObject *(*init)(Dataset::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Dataset_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1486
 * pytact.graph_api_capnp.Dataset.schema = _Schema()._init(getDataset_Schema()).as_struct()
 * 
 * cdef class Exception_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     A list of things that can go wrong during reinforcement learning.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Exception_Reader {
  PyObject *(*init)(Exception::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Exception_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1543
 * pytact.graph_api_capnp.Exception.schema = _Schema()._init(getException_Schema()).as_struct()
 * 
 * cdef class ExecutionResult_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     The result of executing a tactic on a proof state.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader {
  PyObject *(*init)(ExecutionResult::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1615
 * pytact.graph_api_capnp.ExecutionResult.schema = _Schema()._init(getExecutionResult_Schema()).as_struct()
 * 
 * cdef class ExecutionResult_NewState_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     The tactic ran successfully and produced a new proof state.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader {
  PyObject *(*init)(ExecutionResult::NewState::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1651
 *         return self.source.hasState()
 * 
 * cdef class PredictionProtocol_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     This protocol works by exchanging raw messages over a socket. The protocol is fully linear.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader {
  PyObject *(*init)(PredictionProtocol::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1676
 * pytact.graph_api_capnp.PredictionProtocol.schema = _Schema()._init(getPredictionProtocol_Schema()).as_struct()
 * 
 * cdef class PredictionProtocol_Request_Reader:             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader {
  PyObject *(*init)(PredictionProtocol::Request::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1745
 * pytact.graph_api_capnp.PredictionProtocol.Request.schema = _Schema()._init(getPredictionProtocol_Request_Schema()).as_struct()
 * 
 * cdef class PredictionProtocol_Request_Initialize_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     Start a context for making tactical predictions for proof search. The context includes the tactics
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader {
  PyObject *(*init)(PredictionProtocol::Request::Initialize::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1833
 *         return self.source.getRepresentative()
 * 
 * cdef class PredictionProtocol_Request_Predict_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     Request a list of tactic predictions given the graph of a proof state.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader {
  PyObject *(*init)(PredictionProtocol::Request::Predict::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1876
 *         return self.source.hasState()
 * 
 * cdef class PredictionProtocol_Prediction_Reader:             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader {
  PyObject *(*init)(PredictionProtocol::Prediction::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1914
 * pytact.graph_api_capnp.PredictionProtocol.Prediction.schema = _Schema()._init(getPredictionProtocol_Prediction_Schema()).as_struct()
 * 
 * cdef class PredictionProtocol_TextPrediction_Reader:             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader {
  PyObject *(*init)(PredictionProtocol::TextPrediction::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader;


/* "pytact/graph_api_capnp_cython.pyx":1946
 * pytact.graph_api_capnp.PredictionProtocol.TextPrediction.schema = _Schema()._init(getPredictionProtocol_TextPrediction_Schema()).as_struct()
 * 
 * cdef class PredictionProtocol_Response_Reader:             # <<<<<<<<<<<<<<
 *     """
 *     See Request for documentation.
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader {
  PyObject *(*init)(PredictionProtocol::Response::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader;


/* "pytact/graph_api_capnp_cython.pyx":2020
 * pytact.graph_api_capnp.PredictionProtocol.Response.schema = _Schema()._init(getPredictionProtocol_Response_Schema()).as_struct()
 * 
 * cdef class PredictionProtocol_Response_Alignment_Reader:             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader {
  PyObject *(*init)(PredictionProtocol::Response::Alignment::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader;


/* "pytact/graph_api_capnp_cython.pyx":2053
 *         return self.source.hasUnalignedDefinitions()
 * 
 * cdef class ConflatableEdges_Reader:             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader {
  PyObject *(*init)(ConflatableEdges::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader;


/* "pytact/graph_api_capnp_cython.pyx":2074
 * 
 * 
 * cdef class Uint16_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint16_List {
  PyObject *(*init)( ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint16_List;


/* "pytact/graph_api_capnp_cython.pyx":2092
 *         return self.source.size()
 * 
 * cdef class Uint32_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint32_List {
  PyObject *(*init)( ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint32_List;


/* "pytact/graph_api_capnp_cython.pyx":2110
 *         return self.source.size()
 * 
 * cdef class Uint64_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint64_List {
  PyObject *(*init)( ::capnp::List<uint64_t, ::capnp::Kind::PRIMITIVE>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint64_List;


/* "pytact/graph_api_capnp_cython.pyx":2128
 *         return self.source.size()
 * 
 * cdef class Int64_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Int64_List {
  PyObject *(*init)( ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Int64_List;


/* "pytact/graph_api_capnp_cython.pyx":2146
 *         return self.source.size()
 * 
 * cdef class String_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_String_List {
  PyObject *(*init)( ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_String_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_String_List;


/* "pytact/graph_api_capnp_cython.pyx":2164
 *     def __len__(self):
 *         return self.source.size()
 * cdef class PredictionProtocol_Prediction_Reader_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List {
  PyObject *(*init)( ::capnp::List<PredictionProtocol::Prediction, ::capnp::Kind::STRUCT>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List;


/* "pytact/graph_api_capnp_cython.pyx":2181
 *     def __len__(self):
 *         return self.source.size()
 * cdef class Graph_Node_Reader_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List {
  PyObject *(*init)( ::capnp::List<Graph::Node, ::capnp::Kind::STRUCT>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List;


/* "pytact/graph_api_capnp_cython.pyx":2198
 *     def __len__(self):
 *         return self.source.size()
 * cdef class AbstractTactic_Reader_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List {
  PyObject *(*init)( ::capnp::List<AbstractTactic, ::capnp::Kind::STRUCT>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List;


/* "pytact/graph_api_capnp_cython.pyx":2215
 *     def __len__(self):
 *         return self.source.size()
 * cdef class ProofStep_Reader_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List {
  PyObject *(*init)( ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List;


/* "pytact/graph_api_capnp_cython.pyx":2232
 *     def __len__(self):
 *         return self.source.size()
 * cdef class PredictionProtocol_TextPrediction_Reader_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List {
  PyObject *(*init)( ::capnp::List<PredictionProtocol::TextPrediction, ::capnp::Kind::STRUCT>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List;


/* "pytact/graph_api_capnp_cython.pyx":2249
 *     def __len__(self):
 *         return self.source.size()
 * cdef class ProofState_Reader_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Reader_List {
  PyObject *(*init)( ::capnp::List<ProofState, ::capnp::Kind::STRUCT>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Reader_List;


/* "pytact/graph_api_capnp_cython.pyx":2266
 *     def __len__(self):
 *         return self.source.size()
 * cdef class Graph_EdgeTarget_Reader_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List {
  PyObject *(*init)( ::capnp::List<Graph::EdgeTarget, ::capnp::Kind::STRUCT>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List;


/* "pytact/graph_api_capnp_cython.pyx":2283
 *     def __len__(self):
 *         return self.source.size()
 * cdef class Argument_Reader_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Reader_List {
  PyObject *(*init)( ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Reader_List;


/* "pytact/graph_api_capnp_cython.pyx":2300
 *     def __len__(self):
 *         return self.source.size()
 * cdef class Outcome_Reader_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Reader_List {
  PyObject *(*init)( ::capnp::List<Outcome, ::capnp::Kind::STRUCT>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Reader_List;


/* "pytact/graph_api_capnp_cython.pyx":2317
 *     def __len__(self):
 *         return self.source.size()
 * cdef class Node_Reader_List:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Node_Reader_List {
  PyObject *(*init)( ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader, PyObject *);
};
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_vtabptr_6pytact_22graph_api_capnp_cython_Node_Reader_List;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* MoveIfSupported.proto */
#if __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600)
  #include <utility>
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) std::move(x)
#else
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) x
#endif

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
static CYTHON_UNUSED int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
#if __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if __STDC_VERSION__ >= 201112L || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#if !CYTHON_USE_MODULE_STATE
static PyTypeObject *__pyx_CyFunctionType = 0;
#endif
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_IsCyOrPyCFunction(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* SetNewInClass.proto */
static int __Pyx_SetNewInClass(PyObject *ns, PyObject *name, PyObject *value);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* EnumClassDecl.proto */
#if defined (_MSC_VER)
  #if _MSC_VER >= 1910
    #define __PYX_ENUM_CLASS_DECL enum
  #else
    #define __PYX_ENUM_CLASS_DECL
  #endif
#else
  #define __PYX_ENUM_CLASS_DECL enum
#endif

/* None.proto */
#include <new>

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint16_t(uint16_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_Occurred(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str);
#else
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);
#endif

/* #### Code section: module_declarations ### */
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_init(Graph::Node::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_init(Graph::EdgeTarget::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_12Graph_Reader_init(Graph::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11IntP_Reader_init(IntP::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_17Definition_Reader_init(Definition::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_13FloatP_Reader_init(FloatP::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_init(Graph::Node::Label::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_init(Definition::Status::Substituted::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_init(Definition::Status::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_init(ProofStep::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_13Tactic_Reader_init(Tactic::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_init(ProofStep::Tactic::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_14Outcome_Reader_init(Outcome::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_init(Outcome::Term::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_17ProofState_Reader_init(ProofState::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_15Argument_Reader_init(Argument::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_init(ProofState::Root::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11Node_Reader_init(Node::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_init(Argument::Term::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_init(Graph::EdgeTarget::Target::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_init(AbstractTactic::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_init(DataVersion::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_14Dataset_Reader_init(Dataset::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_16Exception_Reader_init(Exception::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_init(ExecutionResult::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_init(ExecutionResult::NewState::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_init(PredictionProtocol::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_init(PredictionProtocol::Request::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_init(PredictionProtocol::Request::Initialize::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_init(PredictionProtocol::Request::Predict::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_init(PredictionProtocol::Prediction::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_init(PredictionProtocol::TextPrediction::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_init(PredictionProtocol::Response::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_init(PredictionProtocol::Response::Alignment::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_init(ConflatableEdges::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11Uint16_List_init( ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11Uint32_List_init( ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11Uint64_List_init( ::capnp::List<uint64_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_10Int64_List_init( ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11String_List_init( ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_init( ::capnp::List<PredictionProtocol::Prediction, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_init( ::capnp::List<Graph::Node, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_init( ::capnp::List<AbstractTactic, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_init( ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_init( ::capnp::List<PredictionProtocol::TextPrediction, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_init( ::capnp::List<ProofState, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_init( ::capnp::List<Graph::EdgeTarget, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_init( ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_init( ::capnp::List<Outcome, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/
static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_16Node_Reader_List_init( ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root); /* proto*/

/* Module declarations from "libc.string" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "libc.stdio" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "__builtin__" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "cpython.type" */
#if !CYTHON_USE_MODULE_STATE
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;
#endif

/* Module declarations from "cpython" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "cpython.object" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "cpython.ref" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "libc.stdint" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "libcpp" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "capnp.includes.types" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "capnp" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "capnp.helpers" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "capnp.helpers.non_circular" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "capnp.includes.schema_cpp" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "capnp.includes.capnp_cpp" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "capnp.includes" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "capnp.helpers.helpers" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "capnp.lib.capnp" */
#if !CYTHON_USE_MODULE_STATE
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__StructSchemaField = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__StringArrayPtr = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp_SchemaParser = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicOrphan = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicStructBuilder = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicEnumField = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__Schema = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__InterfaceSchema = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicEnum = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicListBuilder = 0;
static PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__MessageBuilder = 0;
#endif
static PyObject *(*__pyx_f_5capnp_3lib_5capnp_to_python_reader)( ::capnp::DynamicValue::Reader, PyObject *); /*proto*/
static PyObject *(*__pyx_f_5capnp_3lib_5capnp_wrap_dynamic_struct_reader)( ::capnp::Response< ::capnp::DynamicStruct> &); /*proto*/
static PyObject *(*__pyx_f_5capnp_3lib_5capnp_wrap_remote_call)(PyObject *,  ::capnp::Response< ::capnp::DynamicStruct> &); /*proto*/
static  ::kj::Promise<void>  *(*__pyx_f_5capnp_3lib_5capnp_call_server_method)(PyObject *, char *,  ::capnp::CallContext< ::capnp::DynamicStruct, ::capnp::DynamicStruct> &); /*proto*/
static PyObject *(*__pyx_f_5capnp_3lib_5capnp_convert_array_pyobject)( ::kj::Array<PyObject *> &); /*proto*/
static  ::kj::Promise<PyObject *>  *(*__pyx_f_5capnp_3lib_5capnp_extract_promise)(PyObject *); /*proto*/
static  ::capnp::RemotePromise< ::capnp::DynamicStruct> *(*__pyx_f_5capnp_3lib_5capnp_extract_remote_promise)(PyObject *); /*proto*/
static PyObject *(*__pyx_f_5capnp_3lib_5capnp_wrap_kj_exception)( ::kj::Exception &); /*proto*/
static PyObject *(*__pyx_f_5capnp_3lib_5capnp_wrap_kj_exception_for_reraise)( ::kj::Exception &); /*proto*/
static PyObject *(*__pyx_f_5capnp_3lib_5capnp_get_exception_info)(PyObject *, PyObject *, PyObject *); /*proto*/

/* Module declarations from "pytact.graph_api_capnp_cython" */
#if !CYTHON_USE_MODULE_STATE
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List = 0;
static PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List = 0;
static PyTypeObject *__pyx_ptype___Pyx_EnumMeta = 0;
#endif
static PyObject *__Pyx_OrderedDict = 0;
static PyObject *__Pyx_EnumBase = 0;
static PyObject *__Pyx_globals = 0;
static PyObject *__pyx_unpickle___Pyx_EnumMeta__set_state(struct __pyx_obj___Pyx_EnumMeta *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pytact.graph_api_capnp_cython"
extern int __pyx_module_is_main_pytact__graph_api_capnp_cython;
int __pyx_module_is_main_pytact__graph_api_capnp_cython = 0;

/* Implementation of "pytact.graph_api_capnp_cython" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_ValueError;
/* #### Code section: string_decls ### */
static const char __pyx_k_v[] = "v";
static const char __pyx_k__3[] = ".";
static const char __pyx_k__4[] = "*";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_APP[] = "APP";
static const char __pyx_k_FIX[] = "FIX";
static const char __pyx_k_INT[] = "INT";
static const char __pyx_k_REL[] = "REL";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_dct[] = "dct";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_dyn[] = "dyn";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_s_s[] = "%s.%s";
static const char __pyx_k_str[] = "__str__";
static const char __pyx_k_CASE[] = "CASE";
static const char __pyx_k_CAST[] = "CAST";
static const char __pyx_k_EVAR[] = "EVAR";
static const char __pyx_k_IntP[] = "IntP";
static const char __pyx_k_Node[] = "Node";
static const char __pyx_k_PROD[] = "PROD";
static const char __pyx_k_TERM[] = "TERM";
static const char __pyx_k__118[] = "?";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_FLOAT[] = "FLOAT";
static const char __pyx_k_Graph[] = "Graph";
static const char __pyx_k_KNOWN[] = "KNOWN";
static const char __pyx_k_capnp[] = "capnp";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_s_s_d[] = "<%s.%s: %d>";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_CO_FIX[] = "CO_FIX";
static const char __pyx_k_FloatP[] = "FloatP";
static const char __pyx_k_LAMBDA[] = "LAMBDA";
static const char __pyx_k_LET_IN[] = "LET_IN";
static const char __pyx_k_Tactic[] = "Tactic";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_pytact[] = "pytact";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_schema[] = "schema";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_APP_ARG[] = "APP_ARG";
static const char __pyx_k_APP_FUN[] = "APP_FUN";
static const char __pyx_k_Dataset[] = "Dataset";
static const char __pyx_k_FAILURE[] = "FAILURE";
static const char __pyx_k_FIX_FUN[] = "FIX_FUN";
static const char __pyx_k_IntEnum[] = "IntEnum";
static const char __pyx_k_Outcome[] = "Outcome";
static const char __pyx_k_PREDICT[] = "PREDICT";
static const char __pyx_k_Request[] = "Request";
static const char __pyx_k_UNKNOWN[] = "UNKNOWN";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_dynamic[] = "dynamic";
static const char __pyx_k_members[] = "__members__";
static const char __pyx_k_parents[] = "parents";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_Argument[] = "Argument";
static const char __pyx_k_CASE_IND[] = "CASE_IND";
static const char __pyx_k_COMPLETE[] = "COMPLETE";
static const char __pyx_k_C_B_TERM[] = "C_B_TERM";
static const char __pyx_k_EnumBase[] = "EnumBase";
static const char __pyx_k_IND_TYPE[] = "IND_TYPE";
static const char __pyx_k_ORIGINAL[] = "ORIGINAL";
static const char __pyx_k_Response[] = "Response";
static const char __pyx_k_SORT_SET[] = "SORT_SET";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_set_name[] = "__set_name__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_ALIGNMENT[] = "ALIGNMENT";
static const char __pyx_k_CASE_TERM[] = "CASE_TERM";
static const char __pyx_k_CAST_TERM[] = "CAST_TERM";
static const char __pyx_k_CAST_TYPE[] = "CAST_TYPE";
static const char __pyx_k_CONST_DEF[] = "CONST_DEF";
static const char __pyx_k_Exception[] = "Exception";
static const char __pyx_k_INDUCTIVE[] = "INDUCTIVE";
static const char __pyx_k_NEW_STATE[] = "NEW_STATE";
static const char __pyx_k_PRIMITIVE[] = "PRIMITIVE";
static const char __pyx_k_PROD_TERM[] = "PROD_TERM";
static const char __pyx_k_PROD_TYPE[] = "PROD_TYPE";
static const char __pyx_k_PROJ_TERM[] = "PROJ_TERM";
static const char __pyx_k_ProofStep[] = "ProofStep";
static const char __pyx_k_SORT_PROP[] = "SORT_PROP";
static const char __pyx_k_SORT_TYPE[] = "SORT_TYPE";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_as_struct[] = "as_struct";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_CONST_TYPE[] = "CONST_TYPE";
static const char __pyx_k_CO_FIX_FUN[] = "CO_FIX_FUN";
static const char __pyx_k_DEFINITION[] = "DEFINITION";
static const char __pyx_k_DISCHARGED[] = "DISCHARGED";
static const char __pyx_k_Definition[] = "Definition";
static const char __pyx_k_EVAR_SUBST[] = "EVAR_SUBST";
static const char __pyx_k_EdgeTarget[] = "EdgeTarget";
static const char __pyx_k_FIX_MUTUAL[] = "FIX_MUTUAL";
static const char __pyx_k_FIX_RETURN[] = "FIX_RETURN";
static const char __pyx_k_INITIALIZE[] = "INITIALIZE";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_Int64_List[] = "Int64_List";
static const char __pyx_k_LET_IN_DEF[] = "LET_IN_DEF";
static const char __pyx_k_PREDICTION[] = "PREDICTION";
static const char __pyx_k_PROJECTION[] = "PROJECTION";
static const char __pyx_k_Prediction[] = "Prediction";
static const char __pyx_k_ProofState[] = "ProofState";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_CASE_BRANCH[] = "CASE_BRANCH";
static const char __pyx_k_CASE_RETURN[] = "CASE_RETURN";
static const char __pyx_k_CONSTRUCTOR[] = "CONSTRUCTOR";
static const char __pyx_k_CONST_EMPTY[] = "CONST_EMPTY";
static const char __pyx_k_CONST_UNDEF[] = "CONST_UNDEF";
static const char __pyx_k_CONTEXT_DEF[] = "CONTEXT_DEF";
static const char __pyx_k_DataVersion[] = "DataVersion";
static const char __pyx_k_INITIALIZED[] = "INITIALIZED";
static const char __pyx_k_IntP_Reader[] = "IntP_Reader";
static const char __pyx_k_LAMBDA_TERM[] = "LAMBDA_TERM";
static const char __pyx_k_LAMBDA_TYPE[] = "LAMBDA_TYPE";
static const char __pyx_k_LET_IN_TERM[] = "LET_IN_TERM";
static const char __pyx_k_LET_IN_TYPE[] = "LET_IN_TYPE";
static const char __pyx_k_Node_Reader[] = "Node_Reader";
static const char __pyx_k_OrderedDict[] = "OrderedDict";
static const char __pyx_k_PARSE_ERROR[] = "PARSE_ERROR";
static const char __pyx_k_PROOF_STATE[] = "PROOF_STATE";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_REL_POINTER[] = "REL_POINTER";
static const char __pyx_k_SORT_S_PROP[] = "SORT_S_PROP";
static const char __pyx_k_SUBSTITUTED[] = "SUBSTITUTED";
static const char __pyx_k_SYNCHRONIZE[] = "SYNCHRONIZE";
static const char __pyx_k_String_List[] = "String_List";
static const char __pyx_k_Uint16_List[] = "Uint16_List";
static const char __pyx_k_Uint32_List[] = "Uint32_List";
static const char __pyx_k_Uint64_List[] = "Uint64_List";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_mro_entries[] = "__mro_entries__";
static const char __pyx_k_CONTEXT_ELEM[] = "CONTEXT_ELEM";
static const char __pyx_k_EVAR_SUBJECT[] = "EVAR_SUBJECT";
static const char __pyx_k_FIX_FUN_TERM[] = "FIX_FUN_TERM";
static const char __pyx_k_FIX_FUN_TYPE[] = "FIX_FUN_TYPE";
static const char __pyx_k_Graph_Reader[] = "Graph_Reader";
static const char __pyx_k_Pyx_EnumBase[] = "__Pyx_EnumBase";
static const char __pyx_k_SYNCHRONIZED[] = "SYNCHRONIZED";
static const char __pyx_k_UNRESOLVABLE[] = "UNRESOLVABLE";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_CONTEXT_ASSUM[] = "CONTEXT_ASSUM";
static const char __pyx_k_CO_FIX_MUTUAL[] = "CO_FIX_MUTUAL";
static const char __pyx_k_CO_FIX_RETURN[] = "CO_FIX_RETURN";
static const char __pyx_k_C_B_CONSTRUCT[] = "C_B_CONSTRUCT";
static const char __pyx_k_FloatP_Reader[] = "FloatP_Reader";
static const char __pyx_k_IND_CONSTRUCT[] = "IND_CONSTRUCT";
static const char __pyx_k_Out_of_bounds[] = "Out of bounds";
static const char __pyx_k_Tactic_Reader[] = "Tactic_Reader";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_init_subclass[] = "__init_subclass__";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_AbstractTactic[] = "AbstractTactic";
static const char __pyx_k_Argument_Which[] = "Argument_Which";
static const char __pyx_k_CONSTRUCT_TERM[] = "CONSTRUCT_TERM";
static const char __pyx_k_Dataset_Reader[] = "Dataset_Reader";
static const char __pyx_k_IND_PROJECTION[] = "IND_PROJECTION";
static const char __pyx_k_NO_SUCH_TACTIC[] = "NO_SUCH_TACTIC";
static const char __pyx_k_Outcome_Reader[] = "Outcome_Reader";
static const char __pyx_k_PROTOCOL_ERROR[] = "PROTOCOL_ERROR";
static const char __pyx_k_TextPrediction[] = "TextPrediction";
static const char __pyx_k_Argument_Reader[] = "Argument_Reader";
static const char __pyx_k_CHECK_ALIGNMENT[] = "CHECK_ALIGNMENT";
static const char __pyx_k_CONST_PRIMITIVE[] = "CONST_PRIMITIVE";
static const char __pyx_k_CONTEXT_SUBJECT[] = "CONTEXT_SUBJECT";
static const char __pyx_k_CO_FIX_FUN_TERM[] = "CO_FIX_FUN_TERM";
static const char __pyx_k_CO_FIX_FUN_TYPE[] = "CO_FIX_FUN_TYPE";
static const char __pyx_k_EVAR_SUBST_TERM[] = "EVAR_SUBST_TERM";
static const char __pyx_k_Exception_Which[] = "Exception_Which";
static const char __pyx_k_ExecutionResult[] = "ExecutionResult";
static const char __pyx_k_MANUAL_CONSTANT[] = "MANUAL_CONSTANT";
static const char __pyx_k_TEXT_PREDICTION[] = "TEXT_PREDICTION";
static const char __pyx_k_graph_api_capnp[] = "graph_api_capnp";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_CONST_OPAQUE_DEF[] = "CONST_OPAQUE_DEF";
static const char __pyx_k_CONTEXT_DEF_TERM[] = "CONTEXT_DEF_TERM";
static const char __pyx_k_CONTEXT_DEF_TYPE[] = "CONTEXT_DEF_TYPE";
static const char __pyx_k_ConflatableEdges[] = "ConflatableEdges";
static const char __pyx_k_Definition_Which[] = "Definition_Which";
static const char __pyx_k_Exception_Reader[] = "Exception_Reader";
static const char __pyx_k_ILLEGAL_ARGUMENT[] = "ILLEGAL_ARGUMENT";
static const char __pyx_k_Node_Reader_List[] = "Node_Reader_List";
static const char __pyx_k_ProofStep_Reader[] = "ProofStep_Reader";
static const char __pyx_k_Definition_Reader[] = "Definition_Reader";
static const char __pyx_k_EVAR_SUBST_TARGET[] = "EVAR_SUBST_TARGET";
static const char __pyx_k_Graph_Node_Reader[] = "Graph_Node_Reader";
static const char __pyx_k_ProofState_Reader[] = "ProofState_Reader";
static const char __pyx_k_TACTICAL_CONSTANT[] = "TACTICAL_CONSTANT";
static const char __pyx_k_DataVersion_Reader[] = "DataVersion_Reader";
static const char __pyx_k_EVAR_SUBST_POINTER[] = "EVAR_SUBST_POINTER";
static const char __pyx_k_EdgeClassification[] = "EdgeClassification";
static const char __pyx_k_PredictionProtocol[] = "PredictionProtocol";
static const char __pyx_k_Pyx_EnumBase___new[] = "__Pyx_EnumBase.__new__";
static const char __pyx_k_Pyx_EnumBase___str[] = "__Pyx_EnumBase.__str__";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_CASE_BRANCH_POINTER[] = "CASE_BRANCH_POINTER";
static const char __pyx_k_Outcome_Reader_List[] = "Outcome_Reader_List";
static const char __pyx_k_Outcome_Term_Reader[] = "Outcome_Term_Reader";
static const char __pyx_k_Pyx_EnumBase___repr[] = "__Pyx_EnumBase.__repr__";
static const char __pyx_k_Argument_Reader_List[] = "Argument_Reader_List";
static const char __pyx_k_Argument_Term_Reader[] = "Argument_Term_Reader";
static const char __pyx_k_MISMATCHED_ARGUMENTS[] = "MISMATCHED_ARGUMENTS";
static const char __pyx_k_Unknown_enum_value_s[] = "Unknown enum value: '%s'";
static const char __pyx_k_AbstractTactic_Reader[] = "AbstractTactic_Reader";
static const char __pyx_k_ExecutionResult_Which[] = "ExecutionResult_Which";
static const char __pyx_k_ProofStep_Reader_List[] = "ProofStep_Reader_List";
static const char __pyx_k_ExecutionResult_Reader[] = "ExecutionResult_Reader";
static const char __pyx_k_Graph_Node_Label_Which[] = "Graph_Node_Label_Which";
static const char __pyx_k_Graph_Node_Reader_List[] = "Graph_Node_Reader_List";
static const char __pyx_k_ProofState_Reader_List[] = "ProofState_Reader_List";
static const char __pyx_k_ProofState_Root_Reader[] = "ProofState_Root_Reader";
static const char __pyx_k_ProofStep_Tactic_Which[] = "ProofStep_Tactic_Which";
static const char __pyx_k_pytact_graph_api_capnp[] = "pytact.graph_api_capnp";
static const char __pyx_k_ConflatableEdges_Reader[] = "ConflatableEdges_Reader";
static const char __pyx_k_Definition_Status_Which[] = "Definition_Status_Which";
static const char __pyx_k_Graph_EdgeTarget_Reader[] = "Graph_EdgeTarget_Reader";
static const char __pyx_k_Graph_Node_Label_Reader[] = "Graph_Node_Label_Reader";
static const char __pyx_k_MANUAL_SECTION_CONSTANT[] = "MANUAL_SECTION_CONSTANT";
static const char __pyx_k_ProofStep_Tactic_Reader[] = "ProofStep_Tactic_Reader";
static const char __pyx_k_Definition_Status_Reader[] = "Definition_Status_Reader";
static const char __pyx_k_PredictionProtocol_Reader[] = "PredictionProtocol_Reader";
static const char __pyx_k_TACTICAL_SECTION_CONSTANT[] = "TACTICAL_SECTION_CONSTANT";
static const char __pyx_k_AbstractTactic_Reader_List[] = "AbstractTactic_Reader_List";
static const char __pyx_k_Int64_List___reduce_cython[] = "Int64_List.__reduce_cython__";
static const char __pyx_k_IntP_Reader___reduce_cython[] = "IntP_Reader.__reduce_cython__";
static const char __pyx_k_Node_Reader___reduce_cython[] = "Node_Reader.__reduce_cython__";
static const char __pyx_k_String_List___reduce_cython[] = "String_List.__reduce_cython__";
static const char __pyx_k_Uint16_List___reduce_cython[] = "Uint16_List.__reduce_cython__";
static const char __pyx_k_Uint32_List___reduce_cython[] = "Uint32_List.__reduce_cython__";
static const char __pyx_k_Uint64_List___reduce_cython[] = "Uint64_List.__reduce_cython__";
static const char __pyx_k_pyx_unpickle___Pyx_EnumMeta[] = "__pyx_unpickle___Pyx_EnumMeta";
static const char __pyx_k_Graph_EdgeTarget_Reader_List[] = "Graph_EdgeTarget_Reader_List";
static const char __pyx_k_Graph_Reader___reduce_cython[] = "Graph_Reader.__reduce_cython__";
static const char __pyx_k_Int64_List___setstate_cython[] = "Int64_List.__setstate_cython__";
static const char __pyx_k_Pyx_EnumMeta___reduce_cython[] = "__Pyx_EnumMeta.__reduce_cython__";
static const char __pyx_k_FloatP_Reader___reduce_cython[] = "FloatP_Reader.__reduce_cython__";
static const char __pyx_k_IntP_Reader___setstate_cython[] = "IntP_Reader.__setstate_cython__";
static const char __pyx_k_Node_Reader___setstate_cython[] = "Node_Reader.__setstate_cython__";
static const char __pyx_k_String_List___setstate_cython[] = "String_List.__setstate_cython__";
static const char __pyx_k_Tactic_Reader___reduce_cython[] = "Tactic_Reader.__reduce_cython__";
static const char __pyx_k_Uint16_List___setstate_cython[] = "Uint16_List.__setstate_cython__";
static const char __pyx_k_Uint32_List___setstate_cython[] = "Uint32_List.__setstate_cython__";
static const char __pyx_k_Uint64_List___setstate_cython[] = "Uint64_List.__setstate_cython__";
static const char __pyx_k_pytact_graph_api_capnp_cython[] = "pytact.graph_api_capnp_cython";
static const char __pyx_k_Dataset_Reader___reduce_cython[] = "Dataset_Reader.__reduce_cython__";
static const char __pyx_k_Graph_EdgeTarget_Target_Reader[] = "Graph_EdgeTarget_Target_Reader";
static const char __pyx_k_Graph_Reader___setstate_cython[] = "Graph_Reader.__setstate_cython__";
static const char __pyx_k_Outcome_Reader___reduce_cython[] = "Outcome_Reader.__reduce_cython__";
static const char __pyx_k_Pyx_EnumMeta___setstate_cython[] = "__Pyx_EnumMeta.__setstate_cython__";
static const char __pyx_k_Argument_Reader_List___setstate[] = "Argument_Reader_List.__setstate_cython__";
static const char __pyx_k_Argument_Reader___reduce_cython[] = "Argument_Reader.__reduce_cython__";
static const char __pyx_k_Argument_Term_Reader___setstate[] = "Argument_Term_Reader.__setstate_cython__";
static const char __pyx_k_ExecutionResult_NewState_Reader[] = "ExecutionResult_NewState_Reader";
static const char __pyx_k_ExecutionResult_Reader___reduce[] = "ExecutionResult_Reader.__reduce_cython__";
static const char __pyx_k_FloatP_Reader___setstate_cython[] = "FloatP_Reader.__setstate_cython__";
static const char __pyx_k_Graph_Node_Reader_List___reduce[] = "Graph_Node_Reader_List.__reduce_cython__";
static const char __pyx_k_ProofState_Reader_List___reduce[] = "ProofState_Reader_List.__reduce_cython__";
static const char __pyx_k_ProofState_Root_Reader___reduce[] = "ProofState_Root_Reader.__reduce_cython__";
static const char __pyx_k_Tactic_Reader___setstate_cython[] = "Tactic_Reader.__setstate_cython__";
static const char __pyx_k_self_source_cannot_be_converted[] = "self.source cannot be converted to a Python object for pickling";
static const char __pyx_k_AbstractTactic_Reader_List___red[] = "AbstractTactic_Reader_List.__reduce_cython__";
static const char __pyx_k_AbstractTactic_Reader_List___set[] = "AbstractTactic_Reader_List.__setstate_cython__";
static const char __pyx_k_AbstractTactic_Reader___reduce_c[] = "AbstractTactic_Reader.__reduce_cython__";
static const char __pyx_k_AbstractTactic_Reader___setstate[] = "AbstractTactic_Reader.__setstate_cython__";
static const char __pyx_k_Argument_Reader_List___reduce_cy[] = "Argument_Reader_List.__reduce_cython__";
static const char __pyx_k_Argument_Reader___setstate_cytho[] = "Argument_Reader.__setstate_cython__";
static const char __pyx_k_Argument_Term_Reader___reduce_cy[] = "Argument_Term_Reader.__reduce_cython__";
static const char __pyx_k_ConflatableEdges_Reader___reduce[] = "ConflatableEdges_Reader.__reduce_cython__";
static const char __pyx_k_ConflatableEdges_Reader___setsta[] = "ConflatableEdges_Reader.__setstate_cython__";
static const char __pyx_k_DataVersion_Reader___reduce_cyth[] = "DataVersion_Reader.__reduce_cython__";
static const char __pyx_k_DataVersion_Reader___setstate_cy[] = "DataVersion_Reader.__setstate_cython__";
static const char __pyx_k_Dataset_Reader___setstate_cython[] = "Dataset_Reader.__setstate_cython__";
static const char __pyx_k_Definition_Reader___reduce_cytho[] = "Definition_Reader.__reduce_cython__";
static const char __pyx_k_Definition_Reader___setstate_cyt[] = "Definition_Reader.__setstate_cython__";
static const char __pyx_k_Definition_Status_Reader___reduc[] = "Definition_Status_Reader.__reduce_cython__";
static const char __pyx_k_Definition_Status_Reader___setst[] = "Definition_Status_Reader.__setstate_cython__";
static const char __pyx_k_Definition_Status_Substituted_Re[] = "Definition_Status_Substituted_Reader";
static const char __pyx_k_Exception_Reader___reduce_cython[] = "Exception_Reader.__reduce_cython__";
static const char __pyx_k_Exception_Reader___setstate_cyth[] = "Exception_Reader.__setstate_cython__";
static const char __pyx_k_ExecutionResult_Reader___setstat[] = "ExecutionResult_Reader.__setstate_cython__";
static const char __pyx_k_Graph_EdgeTarget_Reader_List___r[] = "Graph_EdgeTarget_Reader_List.__reduce_cython__";
static const char __pyx_k_Graph_EdgeTarget_Reader_List___s[] = "Graph_EdgeTarget_Reader_List.__setstate_cython__";
static const char __pyx_k_Graph_EdgeTarget_Reader___reduce[] = "Graph_EdgeTarget_Reader.__reduce_cython__";
static const char __pyx_k_Graph_EdgeTarget_Reader___setsta[] = "Graph_EdgeTarget_Reader.__setstate_cython__";
static const char __pyx_k_Graph_EdgeTarget_Target_Reader_2[] = "Graph_EdgeTarget_Target_Reader.__reduce_cython__";
static const char __pyx_k_Graph_EdgeTarget_Target_Reader_3[] = "Graph_EdgeTarget_Target_Reader.__setstate_cython__";
static const char __pyx_k_Graph_Node_Label_Reader___reduce[] = "Graph_Node_Label_Reader.__reduce_cython__";
static const char __pyx_k_Graph_Node_Label_Reader___setsta[] = "Graph_Node_Label_Reader.__setstate_cython__";
static const char __pyx_k_Graph_Node_Reader_List___setstat[] = "Graph_Node_Reader_List.__setstate_cython__";
static const char __pyx_k_Graph_Node_Reader___reduce_cytho[] = "Graph_Node_Reader.__reduce_cython__";
static const char __pyx_k_Graph_Node_Reader___setstate_cyt[] = "Graph_Node_Reader.__setstate_cython__";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())";
static const char __pyx_k_Node_Reader_List___reduce_cython[] = "Node_Reader_List.__reduce_cython__";
static const char __pyx_k_Node_Reader_List___setstate_cyth[] = "Node_Reader_List.__setstate_cython__";
static const char __pyx_k_Outcome_Reader_List___reduce_cyt[] = "Outcome_Reader_List.__reduce_cython__";
static const char __pyx_k_Outcome_Reader_List___setstate_c[] = "Outcome_Reader_List.__setstate_cython__";
static const char __pyx_k_Outcome_Reader___setstate_cython[] = "Outcome_Reader.__setstate_cython__";
static const char __pyx_k_Outcome_Term_Reader___reduce_cyt[] = "Outcome_Term_Reader.__reduce_cython__";
static const char __pyx_k_Outcome_Term_Reader___setstate_c[] = "Outcome_Term_Reader.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_Prediction_Re[] = "PredictionProtocol_Prediction_Reader";
static const char __pyx_k_PredictionProtocol_Reader___redu[] = "PredictionProtocol_Reader.__reduce_cython__";
static const char __pyx_k_PredictionProtocol_Reader___sets[] = "PredictionProtocol_Reader.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_Request_Initi[] = "PredictionProtocol_Request_Initialize_Reader";
static const char __pyx_k_PredictionProtocol_Request_Predi[] = "PredictionProtocol_Request_Predict_Reader";
static const char __pyx_k_PredictionProtocol_Request_Reade[] = "PredictionProtocol_Request_Reader";
static const char __pyx_k_PredictionProtocol_Request_Which[] = "PredictionProtocol_Request_Which";
static const char __pyx_k_PredictionProtocol_Response_Alig[] = "PredictionProtocol_Response_Alignment_Reader";
static const char __pyx_k_PredictionProtocol_Response_Read[] = "PredictionProtocol_Response_Reader";
static const char __pyx_k_PredictionProtocol_Response_Whic[] = "PredictionProtocol_Response_Which";
static const char __pyx_k_PredictionProtocol_TextPredictio[] = "PredictionProtocol_TextPrediction_Reader";
static const char __pyx_k_ProofState_Reader_List___setstat[] = "ProofState_Reader_List.__setstate_cython__";
static const char __pyx_k_ProofState_Reader___reduce_cytho[] = "ProofState_Reader.__reduce_cython__";
static const char __pyx_k_ProofState_Reader___setstate_cyt[] = "ProofState_Reader.__setstate_cython__";
static const char __pyx_k_ProofState_Root_Reader___setstat[] = "ProofState_Root_Reader.__setstate_cython__";
static const char __pyx_k_ProofStep_Reader_List___reduce_c[] = "ProofStep_Reader_List.__reduce_cython__";
static const char __pyx_k_ProofStep_Reader_List___setstate[] = "ProofStep_Reader_List.__setstate_cython__";
static const char __pyx_k_ProofStep_Reader___reduce_cython[] = "ProofStep_Reader.__reduce_cython__";
static const char __pyx_k_ProofStep_Reader___setstate_cyth[] = "ProofStep_Reader.__setstate_cython__";
static const char __pyx_k_ProofStep_Tactic_Reader___reduce[] = "ProofStep_Tactic_Reader.__reduce_cython__";
static const char __pyx_k_ProofStep_Tactic_Reader___setsta[] = "ProofStep_Tactic_Reader.__setstate_cython__";
static const char __pyx_k_ExecutionResult_NewState_Reader_2[] = "ExecutionResult_NewState_Reader.__reduce_cython__";
static const char __pyx_k_ExecutionResult_NewState_Reader_3[] = "ExecutionResult_NewState_Reader.__setstate_cython__";
static const char __pyx_k_Definition_Status_Substituted_Re_2[] = "Definition_Status_Substituted_Reader.__reduce_cython__";
static const char __pyx_k_Definition_Status_Substituted_Re_3[] = "Definition_Status_Substituted_Reader.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_Prediction_Re_2[] = "PredictionProtocol_Prediction_Reader_List";
static const char __pyx_k_PredictionProtocol_Prediction_Re_3[] = "PredictionProtocol_Prediction_Reader.__reduce_cython__";
static const char __pyx_k_PredictionProtocol_Prediction_Re_4[] = "PredictionProtocol_Prediction_Reader.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_Prediction_Re_5[] = "PredictionProtocol_Prediction_Reader_List.__reduce_cython__";
static const char __pyx_k_PredictionProtocol_Prediction_Re_6[] = "PredictionProtocol_Prediction_Reader_List.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_Request_Initi_2[] = "PredictionProtocol_Request_Initialize_Reader.__reduce_cython__";
static const char __pyx_k_PredictionProtocol_Request_Initi_3[] = "PredictionProtocol_Request_Initialize_Reader.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_Request_Predi_2[] = "PredictionProtocol_Request_Predict_Reader.__reduce_cython__";
static const char __pyx_k_PredictionProtocol_Request_Predi_3[] = "PredictionProtocol_Request_Predict_Reader.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_Request_Reade_2[] = "PredictionProtocol_Request_Reader.__reduce_cython__";
static const char __pyx_k_PredictionProtocol_Request_Reade_3[] = "PredictionProtocol_Request_Reader.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_Response_Alig_2[] = "PredictionProtocol_Response_Alignment_Reader.__reduce_cython__";
static const char __pyx_k_PredictionProtocol_Response_Alig_3[] = "PredictionProtocol_Response_Alignment_Reader.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_Response_Read_2[] = "PredictionProtocol_Response_Reader.__reduce_cython__";
static const char __pyx_k_PredictionProtocol_Response_Read_3[] = "PredictionProtocol_Response_Reader.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_TextPredictio_2[] = "PredictionProtocol_TextPrediction_Reader_List";
static const char __pyx_k_PredictionProtocol_TextPredictio_3[] = "PredictionProtocol_TextPrediction_Reader.__reduce_cython__";
static const char __pyx_k_PredictionProtocol_TextPredictio_4[] = "PredictionProtocol_TextPrediction_Reader.__setstate_cython__";
static const char __pyx_k_PredictionProtocol_TextPredictio_5[] = "PredictionProtocol_TextPrediction_Reader_List.__reduce_cython__";
static const char __pyx_k_PredictionProtocol_TextPredictio_6[] = "PredictionProtocol_TextPrediction_Reader_List.__setstate_cython__";
#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_n_s_ALIGNMENT;
static PyObject *__pyx_n_s_APP;
static PyObject *__pyx_n_s_APP_ARG;
static PyObject *__pyx_n_s_APP_FUN;
static PyObject *__pyx_n_s_AbstractTactic;
static PyObject *__pyx_n_s_AbstractTactic_Reader;
static PyObject *__pyx_n_s_AbstractTactic_Reader_List;
static PyObject *__pyx_n_s_AbstractTactic_Reader_List___red;
static PyObject *__pyx_n_s_AbstractTactic_Reader_List___set;
static PyObject *__pyx_n_s_AbstractTactic_Reader___reduce_c;
static PyObject *__pyx_n_s_AbstractTactic_Reader___setstate;
static PyObject *__pyx_n_s_Argument;
static PyObject *__pyx_n_s_Argument_Reader;
static PyObject *__pyx_n_s_Argument_Reader_List;
static PyObject *__pyx_n_s_Argument_Reader_List___reduce_cy;
static PyObject *__pyx_n_s_Argument_Reader_List___setstate;
static PyObject *__pyx_n_s_Argument_Reader___reduce_cython;
static PyObject *__pyx_n_s_Argument_Reader___setstate_cytho;
static PyObject *__pyx_n_s_Argument_Term_Reader;
static PyObject *__pyx_n_s_Argument_Term_Reader___reduce_cy;
static PyObject *__pyx_n_s_Argument_Term_Reader___setstate;
static PyObject *__pyx_n_s_Argument_Which;
static PyObject *__pyx_n_s_CASE;
static PyObject *__pyx_n_s_CASE_BRANCH;
static PyObject *__pyx_n_s_CASE_BRANCH_POINTER;
static PyObject *__pyx_n_s_CASE_IND;
static PyObject *__pyx_n_s_CASE_RETURN;
static PyObject *__pyx_n_s_CASE_TERM;
static PyObject *__pyx_n_s_CAST;
static PyObject *__pyx_n_s_CAST_TERM;
static PyObject *__pyx_n_s_CAST_TYPE;
static PyObject *__pyx_n_s_CHECK_ALIGNMENT;
static PyObject *__pyx_n_s_COMPLETE;
static PyObject *__pyx_n_s_CONSTRUCTOR;
static PyObject *__pyx_n_s_CONSTRUCT_TERM;
static PyObject *__pyx_n_s_CONST_DEF;
static PyObject *__pyx_n_s_CONST_EMPTY;
static PyObject *__pyx_n_s_CONST_OPAQUE_DEF;
static PyObject *__pyx_n_s_CONST_PRIMITIVE;
static PyObject *__pyx_n_s_CONST_TYPE;
static PyObject *__pyx_n_s_CONST_UNDEF;
static PyObject *__pyx_n_s_CONTEXT_ASSUM;
static PyObject *__pyx_n_s_CONTEXT_DEF;
static PyObject *__pyx_n_s_CONTEXT_DEF_TERM;
static PyObject *__pyx_n_s_CONTEXT_DEF_TYPE;
static PyObject *__pyx_n_s_CONTEXT_ELEM;
static PyObject *__pyx_n_s_CONTEXT_SUBJECT;
static PyObject *__pyx_n_s_CO_FIX;
static PyObject *__pyx_n_s_CO_FIX_FUN;
static PyObject *__pyx_n_s_CO_FIX_FUN_TERM;
static PyObject *__pyx_n_s_CO_FIX_FUN_TYPE;
static PyObject *__pyx_n_s_CO_FIX_MUTUAL;
static PyObject *__pyx_n_s_CO_FIX_RETURN;
static PyObject *__pyx_n_s_C_B_CONSTRUCT;
static PyObject *__pyx_n_s_C_B_TERM;
static PyObject *__pyx_n_s_ConflatableEdges;
static PyObject *__pyx_n_s_ConflatableEdges_Reader;
static PyObject *__pyx_n_s_ConflatableEdges_Reader___reduce;
static PyObject *__pyx_n_s_ConflatableEdges_Reader___setsta;
static PyObject *__pyx_n_s_DEFINITION;
static PyObject *__pyx_n_s_DISCHARGED;
static PyObject *__pyx_n_s_DataVersion;
static PyObject *__pyx_n_s_DataVersion_Reader;
static PyObject *__pyx_n_s_DataVersion_Reader___reduce_cyth;
static PyObject *__pyx_n_s_DataVersion_Reader___setstate_cy;
static PyObject *__pyx_n_s_Dataset;
static PyObject *__pyx_n_s_Dataset_Reader;
static PyObject *__pyx_n_s_Dataset_Reader___reduce_cython;
static PyObject *__pyx_n_s_Dataset_Reader___setstate_cython;
static PyObject *__pyx_n_s_Definition;
static PyObject *__pyx_n_s_Definition_Reader;
static PyObject *__pyx_n_s_Definition_Reader___reduce_cytho;
static PyObject *__pyx_n_s_Definition_Reader___setstate_cyt;
static PyObject *__pyx_n_s_Definition_Status_Reader;
static PyObject *__pyx_n_s_Definition_Status_Reader___reduc;
static PyObject *__pyx_n_s_Definition_Status_Reader___setst;
static PyObject *__pyx_n_s_Definition_Status_Substituted_Re;
static PyObject *__pyx_n_s_Definition_Status_Substituted_Re_2;
static PyObject *__pyx_n_s_Definition_Status_Substituted_Re_3;
static PyObject *__pyx_n_s_Definition_Status_Which;
static PyObject *__pyx_n_s_Definition_Which;
static PyObject *__pyx_n_s_EVAR;
static PyObject *__pyx_n_s_EVAR_SUBJECT;
static PyObject *__pyx_n_s_EVAR_SUBST;
static PyObject *__pyx_n_s_EVAR_SUBST_POINTER;
static PyObject *__pyx_n_s_EVAR_SUBST_TARGET;
static PyObject *__pyx_n_s_EVAR_SUBST_TERM;
static PyObject *__pyx_n_s_EdgeClassification;
static PyObject *__pyx_n_s_EdgeTarget;
static PyObject *__pyx_n_s_EnumBase;
static PyObject *__pyx_n_s_Exception;
static PyObject *__pyx_n_s_Exception_Reader;
static PyObject *__pyx_n_s_Exception_Reader___reduce_cython;
static PyObject *__pyx_n_s_Exception_Reader___setstate_cyth;
static PyObject *__pyx_n_s_Exception_Which;
static PyObject *__pyx_n_s_ExecutionResult;
static PyObject *__pyx_n_s_ExecutionResult_NewState_Reader;
static PyObject *__pyx_n_s_ExecutionResult_NewState_Reader_2;
static PyObject *__pyx_n_s_ExecutionResult_NewState_Reader_3;
static PyObject *__pyx_n_s_ExecutionResult_Reader;
static PyObject *__pyx_n_s_ExecutionResult_Reader___reduce;
static PyObject *__pyx_n_s_ExecutionResult_Reader___setstat;
static PyObject *__pyx_n_s_ExecutionResult_Which;
static PyObject *__pyx_n_s_FAILURE;
static PyObject *__pyx_n_s_FIX;
static PyObject *__pyx_n_s_FIX_FUN;
static PyObject *__pyx_n_s_FIX_FUN_TERM;
static PyObject *__pyx_n_s_FIX_FUN_TYPE;
static PyObject *__pyx_n_s_FIX_MUTUAL;
static PyObject *__pyx_n_s_FIX_RETURN;
static PyObject *__pyx_n_s_FLOAT;
static PyObject *__pyx_n_s_FloatP;
static PyObject *__pyx_n_s_FloatP_Reader;
static PyObject *__pyx_n_s_FloatP_Reader___reduce_cython;
static PyObject *__pyx_n_s_FloatP_Reader___setstate_cython;
static PyObject *__pyx_n_s_Graph;
static PyObject *__pyx_n_s_Graph_EdgeTarget_Reader;
static PyObject *__pyx_n_s_Graph_EdgeTarget_Reader_List;
static PyObject *__pyx_n_s_Graph_EdgeTarget_Reader_List___r;
static PyObject *__pyx_n_s_Graph_EdgeTarget_Reader_List___s;
static PyObject *__pyx_n_s_Graph_EdgeTarget_Reader___reduce;
static PyObject *__pyx_n_s_Graph_EdgeTarget_Reader___setsta;
static PyObject *__pyx_n_s_Graph_EdgeTarget_Target_Reader;
static PyObject *__pyx_n_s_Graph_EdgeTarget_Target_Reader_2;
static PyObject *__pyx_n_s_Graph_EdgeTarget_Target_Reader_3;
static PyObject *__pyx_n_s_Graph_Node_Label_Reader;
static PyObject *__pyx_n_s_Graph_Node_Label_Reader___reduce;
static PyObject *__pyx_n_s_Graph_Node_Label_Reader___setsta;
static PyObject *__pyx_n_s_Graph_Node_Label_Which;
static PyObject *__pyx_n_s_Graph_Node_Reader;
static PyObject *__pyx_n_s_Graph_Node_Reader_List;
static PyObject *__pyx_n_s_Graph_Node_Reader_List___reduce;
static PyObject *__pyx_n_s_Graph_Node_Reader_List___setstat;
static PyObject *__pyx_n_s_Graph_Node_Reader___reduce_cytho;
static PyObject *__pyx_n_s_Graph_Node_Reader___setstate_cyt;
static PyObject *__pyx_n_s_Graph_Reader;
static PyObject *__pyx_n_s_Graph_Reader___reduce_cython;
static PyObject *__pyx_n_s_Graph_Reader___setstate_cython;
static PyObject *__pyx_n_s_ILLEGAL_ARGUMENT;
static PyObject *__pyx_n_s_INDUCTIVE;
static PyObject *__pyx_n_s_IND_CONSTRUCT;
static PyObject *__pyx_n_s_IND_PROJECTION;
static PyObject *__pyx_n_s_IND_TYPE;
static PyObject *__pyx_n_s_INITIALIZE;
static PyObject *__pyx_n_s_INITIALIZED;
static PyObject *__pyx_n_s_INT;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_Int64_List;
static PyObject *__pyx_n_s_Int64_List___reduce_cython;
static PyObject *__pyx_n_s_Int64_List___setstate_cython;
static PyObject *__pyx_n_s_IntEnum;
static PyObject *__pyx_n_s_IntP;
static PyObject *__pyx_n_s_IntP_Reader;
static PyObject *__pyx_n_s_IntP_Reader___reduce_cython;
static PyObject *__pyx_n_s_IntP_Reader___setstate_cython;
static PyObject *__pyx_n_s_KNOWN;
static PyObject *__pyx_n_s_LAMBDA;
static PyObject *__pyx_n_s_LAMBDA_TERM;
static PyObject *__pyx_n_s_LAMBDA_TYPE;
static PyObject *__pyx_n_s_LET_IN;
static PyObject *__pyx_n_s_LET_IN_DEF;
static PyObject *__pyx_n_s_LET_IN_TERM;
static PyObject *__pyx_n_s_LET_IN_TYPE;
static PyObject *__pyx_n_s_MANUAL_CONSTANT;
static PyObject *__pyx_n_s_MANUAL_SECTION_CONSTANT;
static PyObject *__pyx_n_s_MISMATCHED_ARGUMENTS;
static PyObject *__pyx_n_s_NEW_STATE;
static PyObject *__pyx_n_s_NO_SUCH_TACTIC;
static PyObject *__pyx_n_s_Node;
static PyObject *__pyx_n_s_Node_Reader;
static PyObject *__pyx_n_s_Node_Reader_List;
static PyObject *__pyx_n_s_Node_Reader_List___reduce_cython;
static PyObject *__pyx_n_s_Node_Reader_List___setstate_cyth;
static PyObject *__pyx_n_s_Node_Reader___reduce_cython;
static PyObject *__pyx_n_s_Node_Reader___setstate_cython;
static PyObject *__pyx_n_s_ORIGINAL;
static PyObject *__pyx_n_s_OrderedDict;
static PyObject *__pyx_kp_u_Out_of_bounds;
static PyObject *__pyx_n_s_Outcome;
static PyObject *__pyx_n_s_Outcome_Reader;
static PyObject *__pyx_n_s_Outcome_Reader_List;
static PyObject *__pyx_n_s_Outcome_Reader_List___reduce_cyt;
static PyObject *__pyx_n_s_Outcome_Reader_List___setstate_c;
static PyObject *__pyx_n_s_Outcome_Reader___reduce_cython;
static PyObject *__pyx_n_s_Outcome_Reader___setstate_cython;
static PyObject *__pyx_n_s_Outcome_Term_Reader;
static PyObject *__pyx_n_s_Outcome_Term_Reader___reduce_cyt;
static PyObject *__pyx_n_s_Outcome_Term_Reader___setstate_c;
static PyObject *__pyx_n_s_PARSE_ERROR;
static PyObject *__pyx_n_s_PREDICT;
static PyObject *__pyx_n_s_PREDICTION;
static PyObject *__pyx_n_s_PRIMITIVE;
static PyObject *__pyx_n_s_PROD;
static PyObject *__pyx_n_s_PROD_TERM;
static PyObject *__pyx_n_s_PROD_TYPE;
static PyObject *__pyx_n_s_PROJECTION;
static PyObject *__pyx_n_s_PROJ_TERM;
static PyObject *__pyx_n_s_PROOF_STATE;
static PyObject *__pyx_n_s_PROTOCOL_ERROR;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_Prediction;
static PyObject *__pyx_n_s_PredictionProtocol;
static PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re;
static PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re_2;
static PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re_3;
static PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re_4;
static PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re_5;
static PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re_6;
static PyObject *__pyx_n_s_PredictionProtocol_Reader;
static PyObject *__pyx_n_s_PredictionProtocol_Reader___redu;
static PyObject *__pyx_n_s_PredictionProtocol_Reader___sets;
static PyObject *__pyx_n_s_PredictionProtocol_Request_Initi;
static PyObject *__pyx_n_s_PredictionProtocol_Request_Initi_2;
static PyObject *__pyx_n_s_PredictionProtocol_Request_Initi_3;
static PyObject *__pyx_n_s_PredictionProtocol_Request_Predi;
static PyObject *__pyx_n_s_PredictionProtocol_Request_Predi_2;
static PyObject *__pyx_n_s_PredictionProtocol_Request_Predi_3;
static PyObject *__pyx_n_s_PredictionProtocol_Request_Reade;
static PyObject *__pyx_n_s_PredictionProtocol_Request_Reade_2;
static PyObject *__pyx_n_s_PredictionProtocol_Request_Reade_3;
static PyObject *__pyx_n_s_PredictionProtocol_Request_Which;
static PyObject *__pyx_n_s_PredictionProtocol_Response_Alig;
static PyObject *__pyx_n_s_PredictionProtocol_Response_Alig_2;
static PyObject *__pyx_n_s_PredictionProtocol_Response_Alig_3;
static PyObject *__pyx_n_s_PredictionProtocol_Response_Read;
static PyObject *__pyx_n_s_PredictionProtocol_Response_Read_2;
static PyObject *__pyx_n_s_PredictionProtocol_Response_Read_3;
static PyObject *__pyx_n_s_PredictionProtocol_Response_Whic;
static PyObject *__pyx_n_s_PredictionProtocol_TextPredictio;
static PyObject *__pyx_n_s_PredictionProtocol_TextPredictio_2;
static PyObject *__pyx_n_s_PredictionProtocol_TextPredictio_3;
static PyObject *__pyx_n_s_PredictionProtocol_TextPredictio_4;
static PyObject *__pyx_n_s_PredictionProtocol_TextPredictio_5;
static PyObject *__pyx_n_s_PredictionProtocol_TextPredictio_6;
static PyObject *__pyx_n_s_ProofState;
static PyObject *__pyx_n_s_ProofState_Reader;
static PyObject *__pyx_n_s_ProofState_Reader_List;
static PyObject *__pyx_n_s_ProofState_Reader_List___reduce;
static PyObject *__pyx_n_s_ProofState_Reader_List___setstat;
static PyObject *__pyx_n_s_ProofState_Reader___reduce_cytho;
static PyObject *__pyx_n_s_ProofState_Reader___setstate_cyt;
static PyObject *__pyx_n_s_ProofState_Root_Reader;
static PyObject *__pyx_n_s_ProofState_Root_Reader___reduce;
static PyObject *__pyx_n_s_ProofState_Root_Reader___setstat;
static PyObject *__pyx_n_s_ProofStep;
static PyObject *__pyx_n_s_ProofStep_Reader;
static PyObject *__pyx_n_s_ProofStep_Reader_List;
static PyObject *__pyx_n_s_ProofStep_Reader_List___reduce_c;
static PyObject *__pyx_n_s_ProofStep_Reader_List___setstate;
static PyObject *__pyx_n_s_ProofStep_Reader___reduce_cython;
static PyObject *__pyx_n_s_ProofStep_Reader___setstate_cyth;
static PyObject *__pyx_n_s_ProofStep_Tactic_Reader;
static PyObject *__pyx_n_s_ProofStep_Tactic_Reader___reduce;
static PyObject *__pyx_n_s_ProofStep_Tactic_Reader___setsta;
static PyObject *__pyx_n_s_ProofStep_Tactic_Which;
static PyObject *__pyx_n_s_Pyx_EnumBase;
static PyObject *__pyx_n_s_Pyx_EnumBase___new;
static PyObject *__pyx_n_s_Pyx_EnumBase___repr;
static PyObject *__pyx_n_s_Pyx_EnumBase___str;
static PyObject *__pyx_n_s_Pyx_EnumMeta___reduce_cython;
static PyObject *__pyx_n_s_Pyx_EnumMeta___setstate_cython;
static PyObject *__pyx_n_s_REL;
static PyObject *__pyx_n_s_REL_POINTER;
static PyObject *__pyx_n_s_Request;
static PyObject *__pyx_n_s_Response;
static PyObject *__pyx_n_s_SORT_PROP;
static PyObject *__pyx_n_s_SORT_SET;
static PyObject *__pyx_n_s_SORT_S_PROP;
static PyObject *__pyx_n_s_SORT_TYPE;
static PyObject *__pyx_n_s_SUBSTITUTED;
static PyObject *__pyx_n_s_SYNCHRONIZE;
static PyObject *__pyx_n_s_SYNCHRONIZED;
static PyObject *__pyx_n_s_String_List;
static PyObject *__pyx_n_s_String_List___reduce_cython;
static PyObject *__pyx_n_s_String_List___setstate_cython;
static PyObject *__pyx_n_s_TACTICAL_CONSTANT;
static PyObject *__pyx_n_s_TACTICAL_SECTION_CONSTANT;
static PyObject *__pyx_n_s_TERM;
static PyObject *__pyx_n_s_TEXT_PREDICTION;
static PyObject *__pyx_n_s_Tactic;
static PyObject *__pyx_n_s_Tactic_Reader;
static PyObject *__pyx_n_s_Tactic_Reader___reduce_cython;
static PyObject *__pyx_n_s_Tactic_Reader___setstate_cython;
static PyObject *__pyx_n_s_TextPrediction;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UNKNOWN;
static PyObject *__pyx_n_s_UNRESOLVABLE;
static PyObject *__pyx_n_s_Uint16_List;
static PyObject *__pyx_n_s_Uint16_List___reduce_cython;
static PyObject *__pyx_n_s_Uint16_List___setstate_cython;
static PyObject *__pyx_n_s_Uint32_List;
static PyObject *__pyx_n_s_Uint32_List___reduce_cython;
static PyObject *__pyx_n_s_Uint32_List___setstate_cython;
static PyObject *__pyx_n_s_Uint64_List;
static PyObject *__pyx_n_s_Uint64_List___reduce_cython;
static PyObject *__pyx_n_s_Uint64_List___setstate_cython;
static PyObject *__pyx_kp_s_Unknown_enum_value_s;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s__118;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_n_s__4;
static PyObject *__pyx_n_s_as_struct;
static PyObject *__pyx_n_s_asyncio_coroutines;
static PyObject *__pyx_n_s_capnp;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_class_getitem;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_cls;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_dct;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dict_2;
static PyObject *__pyx_kp_u_disable;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dyn;
static PyObject *__pyx_n_s_dynamic;
static PyObject *__pyx_kp_u_enable;
static PyObject *__pyx_n_s_enum;
static PyObject *__pyx_kp_u_gc;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_graph_api_capnp;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_init_subclass;
static PyObject *__pyx_n_s_initializing;
static PyObject *__pyx_n_s_is_coroutine;
static PyObject *__pyx_kp_u_isenabled;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_members;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_mro_entries;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_parents;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_pytact;
static PyObject *__pyx_n_s_pytact_graph_api_capnp;
static PyObject *__pyx_n_s_pytact_graph_api_capnp_cython;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle___Pyx_EnumMeta;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_kp_s_s_s;
static PyObject *__pyx_kp_s_s_s_d;
static PyObject *__pyx_n_s_schema;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self_source_cannot_be_converted;
static PyObject *__pyx_n_s_set_name;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_spec;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_use_setstate;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_values;
#endif
/* #### Code section: decls ### */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5label___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_count___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_8identity___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5label___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6target___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_5nodes___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_nodes___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_5edges___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_edges___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_5value___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_4name___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_8has_name___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_8previous___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_17external_previous___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_external_previous___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_6status___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9inductive___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_12is_inductive___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_11constructor___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_14is_constructor___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_10projection___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_13is_projection___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_15manual_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_18is_manual_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_17tactical_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_20is_tactical_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_tactical_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_23manual_section_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_26is_manual_section_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_25tactical_section_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_28is_tactical_section_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_29has_tactical_section_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9type_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_type_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9term_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_term_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5value___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11proof_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_proof_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_15has_proof_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11context_def___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_context_def___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13context_assum___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_16is_context_assum___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10definition___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_definition___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14has_definition___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11const_empty___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_const_empty___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11sort_s_prop___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_sort_s_prop___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_prop___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_prop___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8sort_set___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11is_sort_set___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_type___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_type___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3rel___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_rel___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4evar___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_evar___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10evar_subst___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_evar_subst___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4cast___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_cast___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4prod___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_prod___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7lambda____get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_lambda____get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6let_in___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_let_in___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3app___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_app___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4case___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_case___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11case_branch___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_case_branch___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3fix___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_fix___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7fix_fun___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_fix_fun___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6co_fix___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_co_fix___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10co_fix_fun___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_co_fix_fun___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3int___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_int___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7has_int___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5float___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8is_float___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9has_float___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9primitive___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_primitive___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13has_primitive___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_8original___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11is_original___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_10discharged___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_13is_discharged___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11substituted___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_14is_substituted___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6tactic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_8outcomes___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_12has_outcomes___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5ident___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_8has_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_9base_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_13has_base_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_11interm_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_15has_interm_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5exact___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_18text_non_anonymous___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_22has_text_non_anonymous___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7unknown___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_10is_unknown___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5known___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_8is_known___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9has_known___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6before___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_10has_before___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5after___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9has_after___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4term___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9term_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_13has_term_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_16tactic_arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_20has_tactic_arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4root___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7context___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_11has_context___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_13context_names___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_17has_context_names___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_12context_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_16has_context_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_15conclusion_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_19has_conclusion_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_8has_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_2id___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_12unresolvable___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_15is_unresolvable___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_4term___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_7is_term___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5ident___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_10parameters___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5major___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12dependencies___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_dependencies___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_9has_graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_14representative___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11definitions___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_definitions___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11module_name___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_module_name___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12data_version___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_data_version___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_14no_such_tactic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_17is_no_such_tactic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_20mismatched_arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_23is_mismatched_arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_11parse_error___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_14is_parse_error___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_16illegal_argument___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_19is_illegal_argument___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7failure___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_10is_failure___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_8complete___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_11is_complete___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9new_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_12is_new_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_14protocol_error___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_17is_protocol_error___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_18has_protocol_error___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10initialize___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_13is_initialize___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7predict___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10is_predict___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_11synchronize___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_14is_synchronize___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_15check_alignment___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_18is_check_alignment___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_12data_version___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_16has_data_version___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_10stack_size___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7tactics___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_11has_tactics___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_9has_graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14log_annotation___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_18has_log_annotation___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14representative___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6tactic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10has_tactic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_9arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_13has_arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10confidence___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_11tactic_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_15has_tactic_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_10confidence___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_11initialized___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14is_initialized___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_10prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_13is_prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14has_prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15text_prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_18is_text_prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_19has_text_prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12synchronized___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15is_synchronized___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9alignment___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12is_alignment___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_17unaligned_tactics___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21has_unaligned_tactics___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21unaligned_definitions___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_25has_unaligned_definitions___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11String_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_11String_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11String_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11String_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index); /* proto */
static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name, PyObject *__pyx_v_parents, PyObject *__pyx_v_dct); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_value, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_IntP_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_FloatP_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Status_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Tactic_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Node_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Term_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_DataVersion_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Dataset_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Exception_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint16_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint32_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint64_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Int64_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_String_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Reader_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Reader_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Reader_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Node_Reader_List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_int_222419149;
static PyObject *__pyx_int_228825662;
static PyObject *__pyx_int_238750788;
#endif
#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__9;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__111;
static PyObject *__pyx_tuple__113;
static PyObject *__pyx_tuple__116;
static PyObject *__pyx_codeobj__10;
static PyObject *__pyx_codeobj__11;
static PyObject *__pyx_codeobj__12;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__14;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__41;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__45;
static PyObject *__pyx_codeobj__46;
static PyObject *__pyx_codeobj__47;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__49;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__52;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__54;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__56;
static PyObject *__pyx_codeobj__57;
static PyObject *__pyx_codeobj__58;
static PyObject *__pyx_codeobj__59;
static PyObject *__pyx_codeobj__60;
static PyObject *__pyx_codeobj__61;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__63;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__65;
static PyObject *__pyx_codeobj__66;
static PyObject *__pyx_codeobj__67;
static PyObject *__pyx_codeobj__68;
static PyObject *__pyx_codeobj__69;
static PyObject *__pyx_codeobj__70;
static PyObject *__pyx_codeobj__71;
static PyObject *__pyx_codeobj__72;
static PyObject *__pyx_codeobj__73;
static PyObject *__pyx_codeobj__74;
static PyObject *__pyx_codeobj__75;
static PyObject *__pyx_codeobj__76;
static PyObject *__pyx_codeobj__77;
static PyObject *__pyx_codeobj__78;
static PyObject *__pyx_codeobj__79;
static PyObject *__pyx_codeobj__80;
static PyObject *__pyx_codeobj__81;
static PyObject *__pyx_codeobj__82;
static PyObject *__pyx_codeobj__83;
static PyObject *__pyx_codeobj__84;
static PyObject *__pyx_codeobj__85;
static PyObject *__pyx_codeobj__86;
static PyObject *__pyx_codeobj__87;
static PyObject *__pyx_codeobj__88;
static PyObject *__pyx_codeobj__89;
static PyObject *__pyx_codeobj__90;
static PyObject *__pyx_codeobj__91;
static PyObject *__pyx_codeobj__92;
static PyObject *__pyx_codeobj__93;
static PyObject *__pyx_codeobj__94;
static PyObject *__pyx_codeobj__95;
static PyObject *__pyx_codeobj__96;
static PyObject *__pyx_codeobj__97;
static PyObject *__pyx_codeobj__98;
static PyObject *__pyx_codeobj__99;
static PyObject *__pyx_codeobj__100;
static PyObject *__pyx_codeobj__101;
static PyObject *__pyx_codeobj__102;
static PyObject *__pyx_codeobj__103;
static PyObject *__pyx_codeobj__104;
static PyObject *__pyx_codeobj__105;
static PyObject *__pyx_codeobj__106;
static PyObject *__pyx_codeobj__107;
static PyObject *__pyx_codeobj__109;
static PyObject *__pyx_codeobj__110;
static PyObject *__pyx_codeobj__112;
static PyObject *__pyx_codeobj__114;
static PyObject *__pyx_codeobj__115;
static PyObject *__pyx_codeobj__117;
#endif
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
#if CYTHON_USE_MODULE_STATE
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__StructSchemaField;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__StringArrayPtr;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp_SchemaParser;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicOrphan;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicStructBuilder;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicEnumField;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__Schema;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__InterfaceSchema;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicEnum;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__DynamicListBuilder;
  PyTypeObject *__pyx_ptype_5capnp_3lib_5capnp__MessageBuilder;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Int64_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_String_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List;
  PyTypeObject *__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List;
  PyObject *__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List;
  PyTypeObject *__pyx_ptype___Pyx_EnumMeta;
  PyObject *__Pyx_EnumMeta;
  PyObject *__pyx_n_s_ALIGNMENT;
  PyObject *__pyx_n_s_APP;
  PyObject *__pyx_n_s_APP_ARG;
  PyObject *__pyx_n_s_APP_FUN;
  PyObject *__pyx_n_s_AbstractTactic;
  PyObject *__pyx_n_s_AbstractTactic_Reader;
  PyObject *__pyx_n_s_AbstractTactic_Reader_List;
  PyObject *__pyx_n_s_AbstractTactic_Reader_List___red;
  PyObject *__pyx_n_s_AbstractTactic_Reader_List___set;
  PyObject *__pyx_n_s_AbstractTactic_Reader___reduce_c;
  PyObject *__pyx_n_s_AbstractTactic_Reader___setstate;
  PyObject *__pyx_n_s_Argument;
  PyObject *__pyx_n_s_Argument_Reader;
  PyObject *__pyx_n_s_Argument_Reader_List;
  PyObject *__pyx_n_s_Argument_Reader_List___reduce_cy;
  PyObject *__pyx_n_s_Argument_Reader_List___setstate;
  PyObject *__pyx_n_s_Argument_Reader___reduce_cython;
  PyObject *__pyx_n_s_Argument_Reader___setstate_cytho;
  PyObject *__pyx_n_s_Argument_Term_Reader;
  PyObject *__pyx_n_s_Argument_Term_Reader___reduce_cy;
  PyObject *__pyx_n_s_Argument_Term_Reader___setstate;
  PyObject *__pyx_n_s_Argument_Which;
  PyObject *__pyx_n_s_CASE;
  PyObject *__pyx_n_s_CASE_BRANCH;
  PyObject *__pyx_n_s_CASE_BRANCH_POINTER;
  PyObject *__pyx_n_s_CASE_IND;
  PyObject *__pyx_n_s_CASE_RETURN;
  PyObject *__pyx_n_s_CASE_TERM;
  PyObject *__pyx_n_s_CAST;
  PyObject *__pyx_n_s_CAST_TERM;
  PyObject *__pyx_n_s_CAST_TYPE;
  PyObject *__pyx_n_s_CHECK_ALIGNMENT;
  PyObject *__pyx_n_s_COMPLETE;
  PyObject *__pyx_n_s_CONSTRUCTOR;
  PyObject *__pyx_n_s_CONSTRUCT_TERM;
  PyObject *__pyx_n_s_CONST_DEF;
  PyObject *__pyx_n_s_CONST_EMPTY;
  PyObject *__pyx_n_s_CONST_OPAQUE_DEF;
  PyObject *__pyx_n_s_CONST_PRIMITIVE;
  PyObject *__pyx_n_s_CONST_TYPE;
  PyObject *__pyx_n_s_CONST_UNDEF;
  PyObject *__pyx_n_s_CONTEXT_ASSUM;
  PyObject *__pyx_n_s_CONTEXT_DEF;
  PyObject *__pyx_n_s_CONTEXT_DEF_TERM;
  PyObject *__pyx_n_s_CONTEXT_DEF_TYPE;
  PyObject *__pyx_n_s_CONTEXT_ELEM;
  PyObject *__pyx_n_s_CONTEXT_SUBJECT;
  PyObject *__pyx_n_s_CO_FIX;
  PyObject *__pyx_n_s_CO_FIX_FUN;
  PyObject *__pyx_n_s_CO_FIX_FUN_TERM;
  PyObject *__pyx_n_s_CO_FIX_FUN_TYPE;
  PyObject *__pyx_n_s_CO_FIX_MUTUAL;
  PyObject *__pyx_n_s_CO_FIX_RETURN;
  PyObject *__pyx_n_s_C_B_CONSTRUCT;
  PyObject *__pyx_n_s_C_B_TERM;
  PyObject *__pyx_n_s_ConflatableEdges;
  PyObject *__pyx_n_s_ConflatableEdges_Reader;
  PyObject *__pyx_n_s_ConflatableEdges_Reader___reduce;
  PyObject *__pyx_n_s_ConflatableEdges_Reader___setsta;
  PyObject *__pyx_n_s_DEFINITION;
  PyObject *__pyx_n_s_DISCHARGED;
  PyObject *__pyx_n_s_DataVersion;
  PyObject *__pyx_n_s_DataVersion_Reader;
  PyObject *__pyx_n_s_DataVersion_Reader___reduce_cyth;
  PyObject *__pyx_n_s_DataVersion_Reader___setstate_cy;
  PyObject *__pyx_n_s_Dataset;
  PyObject *__pyx_n_s_Dataset_Reader;
  PyObject *__pyx_n_s_Dataset_Reader___reduce_cython;
  PyObject *__pyx_n_s_Dataset_Reader___setstate_cython;
  PyObject *__pyx_n_s_Definition;
  PyObject *__pyx_n_s_Definition_Reader;
  PyObject *__pyx_n_s_Definition_Reader___reduce_cytho;
  PyObject *__pyx_n_s_Definition_Reader___setstate_cyt;
  PyObject *__pyx_n_s_Definition_Status_Reader;
  PyObject *__pyx_n_s_Definition_Status_Reader___reduc;
  PyObject *__pyx_n_s_Definition_Status_Reader___setst;
  PyObject *__pyx_n_s_Definition_Status_Substituted_Re;
  PyObject *__pyx_n_s_Definition_Status_Substituted_Re_2;
  PyObject *__pyx_n_s_Definition_Status_Substituted_Re_3;
  PyObject *__pyx_n_s_Definition_Status_Which;
  PyObject *__pyx_n_s_Definition_Which;
  PyObject *__pyx_n_s_EVAR;
  PyObject *__pyx_n_s_EVAR_SUBJECT;
  PyObject *__pyx_n_s_EVAR_SUBST;
  PyObject *__pyx_n_s_EVAR_SUBST_POINTER;
  PyObject *__pyx_n_s_EVAR_SUBST_TARGET;
  PyObject *__pyx_n_s_EVAR_SUBST_TERM;
  PyObject *__pyx_n_s_EdgeClassification;
  PyObject *__pyx_n_s_EdgeTarget;
  PyObject *__pyx_n_s_EnumBase;
  PyObject *__pyx_n_s_Exception;
  PyObject *__pyx_n_s_Exception_Reader;
  PyObject *__pyx_n_s_Exception_Reader___reduce_cython;
  PyObject *__pyx_n_s_Exception_Reader___setstate_cyth;
  PyObject *__pyx_n_s_Exception_Which;
  PyObject *__pyx_n_s_ExecutionResult;
  PyObject *__pyx_n_s_ExecutionResult_NewState_Reader;
  PyObject *__pyx_n_s_ExecutionResult_NewState_Reader_2;
  PyObject *__pyx_n_s_ExecutionResult_NewState_Reader_3;
  PyObject *__pyx_n_s_ExecutionResult_Reader;
  PyObject *__pyx_n_s_ExecutionResult_Reader___reduce;
  PyObject *__pyx_n_s_ExecutionResult_Reader___setstat;
  PyObject *__pyx_n_s_ExecutionResult_Which;
  PyObject *__pyx_n_s_FAILURE;
  PyObject *__pyx_n_s_FIX;
  PyObject *__pyx_n_s_FIX_FUN;
  PyObject *__pyx_n_s_FIX_FUN_TERM;
  PyObject *__pyx_n_s_FIX_FUN_TYPE;
  PyObject *__pyx_n_s_FIX_MUTUAL;
  PyObject *__pyx_n_s_FIX_RETURN;
  PyObject *__pyx_n_s_FLOAT;
  PyObject *__pyx_n_s_FloatP;
  PyObject *__pyx_n_s_FloatP_Reader;
  PyObject *__pyx_n_s_FloatP_Reader___reduce_cython;
  PyObject *__pyx_n_s_FloatP_Reader___setstate_cython;
  PyObject *__pyx_n_s_Graph;
  PyObject *__pyx_n_s_Graph_EdgeTarget_Reader;
  PyObject *__pyx_n_s_Graph_EdgeTarget_Reader_List;
  PyObject *__pyx_n_s_Graph_EdgeTarget_Reader_List___r;
  PyObject *__pyx_n_s_Graph_EdgeTarget_Reader_List___s;
  PyObject *__pyx_n_s_Graph_EdgeTarget_Reader___reduce;
  PyObject *__pyx_n_s_Graph_EdgeTarget_Reader___setsta;
  PyObject *__pyx_n_s_Graph_EdgeTarget_Target_Reader;
  PyObject *__pyx_n_s_Graph_EdgeTarget_Target_Reader_2;
  PyObject *__pyx_n_s_Graph_EdgeTarget_Target_Reader_3;
  PyObject *__pyx_n_s_Graph_Node_Label_Reader;
  PyObject *__pyx_n_s_Graph_Node_Label_Reader___reduce;
  PyObject *__pyx_n_s_Graph_Node_Label_Reader___setsta;
  PyObject *__pyx_n_s_Graph_Node_Label_Which;
  PyObject *__pyx_n_s_Graph_Node_Reader;
  PyObject *__pyx_n_s_Graph_Node_Reader_List;
  PyObject *__pyx_n_s_Graph_Node_Reader_List___reduce;
  PyObject *__pyx_n_s_Graph_Node_Reader_List___setstat;
  PyObject *__pyx_n_s_Graph_Node_Reader___reduce_cytho;
  PyObject *__pyx_n_s_Graph_Node_Reader___setstate_cyt;
  PyObject *__pyx_n_s_Graph_Reader;
  PyObject *__pyx_n_s_Graph_Reader___reduce_cython;
  PyObject *__pyx_n_s_Graph_Reader___setstate_cython;
  PyObject *__pyx_n_s_ILLEGAL_ARGUMENT;
  PyObject *__pyx_n_s_INDUCTIVE;
  PyObject *__pyx_n_s_IND_CONSTRUCT;
  PyObject *__pyx_n_s_IND_PROJECTION;
  PyObject *__pyx_n_s_IND_TYPE;
  PyObject *__pyx_n_s_INITIALIZE;
  PyObject *__pyx_n_s_INITIALIZED;
  PyObject *__pyx_n_s_INT;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_n_s_IndexError;
  PyObject *__pyx_n_s_Int64_List;
  PyObject *__pyx_n_s_Int64_List___reduce_cython;
  PyObject *__pyx_n_s_Int64_List___setstate_cython;
  PyObject *__pyx_n_s_IntEnum;
  PyObject *__pyx_n_s_IntP;
  PyObject *__pyx_n_s_IntP_Reader;
  PyObject *__pyx_n_s_IntP_Reader___reduce_cython;
  PyObject *__pyx_n_s_IntP_Reader___setstate_cython;
  PyObject *__pyx_n_s_KNOWN;
  PyObject *__pyx_n_s_LAMBDA;
  PyObject *__pyx_n_s_LAMBDA_TERM;
  PyObject *__pyx_n_s_LAMBDA_TYPE;
  PyObject *__pyx_n_s_LET_IN;
  PyObject *__pyx_n_s_LET_IN_DEF;
  PyObject *__pyx_n_s_LET_IN_TERM;
  PyObject *__pyx_n_s_LET_IN_TYPE;
  PyObject *__pyx_n_s_MANUAL_CONSTANT;
  PyObject *__pyx_n_s_MANUAL_SECTION_CONSTANT;
  PyObject *__pyx_n_s_MISMATCHED_ARGUMENTS;
  PyObject *__pyx_n_s_NEW_STATE;
  PyObject *__pyx_n_s_NO_SUCH_TACTIC;
  PyObject *__pyx_n_s_Node;
  PyObject *__pyx_n_s_Node_Reader;
  PyObject *__pyx_n_s_Node_Reader_List;
  PyObject *__pyx_n_s_Node_Reader_List___reduce_cython;
  PyObject *__pyx_n_s_Node_Reader_List___setstate_cyth;
  PyObject *__pyx_n_s_Node_Reader___reduce_cython;
  PyObject *__pyx_n_s_Node_Reader___setstate_cython;
  PyObject *__pyx_n_s_ORIGINAL;
  PyObject *__pyx_n_s_OrderedDict;
  PyObject *__pyx_kp_u_Out_of_bounds;
  PyObject *__pyx_n_s_Outcome;
  PyObject *__pyx_n_s_Outcome_Reader;
  PyObject *__pyx_n_s_Outcome_Reader_List;
  PyObject *__pyx_n_s_Outcome_Reader_List___reduce_cyt;
  PyObject *__pyx_n_s_Outcome_Reader_List___setstate_c;
  PyObject *__pyx_n_s_Outcome_Reader___reduce_cython;
  PyObject *__pyx_n_s_Outcome_Reader___setstate_cython;
  PyObject *__pyx_n_s_Outcome_Term_Reader;
  PyObject *__pyx_n_s_Outcome_Term_Reader___reduce_cyt;
  PyObject *__pyx_n_s_Outcome_Term_Reader___setstate_c;
  PyObject *__pyx_n_s_PARSE_ERROR;
  PyObject *__pyx_n_s_PREDICT;
  PyObject *__pyx_n_s_PREDICTION;
  PyObject *__pyx_n_s_PRIMITIVE;
  PyObject *__pyx_n_s_PROD;
  PyObject *__pyx_n_s_PROD_TERM;
  PyObject *__pyx_n_s_PROD_TYPE;
  PyObject *__pyx_n_s_PROJECTION;
  PyObject *__pyx_n_s_PROJ_TERM;
  PyObject *__pyx_n_s_PROOF_STATE;
  PyObject *__pyx_n_s_PROTOCOL_ERROR;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_n_s_Prediction;
  PyObject *__pyx_n_s_PredictionProtocol;
  PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re;
  PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re_2;
  PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re_3;
  PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re_4;
  PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re_5;
  PyObject *__pyx_n_s_PredictionProtocol_Prediction_Re_6;
  PyObject *__pyx_n_s_PredictionProtocol_Reader;
  PyObject *__pyx_n_s_PredictionProtocol_Reader___redu;
  PyObject *__pyx_n_s_PredictionProtocol_Reader___sets;
  PyObject *__pyx_n_s_PredictionProtocol_Request_Initi;
  PyObject *__pyx_n_s_PredictionProtocol_Request_Initi_2;
  PyObject *__pyx_n_s_PredictionProtocol_Request_Initi_3;
  PyObject *__pyx_n_s_PredictionProtocol_Request_Predi;
  PyObject *__pyx_n_s_PredictionProtocol_Request_Predi_2;
  PyObject *__pyx_n_s_PredictionProtocol_Request_Predi_3;
  PyObject *__pyx_n_s_PredictionProtocol_Request_Reade;
  PyObject *__pyx_n_s_PredictionProtocol_Request_Reade_2;
  PyObject *__pyx_n_s_PredictionProtocol_Request_Reade_3;
  PyObject *__pyx_n_s_PredictionProtocol_Request_Which;
  PyObject *__pyx_n_s_PredictionProtocol_Response_Alig;
  PyObject *__pyx_n_s_PredictionProtocol_Response_Alig_2;
  PyObject *__pyx_n_s_PredictionProtocol_Response_Alig_3;
  PyObject *__pyx_n_s_PredictionProtocol_Response_Read;
  PyObject *__pyx_n_s_PredictionProtocol_Response_Read_2;
  PyObject *__pyx_n_s_PredictionProtocol_Response_Read_3;
  PyObject *__pyx_n_s_PredictionProtocol_Response_Whic;
  PyObject *__pyx_n_s_PredictionProtocol_TextPredictio;
  PyObject *__pyx_n_s_PredictionProtocol_TextPredictio_2;
  PyObject *__pyx_n_s_PredictionProtocol_TextPredictio_3;
  PyObject *__pyx_n_s_PredictionProtocol_TextPredictio_4;
  PyObject *__pyx_n_s_PredictionProtocol_TextPredictio_5;
  PyObject *__pyx_n_s_PredictionProtocol_TextPredictio_6;
  PyObject *__pyx_n_s_ProofState;
  PyObject *__pyx_n_s_ProofState_Reader;
  PyObject *__pyx_n_s_ProofState_Reader_List;
  PyObject *__pyx_n_s_ProofState_Reader_List___reduce;
  PyObject *__pyx_n_s_ProofState_Reader_List___setstat;
  PyObject *__pyx_n_s_ProofState_Reader___reduce_cytho;
  PyObject *__pyx_n_s_ProofState_Reader___setstate_cyt;
  PyObject *__pyx_n_s_ProofState_Root_Reader;
  PyObject *__pyx_n_s_ProofState_Root_Reader___reduce;
  PyObject *__pyx_n_s_ProofState_Root_Reader___setstat;
  PyObject *__pyx_n_s_ProofStep;
  PyObject *__pyx_n_s_ProofStep_Reader;
  PyObject *__pyx_n_s_ProofStep_Reader_List;
  PyObject *__pyx_n_s_ProofStep_Reader_List___reduce_c;
  PyObject *__pyx_n_s_ProofStep_Reader_List___setstate;
  PyObject *__pyx_n_s_ProofStep_Reader___reduce_cython;
  PyObject *__pyx_n_s_ProofStep_Reader___setstate_cyth;
  PyObject *__pyx_n_s_ProofStep_Tactic_Reader;
  PyObject *__pyx_n_s_ProofStep_Tactic_Reader___reduce;
  PyObject *__pyx_n_s_ProofStep_Tactic_Reader___setsta;
  PyObject *__pyx_n_s_ProofStep_Tactic_Which;
  PyObject *__pyx_n_s_Pyx_EnumBase;
  PyObject *__pyx_n_s_Pyx_EnumBase___new;
  PyObject *__pyx_n_s_Pyx_EnumBase___repr;
  PyObject *__pyx_n_s_Pyx_EnumBase___str;
  PyObject *__pyx_n_s_Pyx_EnumMeta___reduce_cython;
  PyObject *__pyx_n_s_Pyx_EnumMeta___setstate_cython;
  PyObject *__pyx_n_s_REL;
  PyObject *__pyx_n_s_REL_POINTER;
  PyObject *__pyx_n_s_Request;
  PyObject *__pyx_n_s_Response;
  PyObject *__pyx_n_s_SORT_PROP;
  PyObject *__pyx_n_s_SORT_SET;
  PyObject *__pyx_n_s_SORT_S_PROP;
  PyObject *__pyx_n_s_SORT_TYPE;
  PyObject *__pyx_n_s_SUBSTITUTED;
  PyObject *__pyx_n_s_SYNCHRONIZE;
  PyObject *__pyx_n_s_SYNCHRONIZED;
  PyObject *__pyx_n_s_String_List;
  PyObject *__pyx_n_s_String_List___reduce_cython;
  PyObject *__pyx_n_s_String_List___setstate_cython;
  PyObject *__pyx_n_s_TACTICAL_CONSTANT;
  PyObject *__pyx_n_s_TACTICAL_SECTION_CONSTANT;
  PyObject *__pyx_n_s_TERM;
  PyObject *__pyx_n_s_TEXT_PREDICTION;
  PyObject *__pyx_n_s_Tactic;
  PyObject *__pyx_n_s_Tactic_Reader;
  PyObject *__pyx_n_s_Tactic_Reader___reduce_cython;
  PyObject *__pyx_n_s_Tactic_Reader___setstate_cython;
  PyObject *__pyx_n_s_TextPrediction;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_n_s_UNKNOWN;
  PyObject *__pyx_n_s_UNRESOLVABLE;
  PyObject *__pyx_n_s_Uint16_List;
  PyObject *__pyx_n_s_Uint16_List___reduce_cython;
  PyObject *__pyx_n_s_Uint16_List___setstate_cython;
  PyObject *__pyx_n_s_Uint32_List;
  PyObject *__pyx_n_s_Uint32_List___reduce_cython;
  PyObject *__pyx_n_s_Uint32_List___setstate_cython;
  PyObject *__pyx_n_s_Uint64_List;
  PyObject *__pyx_n_s_Uint64_List___reduce_cython;
  PyObject *__pyx_n_s_Uint64_List___setstate_cython;
  PyObject *__pyx_kp_s_Unknown_enum_value_s;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_s__118;
  PyObject *__pyx_kp_u__3;
  PyObject *__pyx_n_s__4;
  PyObject *__pyx_n_s_as_struct;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_capnp;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_cls;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_n_s_dct;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_doc;
  PyObject *__pyx_n_s_dyn;
  PyObject *__pyx_n_s_dynamic;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_enum;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_graph_api_capnp;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_init;
  PyObject *__pyx_n_s_init_subclass;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_members;
  PyObject *__pyx_n_s_metaclass;
  PyObject *__pyx_n_s_module;
  PyObject *__pyx_n_s_mro_entries;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_n_s_os;
  PyObject *__pyx_n_s_parents;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_prepare;
  PyObject *__pyx_n_s_pytact;
  PyObject *__pyx_n_s_pytact_graph_api_capnp;
  PyObject *__pyx_n_s_pytact_graph_api_capnp_cython;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle___Pyx_EnumMeta;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_qualname;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_repr;
  PyObject *__pyx_n_s_res;
  PyObject *__pyx_kp_s_s_s;
  PyObject *__pyx_kp_s_s_s_d;
  PyObject *__pyx_n_s_schema;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_kp_s_self_source_cannot_be_converted;
  PyObject *__pyx_n_s_set_name;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_n_s_v;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_values;
  PyObject *__pyx_int_222419149;
  PyObject *__pyx_int_228825662;
  PyObject *__pyx_int_238750788;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_tuple__2;
  PyObject *__pyx_tuple__5;
  PyObject *__pyx_tuple__6;
  PyObject *__pyx_tuple__8;
  PyObject *__pyx_codeobj__7;
  PyObject *__pyx_codeobj__9;
  PyObject *__pyx_tuple__108;
  PyObject *__pyx_tuple__111;
  PyObject *__pyx_tuple__113;
  PyObject *__pyx_tuple__116;
  PyObject *__pyx_codeobj__10;
  PyObject *__pyx_codeobj__11;
  PyObject *__pyx_codeobj__12;
  PyObject *__pyx_codeobj__13;
  PyObject *__pyx_codeobj__14;
  PyObject *__pyx_codeobj__15;
  PyObject *__pyx_codeobj__16;
  PyObject *__pyx_codeobj__17;
  PyObject *__pyx_codeobj__18;
  PyObject *__pyx_codeobj__19;
  PyObject *__pyx_codeobj__20;
  PyObject *__pyx_codeobj__21;
  PyObject *__pyx_codeobj__22;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__24;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__26;
  PyObject *__pyx_codeobj__27;
  PyObject *__pyx_codeobj__28;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__30;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__32;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__34;
  PyObject *__pyx_codeobj__35;
  PyObject *__pyx_codeobj__36;
  PyObject *__pyx_codeobj__37;
  PyObject *__pyx_codeobj__38;
  PyObject *__pyx_codeobj__39;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__43;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__45;
  PyObject *__pyx_codeobj__46;
  PyObject *__pyx_codeobj__47;
  PyObject *__pyx_codeobj__48;
  PyObject *__pyx_codeobj__49;
  PyObject *__pyx_codeobj__50;
  PyObject *__pyx_codeobj__51;
  PyObject *__pyx_codeobj__52;
  PyObject *__pyx_codeobj__53;
  PyObject *__pyx_codeobj__54;
  PyObject *__pyx_codeobj__55;
  PyObject *__pyx_codeobj__56;
  PyObject *__pyx_codeobj__57;
  PyObject *__pyx_codeobj__58;
  PyObject *__pyx_codeobj__59;
  PyObject *__pyx_codeobj__60;
  PyObject *__pyx_codeobj__61;
  PyObject *__pyx_codeobj__62;
  PyObject *__pyx_codeobj__63;
  PyObject *__pyx_codeobj__64;
  PyObject *__pyx_codeobj__65;
  PyObject *__pyx_codeobj__66;
  PyObject *__pyx_codeobj__67;
  PyObject *__pyx_codeobj__68;
  PyObject *__pyx_codeobj__69;
  PyObject *__pyx_codeobj__70;
  PyObject *__pyx_codeobj__71;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__73;
  PyObject *__pyx_codeobj__74;
  PyObject *__pyx_codeobj__75;
  PyObject *__pyx_codeobj__76;
  PyObject *__pyx_codeobj__77;
  PyObject *__pyx_codeobj__78;
  PyObject *__pyx_codeobj__79;
  PyObject *__pyx_codeobj__80;
  PyObject *__pyx_codeobj__81;
  PyObject *__pyx_codeobj__82;
  PyObject *__pyx_codeobj__83;
  PyObject *__pyx_codeobj__84;
  PyObject *__pyx_codeobj__85;
  PyObject *__pyx_codeobj__86;
  PyObject *__pyx_codeobj__87;
  PyObject *__pyx_codeobj__88;
  PyObject *__pyx_codeobj__89;
  PyObject *__pyx_codeobj__90;
  PyObject *__pyx_codeobj__91;
  PyObject *__pyx_codeobj__92;
  PyObject *__pyx_codeobj__93;
  PyObject *__pyx_codeobj__94;
  PyObject *__pyx_codeobj__95;
  PyObject *__pyx_codeobj__96;
  PyObject *__pyx_codeobj__97;
  PyObject *__pyx_codeobj__98;
  PyObject *__pyx_codeobj__99;
  PyObject *__pyx_codeobj__100;
  PyObject *__pyx_codeobj__101;
  PyObject *__pyx_codeobj__102;
  PyObject *__pyx_codeobj__103;
  PyObject *__pyx_codeobj__104;
  PyObject *__pyx_codeobj__105;
  PyObject *__pyx_codeobj__106;
  PyObject *__pyx_codeobj__107;
  PyObject *__pyx_codeobj__109;
  PyObject *__pyx_codeobj__110;
  PyObject *__pyx_codeobj__112;
  PyObject *__pyx_codeobj__114;
  PyObject *__pyx_codeobj__115;
  PyObject *__pyx_codeobj__117;
} __pyx_mstate;

#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__StructSchemaField);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__StringArrayPtr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp_SchemaParser);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicOrphan);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicStructBuilder);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicEnumField);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__Schema);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__InterfaceSchema);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicEnum);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicListBuilder);
  Py_CLEAR(clear_module_state->__pyx_ptype_5capnp_3lib_5capnp__MessageBuilder);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Int64_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_String_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_ptype___Pyx_EnumMeta);
  Py_CLEAR(clear_module_state->__Pyx_EnumMeta);
  Py_CLEAR(clear_module_state->__pyx_n_s_ALIGNMENT);
  Py_CLEAR(clear_module_state->__pyx_n_s_APP);
  Py_CLEAR(clear_module_state->__pyx_n_s_APP_ARG);
  Py_CLEAR(clear_module_state->__pyx_n_s_APP_FUN);
  Py_CLEAR(clear_module_state->__pyx_n_s_AbstractTactic);
  Py_CLEAR(clear_module_state->__pyx_n_s_AbstractTactic_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_AbstractTactic_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_AbstractTactic_Reader_List___red);
  Py_CLEAR(clear_module_state->__pyx_n_s_AbstractTactic_Reader_List___set);
  Py_CLEAR(clear_module_state->__pyx_n_s_AbstractTactic_Reader___reduce_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_AbstractTactic_Reader___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument_Reader_List___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument_Reader_List___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument_Reader___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument_Term_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument_Term_Reader___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument_Term_Reader___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_Argument_Which);
  Py_CLEAR(clear_module_state->__pyx_n_s_CASE);
  Py_CLEAR(clear_module_state->__pyx_n_s_CASE_BRANCH);
  Py_CLEAR(clear_module_state->__pyx_n_s_CASE_BRANCH_POINTER);
  Py_CLEAR(clear_module_state->__pyx_n_s_CASE_IND);
  Py_CLEAR(clear_module_state->__pyx_n_s_CASE_RETURN);
  Py_CLEAR(clear_module_state->__pyx_n_s_CASE_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_CAST);
  Py_CLEAR(clear_module_state->__pyx_n_s_CAST_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_CAST_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_CHECK_ALIGNMENT);
  Py_CLEAR(clear_module_state->__pyx_n_s_COMPLETE);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONSTRUCTOR);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONSTRUCT_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONST_DEF);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONST_EMPTY);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONST_OPAQUE_DEF);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONST_PRIMITIVE);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONST_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONST_UNDEF);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONTEXT_ASSUM);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONTEXT_DEF);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONTEXT_DEF_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONTEXT_DEF_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONTEXT_ELEM);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONTEXT_SUBJECT);
  Py_CLEAR(clear_module_state->__pyx_n_s_CO_FIX);
  Py_CLEAR(clear_module_state->__pyx_n_s_CO_FIX_FUN);
  Py_CLEAR(clear_module_state->__pyx_n_s_CO_FIX_FUN_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_CO_FIX_FUN_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_CO_FIX_MUTUAL);
  Py_CLEAR(clear_module_state->__pyx_n_s_CO_FIX_RETURN);
  Py_CLEAR(clear_module_state->__pyx_n_s_C_B_CONSTRUCT);
  Py_CLEAR(clear_module_state->__pyx_n_s_C_B_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConflatableEdges);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConflatableEdges_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConflatableEdges_Reader___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConflatableEdges_Reader___setsta);
  Py_CLEAR(clear_module_state->__pyx_n_s_DEFINITION);
  Py_CLEAR(clear_module_state->__pyx_n_s_DISCHARGED);
  Py_CLEAR(clear_module_state->__pyx_n_s_DataVersion);
  Py_CLEAR(clear_module_state->__pyx_n_s_DataVersion_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_DataVersion_Reader___reduce_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_DataVersion_Reader___setstate_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Dataset);
  Py_CLEAR(clear_module_state->__pyx_n_s_Dataset_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Dataset_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Dataset_Reader___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Reader___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Reader___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Status_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Status_Reader___reduc);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Status_Reader___setst);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Status_Substituted_Re);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Status_Substituted_Re_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Status_Substituted_Re_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Status_Which);
  Py_CLEAR(clear_module_state->__pyx_n_s_Definition_Which);
  Py_CLEAR(clear_module_state->__pyx_n_s_EVAR);
  Py_CLEAR(clear_module_state->__pyx_n_s_EVAR_SUBJECT);
  Py_CLEAR(clear_module_state->__pyx_n_s_EVAR_SUBST);
  Py_CLEAR(clear_module_state->__pyx_n_s_EVAR_SUBST_POINTER);
  Py_CLEAR(clear_module_state->__pyx_n_s_EVAR_SUBST_TARGET);
  Py_CLEAR(clear_module_state->__pyx_n_s_EVAR_SUBST_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_EdgeClassification);
  Py_CLEAR(clear_module_state->__pyx_n_s_EdgeTarget);
  Py_CLEAR(clear_module_state->__pyx_n_s_EnumBase);
  Py_CLEAR(clear_module_state->__pyx_n_s_Exception);
  Py_CLEAR(clear_module_state->__pyx_n_s_Exception_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Exception_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Exception_Reader___setstate_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_Exception_Which);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecutionResult);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecutionResult_NewState_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecutionResult_NewState_Reader_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecutionResult_NewState_Reader_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecutionResult_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecutionResult_Reader___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecutionResult_Reader___setstat);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecutionResult_Which);
  Py_CLEAR(clear_module_state->__pyx_n_s_FAILURE);
  Py_CLEAR(clear_module_state->__pyx_n_s_FIX);
  Py_CLEAR(clear_module_state->__pyx_n_s_FIX_FUN);
  Py_CLEAR(clear_module_state->__pyx_n_s_FIX_FUN_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_FIX_FUN_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_FIX_MUTUAL);
  Py_CLEAR(clear_module_state->__pyx_n_s_FIX_RETURN);
  Py_CLEAR(clear_module_state->__pyx_n_s_FLOAT);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloatP);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloatP_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloatP_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloatP_Reader___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_EdgeTarget_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_EdgeTarget_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_EdgeTarget_Reader_List___r);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_EdgeTarget_Reader_List___s);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_EdgeTarget_Reader___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_EdgeTarget_Reader___setsta);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_EdgeTarget_Target_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_EdgeTarget_Target_Reader_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_EdgeTarget_Target_Reader_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Node_Label_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Node_Label_Reader___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Node_Label_Reader___setsta);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Node_Label_Which);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Node_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Node_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Node_Reader_List___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Node_Reader_List___setstat);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Node_Reader___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Node_Reader___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Graph_Reader___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ILLEGAL_ARGUMENT);
  Py_CLEAR(clear_module_state->__pyx_n_s_INDUCTIVE);
  Py_CLEAR(clear_module_state->__pyx_n_s_IND_CONSTRUCT);
  Py_CLEAR(clear_module_state->__pyx_n_s_IND_PROJECTION);
  Py_CLEAR(clear_module_state->__pyx_n_s_IND_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_INITIALIZE);
  Py_CLEAR(clear_module_state->__pyx_n_s_INITIALIZED);
  Py_CLEAR(clear_module_state->__pyx_n_s_INT);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Int64_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_Int64_List___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Int64_List___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntEnum);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntP);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntP_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntP_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntP_Reader___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_KNOWN);
  Py_CLEAR(clear_module_state->__pyx_n_s_LAMBDA);
  Py_CLEAR(clear_module_state->__pyx_n_s_LAMBDA_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_LAMBDA_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_LET_IN);
  Py_CLEAR(clear_module_state->__pyx_n_s_LET_IN_DEF);
  Py_CLEAR(clear_module_state->__pyx_n_s_LET_IN_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_LET_IN_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_MANUAL_CONSTANT);
  Py_CLEAR(clear_module_state->__pyx_n_s_MANUAL_SECTION_CONSTANT);
  Py_CLEAR(clear_module_state->__pyx_n_s_MISMATCHED_ARGUMENTS);
  Py_CLEAR(clear_module_state->__pyx_n_s_NEW_STATE);
  Py_CLEAR(clear_module_state->__pyx_n_s_NO_SUCH_TACTIC);
  Py_CLEAR(clear_module_state->__pyx_n_s_Node);
  Py_CLEAR(clear_module_state->__pyx_n_s_Node_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Node_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_Node_Reader_List___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Node_Reader_List___setstate_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_Node_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Node_Reader___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ORIGINAL);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderedDict);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Out_of_bounds);
  Py_CLEAR(clear_module_state->__pyx_n_s_Outcome);
  Py_CLEAR(clear_module_state->__pyx_n_s_Outcome_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Outcome_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_Outcome_Reader_List___reduce_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_Outcome_Reader_List___setstate_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_Outcome_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Outcome_Reader___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Outcome_Term_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Outcome_Term_Reader___reduce_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_Outcome_Term_Reader___setstate_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_PARSE_ERROR);
  Py_CLEAR(clear_module_state->__pyx_n_s_PREDICT);
  Py_CLEAR(clear_module_state->__pyx_n_s_PREDICTION);
  Py_CLEAR(clear_module_state->__pyx_n_s_PRIMITIVE);
  Py_CLEAR(clear_module_state->__pyx_n_s_PROD);
  Py_CLEAR(clear_module_state->__pyx_n_s_PROD_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_PROD_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_PROJECTION);
  Py_CLEAR(clear_module_state->__pyx_n_s_PROJ_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_PROOF_STATE);
  Py_CLEAR(clear_module_state->__pyx_n_s_PROTOCOL_ERROR);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Prediction);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re_5);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re_6);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Reader___redu);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Reader___sets);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Request_Initi);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Request_Initi_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Request_Initi_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Request_Predi);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Request_Predi_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Request_Predi_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Request_Reade);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Request_Reade_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Request_Reade_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Request_Which);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Response_Alig);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Response_Alig_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Response_Alig_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Response_Read);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Response_Read_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Response_Read_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_Response_Whic);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_TextPredictio);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_TextPredictio_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_TextPredictio_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_TextPredictio_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_TextPredictio_5);
  Py_CLEAR(clear_module_state->__pyx_n_s_PredictionProtocol_TextPredictio_6);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofState);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofState_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofState_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofState_Reader_List___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofState_Reader_List___setstat);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofState_Reader___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofState_Reader___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofState_Root_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofState_Root_Reader___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofState_Root_Reader___setstat);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep_Reader_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep_Reader_List___reduce_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep_Reader_List___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep_Reader___setstate_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep_Tactic_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep_Tactic_Reader___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep_Tactic_Reader___setsta);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProofStep_Tactic_Which);
  Py_CLEAR(clear_module_state->__pyx_n_s_Pyx_EnumBase);
  Py_CLEAR(clear_module_state->__pyx_n_s_Pyx_EnumBase___new);
  Py_CLEAR(clear_module_state->__pyx_n_s_Pyx_EnumBase___repr);
  Py_CLEAR(clear_module_state->__pyx_n_s_Pyx_EnumBase___str);
  Py_CLEAR(clear_module_state->__pyx_n_s_Pyx_EnumMeta___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Pyx_EnumMeta___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_REL);
  Py_CLEAR(clear_module_state->__pyx_n_s_REL_POINTER);
  Py_CLEAR(clear_module_state->__pyx_n_s_Request);
  Py_CLEAR(clear_module_state->__pyx_n_s_Response);
  Py_CLEAR(clear_module_state->__pyx_n_s_SORT_PROP);
  Py_CLEAR(clear_module_state->__pyx_n_s_SORT_SET);
  Py_CLEAR(clear_module_state->__pyx_n_s_SORT_S_PROP);
  Py_CLEAR(clear_module_state->__pyx_n_s_SORT_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_SUBSTITUTED);
  Py_CLEAR(clear_module_state->__pyx_n_s_SYNCHRONIZE);
  Py_CLEAR(clear_module_state->__pyx_n_s_SYNCHRONIZED);
  Py_CLEAR(clear_module_state->__pyx_n_s_String_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_String_List___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_String_List___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TACTICAL_CONSTANT);
  Py_CLEAR(clear_module_state->__pyx_n_s_TACTICAL_SECTION_CONSTANT);
  Py_CLEAR(clear_module_state->__pyx_n_s_TERM);
  Py_CLEAR(clear_module_state->__pyx_n_s_TEXT_PREDICTION);
  Py_CLEAR(clear_module_state->__pyx_n_s_Tactic);
  Py_CLEAR(clear_module_state->__pyx_n_s_Tactic_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Tactic_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Tactic_Reader___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TextPrediction);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_UNKNOWN);
  Py_CLEAR(clear_module_state->__pyx_n_s_UNRESOLVABLE);
  Py_CLEAR(clear_module_state->__pyx_n_s_Uint16_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_Uint16_List___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Uint16_List___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Uint32_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_Uint32_List___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Uint32_List___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Uint64_List);
  Py_CLEAR(clear_module_state->__pyx_n_s_Uint64_List___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Uint64_List___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Unknown_enum_value_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_s__118);
  Py_CLEAR(clear_module_state->__pyx_kp_u__3);
  Py_CLEAR(clear_module_state->__pyx_n_s__4);
  Py_CLEAR(clear_module_state->__pyx_n_s_as_struct);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_capnp);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_cls);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_n_s_dct);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_doc);
  Py_CLEAR(clear_module_state->__pyx_n_s_dyn);
  Py_CLEAR(clear_module_state->__pyx_n_s_dynamic);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_enum);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_graph_api_capnp);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_init);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_subclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_members);
  Py_CLEAR(clear_module_state->__pyx_n_s_metaclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_module);
  Py_CLEAR(clear_module_state->__pyx_n_s_mro_entries);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_n_s_os);
  Py_CLEAR(clear_module_state->__pyx_n_s_parents);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_prepare);
  Py_CLEAR(clear_module_state->__pyx_n_s_pytact);
  Py_CLEAR(clear_module_state->__pyx_n_s_pytact_graph_api_capnp);
  Py_CLEAR(clear_module_state->__pyx_n_s_pytact_graph_api_capnp_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle___Pyx_EnumMeta);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_qualname);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_repr);
  Py_CLEAR(clear_module_state->__pyx_n_s_res);
  Py_CLEAR(clear_module_state->__pyx_kp_s_s_s);
  Py_CLEAR(clear_module_state->__pyx_kp_s_s_s_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self_source_cannot_be_converted);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_v);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_values);
  Py_CLEAR(clear_module_state->__pyx_int_222419149);
  Py_CLEAR(clear_module_state->__pyx_int_228825662);
  Py_CLEAR(clear_module_state->__pyx_int_238750788);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_tuple__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__5);
  Py_CLEAR(clear_module_state->__pyx_tuple__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__8);
  Py_CLEAR(clear_module_state->__pyx_codeobj__7);
  Py_CLEAR(clear_module_state->__pyx_codeobj__9);
  Py_CLEAR(clear_module_state->__pyx_tuple__108);
  Py_CLEAR(clear_module_state->__pyx_tuple__111);
  Py_CLEAR(clear_module_state->__pyx_tuple__113);
  Py_CLEAR(clear_module_state->__pyx_tuple__116);
  Py_CLEAR(clear_module_state->__pyx_codeobj__10);
  Py_CLEAR(clear_module_state->__pyx_codeobj__11);
  Py_CLEAR(clear_module_state->__pyx_codeobj__12);
  Py_CLEAR(clear_module_state->__pyx_codeobj__13);
  Py_CLEAR(clear_module_state->__pyx_codeobj__14);
  Py_CLEAR(clear_module_state->__pyx_codeobj__15);
  Py_CLEAR(clear_module_state->__pyx_codeobj__16);
  Py_CLEAR(clear_module_state->__pyx_codeobj__17);
  Py_CLEAR(clear_module_state->__pyx_codeobj__18);
  Py_CLEAR(clear_module_state->__pyx_codeobj__19);
  Py_CLEAR(clear_module_state->__pyx_codeobj__20);
  Py_CLEAR(clear_module_state->__pyx_codeobj__21);
  Py_CLEAR(clear_module_state->__pyx_codeobj__22);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__24);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__26);
  Py_CLEAR(clear_module_state->__pyx_codeobj__27);
  Py_CLEAR(clear_module_state->__pyx_codeobj__28);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__30);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__34);
  Py_CLEAR(clear_module_state->__pyx_codeobj__35);
  Py_CLEAR(clear_module_state->__pyx_codeobj__36);
  Py_CLEAR(clear_module_state->__pyx_codeobj__37);
  Py_CLEAR(clear_module_state->__pyx_codeobj__38);
  Py_CLEAR(clear_module_state->__pyx_codeobj__39);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__45);
  Py_CLEAR(clear_module_state->__pyx_codeobj__46);
  Py_CLEAR(clear_module_state->__pyx_codeobj__47);
  Py_CLEAR(clear_module_state->__pyx_codeobj__48);
  Py_CLEAR(clear_module_state->__pyx_codeobj__49);
  Py_CLEAR(clear_module_state->__pyx_codeobj__50);
  Py_CLEAR(clear_module_state->__pyx_codeobj__51);
  Py_CLEAR(clear_module_state->__pyx_codeobj__52);
  Py_CLEAR(clear_module_state->__pyx_codeobj__53);
  Py_CLEAR(clear_module_state->__pyx_codeobj__54);
  Py_CLEAR(clear_module_state->__pyx_codeobj__55);
  Py_CLEAR(clear_module_state->__pyx_codeobj__56);
  Py_CLEAR(clear_module_state->__pyx_codeobj__57);
  Py_CLEAR(clear_module_state->__pyx_codeobj__58);
  Py_CLEAR(clear_module_state->__pyx_codeobj__59);
  Py_CLEAR(clear_module_state->__pyx_codeobj__60);
  Py_CLEAR(clear_module_state->__pyx_codeobj__61);
  Py_CLEAR(clear_module_state->__pyx_codeobj__62);
  Py_CLEAR(clear_module_state->__pyx_codeobj__63);
  Py_CLEAR(clear_module_state->__pyx_codeobj__64);
  Py_CLEAR(clear_module_state->__pyx_codeobj__65);
  Py_CLEAR(clear_module_state->__pyx_codeobj__66);
  Py_CLEAR(clear_module_state->__pyx_codeobj__67);
  Py_CLEAR(clear_module_state->__pyx_codeobj__68);
  Py_CLEAR(clear_module_state->__pyx_codeobj__69);
  Py_CLEAR(clear_module_state->__pyx_codeobj__70);
  Py_CLEAR(clear_module_state->__pyx_codeobj__71);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__73);
  Py_CLEAR(clear_module_state->__pyx_codeobj__74);
  Py_CLEAR(clear_module_state->__pyx_codeobj__75);
  Py_CLEAR(clear_module_state->__pyx_codeobj__76);
  Py_CLEAR(clear_module_state->__pyx_codeobj__77);
  Py_CLEAR(clear_module_state->__pyx_codeobj__78);
  Py_CLEAR(clear_module_state->__pyx_codeobj__79);
  Py_CLEAR(clear_module_state->__pyx_codeobj__80);
  Py_CLEAR(clear_module_state->__pyx_codeobj__81);
  Py_CLEAR(clear_module_state->__pyx_codeobj__82);
  Py_CLEAR(clear_module_state->__pyx_codeobj__83);
  Py_CLEAR(clear_module_state->__pyx_codeobj__84);
  Py_CLEAR(clear_module_state->__pyx_codeobj__85);
  Py_CLEAR(clear_module_state->__pyx_codeobj__86);
  Py_CLEAR(clear_module_state->__pyx_codeobj__87);
  Py_CLEAR(clear_module_state->__pyx_codeobj__88);
  Py_CLEAR(clear_module_state->__pyx_codeobj__89);
  Py_CLEAR(clear_module_state->__pyx_codeobj__90);
  Py_CLEAR(clear_module_state->__pyx_codeobj__91);
  Py_CLEAR(clear_module_state->__pyx_codeobj__92);
  Py_CLEAR(clear_module_state->__pyx_codeobj__93);
  Py_CLEAR(clear_module_state->__pyx_codeobj__94);
  Py_CLEAR(clear_module_state->__pyx_codeobj__95);
  Py_CLEAR(clear_module_state->__pyx_codeobj__96);
  Py_CLEAR(clear_module_state->__pyx_codeobj__97);
  Py_CLEAR(clear_module_state->__pyx_codeobj__98);
  Py_CLEAR(clear_module_state->__pyx_codeobj__99);
  Py_CLEAR(clear_module_state->__pyx_codeobj__100);
  Py_CLEAR(clear_module_state->__pyx_codeobj__101);
  Py_CLEAR(clear_module_state->__pyx_codeobj__102);
  Py_CLEAR(clear_module_state->__pyx_codeobj__103);
  Py_CLEAR(clear_module_state->__pyx_codeobj__104);
  Py_CLEAR(clear_module_state->__pyx_codeobj__105);
  Py_CLEAR(clear_module_state->__pyx_codeobj__106);
  Py_CLEAR(clear_module_state->__pyx_codeobj__107);
  Py_CLEAR(clear_module_state->__pyx_codeobj__109);
  Py_CLEAR(clear_module_state->__pyx_codeobj__110);
  Py_CLEAR(clear_module_state->__pyx_codeobj__112);
  Py_CLEAR(clear_module_state->__pyx_codeobj__114);
  Py_CLEAR(clear_module_state->__pyx_codeobj__115);
  Py_CLEAR(clear_module_state->__pyx_codeobj__117);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__StructSchemaField);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__StringArrayPtr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp_SchemaParser);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicOrphan);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicStructBuilder);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicEnumField);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__Schema);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__InterfaceSchema);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicEnum);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__DynamicListBuilder);
  Py_VISIT(traverse_module_state->__pyx_ptype_5capnp_3lib_5capnp__MessageBuilder);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Int64_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_String_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_ptype___Pyx_EnumMeta);
  Py_VISIT(traverse_module_state->__Pyx_EnumMeta);
  Py_VISIT(traverse_module_state->__pyx_n_s_ALIGNMENT);
  Py_VISIT(traverse_module_state->__pyx_n_s_APP);
  Py_VISIT(traverse_module_state->__pyx_n_s_APP_ARG);
  Py_VISIT(traverse_module_state->__pyx_n_s_APP_FUN);
  Py_VISIT(traverse_module_state->__pyx_n_s_AbstractTactic);
  Py_VISIT(traverse_module_state->__pyx_n_s_AbstractTactic_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_AbstractTactic_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_AbstractTactic_Reader_List___red);
  Py_VISIT(traverse_module_state->__pyx_n_s_AbstractTactic_Reader_List___set);
  Py_VISIT(traverse_module_state->__pyx_n_s_AbstractTactic_Reader___reduce_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_AbstractTactic_Reader___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument_Reader_List___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument_Reader_List___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument_Reader___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument_Term_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument_Term_Reader___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument_Term_Reader___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_Argument_Which);
  Py_VISIT(traverse_module_state->__pyx_n_s_CASE);
  Py_VISIT(traverse_module_state->__pyx_n_s_CASE_BRANCH);
  Py_VISIT(traverse_module_state->__pyx_n_s_CASE_BRANCH_POINTER);
  Py_VISIT(traverse_module_state->__pyx_n_s_CASE_IND);
  Py_VISIT(traverse_module_state->__pyx_n_s_CASE_RETURN);
  Py_VISIT(traverse_module_state->__pyx_n_s_CASE_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_CAST);
  Py_VISIT(traverse_module_state->__pyx_n_s_CAST_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_CAST_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_CHECK_ALIGNMENT);
  Py_VISIT(traverse_module_state->__pyx_n_s_COMPLETE);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONSTRUCTOR);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONSTRUCT_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONST_DEF);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONST_EMPTY);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONST_OPAQUE_DEF);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONST_PRIMITIVE);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONST_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONST_UNDEF);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONTEXT_ASSUM);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONTEXT_DEF);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONTEXT_DEF_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONTEXT_DEF_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONTEXT_ELEM);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONTEXT_SUBJECT);
  Py_VISIT(traverse_module_state->__pyx_n_s_CO_FIX);
  Py_VISIT(traverse_module_state->__pyx_n_s_CO_FIX_FUN);
  Py_VISIT(traverse_module_state->__pyx_n_s_CO_FIX_FUN_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_CO_FIX_FUN_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_CO_FIX_MUTUAL);
  Py_VISIT(traverse_module_state->__pyx_n_s_CO_FIX_RETURN);
  Py_VISIT(traverse_module_state->__pyx_n_s_C_B_CONSTRUCT);
  Py_VISIT(traverse_module_state->__pyx_n_s_C_B_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConflatableEdges);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConflatableEdges_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConflatableEdges_Reader___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConflatableEdges_Reader___setsta);
  Py_VISIT(traverse_module_state->__pyx_n_s_DEFINITION);
  Py_VISIT(traverse_module_state->__pyx_n_s_DISCHARGED);
  Py_VISIT(traverse_module_state->__pyx_n_s_DataVersion);
  Py_VISIT(traverse_module_state->__pyx_n_s_DataVersion_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_DataVersion_Reader___reduce_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_DataVersion_Reader___setstate_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Dataset);
  Py_VISIT(traverse_module_state->__pyx_n_s_Dataset_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Dataset_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Dataset_Reader___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Reader___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Reader___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Status_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Status_Reader___reduc);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Status_Reader___setst);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Status_Substituted_Re);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Status_Substituted_Re_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Status_Substituted_Re_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Status_Which);
  Py_VISIT(traverse_module_state->__pyx_n_s_Definition_Which);
  Py_VISIT(traverse_module_state->__pyx_n_s_EVAR);
  Py_VISIT(traverse_module_state->__pyx_n_s_EVAR_SUBJECT);
  Py_VISIT(traverse_module_state->__pyx_n_s_EVAR_SUBST);
  Py_VISIT(traverse_module_state->__pyx_n_s_EVAR_SUBST_POINTER);
  Py_VISIT(traverse_module_state->__pyx_n_s_EVAR_SUBST_TARGET);
  Py_VISIT(traverse_module_state->__pyx_n_s_EVAR_SUBST_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_EdgeClassification);
  Py_VISIT(traverse_module_state->__pyx_n_s_EdgeTarget);
  Py_VISIT(traverse_module_state->__pyx_n_s_EnumBase);
  Py_VISIT(traverse_module_state->__pyx_n_s_Exception);
  Py_VISIT(traverse_module_state->__pyx_n_s_Exception_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Exception_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Exception_Reader___setstate_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_Exception_Which);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecutionResult);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecutionResult_NewState_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecutionResult_NewState_Reader_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecutionResult_NewState_Reader_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecutionResult_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecutionResult_Reader___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecutionResult_Reader___setstat);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecutionResult_Which);
  Py_VISIT(traverse_module_state->__pyx_n_s_FAILURE);
  Py_VISIT(traverse_module_state->__pyx_n_s_FIX);
  Py_VISIT(traverse_module_state->__pyx_n_s_FIX_FUN);
  Py_VISIT(traverse_module_state->__pyx_n_s_FIX_FUN_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_FIX_FUN_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_FIX_MUTUAL);
  Py_VISIT(traverse_module_state->__pyx_n_s_FIX_RETURN);
  Py_VISIT(traverse_module_state->__pyx_n_s_FLOAT);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloatP);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloatP_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloatP_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloatP_Reader___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_EdgeTarget_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_EdgeTarget_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_EdgeTarget_Reader_List___r);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_EdgeTarget_Reader_List___s);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_EdgeTarget_Reader___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_EdgeTarget_Reader___setsta);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_EdgeTarget_Target_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_EdgeTarget_Target_Reader_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_EdgeTarget_Target_Reader_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Node_Label_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Node_Label_Reader___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Node_Label_Reader___setsta);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Node_Label_Which);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Node_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Node_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Node_Reader_List___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Node_Reader_List___setstat);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Node_Reader___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Node_Reader___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Graph_Reader___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ILLEGAL_ARGUMENT);
  Py_VISIT(traverse_module_state->__pyx_n_s_INDUCTIVE);
  Py_VISIT(traverse_module_state->__pyx_n_s_IND_CONSTRUCT);
  Py_VISIT(traverse_module_state->__pyx_n_s_IND_PROJECTION);
  Py_VISIT(traverse_module_state->__pyx_n_s_IND_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_INITIALIZE);
  Py_VISIT(traverse_module_state->__pyx_n_s_INITIALIZED);
  Py_VISIT(traverse_module_state->__pyx_n_s_INT);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Int64_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_Int64_List___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Int64_List___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntEnum);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntP);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntP_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntP_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntP_Reader___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_KNOWN);
  Py_VISIT(traverse_module_state->__pyx_n_s_LAMBDA);
  Py_VISIT(traverse_module_state->__pyx_n_s_LAMBDA_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_LAMBDA_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_LET_IN);
  Py_VISIT(traverse_module_state->__pyx_n_s_LET_IN_DEF);
  Py_VISIT(traverse_module_state->__pyx_n_s_LET_IN_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_LET_IN_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_MANUAL_CONSTANT);
  Py_VISIT(traverse_module_state->__pyx_n_s_MANUAL_SECTION_CONSTANT);
  Py_VISIT(traverse_module_state->__pyx_n_s_MISMATCHED_ARGUMENTS);
  Py_VISIT(traverse_module_state->__pyx_n_s_NEW_STATE);
  Py_VISIT(traverse_module_state->__pyx_n_s_NO_SUCH_TACTIC);
  Py_VISIT(traverse_module_state->__pyx_n_s_Node);
  Py_VISIT(traverse_module_state->__pyx_n_s_Node_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Node_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_Node_Reader_List___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Node_Reader_List___setstate_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_Node_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Node_Reader___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ORIGINAL);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderedDict);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Out_of_bounds);
  Py_VISIT(traverse_module_state->__pyx_n_s_Outcome);
  Py_VISIT(traverse_module_state->__pyx_n_s_Outcome_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Outcome_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_Outcome_Reader_List___reduce_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_Outcome_Reader_List___setstate_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_Outcome_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Outcome_Reader___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Outcome_Term_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Outcome_Term_Reader___reduce_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_Outcome_Term_Reader___setstate_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_PARSE_ERROR);
  Py_VISIT(traverse_module_state->__pyx_n_s_PREDICT);
  Py_VISIT(traverse_module_state->__pyx_n_s_PREDICTION);
  Py_VISIT(traverse_module_state->__pyx_n_s_PRIMITIVE);
  Py_VISIT(traverse_module_state->__pyx_n_s_PROD);
  Py_VISIT(traverse_module_state->__pyx_n_s_PROD_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_PROD_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_PROJECTION);
  Py_VISIT(traverse_module_state->__pyx_n_s_PROJ_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_PROOF_STATE);
  Py_VISIT(traverse_module_state->__pyx_n_s_PROTOCOL_ERROR);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Prediction);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re_5);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Prediction_Re_6);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Reader___redu);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Reader___sets);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Request_Initi);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Request_Initi_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Request_Initi_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Request_Predi);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Request_Predi_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Request_Predi_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Request_Reade);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Request_Reade_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Request_Reade_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Request_Which);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Response_Alig);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Response_Alig_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Response_Alig_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Response_Read);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Response_Read_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Response_Read_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_Response_Whic);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_TextPredictio);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_TextPredictio_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_TextPredictio_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_TextPredictio_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_TextPredictio_5);
  Py_VISIT(traverse_module_state->__pyx_n_s_PredictionProtocol_TextPredictio_6);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofState);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofState_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofState_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofState_Reader_List___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofState_Reader_List___setstat);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofState_Reader___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofState_Reader___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofState_Root_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofState_Root_Reader___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofState_Root_Reader___setstat);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep_Reader_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep_Reader_List___reduce_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep_Reader_List___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep_Reader___setstate_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep_Tactic_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep_Tactic_Reader___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep_Tactic_Reader___setsta);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProofStep_Tactic_Which);
  Py_VISIT(traverse_module_state->__pyx_n_s_Pyx_EnumBase);
  Py_VISIT(traverse_module_state->__pyx_n_s_Pyx_EnumBase___new);
  Py_VISIT(traverse_module_state->__pyx_n_s_Pyx_EnumBase___repr);
  Py_VISIT(traverse_module_state->__pyx_n_s_Pyx_EnumBase___str);
  Py_VISIT(traverse_module_state->__pyx_n_s_Pyx_EnumMeta___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Pyx_EnumMeta___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_REL);
  Py_VISIT(traverse_module_state->__pyx_n_s_REL_POINTER);
  Py_VISIT(traverse_module_state->__pyx_n_s_Request);
  Py_VISIT(traverse_module_state->__pyx_n_s_Response);
  Py_VISIT(traverse_module_state->__pyx_n_s_SORT_PROP);
  Py_VISIT(traverse_module_state->__pyx_n_s_SORT_SET);
  Py_VISIT(traverse_module_state->__pyx_n_s_SORT_S_PROP);
  Py_VISIT(traverse_module_state->__pyx_n_s_SORT_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_SUBSTITUTED);
  Py_VISIT(traverse_module_state->__pyx_n_s_SYNCHRONIZE);
  Py_VISIT(traverse_module_state->__pyx_n_s_SYNCHRONIZED);
  Py_VISIT(traverse_module_state->__pyx_n_s_String_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_String_List___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_String_List___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TACTICAL_CONSTANT);
  Py_VISIT(traverse_module_state->__pyx_n_s_TACTICAL_SECTION_CONSTANT);
  Py_VISIT(traverse_module_state->__pyx_n_s_TERM);
  Py_VISIT(traverse_module_state->__pyx_n_s_TEXT_PREDICTION);
  Py_VISIT(traverse_module_state->__pyx_n_s_Tactic);
  Py_VISIT(traverse_module_state->__pyx_n_s_Tactic_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Tactic_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Tactic_Reader___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TextPrediction);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_UNKNOWN);
  Py_VISIT(traverse_module_state->__pyx_n_s_UNRESOLVABLE);
  Py_VISIT(traverse_module_state->__pyx_n_s_Uint16_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_Uint16_List___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Uint16_List___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Uint32_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_Uint32_List___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Uint32_List___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Uint64_List);
  Py_VISIT(traverse_module_state->__pyx_n_s_Uint64_List___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Uint64_List___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Unknown_enum_value_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_s__118);
  Py_VISIT(traverse_module_state->__pyx_kp_u__3);
  Py_VISIT(traverse_module_state->__pyx_n_s__4);
  Py_VISIT(traverse_module_state->__pyx_n_s_as_struct);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_capnp);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_cls);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_n_s_dct);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_doc);
  Py_VISIT(traverse_module_state->__pyx_n_s_dyn);
  Py_VISIT(traverse_module_state->__pyx_n_s_dynamic);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_enum);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_graph_api_capnp);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_init);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_subclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_members);
  Py_VISIT(traverse_module_state->__pyx_n_s_metaclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_module);
  Py_VISIT(traverse_module_state->__pyx_n_s_mro_entries);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_n_s_os);
  Py_VISIT(traverse_module_state->__pyx_n_s_parents);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_prepare);
  Py_VISIT(traverse_module_state->__pyx_n_s_pytact);
  Py_VISIT(traverse_module_state->__pyx_n_s_pytact_graph_api_capnp);
  Py_VISIT(traverse_module_state->__pyx_n_s_pytact_graph_api_capnp_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle___Pyx_EnumMeta);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_qualname);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_repr);
  Py_VISIT(traverse_module_state->__pyx_n_s_res);
  Py_VISIT(traverse_module_state->__pyx_kp_s_s_s);
  Py_VISIT(traverse_module_state->__pyx_kp_s_s_s_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self_source_cannot_be_converted);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_v);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_values);
  Py_VISIT(traverse_module_state->__pyx_int_222419149);
  Py_VISIT(traverse_module_state->__pyx_int_228825662);
  Py_VISIT(traverse_module_state->__pyx_int_238750788);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_tuple__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__5);
  Py_VISIT(traverse_module_state->__pyx_tuple__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__8);
  Py_VISIT(traverse_module_state->__pyx_codeobj__7);
  Py_VISIT(traverse_module_state->__pyx_codeobj__9);
  Py_VISIT(traverse_module_state->__pyx_tuple__108);
  Py_VISIT(traverse_module_state->__pyx_tuple__111);
  Py_VISIT(traverse_module_state->__pyx_tuple__113);
  Py_VISIT(traverse_module_state->__pyx_tuple__116);
  Py_VISIT(traverse_module_state->__pyx_codeobj__10);
  Py_VISIT(traverse_module_state->__pyx_codeobj__11);
  Py_VISIT(traverse_module_state->__pyx_codeobj__12);
  Py_VISIT(traverse_module_state->__pyx_codeobj__13);
  Py_VISIT(traverse_module_state->__pyx_codeobj__14);
  Py_VISIT(traverse_module_state->__pyx_codeobj__15);
  Py_VISIT(traverse_module_state->__pyx_codeobj__16);
  Py_VISIT(traverse_module_state->__pyx_codeobj__17);
  Py_VISIT(traverse_module_state->__pyx_codeobj__18);
  Py_VISIT(traverse_module_state->__pyx_codeobj__19);
  Py_VISIT(traverse_module_state->__pyx_codeobj__20);
  Py_VISIT(traverse_module_state->__pyx_codeobj__21);
  Py_VISIT(traverse_module_state->__pyx_codeobj__22);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__24);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__26);
  Py_VISIT(traverse_module_state->__pyx_codeobj__27);
  Py_VISIT(traverse_module_state->__pyx_codeobj__28);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__30);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__34);
  Py_VISIT(traverse_module_state->__pyx_codeobj__35);
  Py_VISIT(traverse_module_state->__pyx_codeobj__36);
  Py_VISIT(traverse_module_state->__pyx_codeobj__37);
  Py_VISIT(traverse_module_state->__pyx_codeobj__38);
  Py_VISIT(traverse_module_state->__pyx_codeobj__39);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__45);
  Py_VISIT(traverse_module_state->__pyx_codeobj__46);
  Py_VISIT(traverse_module_state->__pyx_codeobj__47);
  Py_VISIT(traverse_module_state->__pyx_codeobj__48);
  Py_VISIT(traverse_module_state->__pyx_codeobj__49);
  Py_VISIT(traverse_module_state->__pyx_codeobj__50);
  Py_VISIT(traverse_module_state->__pyx_codeobj__51);
  Py_VISIT(traverse_module_state->__pyx_codeobj__52);
  Py_VISIT(traverse_module_state->__pyx_codeobj__53);
  Py_VISIT(traverse_module_state->__pyx_codeobj__54);
  Py_VISIT(traverse_module_state->__pyx_codeobj__55);
  Py_VISIT(traverse_module_state->__pyx_codeobj__56);
  Py_VISIT(traverse_module_state->__pyx_codeobj__57);
  Py_VISIT(traverse_module_state->__pyx_codeobj__58);
  Py_VISIT(traverse_module_state->__pyx_codeobj__59);
  Py_VISIT(traverse_module_state->__pyx_codeobj__60);
  Py_VISIT(traverse_module_state->__pyx_codeobj__61);
  Py_VISIT(traverse_module_state->__pyx_codeobj__62);
  Py_VISIT(traverse_module_state->__pyx_codeobj__63);
  Py_VISIT(traverse_module_state->__pyx_codeobj__64);
  Py_VISIT(traverse_module_state->__pyx_codeobj__65);
  Py_VISIT(traverse_module_state->__pyx_codeobj__66);
  Py_VISIT(traverse_module_state->__pyx_codeobj__67);
  Py_VISIT(traverse_module_state->__pyx_codeobj__68);
  Py_VISIT(traverse_module_state->__pyx_codeobj__69);
  Py_VISIT(traverse_module_state->__pyx_codeobj__70);
  Py_VISIT(traverse_module_state->__pyx_codeobj__71);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__73);
  Py_VISIT(traverse_module_state->__pyx_codeobj__74);
  Py_VISIT(traverse_module_state->__pyx_codeobj__75);
  Py_VISIT(traverse_module_state->__pyx_codeobj__76);
  Py_VISIT(traverse_module_state->__pyx_codeobj__77);
  Py_VISIT(traverse_module_state->__pyx_codeobj__78);
  Py_VISIT(traverse_module_state->__pyx_codeobj__79);
  Py_VISIT(traverse_module_state->__pyx_codeobj__80);
  Py_VISIT(traverse_module_state->__pyx_codeobj__81);
  Py_VISIT(traverse_module_state->__pyx_codeobj__82);
  Py_VISIT(traverse_module_state->__pyx_codeobj__83);
  Py_VISIT(traverse_module_state->__pyx_codeobj__84);
  Py_VISIT(traverse_module_state->__pyx_codeobj__85);
  Py_VISIT(traverse_module_state->__pyx_codeobj__86);
  Py_VISIT(traverse_module_state->__pyx_codeobj__87);
  Py_VISIT(traverse_module_state->__pyx_codeobj__88);
  Py_VISIT(traverse_module_state->__pyx_codeobj__89);
  Py_VISIT(traverse_module_state->__pyx_codeobj__90);
  Py_VISIT(traverse_module_state->__pyx_codeobj__91);
  Py_VISIT(traverse_module_state->__pyx_codeobj__92);
  Py_VISIT(traverse_module_state->__pyx_codeobj__93);
  Py_VISIT(traverse_module_state->__pyx_codeobj__94);
  Py_VISIT(traverse_module_state->__pyx_codeobj__95);
  Py_VISIT(traverse_module_state->__pyx_codeobj__96);
  Py_VISIT(traverse_module_state->__pyx_codeobj__97);
  Py_VISIT(traverse_module_state->__pyx_codeobj__98);
  Py_VISIT(traverse_module_state->__pyx_codeobj__99);
  Py_VISIT(traverse_module_state->__pyx_codeobj__100);
  Py_VISIT(traverse_module_state->__pyx_codeobj__101);
  Py_VISIT(traverse_module_state->__pyx_codeobj__102);
  Py_VISIT(traverse_module_state->__pyx_codeobj__103);
  Py_VISIT(traverse_module_state->__pyx_codeobj__104);
  Py_VISIT(traverse_module_state->__pyx_codeobj__105);
  Py_VISIT(traverse_module_state->__pyx_codeobj__106);
  Py_VISIT(traverse_module_state->__pyx_codeobj__107);
  Py_VISIT(traverse_module_state->__pyx_codeobj__109);
  Py_VISIT(traverse_module_state->__pyx_codeobj__110);
  Py_VISIT(traverse_module_state->__pyx_codeobj__112);
  Py_VISIT(traverse_module_state->__pyx_codeobj__114);
  Py_VISIT(traverse_module_state->__pyx_codeobj__115);
  Py_VISIT(traverse_module_state->__pyx_codeobj__117);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#if CYTHON_USE_MODULE_STATE
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#define __pyx_ptype_5capnp_3lib_5capnp__StructSchemaField __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__StructSchemaField
#define __pyx_ptype_5capnp_3lib_5capnp__StringArrayPtr __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__StringArrayPtr
#define __pyx_ptype_5capnp_3lib_5capnp_SchemaParser __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp_SchemaParser
#define __pyx_ptype_5capnp_3lib_5capnp__DynamicOrphan __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__DynamicOrphan
#define __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader
#define __pyx_ptype_5capnp_3lib_5capnp__DynamicStructBuilder __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__DynamicStructBuilder
#define __pyx_ptype_5capnp_3lib_5capnp__DynamicEnumField __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__DynamicEnumField
#define __pyx_ptype_5capnp_3lib_5capnp__Schema __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__Schema
#define __pyx_ptype_5capnp_3lib_5capnp__InterfaceSchema __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__InterfaceSchema
#define __pyx_ptype_5capnp_3lib_5capnp__DynamicEnum __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__DynamicEnum
#define __pyx_ptype_5capnp_3lib_5capnp__DynamicListBuilder __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__DynamicListBuilder
#define __pyx_ptype_5capnp_3lib_5capnp__MessageBuilder __pyx_mstate_global->__pyx_ptype_5capnp_3lib_5capnp__MessageBuilder
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader
#define __pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_Int64_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Int64_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_String_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_String_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_String_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List
#define __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List __pyx_mstate_global->__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List
#define __pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List __pyx_mstate_global->__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List
#define __pyx_ptype___Pyx_EnumMeta __pyx_mstate_global->__pyx_ptype___Pyx_EnumMeta
#define __Pyx_EnumMeta __pyx_mstate_global->__Pyx_EnumMeta
#define __pyx_n_s_ALIGNMENT __pyx_mstate_global->__pyx_n_s_ALIGNMENT
#define __pyx_n_s_APP __pyx_mstate_global->__pyx_n_s_APP
#define __pyx_n_s_APP_ARG __pyx_mstate_global->__pyx_n_s_APP_ARG
#define __pyx_n_s_APP_FUN __pyx_mstate_global->__pyx_n_s_APP_FUN
#define __pyx_n_s_AbstractTactic __pyx_mstate_global->__pyx_n_s_AbstractTactic
#define __pyx_n_s_AbstractTactic_Reader __pyx_mstate_global->__pyx_n_s_AbstractTactic_Reader
#define __pyx_n_s_AbstractTactic_Reader_List __pyx_mstate_global->__pyx_n_s_AbstractTactic_Reader_List
#define __pyx_n_s_AbstractTactic_Reader_List___red __pyx_mstate_global->__pyx_n_s_AbstractTactic_Reader_List___red
#define __pyx_n_s_AbstractTactic_Reader_List___set __pyx_mstate_global->__pyx_n_s_AbstractTactic_Reader_List___set
#define __pyx_n_s_AbstractTactic_Reader___reduce_c __pyx_mstate_global->__pyx_n_s_AbstractTactic_Reader___reduce_c
#define __pyx_n_s_AbstractTactic_Reader___setstate __pyx_mstate_global->__pyx_n_s_AbstractTactic_Reader___setstate
#define __pyx_n_s_Argument __pyx_mstate_global->__pyx_n_s_Argument
#define __pyx_n_s_Argument_Reader __pyx_mstate_global->__pyx_n_s_Argument_Reader
#define __pyx_n_s_Argument_Reader_List __pyx_mstate_global->__pyx_n_s_Argument_Reader_List
#define __pyx_n_s_Argument_Reader_List___reduce_cy __pyx_mstate_global->__pyx_n_s_Argument_Reader_List___reduce_cy
#define __pyx_n_s_Argument_Reader_List___setstate __pyx_mstate_global->__pyx_n_s_Argument_Reader_List___setstate
#define __pyx_n_s_Argument_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_Argument_Reader___reduce_cython
#define __pyx_n_s_Argument_Reader___setstate_cytho __pyx_mstate_global->__pyx_n_s_Argument_Reader___setstate_cytho
#define __pyx_n_s_Argument_Term_Reader __pyx_mstate_global->__pyx_n_s_Argument_Term_Reader
#define __pyx_n_s_Argument_Term_Reader___reduce_cy __pyx_mstate_global->__pyx_n_s_Argument_Term_Reader___reduce_cy
#define __pyx_n_s_Argument_Term_Reader___setstate __pyx_mstate_global->__pyx_n_s_Argument_Term_Reader___setstate
#define __pyx_n_s_Argument_Which __pyx_mstate_global->__pyx_n_s_Argument_Which
#define __pyx_n_s_CASE __pyx_mstate_global->__pyx_n_s_CASE
#define __pyx_n_s_CASE_BRANCH __pyx_mstate_global->__pyx_n_s_CASE_BRANCH
#define __pyx_n_s_CASE_BRANCH_POINTER __pyx_mstate_global->__pyx_n_s_CASE_BRANCH_POINTER
#define __pyx_n_s_CASE_IND __pyx_mstate_global->__pyx_n_s_CASE_IND
#define __pyx_n_s_CASE_RETURN __pyx_mstate_global->__pyx_n_s_CASE_RETURN
#define __pyx_n_s_CASE_TERM __pyx_mstate_global->__pyx_n_s_CASE_TERM
#define __pyx_n_s_CAST __pyx_mstate_global->__pyx_n_s_CAST
#define __pyx_n_s_CAST_TERM __pyx_mstate_global->__pyx_n_s_CAST_TERM
#define __pyx_n_s_CAST_TYPE __pyx_mstate_global->__pyx_n_s_CAST_TYPE
#define __pyx_n_s_CHECK_ALIGNMENT __pyx_mstate_global->__pyx_n_s_CHECK_ALIGNMENT
#define __pyx_n_s_COMPLETE __pyx_mstate_global->__pyx_n_s_COMPLETE
#define __pyx_n_s_CONSTRUCTOR __pyx_mstate_global->__pyx_n_s_CONSTRUCTOR
#define __pyx_n_s_CONSTRUCT_TERM __pyx_mstate_global->__pyx_n_s_CONSTRUCT_TERM
#define __pyx_n_s_CONST_DEF __pyx_mstate_global->__pyx_n_s_CONST_DEF
#define __pyx_n_s_CONST_EMPTY __pyx_mstate_global->__pyx_n_s_CONST_EMPTY
#define __pyx_n_s_CONST_OPAQUE_DEF __pyx_mstate_global->__pyx_n_s_CONST_OPAQUE_DEF
#define __pyx_n_s_CONST_PRIMITIVE __pyx_mstate_global->__pyx_n_s_CONST_PRIMITIVE
#define __pyx_n_s_CONST_TYPE __pyx_mstate_global->__pyx_n_s_CONST_TYPE
#define __pyx_n_s_CONST_UNDEF __pyx_mstate_global->__pyx_n_s_CONST_UNDEF
#define __pyx_n_s_CONTEXT_ASSUM __pyx_mstate_global->__pyx_n_s_CONTEXT_ASSUM
#define __pyx_n_s_CONTEXT_DEF __pyx_mstate_global->__pyx_n_s_CONTEXT_DEF
#define __pyx_n_s_CONTEXT_DEF_TERM __pyx_mstate_global->__pyx_n_s_CONTEXT_DEF_TERM
#define __pyx_n_s_CONTEXT_DEF_TYPE __pyx_mstate_global->__pyx_n_s_CONTEXT_DEF_TYPE
#define __pyx_n_s_CONTEXT_ELEM __pyx_mstate_global->__pyx_n_s_CONTEXT_ELEM
#define __pyx_n_s_CONTEXT_SUBJECT __pyx_mstate_global->__pyx_n_s_CONTEXT_SUBJECT
#define __pyx_n_s_CO_FIX __pyx_mstate_global->__pyx_n_s_CO_FIX
#define __pyx_n_s_CO_FIX_FUN __pyx_mstate_global->__pyx_n_s_CO_FIX_FUN
#define __pyx_n_s_CO_FIX_FUN_TERM __pyx_mstate_global->__pyx_n_s_CO_FIX_FUN_TERM
#define __pyx_n_s_CO_FIX_FUN_TYPE __pyx_mstate_global->__pyx_n_s_CO_FIX_FUN_TYPE
#define __pyx_n_s_CO_FIX_MUTUAL __pyx_mstate_global->__pyx_n_s_CO_FIX_MUTUAL
#define __pyx_n_s_CO_FIX_RETURN __pyx_mstate_global->__pyx_n_s_CO_FIX_RETURN
#define __pyx_n_s_C_B_CONSTRUCT __pyx_mstate_global->__pyx_n_s_C_B_CONSTRUCT
#define __pyx_n_s_C_B_TERM __pyx_mstate_global->__pyx_n_s_C_B_TERM
#define __pyx_n_s_ConflatableEdges __pyx_mstate_global->__pyx_n_s_ConflatableEdges
#define __pyx_n_s_ConflatableEdges_Reader __pyx_mstate_global->__pyx_n_s_ConflatableEdges_Reader
#define __pyx_n_s_ConflatableEdges_Reader___reduce __pyx_mstate_global->__pyx_n_s_ConflatableEdges_Reader___reduce
#define __pyx_n_s_ConflatableEdges_Reader___setsta __pyx_mstate_global->__pyx_n_s_ConflatableEdges_Reader___setsta
#define __pyx_n_s_DEFINITION __pyx_mstate_global->__pyx_n_s_DEFINITION
#define __pyx_n_s_DISCHARGED __pyx_mstate_global->__pyx_n_s_DISCHARGED
#define __pyx_n_s_DataVersion __pyx_mstate_global->__pyx_n_s_DataVersion
#define __pyx_n_s_DataVersion_Reader __pyx_mstate_global->__pyx_n_s_DataVersion_Reader
#define __pyx_n_s_DataVersion_Reader___reduce_cyth __pyx_mstate_global->__pyx_n_s_DataVersion_Reader___reduce_cyth
#define __pyx_n_s_DataVersion_Reader___setstate_cy __pyx_mstate_global->__pyx_n_s_DataVersion_Reader___setstate_cy
#define __pyx_n_s_Dataset __pyx_mstate_global->__pyx_n_s_Dataset
#define __pyx_n_s_Dataset_Reader __pyx_mstate_global->__pyx_n_s_Dataset_Reader
#define __pyx_n_s_Dataset_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_Dataset_Reader___reduce_cython
#define __pyx_n_s_Dataset_Reader___setstate_cython __pyx_mstate_global->__pyx_n_s_Dataset_Reader___setstate_cython
#define __pyx_n_s_Definition __pyx_mstate_global->__pyx_n_s_Definition
#define __pyx_n_s_Definition_Reader __pyx_mstate_global->__pyx_n_s_Definition_Reader
#define __pyx_n_s_Definition_Reader___reduce_cytho __pyx_mstate_global->__pyx_n_s_Definition_Reader___reduce_cytho
#define __pyx_n_s_Definition_Reader___setstate_cyt __pyx_mstate_global->__pyx_n_s_Definition_Reader___setstate_cyt
#define __pyx_n_s_Definition_Status_Reader __pyx_mstate_global->__pyx_n_s_Definition_Status_Reader
#define __pyx_n_s_Definition_Status_Reader___reduc __pyx_mstate_global->__pyx_n_s_Definition_Status_Reader___reduc
#define __pyx_n_s_Definition_Status_Reader___setst __pyx_mstate_global->__pyx_n_s_Definition_Status_Reader___setst
#define __pyx_n_s_Definition_Status_Substituted_Re __pyx_mstate_global->__pyx_n_s_Definition_Status_Substituted_Re
#define __pyx_n_s_Definition_Status_Substituted_Re_2 __pyx_mstate_global->__pyx_n_s_Definition_Status_Substituted_Re_2
#define __pyx_n_s_Definition_Status_Substituted_Re_3 __pyx_mstate_global->__pyx_n_s_Definition_Status_Substituted_Re_3
#define __pyx_n_s_Definition_Status_Which __pyx_mstate_global->__pyx_n_s_Definition_Status_Which
#define __pyx_n_s_Definition_Which __pyx_mstate_global->__pyx_n_s_Definition_Which
#define __pyx_n_s_EVAR __pyx_mstate_global->__pyx_n_s_EVAR
#define __pyx_n_s_EVAR_SUBJECT __pyx_mstate_global->__pyx_n_s_EVAR_SUBJECT
#define __pyx_n_s_EVAR_SUBST __pyx_mstate_global->__pyx_n_s_EVAR_SUBST
#define __pyx_n_s_EVAR_SUBST_POINTER __pyx_mstate_global->__pyx_n_s_EVAR_SUBST_POINTER
#define __pyx_n_s_EVAR_SUBST_TARGET __pyx_mstate_global->__pyx_n_s_EVAR_SUBST_TARGET
#define __pyx_n_s_EVAR_SUBST_TERM __pyx_mstate_global->__pyx_n_s_EVAR_SUBST_TERM
#define __pyx_n_s_EdgeClassification __pyx_mstate_global->__pyx_n_s_EdgeClassification
#define __pyx_n_s_EdgeTarget __pyx_mstate_global->__pyx_n_s_EdgeTarget
#define __pyx_n_s_EnumBase __pyx_mstate_global->__pyx_n_s_EnumBase
#define __pyx_n_s_Exception __pyx_mstate_global->__pyx_n_s_Exception
#define __pyx_n_s_Exception_Reader __pyx_mstate_global->__pyx_n_s_Exception_Reader
#define __pyx_n_s_Exception_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_Exception_Reader___reduce_cython
#define __pyx_n_s_Exception_Reader___setstate_cyth __pyx_mstate_global->__pyx_n_s_Exception_Reader___setstate_cyth
#define __pyx_n_s_Exception_Which __pyx_mstate_global->__pyx_n_s_Exception_Which
#define __pyx_n_s_ExecutionResult __pyx_mstate_global->__pyx_n_s_ExecutionResult
#define __pyx_n_s_ExecutionResult_NewState_Reader __pyx_mstate_global->__pyx_n_s_ExecutionResult_NewState_Reader
#define __pyx_n_s_ExecutionResult_NewState_Reader_2 __pyx_mstate_global->__pyx_n_s_ExecutionResult_NewState_Reader_2
#define __pyx_n_s_ExecutionResult_NewState_Reader_3 __pyx_mstate_global->__pyx_n_s_ExecutionResult_NewState_Reader_3
#define __pyx_n_s_ExecutionResult_Reader __pyx_mstate_global->__pyx_n_s_ExecutionResult_Reader
#define __pyx_n_s_ExecutionResult_Reader___reduce __pyx_mstate_global->__pyx_n_s_ExecutionResult_Reader___reduce
#define __pyx_n_s_ExecutionResult_Reader___setstat __pyx_mstate_global->__pyx_n_s_ExecutionResult_Reader___setstat
#define __pyx_n_s_ExecutionResult_Which __pyx_mstate_global->__pyx_n_s_ExecutionResult_Which
#define __pyx_n_s_FAILURE __pyx_mstate_global->__pyx_n_s_FAILURE
#define __pyx_n_s_FIX __pyx_mstate_global->__pyx_n_s_FIX
#define __pyx_n_s_FIX_FUN __pyx_mstate_global->__pyx_n_s_FIX_FUN
#define __pyx_n_s_FIX_FUN_TERM __pyx_mstate_global->__pyx_n_s_FIX_FUN_TERM
#define __pyx_n_s_FIX_FUN_TYPE __pyx_mstate_global->__pyx_n_s_FIX_FUN_TYPE
#define __pyx_n_s_FIX_MUTUAL __pyx_mstate_global->__pyx_n_s_FIX_MUTUAL
#define __pyx_n_s_FIX_RETURN __pyx_mstate_global->__pyx_n_s_FIX_RETURN
#define __pyx_n_s_FLOAT __pyx_mstate_global->__pyx_n_s_FLOAT
#define __pyx_n_s_FloatP __pyx_mstate_global->__pyx_n_s_FloatP
#define __pyx_n_s_FloatP_Reader __pyx_mstate_global->__pyx_n_s_FloatP_Reader
#define __pyx_n_s_FloatP_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_FloatP_Reader___reduce_cython
#define __pyx_n_s_FloatP_Reader___setstate_cython __pyx_mstate_global->__pyx_n_s_FloatP_Reader___setstate_cython
#define __pyx_n_s_Graph __pyx_mstate_global->__pyx_n_s_Graph
#define __pyx_n_s_Graph_EdgeTarget_Reader __pyx_mstate_global->__pyx_n_s_Graph_EdgeTarget_Reader
#define __pyx_n_s_Graph_EdgeTarget_Reader_List __pyx_mstate_global->__pyx_n_s_Graph_EdgeTarget_Reader_List
#define __pyx_n_s_Graph_EdgeTarget_Reader_List___r __pyx_mstate_global->__pyx_n_s_Graph_EdgeTarget_Reader_List___r
#define __pyx_n_s_Graph_EdgeTarget_Reader_List___s __pyx_mstate_global->__pyx_n_s_Graph_EdgeTarget_Reader_List___s
#define __pyx_n_s_Graph_EdgeTarget_Reader___reduce __pyx_mstate_global->__pyx_n_s_Graph_EdgeTarget_Reader___reduce
#define __pyx_n_s_Graph_EdgeTarget_Reader___setsta __pyx_mstate_global->__pyx_n_s_Graph_EdgeTarget_Reader___setsta
#define __pyx_n_s_Graph_EdgeTarget_Target_Reader __pyx_mstate_global->__pyx_n_s_Graph_EdgeTarget_Target_Reader
#define __pyx_n_s_Graph_EdgeTarget_Target_Reader_2 __pyx_mstate_global->__pyx_n_s_Graph_EdgeTarget_Target_Reader_2
#define __pyx_n_s_Graph_EdgeTarget_Target_Reader_3 __pyx_mstate_global->__pyx_n_s_Graph_EdgeTarget_Target_Reader_3
#define __pyx_n_s_Graph_Node_Label_Reader __pyx_mstate_global->__pyx_n_s_Graph_Node_Label_Reader
#define __pyx_n_s_Graph_Node_Label_Reader___reduce __pyx_mstate_global->__pyx_n_s_Graph_Node_Label_Reader___reduce
#define __pyx_n_s_Graph_Node_Label_Reader___setsta __pyx_mstate_global->__pyx_n_s_Graph_Node_Label_Reader___setsta
#define __pyx_n_s_Graph_Node_Label_Which __pyx_mstate_global->__pyx_n_s_Graph_Node_Label_Which
#define __pyx_n_s_Graph_Node_Reader __pyx_mstate_global->__pyx_n_s_Graph_Node_Reader
#define __pyx_n_s_Graph_Node_Reader_List __pyx_mstate_global->__pyx_n_s_Graph_Node_Reader_List
#define __pyx_n_s_Graph_Node_Reader_List___reduce __pyx_mstate_global->__pyx_n_s_Graph_Node_Reader_List___reduce
#define __pyx_n_s_Graph_Node_Reader_List___setstat __pyx_mstate_global->__pyx_n_s_Graph_Node_Reader_List___setstat
#define __pyx_n_s_Graph_Node_Reader___reduce_cytho __pyx_mstate_global->__pyx_n_s_Graph_Node_Reader___reduce_cytho
#define __pyx_n_s_Graph_Node_Reader___setstate_cyt __pyx_mstate_global->__pyx_n_s_Graph_Node_Reader___setstate_cyt
#define __pyx_n_s_Graph_Reader __pyx_mstate_global->__pyx_n_s_Graph_Reader
#define __pyx_n_s_Graph_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_Graph_Reader___reduce_cython
#define __pyx_n_s_Graph_Reader___setstate_cython __pyx_mstate_global->__pyx_n_s_Graph_Reader___setstate_cython
#define __pyx_n_s_ILLEGAL_ARGUMENT __pyx_mstate_global->__pyx_n_s_ILLEGAL_ARGUMENT
#define __pyx_n_s_INDUCTIVE __pyx_mstate_global->__pyx_n_s_INDUCTIVE
#define __pyx_n_s_IND_CONSTRUCT __pyx_mstate_global->__pyx_n_s_IND_CONSTRUCT
#define __pyx_n_s_IND_PROJECTION __pyx_mstate_global->__pyx_n_s_IND_PROJECTION
#define __pyx_n_s_IND_TYPE __pyx_mstate_global->__pyx_n_s_IND_TYPE
#define __pyx_n_s_INITIALIZE __pyx_mstate_global->__pyx_n_s_INITIALIZE
#define __pyx_n_s_INITIALIZED __pyx_mstate_global->__pyx_n_s_INITIALIZED
#define __pyx_n_s_INT __pyx_mstate_global->__pyx_n_s_INT
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_n_s_IndexError __pyx_mstate_global->__pyx_n_s_IndexError
#define __pyx_n_s_Int64_List __pyx_mstate_global->__pyx_n_s_Int64_List
#define __pyx_n_s_Int64_List___reduce_cython __pyx_mstate_global->__pyx_n_s_Int64_List___reduce_cython
#define __pyx_n_s_Int64_List___setstate_cython __pyx_mstate_global->__pyx_n_s_Int64_List___setstate_cython
#define __pyx_n_s_IntEnum __pyx_mstate_global->__pyx_n_s_IntEnum
#define __pyx_n_s_IntP __pyx_mstate_global->__pyx_n_s_IntP
#define __pyx_n_s_IntP_Reader __pyx_mstate_global->__pyx_n_s_IntP_Reader
#define __pyx_n_s_IntP_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_IntP_Reader___reduce_cython
#define __pyx_n_s_IntP_Reader___setstate_cython __pyx_mstate_global->__pyx_n_s_IntP_Reader___setstate_cython
#define __pyx_n_s_KNOWN __pyx_mstate_global->__pyx_n_s_KNOWN
#define __pyx_n_s_LAMBDA __pyx_mstate_global->__pyx_n_s_LAMBDA
#define __pyx_n_s_LAMBDA_TERM __pyx_mstate_global->__pyx_n_s_LAMBDA_TERM
#define __pyx_n_s_LAMBDA_TYPE __pyx_mstate_global->__pyx_n_s_LAMBDA_TYPE
#define __pyx_n_s_LET_IN __pyx_mstate_global->__pyx_n_s_LET_IN
#define __pyx_n_s_LET_IN_DEF __pyx_mstate_global->__pyx_n_s_LET_IN_DEF
#define __pyx_n_s_LET_IN_TERM __pyx_mstate_global->__pyx_n_s_LET_IN_TERM
#define __pyx_n_s_LET_IN_TYPE __pyx_mstate_global->__pyx_n_s_LET_IN_TYPE
#define __pyx_n_s_MANUAL_CONSTANT __pyx_mstate_global->__pyx_n_s_MANUAL_CONSTANT
#define __pyx_n_s_MANUAL_SECTION_CONSTANT __pyx_mstate_global->__pyx_n_s_MANUAL_SECTION_CONSTANT
#define __pyx_n_s_MISMATCHED_ARGUMENTS __pyx_mstate_global->__pyx_n_s_MISMATCHED_ARGUMENTS
#define __pyx_n_s_NEW_STATE __pyx_mstate_global->__pyx_n_s_NEW_STATE
#define __pyx_n_s_NO_SUCH_TACTIC __pyx_mstate_global->__pyx_n_s_NO_SUCH_TACTIC
#define __pyx_n_s_Node __pyx_mstate_global->__pyx_n_s_Node
#define __pyx_n_s_Node_Reader __pyx_mstate_global->__pyx_n_s_Node_Reader
#define __pyx_n_s_Node_Reader_List __pyx_mstate_global->__pyx_n_s_Node_Reader_List
#define __pyx_n_s_Node_Reader_List___reduce_cython __pyx_mstate_global->__pyx_n_s_Node_Reader_List___reduce_cython
#define __pyx_n_s_Node_Reader_List___setstate_cyth __pyx_mstate_global->__pyx_n_s_Node_Reader_List___setstate_cyth
#define __pyx_n_s_Node_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_Node_Reader___reduce_cython
#define __pyx_n_s_Node_Reader___setstate_cython __pyx_mstate_global->__pyx_n_s_Node_Reader___setstate_cython
#define __pyx_n_s_ORIGINAL __pyx_mstate_global->__pyx_n_s_ORIGINAL
#define __pyx_n_s_OrderedDict __pyx_mstate_global->__pyx_n_s_OrderedDict
#define __pyx_kp_u_Out_of_bounds __pyx_mstate_global->__pyx_kp_u_Out_of_bounds
#define __pyx_n_s_Outcome __pyx_mstate_global->__pyx_n_s_Outcome
#define __pyx_n_s_Outcome_Reader __pyx_mstate_global->__pyx_n_s_Outcome_Reader
#define __pyx_n_s_Outcome_Reader_List __pyx_mstate_global->__pyx_n_s_Outcome_Reader_List
#define __pyx_n_s_Outcome_Reader_List___reduce_cyt __pyx_mstate_global->__pyx_n_s_Outcome_Reader_List___reduce_cyt
#define __pyx_n_s_Outcome_Reader_List___setstate_c __pyx_mstate_global->__pyx_n_s_Outcome_Reader_List___setstate_c
#define __pyx_n_s_Outcome_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_Outcome_Reader___reduce_cython
#define __pyx_n_s_Outcome_Reader___setstate_cython __pyx_mstate_global->__pyx_n_s_Outcome_Reader___setstate_cython
#define __pyx_n_s_Outcome_Term_Reader __pyx_mstate_global->__pyx_n_s_Outcome_Term_Reader
#define __pyx_n_s_Outcome_Term_Reader___reduce_cyt __pyx_mstate_global->__pyx_n_s_Outcome_Term_Reader___reduce_cyt
#define __pyx_n_s_Outcome_Term_Reader___setstate_c __pyx_mstate_global->__pyx_n_s_Outcome_Term_Reader___setstate_c
#define __pyx_n_s_PARSE_ERROR __pyx_mstate_global->__pyx_n_s_PARSE_ERROR
#define __pyx_n_s_PREDICT __pyx_mstate_global->__pyx_n_s_PREDICT
#define __pyx_n_s_PREDICTION __pyx_mstate_global->__pyx_n_s_PREDICTION
#define __pyx_n_s_PRIMITIVE __pyx_mstate_global->__pyx_n_s_PRIMITIVE
#define __pyx_n_s_PROD __pyx_mstate_global->__pyx_n_s_PROD
#define __pyx_n_s_PROD_TERM __pyx_mstate_global->__pyx_n_s_PROD_TERM
#define __pyx_n_s_PROD_TYPE __pyx_mstate_global->__pyx_n_s_PROD_TYPE
#define __pyx_n_s_PROJECTION __pyx_mstate_global->__pyx_n_s_PROJECTION
#define __pyx_n_s_PROJ_TERM __pyx_mstate_global->__pyx_n_s_PROJ_TERM
#define __pyx_n_s_PROOF_STATE __pyx_mstate_global->__pyx_n_s_PROOF_STATE
#define __pyx_n_s_PROTOCOL_ERROR __pyx_mstate_global->__pyx_n_s_PROTOCOL_ERROR
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_n_s_Prediction __pyx_mstate_global->__pyx_n_s_Prediction
#define __pyx_n_s_PredictionProtocol __pyx_mstate_global->__pyx_n_s_PredictionProtocol
#define __pyx_n_s_PredictionProtocol_Prediction_Re __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Prediction_Re
#define __pyx_n_s_PredictionProtocol_Prediction_Re_2 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Prediction_Re_2
#define __pyx_n_s_PredictionProtocol_Prediction_Re_3 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Prediction_Re_3
#define __pyx_n_s_PredictionProtocol_Prediction_Re_4 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Prediction_Re_4
#define __pyx_n_s_PredictionProtocol_Prediction_Re_5 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Prediction_Re_5
#define __pyx_n_s_PredictionProtocol_Prediction_Re_6 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Prediction_Re_6
#define __pyx_n_s_PredictionProtocol_Reader __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Reader
#define __pyx_n_s_PredictionProtocol_Reader___redu __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Reader___redu
#define __pyx_n_s_PredictionProtocol_Reader___sets __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Reader___sets
#define __pyx_n_s_PredictionProtocol_Request_Initi __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Request_Initi
#define __pyx_n_s_PredictionProtocol_Request_Initi_2 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Request_Initi_2
#define __pyx_n_s_PredictionProtocol_Request_Initi_3 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Request_Initi_3
#define __pyx_n_s_PredictionProtocol_Request_Predi __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Request_Predi
#define __pyx_n_s_PredictionProtocol_Request_Predi_2 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Request_Predi_2
#define __pyx_n_s_PredictionProtocol_Request_Predi_3 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Request_Predi_3
#define __pyx_n_s_PredictionProtocol_Request_Reade __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Request_Reade
#define __pyx_n_s_PredictionProtocol_Request_Reade_2 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Request_Reade_2
#define __pyx_n_s_PredictionProtocol_Request_Reade_3 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Request_Reade_3
#define __pyx_n_s_PredictionProtocol_Request_Which __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Request_Which
#define __pyx_n_s_PredictionProtocol_Response_Alig __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Response_Alig
#define __pyx_n_s_PredictionProtocol_Response_Alig_2 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Response_Alig_2
#define __pyx_n_s_PredictionProtocol_Response_Alig_3 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Response_Alig_3
#define __pyx_n_s_PredictionProtocol_Response_Read __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Response_Read
#define __pyx_n_s_PredictionProtocol_Response_Read_2 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Response_Read_2
#define __pyx_n_s_PredictionProtocol_Response_Read_3 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Response_Read_3
#define __pyx_n_s_PredictionProtocol_Response_Whic __pyx_mstate_global->__pyx_n_s_PredictionProtocol_Response_Whic
#define __pyx_n_s_PredictionProtocol_TextPredictio __pyx_mstate_global->__pyx_n_s_PredictionProtocol_TextPredictio
#define __pyx_n_s_PredictionProtocol_TextPredictio_2 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_TextPredictio_2
#define __pyx_n_s_PredictionProtocol_TextPredictio_3 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_TextPredictio_3
#define __pyx_n_s_PredictionProtocol_TextPredictio_4 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_TextPredictio_4
#define __pyx_n_s_PredictionProtocol_TextPredictio_5 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_TextPredictio_5
#define __pyx_n_s_PredictionProtocol_TextPredictio_6 __pyx_mstate_global->__pyx_n_s_PredictionProtocol_TextPredictio_6
#define __pyx_n_s_ProofState __pyx_mstate_global->__pyx_n_s_ProofState
#define __pyx_n_s_ProofState_Reader __pyx_mstate_global->__pyx_n_s_ProofState_Reader
#define __pyx_n_s_ProofState_Reader_List __pyx_mstate_global->__pyx_n_s_ProofState_Reader_List
#define __pyx_n_s_ProofState_Reader_List___reduce __pyx_mstate_global->__pyx_n_s_ProofState_Reader_List___reduce
#define __pyx_n_s_ProofState_Reader_List___setstat __pyx_mstate_global->__pyx_n_s_ProofState_Reader_List___setstat
#define __pyx_n_s_ProofState_Reader___reduce_cytho __pyx_mstate_global->__pyx_n_s_ProofState_Reader___reduce_cytho
#define __pyx_n_s_ProofState_Reader___setstate_cyt __pyx_mstate_global->__pyx_n_s_ProofState_Reader___setstate_cyt
#define __pyx_n_s_ProofState_Root_Reader __pyx_mstate_global->__pyx_n_s_ProofState_Root_Reader
#define __pyx_n_s_ProofState_Root_Reader___reduce __pyx_mstate_global->__pyx_n_s_ProofState_Root_Reader___reduce
#define __pyx_n_s_ProofState_Root_Reader___setstat __pyx_mstate_global->__pyx_n_s_ProofState_Root_Reader___setstat
#define __pyx_n_s_ProofStep __pyx_mstate_global->__pyx_n_s_ProofStep
#define __pyx_n_s_ProofStep_Reader __pyx_mstate_global->__pyx_n_s_ProofStep_Reader
#define __pyx_n_s_ProofStep_Reader_List __pyx_mstate_global->__pyx_n_s_ProofStep_Reader_List
#define __pyx_n_s_ProofStep_Reader_List___reduce_c __pyx_mstate_global->__pyx_n_s_ProofStep_Reader_List___reduce_c
#define __pyx_n_s_ProofStep_Reader_List___setstate __pyx_mstate_global->__pyx_n_s_ProofStep_Reader_List___setstate
#define __pyx_n_s_ProofStep_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_ProofStep_Reader___reduce_cython
#define __pyx_n_s_ProofStep_Reader___setstate_cyth __pyx_mstate_global->__pyx_n_s_ProofStep_Reader___setstate_cyth
#define __pyx_n_s_ProofStep_Tactic_Reader __pyx_mstate_global->__pyx_n_s_ProofStep_Tactic_Reader
#define __pyx_n_s_ProofStep_Tactic_Reader___reduce __pyx_mstate_global->__pyx_n_s_ProofStep_Tactic_Reader___reduce
#define __pyx_n_s_ProofStep_Tactic_Reader___setsta __pyx_mstate_global->__pyx_n_s_ProofStep_Tactic_Reader___setsta
#define __pyx_n_s_ProofStep_Tactic_Which __pyx_mstate_global->__pyx_n_s_ProofStep_Tactic_Which
#define __pyx_n_s_Pyx_EnumBase __pyx_mstate_global->__pyx_n_s_Pyx_EnumBase
#define __pyx_n_s_Pyx_EnumBase___new __pyx_mstate_global->__pyx_n_s_Pyx_EnumBase___new
#define __pyx_n_s_Pyx_EnumBase___repr __pyx_mstate_global->__pyx_n_s_Pyx_EnumBase___repr
#define __pyx_n_s_Pyx_EnumBase___str __pyx_mstate_global->__pyx_n_s_Pyx_EnumBase___str
#define __pyx_n_s_Pyx_EnumMeta___reduce_cython __pyx_mstate_global->__pyx_n_s_Pyx_EnumMeta___reduce_cython
#define __pyx_n_s_Pyx_EnumMeta___setstate_cython __pyx_mstate_global->__pyx_n_s_Pyx_EnumMeta___setstate_cython
#define __pyx_n_s_REL __pyx_mstate_global->__pyx_n_s_REL
#define __pyx_n_s_REL_POINTER __pyx_mstate_global->__pyx_n_s_REL_POINTER
#define __pyx_n_s_Request __pyx_mstate_global->__pyx_n_s_Request
#define __pyx_n_s_Response __pyx_mstate_global->__pyx_n_s_Response
#define __pyx_n_s_SORT_PROP __pyx_mstate_global->__pyx_n_s_SORT_PROP
#define __pyx_n_s_SORT_SET __pyx_mstate_global->__pyx_n_s_SORT_SET
#define __pyx_n_s_SORT_S_PROP __pyx_mstate_global->__pyx_n_s_SORT_S_PROP
#define __pyx_n_s_SORT_TYPE __pyx_mstate_global->__pyx_n_s_SORT_TYPE
#define __pyx_n_s_SUBSTITUTED __pyx_mstate_global->__pyx_n_s_SUBSTITUTED
#define __pyx_n_s_SYNCHRONIZE __pyx_mstate_global->__pyx_n_s_SYNCHRONIZE
#define __pyx_n_s_SYNCHRONIZED __pyx_mstate_global->__pyx_n_s_SYNCHRONIZED
#define __pyx_n_s_String_List __pyx_mstate_global->__pyx_n_s_String_List
#define __pyx_n_s_String_List___reduce_cython __pyx_mstate_global->__pyx_n_s_String_List___reduce_cython
#define __pyx_n_s_String_List___setstate_cython __pyx_mstate_global->__pyx_n_s_String_List___setstate_cython
#define __pyx_n_s_TACTICAL_CONSTANT __pyx_mstate_global->__pyx_n_s_TACTICAL_CONSTANT
#define __pyx_n_s_TACTICAL_SECTION_CONSTANT __pyx_mstate_global->__pyx_n_s_TACTICAL_SECTION_CONSTANT
#define __pyx_n_s_TERM __pyx_mstate_global->__pyx_n_s_TERM
#define __pyx_n_s_TEXT_PREDICTION __pyx_mstate_global->__pyx_n_s_TEXT_PREDICTION
#define __pyx_n_s_Tactic __pyx_mstate_global->__pyx_n_s_Tactic
#define __pyx_n_s_Tactic_Reader __pyx_mstate_global->__pyx_n_s_Tactic_Reader
#define __pyx_n_s_Tactic_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_Tactic_Reader___reduce_cython
#define __pyx_n_s_Tactic_Reader___setstate_cython __pyx_mstate_global->__pyx_n_s_Tactic_Reader___setstate_cython
#define __pyx_n_s_TextPrediction __pyx_mstate_global->__pyx_n_s_TextPrediction
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_n_s_UNKNOWN __pyx_mstate_global->__pyx_n_s_UNKNOWN
#define __pyx_n_s_UNRESOLVABLE __pyx_mstate_global->__pyx_n_s_UNRESOLVABLE
#define __pyx_n_s_Uint16_List __pyx_mstate_global->__pyx_n_s_Uint16_List
#define __pyx_n_s_Uint16_List___reduce_cython __pyx_mstate_global->__pyx_n_s_Uint16_List___reduce_cython
#define __pyx_n_s_Uint16_List___setstate_cython __pyx_mstate_global->__pyx_n_s_Uint16_List___setstate_cython
#define __pyx_n_s_Uint32_List __pyx_mstate_global->__pyx_n_s_Uint32_List
#define __pyx_n_s_Uint32_List___reduce_cython __pyx_mstate_global->__pyx_n_s_Uint32_List___reduce_cython
#define __pyx_n_s_Uint32_List___setstate_cython __pyx_mstate_global->__pyx_n_s_Uint32_List___setstate_cython
#define __pyx_n_s_Uint64_List __pyx_mstate_global->__pyx_n_s_Uint64_List
#define __pyx_n_s_Uint64_List___reduce_cython __pyx_mstate_global->__pyx_n_s_Uint64_List___reduce_cython
#define __pyx_n_s_Uint64_List___setstate_cython __pyx_mstate_global->__pyx_n_s_Uint64_List___setstate_cython
#define __pyx_kp_s_Unknown_enum_value_s __pyx_mstate_global->__pyx_kp_s_Unknown_enum_value_s
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_s__118 __pyx_mstate_global->__pyx_n_s__118
#define __pyx_kp_u__3 __pyx_mstate_global->__pyx_kp_u__3
#define __pyx_n_s__4 __pyx_mstate_global->__pyx_n_s__4
#define __pyx_n_s_as_struct __pyx_mstate_global->__pyx_n_s_as_struct
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_capnp __pyx_mstate_global->__pyx_n_s_capnp
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_cls __pyx_mstate_global->__pyx_n_s_cls
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_n_s_dct __pyx_mstate_global->__pyx_n_s_dct
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_doc __pyx_mstate_global->__pyx_n_s_doc
#define __pyx_n_s_dyn __pyx_mstate_global->__pyx_n_s_dyn
#define __pyx_n_s_dynamic __pyx_mstate_global->__pyx_n_s_dynamic
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_enum __pyx_mstate_global->__pyx_n_s_enum
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_graph_api_capnp __pyx_mstate_global->__pyx_n_s_graph_api_capnp
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_init __pyx_mstate_global->__pyx_n_s_init
#define __pyx_n_s_init_subclass __pyx_mstate_global->__pyx_n_s_init_subclass
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_members __pyx_mstate_global->__pyx_n_s_members
#define __pyx_n_s_metaclass __pyx_mstate_global->__pyx_n_s_metaclass
#define __pyx_n_s_module __pyx_mstate_global->__pyx_n_s_module
#define __pyx_n_s_mro_entries __pyx_mstate_global->__pyx_n_s_mro_entries
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_n_s_os __pyx_mstate_global->__pyx_n_s_os
#define __pyx_n_s_parents __pyx_mstate_global->__pyx_n_s_parents
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_prepare __pyx_mstate_global->__pyx_n_s_prepare
#define __pyx_n_s_pytact __pyx_mstate_global->__pyx_n_s_pytact
#define __pyx_n_s_pytact_graph_api_capnp __pyx_mstate_global->__pyx_n_s_pytact_graph_api_capnp
#define __pyx_n_s_pytact_graph_api_capnp_cython __pyx_mstate_global->__pyx_n_s_pytact_graph_api_capnp_cython
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle___Pyx_EnumMeta __pyx_mstate_global->__pyx_n_s_pyx_unpickle___Pyx_EnumMeta
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_qualname __pyx_mstate_global->__pyx_n_s_qualname
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_repr __pyx_mstate_global->__pyx_n_s_repr
#define __pyx_n_s_res __pyx_mstate_global->__pyx_n_s_res
#define __pyx_kp_s_s_s __pyx_mstate_global->__pyx_kp_s_s_s
#define __pyx_kp_s_s_s_d __pyx_mstate_global->__pyx_kp_s_s_s_d
#define __pyx_n_s_schema __pyx_mstate_global->__pyx_n_s_schema
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_kp_s_self_source_cannot_be_converted __pyx_mstate_global->__pyx_kp_s_self_source_cannot_be_converted
#define __pyx_n_s_set_name __pyx_mstate_global->__pyx_n_s_set_name
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_n_s_v __pyx_mstate_global->__pyx_n_s_v
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_values __pyx_mstate_global->__pyx_n_s_values
#define __pyx_int_222419149 __pyx_mstate_global->__pyx_int_222419149
#define __pyx_int_228825662 __pyx_mstate_global->__pyx_int_228825662
#define __pyx_int_238750788 __pyx_mstate_global->__pyx_int_238750788
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_tuple__2 __pyx_mstate_global->__pyx_tuple__2
#define __pyx_tuple__5 __pyx_mstate_global->__pyx_tuple__5
#define __pyx_tuple__6 __pyx_mstate_global->__pyx_tuple__6
#define __pyx_tuple__8 __pyx_mstate_global->__pyx_tuple__8
#define __pyx_codeobj__7 __pyx_mstate_global->__pyx_codeobj__7
#define __pyx_codeobj__9 __pyx_mstate_global->__pyx_codeobj__9
#define __pyx_tuple__108 __pyx_mstate_global->__pyx_tuple__108
#define __pyx_tuple__111 __pyx_mstate_global->__pyx_tuple__111
#define __pyx_tuple__113 __pyx_mstate_global->__pyx_tuple__113
#define __pyx_tuple__116 __pyx_mstate_global->__pyx_tuple__116
#define __pyx_codeobj__10 __pyx_mstate_global->__pyx_codeobj__10
#define __pyx_codeobj__11 __pyx_mstate_global->__pyx_codeobj__11
#define __pyx_codeobj__12 __pyx_mstate_global->__pyx_codeobj__12
#define __pyx_codeobj__13 __pyx_mstate_global->__pyx_codeobj__13
#define __pyx_codeobj__14 __pyx_mstate_global->__pyx_codeobj__14
#define __pyx_codeobj__15 __pyx_mstate_global->__pyx_codeobj__15
#define __pyx_codeobj__16 __pyx_mstate_global->__pyx_codeobj__16
#define __pyx_codeobj__17 __pyx_mstate_global->__pyx_codeobj__17
#define __pyx_codeobj__18 __pyx_mstate_global->__pyx_codeobj__18
#define __pyx_codeobj__19 __pyx_mstate_global->__pyx_codeobj__19
#define __pyx_codeobj__20 __pyx_mstate_global->__pyx_codeobj__20
#define __pyx_codeobj__21 __pyx_mstate_global->__pyx_codeobj__21
#define __pyx_codeobj__22 __pyx_mstate_global->__pyx_codeobj__22
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__24 __pyx_mstate_global->__pyx_codeobj__24
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__26 __pyx_mstate_global->__pyx_codeobj__26
#define __pyx_codeobj__27 __pyx_mstate_global->__pyx_codeobj__27
#define __pyx_codeobj__28 __pyx_mstate_global->__pyx_codeobj__28
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__30 __pyx_mstate_global->__pyx_codeobj__30
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__34 __pyx_mstate_global->__pyx_codeobj__34
#define __pyx_codeobj__35 __pyx_mstate_global->__pyx_codeobj__35
#define __pyx_codeobj__36 __pyx_mstate_global->__pyx_codeobj__36
#define __pyx_codeobj__37 __pyx_mstate_global->__pyx_codeobj__37
#define __pyx_codeobj__38 __pyx_mstate_global->__pyx_codeobj__38
#define __pyx_codeobj__39 __pyx_mstate_global->__pyx_codeobj__39
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__45 __pyx_mstate_global->__pyx_codeobj__45
#define __pyx_codeobj__46 __pyx_mstate_global->__pyx_codeobj__46
#define __pyx_codeobj__47 __pyx_mstate_global->__pyx_codeobj__47
#define __pyx_codeobj__48 __pyx_mstate_global->__pyx_codeobj__48
#define __pyx_codeobj__49 __pyx_mstate_global->__pyx_codeobj__49
#define __pyx_codeobj__50 __pyx_mstate_global->__pyx_codeobj__50
#define __pyx_codeobj__51 __pyx_mstate_global->__pyx_codeobj__51
#define __pyx_codeobj__52 __pyx_mstate_global->__pyx_codeobj__52
#define __pyx_codeobj__53 __pyx_mstate_global->__pyx_codeobj__53
#define __pyx_codeobj__54 __pyx_mstate_global->__pyx_codeobj__54
#define __pyx_codeobj__55 __pyx_mstate_global->__pyx_codeobj__55
#define __pyx_codeobj__56 __pyx_mstate_global->__pyx_codeobj__56
#define __pyx_codeobj__57 __pyx_mstate_global->__pyx_codeobj__57
#define __pyx_codeobj__58 __pyx_mstate_global->__pyx_codeobj__58
#define __pyx_codeobj__59 __pyx_mstate_global->__pyx_codeobj__59
#define __pyx_codeobj__60 __pyx_mstate_global->__pyx_codeobj__60
#define __pyx_codeobj__61 __pyx_mstate_global->__pyx_codeobj__61
#define __pyx_codeobj__62 __pyx_mstate_global->__pyx_codeobj__62
#define __pyx_codeobj__63 __pyx_mstate_global->__pyx_codeobj__63
#define __pyx_codeobj__64 __pyx_mstate_global->__pyx_codeobj__64
#define __pyx_codeobj__65 __pyx_mstate_global->__pyx_codeobj__65
#define __pyx_codeobj__66 __pyx_mstate_global->__pyx_codeobj__66
#define __pyx_codeobj__67 __pyx_mstate_global->__pyx_codeobj__67
#define __pyx_codeobj__68 __pyx_mstate_global->__pyx_codeobj__68
#define __pyx_codeobj__69 __pyx_mstate_global->__pyx_codeobj__69
#define __pyx_codeobj__70 __pyx_mstate_global->__pyx_codeobj__70
#define __pyx_codeobj__71 __pyx_mstate_global->__pyx_codeobj__71
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__73 __pyx_mstate_global->__pyx_codeobj__73
#define __pyx_codeobj__74 __pyx_mstate_global->__pyx_codeobj__74
#define __pyx_codeobj__75 __pyx_mstate_global->__pyx_codeobj__75
#define __pyx_codeobj__76 __pyx_mstate_global->__pyx_codeobj__76
#define __pyx_codeobj__77 __pyx_mstate_global->__pyx_codeobj__77
#define __pyx_codeobj__78 __pyx_mstate_global->__pyx_codeobj__78
#define __pyx_codeobj__79 __pyx_mstate_global->__pyx_codeobj__79
#define __pyx_codeobj__80 __pyx_mstate_global->__pyx_codeobj__80
#define __pyx_codeobj__81 __pyx_mstate_global->__pyx_codeobj__81
#define __pyx_codeobj__82 __pyx_mstate_global->__pyx_codeobj__82
#define __pyx_codeobj__83 __pyx_mstate_global->__pyx_codeobj__83
#define __pyx_codeobj__84 __pyx_mstate_global->__pyx_codeobj__84
#define __pyx_codeobj__85 __pyx_mstate_global->__pyx_codeobj__85
#define __pyx_codeobj__86 __pyx_mstate_global->__pyx_codeobj__86
#define __pyx_codeobj__87 __pyx_mstate_global->__pyx_codeobj__87
#define __pyx_codeobj__88 __pyx_mstate_global->__pyx_codeobj__88
#define __pyx_codeobj__89 __pyx_mstate_global->__pyx_codeobj__89
#define __pyx_codeobj__90 __pyx_mstate_global->__pyx_codeobj__90
#define __pyx_codeobj__91 __pyx_mstate_global->__pyx_codeobj__91
#define __pyx_codeobj__92 __pyx_mstate_global->__pyx_codeobj__92
#define __pyx_codeobj__93 __pyx_mstate_global->__pyx_codeobj__93
#define __pyx_codeobj__94 __pyx_mstate_global->__pyx_codeobj__94
#define __pyx_codeobj__95 __pyx_mstate_global->__pyx_codeobj__95
#define __pyx_codeobj__96 __pyx_mstate_global->__pyx_codeobj__96
#define __pyx_codeobj__97 __pyx_mstate_global->__pyx_codeobj__97
#define __pyx_codeobj__98 __pyx_mstate_global->__pyx_codeobj__98
#define __pyx_codeobj__99 __pyx_mstate_global->__pyx_codeobj__99
#define __pyx_codeobj__100 __pyx_mstate_global->__pyx_codeobj__100
#define __pyx_codeobj__101 __pyx_mstate_global->__pyx_codeobj__101
#define __pyx_codeobj__102 __pyx_mstate_global->__pyx_codeobj__102
#define __pyx_codeobj__103 __pyx_mstate_global->__pyx_codeobj__103
#define __pyx_codeobj__104 __pyx_mstate_global->__pyx_codeobj__104
#define __pyx_codeobj__105 __pyx_mstate_global->__pyx_codeobj__105
#define __pyx_codeobj__106 __pyx_mstate_global->__pyx_codeobj__106
#define __pyx_codeobj__107 __pyx_mstate_global->__pyx_codeobj__107
#define __pyx_codeobj__109 __pyx_mstate_global->__pyx_codeobj__109
#define __pyx_codeobj__110 __pyx_mstate_global->__pyx_codeobj__110
#define __pyx_codeobj__112 __pyx_mstate_global->__pyx_codeobj__112
#define __pyx_codeobj__114 __pyx_mstate_global->__pyx_codeobj__114
#define __pyx_codeobj__115 __pyx_mstate_global->__pyx_codeobj__115
#define __pyx_codeobj__117 __pyx_mstate_global->__pyx_codeobj__117
#endif
/* #### Code section: module_code ### */

/* "pytact/graph_api_capnp_cython.pyx":20
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_Node()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 20, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 20, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 20, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 20, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":21
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_Node()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Graph::Node>();

  /* "pytact/graph_api_capnp_cython.pyx":22
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_Node()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":20
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_Node()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":25
 * 
 *     @staticmethod
 *     cdef init(C_Graph_Node_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_Node_Reader wrapper = Graph_Node_Reader.__new__(Graph_Node_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_init(Graph::Node::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":26
 *     @staticmethod
 *     cdef init(C_Graph_Node_Reader source, object root):
 *         cdef Graph_Node_Reader wrapper = Graph_Node_Reader.__new__(Graph_Node_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":27
 *     cdef init(C_Graph_Node_Reader source, object root):
 *         cdef Graph_Node_Reader wrapper = Graph_Node_Reader.__new__(Graph_Node_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":28
 *         cdef Graph_Node_Reader wrapper = Graph_Node_Reader.__new__(Graph_Node_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":29
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":25
 * 
 *     @staticmethod
 *     cdef init(C_Graph_Node_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_Node_Reader wrapper = Graph_Node_Reader.__new__(Graph_Node_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":31
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":33
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":31
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":34
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":35
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":34
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":37
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def label(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5label_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5label_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5label___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5label___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Graph::Node::Label::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":45
 *         contents but do not point to the same object nonetheless
 *         """
 *         return Graph_Node_Label_Reader.init(self.source.getLabel(), self.root)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getLabel();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 45, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":37
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def label(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.label.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":47
 *         return Graph_Node_Label_Reader.init(self.source.getLabel(), self.root)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def children_index(self):
 *         return self.source.getChildrenIndex()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":49
 *     @property
 *     def children_index(self):
 *         return self.source.getChildrenIndex()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getChildrenIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 49, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":47
 *         return Graph_Node_Label_Reader.init(self.source.getLabel(), self.root)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def children_index(self):
 *         return self.source.getChildrenIndex()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.children_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":51
 *         return self.source.getChildrenIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def children_count(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_count___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_count___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":56
 *         The children of a node are encoded as a range withing the `edges`-list of the graph.
 *         """
 *         return self.source.getChildrenCount()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getChildrenCount();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 56, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":51
 *         return self.source.getChildrenIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def children_count(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.children_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":58
 *         return self.source.getChildrenCount()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def identity(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_8identity_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_8identity_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_8identity___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_8identity___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":93
 *         for machine learning purposes.)
 *         """
 *         return self.source.getIdentity()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.Graph.Node.schema = _Schema()._init(getGraph_Node_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getIdentity();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 93, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":58
 *         return self.source.getChildrenCount()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def identity(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.identity.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_4__reduce_cython__, "Graph_Node_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_6__setstate_cython__, "Graph_Node_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":102
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_EdgeTarget()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 102, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 102, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 102, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 102, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":103
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_EdgeTarget()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Graph::EdgeTarget>();

  /* "pytact/graph_api_capnp_cython.pyx":104
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_EdgeTarget()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":102
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_EdgeTarget()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":107
 * 
 *     @staticmethod
 *     cdef init(C_Graph_EdgeTarget_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_EdgeTarget_Reader wrapper = Graph_EdgeTarget_Reader.__new__(Graph_EdgeTarget_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_init(Graph::EdgeTarget::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":108
 *     @staticmethod
 *     cdef init(C_Graph_EdgeTarget_Reader source, object root):
 *         cdef Graph_EdgeTarget_Reader wrapper = Graph_EdgeTarget_Reader.__new__(Graph_EdgeTarget_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":109
 *     cdef init(C_Graph_EdgeTarget_Reader source, object root):
 *         cdef Graph_EdgeTarget_Reader wrapper = Graph_EdgeTarget_Reader.__new__(Graph_EdgeTarget_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":110
 *         cdef Graph_EdgeTarget_Reader wrapper = Graph_EdgeTarget_Reader.__new__(Graph_EdgeTarget_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":111
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":107
 * 
 *     @staticmethod
 *     cdef init(C_Graph_EdgeTarget_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_EdgeTarget_Reader wrapper = Graph_EdgeTarget_Reader.__new__(Graph_EdgeTarget_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":113
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":115
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":113
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":116
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":117
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":116
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":119
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def label(self):
 *         return None(self.source.getLabel())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5label_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5label_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5label___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5label___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __PYX_ENUM_CLASS_DECL capnp::schemas::EdgeClassification_e40ecefe7a595f8e __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":121
 *     @property
 *     def label(self):
 *         return None(self.source.getLabel())             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_2 = __pyx_v_self->source.getLabel();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 121, __pyx_L1_error)
  }
  __pyx_t_3 = [](const capnp::schemas::EdgeClassification_e40ecefe7a595f8e& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __pyx_t_4 = Py_None; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":119
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def label(self):
 *         return None(self.source.getLabel())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader.label.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":123
 *         return None(self.source.getLabel())
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def target(self):
 *         return Graph_EdgeTarget_Target_Reader.init(self.source.getTarget(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6target_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6target_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6target___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6target___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Graph::EdgeTarget::Target::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":125
 *     @property
 *     def target(self):
 *         return Graph_EdgeTarget_Target_Reader.init(self.source.getTarget(), self.root)             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.Graph.EdgeTarget.schema = _Schema()._init(getGraph_EdgeTarget_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getTarget();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 125, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":123
 *         return None(self.source.getLabel())
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def target(self):
 *         return Graph_EdgeTarget_Target_Reader.init(self.source.getTarget(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader.target.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_4__reduce_cython__, "Graph_EdgeTarget_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6__setstate_cython__, "Graph_EdgeTarget_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":136
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 136, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 136, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 136, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 136, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":137
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Graph>();

  /* "pytact/graph_api_capnp_cython.pyx":138
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":136
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":141
 * 
 *     @staticmethod
 *     cdef init(C_Graph_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_Reader wrapper = Graph_Reader.__new__(Graph_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_12Graph_Reader_init(Graph::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":142
 *     @staticmethod
 *     cdef init(C_Graph_Reader source, object root):
 *         cdef Graph_Reader wrapper = Graph_Reader.__new__(Graph_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 142, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":143
 *     cdef init(C_Graph_Reader source, object root):
 *         cdef Graph_Reader wrapper = Graph_Reader.__new__(Graph_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":144
 *         cdef Graph_Reader wrapper = Graph_Reader.__new__(Graph_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":145
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":141
 * 
 *     @staticmethod
 *     cdef init(C_Graph_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_Reader wrapper = Graph_Reader.__new__(Graph_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":147
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":149
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":147
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":150
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":151
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":150
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":153
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def nodes(self):
 *         return Graph_Node_Reader_List.init(self.source.getNodes(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_5nodes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_5nodes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_5nodes___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_5nodes___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Graph::Node, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":155
 *     @property
 *     def nodes(self):
 *         return Graph_Node_Reader_List.init(self.source.getNodes(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_nodes(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getNodes();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 155, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":153
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def nodes(self):
 *         return Graph_Node_Reader_List.init(self.source.getNodes(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.nodes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":156
 *     def nodes(self):
 *         return Graph_Node_Reader_List.init(self.source.getNodes(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_nodes(self):
 *         return self.source.hasNodes()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_nodes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_nodes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_nodes___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_nodes___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":158
 *     @property
 *     def has_nodes(self):
 *         return self.source.hasNodes()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasNodes();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 158, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":156
 *     def nodes(self):
 *         return Graph_Node_Reader_List.init(self.source.getNodes(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_nodes(self):
 *         return self.source.hasNodes()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.has_nodes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":160
 *         return self.source.hasNodes()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def edges(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_5edges_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_5edges_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_5edges___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_5edges___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Graph::EdgeTarget, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":172
 *         Instead, you should access the heap through various entry-points that are provided.
 *         """
 *         return Graph_EdgeTarget_Reader_List.init(self.source.getEdges(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_edges(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getEdges();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 172, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":160
 *         return self.source.hasNodes()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def edges(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.edges.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":173
 *         """
 *         return Graph_EdgeTarget_Reader_List.init(self.source.getEdges(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_edges(self):
 *         return self.source.hasEdges()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_edges_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_edges_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_edges___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_edges___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":175
 *     @property
 *     def has_edges(self):
 *         return self.source.hasEdges()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.Graph.schema = _Schema()._init(getGraph_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasEdges();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 175, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":173
 *         """
 *         return Graph_EdgeTarget_Reader_List.init(self.source.getEdges(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_edges(self):
 *         return self.source.hasEdges()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.has_edges.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_12Graph_Reader_4__reduce_cython__, "Graph_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_12Graph_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_12Graph_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_12Graph_Reader_6__setstate_cython__, "Graph_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_12Graph_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_12Graph_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_12Graph_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":180
 * cdef class IntP_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asIntP()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 180, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 180, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 180, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.IntP_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 180, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":181
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asIntP()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<IntP>();

  /* "pytact/graph_api_capnp_cython.pyx":182
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asIntP()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":180
 * cdef class IntP_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asIntP()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":185
 * 
 *     @staticmethod
 *     cdef init(C_IntP_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef IntP_Reader wrapper = IntP_Reader.__new__(IntP_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11IntP_Reader_init(IntP::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":186
 *     @staticmethod
 *     cdef init(C_IntP_Reader source, object root):
 *         cdef IntP_Reader wrapper = IntP_Reader.__new__(IntP_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_IntP_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":187
 *     cdef init(C_IntP_Reader source, object root):
 *         cdef IntP_Reader wrapper = IntP_Reader.__new__(IntP_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":188
 *         cdef IntP_Reader wrapper = IntP_Reader.__new__(IntP_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":189
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":185
 * 
 *     @staticmethod
 *     cdef init(C_IntP_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef IntP_Reader wrapper = IntP_Reader.__new__(IntP_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.IntP_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":191
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":193
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":191
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.IntP_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":194
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":195
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":194
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.IntP_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":197
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def value(self):
 *         return self.source.getValue()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_5value___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_5value___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":199
 *     @property
 *     def value(self):
 *         return self.source.getValue()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.IntP.schema = _Schema()._init(getIntP_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getValue();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 199, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":197
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def value(self):
 *         return self.source.getValue()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.IntP_Reader.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11IntP_Reader_4__reduce_cython__, "IntP_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11IntP_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11IntP_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.IntP_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11IntP_Reader_6__setstate_cython__, "IntP_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11IntP_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11IntP_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.IntP_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11IntP_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.IntP_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":208
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 208, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 208, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 208, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 208, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":209
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Definition>();

  /* "pytact/graph_api_capnp_cython.pyx":210
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":208
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":213
 * 
 *     @staticmethod
 *     cdef init(C_Definition_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Definition_Reader wrapper = Definition_Reader.__new__(Definition_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_17Definition_Reader_init(Definition::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":214
 *     @staticmethod
 *     cdef init(C_Definition_Reader source, object root):
 *         cdef Definition_Reader wrapper = Definition_Reader.__new__(Definition_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 214, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":215
 *     cdef init(C_Definition_Reader source, object root):
 *         cdef Definition_Reader wrapper = Definition_Reader.__new__(Definition_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":216
 *         cdef Definition_Reader wrapper = Definition_Reader.__new__(Definition_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":217
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":213
 * 
 *     @staticmethod
 *     cdef init(C_Definition_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Definition_Reader wrapper = Definition_Reader.__new__(Definition_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":219
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":221
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":219
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":222
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":223
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 *     @property
 *     def which(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":222
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":224
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return Definition_Which(self.source.which())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_5which_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_5which_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_5which___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __PYX_ENUM_CLASS_DECL Definition::Which __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":226
 *     @property
 *     def which(self):
 *         return Definition_Which(self.source.which())             # <<<<<<<<<<<<<<
 *     @property
 *     def which_raw(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Definition_Which); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    __pyx_t_3 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 226, __pyx_L1_error)
  }
  __pyx_t_4 = [](const Definition::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":224
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return Definition_Which(self.source.which())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.which.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":227
 *     def which(self):
 *         return Definition_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9which_raw_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9which_raw_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9which_raw___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __PYX_ENUM_CLASS_DECL Definition::Which __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":229
 *     @property
 *     def which_raw(self):
 *         return self.source.which()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 229, __pyx_L1_error)
  }
  __pyx_t_2 = [](const Definition::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":227
 *     def which(self):
 *         return Definition_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.which_raw.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":231
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_4name___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_4name___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":237
 *         but is not unique among different branches of the global in a dataset.
 *         """
 *         temp = self.source.getName()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getName();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 237, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":238
 *         """
 *         temp = self.source.getName()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_name(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":231
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":239
 *         temp = self.source.getName()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_name(self):
 *         return self.source.hasName()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_8has_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_8has_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_8has_name___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_8has_name___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":241
 *     @property
 *     def has_name(self):
 *         return self.source.hasName()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasName();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 241, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":239
 *         temp = self.source.getName()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_name(self):
 *         return self.source.hasName()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.has_name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":243
 *         return self.source.hasName()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def previous(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_8previous_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_8previous_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_8previous___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_8previous___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":252
 *         recursive definitions. Those nodes are placed into the global context in an arbitrary ordering.
 *         """
 *         return self.source.getPrevious()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getPrevious();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 252, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":243
 *         return self.source.hasName()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def previous(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.previous.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":254
 *         return self.source.getPrevious()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def external_previous(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_17external_previous_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_17external_previous_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_17external_previous___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_17external_previous___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":262
 *         definitions.
 *         """
 *         return Uint16_List.init(self.source.getExternalPrevious(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_external_previous(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getExternalPrevious();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 262, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_11Uint16_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":254
 *         return self.source.getPrevious()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def external_previous(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.external_previous.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":263
 *         """
 *         return Uint16_List.init(self.source.getExternalPrevious(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_external_previous(self):
 *         return self.source.hasExternalPrevious()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_external_previous_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_external_previous_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_external_previous___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_external_previous___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":265
 *     @property
 *     def has_external_previous(self):
 *         return self.source.hasExternalPrevious()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasExternalPrevious();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 265, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":263
 *         """
 *         return Uint16_List.init(self.source.getExternalPrevious(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_external_previous(self):
 *         return self.source.hasExternalPrevious()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.has_external_previous.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":267
 *         return self.source.hasExternalPrevious()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def status(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_6status_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_6status_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_6status___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_6status___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Definition::Status::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":272
 *         A definition can have different origins, which are encoded here
 *         """
 *         return Definition_Status_Reader.init(self.source.getStatus(), self.root)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getStatus();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 272, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":267
 *         return self.source.hasExternalPrevious()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def status(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.status.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":274
 *         return Definition_Status_Reader.init(self.source.getStatus(), self.root)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inductive(self):
 *         return self.source.getInductive()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9inductive_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9inductive_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9inductive___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9inductive___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":276
 *     @property
 *     def inductive(self):
 *         return self.source.getInductive()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_inductive(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getInductive();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 276, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":274
 *         return Definition_Status_Reader.init(self.source.getStatus(), self.root)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inductive(self):
 *         return self.source.getInductive()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.inductive.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":277
 *     def inductive(self):
 *         return self.source.getInductive()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_inductive(self):
 *         return self.source.isInductive()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_12is_inductive_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_12is_inductive_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_12is_inductive___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_12is_inductive___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":279
 *     @property
 *     def is_inductive(self):
 *         return self.source.isInductive()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isInductive();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 279, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":277
 *     def inductive(self):
 *         return self.source.getInductive()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_inductive(self):
 *         return self.source.isInductive()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.is_inductive.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":281
 *         return self.source.isInductive()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def constructor(self):
 *         return self.source.getConstructor()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_11constructor_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_11constructor_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_11constructor___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_11constructor___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":283
 *     @property
 *     def constructor(self):
 *         return self.source.getConstructor()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_constructor(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getConstructor();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 283, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":281
 *         return self.source.isInductive()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def constructor(self):
 *         return self.source.getConstructor()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.constructor.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":284
 *     def constructor(self):
 *         return self.source.getConstructor()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_constructor(self):
 *         return self.source.isConstructor()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_14is_constructor_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_14is_constructor_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_14is_constructor___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_14is_constructor___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":286
 *     @property
 *     def is_constructor(self):
 *         return self.source.isConstructor()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isConstructor();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 286, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":284
 *     def constructor(self):
 *         return self.source.getConstructor()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_constructor(self):
 *         return self.source.isConstructor()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.is_constructor.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":288
 *         return self.source.isConstructor()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def projection(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_10projection_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_10projection_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_10projection___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_10projection___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":297
 *         cluster are interleaved within the chain of `previous` nodes.
 *         """
 *         return self.source.getProjection()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_projection(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getProjection();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 297, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":288
 *         return self.source.isConstructor()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def projection(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.projection.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":298
 *         """
 *         return self.source.getProjection()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_projection(self):
 *         return self.source.isProjection()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_13is_projection_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_13is_projection_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_13is_projection___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_13is_projection___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":300
 *     @property
 *     def is_projection(self):
 *         return self.source.isProjection()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isProjection();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 300, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":298
 *         """
 *         return self.source.getProjection()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_projection(self):
 *         return self.source.isProjection()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.is_projection.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":302
 *         return self.source.isProjection()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def manual_constant(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_15manual_constant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_15manual_constant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_15manual_constant___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_15manual_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":309
 *         refinement proofs that build the term iteratively.
 *         """
 *         return self.source.getManualConstant()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_manual_constant(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getManualConstant();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 309, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":302
 *         return self.source.isProjection()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def manual_constant(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.manual_constant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":310
 *         """
 *         return self.source.getManualConstant()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_manual_constant(self):
 *         return self.source.isManualConstant()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_18is_manual_constant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_18is_manual_constant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_18is_manual_constant___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_18is_manual_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":312
 *     @property
 *     def is_manual_constant(self):
 *         return self.source.isManualConstant()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isManualConstant();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 312, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":310
 *         """
 *         return self.source.getManualConstant()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_manual_constant(self):
 *         return self.source.isManualConstant()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.is_manual_constant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":314
 *         return self.source.isManualConstant()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactical_constant(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_17tactical_constant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_17tactical_constant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_17tactical_constant___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_17tactical_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":320
 *         Note that for non-original constants, the proof step sequence may not be very accurate.
 *         """
 *         return ProofStep_Reader_List.init(self.source.getTacticalConstant(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_tactical_constant(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getTacticalConstant();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 320, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":314
 *         return self.source.isManualConstant()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactical_constant(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.tactical_constant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":321
 *         """
 *         return ProofStep_Reader_List.init(self.source.getTacticalConstant(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_tactical_constant(self):
 *         return self.source.isTacticalConstant()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_20is_tactical_constant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_20is_tactical_constant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_20is_tactical_constant___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_20is_tactical_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":323
 *     @property
 *     def is_tactical_constant(self):
 *         return self.source.isTacticalConstant()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_tactical_constant(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isTacticalConstant();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 323, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":321
 *         """
 *         return ProofStep_Reader_List.init(self.source.getTacticalConstant(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_tactical_constant(self):
 *         return self.source.isTacticalConstant()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.is_tactical_constant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":324
 *     def is_tactical_constant(self):
 *         return self.source.isTacticalConstant()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactical_constant(self):
 *         return self.source.hasTacticalConstant()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_tactical_constant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_tactical_constant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_tactical_constant___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_tactical_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":326
 *     @property
 *     def has_tactical_constant(self):
 *         return self.source.hasTacticalConstant()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTacticalConstant();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 326, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":324
 *     def is_tactical_constant(self):
 *         return self.source.isTacticalConstant()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactical_constant(self):
 *         return self.source.hasTacticalConstant()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.has_tactical_constant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":328
 *         return self.source.hasTacticalConstant()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def manual_section_constant(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_23manual_section_constant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_23manual_section_constant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_23manual_section_constant___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_23manual_section_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":333
 *         A section variable or local section definition.
 *         """
 *         return self.source.getManualSectionConstant()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_manual_section_constant(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getManualSectionConstant();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 333, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":328
 *         return self.source.hasTacticalConstant()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def manual_section_constant(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.manual_section_constant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":334
 *         """
 *         return self.source.getManualSectionConstant()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_manual_section_constant(self):
 *         return self.source.isManualSectionConstant()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_26is_manual_section_constant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_26is_manual_section_constant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_26is_manual_section_constant___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_26is_manual_section_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":336
 *     @property
 *     def is_manual_section_constant(self):
 *         return self.source.isManualSectionConstant()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isManualSectionConstant();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 336, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":334
 *         """
 *         return self.source.getManualSectionConstant()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_manual_section_constant(self):
 *         return self.source.isManualSectionConstant()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.is_manual_section_constant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":338
 *         return self.source.isManualSectionConstant()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactical_section_constant(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_25tactical_section_constant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_25tactical_section_constant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_25tactical_section_constant___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_25tactical_section_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":343
 *         Local section definitions can also be defined using a tactical proof.
 *         """
 *         return ProofStep_Reader_List.init(self.source.getTacticalSectionConstant(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_tactical_section_constant(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getTacticalSectionConstant();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 343, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":338
 *         return self.source.isManualSectionConstant()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactical_section_constant(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.tactical_section_constant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":344
 *         """
 *         return ProofStep_Reader_List.init(self.source.getTacticalSectionConstant(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_tactical_section_constant(self):
 *         return self.source.isTacticalSectionConstant()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_28is_tactical_section_constant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_28is_tactical_section_constant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_28is_tactical_section_constant___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_28is_tactical_section_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":346
 *     @property
 *     def is_tactical_section_constant(self):
 *         return self.source.isTacticalSectionConstant()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_tactical_section_constant(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isTacticalSectionConstant();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 346, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":344
 *         """
 *         return ProofStep_Reader_List.init(self.source.getTacticalSectionConstant(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_tactical_section_constant(self):
 *         return self.source.isTacticalSectionConstant()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.is_tactical_section_constant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":347
 *     def is_tactical_section_constant(self):
 *         return self.source.isTacticalSectionConstant()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactical_section_constant(self):
 *         return self.source.hasTacticalSectionConstant()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_29has_tactical_section_constant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_29has_tactical_section_constant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_29has_tactical_section_constant___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_29has_tactical_section_constant___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":349
 *     @property
 *     def has_tactical_section_constant(self):
 *         return self.source.hasTacticalSectionConstant()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTacticalSectionConstant();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 349, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":347
 *     def is_tactical_section_constant(self):
 *         return self.source.isTacticalSectionConstant()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactical_section_constant(self):
 *         return self.source.hasTacticalSectionConstant()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.has_tactical_section_constant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":351
 *         return self.source.hasTacticalSectionConstant()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_text(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9type_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9type_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9type_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9type_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":356
 *         A textual representation of the type of this definition.
 *         """
 *         temp = self.source.getTypeText()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getTypeText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 356, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":357
 *         """
 *         temp = self.source.getTypeText()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_type_text(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":351
 *         return self.source.hasTacticalSectionConstant()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_text(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.type_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":358
 *         temp = self.source.getTypeText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_type_text(self):
 *         return self.source.hasTypeText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_type_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_type_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_type_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_type_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":360
 *     @property
 *     def has_type_text(self):
 *         return self.source.hasTypeText()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTypeText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 360, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":358
 *         temp = self.source.getTypeText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_type_text(self):
 *         return self.source.hasTypeText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.has_type_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":362
 *         return self.source.hasTypeText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def term_text(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9term_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9term_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9term_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_9term_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":368
 *         For inductives, constructors, projections, section variables and axioms the string is empty.
 *         """
 *         temp = self.source.getTermText()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getTermText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 368, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":369
 *         """
 *         temp = self.source.getTermText()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_term_text(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":362
 *         return self.source.hasTypeText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def term_text(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.term_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":370
 *         temp = self.source.getTermText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_term_text(self):
 *         return self.source.hasTermText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_term_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_term_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_term_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_term_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":372
 *     @property
 *     def has_term_text(self):
 *         return self.source.hasTermText()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.Definition.schema = _Schema()._init(getDefinition_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTermText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 372, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":370
 *         temp = self.source.getTermText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_term_text(self):
 *         return self.source.hasTermText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.has_term_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_17Definition_Reader_4__reduce_cython__, "Definition_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_17Definition_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17Definition_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_17Definition_Reader_6__setstate_cython__, "Definition_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_17Definition_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17Definition_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17Definition_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":377
 * cdef class FloatP_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asFloatP()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 377, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 377, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 377, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.FloatP_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 377, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":378
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asFloatP()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<FloatP>();

  /* "pytact/graph_api_capnp_cython.pyx":379
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asFloatP()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":377
 * cdef class FloatP_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asFloatP()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":382
 * 
 *     @staticmethod
 *     cdef init(C_FloatP_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef FloatP_Reader wrapper = FloatP_Reader.__new__(FloatP_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_13FloatP_Reader_init(FloatP::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":383
 *     @staticmethod
 *     cdef init(C_FloatP_Reader source, object root):
 *         cdef FloatP_Reader wrapper = FloatP_Reader.__new__(FloatP_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_FloatP_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 383, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":384
 *     cdef init(C_FloatP_Reader source, object root):
 *         cdef FloatP_Reader wrapper = FloatP_Reader.__new__(FloatP_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":385
 *         cdef FloatP_Reader wrapper = FloatP_Reader.__new__(FloatP_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":386
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":382
 * 
 *     @staticmethod
 *     cdef init(C_FloatP_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef FloatP_Reader wrapper = FloatP_Reader.__new__(FloatP_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.FloatP_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":388
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":390
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":388
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.FloatP_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":391
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":392
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":391
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.FloatP_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":394
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def value(self):
 *         return self.source.getValue()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5value___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5value___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":396
 *     @property
 *     def value(self):
 *         return self.source.getValue()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.FloatP.schema = _Schema()._init(getFloatP_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getValue();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 396, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":394
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def value(self):
 *         return self.source.getValue()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.FloatP_Reader.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_13FloatP_Reader_4__reduce_cython__, "FloatP_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_13FloatP_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.FloatP_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_13FloatP_Reader_6__setstate_cython__, "FloatP_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_13FloatP_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.FloatP_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13FloatP_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.FloatP_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":407
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_Node_Label()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 407, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 407, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 407, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 407, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":408
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_Node_Label()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Graph::Node::Label>();

  /* "pytact/graph_api_capnp_cython.pyx":409
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_Node_Label()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":407
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_Node_Label()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":412
 * 
 *     @staticmethod
 *     cdef init(C_Graph_Node_Label_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_Node_Label_Reader wrapper = Graph_Node_Label_Reader.__new__(Graph_Node_Label_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_init(Graph::Node::Label::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":413
 *     @staticmethod
 *     cdef init(C_Graph_Node_Label_Reader source, object root):
 *         cdef Graph_Node_Label_Reader wrapper = Graph_Node_Label_Reader.__new__(Graph_Node_Label_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 413, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":414
 *     cdef init(C_Graph_Node_Label_Reader source, object root):
 *         cdef Graph_Node_Label_Reader wrapper = Graph_Node_Label_Reader.__new__(Graph_Node_Label_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":415
 *         cdef Graph_Node_Label_Reader wrapper = Graph_Node_Label_Reader.__new__(Graph_Node_Label_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":416
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":412
 * 
 *     @staticmethod
 *     cdef init(C_Graph_Node_Label_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_Node_Label_Reader wrapper = Graph_Node_Label_Reader.__new__(Graph_Node_Label_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":418
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":420
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":418
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":421
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":422
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 *     @property
 *     def which(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":421
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":423
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return Graph_Node_Label_Which(self.source.which())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5which_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5which_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5which___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __PYX_ENUM_CLASS_DECL Graph::Node::Label::Which __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":425
 *     @property
 *     def which(self):
 *         return Graph_Node_Label_Which(self.source.which())             # <<<<<<<<<<<<<<
 *     @property
 *     def which_raw(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    __pyx_t_3 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 425, __pyx_L1_error)
  }
  __pyx_t_4 = [](const Graph::Node::Label::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":423
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return Graph_Node_Label_Which(self.source.which())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.which.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":426
 *     def which(self):
 *         return Graph_Node_Label_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9which_raw_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9which_raw_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9which_raw___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __PYX_ENUM_CLASS_DECL Graph::Node::Label::Which __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":428
 *     @property
 *     def which_raw(self):
 *         return self.source.which()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 428, __pyx_L1_error)
  }
  __pyx_t_2 = [](const Graph::Node::Label::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":426
 *     def which(self):
 *         return Graph_Node_Label_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.which_raw.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":430
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def proof_state(self):
 *         return IntP_Reader.init(self.source.getProofState(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11proof_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11proof_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11proof_state___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11proof_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  IntP::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":432
 *     @property
 *     def proof_state(self):
 *         return IntP_Reader.init(self.source.getProofState(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_proof_state(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getProofState();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 432, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_11IntP_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":430
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def proof_state(self):
 *         return IntP_Reader.init(self.source.getProofState(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.proof_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":433
 *     def proof_state(self):
 *         return IntP_Reader.init(self.source.getProofState(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_proof_state(self):
 *         return self.source.isProofState()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_proof_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_proof_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_proof_state___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_proof_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":435
 *     @property
 *     def is_proof_state(self):
 *         return self.source.isProofState()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_proof_state(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isProofState();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 435, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":433
 *     def proof_state(self):
 *         return IntP_Reader.init(self.source.getProofState(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_proof_state(self):
 *         return self.source.isProofState()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_proof_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":436
 *     def is_proof_state(self):
 *         return self.source.isProofState()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_proof_state(self):
 *         return self.source.hasProofState()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_15has_proof_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_15has_proof_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_15has_proof_state___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_15has_proof_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":438
 *     @property
 *     def has_proof_state(self):
 *         return self.source.hasProofState()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasProofState();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 438, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":436
 *     def is_proof_state(self):
 *         return self.source.isProofState()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_proof_state(self):
 *         return self.source.hasProofState()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.has_proof_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":440
 *         return self.source.hasProofState()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context_def(self):
 *         return self.source.getContextDef()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11context_def_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11context_def_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11context_def___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11context_def___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":442
 *     @property
 *     def context_def(self):
 *         return self.source.getContextDef()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_context_def(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getContextDef();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 442, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":440
 *         return self.source.hasProofState()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context_def(self):
 *         return self.source.getContextDef()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.context_def.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":443
 *     def context_def(self):
 *         return self.source.getContextDef()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_context_def(self):
 *         return self.source.isContextDef()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_context_def_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_context_def_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_context_def___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_context_def___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":445
 *     @property
 *     def is_context_def(self):
 *         return self.source.isContextDef()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isContextDef();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 445, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":443
 *     def context_def(self):
 *         return self.source.getContextDef()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_context_def(self):
 *         return self.source.isContextDef()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_context_def.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":447
 *         return self.source.isContextDef()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context_assum(self):
 *         return self.source.getContextAssum()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13context_assum_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13context_assum_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13context_assum___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13context_assum___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":449
 *     @property
 *     def context_assum(self):
 *         return self.source.getContextAssum()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_context_assum(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getContextAssum();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 449, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":447
 *         return self.source.isContextDef()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context_assum(self):
 *         return self.source.getContextAssum()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.context_assum.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":450
 *     def context_assum(self):
 *         return self.source.getContextAssum()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_context_assum(self):
 *         return self.source.isContextAssum()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_16is_context_assum_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_16is_context_assum_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_16is_context_assum___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_16is_context_assum___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":452
 *     @property
 *     def is_context_assum(self):
 *         return self.source.isContextAssum()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isContextAssum();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 452, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":450
 *     def context_assum(self):
 *         return self.source.getContextAssum()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_context_assum(self):
 *         return self.source.isContextAssum()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_context_assum.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":454
 *         return self.source.isContextAssum()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def definition(self):
 *         return Definition_Reader.init(self.source.getDefinition(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10definition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10definition_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10definition___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10definition___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Definition::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":456
 *     @property
 *     def definition(self):
 *         return Definition_Reader.init(self.source.getDefinition(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_definition(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDefinition();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 456, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_17Definition_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":454
 *         return self.source.isContextAssum()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def definition(self):
 *         return Definition_Reader.init(self.source.getDefinition(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.definition.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":457
 *     def definition(self):
 *         return Definition_Reader.init(self.source.getDefinition(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_definition(self):
 *         return self.source.isDefinition()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_definition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_definition_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_definition___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_definition___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":459
 *     @property
 *     def is_definition(self):
 *         return self.source.isDefinition()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_definition(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isDefinition();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 459, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":457
 *     def definition(self):
 *         return Definition_Reader.init(self.source.getDefinition(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_definition(self):
 *         return self.source.isDefinition()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_definition.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":460
 *     def is_definition(self):
 *         return self.source.isDefinition()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_definition(self):
 *         return self.source.hasDefinition()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14has_definition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14has_definition_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14has_definition___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14has_definition___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":462
 *     @property
 *     def has_definition(self):
 *         return self.source.hasDefinition()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasDefinition();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 462, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":460
 *     def is_definition(self):
 *         return self.source.isDefinition()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_definition(self):
 *         return self.source.hasDefinition()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.has_definition.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":464
 *         return self.source.hasDefinition()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def const_empty(self):
 *         return self.source.getConstEmpty()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11const_empty_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11const_empty_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11const_empty___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11const_empty___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":466
 *     @property
 *     def const_empty(self):
 *         return self.source.getConstEmpty()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_const_empty(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getConstEmpty();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 466, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":464
 *         return self.source.hasDefinition()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def const_empty(self):
 *         return self.source.getConstEmpty()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.const_empty.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":467
 *     def const_empty(self):
 *         return self.source.getConstEmpty()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_const_empty(self):
 *         return self.source.isConstEmpty()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_const_empty_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_const_empty_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_const_empty___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_const_empty___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":469
 *     @property
 *     def is_const_empty(self):
 *         return self.source.isConstEmpty()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isConstEmpty();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 469, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":467
 *     def const_empty(self):
 *         return self.source.getConstEmpty()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_const_empty(self):
 *         return self.source.isConstEmpty()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_const_empty.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":471
 *         return self.source.isConstEmpty()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sort_s_prop(self):
 *         return self.source.getSortSProp()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11sort_s_prop_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11sort_s_prop_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11sort_s_prop___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11sort_s_prop___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":473
 *     @property
 *     def sort_s_prop(self):
 *         return self.source.getSortSProp()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_sort_s_prop(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getSortSProp();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 473, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":471
 *         return self.source.isConstEmpty()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sort_s_prop(self):
 *         return self.source.getSortSProp()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.sort_s_prop.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":474
 *     def sort_s_prop(self):
 *         return self.source.getSortSProp()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_sort_s_prop(self):
 *         return self.source.isSortSProp()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_sort_s_prop_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_sort_s_prop_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_sort_s_prop___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_sort_s_prop___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":476
 *     @property
 *     def is_sort_s_prop(self):
 *         return self.source.isSortSProp()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isSortSProp();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 476, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":474
 *     def sort_s_prop(self):
 *         return self.source.getSortSProp()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_sort_s_prop(self):
 *         return self.source.isSortSProp()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_sort_s_prop.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":478
 *         return self.source.isSortSProp()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sort_prop(self):
 *         return self.source.getSortProp()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_prop_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_prop_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_prop___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_prop___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":480
 *     @property
 *     def sort_prop(self):
 *         return self.source.getSortProp()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_sort_prop(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getSortProp();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 480, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":478
 *         return self.source.isSortSProp()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sort_prop(self):
 *         return self.source.getSortProp()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.sort_prop.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":481
 *     def sort_prop(self):
 *         return self.source.getSortProp()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_sort_prop(self):
 *         return self.source.isSortProp()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_prop_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_prop_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_prop___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_prop___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":483
 *     @property
 *     def is_sort_prop(self):
 *         return self.source.isSortProp()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isSortProp();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 483, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":481
 *     def sort_prop(self):
 *         return self.source.getSortProp()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_sort_prop(self):
 *         return self.source.isSortProp()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_sort_prop.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":485
 *         return self.source.isSortProp()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sort_set(self):
 *         return self.source.getSortSet()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8sort_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8sort_set_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8sort_set___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8sort_set___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":487
 *     @property
 *     def sort_set(self):
 *         return self.source.getSortSet()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_sort_set(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getSortSet();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 487, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":485
 *         return self.source.isSortProp()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sort_set(self):
 *         return self.source.getSortSet()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.sort_set.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":488
 *     def sort_set(self):
 *         return self.source.getSortSet()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_sort_set(self):
 *         return self.source.isSortSet()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11is_sort_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11is_sort_set_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11is_sort_set___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11is_sort_set___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":490
 *     @property
 *     def is_sort_set(self):
 *         return self.source.isSortSet()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isSortSet();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 490, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":488
 *     def sort_set(self):
 *         return self.source.getSortSet()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_sort_set(self):
 *         return self.source.isSortSet()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_sort_set.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":492
 *         return self.source.isSortSet()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sort_type(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_type___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_type___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":497
 *         Collapsed universe
 *         """
 *         return self.source.getSortType()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_sort_type(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getSortType();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 497, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":492
 *         return self.source.isSortSet()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sort_type(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.sort_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":498
 *         """
 *         return self.source.getSortType()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_sort_type(self):
 *         return self.source.isSortType()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_type___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_type___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":500
 *     @property
 *     def is_sort_type(self):
 *         return self.source.isSortType()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isSortType();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 500, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":498
 *         """
 *         return self.source.getSortType()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_sort_type(self):
 *         return self.source.isSortType()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_sort_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":502
 *         return self.source.isSortType()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rel(self):
 *         return self.source.getRel()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3rel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3rel_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3rel___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3rel___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":504
 *     @property
 *     def rel(self):
 *         return self.source.getRel()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_rel(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getRel();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 504, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":502
 *         return self.source.isSortType()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rel(self):
 *         return self.source.getRel()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.rel.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":505
 *     def rel(self):
 *         return self.source.getRel()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_rel(self):
 *         return self.source.isRel()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_rel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_rel_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_rel___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_rel___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":507
 *     @property
 *     def is_rel(self):
 *         return self.source.isRel()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isRel();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 507, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":505
 *     def rel(self):
 *         return self.source.getRel()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_rel(self):
 *         return self.source.isRel()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_rel.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":509
 *         return self.source.isRel()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def evar(self):
 *         return self.source.getEvar()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4evar_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4evar_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4evar___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4evar___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":511
 *     @property
 *     def evar(self):
 *         return self.source.getEvar()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_evar(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getEvar();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 511, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":509
 *         return self.source.isRel()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def evar(self):
 *         return self.source.getEvar()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.evar.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":512
 *     def evar(self):
 *         return self.source.getEvar()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_evar(self):
 *         return self.source.isEvar()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_evar_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_evar_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_evar___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_evar___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":514
 *     @property
 *     def is_evar(self):
 *         return self.source.isEvar()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isEvar();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 514, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":512
 *     def evar(self):
 *         return self.source.getEvar()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_evar(self):
 *         return self.source.isEvar()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_evar.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":516
 *         return self.source.isEvar()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def evar_subst(self):
 *         return self.source.getEvarSubst()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10evar_subst_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10evar_subst_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10evar_subst___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10evar_subst___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":518
 *     @property
 *     def evar_subst(self):
 *         return self.source.getEvarSubst()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_evar_subst(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getEvarSubst();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 518, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":516
 *         return self.source.isEvar()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def evar_subst(self):
 *         return self.source.getEvarSubst()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.evar_subst.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":519
 *     def evar_subst(self):
 *         return self.source.getEvarSubst()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_evar_subst(self):
 *         return self.source.isEvarSubst()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_evar_subst_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_evar_subst_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_evar_subst___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_evar_subst___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":521
 *     @property
 *     def is_evar_subst(self):
 *         return self.source.isEvarSubst()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isEvarSubst();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 521, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":519
 *     def evar_subst(self):
 *         return self.source.getEvarSubst()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_evar_subst(self):
 *         return self.source.isEvarSubst()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_evar_subst.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":523
 *         return self.source.isEvarSubst()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cast(self):
 *         return self.source.getCast()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4cast_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4cast_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4cast___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4cast___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":525
 *     @property
 *     def cast(self):
 *         return self.source.getCast()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_cast(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getCast();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 525, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":523
 *         return self.source.isEvarSubst()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cast(self):
 *         return self.source.getCast()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.cast.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":526
 *     def cast(self):
 *         return self.source.getCast()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_cast(self):
 *         return self.source.isCast()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_cast_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_cast_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_cast___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_cast___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":528
 *     @property
 *     def is_cast(self):
 *         return self.source.isCast()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isCast();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 528, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":526
 *     def cast(self):
 *         return self.source.getCast()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_cast(self):
 *         return self.source.isCast()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_cast.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":530
 *         return self.source.isCast()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def prod(self):
 *         return self.source.getProd()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4prod_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4prod_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4prod___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4prod___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":532
 *     @property
 *     def prod(self):
 *         return self.source.getProd()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_prod(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getProd();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 532, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":530
 *         return self.source.isCast()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def prod(self):
 *         return self.source.getProd()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.prod.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":533
 *     def prod(self):
 *         return self.source.getProd()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_prod(self):
 *         return self.source.isProd()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_prod_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_prod_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_prod___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_prod___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":535
 *     @property
 *     def is_prod(self):
 *         return self.source.isProd()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isProd();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 535, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":533
 *     def prod(self):
 *         return self.source.getProd()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_prod(self):
 *         return self.source.isProd()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_prod.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":537
 *         return self.source.isProd()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def lambda_(self):
 *         return self.source.getLambda()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7lambda__1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7lambda__1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7lambda____get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7lambda____get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":539
 *     @property
 *     def lambda_(self):
 *         return self.source.getLambda()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_lambda_(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getLambda();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 539, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":537
 *         return self.source.isProd()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def lambda_(self):
 *         return self.source.getLambda()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.lambda_.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":540
 *     def lambda_(self):
 *         return self.source.getLambda()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_lambda_(self):
 *         return self.source.isLambda()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_lambda__1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_lambda__1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_lambda____get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_lambda____get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":542
 *     @property
 *     def is_lambda_(self):
 *         return self.source.isLambda()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isLambda();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 542, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":540
 *     def lambda_(self):
 *         return self.source.getLambda()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_lambda_(self):
 *         return self.source.isLambda()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_lambda_.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":544
 *         return self.source.isLambda()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def let_in(self):
 *         return self.source.getLetIn()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6let_in_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6let_in_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6let_in___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6let_in___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":546
 *     @property
 *     def let_in(self):
 *         return self.source.getLetIn()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_let_in(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getLetIn();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 546, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":544
 *         return self.source.isLambda()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def let_in(self):
 *         return self.source.getLetIn()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.let_in.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":547
 *     def let_in(self):
 *         return self.source.getLetIn()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_let_in(self):
 *         return self.source.isLetIn()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_let_in_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_let_in_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_let_in___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_let_in___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":549
 *     @property
 *     def is_let_in(self):
 *         return self.source.isLetIn()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isLetIn();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 549, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":547
 *     def let_in(self):
 *         return self.source.getLetIn()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_let_in(self):
 *         return self.source.isLetIn()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_let_in.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":551
 *         return self.source.isLetIn()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def app(self):
 *         return self.source.getApp()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3app_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3app_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3app___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3app___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":553
 *     @property
 *     def app(self):
 *         return self.source.getApp()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_app(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getApp();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 553, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":551
 *         return self.source.isLetIn()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def app(self):
 *         return self.source.getApp()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.app.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":554
 *     def app(self):
 *         return self.source.getApp()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_app(self):
 *         return self.source.isApp()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_app_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_app_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_app___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_app___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":556
 *     @property
 *     def is_app(self):
 *         return self.source.isApp()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isApp();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 556, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":554
 *     def app(self):
 *         return self.source.getApp()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_app(self):
 *         return self.source.isApp()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_app.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":558
 *         return self.source.isApp()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def case(self):
 *         return self.source.getCase()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4case_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4case_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4case___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4case___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":560
 *     @property
 *     def case(self):
 *         return self.source.getCase()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_case(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getCase();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 560, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":558
 *         return self.source.isApp()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def case(self):
 *         return self.source.getCase()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.case.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":561
 *     def case(self):
 *         return self.source.getCase()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_case(self):
 *         return self.source.isCase()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_case_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_case_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_case___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_case___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":563
 *     @property
 *     def is_case(self):
 *         return self.source.isCase()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isCase();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 563, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":561
 *     def case(self):
 *         return self.source.getCase()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_case(self):
 *         return self.source.isCase()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_case.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":565
 *         return self.source.isCase()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def case_branch(self):
 *         return self.source.getCaseBranch()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11case_branch_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11case_branch_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11case_branch___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11case_branch___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":567
 *     @property
 *     def case_branch(self):
 *         return self.source.getCaseBranch()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_case_branch(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getCaseBranch();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 567, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":565
 *         return self.source.isCase()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def case_branch(self):
 *         return self.source.getCaseBranch()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.case_branch.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":568
 *     def case_branch(self):
 *         return self.source.getCaseBranch()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_case_branch(self):
 *         return self.source.isCaseBranch()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_case_branch_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_case_branch_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_case_branch___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_case_branch___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":570
 *     @property
 *     def is_case_branch(self):
 *         return self.source.isCaseBranch()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isCaseBranch();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 570, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":568
 *     def case_branch(self):
 *         return self.source.getCaseBranch()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_case_branch(self):
 *         return self.source.isCaseBranch()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_case_branch.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":572
 *         return self.source.isCaseBranch()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fix(self):
 *         return self.source.getFix()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3fix_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3fix_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3fix___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3fix___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":574
 *     @property
 *     def fix(self):
 *         return self.source.getFix()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_fix(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getFix();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 574, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":572
 *         return self.source.isCaseBranch()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fix(self):
 *         return self.source.getFix()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.fix.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":575
 *     def fix(self):
 *         return self.source.getFix()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_fix(self):
 *         return self.source.isFix()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_fix_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_fix_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_fix___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_fix___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":577
 *     @property
 *     def is_fix(self):
 *         return self.source.isFix()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isFix();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 577, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":575
 *     def fix(self):
 *         return self.source.getFix()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_fix(self):
 *         return self.source.isFix()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_fix.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":579
 *         return self.source.isFix()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fix_fun(self):
 *         return self.source.getFixFun()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7fix_fun_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7fix_fun_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7fix_fun___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7fix_fun___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":581
 *     @property
 *     def fix_fun(self):
 *         return self.source.getFixFun()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_fix_fun(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getFixFun();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 581, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":579
 *         return self.source.isFix()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fix_fun(self):
 *         return self.source.getFixFun()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.fix_fun.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":582
 *     def fix_fun(self):
 *         return self.source.getFixFun()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_fix_fun(self):
 *         return self.source.isFixFun()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_fix_fun_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_fix_fun_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_fix_fun___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_fix_fun___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":584
 *     @property
 *     def is_fix_fun(self):
 *         return self.source.isFixFun()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isFixFun();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 584, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":582
 *     def fix_fun(self):
 *         return self.source.getFixFun()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_fix_fun(self):
 *         return self.source.isFixFun()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_fix_fun.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":586
 *         return self.source.isFixFun()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def co_fix(self):
 *         return self.source.getCoFix()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6co_fix_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6co_fix_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6co_fix___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6co_fix___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":588
 *     @property
 *     def co_fix(self):
 *         return self.source.getCoFix()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_co_fix(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getCoFix();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 588, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":586
 *         return self.source.isFixFun()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def co_fix(self):
 *         return self.source.getCoFix()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.co_fix.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":589
 *     def co_fix(self):
 *         return self.source.getCoFix()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_co_fix(self):
 *         return self.source.isCoFix()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_co_fix_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_co_fix_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_co_fix___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_co_fix___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":591
 *     @property
 *     def is_co_fix(self):
 *         return self.source.isCoFix()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isCoFix();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 591, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":589
 *     def co_fix(self):
 *         return self.source.getCoFix()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_co_fix(self):
 *         return self.source.isCoFix()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_co_fix.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":593
 *         return self.source.isCoFix()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def co_fix_fun(self):
 *         return self.source.getCoFixFun()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10co_fix_fun_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10co_fix_fun_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10co_fix_fun___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10co_fix_fun___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":595
 *     @property
 *     def co_fix_fun(self):
 *         return self.source.getCoFixFun()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_co_fix_fun(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getCoFixFun();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 595, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":593
 *         return self.source.isCoFix()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def co_fix_fun(self):
 *         return self.source.getCoFixFun()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.co_fix_fun.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":596
 *     def co_fix_fun(self):
 *         return self.source.getCoFixFun()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_co_fix_fun(self):
 *         return self.source.isCoFixFun()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_co_fix_fun_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_co_fix_fun_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_co_fix_fun___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_co_fix_fun___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":598
 *     @property
 *     def is_co_fix_fun(self):
 *         return self.source.isCoFixFun()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isCoFixFun();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 598, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":596
 *     def co_fix_fun(self):
 *         return self.source.getCoFixFun()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_co_fix_fun(self):
 *         return self.source.isCoFixFun()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_co_fix_fun.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":600
 *         return self.source.isCoFixFun()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def int(self):
 *         return IntP_Reader.init(self.source.getInt(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3int_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3int_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3int___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3int___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  IntP::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":602
 *     @property
 *     def int(self):
 *         return IntP_Reader.init(self.source.getInt(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_int(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getInt();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 602, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_11IntP_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":600
 *         return self.source.isCoFixFun()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def int(self):
 *         return IntP_Reader.init(self.source.getInt(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.int.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":603
 *     def int(self):
 *         return IntP_Reader.init(self.source.getInt(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_int(self):
 *         return self.source.isInt()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_int_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_int_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_int___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_int___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":605
 *     @property
 *     def is_int(self):
 *         return self.source.isInt()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_int(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isInt();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 605, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":603
 *     def int(self):
 *         return IntP_Reader.init(self.source.getInt(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_int(self):
 *         return self.source.isInt()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_int.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":606
 *     def is_int(self):
 *         return self.source.isInt()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_int(self):
 *         return self.source.hasInt()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7has_int_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7has_int_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7has_int___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7has_int___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":608
 *     @property
 *     def has_int(self):
 *         return self.source.hasInt()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasInt();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 608, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":606
 *     def is_int(self):
 *         return self.source.isInt()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_int(self):
 *         return self.source.hasInt()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.has_int.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":610
 *         return self.source.hasInt()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def float(self):
 *         return FloatP_Reader.init(self.source.getFloat(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5float_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5float_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5float___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5float___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  FloatP::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":612
 *     @property
 *     def float(self):
 *         return FloatP_Reader.init(self.source.getFloat(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_float(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getFloat();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 612, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_13FloatP_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":610
 *         return self.source.hasInt()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def float(self):
 *         return FloatP_Reader.init(self.source.getFloat(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.float.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":613
 *     def float(self):
 *         return FloatP_Reader.init(self.source.getFloat(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_float(self):
 *         return self.source.isFloat()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8is_float_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8is_float_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8is_float___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8is_float___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":615
 *     @property
 *     def is_float(self):
 *         return self.source.isFloat()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_float(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isFloat();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 615, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":613
 *     def float(self):
 *         return FloatP_Reader.init(self.source.getFloat(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_float(self):
 *         return self.source.isFloat()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_float.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":616
 *     def is_float(self):
 *         return self.source.isFloat()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_float(self):
 *         return self.source.hasFloat()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9has_float_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9has_float_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9has_float___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9has_float___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":618
 *     @property
 *     def has_float(self):
 *         return self.source.hasFloat()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasFloat();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 618, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":616
 *     def is_float(self):
 *         return self.source.isFloat()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_float(self):
 *         return self.source.hasFloat()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.has_float.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":620
 *         return self.source.hasFloat()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def primitive(self):
 *         temp = self.source.getPrimitive()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9primitive_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9primitive_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9primitive___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9primitive___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":622
 *     @property
 *     def primitive(self):
 *         temp = self.source.getPrimitive()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getPrimitive();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 622, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":623
 *     def primitive(self):
 *         temp = self.source.getPrimitive()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def is_primitive(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":620
 *         return self.source.hasFloat()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def primitive(self):
 *         temp = self.source.getPrimitive()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.primitive.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":624
 *         temp = self.source.getPrimitive()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def is_primitive(self):
 *         return self.source.isPrimitive()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_primitive_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_primitive_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_primitive___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_primitive___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":626
 *     @property
 *     def is_primitive(self):
 *         return self.source.isPrimitive()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_primitive(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isPrimitive();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 626, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":624
 *         temp = self.source.getPrimitive()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def is_primitive(self):
 *         return self.source.isPrimitive()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.is_primitive.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":627
 *     def is_primitive(self):
 *         return self.source.isPrimitive()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_primitive(self):
 *         return self.source.hasPrimitive()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13has_primitive_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13has_primitive_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13has_primitive___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13has_primitive___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":629
 *     @property
 *     def has_primitive(self):
 *         return self.source.hasPrimitive()             # <<<<<<<<<<<<<<
 * 
 * cdef class Definition_Status_Substituted_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasPrimitive();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 629, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":627
 *     def is_primitive(self):
 *         return self.source.isPrimitive()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_primitive(self):
 *         return self.source.hasPrimitive()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.has_primitive.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4__reduce_cython__, "Graph_Node_Label_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6__setstate_cython__, "Graph_Node_Label_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Label_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":637
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition_Status_Substituted()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 637, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 637, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 637, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Substituted_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 637, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":638
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition_Status_Substituted()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Definition::Status::Substituted>();

  /* "pytact/graph_api_capnp_cython.pyx":639
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition_Status_Substituted()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":637
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition_Status_Substituted()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":642
 * 
 *     @staticmethod
 *     cdef init(C_Definition_Status_Substituted_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Definition_Status_Substituted_Reader wrapper = Definition_Status_Substituted_Reader.__new__(Definition_Status_Substituted_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_init(Definition::Status::Substituted::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":643
 *     @staticmethod
 *     cdef init(C_Definition_Status_Substituted_Reader source, object root):
 *         cdef Definition_Status_Substituted_Reader wrapper = Definition_Status_Substituted_Reader.__new__(Definition_Status_Substituted_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 643, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":644
 *     cdef init(C_Definition_Status_Substituted_Reader source, object root):
 *         cdef Definition_Status_Substituted_Reader wrapper = Definition_Status_Substituted_Reader.__new__(Definition_Status_Substituted_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":645
 *         cdef Definition_Status_Substituted_Reader wrapper = Definition_Status_Substituted_Reader.__new__(Definition_Status_Substituted_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":646
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":642
 * 
 *     @staticmethod
 *     cdef init(C_Definition_Status_Substituted_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Definition_Status_Substituted_Reader wrapper = Definition_Status_Substituted_Reader.__new__(Definition_Status_Substituted_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Substituted_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":648
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":650
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":648
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Substituted_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":651
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":652
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":651
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Substituted_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":654
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_9dep_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_9dep_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_9dep_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":660
 *         However, a value of zero always points to the 'current' graph.
 *         """
 *         return self.source.getDepIndex()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDepIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 660, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":654
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Substituted_Reader.dep_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":662
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_10node_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_10node_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_10node_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":667
 *         The index into `Graph.nodes` where this node can be found.
 *         """
 *         return self.source.getNodeIndex()             # <<<<<<<<<<<<<<
 * 
 * cdef class Definition_Status_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getNodeIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 667, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":662
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Substituted_Reader.node_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_4__reduce_cython__, "Definition_Status_Substituted_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Substituted_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_6__setstate_cython__, "Definition_Status_Substituted_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Substituted_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Substituted_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":674
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition_Status()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 674, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 674, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 674, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 674, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":675
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition_Status()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Definition::Status>();

  /* "pytact/graph_api_capnp_cython.pyx":676
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition_Status()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":674
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition_Status()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":679
 * 
 *     @staticmethod
 *     cdef init(C_Definition_Status_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Definition_Status_Reader wrapper = Definition_Status_Reader.__new__(Definition_Status_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_init(Definition::Status::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":680
 *     @staticmethod
 *     cdef init(C_Definition_Status_Reader source, object root):
 *         cdef Definition_Status_Reader wrapper = Definition_Status_Reader.__new__(Definition_Status_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Status_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 680, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":681
 *     cdef init(C_Definition_Status_Reader source, object root):
 *         cdef Definition_Status_Reader wrapper = Definition_Status_Reader.__new__(Definition_Status_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":682
 *         cdef Definition_Status_Reader wrapper = Definition_Status_Reader.__new__(Definition_Status_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":683
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":679
 * 
 *     @staticmethod
 *     cdef init(C_Definition_Status_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Definition_Status_Reader wrapper = Definition_Status_Reader.__new__(Definition_Status_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":685
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":687
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":685
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":688
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":689
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 *     @property
 *     def which(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":688
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":690
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return Definition_Status_Which(self.source.which())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5which_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5which_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5which___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __PYX_ENUM_CLASS_DECL Definition::Status::Which __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":692
 *     @property
 *     def which(self):
 *         return Definition_Status_Which(self.source.which())             # <<<<<<<<<<<<<<
 *     @property
 *     def which_raw(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Definition_Status_Which); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    __pyx_t_3 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 692, __pyx_L1_error)
  }
  __pyx_t_4 = [](const Definition::Status::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":690
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return Definition_Status_Which(self.source.which())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.which.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":693
 *     def which(self):
 *         return Definition_Status_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_9which_raw_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_9which_raw_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_9which_raw___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __PYX_ENUM_CLASS_DECL Definition::Status::Which __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":695
 *     @property
 *     def which_raw(self):
 *         return self.source.which()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 695, __pyx_L1_error)
  }
  __pyx_t_2 = [](const Definition::Status::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":693
 *     def which(self):
 *         return Definition_Status_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.which_raw.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":697
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def original(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_8original_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_8original_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_8original___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_8original___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":702
 *         An object as originally inputted by the user.
 *         """
 *         return self.source.getOriginal()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_original(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getOriginal();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 702, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":697
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def original(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.original.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":703
 *         """
 *         return self.source.getOriginal()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_original(self):
 *         return self.source.isOriginal()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11is_original_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11is_original_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11is_original___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11is_original___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":705
 *     @property
 *     def is_original(self):
 *         return self.source.isOriginal()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isOriginal();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 705, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":703
 *         """
 *         return self.source.getOriginal()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_original(self):
 *         return self.source.isOriginal()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.is_original.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":707
 *         return self.source.isOriginal()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def discharged(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_10discharged_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_10discharged_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_10discharged___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_10discharged___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":713
 *         variables discharged into it. The node index references the original definition.
 *         """
 *         return self.source.getDischarged()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_discharged(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDischarged();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 713, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":707
 *         return self.source.isOriginal()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def discharged(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.discharged.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":714
 *         """
 *         return self.source.getDischarged()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_discharged(self):
 *         return self.source.isDischarged()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_13is_discharged_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_13is_discharged_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_13is_discharged___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_13is_discharged___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":716
 *     @property
 *     def is_discharged(self):
 *         return self.source.isDischarged()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isDischarged();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 716, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":714
 *         """
 *         return self.source.getDischarged()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_discharged(self):
 *         return self.source.isDischarged()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.is_discharged.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":718
 *         return self.source.isDischarged()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def substituted(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11substituted_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11substituted_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11substituted___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11substituted___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Definition::Status::Substituted::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":724
 *         We reference the original definition.
 *         """
 *         return Definition_Status_Substituted_Reader.init(self.source.getSubstituted(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_substituted(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getSubstituted();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 724, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":718
 *         return self.source.isDischarged()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def substituted(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.substituted.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":725
 *         """
 *         return Definition_Status_Substituted_Reader.init(self.source.getSubstituted(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_substituted(self):
 *         return self.source.isSubstituted()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_14is_substituted_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_14is_substituted_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_14is_substituted___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_14is_substituted___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":727
 *     @property
 *     def is_substituted(self):
 *         return self.source.isSubstituted()             # <<<<<<<<<<<<<<
 * 
 * cdef class ProofStep_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isSubstituted();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 727, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":725
 *         """
 *         return Definition_Status_Substituted_Reader.init(self.source.getSubstituted(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_substituted(self):
 *         return self.source.isSubstituted()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.is_substituted.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_4__reduce_cython__, "Definition_Status_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_6__setstate_cython__, "Definition_Status_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Definition_Status_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":734
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofStep()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 734, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 734, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 734, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 734, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":735
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofStep()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<ProofStep>();

  /* "pytact/graph_api_capnp_cython.pyx":736
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofStep()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":734
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofStep()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":739
 * 
 *     @staticmethod
 *     cdef init(C_ProofStep_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofStep_Reader wrapper = ProofStep_Reader.__new__(ProofStep_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_init(ProofStep::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":740
 *     @staticmethod
 *     cdef init(C_ProofStep_Reader source, object root):
 *         cdef ProofStep_Reader wrapper = ProofStep_Reader.__new__(ProofStep_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 740, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":741
 *     cdef init(C_ProofStep_Reader source, object root):
 *         cdef ProofStep_Reader wrapper = ProofStep_Reader.__new__(ProofStep_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":742
 *         cdef ProofStep_Reader wrapper = ProofStep_Reader.__new__(ProofStep_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":743
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":739
 * 
 *     @staticmethod
 *     cdef init(C_ProofStep_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofStep_Reader wrapper = ProofStep_Reader.__new__(ProofStep_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":745
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":747
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":745
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":748
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":749
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":748
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":751
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactic(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6tactic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6tactic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6tactic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6tactic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  ProofStep::Tactic::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":757
 *         individual outcomes, because they may be different for each outcome.
 *         """
 *         return ProofStep_Tactic_Reader.init(self.source.getTactic(), self.root)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getTactic();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 757, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":751
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactic(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader.tactic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":759
 *         return ProofStep_Tactic_Reader.init(self.source.getTactic(), self.root)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def outcomes(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_8outcomes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_8outcomes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_8outcomes___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_8outcomes___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Outcome, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":764
 *         A list of transformations of proof states to other proof states, as executed by the tactic of the proof step
 *         """
 *         return Outcome_Reader_List.init(self.source.getOutcomes(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_outcomes(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getOutcomes();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 764, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":759
 *         return ProofStep_Tactic_Reader.init(self.source.getTactic(), self.root)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def outcomes(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader.outcomes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":765
 *         """
 *         return Outcome_Reader_List.init(self.source.getOutcomes(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_outcomes(self):
 *         return self.source.hasOutcomes()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_12has_outcomes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_12has_outcomes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_12has_outcomes___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_12has_outcomes___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":767
 *     @property
 *     def has_outcomes(self):
 *         return self.source.hasOutcomes()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.ProofStep.schema = _Schema()._init(getProofStep_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasOutcomes();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 767, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":765
 *         """
 *         return Outcome_Reader_List.init(self.source.getOutcomes(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_outcomes(self):
 *         return self.source.hasOutcomes()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader.has_outcomes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_4__reduce_cython__, "ProofStep_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6__setstate_cython__, "ProofStep_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":778
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asTactic()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 778, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 778, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 778, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 778, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":779
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asTactic()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Tactic>();

  /* "pytact/graph_api_capnp_cython.pyx":780
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asTactic()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":778
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asTactic()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":783
 * 
 *     @staticmethod
 *     cdef init(C_Tactic_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Tactic_Reader wrapper = Tactic_Reader.__new__(Tactic_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_13Tactic_Reader_init(Tactic::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":784
 *     @staticmethod
 *     cdef init(C_Tactic_Reader source, object root):
 *         cdef Tactic_Reader wrapper = Tactic_Reader.__new__(Tactic_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Tactic_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 784, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":785
 *     cdef init(C_Tactic_Reader source, object root):
 *         cdef Tactic_Reader wrapper = Tactic_Reader.__new__(Tactic_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":786
 *         cdef Tactic_Reader wrapper = Tactic_Reader.__new__(Tactic_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":787
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":783
 * 
 *     @staticmethod
 *     cdef init(C_Tactic_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Tactic_Reader wrapper = Tactic_Reader.__new__(Tactic_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":789
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":791
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":789
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":792
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":793
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":792
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":795
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ident(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5ident_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5ident_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5ident___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5ident___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":802
 *         of the syntax tree of the tactic with all of it's arguments removed.
 *         """
 *         return self.source.getIdent()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getIdent();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 802, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":795
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ident(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.ident.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":804
 *         return self.source.getIdent()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def text(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":812
 *         hypotheses as much as possible, and instead asking Coq to invent a name.
 *         """
 *         temp = self.source.getText()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 812, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":813
 *         """
 *         temp = self.source.getText()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_text(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":804
 *         return self.source.getIdent()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def text(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":814
 *         temp = self.source.getText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_text(self):
 *         return self.source.hasText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_8has_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_8has_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_8has_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_8has_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":816
 *     @property
 *     def has_text(self):
 *         return self.source.hasText()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 816, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":814
 *         temp = self.source.getText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_text(self):
 *         return self.source.hasText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.has_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":818
 *         return self.source.hasText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base_text(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_9base_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_9base_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_9base_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_9base_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":827
 *         using different partially-qualified names.
 *         """
 *         temp = self.source.getBaseText()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getBaseText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 827, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":828
 *         """
 *         temp = self.source.getBaseText()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_base_text(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":818
 *         return self.source.hasText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base_text(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.base_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":829
 *         temp = self.source.getBaseText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_base_text(self):
 *         return self.source.hasBaseText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_13has_base_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_13has_base_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_13has_base_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_13has_base_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":831
 *     @property
 *     def has_base_text(self):
 *         return self.source.hasBaseText()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasBaseText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 831, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":829
 *         temp = self.source.getBaseText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_base_text(self):
 *         return self.source.hasBaseText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.has_base_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":833
 *         return self.source.hasBaseText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def interm_text(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_11interm_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_11interm_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_11interm_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_11interm_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":839
 *         It comes with the same caveats as `baseText`.
 *         """
 *         temp = self.source.getIntermText()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getIntermText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 839, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":840
 *         """
 *         temp = self.source.getIntermText()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_interm_text(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":833
 *         return self.source.hasBaseText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def interm_text(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.interm_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":841
 *         temp = self.source.getIntermText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_interm_text(self):
 *         return self.source.hasIntermText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_15has_interm_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_15has_interm_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_15has_interm_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_15has_interm_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":843
 *     @property
 *     def has_interm_text(self):
 *         return self.source.hasIntermText()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasIntermText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 843, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":841
 *         temp = self.source.getIntermText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_interm_text(self):
 *         return self.source.hasIntermText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.has_interm_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":845
 *         return self.source.hasIntermText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def exact(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5exact_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5exact_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5exact___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5exact___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":854
 *         original tactic inputted by the user.
 *         """
 *         return self.source.getExact()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getExact();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 854, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":845
 *         return self.source.hasIntermText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def exact(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.exact.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":856
 *         return self.source.getExact()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def text_non_anonymous(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_18text_non_anonymous_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_18text_non_anonymous_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_18text_non_anonymous___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_18text_non_anonymous___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":861
 *         Same as `text` except that the tactic is not postprocessed to remove explicit naming of new hypotheses.
 *         """
 *         temp = self.source.getTextNonAnonymous()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getTextNonAnonymous();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 861, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":862
 *         """
 *         temp = self.source.getTextNonAnonymous()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_text_non_anonymous(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":856
 *         return self.source.getExact()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def text_non_anonymous(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.text_non_anonymous.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":863
 *         temp = self.source.getTextNonAnonymous()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_text_non_anonymous(self):
 *         return self.source.hasTextNonAnonymous()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_22has_text_non_anonymous_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_22has_text_non_anonymous_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_22has_text_non_anonymous___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_22has_text_non_anonymous___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":865
 *     @property
 *     def has_text_non_anonymous(self):
 *         return self.source.hasTextNonAnonymous()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.Tactic.schema = _Schema()._init(getTactic_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTextNonAnonymous();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 865, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":863
 *         temp = self.source.getTextNonAnonymous()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_text_non_anonymous(self):
 *         return self.source.hasTextNonAnonymous()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.has_text_non_anonymous.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4__reduce_cython__, "Tactic_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_13Tactic_Reader_6__setstate_cython__, "Tactic_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_13Tactic_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_13Tactic_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Tactic_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":874
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofStep_Tactic()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 874, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 874, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 874, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 874, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":875
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofStep_Tactic()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<ProofStep::Tactic>();

  /* "pytact/graph_api_capnp_cython.pyx":876
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofStep_Tactic()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":874
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofStep_Tactic()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":879
 * 
 *     @staticmethod
 *     cdef init(C_ProofStep_Tactic_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofStep_Tactic_Reader wrapper = ProofStep_Tactic_Reader.__new__(ProofStep_Tactic_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_init(ProofStep::Tactic::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":880
 *     @staticmethod
 *     cdef init(C_ProofStep_Tactic_Reader source, object root):
 *         cdef ProofStep_Tactic_Reader wrapper = ProofStep_Tactic_Reader.__new__(ProofStep_Tactic_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 880, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":881
 *     cdef init(C_ProofStep_Tactic_Reader source, object root):
 *         cdef ProofStep_Tactic_Reader wrapper = ProofStep_Tactic_Reader.__new__(ProofStep_Tactic_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":882
 *         cdef ProofStep_Tactic_Reader wrapper = ProofStep_Tactic_Reader.__new__(ProofStep_Tactic_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":883
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":879
 * 
 *     @staticmethod
 *     cdef init(C_ProofStep_Tactic_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofStep_Tactic_Reader wrapper = ProofStep_Tactic_Reader.__new__(ProofStep_Tactic_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":885
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":887
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":885
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":888
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":889
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 *     @property
 *     def which(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":888
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":890
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return ProofStep_Tactic_Which(self.source.which())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5which_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5which_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5which___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __PYX_ENUM_CLASS_DECL ProofStep::Tactic::Which __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":892
 *     @property
 *     def which(self):
 *         return ProofStep_Tactic_Which(self.source.which())             # <<<<<<<<<<<<<<
 *     @property
 *     def which_raw(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ProofStep_Tactic_Which); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    __pyx_t_3 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 892, __pyx_L1_error)
  }
  __pyx_t_4 = [](const ProofStep::Tactic::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":890
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return ProofStep_Tactic_Which(self.source.which())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.which.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":893
 *     def which(self):
 *         return ProofStep_Tactic_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9which_raw_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9which_raw_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9which_raw___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __PYX_ENUM_CLASS_DECL ProofStep::Tactic::Which __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":895
 *     @property
 *     def which_raw(self):
 *         return self.source.which()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 895, __pyx_L1_error)
  }
  __pyx_t_2 = [](const ProofStep::Tactic::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":893
 *     def which(self):
 *         return ProofStep_Tactic_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.which_raw.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":897
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unknown(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7unknown_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7unknown_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7unknown___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7unknown___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":904
 *         happens for tactics that are known to be unsafe like `change_no_check`, `fix`, `cofix` and more.
 *         """
 *         return self.source.getUnknown()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_unknown(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getUnknown();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 904, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":897
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unknown(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.unknown.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":905
 *         """
 *         return self.source.getUnknown()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_unknown(self):
 *         return self.source.isUnknown()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_10is_unknown_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_10is_unknown_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_10is_unknown___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_10is_unknown___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":907
 *     @property
 *     def is_unknown(self):
 *         return self.source.isUnknown()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isUnknown();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 907, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":905
 *         """
 *         return self.source.getUnknown()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_unknown(self):
 *         return self.source.isUnknown()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.is_unknown.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":909
 *         return self.source.isUnknown()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def known(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5known_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5known_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5known___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5known___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Tactic::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":914
 *         The tactic
 *         """
 *         return Tactic_Reader.init(self.source.getKnown(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_known(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getKnown();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 914, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_13Tactic_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":909
 *         return self.source.isUnknown()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def known(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.known.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":915
 *         """
 *         return Tactic_Reader.init(self.source.getKnown(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_known(self):
 *         return self.source.isKnown()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_8is_known_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_8is_known_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_8is_known___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_8is_known___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":917
 *     @property
 *     def is_known(self):
 *         return self.source.isKnown()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_known(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isKnown();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 917, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":915
 *         """
 *         return Tactic_Reader.init(self.source.getKnown(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_known(self):
 *         return self.source.isKnown()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.is_known.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":918
 *     def is_known(self):
 *         return self.source.isKnown()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_known(self):
 *         return self.source.hasKnown()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9has_known_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9has_known_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9has_known___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9has_known___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":920
 *     @property
 *     def has_known(self):
 *         return self.source.hasKnown()             # <<<<<<<<<<<<<<
 * 
 * cdef class Outcome_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasKnown();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 920, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":918
 *     def is_known(self):
 *         return self.source.isKnown()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_known(self):
 *         return self.source.hasKnown()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.has_known.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_4__reduce_cython__, "ProofStep_Tactic_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_6__setstate_cython__, "ProofStep_Tactic_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":927
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asOutcome()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 927, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 927, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 927, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 927, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":928
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asOutcome()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Outcome>();

  /* "pytact/graph_api_capnp_cython.pyx":929
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asOutcome()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":927
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asOutcome()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":932
 * 
 *     @staticmethod
 *     cdef init(C_Outcome_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Outcome_Reader wrapper = Outcome_Reader.__new__(Outcome_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_14Outcome_Reader_init(Outcome::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":933
 *     @staticmethod
 *     cdef init(C_Outcome_Reader source, object root):
 *         cdef Outcome_Reader wrapper = Outcome_Reader.__new__(Outcome_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 933, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":934
 *     cdef init(C_Outcome_Reader source, object root):
 *         cdef Outcome_Reader wrapper = Outcome_Reader.__new__(Outcome_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":935
 *         cdef Outcome_Reader wrapper = Outcome_Reader.__new__(Outcome_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":936
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":932
 * 
 *     @staticmethod
 *     cdef init(C_Outcome_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Outcome_Reader wrapper = Outcome_Reader.__new__(Outcome_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":938
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":940
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":938
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":941
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":942
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":941
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":944
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def before(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6before_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6before_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6before___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6before___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  ProofState::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":949
 *         The proof state before the tactic execution.
 *         """
 *         return ProofState_Reader.init(self.source.getBefore(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_before(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getBefore();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 949, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_17ProofState_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":944
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def before(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.before.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":950
 *         """
 *         return ProofState_Reader.init(self.source.getBefore(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_before(self):
 *         return self.source.hasBefore()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_10has_before_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_10has_before_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_10has_before___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_10has_before___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":952
 *     @property
 *     def has_before(self):
 *         return self.source.hasBefore()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasBefore();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 952, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":950
 *         """
 *         return ProofState_Reader.init(self.source.getBefore(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_before(self):
 *         return self.source.hasBefore()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.has_before.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":954
 *         return self.source.hasBefore()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def after(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5after_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5after_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5after___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5after___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<ProofState, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":959
 *         The new proof states that were generated by the tactic.
 *         """
 *         return ProofState_Reader_List.init(self.source.getAfter(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_after(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getAfter();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 959, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":954
 *         return self.source.hasBefore()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def after(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.after.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":960
 *         """
 *         return ProofState_Reader_List.init(self.source.getAfter(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_after(self):
 *         return self.source.hasAfter()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9has_after_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9has_after_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9has_after___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9has_after___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":962
 *     @property
 *     def has_after(self):
 *         return self.source.hasAfter()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasAfter();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 962, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":960
 *         """
 *         return ProofState_Reader_List.init(self.source.getAfter(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_after(self):
 *         return self.source.hasAfter()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.has_after.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":964
 *         return self.source.hasAfter()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def term(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4term_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4term_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4term___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4term___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Outcome::Term::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":971
 *         before state.
 *         """
 *         return Outcome_Term_Reader.init(self.source.getTerm(), self.root)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getTerm();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 971, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":964
 *         return self.source.hasAfter()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def term(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.term.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":973
 *         return Outcome_Term_Reader.init(self.source.getTerm(), self.root)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def term_text(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9term_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9term_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9term_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9term_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":978
 *         A textual representation of the proof term.
 *         """
 *         temp = self.source.getTermText()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getTermText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 978, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":979
 *         """
 *         temp = self.source.getTermText()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_term_text(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":973
 *         return Outcome_Term_Reader.init(self.source.getTerm(), self.root)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def term_text(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.term_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":980
 *         temp = self.source.getTermText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_term_text(self):
 *         return self.source.hasTermText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_13has_term_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_13has_term_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_13has_term_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_13has_term_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":982
 *     @property
 *     def has_term_text(self):
 *         return self.source.hasTermText()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTermText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 982, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":980
 *         temp = self.source.getTermText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_term_text(self):
 *         return self.source.hasTermText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.has_term_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":984
 *         return self.source.hasTermText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactic_arguments(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_16tactic_arguments_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_16tactic_arguments_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_16tactic_arguments___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_16tactic_arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":990
 *         `ProofStep.tactic`.
 *         """
 *         return Argument_Reader_List.init(self.source.getTacticArguments(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_tactic_arguments(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getTacticArguments();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 990, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":984
 *         return self.source.hasTermText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactic_arguments(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.tactic_arguments.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":991
 *         """
 *         return Argument_Reader_List.init(self.source.getTacticArguments(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactic_arguments(self):
 *         return self.source.hasTacticArguments()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_20has_tactic_arguments_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_20has_tactic_arguments_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_20has_tactic_arguments___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_20has_tactic_arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":993
 *     @property
 *     def has_tactic_arguments(self):
 *         return self.source.hasTacticArguments()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.Outcome.schema = _Schema()._init(getOutcome_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTacticArguments();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 993, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":991
 *         """
 *         return Argument_Reader_List.init(self.source.getTacticArguments(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactic_arguments(self):
 *         return self.source.hasTacticArguments()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.has_tactic_arguments.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4__reduce_cython__, "Outcome_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6__setstate_cython__, "Outcome_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1003
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asOutcome_Term()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1003, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1003, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1003, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Term_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1003, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1004
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asOutcome_Term()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Outcome::Term>();

  /* "pytact/graph_api_capnp_cython.pyx":1005
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asOutcome_Term()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1003
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asOutcome_Term()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1008
 * 
 *     @staticmethod
 *     cdef init(C_Outcome_Term_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Outcome_Term_Reader wrapper = Outcome_Term_Reader.__new__(Outcome_Term_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_init(Outcome::Term::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1009
 *     @staticmethod
 *     cdef init(C_Outcome_Term_Reader source, object root):
 *         cdef Outcome_Term_Reader wrapper = Outcome_Term_Reader.__new__(Outcome_Term_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1009, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1010
 *     cdef init(C_Outcome_Term_Reader source, object root):
 *         cdef Outcome_Term_Reader wrapper = Outcome_Term_Reader.__new__(Outcome_Term_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1011
 *         cdef Outcome_Term_Reader wrapper = Outcome_Term_Reader.__new__(Outcome_Term_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1012
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1008
 * 
 *     @staticmethod
 *     cdef init(C_Outcome_Term_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Outcome_Term_Reader wrapper = Outcome_Term_Reader.__new__(Outcome_Term_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Term_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1014
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1016
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1014
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Term_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1017
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1018
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1017
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Term_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1020
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_9dep_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_9dep_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_9dep_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1026
 *         However, a value of zero always points to the 'current' graph.
 *         """
 *         return self.source.getDepIndex()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDepIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1026, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1020
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Term_Reader.dep_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1028
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_10node_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_10node_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_10node_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1033
 *         The index into `Graph.nodes` where this node can be found.
 *         """
 *         return self.source.getNodeIndex()             # <<<<<<<<<<<<<<
 * 
 * cdef class ProofState_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getNodeIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1033, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1028
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Term_Reader.node_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_4__reduce_cython__, "Outcome_Term_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Term_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_6__setstate_cython__, "Outcome_Term_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Term_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Term_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1040
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofState()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1040, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1040, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1040, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1040, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1041
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofState()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<ProofState>();

  /* "pytact/graph_api_capnp_cython.pyx":1042
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofState()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1040
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofState()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1045
 * 
 *     @staticmethod
 *     cdef init(C_ProofState_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofState_Reader wrapper = ProofState_Reader.__new__(ProofState_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_17ProofState_Reader_init(ProofState::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1046
 *     @staticmethod
 *     cdef init(C_ProofState_Reader source, object root):
 *         cdef ProofState_Reader wrapper = ProofState_Reader.__new__(ProofState_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1046, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1047
 *     cdef init(C_ProofState_Reader source, object root):
 *         cdef ProofState_Reader wrapper = ProofState_Reader.__new__(ProofState_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1048
 *         cdef ProofState_Reader wrapper = ProofState_Reader.__new__(ProofState_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1049
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1045
 * 
 *     @staticmethod
 *     cdef init(C_ProofState_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofState_Reader wrapper = ProofState_Reader.__new__(ProofState_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1051
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1053
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1051
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1054
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1055
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1054
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1057
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def root(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4root_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4root_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4root___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4root___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  ProofState::Root::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1062
 *         The entry-point of the proof state, all nodes that are 'part of' the proof state are reachable from here.
 *         """
 *         return ProofState_Root_Reader.init(self.source.getRoot(), self.root)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getRoot();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1062, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1057
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def root(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.root.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1064
 *         return ProofState_Root_Reader.init(self.source.getRoot(), self.root)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7context_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7context_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7context___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7context___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1070
 *         these nodes are also reachable from the root of the proof state.
 *         """
 *         return Node_Reader_List.init(self.source.getContext(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_context(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getContext();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1070, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_16Node_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1064
 *         return ProofState_Root_Reader.init(self.source.getRoot(), self.root)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.context.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1071
 *         """
 *         return Node_Reader_List.init(self.source.getContext(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_context(self):
 *         return self.source.hasContext()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_11has_context_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_11has_context_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_11has_context___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_11has_context___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1073
 *     @property
 *     def has_context(self):
 *         return self.source.hasContext()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasContext();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1073, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1071
 *         """
 *         return Node_Reader_List.init(self.source.getContext(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_context(self):
 *         return self.source.hasContext()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.has_context.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1075
 *         return self.source.hasContext()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context_names(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_13context_names_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_13context_names_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_13context_names___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_13context_names___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1083
 *         the context.
 *         """
 *         return String_List.init(self.source.getContextNames(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_context_names(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getContextNames();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1083, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_11String_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1075
 *         return self.source.hasContext()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context_names(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.context_names.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1084
 *         """
 *         return String_List.init(self.source.getContextNames(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_context_names(self):
 *         return self.source.hasContextNames()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_17has_context_names_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_17has_context_names_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_17has_context_names___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_17has_context_names___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1086
 *     @property
 *     def has_context_names(self):
 *         return self.source.hasContextNames()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasContextNames();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1086, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1084
 *         """
 *         return String_List.init(self.source.getContextNames(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_context_names(self):
 *         return self.source.hasContextNames()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.has_context_names.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1088
 *         return self.source.hasContextNames()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context_text(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_12context_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_12context_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_12context_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_12context_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1093
 *         A textual representation of the type/definition of context nodes
 *         """
 *         return String_List.init(self.source.getContextText(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_context_text(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getContextText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1093, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_11String_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1088
 *         return self.source.hasContextNames()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context_text(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.context_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1094
 *         """
 *         return String_List.init(self.source.getContextText(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_context_text(self):
 *         return self.source.hasContextText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_16has_context_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_16has_context_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_16has_context_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_16has_context_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1096
 *     @property
 *     def has_context_text(self):
 *         return self.source.hasContextText()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasContextText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1096, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1094
 *         """
 *         return String_List.init(self.source.getContextText(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_context_text(self):
 *         return self.source.hasContextText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.has_context_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1098
 *         return self.source.hasContextText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def conclusion_text(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_15conclusion_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_15conclusion_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_15conclusion_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_15conclusion_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1103
 *         A textual representation of the conclusion of the proof state.
 *         """
 *         temp = self.source.getConclusionText()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getConclusionText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1103, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":1104
 *         """
 *         temp = self.source.getConclusionText()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_conclusion_text(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1098
 *         return self.source.hasContextText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def conclusion_text(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.conclusion_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1105
 *         temp = self.source.getConclusionText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_conclusion_text(self):
 *         return self.source.hasConclusionText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_19has_conclusion_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_19has_conclusion_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_19has_conclusion_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_19has_conclusion_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1107
 *     @property
 *     def has_conclusion_text(self):
 *         return self.source.hasConclusionText()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasConclusionText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1107, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1105
 *         temp = self.source.getConclusionText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_conclusion_text(self):
 *         return self.source.hasConclusionText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.has_conclusion_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1109
 *         return self.source.hasConclusionText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def text(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1114
 *         A textual representation of the proof state.
 *         """
 *         temp = self.source.getText()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1114, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":1115
 *         """
 *         temp = self.source.getText()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_text(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1109
 *         return self.source.hasConclusionText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def text(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1116
 *         temp = self.source.getText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_text(self):
 *         return self.source.hasText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_8has_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_8has_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_8has_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_8has_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1118
 *     @property
 *     def has_text(self):
 *         return self.source.hasText()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1118, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1116
 *         temp = self.source.getText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_text(self):
 *         return self.source.hasText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.has_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1120
 *         return self.source.hasText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def id(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_2id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_2id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_2id___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_2id___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1132
 *               in the final proof term that was generated.
 *         """
 *         return self.source.getId()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.ProofState.schema = _Schema()._init(getProofState_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getId();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1132, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1120
 *         return self.source.hasText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def id(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4__reduce_cython__, "ProofState_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_17ProofState_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_17ProofState_Reader_6__setstate_cython__, "ProofState_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17ProofState_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_17ProofState_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1140
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asArgument()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1140, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1140, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1140, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1140, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1141
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asArgument()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Argument>();

  /* "pytact/graph_api_capnp_cython.pyx":1142
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asArgument()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1140
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asArgument()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1145
 * 
 *     @staticmethod
 *     cdef init(C_Argument_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Argument_Reader wrapper = Argument_Reader.__new__(Argument_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_15Argument_Reader_init(Argument::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1146
 *     @staticmethod
 *     cdef init(C_Argument_Reader source, object root):
 *         cdef Argument_Reader wrapper = Argument_Reader.__new__(Argument_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1146, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1147
 *     cdef init(C_Argument_Reader source, object root):
 *         cdef Argument_Reader wrapper = Argument_Reader.__new__(Argument_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1148
 *         cdef Argument_Reader wrapper = Argument_Reader.__new__(Argument_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1149
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1145
 * 
 *     @staticmethod
 *     cdef init(C_Argument_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Argument_Reader wrapper = Argument_Reader.__new__(Argument_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1151
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1153
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1151
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1154
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1155
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 *     @property
 *     def which(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1154
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1156
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return Argument_Which(self.source.which())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_5which_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_5which_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_5which___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __PYX_ENUM_CLASS_DECL Argument::Which __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1158
 *     @property
 *     def which(self):
 *         return Argument_Which(self.source.which())             # <<<<<<<<<<<<<<
 *     @property
 *     def which_raw(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Argument_Which); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    __pyx_t_3 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1158, __pyx_L1_error)
  }
  __pyx_t_4 = [](const Argument::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1156
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return Argument_Which(self.source.which())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.which.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1159
 *     def which(self):
 *         return Argument_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_9which_raw_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_9which_raw_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_9which_raw___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __PYX_ENUM_CLASS_DECL Argument::Which __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1161
 *     @property
 *     def which_raw(self):
 *         return self.source.which()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1161, __pyx_L1_error)
  }
  __pyx_t_2 = [](const Argument::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1159
 *     def which(self):
 *         return Argument_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.which_raw.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1163
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unresolvable(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_12unresolvable_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_12unresolvable_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_12unresolvable___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_12unresolvable___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1168
 *         An argument that is currently unresolvable due to limitations of the extraction process.
 *         """
 *         return self.source.getUnresolvable()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_unresolvable(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getUnresolvable();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1168, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1163
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unresolvable(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.unresolvable.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1169
 *         """
 *         return self.source.getUnresolvable()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_unresolvable(self):
 *         return self.source.isUnresolvable()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_15is_unresolvable_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_15is_unresolvable_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_15is_unresolvable___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_15is_unresolvable___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1171
 *     @property
 *     def is_unresolvable(self):
 *         return self.source.isUnresolvable()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isUnresolvable();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1171, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1169
 *         """
 *         return self.source.getUnresolvable()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_unresolvable(self):
 *         return self.source.isUnresolvable()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.is_unresolvable.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1173
 *         return self.source.isUnresolvable()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def term(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_4term_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_4term_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_4term___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_4term___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Argument::Term::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1178
 *         The root of a graph representing an argument that is a term in the calculus of constructions.
 *         """
 *         return Argument_Term_Reader.init(self.source.getTerm(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_term(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getTerm();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1178, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1173
 *         return self.source.isUnresolvable()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def term(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.term.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1179
 *         """
 *         return Argument_Term_Reader.init(self.source.getTerm(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_term(self):
 *         return self.source.isTerm()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_7is_term_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_7is_term_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_7is_term___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_7is_term___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1181
 *     @property
 *     def is_term(self):
 *         return self.source.isTerm()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.Argument.schema = _Schema()._init(getArgument_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isTerm();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1181, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1179
 *         """
 *         return Argument_Term_Reader.init(self.source.getTerm(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_term(self):
 *         return self.source.isTerm()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.is_term.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_15Argument_Reader_4__reduce_cython__, "Argument_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_15Argument_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_15Argument_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_15Argument_Reader_6__setstate_cython__, "Argument_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_15Argument_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_15Argument_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_15Argument_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1189
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofState_Root()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1189, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1189, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1189, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Root_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1189, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1190
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofState_Root()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<ProofState::Root>();

  /* "pytact/graph_api_capnp_cython.pyx":1191
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofState_Root()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1189
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofState_Root()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1194
 * 
 *     @staticmethod
 *     cdef init(C_ProofState_Root_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofState_Root_Reader wrapper = ProofState_Root_Reader.__new__(ProofState_Root_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_init(ProofState::Root::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1195
 *     @staticmethod
 *     cdef init(C_ProofState_Root_Reader source, object root):
 *         cdef ProofState_Root_Reader wrapper = ProofState_Root_Reader.__new__(ProofState_Root_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1195, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1196
 *     cdef init(C_ProofState_Root_Reader source, object root):
 *         cdef ProofState_Root_Reader wrapper = ProofState_Root_Reader.__new__(ProofState_Root_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1197
 *         cdef ProofState_Root_Reader wrapper = ProofState_Root_Reader.__new__(ProofState_Root_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1198
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1194
 * 
 *     @staticmethod
 *     cdef init(C_ProofState_Root_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofState_Root_Reader wrapper = ProofState_Root_Reader.__new__(ProofState_Root_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Root_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1200
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1202
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1200
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Root_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1203
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1204
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1203
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Root_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1206
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_9dep_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_9dep_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_9dep_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1212
 *         However, a value of zero always points to the 'current' graph.
 *         """
 *         return self.source.getDepIndex()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDepIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1212, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1206
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Root_Reader.dep_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1214
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_10node_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_10node_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_10node_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1219
 *         The index into `Graph.nodes` where this node can be found.
 *         """
 *         return self.source.getNodeIndex()             # <<<<<<<<<<<<<<
 * 
 * cdef class Node_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getNodeIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1219, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1214
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Root_Reader.node_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_4__reduce_cython__, "ProofState_Root_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Root_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_6__setstate_cython__, "ProofState_Root_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Root_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Root_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1226
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asNode()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1226, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1226, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1226, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1226, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1227
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asNode()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Node>();

  /* "pytact/graph_api_capnp_cython.pyx":1228
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asNode()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1226
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asNode()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1231
 * 
 *     @staticmethod
 *     cdef init(C_Node_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Node_Reader wrapper = Node_Reader.__new__(Node_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11Node_Reader_init(Node::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1232
 *     @staticmethod
 *     cdef init(C_Node_Reader source, object root):
 *         cdef Node_Reader wrapper = Node_Reader.__new__(Node_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Node_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1232, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1233
 *     cdef init(C_Node_Reader source, object root):
 *         cdef Node_Reader wrapper = Node_Reader.__new__(Node_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1234
 *         cdef Node_Reader wrapper = Node_Reader.__new__(Node_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1235
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1231
 * 
 *     @staticmethod
 *     cdef init(C_Node_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Node_Reader wrapper = Node_Reader.__new__(Node_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1237
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1239
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1237
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1240
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1241
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1240
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1243
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_9dep_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_9dep_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_9dep_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1249
 *         However, a value of zero always points to the 'current' graph.
 *         """
 *         return self.source.getDepIndex()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDepIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1249, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1243
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader.dep_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1251
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_10node_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_10node_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_10node_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1256
 *         The index into `Graph.nodes` where this node can be found.
 *         """
 *         return self.source.getNodeIndex()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.Node.schema = _Schema()._init(getNode_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getNodeIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1256, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1251
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader.node_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11Node_Reader_4__reduce_cython__, "Node_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11Node_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Node_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11Node_Reader_6__setstate_cython__, "Node_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11Node_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Node_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Node_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1264
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asArgument_Term()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1264, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1264, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1264, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Term_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1264, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1265
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asArgument_Term()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Argument::Term>();

  /* "pytact/graph_api_capnp_cython.pyx":1266
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asArgument_Term()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1264
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asArgument_Term()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1269
 * 
 *     @staticmethod
 *     cdef init(C_Argument_Term_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Argument_Term_Reader wrapper = Argument_Term_Reader.__new__(Argument_Term_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_init(Argument::Term::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1270
 *     @staticmethod
 *     cdef init(C_Argument_Term_Reader source, object root):
 *         cdef Argument_Term_Reader wrapper = Argument_Term_Reader.__new__(Argument_Term_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Term_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1270, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1271
 *     cdef init(C_Argument_Term_Reader source, object root):
 *         cdef Argument_Term_Reader wrapper = Argument_Term_Reader.__new__(Argument_Term_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1272
 *         cdef Argument_Term_Reader wrapper = Argument_Term_Reader.__new__(Argument_Term_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1273
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1269
 * 
 *     @staticmethod
 *     cdef init(C_Argument_Term_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Argument_Term_Reader wrapper = Argument_Term_Reader.__new__(Argument_Term_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Term_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1275
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1277
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1275
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Term_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1278
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1279
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1278
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Term_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1281
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_9dep_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_9dep_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_9dep_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1287
 *         However, a value of zero always points to the 'current' graph.
 *         """
 *         return self.source.getDepIndex()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDepIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1287, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1281
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Term_Reader.dep_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1289
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_10node_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_10node_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_10node_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1294
 *         The index into `Graph.nodes` where this node can be found.
 *         """
 *         return self.source.getNodeIndex()             # <<<<<<<<<<<<<<
 * 
 * cdef class Graph_EdgeTarget_Target_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getNodeIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1294, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1289
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Term_Reader.node_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_4__reduce_cython__, "Argument_Term_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Term_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_6__setstate_cython__, "Argument_Term_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Term_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Term_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1298
 * cdef class Graph_EdgeTarget_Target_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_EdgeTarget_Target()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1298, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1298, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1298, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Target_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1298, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1299
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_EdgeTarget_Target()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Graph::EdgeTarget::Target>();

  /* "pytact/graph_api_capnp_cython.pyx":1300
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_EdgeTarget_Target()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1298
 * cdef class Graph_EdgeTarget_Target_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_EdgeTarget_Target()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1303
 * 
 *     @staticmethod
 *     cdef init(C_Graph_EdgeTarget_Target_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_EdgeTarget_Target_Reader wrapper = Graph_EdgeTarget_Target_Reader.__new__(Graph_EdgeTarget_Target_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_init(Graph::EdgeTarget::Target::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1304
 *     @staticmethod
 *     cdef init(C_Graph_EdgeTarget_Target_Reader source, object root):
 *         cdef Graph_EdgeTarget_Target_Reader wrapper = Graph_EdgeTarget_Target_Reader.__new__(Graph_EdgeTarget_Target_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1304, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1305
 *     cdef init(C_Graph_EdgeTarget_Target_Reader source, object root):
 *         cdef Graph_EdgeTarget_Target_Reader wrapper = Graph_EdgeTarget_Target_Reader.__new__(Graph_EdgeTarget_Target_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1306
 *         cdef Graph_EdgeTarget_Target_Reader wrapper = Graph_EdgeTarget_Target_Reader.__new__(Graph_EdgeTarget_Target_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1307
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1303
 * 
 *     @staticmethod
 *     cdef init(C_Graph_EdgeTarget_Target_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_EdgeTarget_Target_Reader wrapper = Graph_EdgeTarget_Target_Reader.__new__(Graph_EdgeTarget_Target_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Target_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1309
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1311
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1309
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Target_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1312
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1313
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1312
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Target_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1315
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_9dep_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_9dep_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_9dep_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_9dep_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1321
 *         However, a value of zero always points to the 'current' graph.
 *         """
 *         return self.source.getDepIndex()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDepIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1321, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1315
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dep_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Target_Reader.dep_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1323
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_10node_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_10node_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_10node_index___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_10node_index___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1328
 *         The index into `Graph.nodes` where this node can be found.
 *         """
 *         return self.source.getNodeIndex()             # <<<<<<<<<<<<<<
 * 
 * cdef class AbstractTactic_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getNodeIndex();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1328, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1323
 *         return self.source.getDepIndex()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def node_index(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Target_Reader.node_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_4__reduce_cython__, "Graph_EdgeTarget_Target_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Target_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_6__setstate_cython__, "Graph_EdgeTarget_Target_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Target_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Target_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1332
 * cdef class AbstractTactic_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asAbstractTactic()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1332, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1332, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1332, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1332, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1333
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asAbstractTactic()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<AbstractTactic>();

  /* "pytact/graph_api_capnp_cython.pyx":1334
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asAbstractTactic()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1332
 * cdef class AbstractTactic_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asAbstractTactic()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1337
 * 
 *     @staticmethod
 *     cdef init(C_AbstractTactic_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef AbstractTactic_Reader wrapper = AbstractTactic_Reader.__new__(AbstractTactic_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_init(AbstractTactic::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1338
 *     @staticmethod
 *     cdef init(C_AbstractTactic_Reader source, object root):
 *         cdef AbstractTactic_Reader wrapper = AbstractTactic_Reader.__new__(AbstractTactic_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1338, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1339
 *     cdef init(C_AbstractTactic_Reader source, object root):
 *         cdef AbstractTactic_Reader wrapper = AbstractTactic_Reader.__new__(AbstractTactic_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1340
 *         cdef AbstractTactic_Reader wrapper = AbstractTactic_Reader.__new__(AbstractTactic_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1341
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1337
 * 
 *     @staticmethod
 *     cdef init(C_AbstractTactic_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef AbstractTactic_Reader wrapper = AbstractTactic_Reader.__new__(AbstractTactic_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1343
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1345
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1343
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1346
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1347
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1346
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1349
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ident(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5ident_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5ident_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5ident___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5ident___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1354
 *         An abstract tactic is referenced to using a identifier (hash).
 *         """
 *         return self.source.getIdent()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getIdent();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1354, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1349
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ident(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader.ident.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1356
 *         return self.source.getIdent()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def parameters(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_10parameters_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_10parameters_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_10parameters___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_10parameters___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint8_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1361
 *         Every tactic has a constant number of parameters that need to be filled in.
 *         """
 *         return self.source.getParameters()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.AbstractTactic.schema = _Schema()._init(getAbstractTactic_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getParameters();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1361, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1356
 *         return self.source.getIdent()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def parameters(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader.parameters.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_4__reduce_cython__, "AbstractTactic_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_6__setstate_cython__, "AbstractTactic_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1369
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDataVersion()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1369, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1369, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1369, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.DataVersion_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1369, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1370
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDataVersion()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<DataVersion>();

  /* "pytact/graph_api_capnp_cython.pyx":1371
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDataVersion()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1369
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDataVersion()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1374
 * 
 *     @staticmethod
 *     cdef init(C_DataVersion_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef DataVersion_Reader wrapper = DataVersion_Reader.__new__(DataVersion_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_init(DataVersion::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1375
 *     @staticmethod
 *     cdef init(C_DataVersion_Reader source, object root):
 *         cdef DataVersion_Reader wrapper = DataVersion_Reader.__new__(DataVersion_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_DataVersion_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1375, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1376
 *     cdef init(C_DataVersion_Reader source, object root):
 *         cdef DataVersion_Reader wrapper = DataVersion_Reader.__new__(DataVersion_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1377
 *         cdef DataVersion_Reader wrapper = DataVersion_Reader.__new__(DataVersion_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1378
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1374
 * 
 *     @staticmethod
 *     cdef init(C_DataVersion_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef DataVersion_Reader wrapper = DataVersion_Reader.__new__(DataVersion_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.DataVersion_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1380
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1382
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1380
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.DataVersion_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1383
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1384
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1383
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.DataVersion_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1386
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def major(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5major_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5major_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5major___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5major___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1395
 *         In the future, a minor version id might be introduced.
 *         """
 *         return self.source.getMajor()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.DataVersion.schema = _Schema()._init(getDataVersion_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getMajor();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1395, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1386
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def major(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.DataVersion_Reader.major.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_4__reduce_cython__, "DataVersion_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.DataVersion_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_6__setstate_cython__, "DataVersion_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.DataVersion_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.DataVersion_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1405
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDataset()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1405, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1405, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1405, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1405, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1406
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDataset()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Dataset>();

  /* "pytact/graph_api_capnp_cython.pyx":1407
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDataset()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1405
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDataset()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1410
 * 
 *     @staticmethod
 *     cdef init(C_Dataset_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Dataset_Reader wrapper = Dataset_Reader.__new__(Dataset_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_14Dataset_Reader_init(Dataset::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1411
 *     @staticmethod
 *     cdef init(C_Dataset_Reader source, object root):
 *         cdef Dataset_Reader wrapper = Dataset_Reader.__new__(Dataset_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Dataset_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1411, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1412
 *     cdef init(C_Dataset_Reader source, object root):
 *         cdef Dataset_Reader wrapper = Dataset_Reader.__new__(Dataset_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1413
 *         cdef Dataset_Reader wrapper = Dataset_Reader.__new__(Dataset_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1414
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1410
 * 
 *     @staticmethod
 *     cdef init(C_Dataset_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Dataset_Reader wrapper = Dataset_Reader.__new__(Dataset_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1416
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1418
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1416
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1419
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1420
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1419
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1422
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dependencies(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12dependencies_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12dependencies_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12dependencies___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12dependencies___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1430
 *         the dependency relation between files induced by this field (except for the self-reference of the file).
 *         """
 *         return String_List.init(self.source.getDependencies(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_dependencies(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDependencies();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1430, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_11String_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1422
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dependencies(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.dependencies.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1431
 *         """
 *         return String_List.init(self.source.getDependencies(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_dependencies(self):
 *         return self.source.hasDependencies()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_dependencies_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_dependencies_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_dependencies___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_dependencies___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1433
 *     @property
 *     def has_dependencies(self):
 *         return self.source.hasDependencies()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasDependencies();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1433, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1431
 *         """
 *         return String_List.init(self.source.getDependencies(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_dependencies(self):
 *         return self.source.hasDependencies()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.has_dependencies.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1435
 *         return self.source.hasDependencies()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def graph(self):
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5graph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5graph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5graph___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Graph::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1437
 *     @property
 *     def graph(self):
 *         return Graph_Reader.init(self.source.getGraph(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_graph(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getGraph();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1437, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_12Graph_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1435
 *         return self.source.hasDependencies()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def graph(self):
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.graph.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1438
 *     def graph(self):
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_graph(self):
 *         return self.source.hasGraph()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_9has_graph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_9has_graph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_9has_graph___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_9has_graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1440
 *     @property
 *     def has_graph(self):
 *         return self.source.hasGraph()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasGraph();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1440, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1438
 *     def graph(self):
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_graph(self):
 *         return self.source.hasGraph()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.has_graph.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1442
 *         return self.source.hasGraph()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def representative(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_14representative_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_14representative_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_14representative___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_14representative___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1449
 *         If the compilation unit does not contain any 'super'-global definitions this is set to `len(graph.nodes)`
 *         """
 *         return self.source.getRepresentative()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getRepresentative();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1449, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1442
 *         return self.source.hasGraph()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def representative(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.representative.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1451
 *         return self.source.getRepresentative()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def definitions(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11definitions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11definitions_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11definitions___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11definitions___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1459
 *         and the 'super-global' context is only the main spine of this forest.
 *         """
 *         return Uint32_List.init(self.source.getDefinitions(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_definitions(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDefinitions();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1459, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_11Uint32_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1451
 *         return self.source.getRepresentative()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def definitions(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.definitions.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1460
 *         """
 *         return Uint32_List.init(self.source.getDefinitions(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_definitions(self):
 *         return self.source.hasDefinitions()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_definitions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_definitions_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_definitions___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_definitions___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1462
 *     @property
 *     def has_definitions(self):
 *         return self.source.hasDefinitions()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasDefinitions();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1462, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1460
 *         """
 *         return Uint32_List.init(self.source.getDefinitions(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_definitions(self):
 *         return self.source.hasDefinitions()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.has_definitions.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1464
 *         return self.source.hasDefinitions()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def module_name(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11module_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11module_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11module_name___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11module_name___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1469
 *         The name of the module defined by this file.
 *         """
 *         temp = self.source.getModuleName()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getModuleName();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1469, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":1470
 *         """
 *         temp = self.source.getModuleName()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_module_name(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1464
 *         return self.source.hasDefinitions()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def module_name(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.module_name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1471
 *         temp = self.source.getModuleName()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_module_name(self):
 *         return self.source.hasModuleName()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_module_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_module_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_module_name___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_module_name___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1473
 *     @property
 *     def has_module_name(self):
 *         return self.source.hasModuleName()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasModuleName();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1473, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1471
 *         temp = self.source.getModuleName()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_module_name(self):
 *         return self.source.hasModuleName()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.has_module_name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1475
 *         return self.source.hasModuleName()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def data_version(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12data_version_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12data_version_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12data_version___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12data_version___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  DataVersion::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1480
 *         The version of the data stored in this dataset.
 *         """
 *         return DataVersion_Reader.init(self.source.getDataVersion(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_data_version(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDataVersion();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1480, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1475
 *         return self.source.hasModuleName()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def data_version(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.data_version.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1481
 *         """
 *         return DataVersion_Reader.init(self.source.getDataVersion(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_data_version(self):
 *         return self.source.hasDataVersion()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_data_version_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_data_version_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_data_version___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_data_version___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1483
 *     @property
 *     def has_data_version(self):
 *         return self.source.hasDataVersion()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.Dataset.schema = _Schema()._init(getDataset_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasDataVersion();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1483, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1481
 *         """
 *         return DataVersion_Reader.init(self.source.getDataVersion(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_data_version(self):
 *         return self.source.hasDataVersion()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.has_data_version.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_14Dataset_Reader_4__reduce_cython__, "Dataset_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_14Dataset_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_14Dataset_Reader_6__setstate_cython__, "Dataset_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_14Dataset_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_14Dataset_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Dataset_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1491
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asException()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1491, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1491, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1491, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1491, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1492
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asException()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<Exception>();

  /* "pytact/graph_api_capnp_cython.pyx":1493
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asException()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1491
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asException()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1496
 * 
 *     @staticmethod
 *     cdef init(C_Exception_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Exception_Reader wrapper = Exception_Reader.__new__(Exception_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_16Exception_Reader_init(Exception::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1497
 *     @staticmethod
 *     cdef init(C_Exception_Reader source, object root):
 *         cdef Exception_Reader wrapper = Exception_Reader.__new__(Exception_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Exception_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1497, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1498
 *     cdef init(C_Exception_Reader source, object root):
 *         cdef Exception_Reader wrapper = Exception_Reader.__new__(Exception_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1499
 *         cdef Exception_Reader wrapper = Exception_Reader.__new__(Exception_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1500
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1496
 * 
 *     @staticmethod
 *     cdef init(C_Exception_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef Exception_Reader wrapper = Exception_Reader.__new__(Exception_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1502
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1504
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1502
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1505
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1506
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 *     @property
 *     def which(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1505
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1507
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return Exception_Which(self.source.which())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_5which_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_5which_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_5which___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __PYX_ENUM_CLASS_DECL Exception::Which __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1509
 *     @property
 *     def which(self):
 *         return Exception_Which(self.source.which())             # <<<<<<<<<<<<<<
 *     @property
 *     def which_raw(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Exception_Which); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    __pyx_t_3 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1509, __pyx_L1_error)
  }
  __pyx_t_4 = [](const Exception::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1507
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return Exception_Which(self.source.which())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.which.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1510
 *     def which(self):
 *         return Exception_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_9which_raw_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_9which_raw_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_9which_raw___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __PYX_ENUM_CLASS_DECL Exception::Which __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1512
 *     @property
 *     def which_raw(self):
 *         return self.source.which()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1512, __pyx_L1_error)
  }
  __pyx_t_2 = [](const Exception::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1510
 *     def which(self):
 *         return Exception_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.which_raw.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1514
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def no_such_tactic(self):
 *         return self.source.getNoSuchTactic()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_14no_such_tactic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_14no_such_tactic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_14no_such_tactic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_14no_such_tactic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1516
 *     @property
 *     def no_such_tactic(self):
 *         return self.source.getNoSuchTactic()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_no_such_tactic(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getNoSuchTactic();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1516, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1514
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def no_such_tactic(self):
 *         return self.source.getNoSuchTactic()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.no_such_tactic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1517
 *     def no_such_tactic(self):
 *         return self.source.getNoSuchTactic()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_no_such_tactic(self):
 *         return self.source.isNoSuchTactic()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_17is_no_such_tactic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_17is_no_such_tactic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_17is_no_such_tactic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_17is_no_such_tactic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1519
 *     @property
 *     def is_no_such_tactic(self):
 *         return self.source.isNoSuchTactic()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isNoSuchTactic();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1519, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1517
 *     def no_such_tactic(self):
 *         return self.source.getNoSuchTactic()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_no_such_tactic(self):
 *         return self.source.isNoSuchTactic()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.is_no_such_tactic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1521
 *         return self.source.isNoSuchTactic()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def mismatched_arguments(self):
 *         return self.source.getMismatchedArguments()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_20mismatched_arguments_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_20mismatched_arguments_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_20mismatched_arguments___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_20mismatched_arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1523
 *     @property
 *     def mismatched_arguments(self):
 *         return self.source.getMismatchedArguments()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_mismatched_arguments(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getMismatchedArguments();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1523, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1521
 *         return self.source.isNoSuchTactic()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def mismatched_arguments(self):
 *         return self.source.getMismatchedArguments()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.mismatched_arguments.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1524
 *     def mismatched_arguments(self):
 *         return self.source.getMismatchedArguments()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_mismatched_arguments(self):
 *         return self.source.isMismatchedArguments()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_23is_mismatched_arguments_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_23is_mismatched_arguments_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_23is_mismatched_arguments___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_23is_mismatched_arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1526
 *     @property
 *     def is_mismatched_arguments(self):
 *         return self.source.isMismatchedArguments()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isMismatchedArguments();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1526, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1524
 *     def mismatched_arguments(self):
 *         return self.source.getMismatchedArguments()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_mismatched_arguments(self):
 *         return self.source.isMismatchedArguments()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.is_mismatched_arguments.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1528
 *         return self.source.isMismatchedArguments()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def parse_error(self):
 *         return self.source.getParseError()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_11parse_error_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_11parse_error_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_11parse_error___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_11parse_error___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1530
 *     @property
 *     def parse_error(self):
 *         return self.source.getParseError()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_parse_error(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getParseError();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1530, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1528
 *         return self.source.isMismatchedArguments()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def parse_error(self):
 *         return self.source.getParseError()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.parse_error.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1531
 *     def parse_error(self):
 *         return self.source.getParseError()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_parse_error(self):
 *         return self.source.isParseError()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_14is_parse_error_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_14is_parse_error_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_14is_parse_error___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_14is_parse_error___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1533
 *     @property
 *     def is_parse_error(self):
 *         return self.source.isParseError()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isParseError();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1533, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1531
 *     def parse_error(self):
 *         return self.source.getParseError()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_parse_error(self):
 *         return self.source.isParseError()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.is_parse_error.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1535
 *         return self.source.isParseError()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def illegal_argument(self):
 *         return self.source.getIllegalArgument()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_16illegal_argument_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_16illegal_argument_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_16illegal_argument___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_16illegal_argument___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1537
 *     @property
 *     def illegal_argument(self):
 *         return self.source.getIllegalArgument()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_illegal_argument(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getIllegalArgument();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1537, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1535
 *         return self.source.isParseError()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def illegal_argument(self):
 *         return self.source.getIllegalArgument()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.illegal_argument.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1538
 *     def illegal_argument(self):
 *         return self.source.getIllegalArgument()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_illegal_argument(self):
 *         return self.source.isIllegalArgument()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_19is_illegal_argument_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_19is_illegal_argument_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_19is_illegal_argument___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_19is_illegal_argument___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1540
 *     @property
 *     def is_illegal_argument(self):
 *         return self.source.isIllegalArgument()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.Exception.schema = _Schema()._init(getException_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isIllegalArgument();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1540, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1538
 *     def illegal_argument(self):
 *         return self.source.getIllegalArgument()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_illegal_argument(self):
 *         return self.source.isIllegalArgument()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.is_illegal_argument.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_16Exception_Reader_4__reduce_cython__, "Exception_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_16Exception_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16Exception_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_16Exception_Reader_6__setstate_cython__, "Exception_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_16Exception_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16Exception_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Exception_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Exception_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1548
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asExecutionResult()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1548, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1548, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1548, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1548, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1549
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asExecutionResult()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<ExecutionResult>();

  /* "pytact/graph_api_capnp_cython.pyx":1550
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asExecutionResult()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1548
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asExecutionResult()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1553
 * 
 *     @staticmethod
 *     cdef init(C_ExecutionResult_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ExecutionResult_Reader wrapper = ExecutionResult_Reader.__new__(ExecutionResult_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_init(ExecutionResult::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1554
 *     @staticmethod
 *     cdef init(C_ExecutionResult_Reader source, object root):
 *         cdef ExecutionResult_Reader wrapper = ExecutionResult_Reader.__new__(ExecutionResult_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1554, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1555
 *     cdef init(C_ExecutionResult_Reader source, object root):
 *         cdef ExecutionResult_Reader wrapper = ExecutionResult_Reader.__new__(ExecutionResult_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1556
 *         cdef ExecutionResult_Reader wrapper = ExecutionResult_Reader.__new__(ExecutionResult_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1557
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1553
 * 
 *     @staticmethod
 *     cdef init(C_ExecutionResult_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ExecutionResult_Reader wrapper = ExecutionResult_Reader.__new__(ExecutionResult_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1559
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1561
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1559
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1562
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1563
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 *     @property
 *     def which(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1562
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1564
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return ExecutionResult_Which(self.source.which())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5which_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5which_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5which___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __PYX_ENUM_CLASS_DECL ExecutionResult::Which __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1566
 *     @property
 *     def which(self):
 *         return ExecutionResult_Which(self.source.which())             # <<<<<<<<<<<<<<
 *     @property
 *     def which_raw(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ExecutionResult_Which); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    __pyx_t_3 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1566, __pyx_L1_error)
  }
  __pyx_t_4 = [](const ExecutionResult::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1564
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return ExecutionResult_Which(self.source.which())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.which.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1567
 *     def which(self):
 *         return ExecutionResult_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9which_raw_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9which_raw_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9which_raw___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __PYX_ENUM_CLASS_DECL ExecutionResult::Which __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1569
 *     @property
 *     def which_raw(self):
 *         return self.source.which()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1569, __pyx_L1_error)
  }
  __pyx_t_2 = [](const ExecutionResult::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1567
 *     def which(self):
 *         return ExecutionResult_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.which_raw.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1571
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def failure(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7failure_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7failure_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7failure___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7failure___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1576
 *         The tactic execution failed. This is not an error condition, but rather the natural failure of the tactic.
 *         """
 *         return self.source.getFailure()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_failure(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getFailure();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1576, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1571
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def failure(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.failure.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1577
 *         """
 *         return self.source.getFailure()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_failure(self):
 *         return self.source.isFailure()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_10is_failure_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_10is_failure_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_10is_failure___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_10is_failure___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1579
 *     @property
 *     def is_failure(self):
 *         return self.source.isFailure()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isFailure();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1579, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1577
 *         """
 *         return self.source.getFailure()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_failure(self):
 *         return self.source.isFailure()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.is_failure.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1581
 *         return self.source.isFailure()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def complete(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_8complete_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_8complete_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_8complete___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_8complete___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1586
 *         The proof has been completed.
 *         """
 *         return self.source.getComplete()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_complete(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getComplete();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1586, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1581
 *         return self.source.isFailure()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def complete(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.complete.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1587
 *         """
 *         return self.source.getComplete()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_complete(self):
 *         return self.source.isComplete()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_11is_complete_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_11is_complete_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_11is_complete___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_11is_complete___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1589
 *     @property
 *     def is_complete(self):
 *         return self.source.isComplete()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isComplete();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1589, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1587
 *         """
 *         return self.source.getComplete()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_complete(self):
 *         return self.source.isComplete()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.is_complete.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1591
 *         return self.source.isComplete()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def new_state(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9new_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9new_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9new_state___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9new_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  ExecutionResult::NewState::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1596
 *         The tactic ran successfully and produced a new proof state.
 *         """
 *         return ExecutionResult_NewState_Reader.init(self.source.getNewState(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_new_state(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getNewState();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1596, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1591
 *         return self.source.isComplete()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def new_state(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.new_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1597
 *         """
 *         return ExecutionResult_NewState_Reader.init(self.source.getNewState(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_new_state(self):
 *         return self.source.isNewState()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_12is_new_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_12is_new_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_12is_new_state___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_12is_new_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1599
 *     @property
 *     def is_new_state(self):
 *         return self.source.isNewState()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isNewState();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1599, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1597
 *         """
 *         return ExecutionResult_NewState_Reader.init(self.source.getNewState(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_new_state(self):
 *         return self.source.isNewState()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.is_new_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1601
 *         return self.source.isNewState()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def protocol_error(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_14protocol_error_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_14protocol_error_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_14protocol_error___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_14protocol_error___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Exception::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1606
 *         Indicates a programmer error.
 *         """
 *         return Exception_Reader.init(self.source.getProtocolError(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_protocol_error(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getProtocolError();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1606, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_16Exception_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1601
 *         return self.source.isNewState()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def protocol_error(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.protocol_error.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1607
 *         """
 *         return Exception_Reader.init(self.source.getProtocolError(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_protocol_error(self):
 *         return self.source.isProtocolError()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_17is_protocol_error_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_17is_protocol_error_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_17is_protocol_error___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_17is_protocol_error___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1609
 *     @property
 *     def is_protocol_error(self):
 *         return self.source.isProtocolError()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_protocol_error(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isProtocolError();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1609, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1607
 *         """
 *         return Exception_Reader.init(self.source.getProtocolError(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_protocol_error(self):
 *         return self.source.isProtocolError()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.is_protocol_error.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1610
 *     def is_protocol_error(self):
 *         return self.source.isProtocolError()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_protocol_error(self):
 *         return self.source.hasProtocolError()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_18has_protocol_error_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_18has_protocol_error_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_18has_protocol_error___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_18has_protocol_error___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1612
 *     @property
 *     def has_protocol_error(self):
 *         return self.source.hasProtocolError()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.ExecutionResult.schema = _Schema()._init(getExecutionResult_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasProtocolError();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1612, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1610
 *     def is_protocol_error(self):
 *         return self.source.isProtocolError()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_protocol_error(self):
 *         return self.source.hasProtocolError()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.has_protocol_error.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_4__reduce_cython__, "ExecutionResult_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_6__setstate_cython__, "ExecutionResult_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1620
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asExecutionResult_NewState()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1620, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1620, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1620, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1620, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1621
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asExecutionResult_NewState()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<ExecutionResult::NewState>();

  /* "pytact/graph_api_capnp_cython.pyx":1622
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asExecutionResult_NewState()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1620
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asExecutionResult_NewState()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1625
 * 
 *     @staticmethod
 *     cdef init(C_ExecutionResult_NewState_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ExecutionResult_NewState_Reader wrapper = ExecutionResult_NewState_Reader.__new__(ExecutionResult_NewState_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_init(ExecutionResult::NewState::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1626
 *     @staticmethod
 *     cdef init(C_ExecutionResult_NewState_Reader source, object root):
 *         cdef ExecutionResult_NewState_Reader wrapper = ExecutionResult_NewState_Reader.__new__(ExecutionResult_NewState_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1626, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1627
 *     cdef init(C_ExecutionResult_NewState_Reader source, object root):
 *         cdef ExecutionResult_NewState_Reader wrapper = ExecutionResult_NewState_Reader.__new__(ExecutionResult_NewState_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1628
 *         cdef ExecutionResult_NewState_Reader wrapper = ExecutionResult_NewState_Reader.__new__(ExecutionResult_NewState_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1629
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1625
 * 
 *     @staticmethod
 *     cdef init(C_ExecutionResult_NewState_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ExecutionResult_NewState_Reader wrapper = ExecutionResult_NewState_Reader.__new__(ExecutionResult_NewState_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1631
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1633
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1631
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1634
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1635
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1634
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1637
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def graph(self):
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5graph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5graph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5graph___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Graph::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1639
 *     @property
 *     def graph(self):
 *         return Graph_Reader.init(self.source.getGraph(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_graph(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getGraph();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1639, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_12Graph_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1637
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def graph(self):
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.graph.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1640
 *     def graph(self):
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_graph(self):
 *         return self.source.hasGraph()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_graph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_graph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_graph___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1642
 *     @property
 *     def has_graph(self):
 *         return self.source.hasGraph()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasGraph();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1642, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1640
 *     def graph(self):
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_graph(self):
 *         return self.source.hasGraph()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.has_graph.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1644
 *         return self.source.hasGraph()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def state(self):
 *         return ProofState_Reader.init(self.source.getState(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5state___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  ProofState::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1646
 *     @property
 *     def state(self):
 *         return ProofState_Reader.init(self.source.getState(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_state(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getState();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1646, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_17ProofState_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1644
 *         return self.source.hasGraph()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def state(self):
 *         return ProofState_Reader.init(self.source.getState(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1647
 *     def state(self):
 *         return ProofState_Reader.init(self.source.getState(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_state(self):
 *         return self.source.hasState()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_state___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1649
 *     @property
 *     def has_state(self):
 *         return self.source.hasState()             # <<<<<<<<<<<<<<
 * 
 * cdef class PredictionProtocol_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasState();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1649, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1647
 *     def state(self):
 *         return ProofState_Reader.init(self.source.getState(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_state(self):
 *         return self.source.hasState()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.has_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_4__reduce_cython__, "ExecutionResult_NewState_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_6__setstate_cython__, "ExecutionResult_NewState_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1658
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1658, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1658, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1658, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1658, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1659
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<PredictionProtocol>();

  /* "pytact/graph_api_capnp_cython.pyx":1660
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1658
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1663
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Reader wrapper = PredictionProtocol_Reader.__new__(PredictionProtocol_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_init(PredictionProtocol::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1664
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Reader source, object root):
 *         cdef PredictionProtocol_Reader wrapper = PredictionProtocol_Reader.__new__(PredictionProtocol_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1664, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1665
 *     cdef init(C_PredictionProtocol_Reader source, object root):
 *         cdef PredictionProtocol_Reader wrapper = PredictionProtocol_Reader.__new__(PredictionProtocol_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1666
 *         cdef PredictionProtocol_Reader wrapper = PredictionProtocol_Reader.__new__(PredictionProtocol_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1667
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1663
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Reader wrapper = PredictionProtocol_Reader.__new__(PredictionProtocol_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1669
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1671
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1669
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1672
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * pytact.graph_api_capnp.PredictionProtocol.schema = _Schema()._init(getPredictionProtocol_Schema()).as_struct()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1673
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.PredictionProtocol.schema = _Schema()._init(getPredictionProtocol_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1672
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * pytact.graph_api_capnp.PredictionProtocol.schema = _Schema()._init(getPredictionProtocol_Schema()).as_struct()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_4__reduce_cython__, "PredictionProtocol_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_6__setstate_cython__, "PredictionProtocol_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1678
 * cdef class PredictionProtocol_Request_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1678, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1678, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1678, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1678, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1679
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<PredictionProtocol::Request>();

  /* "pytact/graph_api_capnp_cython.pyx":1680
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1678
 * cdef class PredictionProtocol_Request_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1683
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Request_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Request_Reader wrapper = PredictionProtocol_Request_Reader.__new__(PredictionProtocol_Request_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_init(PredictionProtocol::Request::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1684
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Request_Reader source, object root):
 *         cdef PredictionProtocol_Request_Reader wrapper = PredictionProtocol_Request_Reader.__new__(PredictionProtocol_Request_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1684, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1685
 *     cdef init(C_PredictionProtocol_Request_Reader source, object root):
 *         cdef PredictionProtocol_Request_Reader wrapper = PredictionProtocol_Request_Reader.__new__(PredictionProtocol_Request_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1686
 *         cdef PredictionProtocol_Request_Reader wrapper = PredictionProtocol_Request_Reader.__new__(PredictionProtocol_Request_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1687
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1683
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Request_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Request_Reader wrapper = PredictionProtocol_Request_Reader.__new__(PredictionProtocol_Request_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1689
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1691
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1689
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1692
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1693
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 *     @property
 *     def which(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1692
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1694
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return PredictionProtocol_Request_Which(self.source.which())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5which_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5which_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5which___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __PYX_ENUM_CLASS_DECL PredictionProtocol::Request::Which __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1696
 *     @property
 *     def which(self):
 *         return PredictionProtocol_Request_Which(self.source.which())             # <<<<<<<<<<<<<<
 *     @property
 *     def which_raw(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PredictionProtocol_Request_Which); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    __pyx_t_3 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1696, __pyx_L1_error)
  }
  __pyx_t_4 = [](const PredictionProtocol::Request::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1694
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return PredictionProtocol_Request_Which(self.source.which())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.which.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1697
 *     def which(self):
 *         return PredictionProtocol_Request_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_9which_raw_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_9which_raw_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_9which_raw___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __PYX_ENUM_CLASS_DECL PredictionProtocol::Request::Which __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1699
 *     @property
 *     def which_raw(self):
 *         return self.source.which()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1699, __pyx_L1_error)
  }
  __pyx_t_2 = [](const PredictionProtocol::Request::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1697
 *     def which(self):
 *         return PredictionProtocol_Request_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.which_raw.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1701
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def initialize(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10initialize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10initialize_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10initialize___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10initialize___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PredictionProtocol::Request::Initialize::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1707
 *         that are currently available, the definitions that are available.
 *         """
 *         return PredictionProtocol_Request_Initialize_Reader.init(self.source.getInitialize(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_initialize(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getInitialize();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1707, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1701
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def initialize(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.initialize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1708
 *         """
 *         return PredictionProtocol_Request_Initialize_Reader.init(self.source.getInitialize(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_initialize(self):
 *         return self.source.isInitialize()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_13is_initialize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_13is_initialize_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_13is_initialize___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_13is_initialize___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1710
 *     @property
 *     def is_initialize(self):
 *         return self.source.isInitialize()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isInitialize();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1710, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1708
 *         """
 *         return PredictionProtocol_Request_Initialize_Reader.init(self.source.getInitialize(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_initialize(self):
 *         return self.source.isInitialize()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.is_initialize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1712
 *         return self.source.isInitialize()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def predict(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7predict_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7predict_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7predict___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7predict___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PredictionProtocol::Request::Predict::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1717
 *         Request a list of tactic predictions given the graph of a proof state.
 *         """
 *         return PredictionProtocol_Request_Predict_Reader.init(self.source.getPredict(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_predict(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getPredict();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1717, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1712
 *         return self.source.isInitialize()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def predict(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.predict.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1718
 *         """
 *         return PredictionProtocol_Request_Predict_Reader.init(self.source.getPredict(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_predict(self):
 *         return self.source.isPredict()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10is_predict_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10is_predict_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10is_predict___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10is_predict___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1720
 *     @property
 *     def is_predict(self):
 *         return self.source.isPredict()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isPredict();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1720, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1718
 *         """
 *         return PredictionProtocol_Request_Predict_Reader.init(self.source.getPredict(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_predict(self):
 *         return self.source.isPredict()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.is_predict.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1722
 *         return self.source.isPredict()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def synchronize(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_11synchronize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_11synchronize_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_11synchronize___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_11synchronize___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1728
 *         The contract is that the given integer needs to be echo'd back verbatim.
 *         """
 *         return self.source.getSynchronize()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_synchronize(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getSynchronize();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1728, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1722
 *         return self.source.isPredict()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def synchronize(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.synchronize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1729
 *         """
 *         return self.source.getSynchronize()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_synchronize(self):
 *         return self.source.isSynchronize()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_14is_synchronize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_14is_synchronize_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_14is_synchronize___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_14is_synchronize___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1731
 *     @property
 *     def is_synchronize(self):
 *         return self.source.isSynchronize()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isSynchronize();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1731, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1729
 *         """
 *         return self.source.getSynchronize()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_synchronize(self):
 *         return self.source.isSynchronize()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.is_synchronize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1733
 *         return self.source.isSynchronize()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def check_alignment(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_15check_alignment_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_15check_alignment_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_15check_alignment___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_15check_alignment___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1739
 *         and report back any tactics and definitions that were not found
 *         """
 *         return self.source.getCheckAlignment()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_check_alignment(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getCheckAlignment();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1739, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1733
 *         return self.source.isSynchronize()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def check_alignment(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.check_alignment.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1740
 *         """
 *         return self.source.getCheckAlignment()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_check_alignment(self):
 *         return self.source.isCheckAlignment()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_18is_check_alignment_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_18is_check_alignment_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_18is_check_alignment___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_18is_check_alignment___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1742
 *     @property
 *     def is_check_alignment(self):
 *         return self.source.isCheckAlignment()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.PredictionProtocol.Request.schema = _Schema()._init(getPredictionProtocol_Request_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isCheckAlignment();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1742, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1740
 *         """
 *         return self.source.getCheckAlignment()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_check_alignment(self):
 *         return self.source.isCheckAlignment()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.is_check_alignment.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_4__reduce_cython__, "PredictionProtocol_Request_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_6__setstate_cython__, "PredictionProtocol_Request_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1751
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request_Initialize()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1751, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1751, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1751, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1751, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1752
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request_Initialize()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<PredictionProtocol::Request::Initialize>();

  /* "pytact/graph_api_capnp_cython.pyx":1753
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request_Initialize()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1751
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request_Initialize()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1756
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Request_Initialize_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Request_Initialize_Reader wrapper = PredictionProtocol_Request_Initialize_Reader.__new__(PredictionProtocol_Request_Initialize_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_init(PredictionProtocol::Request::Initialize::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1757
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Request_Initialize_Reader source, object root):
 *         cdef PredictionProtocol_Request_Initialize_Reader wrapper = PredictionProtocol_Request_Initialize_Reader.__new__(PredictionProtocol_Request_Initialize_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1757, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1758
 *     cdef init(C_PredictionProtocol_Request_Initialize_Reader source, object root):
 *         cdef PredictionProtocol_Request_Initialize_Reader wrapper = PredictionProtocol_Request_Initialize_Reader.__new__(PredictionProtocol_Request_Initialize_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1759
 *         cdef PredictionProtocol_Request_Initialize_Reader wrapper = PredictionProtocol_Request_Initialize_Reader.__new__(PredictionProtocol_Request_Initialize_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1760
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1756
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Request_Initialize_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Request_Initialize_Reader wrapper = PredictionProtocol_Request_Initialize_Reader.__new__(PredictionProtocol_Request_Initialize_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1762
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1764
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1762
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1765
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1766
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1765
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1768
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def data_version(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_12data_version_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_12data_version_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_12data_version___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_12data_version___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  DataVersion::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1773
 *         The version number this message and subsequent messages is compatible with.
 *         """
 *         return DataVersion_Reader.init(self.source.getDataVersion(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_data_version(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getDataVersion();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1773, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1768
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def data_version(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.data_version.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1774
 *         """
 *         return DataVersion_Reader.init(self.source.getDataVersion(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_data_version(self):
 *         return self.source.hasDataVersion()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_16has_data_version_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_16has_data_version_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_16has_data_version___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_16has_data_version___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1776
 *     @property
 *     def has_data_version(self):
 *         return self.source.hasDataVersion()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasDataVersion();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1776, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1774
 *         """
 *         return DataVersion_Reader.init(self.source.getDataVersion(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_data_version(self):
 *         return self.source.hasDataVersion()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.has_data_version.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1778
 *         return self.source.hasDataVersion()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def stack_size(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_10stack_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_10stack_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_10stack_size___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_10stack_size___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1791
 *         `j - i`. (This scheme keeps the guarantee that dependency index `0` always points to the current graph.)
 *         """
 *         return self.source.getStackSize()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getStackSize();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1791, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_int32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1778
 *         return self.source.hasDataVersion()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def stack_size(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.stack_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1793
 *         return self.source.getStackSize()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactics(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7tactics_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7tactics_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7tactics___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7tactics___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<AbstractTactic, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1798
 *         A list of tactics that Coq currently knows about.
 *         """
 *         return AbstractTactic_Reader_List.init(self.source.getTactics(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_tactics(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getTactics();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1798, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1793
 *         return self.source.getStackSize()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactics(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.tactics.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1799
 *         """
 *         return AbstractTactic_Reader_List.init(self.source.getTactics(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactics(self):
 *         return self.source.hasTactics()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_11has_tactics_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_11has_tactics_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_11has_tactics___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_11has_tactics___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1801
 *     @property
 *     def has_tactics(self):
 *         return self.source.hasTactics()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTactics();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1801, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1799
 *         """
 *         return AbstractTactic_Reader_List.init(self.source.getTactics(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactics(self):
 *         return self.source.hasTactics()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.has_tactics.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1803
 *         return self.source.hasTactics()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def graph(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5graph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5graph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5graph___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Graph::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1808
 *         The graph of this message. See `stackSize` on how to resolve dependency indexes in this graph.
 *         """
 *         return Graph_Reader.init(self.source.getGraph(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_graph(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getGraph();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1808, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_12Graph_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1803
 *         return self.source.hasTactics()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def graph(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.graph.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1809
 *         """
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_graph(self):
 *         return self.source.hasGraph()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_9has_graph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_9has_graph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_9has_graph___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_9has_graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1811
 *     @property
 *     def has_graph(self):
 *         return self.source.hasGraph()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasGraph();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1811, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1809
 *         """
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_graph(self):
 *         return self.source.hasGraph()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.has_graph.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1813
 *         return self.source.hasGraph()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def log_annotation(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14log_annotation_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14log_annotation_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14log_annotation___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14log_annotation___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1818
 *         An annotation containing file and line information on where Coq is currently processing.
 *         """
 *         temp = self.source.getLogAnnotation()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getLogAnnotation();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1818, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":1819
 *         """
 *         temp = self.source.getLogAnnotation()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_log_annotation(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1813
 *         return self.source.hasGraph()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def log_annotation(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.log_annotation.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1820
 *         temp = self.source.getLogAnnotation()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_log_annotation(self):
 *         return self.source.hasLogAnnotation()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_18has_log_annotation_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_18has_log_annotation_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_18has_log_annotation___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_18has_log_annotation___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1822
 *     @property
 *     def has_log_annotation(self):
 *         return self.source.hasLogAnnotation()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasLogAnnotation();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1822, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1820
 *         temp = self.source.getLogAnnotation()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_log_annotation(self):
 *         return self.source.hasLogAnnotation()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.has_log_annotation.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1824
 *         return self.source.hasLogAnnotation()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def representative(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14representative_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14representative_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14representative___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14representative___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1831
 *         this is equal to `len(graph.nodes)`.
 *         """
 *         return self.source.getRepresentative()             # <<<<<<<<<<<<<<
 * 
 * cdef class PredictionProtocol_Request_Predict_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getRepresentative();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1831, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1824
 *         return self.source.hasLogAnnotation()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def representative(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.representative.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_4__reduce_cython__, "PredictionProtocol_Request_Initialize_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_6__setstate_cython__, "PredictionProtocol_Request_Initialize_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1838
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request_Predict()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1838, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1838, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1838, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1838, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1839
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request_Predict()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<PredictionProtocol::Request::Predict>();

  /* "pytact/graph_api_capnp_cython.pyx":1840
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request_Predict()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1838
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request_Predict()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1843
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Request_Predict_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Request_Predict_Reader wrapper = PredictionProtocol_Request_Predict_Reader.__new__(PredictionProtocol_Request_Predict_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_init(PredictionProtocol::Request::Predict::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1844
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Request_Predict_Reader source, object root):
 *         cdef PredictionProtocol_Request_Predict_Reader wrapper = PredictionProtocol_Request_Predict_Reader.__new__(PredictionProtocol_Request_Predict_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1844, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1845
 *     cdef init(C_PredictionProtocol_Request_Predict_Reader source, object root):
 *         cdef PredictionProtocol_Request_Predict_Reader wrapper = PredictionProtocol_Request_Predict_Reader.__new__(PredictionProtocol_Request_Predict_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1846
 *         cdef PredictionProtocol_Request_Predict_Reader wrapper = PredictionProtocol_Request_Predict_Reader.__new__(PredictionProtocol_Request_Predict_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1847
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1843
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Request_Predict_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Request_Predict_Reader wrapper = PredictionProtocol_Request_Predict_Reader.__new__(PredictionProtocol_Request_Predict_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1849
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1851
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1849
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1852
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1853
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1852
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1855
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def graph(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5graph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5graph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5graph___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Graph::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1861
 *         `initialize.stackSize` on how dependency indices can be resolved to a graph in the stack.
 *         """
 *         return Graph_Reader.init(self.source.getGraph(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_graph(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getGraph();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1861, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_12Graph_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1855
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def graph(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.graph.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1862
 *         """
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_graph(self):
 *         return self.source.hasGraph()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_graph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_graph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_graph___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_graph___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1864
 *     @property
 *     def has_graph(self):
 *         return self.source.hasGraph()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasGraph();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1864, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1862
 *         """
 *         return Graph_Reader.init(self.source.getGraph(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_graph(self):
 *         return self.source.hasGraph()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.has_graph.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1866
 *         return self.source.hasGraph()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def state(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5state___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  ProofState::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1871
 *         The proof state for which a prediction is requested.
 *         """
 *         return ProofState_Reader.init(self.source.getState(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_state(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getState();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1871, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_17ProofState_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1866
 *         return self.source.hasGraph()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def state(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1872
 *         """
 *         return ProofState_Reader.init(self.source.getState(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_state(self):
 *         return self.source.hasState()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_state___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_state___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1874
 *     @property
 *     def has_state(self):
 *         return self.source.hasState()             # <<<<<<<<<<<<<<
 * 
 * cdef class PredictionProtocol_Prediction_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasState();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1874, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1872
 *         """
 *         return ProofState_Reader.init(self.source.getState(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_state(self):
 *         return self.source.hasState()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.has_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_4__reduce_cython__, "PredictionProtocol_Request_Predict_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_6__setstate_cython__, "PredictionProtocol_Request_Predict_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1878
 * cdef class PredictionProtocol_Prediction_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Prediction()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1878, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1878, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1878, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1878, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1879
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Prediction()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<PredictionProtocol::Prediction>();

  /* "pytact/graph_api_capnp_cython.pyx":1880
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Prediction()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1878
 * cdef class PredictionProtocol_Prediction_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Prediction()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1883
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Prediction_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Prediction_Reader wrapper = PredictionProtocol_Prediction_Reader.__new__(PredictionProtocol_Prediction_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_init(PredictionProtocol::Prediction::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1884
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Prediction_Reader source, object root):
 *         cdef PredictionProtocol_Prediction_Reader wrapper = PredictionProtocol_Prediction_Reader.__new__(PredictionProtocol_Prediction_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1884, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1885
 *     cdef init(C_PredictionProtocol_Prediction_Reader source, object root):
 *         cdef PredictionProtocol_Prediction_Reader wrapper = PredictionProtocol_Prediction_Reader.__new__(PredictionProtocol_Prediction_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1886
 *         cdef PredictionProtocol_Prediction_Reader wrapper = PredictionProtocol_Prediction_Reader.__new__(PredictionProtocol_Prediction_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1887
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1883
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Prediction_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Prediction_Reader wrapper = PredictionProtocol_Prediction_Reader.__new__(PredictionProtocol_Prediction_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1889
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1891
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1889
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1892
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1893
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1892
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1895
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactic(self):
 *         return Tactic_Reader.init(self.source.getTactic(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6tactic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6tactic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6tactic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6tactic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Tactic::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1897
 *     @property
 *     def tactic(self):
 *         return Tactic_Reader.init(self.source.getTactic(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_tactic(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getTactic();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1897, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_13Tactic_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1895
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactic(self):
 *         return Tactic_Reader.init(self.source.getTactic(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.tactic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1898
 *     def tactic(self):
 *         return Tactic_Reader.init(self.source.getTactic(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactic(self):
 *         return self.source.hasTactic()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10has_tactic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10has_tactic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10has_tactic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10has_tactic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1900
 *     @property
 *     def has_tactic(self):
 *         return self.source.hasTactic()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTactic();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1900, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1898
 *     def tactic(self):
 *         return Tactic_Reader.init(self.source.getTactic(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactic(self):
 *         return self.source.hasTactic()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.has_tactic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1902
 *         return self.source.hasTactic()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def arguments(self):
 *         return Argument_Reader_List.init(self.source.getArguments(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_9arguments_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_9arguments_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_9arguments___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_9arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1904
 *     @property
 *     def arguments(self):
 *         return Argument_Reader_List.init(self.source.getArguments(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_arguments(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getArguments();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1904, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1902
 *         return self.source.hasTactic()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def arguments(self):
 *         return Argument_Reader_List.init(self.source.getArguments(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.arguments.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1905
 *     def arguments(self):
 *         return Argument_Reader_List.init(self.source.getArguments(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_arguments(self):
 *         return self.source.hasArguments()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_13has_arguments_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_13has_arguments_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_13has_arguments___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_13has_arguments___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1907
 *     @property
 *     def has_arguments(self):
 *         return self.source.hasArguments()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasArguments();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1907, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1905
 *     def arguments(self):
 *         return Argument_Reader_List.init(self.source.getArguments(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_arguments(self):
 *         return self.source.hasArguments()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.has_arguments.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1909
 *         return self.source.hasArguments()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def confidence(self):
 *         return self.source.getConfidence()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10confidence_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10confidence_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10confidence___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10confidence___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1911
 *     @property
 *     def confidence(self):
 *         return self.source.getConfidence()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.PredictionProtocol.Prediction.schema = _Schema()._init(getPredictionProtocol_Prediction_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getConfidence();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1911, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1909
 *         return self.source.hasArguments()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def confidence(self):
 *         return self.source.getConfidence()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.confidence.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_4__reduce_cython__, "PredictionProtocol_Prediction_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6__setstate_cython__, "PredictionProtocol_Prediction_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1916
 * cdef class PredictionProtocol_TextPrediction_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_TextPrediction()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1916, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1916, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1916, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1916, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1917
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_TextPrediction()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<PredictionProtocol::TextPrediction>();

  /* "pytact/graph_api_capnp_cython.pyx":1918
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_TextPrediction()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1916
 * cdef class PredictionProtocol_TextPrediction_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_TextPrediction()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1921
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_TextPrediction_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_TextPrediction_Reader wrapper = PredictionProtocol_TextPrediction_Reader.__new__(PredictionProtocol_TextPrediction_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_init(PredictionProtocol::TextPrediction::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1922
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_TextPrediction_Reader source, object root):
 *         cdef PredictionProtocol_TextPrediction_Reader wrapper = PredictionProtocol_TextPrediction_Reader.__new__(PredictionProtocol_TextPrediction_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1922, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1923
 *     cdef init(C_PredictionProtocol_TextPrediction_Reader source, object root):
 *         cdef PredictionProtocol_TextPrediction_Reader wrapper = PredictionProtocol_TextPrediction_Reader.__new__(PredictionProtocol_TextPrediction_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1924
 *         cdef PredictionProtocol_TextPrediction_Reader wrapper = PredictionProtocol_TextPrediction_Reader.__new__(PredictionProtocol_TextPrediction_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1925
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1921
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_TextPrediction_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_TextPrediction_Reader wrapper = PredictionProtocol_TextPrediction_Reader.__new__(PredictionProtocol_TextPrediction_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1927
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1929
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1927
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1930
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1931
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1930
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1933
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactic_text(self):
 *         temp = self.source.getTacticText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_11tactic_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_11tactic_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_11tactic_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_11tactic_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self) {
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::kj::StringPtr __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1935
 *     @property
 *     def tactic_text(self):
 *         temp = self.source.getTacticText()             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property
 */
  try {
    __pyx_t_1 = __pyx_v_self->source.getTacticText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1935, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":1936
 *     def tactic_text(self):
 *         temp = self.source.getTacticText()
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 *     @property
 *     def has_tactic_text(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1933
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tactic_text(self):
 *         temp = self.source.getTacticText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader.tactic_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1937
 *         temp = self.source.getTacticText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactic_text(self):
 *         return self.source.hasTacticText()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_15has_tactic_text_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_15has_tactic_text_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_15has_tactic_text___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_15has_tactic_text___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1939
 *     @property
 *     def has_tactic_text(self):
 *         return self.source.hasTacticText()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTacticText();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1939, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1937
 *         temp = self.source.getTacticText()
 *         return (<char*>temp.begin())[:temp.size()]
 *     @property             # <<<<<<<<<<<<<<
 *     def has_tactic_text(self):
 *         return self.source.hasTacticText()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader.has_tactic_text.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1941
 *         return self.source.hasTacticText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def confidence(self):
 *         return self.source.getConfidence()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_10confidence_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_10confidence_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_10confidence___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_10confidence___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1943
 *     @property
 *     def confidence(self):
 *         return self.source.getConfidence()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.PredictionProtocol.TextPrediction.schema = _Schema()._init(getPredictionProtocol_TextPrediction_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getConfidence();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1943, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1941
 *         return self.source.hasTacticText()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def confidence(self):
 *         return self.source.getConfidence()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader.confidence.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_4__reduce_cython__, "PredictionProtocol_TextPrediction_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_6__setstate_cython__, "PredictionProtocol_TextPrediction_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1951
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Response()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1951, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 1951, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1951, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 1951, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1952
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Response()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<PredictionProtocol::Response>();

  /* "pytact/graph_api_capnp_cython.pyx":1953
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Response()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":1951
 *     """
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Response()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1956
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Response_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Response_Reader wrapper = PredictionProtocol_Response_Reader.__new__(PredictionProtocol_Response_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_init(PredictionProtocol::Response::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1957
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Response_Reader source, object root):
 *         cdef PredictionProtocol_Response_Reader wrapper = PredictionProtocol_Response_Reader.__new__(PredictionProtocol_Response_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1957, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1958
 *     cdef init(C_PredictionProtocol_Response_Reader source, object root):
 *         cdef PredictionProtocol_Response_Reader wrapper = PredictionProtocol_Response_Reader.__new__(PredictionProtocol_Response_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":1959
 *         cdef PredictionProtocol_Response_Reader wrapper = PredictionProtocol_Response_Reader.__new__(PredictionProtocol_Response_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":1960
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1956
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Response_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Response_Reader wrapper = PredictionProtocol_Response_Reader.__new__(PredictionProtocol_Response_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1962
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1964
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1962
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1965
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1966
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 *     @property
 *     def which(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1965
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 *     @property
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1967
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return PredictionProtocol_Response_Which(self.source.which())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5which_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5which_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5which___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5which___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __PYX_ENUM_CLASS_DECL PredictionProtocol::Response::Which __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1969
 *     @property
 *     def which(self):
 *         return PredictionProtocol_Response_Which(self.source.which())             # <<<<<<<<<<<<<<
 *     @property
 *     def which_raw(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PredictionProtocol_Response_Whic); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    __pyx_t_3 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1969, __pyx_L1_error)
  }
  __pyx_t_4 = [](const PredictionProtocol::Response::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1967
 *     def __repr__(self):
 *         return repr(self.dynamic)
 *     @property             # <<<<<<<<<<<<<<
 *     def which(self):
 *         return PredictionProtocol_Response_Which(self.source.which())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.which.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1970
 *     def which(self):
 *         return PredictionProtocol_Response_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9which_raw_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9which_raw_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9which_raw___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9which_raw___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __PYX_ENUM_CLASS_DECL PredictionProtocol::Response::Which __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1972
 *     @property
 *     def which_raw(self):
 *         return self.source.which()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.which();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1972, __pyx_L1_error)
  }
  __pyx_t_2 = [](const PredictionProtocol::Response::Which& x){return __Pyx_PyInt_From_uint16_t((uint16_t)x);}(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1970
 *     def which(self):
 *         return PredictionProtocol_Response_Which(self.source.which())
 *     @property             # <<<<<<<<<<<<<<
 *     def which_raw(self):
 *         return self.source.which()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.which_raw.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1974
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def initialized(self):
 *         return self.source.getInitialized()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_11initialized_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_11initialized_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_11initialized___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_11initialized___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1976
 *     @property
 *     def initialized(self):
 *         return self.source.getInitialized()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_initialized(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_v_self->source.getInitialized();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1976, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1974
 *         return self.source.which()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def initialized(self):
 *         return self.source.getInitialized()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.initialized.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1977
 *     def initialized(self):
 *         return self.source.getInitialized()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_initialized(self):
 *         return self.source.isInitialized()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14is_initialized_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14is_initialized_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14is_initialized___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14is_initialized___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1979
 *     @property
 *     def is_initialized(self):
 *         return self.source.isInitialized()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isInitialized();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1979, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1977
 *     def initialized(self):
 *         return self.source.getInitialized()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_initialized(self):
 *         return self.source.isInitialized()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.is_initialized.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1981
 *         return self.source.isInitialized()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def prediction(self):
 *         return PredictionProtocol_Prediction_Reader_List.init(self.source.getPrediction(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_10prediction_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_10prediction_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_10prediction___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_10prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<PredictionProtocol::Prediction, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1983
 *     @property
 *     def prediction(self):
 *         return PredictionProtocol_Prediction_Reader_List.init(self.source.getPrediction(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_prediction(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getPrediction();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1983, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1981
 *         return self.source.isInitialized()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def prediction(self):
 *         return PredictionProtocol_Prediction_Reader_List.init(self.source.getPrediction(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.prediction.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1984
 *     def prediction(self):
 *         return PredictionProtocol_Prediction_Reader_List.init(self.source.getPrediction(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_prediction(self):
 *         return self.source.isPrediction()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_13is_prediction_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_13is_prediction_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_13is_prediction___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_13is_prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1986
 *     @property
 *     def is_prediction(self):
 *         return self.source.isPrediction()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_prediction(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isPrediction();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1986, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1984
 *     def prediction(self):
 *         return PredictionProtocol_Prediction_Reader_List.init(self.source.getPrediction(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_prediction(self):
 *         return self.source.isPrediction()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.is_prediction.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1987
 *     def is_prediction(self):
 *         return self.source.isPrediction()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_prediction(self):
 *         return self.source.hasPrediction()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14has_prediction_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14has_prediction_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14has_prediction___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14has_prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1989
 *     @property
 *     def has_prediction(self):
 *         return self.source.hasPrediction()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasPrediction();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1989, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1987
 *     def is_prediction(self):
 *         return self.source.isPrediction()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_prediction(self):
 *         return self.source.hasPrediction()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.has_prediction.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1991
 *         return self.source.hasPrediction()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def text_prediction(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15text_prediction_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15text_prediction_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15text_prediction___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15text_prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<PredictionProtocol::TextPrediction, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":1997
 *         sorted by decreasing confidence.
 *         """
 *         return PredictionProtocol_TextPrediction_Reader_List.init(self.source.getTextPrediction(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_text_prediction(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getTextPrediction();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 1997, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1991
 *         return self.source.hasPrediction()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def text_prediction(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.text_prediction.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":1998
 *         """
 *         return PredictionProtocol_TextPrediction_Reader_List.init(self.source.getTextPrediction(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_text_prediction(self):
 *         return self.source.isTextPrediction()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_18is_text_prediction_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_18is_text_prediction_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_18is_text_prediction___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_18is_text_prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2000
 *     @property
 *     def is_text_prediction(self):
 *         return self.source.isTextPrediction()             # <<<<<<<<<<<<<<
 *     @property
 *     def has_text_prediction(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isTextPrediction();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2000, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":1998
 *         """
 *         return PredictionProtocol_TextPrediction_Reader_List.init(self.source.getTextPrediction(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_text_prediction(self):
 *         return self.source.isTextPrediction()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.is_text_prediction.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2001
 *     def is_text_prediction(self):
 *         return self.source.isTextPrediction()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_text_prediction(self):
 *         return self.source.hasTextPrediction()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_19has_text_prediction_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_19has_text_prediction_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_19has_text_prediction___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_19has_text_prediction___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2003
 *     @property
 *     def has_text_prediction(self):
 *         return self.source.hasTextPrediction()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasTextPrediction();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2003, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2001
 *     def is_text_prediction(self):
 *         return self.source.isTextPrediction()
 *     @property             # <<<<<<<<<<<<<<
 *     def has_text_prediction(self):
 *         return self.source.hasTextPrediction()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.has_text_prediction.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2005
 *         return self.source.hasTextPrediction()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def synchronized(self):
 *         return self.source.getSynchronized()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12synchronized_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12synchronized_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12synchronized___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12synchronized___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2007
 *     @property
 *     def synchronized(self):
 *         return self.source.getSynchronized()             # <<<<<<<<<<<<<<
 *     @property
 *     def is_synchronized(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getSynchronized();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2007, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_uint64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2005
 *         return self.source.hasTextPrediction()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def synchronized(self):
 *         return self.source.getSynchronized()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.synchronized.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2008
 *     def synchronized(self):
 *         return self.source.getSynchronized()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_synchronized(self):
 *         return self.source.isSynchronized()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15is_synchronized_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15is_synchronized_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15is_synchronized___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15is_synchronized___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2010
 *     @property
 *     def is_synchronized(self):
 *         return self.source.isSynchronized()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isSynchronized();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2010, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2008
 *     def synchronized(self):
 *         return self.source.getSynchronized()
 *     @property             # <<<<<<<<<<<<<<
 *     def is_synchronized(self):
 *         return self.source.isSynchronized()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.is_synchronized.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2012
 *         return self.source.isSynchronized()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alignment(self):
 *         return PredictionProtocol_Response_Alignment_Reader.init(self.source.getAlignment(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9alignment_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9alignment_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9alignment___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9alignment___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PredictionProtocol::Response::Alignment::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2014
 *     @property
 *     def alignment(self):
 *         return PredictionProtocol_Response_Alignment_Reader.init(self.source.getAlignment(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def is_alignment(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getAlignment();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2014, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2012
 *         return self.source.isSynchronized()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alignment(self):
 *         return PredictionProtocol_Response_Alignment_Reader.init(self.source.getAlignment(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.alignment.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2015
 *     def alignment(self):
 *         return PredictionProtocol_Response_Alignment_Reader.init(self.source.getAlignment(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_alignment(self):
 *         return self.source.isAlignment()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12is_alignment_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12is_alignment_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12is_alignment___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12is_alignment___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2017
 *     @property
 *     def is_alignment(self):
 *         return self.source.isAlignment()             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.PredictionProtocol.Response.schema = _Schema()._init(getPredictionProtocol_Response_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.isAlignment();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2017, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2015
 *     def alignment(self):
 *         return PredictionProtocol_Response_Alignment_Reader.init(self.source.getAlignment(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def is_alignment(self):
 *         return self.source.isAlignment()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.is_alignment.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_4__reduce_cython__, "PredictionProtocol_Response_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_6__setstate_cython__, "PredictionProtocol_Response_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2022
 * cdef class PredictionProtocol_Response_Alignment_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Response_Alignment()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 2022, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 2022, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 2022, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 2022, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2023
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Response_Alignment()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<PredictionProtocol::Response::Alignment>();

  /* "pytact/graph_api_capnp_cython.pyx":2024
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Response_Alignment()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":2022
 * cdef class PredictionProtocol_Response_Alignment_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Response_Alignment()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2027
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Response_Alignment_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Response_Alignment_Reader wrapper = PredictionProtocol_Response_Alignment_Reader.__new__(PredictionProtocol_Response_Alignment_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_init(PredictionProtocol::Response::Alignment::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2028
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Response_Alignment_Reader source, object root):
 *         cdef PredictionProtocol_Response_Alignment_Reader wrapper = PredictionProtocol_Response_Alignment_Reader.__new__(PredictionProtocol_Response_Alignment_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2028, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2029
 *     cdef init(C_PredictionProtocol_Response_Alignment_Reader source, object root):
 *         cdef PredictionProtocol_Response_Alignment_Reader wrapper = PredictionProtocol_Response_Alignment_Reader.__new__(PredictionProtocol_Response_Alignment_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2030
 *         cdef PredictionProtocol_Response_Alignment_Reader wrapper = PredictionProtocol_Response_Alignment_Reader.__new__(PredictionProtocol_Response_Alignment_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2031
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2027
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Response_Alignment_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Response_Alignment_Reader wrapper = PredictionProtocol_Response_Alignment_Reader.__new__(PredictionProtocol_Response_Alignment_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2033
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2035
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2033
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2036
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2037
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2036
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2039
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unaligned_tactics(self):
 *         return Int64_List.init(self.source.getUnalignedTactics(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_17unaligned_tactics_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_17unaligned_tactics_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_17unaligned_tactics___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_17unaligned_tactics___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2041
 *     @property
 *     def unaligned_tactics(self):
 *         return Int64_List.init(self.source.getUnalignedTactics(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_unaligned_tactics(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getUnalignedTactics();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2041, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_10Int64_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2039
 *         return repr(self.dynamic)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unaligned_tactics(self):
 *         return Int64_List.init(self.source.getUnalignedTactics(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.unaligned_tactics.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2042
 *     def unaligned_tactics(self):
 *         return Int64_List.init(self.source.getUnalignedTactics(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_unaligned_tactics(self):
 *         return self.source.hasUnalignedTactics()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21has_unaligned_tactics_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21has_unaligned_tactics_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21has_unaligned_tactics___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21has_unaligned_tactics___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2044
 *     @property
 *     def has_unaligned_tactics(self):
 *         return self.source.hasUnalignedTactics()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasUnalignedTactics();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2044, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2042
 *     def unaligned_tactics(self):
 *         return Int64_List.init(self.source.getUnalignedTactics(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_unaligned_tactics(self):
 *         return self.source.hasUnalignedTactics()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.has_unaligned_tactics.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2046
 *         return self.source.hasUnalignedTactics()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unaligned_definitions(self):
 *         return Node_Reader_List.init(self.source.getUnalignedDefinitions(), self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21unaligned_definitions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21unaligned_definitions_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21unaligned_definitions___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21unaligned_definitions___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2048
 *     @property
 *     def unaligned_definitions(self):
 *         return Node_Reader_List.init(self.source.getUnalignedDefinitions(), self.root)             # <<<<<<<<<<<<<<
 *     @property
 *     def has_unaligned_definitions(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.getUnalignedDefinitions();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2048, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6pytact_22graph_api_capnp_cython_16Node_Reader_List_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2046
 *         return self.source.hasUnalignedTactics()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unaligned_definitions(self):
 *         return Node_Reader_List.init(self.source.getUnalignedDefinitions(), self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.unaligned_definitions.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2049
 *     def unaligned_definitions(self):
 *         return Node_Reader_List.init(self.source.getUnalignedDefinitions(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_unaligned_definitions(self):
 *         return self.source.hasUnalignedDefinitions()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_25has_unaligned_definitions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_25has_unaligned_definitions_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_25has_unaligned_definitions___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_25has_unaligned_definitions___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2051
 *     @property
 *     def has_unaligned_definitions(self):
 *         return self.source.hasUnalignedDefinitions()             # <<<<<<<<<<<<<<
 * 
 * cdef class ConflatableEdges_Reader:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->source.hasUnalignedDefinitions();
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2051, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2049
 *     def unaligned_definitions(self):
 *         return Node_Reader_List.init(self.source.getUnalignedDefinitions(), self.root)
 *     @property             # <<<<<<<<<<<<<<
 *     def has_unaligned_definitions(self):
 *         return self.source.hasUnalignedDefinitions()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.has_unaligned_definitions.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_4__reduce_cython__, "PredictionProtocol_Response_Alignment_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_6__setstate_cython__, "PredictionProtocol_Response_Alignment_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2055
 * cdef class ConflatableEdges_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asConflatableEdges()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

/* Python wrapper */
static int __pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dyn,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dyn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 2055, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 2055, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_dyn = ((struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 2055, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ConflatableEdges_Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dyn), __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader, 1, "dyn", 0))) __PYX_ERR(1, 2055, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader___init__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *)__pyx_v_self), __pyx_v_dyn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader___init__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_self, struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader *__pyx_v_dyn) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2056
 * 
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asConflatableEdges()             # <<<<<<<<<<<<<<
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 * 
 */
  __pyx_v_self->source = (( ::capnp::DynamicStruct::Reader)__pyx_v_dyn->thisptr).as<ConflatableEdges>();

  /* "pytact/graph_api_capnp_cython.pyx":2057
 *     def __init__(self, _DynamicStructReader dyn):
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asConflatableEdges()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF((PyObject *)__pyx_v_dyn);
  __Pyx_GIVEREF((PyObject *)__pyx_v_dyn);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = ((PyObject *)__pyx_v_dyn);

  /* "pytact/graph_api_capnp_cython.pyx":2055
 * cdef class ConflatableEdges_Reader:
 * 
 *     def __init__(self, _DynamicStructReader dyn):             # <<<<<<<<<<<<<<
 *         self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asConflatableEdges()
 *         self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2060
 * 
 *     @staticmethod
 *     cdef init(C_ConflatableEdges_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ConflatableEdges_Reader wrapper = ConflatableEdges_Reader.__new__(ConflatableEdges_Reader)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_init(ConflatableEdges::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2061
 *     @staticmethod
 *     cdef init(C_ConflatableEdges_Reader source, object root):
 *         cdef ConflatableEdges_Reader wrapper = ConflatableEdges_Reader.__new__(ConflatableEdges_Reader)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2061, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2062
 *     cdef init(C_ConflatableEdges_Reader source, object root):
 *         cdef ConflatableEdges_Reader wrapper = ConflatableEdges_Reader.__new__(ConflatableEdges_Reader)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2063
 *         cdef ConflatableEdges_Reader wrapper = ConflatableEdges_Reader.__new__(ConflatableEdges_Reader)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2064
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2060
 * 
 *     @staticmethod
 *     cdef init(C_ConflatableEdges_Reader source, object root):             # <<<<<<<<<<<<<<
 *         cdef ConflatableEdges_Reader wrapper = ConflatableEdges_Reader.__new__(ConflatableEdges_Reader)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ConflatableEdges_Reader.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2066
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7dynamic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7dynamic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7dynamic___get__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7dynamic___get__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2068
 *     @property
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return repr(self.dynamic)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5capnp_3lib_5capnp_to_python_reader((( ::capnp::DynamicValue::Reader)__pyx_v_self->source), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2066
 *         return wrapper
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ConflatableEdges_Reader.dynamic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2069
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * pytact.graph_api_capnp.ConflatableEdges.schema = _Schema()._init(getConflatableEdges_Schema()).as_struct()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_2__repr__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_2__repr__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2070
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):
 *         return repr(self.dynamic)             # <<<<<<<<<<<<<<
 * pytact.graph_api_capnp.ConflatableEdges.schema = _Schema()._init(getConflatableEdges_Schema()).as_struct()
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2069
 *     def dynamic(self):
 *         return to_python_reader(<DynamicValue.Reader>self.source, self.root)
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.dynamic)
 * pytact.graph_api_capnp.ConflatableEdges.schema = _Schema()._init(getConflatableEdges_Schema()).as_struct()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ConflatableEdges_Reader.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_4__reduce_cython__, "ConflatableEdges_Reader.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ConflatableEdges_Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_6__setstate_cython__, "ConflatableEdges_Reader.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ConflatableEdges_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ConflatableEdges_Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2077
 * 
 *     @staticmethod
 *     cdef init(C_Uint16_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Uint16_List wrapper = Uint16_List.__new__(Uint16_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11Uint16_List_init( ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2078
 *     @staticmethod
 *     cdef init(C_Uint16_List source, object root):
 *         cdef Uint16_List wrapper = Uint16_List.__new__(Uint16_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint16_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2078, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2079
 *     cdef init(C_Uint16_List source, object root):
 *         cdef Uint16_List wrapper = Uint16_List.__new__(Uint16_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2080
 *         cdef Uint16_List wrapper = Uint16_List.__new__(Uint16_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2081
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2077
 * 
 *     @staticmethod
 *     cdef init(C_Uint16_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Uint16_List wrapper = Uint16_List.__new__(Uint16_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint16_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2083
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2083, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint16_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  uint16_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2084
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2085
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return source[index]
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2086
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return source[index]
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2086, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2085
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return source[index]
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2087
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return source[index]             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2087, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_From_uint16_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2083
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint16_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2089
 *         return source[index]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2090
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * 
 * cdef class Uint32_List:
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2089
 *         return source[index]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11Uint16_List_4__reduce_cython__, "Uint16_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint16_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint16_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint16_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11Uint16_List_6__setstate_cython__, "Uint16_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint16_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint16_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint16_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint16_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint16_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2095
 * 
 *     @staticmethod
 *     cdef init(C_Uint32_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Uint32_List wrapper = Uint32_List.__new__(Uint32_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11Uint32_List_init( ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2096
 *     @staticmethod
 *     cdef init(C_Uint32_List source, object root):
 *         cdef Uint32_List wrapper = Uint32_List.__new__(Uint32_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint32_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2096, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2097
 *     cdef init(C_Uint32_List source, object root):
 *         cdef Uint32_List wrapper = Uint32_List.__new__(Uint32_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2098
 *         cdef Uint32_List wrapper = Uint32_List.__new__(Uint32_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2099
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2095
 * 
 *     @staticmethod
 *     cdef init(C_Uint32_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Uint32_List wrapper = Uint32_List.__new__(Uint32_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint32_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2101
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2101, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint32_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2102
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2103
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return source[index]
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2104
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return source[index]
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2104, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2103
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return source[index]
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2105
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return source[index]             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2105, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_From_uint32_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2101
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint32_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2107
 *         return source[index]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2108
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * 
 * cdef class Uint64_List:
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2107
 *         return source[index]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11Uint32_List_4__reduce_cython__, "Uint32_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint32_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint32_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint32_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11Uint32_List_6__setstate_cython__, "Uint32_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint32_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint32_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint32_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint32_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint32_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2113
 * 
 *     @staticmethod
 *     cdef init(C_Uint64_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Uint64_List wrapper = Uint64_List.__new__(Uint64_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11Uint64_List_init( ::capnp::List<uint64_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2114
 *     @staticmethod
 *     cdef init(C_Uint64_List source, object root):
 *         cdef Uint64_List wrapper = Uint64_List.__new__(Uint64_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint64_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2114, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2115
 *     cdef init(C_Uint64_List source, object root):
 *         cdef Uint64_List wrapper = Uint64_List.__new__(Uint64_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2116
 *         cdef Uint64_List wrapper = Uint64_List.__new__(Uint64_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2117
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2113
 * 
 *     @staticmethod
 *     cdef init(C_Uint64_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Uint64_List wrapper = Uint64_List.__new__(Uint64_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint64_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2119
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2119, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint64_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<uint64_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<uint64_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  uint64_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2120
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2121
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return source[index]
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2122
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return source[index]
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2122, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2121
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return source[index]
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2123
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return source[index]             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2123, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_From_uint64_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2119
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint64_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2125
 *         return source[index]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2126
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * 
 * cdef class Int64_List:
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2125
 *         return source[index]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11Uint64_List_4__reduce_cython__, "Uint64_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint64_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint64_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint64_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11Uint64_List_6__setstate_cython__, "Uint64_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint64_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint64_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint64_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11Uint64_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Uint64_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2131
 * 
 *     @staticmethod
 *     cdef init(C_Int64_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Int64_List wrapper = Int64_List.__new__(Int64_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_10Int64_List_init( ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2132
 *     @staticmethod
 *     cdef init(C_Int64_List source, object root):
 *         cdef Int64_List wrapper = Int64_List.__new__(Int64_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Int64_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2132, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2133
 *     cdef init(C_Int64_List source, object root):
 *         cdef Int64_List wrapper = Int64_List.__new__(Int64_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2134
 *         cdef Int64_List wrapper = Int64_List.__new__(Int64_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2135
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2131
 * 
 *     @staticmethod
 *     cdef init(C_Int64_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Int64_List wrapper = Int64_List.__new__(Int64_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Int64_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2137
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2137, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Int64_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int64_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2138
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2139
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return source[index]
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2140
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return source[index]
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2140, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2139
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return source[index]
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2141
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return source[index]             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2141, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2137
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Int64_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2143
 *         return source[index]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2144
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * 
 * cdef class String_List:
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2143
 *         return source[index]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_10Int64_List_4__reduce_cython__, "Int64_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_10Int64_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_10Int64_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Int64_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_10Int64_List_6__setstate_cython__, "Int64_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_10Int64_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_10Int64_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Int64_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_10Int64_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Int64_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2149
 * 
 *     @staticmethod
 *     cdef init(C_String_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef String_List wrapper = String_List.__new__(String_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_11String_List_init( ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2150
 *     @staticmethod
 *     cdef init(C_String_List source, object root):
 *         cdef String_List wrapper = String_List.__new__(String_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_String_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2150, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2151
 *     cdef init(C_String_List source, object root):
 *         cdef String_List wrapper = String_List.__new__(String_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2152
 *         cdef String_List wrapper = String_List.__new__(String_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2153
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2149
 * 
 *     @staticmethod
 *     cdef init(C_String_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef String_List wrapper = String_List.__new__(String_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.String_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2155
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2155, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.String_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11String_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11String_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader __pyx_v_source;
   ::kj::StringPtr __pyx_v_temp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
   ::kj::StringPtr __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2156
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2157
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         temp = source[index]
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2158
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         temp = source[index]
 *         return (<char*>temp.begin())[:temp.size()]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2158, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2157
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         temp = source[index]
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2159
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         temp = source[index]             # <<<<<<<<<<<<<<
 *         return (<char*>temp.begin())[:temp.size()]
 * 
 */
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2159, __pyx_L1_error)
  }
  __pyx_v_temp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4);

  /* "pytact/graph_api_capnp_cython.pyx":2160
 *             raise IndexError('Out of bounds')
 *         temp = source[index]
 *         return (<char*>temp.begin())[:temp.size()]             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyStr_FromStringAndSize(((char *)__pyx_v_temp.begin()) + 0, __pyx_v_temp.size() - 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2155
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.String_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2162
 *         return (<char*>temp.begin())[:temp.size()]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class PredictionProtocol_Prediction_Reader_List:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11String_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_11String_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2163
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * cdef class PredictionProtocol_Prediction_Reader_List:
 * 
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2162
 *         return (<char*>temp.begin())[:temp.size()]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class PredictionProtocol_Prediction_Reader_List:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11String_List_4__reduce_cython__, "String_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11String_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11String_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11String_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11String_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.String_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_11String_List_6__setstate_cython__, "String_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_11String_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11String_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.String_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_11String_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_11String_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.String_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2167
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Prediction_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Prediction_Reader_List wrapper = PredictionProtocol_Prediction_Reader_List.__new__(PredictionProtocol_Prediction_Reader_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_init( ::capnp::List<PredictionProtocol::Prediction, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2168
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Prediction_Reader_List source, object root):
 *         cdef PredictionProtocol_Prediction_Reader_List wrapper = PredictionProtocol_Prediction_Reader_List.__new__(PredictionProtocol_Prediction_Reader_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2168, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2169
 *     cdef init(C_PredictionProtocol_Prediction_Reader_List source, object root):
 *         cdef PredictionProtocol_Prediction_Reader_List wrapper = PredictionProtocol_Prediction_Reader_List.__new__(PredictionProtocol_Prediction_Reader_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2170
 *         cdef PredictionProtocol_Prediction_Reader_List wrapper = PredictionProtocol_Prediction_Reader_List.__new__(PredictionProtocol_Prediction_Reader_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2171
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2167
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_Prediction_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_Prediction_Reader_List wrapper = PredictionProtocol_Prediction_Reader_List.__new__(PredictionProtocol_Prediction_Reader_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2173
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2173, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<PredictionProtocol::Prediction, ::capnp::Kind::STRUCT>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<PredictionProtocol::Prediction, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PredictionProtocol::Prediction::Reader __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2174
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2175
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return PredictionProtocol_Prediction_Reader.init(source[index], self.root)
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2176
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return PredictionProtocol_Prediction_Reader.init(source[index], self.root)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2176, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2175
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return PredictionProtocol_Prediction_Reader.init(source[index], self.root)
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2177
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return PredictionProtocol_Prediction_Reader.init(source[index], self.root)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2177, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2173
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2179
 *         return PredictionProtocol_Prediction_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class Graph_Node_Reader_List:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2180
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * cdef class Graph_Node_Reader_List:
 * 
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2179
 *         return PredictionProtocol_Prediction_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class Graph_Node_Reader_List:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_4__reduce_cython__, "PredictionProtocol_Prediction_Reader_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_6__setstate_cython__, "PredictionProtocol_Prediction_Reader_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2184
 * 
 *     @staticmethod
 *     cdef init(C_Graph_Node_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_Node_Reader_List wrapper = Graph_Node_Reader_List.__new__(Graph_Node_Reader_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_init( ::capnp::List<Graph::Node, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2185
 *     @staticmethod
 *     cdef init(C_Graph_Node_Reader_List source, object root):
 *         cdef Graph_Node_Reader_List wrapper = Graph_Node_Reader_List.__new__(Graph_Node_Reader_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2185, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2186
 *     cdef init(C_Graph_Node_Reader_List source, object root):
 *         cdef Graph_Node_Reader_List wrapper = Graph_Node_Reader_List.__new__(Graph_Node_Reader_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2187
 *         cdef Graph_Node_Reader_List wrapper = Graph_Node_Reader_List.__new__(Graph_Node_Reader_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2188
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2184
 * 
 *     @staticmethod
 *     cdef init(C_Graph_Node_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_Node_Reader_List wrapper = Graph_Node_Reader_List.__new__(Graph_Node_Reader_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2190
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2190, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<Graph::Node, ::capnp::Kind::STRUCT>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Graph::Node, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Graph::Node::Reader __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2191
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2192
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return Graph_Node_Reader.init(source[index], self.root)
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2193
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return Graph_Node_Reader.init(source[index], self.root)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2193, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2192
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return Graph_Node_Reader.init(source[index], self.root)
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2194
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return Graph_Node_Reader.init(source[index], self.root)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2194, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2190
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2196
 *         return Graph_Node_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class AbstractTactic_Reader_List:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2197
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * cdef class AbstractTactic_Reader_List:
 * 
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2196
 *         return Graph_Node_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class AbstractTactic_Reader_List:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_4__reduce_cython__, "Graph_Node_Reader_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_6__setstate_cython__, "Graph_Node_Reader_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_Node_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2201
 * 
 *     @staticmethod
 *     cdef init(C_AbstractTactic_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef AbstractTactic_Reader_List wrapper = AbstractTactic_Reader_List.__new__(AbstractTactic_Reader_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_init( ::capnp::List<AbstractTactic, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2202
 *     @staticmethod
 *     cdef init(C_AbstractTactic_Reader_List source, object root):
 *         cdef AbstractTactic_Reader_List wrapper = AbstractTactic_Reader_List.__new__(AbstractTactic_Reader_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2202, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2203
 *     cdef init(C_AbstractTactic_Reader_List source, object root):
 *         cdef AbstractTactic_Reader_List wrapper = AbstractTactic_Reader_List.__new__(AbstractTactic_Reader_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2204
 *         cdef AbstractTactic_Reader_List wrapper = AbstractTactic_Reader_List.__new__(AbstractTactic_Reader_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2205
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2201
 * 
 *     @staticmethod
 *     cdef init(C_AbstractTactic_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef AbstractTactic_Reader_List wrapper = AbstractTactic_Reader_List.__new__(AbstractTactic_Reader_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2207
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2207, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<AbstractTactic, ::capnp::Kind::STRUCT>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<AbstractTactic, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  AbstractTactic::Reader __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2208
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2209
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return AbstractTactic_Reader.init(source[index], self.root)
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2210
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return AbstractTactic_Reader.init(source[index], self.root)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2210, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2209
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return AbstractTactic_Reader.init(source[index], self.root)
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2211
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return AbstractTactic_Reader.init(source[index], self.root)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2211, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2207
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2213
 *         return AbstractTactic_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class ProofStep_Reader_List:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2214
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * cdef class ProofStep_Reader_List:
 * 
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2213
 *         return AbstractTactic_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class ProofStep_Reader_List:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_4__reduce_cython__, "AbstractTactic_Reader_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_6__setstate_cython__, "AbstractTactic_Reader_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.AbstractTactic_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2218
 * 
 *     @staticmethod
 *     cdef init(C_ProofStep_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofStep_Reader_List wrapper = ProofStep_Reader_List.__new__(ProofStep_Reader_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_init( ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2219
 *     @staticmethod
 *     cdef init(C_ProofStep_Reader_List source, object root):
 *         cdef ProofStep_Reader_List wrapper = ProofStep_Reader_List.__new__(ProofStep_Reader_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2219, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2220
 *     cdef init(C_ProofStep_Reader_List source, object root):
 *         cdef ProofStep_Reader_List wrapper = ProofStep_Reader_List.__new__(ProofStep_Reader_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2221
 *         cdef ProofStep_Reader_List wrapper = ProofStep_Reader_List.__new__(ProofStep_Reader_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2222
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2218
 * 
 *     @staticmethod
 *     cdef init(C_ProofStep_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofStep_Reader_List wrapper = ProofStep_Reader_List.__new__(ProofStep_Reader_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2224
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2224, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  ProofStep::Reader __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2225
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2226
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return ProofStep_Reader.init(source[index], self.root)
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2227
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return ProofStep_Reader.init(source[index], self.root)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2227, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2226
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return ProofStep_Reader.init(source[index], self.root)
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2228
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return ProofStep_Reader.init(source[index], self.root)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2228, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2224
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2230
 *         return ProofStep_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class PredictionProtocol_TextPrediction_Reader_List:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2231
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * cdef class PredictionProtocol_TextPrediction_Reader_List:
 * 
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2230
 *         return ProofStep_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class PredictionProtocol_TextPrediction_Reader_List:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_4__reduce_cython__, "ProofStep_Reader_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_6__setstate_cython__, "ProofStep_Reader_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofStep_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2235
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_TextPrediction_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_TextPrediction_Reader_List wrapper = PredictionProtocol_TextPrediction_Reader_List.__new__(PredictionProtocol_TextPrediction_Reader_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_init( ::capnp::List<PredictionProtocol::TextPrediction, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2236
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_TextPrediction_Reader_List source, object root):
 *         cdef PredictionProtocol_TextPrediction_Reader_List wrapper = PredictionProtocol_TextPrediction_Reader_List.__new__(PredictionProtocol_TextPrediction_Reader_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2236, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2237
 *     cdef init(C_PredictionProtocol_TextPrediction_Reader_List source, object root):
 *         cdef PredictionProtocol_TextPrediction_Reader_List wrapper = PredictionProtocol_TextPrediction_Reader_List.__new__(PredictionProtocol_TextPrediction_Reader_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2238
 *         cdef PredictionProtocol_TextPrediction_Reader_List wrapper = PredictionProtocol_TextPrediction_Reader_List.__new__(PredictionProtocol_TextPrediction_Reader_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2239
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2235
 * 
 *     @staticmethod
 *     cdef init(C_PredictionProtocol_TextPrediction_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef PredictionProtocol_TextPrediction_Reader_List wrapper = PredictionProtocol_TextPrediction_Reader_List.__new__(PredictionProtocol_TextPrediction_Reader_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2241
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2241, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<PredictionProtocol::TextPrediction, ::capnp::Kind::STRUCT>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<PredictionProtocol::TextPrediction, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PredictionProtocol::TextPrediction::Reader __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2242
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2243
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return PredictionProtocol_TextPrediction_Reader.init(source[index], self.root)
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2244
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return PredictionProtocol_TextPrediction_Reader.init(source[index], self.root)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2244, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2243
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return PredictionProtocol_TextPrediction_Reader.init(source[index], self.root)
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2245
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return PredictionProtocol_TextPrediction_Reader.init(source[index], self.root)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2245, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2241
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2247
 *         return PredictionProtocol_TextPrediction_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class ProofState_Reader_List:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2248
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * cdef class ProofState_Reader_List:
 * 
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2247
 *         return PredictionProtocol_TextPrediction_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class ProofState_Reader_List:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_4__reduce_cython__, "PredictionProtocol_TextPrediction_Reader_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_6__setstate_cython__, "PredictionProtocol_TextPrediction_Reader_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2252
 * 
 *     @staticmethod
 *     cdef init(C_ProofState_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofState_Reader_List wrapper = ProofState_Reader_List.__new__(ProofState_Reader_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_init( ::capnp::List<ProofState, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2253
 *     @staticmethod
 *     cdef init(C_ProofState_Reader_List source, object root):
 *         cdef ProofState_Reader_List wrapper = ProofState_Reader_List.__new__(ProofState_Reader_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Reader_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2253, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2254
 *     cdef init(C_ProofState_Reader_List source, object root):
 *         cdef ProofState_Reader_List wrapper = ProofState_Reader_List.__new__(ProofState_Reader_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2255
 *         cdef ProofState_Reader_List wrapper = ProofState_Reader_List.__new__(ProofState_Reader_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2256
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2252
 * 
 *     @staticmethod
 *     cdef init(C_ProofState_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef ProofState_Reader_List wrapper = ProofState_Reader_List.__new__(ProofState_Reader_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2258
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2258, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<ProofState, ::capnp::Kind::STRUCT>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<ProofState, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  ProofState::Reader __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2259
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2260
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return ProofState_Reader.init(source[index], self.root)
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2261
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return ProofState_Reader.init(source[index], self.root)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2261, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2260
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return ProofState_Reader.init(source[index], self.root)
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2262
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return ProofState_Reader.init(source[index], self.root)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2262, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_6pytact_22graph_api_capnp_cython_17ProofState_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2258
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2264
 *         return ProofState_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class Graph_EdgeTarget_Reader_List:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2265
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * cdef class Graph_EdgeTarget_Reader_List:
 * 
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2264
 *         return ProofState_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class Graph_EdgeTarget_Reader_List:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_4__reduce_cython__, "ProofState_Reader_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_6__setstate_cython__, "ProofState_Reader_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.ProofState_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2269
 * 
 *     @staticmethod
 *     cdef init(C_Graph_EdgeTarget_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_EdgeTarget_Reader_List wrapper = Graph_EdgeTarget_Reader_List.__new__(Graph_EdgeTarget_Reader_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_init( ::capnp::List<Graph::EdgeTarget, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2270
 *     @staticmethod
 *     cdef init(C_Graph_EdgeTarget_Reader_List source, object root):
 *         cdef Graph_EdgeTarget_Reader_List wrapper = Graph_EdgeTarget_Reader_List.__new__(Graph_EdgeTarget_Reader_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2270, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2271
 *     cdef init(C_Graph_EdgeTarget_Reader_List source, object root):
 *         cdef Graph_EdgeTarget_Reader_List wrapper = Graph_EdgeTarget_Reader_List.__new__(Graph_EdgeTarget_Reader_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2272
 *         cdef Graph_EdgeTarget_Reader_List wrapper = Graph_EdgeTarget_Reader_List.__new__(Graph_EdgeTarget_Reader_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2273
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2269
 * 
 *     @staticmethod
 *     cdef init(C_Graph_EdgeTarget_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Graph_EdgeTarget_Reader_List wrapper = Graph_EdgeTarget_Reader_List.__new__(Graph_EdgeTarget_Reader_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2275
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2275, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<Graph::EdgeTarget, ::capnp::Kind::STRUCT>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Graph::EdgeTarget, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Graph::EdgeTarget::Reader __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2276
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2277
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return Graph_EdgeTarget_Reader.init(source[index], self.root)
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2278
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return Graph_EdgeTarget_Reader.init(source[index], self.root)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2278, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2277
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return Graph_EdgeTarget_Reader.init(source[index], self.root)
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2279
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return Graph_EdgeTarget_Reader.init(source[index], self.root)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2279, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2275
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2281
 *         return Graph_EdgeTarget_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class Argument_Reader_List:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2282
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * cdef class Argument_Reader_List:
 * 
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2281
 *         return Graph_EdgeTarget_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class Argument_Reader_List:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_4__reduce_cython__, "Graph_EdgeTarget_Reader_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_6__setstate_cython__, "Graph_EdgeTarget_Reader_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2286
 * 
 *     @staticmethod
 *     cdef init(C_Argument_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Argument_Reader_List wrapper = Argument_Reader_List.__new__(Argument_Reader_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_init( ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2287
 *     @staticmethod
 *     cdef init(C_Argument_Reader_List source, object root):
 *         cdef Argument_Reader_List wrapper = Argument_Reader_List.__new__(Argument_Reader_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Reader_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2287, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2288
 *     cdef init(C_Argument_Reader_List source, object root):
 *         cdef Argument_Reader_List wrapper = Argument_Reader_List.__new__(Argument_Reader_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2289
 *         cdef Argument_Reader_List wrapper = Argument_Reader_List.__new__(Argument_Reader_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2290
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2286
 * 
 *     @staticmethod
 *     cdef init(C_Argument_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Argument_Reader_List wrapper = Argument_Reader_List.__new__(Argument_Reader_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2292
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2292, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Argument::Reader __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2293
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2294
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return Argument_Reader.init(source[index], self.root)
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2295
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return Argument_Reader.init(source[index], self.root)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2295, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2294
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return Argument_Reader.init(source[index], self.root)
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2296
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return Argument_Reader.init(source[index], self.root)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2296, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_6pytact_22graph_api_capnp_cython_15Argument_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2292
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2298
 *         return Argument_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class Outcome_Reader_List:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2299
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * cdef class Outcome_Reader_List:
 * 
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2298
 *         return Argument_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class Outcome_Reader_List:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_4__reduce_cython__, "Argument_Reader_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_6__setstate_cython__, "Argument_Reader_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Argument_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2303
 * 
 *     @staticmethod
 *     cdef init(C_Outcome_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Outcome_Reader_List wrapper = Outcome_Reader_List.__new__(Outcome_Reader_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_init( ::capnp::List<Outcome, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2304
 *     @staticmethod
 *     cdef init(C_Outcome_Reader_List source, object root):
 *         cdef Outcome_Reader_List wrapper = Outcome_Reader_List.__new__(Outcome_Reader_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Reader_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2304, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2305
 *     cdef init(C_Outcome_Reader_List source, object root):
 *         cdef Outcome_Reader_List wrapper = Outcome_Reader_List.__new__(Outcome_Reader_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2306
 *         cdef Outcome_Reader_List wrapper = Outcome_Reader_List.__new__(Outcome_Reader_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2307
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2303
 * 
 *     @staticmethod
 *     cdef init(C_Outcome_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Outcome_Reader_List wrapper = Outcome_Reader_List.__new__(Outcome_Reader_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2309
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2309, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<Outcome, ::capnp::Kind::STRUCT>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Outcome, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Outcome::Reader __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2310
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2311
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return Outcome_Reader.init(source[index], self.root)
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2312
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return Outcome_Reader.init(source[index], self.root)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2312, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2311
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return Outcome_Reader.init(source[index], self.root)
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2313
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return Outcome_Reader.init(source[index], self.root)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2313, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_6pytact_22graph_api_capnp_cython_14Outcome_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2309
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2315
 *         return Outcome_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class Node_Reader_List:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2316
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 * cdef class Node_Reader_List:
 * 
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2315
 *         return Outcome_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 * cdef class Node_Reader_List:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_4__reduce_cython__, "Outcome_Reader_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_6__setstate_cython__, "Outcome_Reader_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Outcome_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2320
 * 
 *     @staticmethod
 *     cdef init(C_Node_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Node_Reader_List wrapper = Node_Reader_List.__new__(Node_Reader_List)
 *         wrapper.source = source
 */

static PyObject *__pyx_f_6pytact_22graph_api_capnp_cython_16Node_Reader_List_init( ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader __pyx_v_source, PyObject *__pyx_v_root) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_v_wrapper = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2321
 *     @staticmethod
 *     cdef init(C_Node_Reader_List source, object root):
 *         cdef Node_Reader_List wrapper = Node_Reader_List.__new__(Node_Reader_List)             # <<<<<<<<<<<<<<
 *         wrapper.source = source
 *         wrapper.root = root
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Node_Reader_List(((PyTypeObject *)__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2321, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2322
 *     cdef init(C_Node_Reader_List source, object root):
 *         cdef Node_Reader_List wrapper = Node_Reader_List.__new__(Node_Reader_List)
 *         wrapper.source = source             # <<<<<<<<<<<<<<
 *         wrapper.root = root
 *         return wrapper
 */
  __pyx_v_wrapper->source = __pyx_v_source;

  /* "pytact/graph_api_capnp_cython.pyx":2323
 *         cdef Node_Reader_List wrapper = Node_Reader_List.__new__(Node_Reader_List)
 *         wrapper.source = source
 *         wrapper.root = root             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __Pyx_INCREF(__pyx_v_root);
  __Pyx_GIVEREF(__pyx_v_root);
  __Pyx_GOTREF(__pyx_v_wrapper->root);
  __Pyx_DECREF(__pyx_v_wrapper->root);
  __pyx_v_wrapper->root = __pyx_v_root;

  /* "pytact/graph_api_capnp_cython.pyx":2324
 *         wrapper.source = source
 *         wrapper.root = root
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, uint index):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = ((PyObject *)__pyx_v_wrapper);
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2320
 * 
 *     @staticmethod
 *     cdef init(C_Node_Reader_List source, object root):             # <<<<<<<<<<<<<<
 *         cdef Node_Reader_List wrapper = Node_Reader_List.__new__(Node_Reader_List)
 *         wrapper.source = source
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader_List.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2326
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index); /*proto*/
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_1__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_index) {
  __pyx_t_5capnp_8includes_5types_uint __pyx_v_index;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_index); {
    __pyx_v_index = __Pyx_PyInt_As_unsigned_int(__pyx_arg_index); if (unlikely((__pyx_v_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2326, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List___getitem__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *)__pyx_v_self), ((__pyx_t_5capnp_8includes_5types_uint)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List___getitem__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_v_self, __pyx_t_5capnp_8includes_5types_uint __pyx_v_index) {
   ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader __pyx_v_source;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
   ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Node::Reader __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2327
 * 
 *     def __getitem__(self, uint index):
 *         source = self.source             # <<<<<<<<<<<<<<
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 */
  __pyx_t_1 = __pyx_v_self->source;
  __pyx_v_source = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "pytact/graph_api_capnp_cython.pyx":2328
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return Node_Reader.init(source[index], self.root)
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_source.size());
  if (unlikely(__pyx_t_2)) {

    /* "pytact/graph_api_capnp_cython.pyx":2329
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return Node_Reader.init(source[index], self.root)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 2329, __pyx_L1_error)

    /* "pytact/graph_api_capnp_cython.pyx":2328
 *     def __getitem__(self, uint index):
 *         source = self.source
 *         if index >= source.size():             # <<<<<<<<<<<<<<
 *             raise IndexError('Out of bounds')
 *         return Node_Reader.init(source[index], self.root)
 */
  }

  /* "pytact/graph_api_capnp_cython.pyx":2330
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')
 *         return Node_Reader.init(source[index], self.root)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_source[__pyx_v_index];
  } catch(...) {
    reraise_kj_exception(); if (!PyErr_Occurred())PyErr_SetString(PyExc_RuntimeError, "Error converting c++ exception.");
    __PYX_ERR(1, 2330, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->root;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_6pytact_22graph_api_capnp_cython_11Node_Reader_init(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2326
 *         return wrapper
 * 
 *     def __getitem__(self, uint index):             # <<<<<<<<<<<<<<
 *         source = self.source
 *         if index >= source.size():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader_List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pytact/graph_api_capnp_cython.pyx":2332
 *         return Node_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List_2__len__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List_2__len__(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2333
 * 
 *     def __len__(self):
 *         return self.source.size()             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_self->source.size();
  goto __pyx_L0;

  /* "pytact/graph_api_capnp_cython.pyx":2332
 *         return Node_Reader.init(source[index], self.root)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.source.size()
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_16Node_Reader_List_4__reduce_cython__, "Node_Reader_List.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_16Node_Reader_List_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16Node_Reader_List_4__reduce_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List_4__reduce_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader_List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pytact_22graph_api_capnp_cython_16Node_Reader_List_6__setstate_cython__, "Node_Reader_List.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6pytact_22graph_api_capnp_cython_16Node_Reader_List_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16Node_Reader_List_6__setstate_cython__};
static PyObject *__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List_6__setstate_cython__(((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pytact_22graph_api_capnp_cython_16Node_Reader_List_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_source_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pytact.graph_api_capnp_cython.Node_Reader_List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "capnp/includes/capnp_cpp.pxd":127
 *         VoidPromise afterDelay(Duration delay)
 * 
 * cdef inline Duration Nanoseconds(int64_t nanos):             # <<<<<<<<<<<<<<
 *     return NANOSECONDS * nanos
 * 
 */

static CYTHON_INLINE  ::kj::Duration __pyx_f_5capnp_8includes_9capnp_cpp_Nanoseconds(int64_t __pyx_v_nanos) {
   ::kj::Duration __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Nanoseconds", 0);

  /* "capnp/includes/capnp_cpp.pxd":128
 * 
 * cdef inline Duration Nanoseconds(int64_t nanos):
 *     return NANOSECONDS * nanos             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "kj/async-io.h" namespace " ::kj":
 */
  __pyx_r = ( ::kj::NANOSECONDS * __pyx_v_nanos);
  goto __pyx_L0;

  /* "capnp/includes/capnp_cpp.pxd":127
 *         VoidPromise afterDelay(Duration delay)
 * 
 * cdef inline Duration Nanoseconds(int64_t nanos):             # <<<<<<<<<<<<<<
 *     return NANOSECONDS * nanos
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":13
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):             # <<<<<<<<<<<<<<
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 */

/* Python wrapper */
static int __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_parents = 0;
  PyObject *__pyx_v_dct = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_parents,&__pyx_n_s_dct,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_parents,&__pyx_n_s_dct,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 13, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_parents)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 13, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 13, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dct)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 13, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 13, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 13, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_name = values[0];
    __pyx_v_parents = values[1];
    __pyx_v_dct = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 13, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name, PyObject *__pyx_v_parents, PyObject *__pyx_v_dct) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "EnumBase":14
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)             # <<<<<<<<<<<<<<
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyType_Type)), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_3, ((PyObject *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 4+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":15
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()             # <<<<<<<<<<<<<<
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 */
  __Pyx_INCREF(__Pyx_OrderedDict);
  __pyx_t_2 = __Pyx_OrderedDict; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":13
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):             # <<<<<<<<<<<<<<
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":16
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):             # <<<<<<<<<<<<<<
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__(PyObject *__pyx_v_cls); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__(PyObject *__pyx_v_cls) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "EnumBase":17
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())             # <<<<<<<<<<<<<<
 *     def __getitem__(cls, name):
 *         return cls.__members__[name]
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_2, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "EnumBase":16
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):             # <<<<<<<<<<<<<<
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":18
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):             # <<<<<<<<<<<<<<
 *         return cls.__members__[name]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__(PyObject *__pyx_v_cls, PyObject *__pyx_v_name); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__(PyObject *__pyx_v_cls, PyObject *__pyx_v_name) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls), ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "EnumBase":19
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 *         return cls.__members__[name]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "EnumBase":18
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):             # <<<<<<<<<<<<<<
 *         return cls.__members__[name]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle___Pyx_EnumMeta__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":24
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int, metaclass=__Pyx_EnumMeta):
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_1__new__ = {"__new__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_name = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__new__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_value,&__pyx_n_s_name,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_value,&__pyx_n_s_name,0};
    #endif
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_cls)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 24, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 24, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__new__", 0, 2, 3, 1); __PYX_ERR(0, 24, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 24, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__new__") < 0)) __PYX_ERR(0, 24, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cls = values[0];
    __pyx_v_value = values[1];
    __pyx_v_name = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__new__", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 24, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(__pyx_self, __pyx_v_cls, __pyx_v_value, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_value, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__new__", 0);

  /* "EnumBase":25
 * class __Pyx_EnumBase(int, metaclass=__Pyx_EnumMeta):
 *     def __new__(cls, value, name=None):
 *         for v in cls:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return v
 */
  if (likely(PyList_CheckExact(__pyx_v_cls)) || PyTuple_CheckExact(__pyx_v_cls)) {
    __pyx_t_1 = __pyx_v_cls; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_cls); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 25, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 25, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 25, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "EnumBase":26
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return v
 *         if name is None:
 */
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_v, __pyx_v_value, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 26, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 26, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "EnumBase":27
 *         for v in cls:
 *             if v == value:
 *                 return v             # <<<<<<<<<<<<<<
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_v);
      __pyx_r = __pyx_v_v;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "EnumBase":26
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return v
 *         if name is None:
 */
    }

    /* "EnumBase":25
 * class __Pyx_EnumBase(int, metaclass=__Pyx_EnumMeta):
 *     def __new__(cls, value, name=None):
 *         for v in cls:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":28
 *             if v == value:
 *                 return v
 *         if name is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 */
  __pyx_t_5 = (__pyx_v_name == Py_None);
  if (unlikely(__pyx_t_5)) {

    /* "EnumBase":29
 *                 return v
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)             # <<<<<<<<<<<<<<
 *         res = int.__new__(cls, value)
 *         res.name = name
 */
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Unknown_enum_value_s, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 29, __pyx_L1_error)

    /* "EnumBase":28
 *             if v == value:
 *                 return v
 *         if name is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 */
  }

  /* "EnumBase":30
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)             # <<<<<<<<<<<<<<
 *         res.name = name
 *         setattr(cls, name, res)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyInt_Type)), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_7 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_cls, __pyx_v_value};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_res = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "EnumBase":31
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 *         res.name = name             # <<<<<<<<<<<<<<
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_res, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 31, __pyx_L1_error)

  /* "EnumBase":32
 *         res = int.__new__(cls, value)
 *         res.name = name
 *         setattr(cls, name, res)             # <<<<<<<<<<<<<<
 *         cls.__members__[name] = res
 *         return res
 */
  __pyx_t_8 = PyObject_SetAttr(__pyx_v_cls, __pyx_v_name, __pyx_v_res); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 32, __pyx_L1_error)

  /* "EnumBase":33
 *         res.name = name
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res             # <<<<<<<<<<<<<<
 *         return res
 *     def __repr__(self):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely((PyObject_SetItem(__pyx_t_4, __pyx_v_name, __pyx_v_res) < 0))) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumBase":34
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res
 *         return res             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "EnumBase":24
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int, metaclass=__Pyx_EnumMeta):
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":35
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_3__repr__ = {"__repr__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 35, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__repr__") < 0)) __PYX_ERR(0, 35, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__repr__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 35, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(__pyx_self, __pyx_v_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "EnumBase":36
 *         return res
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)             # <<<<<<<<<<<<<<
 *     def __str__(self):
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s_d, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "EnumBase":35
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":37
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_5__str__ = {"__str__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 37, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__str__") < 0)) __PYX_ERR(0, 37, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__str__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 37, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(__pyx_self, __pyx_v_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "EnumBase":38
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 *         return "%s.%s" % (self.__class__.__name__, self.name)             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "EnumBase":37
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta = {"__pyx_unpickle___Pyx_EnumMeta", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle___Pyx_EnumMeta") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__2, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype___Pyx_EnumMeta), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_unpickle___Pyx_EnumMeta__set_state(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_unpickle___Pyx_EnumMeta__set_state(struct __pyx_obj___Pyx_EnumMeta *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Node_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Reader;
  new((void*)&(p->source)) Graph::Node::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Node_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Node_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_label(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5label_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_children_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_index_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_children_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_14children_count_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_identity(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_8identity_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Node_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_Node_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"label", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_label, 0, (char *)PyDoc_STR("\n        Inlined for efficiency purposes\n        Proof state\n        Hash a unique id (evar) for the proof state that distinquishes proof states with identical\n        contents but do not point to the same object nonetheless\n        "), 0},
  {(char *)"children_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_children_index, 0, (char *)0, 0},
  {(char *)"children_count", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_children_count, 0, (char *)PyDoc_STR("\n        The children of a node are encoded as a range withing the `edges`-list of the graph.\n        "), 0},
  {(char *)"identity", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_identity, 0, (char *)PyDoc_STR("\n        The identity of a node uniquely determines it. That is, one can consider any to nodes with the same\n        identity to be equal. The notion of equal we use is as follows:\n        1. Graph perspective: Two nodes have the same identity if and only if they are bisimilar.\n           In this notion, bisimilarity does take into consideration the label of the nodes, modulo some\n           equivalence class that is not fully specified here. One aspect of the equivalence class is that\n           for definition nodes their associated global context (accessed through `Definition.previous`) is\n           not taken into account.\n        2. Lambda calculus perspective: Two nodes have the same identity if their corresponding lambda terms\n           are alpha-equivalent. Note that two definitions with the same body are not considered\n           alpha-equivalent.\n\n        The identity of a node is used to perform partial graph-sharing. That is, two nodes with the same\n        identity are merged when the graph is generated. There are three reasons why two nodes with the same\n        semantic identity might have a different physical identity:\n        1. Nodes are only merged when the first node exists in the same graph as the second node, or exists\n           in a dependent graph. Hence, nodes originating from developments that do not depend on each other\n           are never merged. Full graph-sharing would require global analysis on a dataset, which any consumer\n           can optionally do as a post-processing step.\n        2. Two definition nodes with the same name and body have the same identity. But if they occur in\n           different global contexts, these nodes are physically different to ensure the integrity of their\n           global contexts.\n        3. For definitions with an opaque proof, the sub-graph that represents this opaque proof is ignored\n           while calculating the identity. Morally we defend this decision with the concept"" that opaque proofs\n           are invisible in Coq (proof irrelevance) and two constants with different proofs might as well be\n           considered equal. Practically speaking, this is done so that one can compare the identity of a\n           definition in a dataset quickly to a identity of a live definition during Coq interaction without\n           having to produce the entire graph of opaque proofs (which tend to be large).\n\n        Beware that the identity is currently a 64 bit field. For datasets that have graphs of size in the\n        order of billions of nodes there is a non-trivial chance of a collision. (We consider this acceptable\n        for machine learning purposes.)\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Graph_Node_Reader(_DynamicStructReader dyn)\n\n    Fits exactly in 128 bits.\n    A node has a label that optionally contains some additional information, together with a list of\n    outgoing edges (children).\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Node_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Node_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Node_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_Node_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_spec = {
  "pytact.graph_api_capnp_cython.Graph_Node_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Graph_Node_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Graph_Node_Reader(_DynamicStructReader dyn)\n\n    Fits exactly in 128 bits.\n    A node has a label that optionally contains some additional information, together with a list of\n    outgoing edges (children).\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Node_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Node_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Node_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_Node_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader;
  new((void*)&(p->source)) Graph::EdgeTarget::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_label(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5label_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_target(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6target_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"label", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_label, 0, (char *)0, 0},
  {(char *)"target", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_target, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Graph_EdgeTarget_Reader(_DynamicStructReader dyn)\n\n    Fits exactly in 64 bits. Let's keep it that way.\n    The 'pointy' end of an edge, together with the label of that edge.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_spec = {
  "pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Graph_EdgeTarget_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Graph_EdgeTarget_Reader(_DynamicStructReader dyn)\n\n    Fits exactly in 64 bits. Let's keep it that way.\n    The 'pointy' end of an edge, together with the label of that edge.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Reader;
  new((void*)&(p->source)) Graph::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_12Graph_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_12Graph_Reader_nodes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_5nodes_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_12Graph_Reader_has_nodes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_nodes_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_12Graph_Reader_edges(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_5edges_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_12Graph_Reader_has_edges(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_9has_edges_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_12Graph_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_12Graph_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_12Graph_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"nodes", __pyx_getprop_6pytact_22graph_api_capnp_cython_12Graph_Reader_nodes, 0, (char *)0, 0},
  {(char *)"has_nodes", __pyx_getprop_6pytact_22graph_api_capnp_cython_12Graph_Reader_has_nodes, 0, (char *)0, 0},
  {(char *)"edges", __pyx_getprop_6pytact_22graph_api_capnp_cython_12Graph_Reader_edges, 0, (char *)PyDoc_STR("\n        The main memory store of the graph. It acts as a heap similar to the main memory of a C/C++ program.\n        The heap is accessed by indexing the `nodes` list using a `NodeIndex` which returns a `Node`.\n        Every node has a label and a list of children, which is indicated as a range within the `edges` list using\n        `childrenIndex` and `childrenCount`. The targets of the edges can again be found in the `nodes` list of the\n        current file or of a dependency.\n        Note that just like in C/C++ doing pointer arithmetic on the heap is undefined behavior, and you may\n        encounter arbitrary garbage if you do this. In particular, iterating over the heap is discouraged.\n        Instead, you should access the heap through various entry-points that are provided.\n        "), 0},
  {(char *)"has_edges", __pyx_getprop_6pytact_22graph_api_capnp_cython_12Graph_Reader_has_edges, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Graph_Reader(_DynamicStructReader dyn)\n\n    A graph is the main data store (the 'heap') that contains the bulk of the data in the dataset and\n    during communication with Coq. A graph is a collection of labeled nodes with directed, labeled edges\n    between them. A graph may reference nodes from other graphs. For an edge 'A --> B' in the graph it\n    always holds that 'A' is part of the current graph, but 'B' may (or may not) be part of another graph.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader_spec = {
  "pytact.graph_api_capnp_cython.Graph_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Graph_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Graph_Reader(_DynamicStructReader dyn)\n\n    A graph is the main data store (the 'heap') that contains the bulk of the data in the dataset and\n    during communication with Coq. A graph is a collection of labeled nodes with directed, labeled edges\n    between them. A graph may reference nodes from other graphs. For an edge 'A --> B' in the graph it\n    always holds that 'A' is part of the current graph, but 'B' may (or may not) be part of another graph.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_12Graph_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_IntP_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_IntP_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_IntP_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_IntP_Reader;
  new((void*)&(p->source)) IntP::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_IntP_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_IntP_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_IntP_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_IntP_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_11IntP_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_11IntP_Reader_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_5value_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_IntP_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11IntP_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11IntP_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_IntP_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_11IntP_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"value", __pyx_getprop_6pytact_22graph_api_capnp_cython_11IntP_Reader_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_IntP_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("IntP_Reader(_DynamicStructReader dyn)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_IntP_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_IntP_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_IntP_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_IntP_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_IntP_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader_spec = {
  "pytact.graph_api_capnp_cython.IntP_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""IntP_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_IntP_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_IntP_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("IntP_Reader(_DynamicStructReader dyn)"), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_IntP_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_IntP_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_IntP_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_IntP_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_11IntP_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_IntP_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Definition_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Reader;
  new((void*)&(p->source)) Definition::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Definition_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Definition_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_which(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_5which_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_which_raw(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9which_raw_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_4name_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_8has_name_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_previous(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_8previous_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_external_previous(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_17external_previous_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_external_previous(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_external_previous_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_status(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_6status_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_inductive(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9inductive_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_inductive(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_12is_inductive_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_constructor(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_11constructor_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_constructor(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_14is_constructor_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_projection(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_10projection_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_projection(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_13is_projection_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_manual_constant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_15manual_constant_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_manual_constant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_18is_manual_constant_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_tactical_constant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_17tactical_constant_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_tactical_constant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_20is_tactical_constant_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_tactical_constant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_21has_tactical_constant_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_manual_section_constant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_23manual_section_constant_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_manual_section_constant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_26is_manual_section_constant_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_tactical_section_constant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_25tactical_section_constant_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_tactical_section_constant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_28is_tactical_section_constant_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_tactical_section_constant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_29has_tactical_section_constant_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_type_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9type_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_type_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_type_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_term_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_9term_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_term_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_13has_term_text_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Definition_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17Definition_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17Definition_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Definition_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"which", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_which, 0, (char *)0, 0},
  {(char *)"which_raw", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_which_raw, 0, (char *)0, 0},
  {(char *)"name", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_name, 0, (char *)PyDoc_STR("\n        The fully-qualified name of the definition. The name should be unique in a particular global context,\n        but is not unique among different branches of the global in a dataset.\n        "), 0},
  {(char *)"has_name", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_name, 0, (char *)0, 0},
  {(char *)"previous", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_previous, 0, (char *)PyDoc_STR("\n        The previous definition within the global context of the current file.\n        For the first definition this field is set to `len(graph.nodes)`.\n        The contract on this field is that any definition nodes reachable from the forward closure of the definition\n        must also be reachable through the chain of previous fields. An exception to this rule are mutually\n        recursive definitions. Those nodes are placed into the global context in an arbitrary ordering.\n        "), 0},
  {(char *)"external_previous", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_external_previous, 0, (char *)PyDoc_STR("\n        This field provides information about the external global context.\n        At any point in a source file other files 'X' can be loaded through 'Require X'. When this happens, the\n        definitions of X that are reachable through its 'representative' field become available to all subsequent\n        definitions.\n        "), 0},
  {(char *)"has_external_previous", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_external_previous, 0, (char *)0, 0},
  {(char *)"status", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_status, 0, (char *)PyDoc_STR("\n        A definition can have different origins, which are encoded here\n        "), 0},
  {(char *)"inductive", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_inductive, 0, (char *)0, 0},
  {(char *)"is_inductive", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_inductive, 0, (char *)0, 0},
  {(char *)"constructor", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_constructor, 0, (char *)0, 0},
  {(char *)"is_constructor", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_constructor, 0, (char *)0, 0},
  {(char *)"projection", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_projection, 0, (char *)PyDoc_STR("\n        These definitions are part of a mutually recursive cluster. They hold a reference to another definition\n        that acts as the 'representative' of the mutually recursive cluster. The representative is chosen such\n        that all definitions of the cluster are reachable through its `previous` pointer. Additionally, all\n        definitions within the cluster have the same representative, and no definitions that are not part of the\n        cluster are interleaved within the chain of `previous` nodes.\n        "), 0},
  {(char *)"is_projection", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_projection, 0, (char *)0, 0},
  {(char *)"manual_constant", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_manual_constant, 0, (char *)PyDoc_STR("\n        A constant defined by directly inputting a term\n        In the future, we might augment such constants with tactical\n        refinement proofs that build the term iteratively.\n        "), 0},
  {(char *)"is_manual_constant", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_manual_constant, 0, (char *)0, 0},
  {(char *)"tactical_constant", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_tactical_constant, 0, (char *)PyDoc_STR("\n        A constant that was either directly or indirectly generated by a tactical proof.\n        Note that for non-original constants, the proof step sequence may not be very accurate.\n        "), 0},
  {(char *)"is_tactical_constant", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_tactical_constant, 0, (char *)0, 0},
  {(char *)"has_tactical_constant", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_tactical_constant, 0, (char *)0, 0},
  {(char *)"manual_section_constant", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_manual_section_constant, 0, (char *)PyDoc_STR("\n        A section variable or local section definition.\n        "), 0},
  {(char *)"is_manual_section_constant", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_manual_section_constant, 0, (char *)0, 0},
  {(char *)"tactical_section_constant", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_tactical_section_constant, 0, (char *)PyDoc_STR("\n        Local section definitions can also be defined using a tactical proof.\n        "), 0},
  {(char *)"is_tactical_section_constant", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_is_tactical_section_constant, 0, (char *)0, 0},
  {(char *)"has_tactical_section_constant", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_tactical_section_constant, 0, (char *)0, 0},
  {(char *)"type_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_type_text, 0, (char *)PyDoc_STR("\n        A textual representation of the type of this definition.\n        "), 0},
  {(char *)"has_type_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_type_text, 0, (char *)0, 0},
  {(char *)"term_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_term_text, 0, (char *)PyDoc_STR("\n        A textual representation of the body of this definition.\n        For inductives, constructors, projections, section variables and axioms the string is empty.\n        "), 0},
  {(char *)"has_term_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_17Definition_Reader_has_term_text, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Definition_Reader(_DynamicStructReader dyn)\n\n    A definition of the CIC, which is either an constant, inductive, constructor, projection or section\n    variable. Constants and section variables can have tactical proofs associated to them.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Definition_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Definition_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Definition_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Definition_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader_spec = {
  "pytact.graph_api_capnp_cython.Definition_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Definition_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Definition_Reader(_DynamicStructReader dyn)\n\n    A definition of the CIC, which is either an constant, inductive, constructor, projection or section\n    variable. Constants and section variables can have tactical proofs associated to them.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Definition_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Definition_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Definition_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Definition_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_17Definition_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_FloatP_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_FloatP_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_FloatP_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_FloatP_Reader;
  new((void*)&(p->source)) FloatP::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_FloatP_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_FloatP_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_FloatP_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_FloatP_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13FloatP_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13FloatP_Reader_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5value_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_FloatP_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_13FloatP_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_13FloatP_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_FloatP_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_13FloatP_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"value", __pyx_getprop_6pytact_22graph_api_capnp_cython_13FloatP_Reader_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_FloatP_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("FloatP_Reader(_DynamicStructReader dyn)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_FloatP_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_FloatP_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_FloatP_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_FloatP_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_FloatP_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader_spec = {
  "pytact.graph_api_capnp_cython.FloatP_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""FloatP_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_FloatP_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_FloatP_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("FloatP_Reader(_DynamicStructReader dyn)"), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_FloatP_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_FloatP_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_FloatP_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_FloatP_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_13FloatP_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_FloatP_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader;
  new((void*)&(p->source)) Graph::Node::Label::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_which(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5which_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_which_raw(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9which_raw_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_proof_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11proof_state_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_proof_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_proof_state_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_has_proof_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_15has_proof_state_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_context_def(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11context_def_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_context_def(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_context_def_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_context_assum(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13context_assum_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_context_assum(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_16is_context_assum_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_definition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10definition_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_definition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_definition_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_has_definition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14has_definition_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_const_empty(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11const_empty_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_const_empty(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_const_empty_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_sort_s_prop(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11sort_s_prop_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_sort_s_prop(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_sort_s_prop_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_sort_prop(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_prop_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_sort_prop(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_prop_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_sort_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8sort_set_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_sort_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11is_sort_set_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_sort_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9sort_type_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_sort_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_sort_type_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_rel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3rel_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_rel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_rel_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_evar(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4evar_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_evar(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_evar_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_evar_subst(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10evar_subst_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_evar_subst(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_evar_subst_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_cast(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4cast_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_cast(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_cast_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_prod(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4prod_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_prod(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_prod_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_lambda_(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7lambda__1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_lambda_(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_lambda__1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_let_in(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6let_in_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_let_in(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_let_in_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_app(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3app_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_app(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_app_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_case(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4case_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_case(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7is_case_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_case_branch(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_11case_branch_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_case_branch(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_14is_case_branch_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_fix(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3fix_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_fix(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_fix_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_fix_fun(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7fix_fun_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_fix_fun(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10is_fix_fun_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_co_fix(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6co_fix_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_co_fix(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9is_co_fix_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_co_fix_fun(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_10co_fix_fun_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_co_fix_fun(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13is_co_fix_fun_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_int(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3int_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_int(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6is_int_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_has_int(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7has_int_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_float(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5float_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_float(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_8is_float_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_has_float(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9has_float_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_primitive(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_9primitive_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_primitive(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_12is_primitive_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_has_primitive(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_13has_primitive_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"which", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_which, 0, (char *)0, 0},
  {(char *)"which_raw", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_which_raw, 0, (char *)0, 0},
  {(char *)"proof_state", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_proof_state, 0, (char *)0, 0},
  {(char *)"is_proof_state", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_proof_state, 0, (char *)0, 0},
  {(char *)"has_proof_state", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_has_proof_state, 0, (char *)0, 0},
  {(char *)"context_def", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_context_def, 0, (char *)0, 0},
  {(char *)"is_context_def", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_context_def, 0, (char *)0, 0},
  {(char *)"context_assum", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_context_assum, 0, (char *)0, 0},
  {(char *)"is_context_assum", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_context_assum, 0, (char *)0, 0},
  {(char *)"definition", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_definition, 0, (char *)0, 0},
  {(char *)"is_definition", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_definition, 0, (char *)0, 0},
  {(char *)"has_definition", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_has_definition, 0, (char *)0, 0},
  {(char *)"const_empty", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_const_empty, 0, (char *)0, 0},
  {(char *)"is_const_empty", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_const_empty, 0, (char *)0, 0},
  {(char *)"sort_s_prop", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_sort_s_prop, 0, (char *)0, 0},
  {(char *)"is_sort_s_prop", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_sort_s_prop, 0, (char *)0, 0},
  {(char *)"sort_prop", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_sort_prop, 0, (char *)0, 0},
  {(char *)"is_sort_prop", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_sort_prop, 0, (char *)0, 0},
  {(char *)"sort_set", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_sort_set, 0, (char *)0, 0},
  {(char *)"is_sort_set", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_sort_set, 0, (char *)0, 0},
  {(char *)"sort_type", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_sort_type, 0, (char *)PyDoc_STR("\n        Collapsed universe\n        "), 0},
  {(char *)"is_sort_type", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_sort_type, 0, (char *)0, 0},
  {(char *)"rel", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_rel, 0, (char *)0, 0},
  {(char *)"is_rel", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_rel, 0, (char *)0, 0},
  {(char *)"evar", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_evar, 0, (char *)0, 0},
  {(char *)"is_evar", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_evar, 0, (char *)0, 0},
  {(char *)"evar_subst", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_evar_subst, 0, (char *)0, 0},
  {(char *)"is_evar_subst", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_evar_subst, 0, (char *)0, 0},
  {(char *)"cast", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_cast, 0, (char *)0, 0},
  {(char *)"is_cast", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_cast, 0, (char *)0, 0},
  {(char *)"prod", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_prod, 0, (char *)0, 0},
  {(char *)"is_prod", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_prod, 0, (char *)0, 0},
  {(char *)"lambda_", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_lambda_, 0, (char *)0, 0},
  {(char *)"is_lambda_", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_lambda_, 0, (char *)0, 0},
  {(char *)"let_in", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_let_in, 0, (char *)0, 0},
  {(char *)"is_let_in", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_let_in, 0, (char *)0, 0},
  {(char *)"app", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_app, 0, (char *)0, 0},
  {(char *)"is_app", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_app, 0, (char *)0, 0},
  {(char *)"case", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_case, 0, (char *)0, 0},
  {(char *)"is_case", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_case, 0, (char *)0, 0},
  {(char *)"case_branch", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_case_branch, 0, (char *)0, 0},
  {(char *)"is_case_branch", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_case_branch, 0, (char *)0, 0},
  {(char *)"fix", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_fix, 0, (char *)0, 0},
  {(char *)"is_fix", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_fix, 0, (char *)0, 0},
  {(char *)"fix_fun", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_fix_fun, 0, (char *)0, 0},
  {(char *)"is_fix_fun", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_fix_fun, 0, (char *)0, 0},
  {(char *)"co_fix", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_co_fix, 0, (char *)0, 0},
  {(char *)"is_co_fix", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_co_fix, 0, (char *)0, 0},
  {(char *)"co_fix_fun", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_co_fix_fun, 0, (char *)0, 0},
  {(char *)"is_co_fix_fun", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_co_fix_fun, 0, (char *)0, 0},
  {(char *)"int", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_int, 0, (char *)0, 0},
  {(char *)"is_int", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_int, 0, (char *)0, 0},
  {(char *)"has_int", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_has_int, 0, (char *)0, 0},
  {(char *)"float", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_float, 0, (char *)0, 0},
  {(char *)"is_float", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_float, 0, (char *)0, 0},
  {(char *)"has_float", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_has_float, 0, (char *)0, 0},
  {(char *)"primitive", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_primitive, 0, (char *)0, 0},
  {(char *)"is_primitive", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_is_primitive, 0, (char *)0, 0},
  {(char *)"has_primitive", __pyx_getprop_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_has_primitive, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Graph_Node_Label_Reader(_DynamicStructReader dyn)\n\n    Inlined for efficiency purposes\n    Proof state\n    Hash a unique id (evar) for the proof state that distinquishes proof states with identical\n    contents but do not point to the same object nonetheless\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader_spec = {
  "pytact.graph_api_capnp_cython.Graph_Node_Label_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Graph_Node_Label_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Graph_Node_Label_Reader(_DynamicStructReader dyn)\n\n    Inlined for efficiency purposes\n    Proof state\n    Hash a unique id (evar) for the proof state that distinquishes proof states with identical\n    contents but do not point to the same object nonetheless\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader;
  new((void*)&(p->source)) Definition::Status::Substituted::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_dep_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_9dep_index_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_node_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_10node_index_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"dep_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_dep_index, 0, (char *)PyDoc_STR("\n        Indicates to which graph a node belongs. How this should be resolved is not specified here.\n        However, a value of zero always points to the 'current' graph.\n        "), 0},
  {(char *)"node_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_node_index, 0, (char *)PyDoc_STR("\n        The index into `Graph.nodes` where this node can be found.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Definition_Status_Substituted_Reader(_DynamicStructReader dyn)\n\n    A definition that was obtained by performing some sort of module functor substitution.\n    We reference the original definition.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader_spec = {
  "pytact.graph_api_capnp_cython.Definition_Status_Substituted_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Definition_Status_Substituted_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Definition_Status_Substituted_Reader(_DynamicStructReader dyn)\n\n    A definition that was obtained by performing some sort of module functor substitution.\n    We reference the original definition.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Definition_Status_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Definition_Status_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Status_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Status_Reader;
  new((void*)&(p->source)) Definition::Status::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Status_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Status_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Definition_Status_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Definition_Status_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_which(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5which_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_which_raw(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_9which_raw_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_original(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_8original_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_is_original(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11is_original_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_discharged(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_10discharged_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_is_discharged(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_13is_discharged_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_substituted(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_11substituted_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_is_substituted(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_14is_substituted_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Definition_Status_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Definition_Status_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"which", __pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_which, 0, (char *)0, 0},
  {(char *)"which_raw", __pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_which_raw, 0, (char *)0, 0},
  {(char *)"original", __pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_original, 0, (char *)PyDoc_STR("\n        An object as originally inputted by the user.\n        "), 0},
  {(char *)"is_original", __pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_is_original, 0, (char *)0, 0},
  {(char *)"discharged", __pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_discharged, 0, (char *)PyDoc_STR("\n        a definition that was originally defined in a section and has now had the section\n        variables discharged into it. The node index references the original definition.\n        "), 0},
  {(char *)"is_discharged", __pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_is_discharged, 0, (char *)0, 0},
  {(char *)"substituted", __pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_substituted, 0, (char *)PyDoc_STR("\n        A definition that was obtained by performing some sort of module functor substitution.\n        We reference the original definition.\n        "), 0},
  {(char *)"is_substituted", __pyx_getprop_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_is_substituted, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Status_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Definition_Status_Reader(_DynamicStructReader dyn)\n\n    A definition can have different origins, which are encoded here\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Definition_Status_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Definition_Status_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Definition_Status_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Definition_Status_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Status_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader_spec = {
  "pytact.graph_api_capnp_cython.Definition_Status_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Definition_Status_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Definition_Status_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Definition_Status_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Definition_Status_Reader(_DynamicStructReader dyn)\n\n    A definition can have different origins, which are encoded here\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Definition_Status_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Definition_Status_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Definition_Status_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Definition_Status_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Definition_Status_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofStep_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Reader;
  new((void*)&(p->source)) ProofStep::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofStep_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofStep_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_tactic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6tactic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_outcomes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_8outcomes_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_has_outcomes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_12has_outcomes_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_ProofStep_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_ProofStep_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"tactic", __pyx_getprop_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_tactic, 0, (char *)PyDoc_STR("\n        The tactic that generated the proof step. Note that the arguments of the tactic can be found in the\n        individual outcomes, because they may be different for each outcome.\n        "), 0},
  {(char *)"outcomes", __pyx_getprop_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_outcomes, 0, (char *)PyDoc_STR("\n        A list of transformations of proof states to other proof states, as executed by the tactic of the proof step\n        "), 0},
  {(char *)"has_outcomes", __pyx_getprop_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_has_outcomes, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("ProofStep_Reader(_DynamicStructReader dyn)\n\n    A proof step is the execution of a single tactic on one or more proof states, producing a list of outcomes.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofStep_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofStep_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_ProofStep_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_ProofStep_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_spec = {
  "pytact.graph_api_capnp_cython.ProofStep_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""ProofStep_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("ProofStep_Reader(_DynamicStructReader dyn)\n\n    A proof step is the execution of a single tactic on one or more proof states, producing a list of outcomes.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofStep_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofStep_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_ProofStep_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_ProofStep_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Tactic_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Tactic_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Tactic_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Tactic_Reader;
  new((void*)&(p->source)) Tactic::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Tactic_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Tactic_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Tactic_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Tactic_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_ident(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5ident_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_has_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_8has_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_base_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_9base_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_has_base_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_13has_base_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_interm_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_11interm_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_has_interm_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_15has_interm_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_exact(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5exact_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_text_non_anonymous(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_18text_non_anonymous_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_has_text_non_anonymous(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_22has_text_non_anonymous_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Tactic_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_13Tactic_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_13Tactic_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Tactic_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"ident", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_ident, 0, (char *)PyDoc_STR("\n        A hash representing the identity of a tactic without it's arguments. Due to the complexity of the syntax\n        trees of Coq's tactics, we do not currently encode the syntax tree. Instead, this hash is a representative\n        of the syntax tree of the tactic with all of it's arguments removed.\n        "), 0},
  {(char *)"text", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_text, 0, (char *)PyDoc_STR("\n        The full text of the tactic including the full arguments. This does not currently correspond to\n        (ident, arguments) because in this dataset arguments do not include full terms, but only references to\n        definitions and local context elements. Tactics are postprocessed to remove explicit naming of new\n        hypotheses as much as possible, and instead asking Coq to invent a name.\n        "), 0},
  {(char *)"has_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_has_text, 0, (char *)0, 0},
  {(char *)"base_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_base_text, 0, (char *)PyDoc_STR("\n        A textual representation of the base tactic without arguments. It tries to roughly correspond to `ident`.\n        Note, however, that this is both an under-approximation and an over-approximation. The reason is that tactic\n        printing is not 100% isomorphic to Coq's internal AST of tactics. Sometimes, different tactics get mapped to\n        the same text. Conversely, the same tactic may be mapped to different texts when identifiers are printed\n        using different partially-qualified names.\n        "), 0},
  {(char *)"has_base_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_has_base_text, 0, (char *)0, 0},
  {(char *)"interm_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_interm_text, 0, (char *)PyDoc_STR("\n        A textual representation that tries to come as close as possible to (ident, arguments).\n        It comes with the same caveats as `baseText`.\n        "), 0},
  {(char *)"has_interm_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_has_interm_text, 0, (char *)0, 0},
  {(char *)"exact", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_exact, 0, (char *)PyDoc_STR("\n        Indicates whether or not `ident` + `arguments` is faithfully reversible into the original \"strictified\" tactic.\n        Note that this does not necessarily mean that it represents exactly the tactic that was inputted by the user.\n        All tactics are modified to be 'strict' (meaning that tactics that have delayed variables in them break).\n        This flag measures the faithfulness of the representation w.r.t. the strict version of the tactic, not the\n        original tactic inputted by the user.\n        "), 0},
  {(char *)"text_non_anonymous", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_text_non_anonymous, 0, (char *)PyDoc_STR("\n        Same as `text` except that the tactic is not postprocessed to remove explicit naming of new hypotheses.\n        "), 0},
  {(char *)"has_text_non_anonymous", __pyx_getprop_6pytact_22graph_api_capnp_cython_13Tactic_Reader_has_text_non_anonymous, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Tactic_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Tactic_Reader(_DynamicStructReader dyn)\n\n    A concrete tactic with it's parameters determined. Somewhat strangely, this struct does not actually include\n    these parameters. They can instead be found in `Outcome.tacticArguments`. The reason for this is that one\n    tactic can run on multiple proof states at the same time and for all of those proof states, the arguments\n    may be resolved differently.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Tactic_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Tactic_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Tactic_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Tactic_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Tactic_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader_spec = {
  "pytact.graph_api_capnp_cython.Tactic_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Tactic_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Tactic_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Tactic_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Tactic_Reader(_DynamicStructReader dyn)\n\n    A concrete tactic with it's parameters determined. Somewhat strangely, this struct does not actually include\n    these parameters. They can instead be found in `Outcome.tacticArguments`. The reason for this is that one\n    tactic can run on multiple proof states at the same time and for all of those proof states, the arguments\n    may be resolved differently.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Tactic_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Tactic_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Tactic_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Tactic_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_13Tactic_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Tactic_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader;
  new((void*)&(p->source)) ProofStep::Tactic::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_which(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5which_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_which_raw(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9which_raw_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_unknown(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7unknown_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_is_unknown(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_10is_unknown_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_known(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5known_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_is_known(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_8is_known_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_has_known(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_9has_known_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"which", __pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_which, 0, (char *)0, 0},
  {(char *)"which_raw", __pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_which_raw, 0, (char *)0, 0},
  {(char *)"unknown", __pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_unknown, 0, (char *)PyDoc_STR("\n        Sometimes a tactic cannot or should not be recorded. In those cases, it is marked as 'unknown'.\n        This currently happens with tactics that are run as a result of the `Proof with tac` construct and it\n        happens for tactics that are known to be unsafe like `change_no_check`, `fix`, `cofix` and more.\n        "), 0},
  {(char *)"is_unknown", __pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_is_unknown, 0, (char *)0, 0},
  {(char *)"known", __pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_known, 0, (char *)PyDoc_STR("\n        The tactic\n        "), 0},
  {(char *)"is_known", __pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_is_known, 0, (char *)0, 0},
  {(char *)"has_known", __pyx_getprop_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_has_known, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("ProofStep_Tactic_Reader(_DynamicStructReader dyn)\n\n    The tactic that generated the proof step. Note that the arguments of the tactic can be found in the\n    individual outcomes, because they may be different for each outcome.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader_spec = {
  "pytact.graph_api_capnp_cython.ProofStep_Tactic_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""ProofStep_Tactic_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("ProofStep_Tactic_Reader(_DynamicStructReader dyn)\n\n    The tactic that generated the proof step. Note that the arguments of the tactic can be found in the\n    individual outcomes, because they may be different for each outcome.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Outcome_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Reader;
  new((void*)&(p->source)) Outcome::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Outcome_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Outcome_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_before(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6before_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_has_before(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_10has_before_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_after(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5after_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_has_after(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9has_after_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_term(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4term_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_term_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_9term_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_has_term_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_13has_term_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_tactic_arguments(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_16tactic_arguments_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_has_tactic_arguments(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_20has_tactic_arguments_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Outcome_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_14Outcome_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_14Outcome_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Outcome_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"before", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_before, 0, (char *)PyDoc_STR("\n        The proof state before the tactic execution.\n        "), 0},
  {(char *)"has_before", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_has_before, 0, (char *)0, 0},
  {(char *)"after", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_after, 0, (char *)PyDoc_STR("\n        The new proof states that were generated by the tactic.\n        "), 0},
  {(char *)"has_after", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_has_after, 0, (char *)0, 0},
  {(char *)"term", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_term, 0, (char *)PyDoc_STR("\n        The proof term that witnesses the transition from the before state to the after states. It contains a hole\n        (an `evar` node) for each of the after states. It may also refer to elements of the local context of the\n        before state.\n        "), 0},
  {(char *)"term_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_term_text, 0, (char *)PyDoc_STR("\n        A textual representation of the proof term.\n        "), 0},
  {(char *)"has_term_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_has_term_text, 0, (char *)0, 0},
  {(char *)"tactic_arguments", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_tactic_arguments, 0, (char *)PyDoc_STR("\n        The arguments of the tactic that produced this outcome. Note that these arguments belong to the tactic in\n        `ProofStep.tactic`.\n        "), 0},
  {(char *)"has_tactic_arguments", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Outcome_Reader_has_tactic_arguments, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Outcome_Reader(_DynamicStructReader dyn)\n\n    An outcome is the result of running a tactic on a proof state. A tactic may run on multiple proof states.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Outcome_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Outcome_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Outcome_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Outcome_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_spec = {
  "pytact.graph_api_capnp_cython.Outcome_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Outcome_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Outcome_Reader(_DynamicStructReader dyn)\n\n    An outcome is the result of running a tactic on a proof state. A tactic may run on multiple proof states.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Outcome_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Outcome_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Outcome_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Outcome_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_14Outcome_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader;
  new((void*)&(p->source)) Outcome::Term::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_dep_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_9dep_index_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_node_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_10node_index_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"dep_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_dep_index, 0, (char *)PyDoc_STR("\n        Indicates to which graph a node belongs. How this should be resolved is not specified here.\n        However, a value of zero always points to the 'current' graph.\n        "), 0},
  {(char *)"node_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_node_index, 0, (char *)PyDoc_STR("\n        The index into `Graph.nodes` where this node can be found.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Outcome_Term_Reader(_DynamicStructReader dyn)\n\n    The proof term that witnesses the transition from the before state to the after states. It contains a hole\n    (an `evar` node) for each of the after states. It may also refer to elements of the local context of the\n    before state.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader_spec = {
  "pytact.graph_api_capnp_cython.Outcome_Term_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Outcome_Term_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Outcome_Term_Reader(_DynamicStructReader dyn)\n\n    The proof term that witnesses the transition from the before state to the after states. It contains a hole\n    (an `evar` node) for each of the after states. It may also refer to elements of the local context of the\n    before state.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofState_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Reader;
  new((void*)&(p->source)) ProofState::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofState_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofState_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_root(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4root_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_context(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7context_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_has_context(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_11has_context_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_context_names(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_13context_names_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_has_context_names(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_17has_context_names_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_context_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_12context_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_has_context_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_16has_context_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_conclusion_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_15conclusion_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_has_conclusion_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_19has_conclusion_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_has_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_8has_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_2id_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_ProofState_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17ProofState_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_17ProofState_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_ProofState_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"root", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_root, 0, (char *)PyDoc_STR("\n        The entry-point of the proof state, all nodes that are 'part of' the proof state are reachable from here.\n        "), 0},
  {(char *)"context", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_context, 0, (char *)PyDoc_STR("\n        The local context of the proof state. These nodes label's are either `contextAssum` or `contextDef`. Note that\n        these nodes are also reachable from the root of the proof state.\n        "), 0},
  {(char *)"has_context", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_has_context, 0, (char *)0, 0},
  {(char *)"context_names", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_context_names, 0, (char *)PyDoc_STR("\n        The names of the local context nodes of the proof state, as they originally appeared in the proof.\n        These names should be used for debugging and viewing purposes only, because hypothesis-generating tactics have\n        been modified to use auto-generated names. Hence, tactics should not be concerned about the names of\n        the context.\n        "), 0},
  {(char *)"has_context_names", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_has_context_names, 0, (char *)0, 0},
  {(char *)"context_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_context_text, 0, (char *)PyDoc_STR("\n        A textual representation of the type/definition of context nodes\n        "), 0},
  {(char *)"has_context_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_has_context_text, 0, (char *)0, 0},
  {(char *)"conclusion_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_conclusion_text, 0, (char *)PyDoc_STR("\n        A textual representation of the conclusion of the proof state.\n        "), 0},
  {(char *)"has_conclusion_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_has_conclusion_text, 0, (char *)0, 0},
  {(char *)"text", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_text, 0, (char *)PyDoc_STR("\n        A textual representation of the proof state.\n        "), 0},
  {(char *)"has_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_has_text, 0, (char *)0, 0},
  {(char *)"id", __pyx_getprop_6pytact_22graph_api_capnp_cython_17ProofState_Reader_id, 0, (char *)PyDoc_STR("\n        A unique identifier of the proof state. Any two proof states in a tactical proof that have an equal id\n        can morally be regarded to be 'the same' proof state.\n        IMPORTANT: Two proof states with the same id may still have different contents. This is because proof states\n                   can contain existential variables (represented by the `evar` node) that can be filled as a\n                   side-effect by a tactic running on another proof state.\n        TODO: This is currently duplicated with the id attached to the `ProofState` node that `root` points to.\n              Eventually, this id should be resolved to a NodeIndex that represents the location of this proof state\n              in the final proof term that was generated.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("ProofState_Reader(_DynamicStructReader dyn)\n\n    A proof state represents a particular point in the tactical proof of a constant.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofState_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofState_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_ProofState_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_ProofState_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_spec = {
  "pytact.graph_api_capnp_cython.ProofState_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""ProofState_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("ProofState_Reader(_DynamicStructReader dyn)\n\n    A proof state represents a particular point in the tactical proof of a constant.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofState_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofState_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_ProofState_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_ProofState_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_17ProofState_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Argument_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Reader;
  new((void*)&(p->source)) Argument::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Argument_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Argument_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_which(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_5which_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_which_raw(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_9which_raw_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_unresolvable(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_12unresolvable_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_is_unresolvable(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_15is_unresolvable_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_term(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_4term_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_is_term(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_7is_term_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Argument_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_15Argument_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_15Argument_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Argument_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"which", __pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_which, 0, (char *)0, 0},
  {(char *)"which_raw", __pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_which_raw, 0, (char *)0, 0},
  {(char *)"unresolvable", __pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_unresolvable, 0, (char *)PyDoc_STR("\n        An argument that is currently unresolvable due to limitations of the extraction process.\n        "), 0},
  {(char *)"is_unresolvable", __pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_is_unresolvable, 0, (char *)0, 0},
  {(char *)"term", __pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_term, 0, (char *)PyDoc_STR("\n        The root of a graph representing an argument that is a term in the calculus of constructions.\n        "), 0},
  {(char *)"is_term", __pyx_getprop_6pytact_22graph_api_capnp_cython_15Argument_Reader_is_term, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Argument_Reader(_DynamicStructReader dyn)\n\n    A concrete argument of a tactic.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Argument_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Argument_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Argument_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Argument_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_spec = {
  "pytact.graph_api_capnp_cython.Argument_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Argument_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Argument_Reader(_DynamicStructReader dyn)\n\n    A concrete argument of a tactic.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Argument_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Argument_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Argument_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Argument_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_15Argument_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader;
  new((void*)&(p->source)) ProofState::Root::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_dep_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_9dep_index_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_node_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_10node_index_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"dep_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_dep_index, 0, (char *)PyDoc_STR("\n        Indicates to which graph a node belongs. How this should be resolved is not specified here.\n        However, a value of zero always points to the 'current' graph.\n        "), 0},
  {(char *)"node_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_node_index, 0, (char *)PyDoc_STR("\n        The index into `Graph.nodes` where this node can be found.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("ProofState_Root_Reader(_DynamicStructReader dyn)\n\n    The entry-point of the proof state, all nodes that are 'part of' the proof state are reachable from here.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader_spec = {
  "pytact.graph_api_capnp_cython.ProofState_Root_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""ProofState_Root_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("ProofState_Root_Reader(_DynamicStructReader dyn)\n\n    The entry-point of the proof state, all nodes that are 'part of' the proof state are reachable from here.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Node_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Node_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Node_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Node_Reader;
  new((void*)&(p->source)) Node::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Node_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Node_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Node_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Node_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_11Node_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_11Node_Reader_dep_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_9dep_index_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_11Node_Reader_node_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_10node_index_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Node_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Node_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Node_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Node_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_11Node_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"dep_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_11Node_Reader_dep_index, 0, (char *)PyDoc_STR("\n        Indicates to which graph a node belongs. How this should be resolved is not specified here.\n        However, a value of zero always points to the 'current' graph.\n        "), 0},
  {(char *)"node_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_11Node_Reader_node_index, 0, (char *)PyDoc_STR("\n        The index into `Graph.nodes` where this node can be found.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Node_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Node_Reader(_DynamicStructReader dyn)\n\n    A node in the CIC graph\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Node_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Node_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Node_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Node_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Node_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_spec = {
  "pytact.graph_api_capnp_cython.Node_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Node_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Node_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Node_Reader(_DynamicStructReader dyn)\n\n    A node in the CIC graph\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Node_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Node_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Node_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Node_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Node_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Node_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Term_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Argument_Term_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Term_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Term_Reader;
  new((void*)&(p->source)) Argument::Term::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Term_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Term_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Argument_Term_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Argument_Term_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_dep_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_9dep_index_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_node_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_10node_index_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Argument_Term_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Argument_Term_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"dep_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_dep_index, 0, (char *)PyDoc_STR("\n        Indicates to which graph a node belongs. How this should be resolved is not specified here.\n        However, a value of zero always points to the 'current' graph.\n        "), 0},
  {(char *)"node_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_node_index, 0, (char *)PyDoc_STR("\n        The index into `Graph.nodes` where this node can be found.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Term_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Argument_Term_Reader(_DynamicStructReader dyn)\n\n    The root of a graph representing an argument that is a term in the calculus of constructions.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Argument_Term_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Argument_Term_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Argument_Term_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Argument_Term_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Term_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader_spec = {
  "pytact.graph_api_capnp_cython.Argument_Term_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Argument_Term_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Term_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Term_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Argument_Term_Reader(_DynamicStructReader dyn)\n\n    The root of a graph representing an argument that is a term in the calculus of constructions.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Argument_Term_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Argument_Term_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Argument_Term_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Argument_Term_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Term_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader;
  new((void*)&(p->source)) Graph::EdgeTarget::Target::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_dep_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_9dep_index_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_node_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_10node_index_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"dep_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_dep_index, 0, (char *)PyDoc_STR("\n        Indicates to which graph a node belongs. How this should be resolved is not specified here.\n        However, a value of zero always points to the 'current' graph.\n        "), 0},
  {(char *)"node_index", __pyx_getprop_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_node_index, 0, (char *)PyDoc_STR("\n        The index into `Graph.nodes` where this node can be found.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Graph_EdgeTarget_Target_Reader(_DynamicStructReader dyn)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader_spec = {
  "pytact.graph_api_capnp_cython.Graph_EdgeTarget_Target_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Graph_EdgeTarget_Target_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Graph_EdgeTarget_Target_Reader(_DynamicStructReader dyn)"), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader;
  new((void*)&(p->source)) AbstractTactic::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_ident(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5ident_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_parameters(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_10parameters_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"ident", __pyx_getprop_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_ident, 0, (char *)PyDoc_STR("\n        An abstract tactic is referenced to using a identifier (hash).\n        "), 0},
  {(char *)"parameters", __pyx_getprop_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_parameters, 0, (char *)PyDoc_STR("\n        Every tactic has a constant number of parameters that need to be filled in.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("AbstractTactic_Reader(_DynamicStructReader dyn)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_spec = {
  "pytact.graph_api_capnp_cython.AbstractTactic_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""AbstractTactic_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("AbstractTactic_Reader(_DynamicStructReader dyn)"), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_DataVersion_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_DataVersion_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_DataVersion_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_DataVersion_Reader;
  new((void*)&(p->source)) DataVersion::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_DataVersion_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_DataVersion_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_DataVersion_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_DataVersion_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_major(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5major_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_DataVersion_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_DataVersion_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"major", __pyx_getprop_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_major, 0, (char *)PyDoc_STR("\n        Currently we only have a major version. Any change to the graph format, Cap'n Proto schema or\n        communication protocol is considered a breaking change and will increment the major version number.\n        Note that currently we don't distinquish backwards-compatible Cap'n Proto schema changes from\n        non-backwards-compatible changes. The schema might change in a non-compatible way without notice.\n        In the future, a minor version id might be introduced.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_DataVersion_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("DataVersion_Reader(_DynamicStructReader dyn)\n\n    Version info of a dataset.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_DataVersion_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_DataVersion_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_DataVersion_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_DataVersion_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_DataVersion_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader_spec = {
  "pytact.graph_api_capnp_cython.DataVersion_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""DataVersion_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_DataVersion_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_DataVersion_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("DataVersion_Reader(_DynamicStructReader dyn)\n\n    Version info of a dataset.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_DataVersion_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_DataVersion_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_DataVersion_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_DataVersion_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_DataVersion_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Dataset_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Dataset_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Dataset_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Dataset_Reader;
  new((void*)&(p->source)) Dataset::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Dataset_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Dataset_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Dataset_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Dataset_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_dependencies(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12dependencies_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_has_dependencies(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_dependencies_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_graph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5graph_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_has_graph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_9has_graph_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_representative(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_14representative_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_definitions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11definitions_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_has_definitions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_definitions_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_module_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_11module_name_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_has_module_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_15has_module_name_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_data_version(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_12data_version_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_has_data_version(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_16has_data_version_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Dataset_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_14Dataset_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_14Dataset_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Dataset_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"dependencies", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_dependencies, 0, (char *)PyDoc_STR("\n        The graph contained in a file may reference nodes from the graph of other files. This field maps\n        a `DepIndex` into the a file that contains that particular node.\n        The first file in this list is always the current file. It is guaranteed that no cycles exist in\n        the dependency relation between files induced by this field (except for the self-reference of the file).\n        "), 0},
  {(char *)"has_dependencies", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_has_dependencies, 0, (char *)0, 0},
  {(char *)"graph", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_graph, 0, (char *)0, 0},
  {(char *)"has_graph", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_has_graph, 0, (char *)0, 0},
  {(char *)"representative", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_representative, 0, (char *)PyDoc_STR("\n        The entry point of the global context of definitions that are available when this file is 'Required' by\n        another file. The full global context can be obtained by following the `previous` node of definitions.\n        If the compilation unit does not contain any 'super'-global definitions this is set to `len(graph.nodes)`\n        "), 0},
  {(char *)"definitions", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_definitions, 0, (char *)PyDoc_STR("\n        All of the definitions present in the graph.\n        Note that some of these nodes may not be part of the 'super-global' context that is reachable using the\n        `representative` field as an entry point. The reason is that the global context is a forest (list of tree's)\n        and the 'super-global' context is only the main spine of this forest.\n        "), 0},
  {(char *)"has_definitions", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_has_definitions, 0, (char *)0, 0},
  {(char *)"module_name", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_module_name, 0, (char *)PyDoc_STR("\n        The name of the module defined by this file.\n        "), 0},
  {(char *)"has_module_name", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_has_module_name, 0, (char *)0, 0},
  {(char *)"data_version", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_data_version, 0, (char *)PyDoc_STR("\n        The version of the data stored in this dataset.\n        "), 0},
  {(char *)"has_data_version", __pyx_getprop_6pytact_22graph_api_capnp_cython_14Dataset_Reader_has_data_version, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Dataset_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Dataset_Reader(_DynamicStructReader dyn)\n\n    Every file in the dataset contains a single message of type `Dataset`. Every file corresponds to\n    a Coq source file, and contains a representation of all definitions that have existed at any point\n    throughout the compilation of the source file.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Dataset_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Dataset_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Dataset_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Dataset_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Dataset_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader_spec = {
  "pytact.graph_api_capnp_cython.Dataset_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Dataset_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Dataset_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Dataset_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Dataset_Reader(_DynamicStructReader dyn)\n\n    Every file in the dataset contains a single message of type `Dataset`. Every file corresponds to\n    a Coq source file, and contains a representation of all definitions that have existed at any point\n    throughout the compilation of the source file.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Dataset_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Dataset_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Dataset_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Dataset_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_14Dataset_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Dataset_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Exception_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_Exception_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Exception_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Exception_Reader;
  new((void*)&(p->source)) Exception::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Exception_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Exception_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Exception_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Exception_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_which(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_5which_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_which_raw(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_9which_raw_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_no_such_tactic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_14no_such_tactic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_is_no_such_tactic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_17is_no_such_tactic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_mismatched_arguments(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_20mismatched_arguments_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_is_mismatched_arguments(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_23is_mismatched_arguments_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_parse_error(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_11parse_error_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_is_parse_error(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_14is_parse_error_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_illegal_argument(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_16illegal_argument_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_is_illegal_argument(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_19is_illegal_argument_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Exception_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16Exception_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16Exception_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_Exception_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"which", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_which, 0, (char *)0, 0},
  {(char *)"which_raw", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_which_raw, 0, (char *)0, 0},
  {(char *)"no_such_tactic", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_no_such_tactic, 0, (char *)0, 0},
  {(char *)"is_no_such_tactic", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_is_no_such_tactic, 0, (char *)0, 0},
  {(char *)"mismatched_arguments", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_mismatched_arguments, 0, (char *)0, 0},
  {(char *)"is_mismatched_arguments", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_is_mismatched_arguments, 0, (char *)0, 0},
  {(char *)"parse_error", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_parse_error, 0, (char *)0, 0},
  {(char *)"is_parse_error", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_is_parse_error, 0, (char *)0, 0},
  {(char *)"illegal_argument", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_illegal_argument, 0, (char *)0, 0},
  {(char *)"is_illegal_argument", __pyx_getprop_6pytact_22graph_api_capnp_cython_16Exception_Reader_is_illegal_argument, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Exception_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Exception_Reader(_DynamicStructReader dyn)\n\n    A list of things that can go wrong during reinforcement learning.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Exception_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Exception_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Exception_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_Exception_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Exception_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader_spec = {
  "pytact.graph_api_capnp_cython.Exception_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Exception_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Exception_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Exception_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Exception_Reader(_DynamicStructReader dyn)\n\n    A list of things that can go wrong during reinforcement learning.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Exception_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Exception_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Exception_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_Exception_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_16Exception_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Exception_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader;
  new((void*)&(p->source)) ExecutionResult::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_which(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5which_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_which_raw(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9which_raw_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_failure(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7failure_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_is_failure(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_10is_failure_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_complete(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_8complete_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_is_complete(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_11is_complete_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_new_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_9new_state_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_is_new_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_12is_new_state_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_protocol_error(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_14protocol_error_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_is_protocol_error(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_17is_protocol_error_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_has_protocol_error(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_18has_protocol_error_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"which", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_which, 0, (char *)0, 0},
  {(char *)"which_raw", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_which_raw, 0, (char *)0, 0},
  {(char *)"failure", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_failure, 0, (char *)PyDoc_STR("\n        The tactic execution failed. This is not an error condition, but rather the natural failure of the tactic.\n        "), 0},
  {(char *)"is_failure", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_is_failure, 0, (char *)0, 0},
  {(char *)"complete", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_complete, 0, (char *)PyDoc_STR("\n        The proof has been completed.\n        "), 0},
  {(char *)"is_complete", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_is_complete, 0, (char *)0, 0},
  {(char *)"new_state", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_new_state, 0, (char *)PyDoc_STR("\n        The tactic ran successfully and produced a new proof state.\n        "), 0},
  {(char *)"is_new_state", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_is_new_state, 0, (char *)0, 0},
  {(char *)"protocol_error", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_protocol_error, 0, (char *)PyDoc_STR("\n        Indicates a programmer error.\n        "), 0},
  {(char *)"is_protocol_error", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_is_protocol_error, 0, (char *)0, 0},
  {(char *)"has_protocol_error", __pyx_getprop_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_has_protocol_error, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("ExecutionResult_Reader(_DynamicStructReader dyn)\n\n    The result of executing a tactic on a proof state.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader_spec = {
  "pytact.graph_api_capnp_cython.ExecutionResult_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""ExecutionResult_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("ExecutionResult_Reader(_DynamicStructReader dyn)\n\n    The result of executing a tactic on a proof state.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader;
  new((void*)&(p->source)) ExecutionResult::NewState::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_graph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5graph_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_has_graph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_graph_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5state_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_has_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_9has_state_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"graph", __pyx_getprop_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_graph, 0, (char *)0, 0},
  {(char *)"has_graph", __pyx_getprop_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_has_graph, 0, (char *)0, 0},
  {(char *)"state", __pyx_getprop_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_state, 0, (char *)0, 0},
  {(char *)"has_state", __pyx_getprop_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_has_state, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("ExecutionResult_NewState_Reader(_DynamicStructReader dyn)\n\n    The tactic ran successfully and produced a new proof state.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader_spec = {
  "pytact.graph_api_capnp_cython.ExecutionResult_NewState_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""ExecutionResult_NewState_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("ExecutionResult_NewState_Reader(_DynamicStructReader dyn)\n\n    The tactic ran successfully and produced a new proof state.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader;
  new((void*)&(p->source)) PredictionProtocol::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_dynamic, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("PredictionProtocol_Reader(_DynamicStructReader dyn)\n\n    This protocol works by exchanging raw messages over a socket. The protocol is fully linear.\n    Coq sends a `Request` and waits for a corresponding `Response` message. A message of a given\n    type should be responded with using the obviously corresponding response type.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader_spec = {
  "pytact.graph_api_capnp_cython.PredictionProtocol_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""PredictionProtocol_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("PredictionProtocol_Reader(_DynamicStructReader dyn)\n\n    This protocol works by exchanging raw messages over a socket. The protocol is fully linear.\n    Coq sends a `Request` and waits for a corresponding `Response` message. A message of a given\n    type should be responded with using the obviously corresponding response type.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader;
  new((void*)&(p->source)) PredictionProtocol::Request::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_which(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5which_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_which_raw(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_9which_raw_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_initialize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10initialize_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_is_initialize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_13is_initialize_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_predict(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7predict_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_is_predict(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_10is_predict_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_synchronize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_11synchronize_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_is_synchronize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_14is_synchronize_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_check_alignment(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_15check_alignment_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_is_check_alignment(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_18is_check_alignment_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"which", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_which, 0, (char *)0, 0},
  {(char *)"which_raw", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_which_raw, 0, (char *)0, 0},
  {(char *)"initialize", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_initialize, 0, (char *)PyDoc_STR("\n        Start a context for making tactical predictions for proof search. The context includes the tactics\n        that are currently available, the definitions that are available.\n        "), 0},
  {(char *)"is_initialize", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_is_initialize, 0, (char *)0, 0},
  {(char *)"predict", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_predict, 0, (char *)PyDoc_STR("\n        Request a list of tactic predictions given the graph of a proof state.\n        "), 0},
  {(char *)"is_predict", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_is_predict, 0, (char *)0, 0},
  {(char *)"synchronize", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_synchronize, 0, (char *)PyDoc_STR("\n        Coq uses this message to synchronize the state of the protocol when exceptions have occurred.\n        The contract is that the given integer needs to be echo'd back verbatim.\n        "), 0},
  {(char *)"is_synchronize", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_is_synchronize, 0, (char *)0, 0},
  {(char *)"check_alignment", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_check_alignment, 0, (char *)PyDoc_STR("\n        Request for the server to align the given tactics and definition to it's internal knowledge\n        and report back any tactics and definitions that were not found\n        "), 0},
  {(char *)"is_check_alignment", __pyx_getprop_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_is_check_alignment, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("PredictionProtocol_Request_Reader(_DynamicStructReader dyn)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader_spec = {
  "pytact.graph_api_capnp_cython.PredictionProtocol_Request_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""PredictionProtocol_Request_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("PredictionProtocol_Request_Reader(_DynamicStructReader dyn)"), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader;
  new((void*)&(p->source)) PredictionProtocol::Request::Initialize::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_data_version(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_12data_version_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_has_data_version(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_16has_data_version_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_stack_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_10stack_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_tactics(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7tactics_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_has_tactics(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_11has_tactics_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_graph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5graph_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_has_graph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_9has_graph_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_log_annotation(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14log_annotation_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_has_log_annotation(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_18has_log_annotation_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_representative(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_14representative_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"data_version", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_data_version, 0, (char *)PyDoc_STR("\n        The version number this message and subsequent messages is compatible with.\n        "), 0},
  {(char *)"has_data_version", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_has_data_version, 0, (char *)0, 0},
  {(char *)"stack_size", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_stack_size, 0, (char *)PyDoc_STR("\n        Initialize messages can depend on previously sent initialize messages. These messages form a stack.\n        This field indicates the size of the stack that this message depends on (if a message has no\n        dependencies, this field is zero). If the current stack is larger than the stack size of this message,\n        then the items excess items on the stack should be discarded before adding the current message to the\n        stack. It is guaranteed that the excess items will never be referenced again.\n\n        The dependency indices of nodes contained in the graphs of the stack are indexed as follows:\n        A dependency index `i` found in the graph at stack height `j` is resolved to the graph at stack height\n        `j - i`. (This scheme keeps the guarantee that dependency index `0` always points to the current graph.)\n        "), 0},
  {(char *)"tactics", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_tactics, 0, (char *)PyDoc_STR("\n        A list of tactics that Coq currently knows about.\n        "), 0},
  {(char *)"has_tactics", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_has_tactics, 0, (char *)0, 0},
  {(char *)"graph", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_graph, 0, (char *)PyDoc_STR("\n        The graph of this message. See `stackSize` on how to resolve dependency indexes in this graph.\n        "), 0},
  {(char *)"has_graph", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_has_graph, 0, (char *)0, 0},
  {(char *)"log_annotation", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_log_annotation, 0, (char *)PyDoc_STR("\n        An annotation containing file and line information on where Coq is currently processing.\n        "), 0},
  {(char *)"has_log_annotation", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_has_log_annotation, 0, (char *)0, 0},
  {(char *)"representative", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_representative, 0, (char *)PyDoc_STR("\n        Points to the last definition of the global context. All other definitions can be accessed by following\n        the `Definition.previous` chain starting from this definition. If the global context is empty\n        this is equal to `len(graph.nodes)`.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("PredictionProtocol_Request_Initialize_Reader(_DynamicStructReader dyn)\n\n    Start a context for making tactical predictions for proof search. The context includes the tactics\n    that are currently available, the definitions that are available.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader_spec = {
  "pytact.graph_api_capnp_cython.PredictionProtocol_Request_Initialize_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""PredictionProtocol_Request_Initialize_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("PredictionProtocol_Request_Initialize_Reader(_DynamicStructReader dyn)\n\n    Start a context for making tactical predictions for proof search. The context includes the tactics\n    that are currently available, the definitions that are available.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader;
  new((void*)&(p->source)) PredictionProtocol::Request::Predict::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_graph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5graph_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_has_graph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_graph_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5state_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_has_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_9has_state_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"graph", __pyx_getprop_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_graph, 0, (char *)PyDoc_STR("\n        The graph may reference definitions present in the stack of initialize messages. See\n        `initialize.stackSize` on how dependency indices can be resolved to a graph in the stack.\n        "), 0},
  {(char *)"has_graph", __pyx_getprop_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_has_graph, 0, (char *)0, 0},
  {(char *)"state", __pyx_getprop_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_state, 0, (char *)PyDoc_STR("\n        The proof state for which a prediction is requested.\n        "), 0},
  {(char *)"has_state", __pyx_getprop_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_has_state, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("PredictionProtocol_Request_Predict_Reader(_DynamicStructReader dyn)\n\n    Request a list of tactic predictions given the graph of a proof state.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader_spec = {
  "pytact.graph_api_capnp_cython.PredictionProtocol_Request_Predict_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""PredictionProtocol_Request_Predict_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("PredictionProtocol_Request_Predict_Reader(_DynamicStructReader dyn)\n\n    Request a list of tactic predictions given the graph of a proof state.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader;
  new((void*)&(p->source)) PredictionProtocol::Prediction::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_tactic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6tactic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_has_tactic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10has_tactic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_arguments(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_9arguments_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_has_arguments(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_13has_arguments_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_confidence(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_10confidence_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"tactic", __pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_tactic, 0, (char *)0, 0},
  {(char *)"has_tactic", __pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_has_tactic, 0, (char *)0, 0},
  {(char *)"arguments", __pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_arguments, 0, (char *)0, 0},
  {(char *)"has_arguments", __pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_has_arguments, 0, (char *)0, 0},
  {(char *)"confidence", __pyx_getprop_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_confidence, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("PredictionProtocol_Prediction_Reader(_DynamicStructReader dyn)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_spec = {
  "pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""PredictionProtocol_Prediction_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("PredictionProtocol_Prediction_Reader(_DynamicStructReader dyn)"), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader;
  new((void*)&(p->source)) PredictionProtocol::TextPrediction::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_tactic_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_11tactic_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_has_tactic_text(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_15has_tactic_text_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_confidence(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_10confidence_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"tactic_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_tactic_text, 0, (char *)0, 0},
  {(char *)"has_tactic_text", __pyx_getprop_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_has_tactic_text, 0, (char *)0, 0},
  {(char *)"confidence", __pyx_getprop_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_confidence, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("PredictionProtocol_TextPrediction_Reader(_DynamicStructReader dyn)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_spec = {
  "pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""PredictionProtocol_TextPrediction_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("PredictionProtocol_TextPrediction_Reader(_DynamicStructReader dyn)"), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader;
  new((void*)&(p->source)) PredictionProtocol::Response::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_which(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5which_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_which_raw(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9which_raw_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_initialized(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_11initialized_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_is_initialized(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14is_initialized_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_prediction(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_10prediction_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_is_prediction(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_13is_prediction_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_has_prediction(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_14has_prediction_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_text_prediction(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15text_prediction_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_is_text_prediction(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_18is_text_prediction_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_has_text_prediction(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_19has_text_prediction_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_synchronized(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12synchronized_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_is_synchronized(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_15is_synchronized_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_alignment(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_9alignment_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_is_alignment(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_12is_alignment_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"which", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_which, 0, (char *)0, 0},
  {(char *)"which_raw", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_which_raw, 0, (char *)0, 0},
  {(char *)"initialized", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_initialized, 0, (char *)0, 0},
  {(char *)"is_initialized", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_is_initialized, 0, (char *)0, 0},
  {(char *)"prediction", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_prediction, 0, (char *)0, 0},
  {(char *)"is_prediction", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_is_prediction, 0, (char *)0, 0},
  {(char *)"has_prediction", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_has_prediction, 0, (char *)0, 0},
  {(char *)"text_prediction", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_text_prediction, 0, (char *)PyDoc_STR("\n        Output is a list of predictions with a confidence. The list is expected to be\n        sorted by decreasing confidence.\n        "), 0},
  {(char *)"is_text_prediction", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_is_text_prediction, 0, (char *)0, 0},
  {(char *)"has_text_prediction", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_has_text_prediction, 0, (char *)0, 0},
  {(char *)"synchronized", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_synchronized, 0, (char *)0, 0},
  {(char *)"is_synchronized", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_is_synchronized, 0, (char *)0, 0},
  {(char *)"alignment", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_alignment, 0, (char *)0, 0},
  {(char *)"is_alignment", __pyx_getprop_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_is_alignment, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("PredictionProtocol_Response_Reader(_DynamicStructReader dyn)\n\n    See Request for documentation.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader_spec = {
  "pytact.graph_api_capnp_cython.PredictionProtocol_Response_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""PredictionProtocol_Response_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("PredictionProtocol_Response_Reader(_DynamicStructReader dyn)\n\n    See Request for documentation.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader;
  new((void*)&(p->source)) PredictionProtocol::Response::Alignment::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_unaligned_tactics(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_17unaligned_tactics_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_has_unaligned_tactics(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21has_unaligned_tactics_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_unaligned_definitions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_21unaligned_definitions_1__get__(o);
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_has_unaligned_definitions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_25has_unaligned_definitions_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_dynamic, 0, (char *)0, 0},
  {(char *)"unaligned_tactics", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_unaligned_tactics, 0, (char *)0, 0},
  {(char *)"has_unaligned_tactics", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_has_unaligned_tactics, 0, (char *)0, 0},
  {(char *)"unaligned_definitions", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_unaligned_definitions, 0, (char *)0, 0},
  {(char *)"has_unaligned_definitions", __pyx_getprop_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_has_unaligned_definitions, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("PredictionProtocol_Response_Alignment_Reader(_DynamicStructReader dyn)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader_spec = {
  "pytact.graph_api_capnp_cython.PredictionProtocol_Response_Alignment_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""PredictionProtocol_Response_Alignment_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("PredictionProtocol_Response_Alignment_Reader(_DynamicStructReader dyn)"), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader __pyx_vtable_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader;
  new((void*)&(p->source)) ConflatableEdges::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_dynamic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7dynamic_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_3__repr__(self);
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader[] = {
  {(char *)"dynamic", __pyx_getprop_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_dynamic, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader},
  {Py_tp_repr, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("ConflatableEdges_Reader(_DynamicStructReader dyn)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader},
  {Py_tp_getset, (void *)__pyx_getsets_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader},
  {Py_tp_init, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader_spec = {
  "pytact.graph_api_capnp_cython.ConflatableEdges_Reader",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""ConflatableEdges_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("ConflatableEdges_Reader(_DynamicStructReader dyn)"), /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint16_List __pyx_vtable_6pytact_22graph_api_capnp_cython_Uint16_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint16_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint16_List;
  new((void*)&(p->source))  ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint16_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint16_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Uint16_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Uint16_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_Uint16_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Uint16_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint16_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint16_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint16_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_Uint16_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Uint16_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Uint16_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Uint16_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint16_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List_spec = {
  "pytact.graph_api_capnp_cython.Uint16_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Uint16_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_Uint16_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Uint16_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint16_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Uint16_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint16_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint16_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Uint16_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Uint16_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Uint16_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Uint16_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Uint16_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint16_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint32_List __pyx_vtable_6pytact_22graph_api_capnp_cython_Uint32_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint32_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint32_List;
  new((void*)&(p->source))  ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint32_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint32_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Uint32_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Uint32_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_Uint32_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Uint32_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint32_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint32_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint32_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_Uint32_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Uint32_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Uint32_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Uint32_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint32_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List_spec = {
  "pytact.graph_api_capnp_cython.Uint32_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Uint32_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_Uint32_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Uint32_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint32_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Uint32_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint32_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint32_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Uint32_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Uint32_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Uint32_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Uint32_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Uint32_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint32_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Uint64_List __pyx_vtable_6pytact_22graph_api_capnp_cython_Uint64_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint64_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint64_List;
  new((void*)&(p->source))  ::capnp::List<uint64_t, ::capnp::Kind::PRIMITIVE>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint64_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint64_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Uint64_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Uint64_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_Uint64_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Uint64_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint64_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11Uint64_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint64_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_Uint64_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Uint64_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Uint64_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Uint64_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint64_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List_spec = {
  "pytact.graph_api_capnp_cython.Uint64_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Uint64_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_Uint64_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Uint64_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_11Uint64_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Uint64_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Uint64_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Uint64_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Uint64_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Uint64_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Uint64_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Uint64_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Uint64_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Uint64_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Int64_List __pyx_vtable_6pytact_22graph_api_capnp_cython_Int64_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Int64_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Int64_List;
  new((void*)&(p->source))  ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Int64_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Int64_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Int64_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Int64_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_Int64_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Int64_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_10Int64_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_10Int64_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Int64_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Int64_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_Int64_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Int64_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Int64_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Int64_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Int64_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Int64_List_spec = {
  "pytact.graph_api_capnp_cython.Int64_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Int64_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Int64_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_Int64_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Int64_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_10Int64_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Int64_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Int64_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Int64_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Int64_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Int64_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Int64_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Int64_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Int64_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Int64_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Int64_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_String_List __pyx_vtable_6pytact_22graph_api_capnp_cython_String_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_String_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_String_List;
  new((void*)&(p->source))  ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_String_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_String_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_String_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_String_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_String_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_String_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11String_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_11String_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_String_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_String_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_String_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_String_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_String_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_String_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_String_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_String_List_spec = {
  "pytact.graph_api_capnp_cython.String_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_String_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_String_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_String_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_String_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_11String_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_String_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""String_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_String_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_String_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_String_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_String_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_String_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_String_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_String_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_String_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List;
  new((void*)&(p->source))  ::capnp::List<PredictionProtocol::Prediction, ::capnp::Kind::STRUCT>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List_spec = {
  "pytact.graph_api_capnp_cython.PredictionProtocol_Prediction_Reader_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_PredictionProtocol_Prediction_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_PredictionProtocol_Prediction_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""PredictionProtocol_Prediction_Reader_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_PredictionProtocol_Prediction_Reader_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_PredictionProtocol_Prediction_Reader_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List;
  new((void*)&(p->source))  ::capnp::List<Graph::Node, ::capnp::Kind::STRUCT>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List_spec = {
  "pytact.graph_api_capnp_cython.Graph_Node_Reader_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Graph_Node_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Graph_Node_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Graph_Node_Reader_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Graph_Node_Reader_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Graph_Node_Reader_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List __pyx_vtable_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List;
  new((void*)&(p->source))  ::capnp::List<AbstractTactic, ::capnp::Kind::STRUCT>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List_spec = {
  "pytact.graph_api_capnp_cython.AbstractTactic_Reader_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_AbstractTactic_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_AbstractTactic_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""AbstractTactic_Reader_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_AbstractTactic_Reader_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_AbstractTactic_Reader_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List;
  new((void*)&(p->source))  ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List_spec = {
  "pytact.graph_api_capnp_cython.ProofStep_Reader_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_ProofStep_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_ProofStep_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""ProofStep_Reader_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_ProofStep_Reader_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_ProofStep_Reader_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List;
  new((void*)&(p->source))  ::capnp::List<PredictionProtocol::TextPrediction, ::capnp::Kind::STRUCT>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List_spec = {
  "pytact.graph_api_capnp_cython.PredictionProtocol_TextPrediction_Reader_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_PredictionProtocol_TextPrediction_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_PredictionProtocol_TextPrediction_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""PredictionProtocol_TextPrediction_Reader_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_PredictionProtocol_TextPrediction_Reader_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_PredictionProtocol_TextPrediction_Reader_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_ProofState_Reader_List __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofState_Reader_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Reader_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Reader_List;
  new((void*)&(p->source))  ::capnp::List<ProofState, ::capnp::Kind::STRUCT>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Reader_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Reader_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofState_Reader_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofState_Reader_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_ProofState_Reader_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_ProofState_Reader_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Reader_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_ProofState_Reader_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofState_Reader_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofState_Reader_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_ProofState_Reader_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Reader_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List_spec = {
  "pytact.graph_api_capnp_cython.ProofState_Reader_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_ProofState_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_ProofState_Reader_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_ProofState_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""ProofState_Reader_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_ProofState_Reader_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_ProofState_Reader_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_ProofState_Reader_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_ProofState_Reader_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_ProofState_Reader_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_ProofState_Reader_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_ProofState_Reader_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_ProofState_Reader_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List;
  new((void*)&(p->source))  ::capnp::List<Graph::EdgeTarget, ::capnp::Kind::STRUCT>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List_spec = {
  "pytact.graph_api_capnp_cython.Graph_EdgeTarget_Reader_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Graph_EdgeTarget_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Graph_EdgeTarget_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Graph_EdgeTarget_Reader_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Graph_EdgeTarget_Reader_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Graph_EdgeTarget_Reader_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Argument_Reader_List __pyx_vtable_6pytact_22graph_api_capnp_cython_Argument_Reader_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Reader_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Reader_List;
  new((void*)&(p->source))  ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Reader_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Reader_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Argument_Reader_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Argument_Reader_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_Argument_Reader_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Argument_Reader_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Reader_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_Argument_Reader_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Argument_Reader_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Argument_Reader_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Argument_Reader_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Reader_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List_spec = {
  "pytact.graph_api_capnp_cython.Argument_Reader_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Argument_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_Argument_Reader_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Argument_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Argument_Reader_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Argument_Reader_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Argument_Reader_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Argument_Reader_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Argument_Reader_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Argument_Reader_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Argument_Reader_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Argument_Reader_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Argument_Reader_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Outcome_Reader_List __pyx_vtable_6pytact_22graph_api_capnp_cython_Outcome_Reader_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Reader_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Reader_List;
  new((void*)&(p->source))  ::capnp::List<Outcome, ::capnp::Kind::STRUCT>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Reader_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Reader_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Outcome_Reader_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Outcome_Reader_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_Outcome_Reader_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Outcome_Reader_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Reader_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_Outcome_Reader_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Outcome_Reader_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Outcome_Reader_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Outcome_Reader_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Reader_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List_spec = {
  "pytact.graph_api_capnp_cython.Outcome_Reader_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Outcome_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_Outcome_Reader_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Outcome_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Outcome_Reader_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Outcome_Reader_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Outcome_Reader_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Outcome_Reader_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Outcome_Reader_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Outcome_Reader_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Outcome_Reader_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Outcome_Reader_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Outcome_Reader_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pytact_22graph_api_capnp_cython_Node_Reader_List __pyx_vtable_6pytact_22graph_api_capnp_cython_Node_Reader_List;

static PyObject *__pyx_tp_new_6pytact_22graph_api_capnp_cython_Node_Reader_List(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Node_Reader_List;
  new((void*)&(p->source))  ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader();
  p->root = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Node_Reader_List(PyObject *o) {
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Node_Reader_List) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->source);
  Py_CLEAR(p->root);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Node_Reader_List(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *)o;
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Node_Reader_List(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *p = (struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List *)o;
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6pytact_22graph_api_capnp_cython_Node_Reader_List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_6pytact_22graph_api_capnp_cython_Node_Reader_List[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16Node_Reader_List_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pytact_22graph_api_capnp_cython_16Node_Reader_List_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Node_Reader_List},
  {Py_sq_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_3__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6pytact_22graph_api_capnp_cython_Node_Reader_List},
  {Py_mp_length, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_3__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_1__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Node_Reader_List},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pytact_22graph_api_capnp_cython_Node_Reader_List},
  {Py_tp_methods, (void *)__pyx_methods_6pytact_22graph_api_capnp_cython_Node_Reader_List},
  {Py_tp_new, (void *)__pyx_tp_new_6pytact_22graph_api_capnp_cython_Node_Reader_List},
  {0, 0},
};
static PyType_Spec __pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List_spec = {
  "pytact.graph_api_capnp_cython.Node_Reader_List",
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Node_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6pytact_22graph_api_capnp_cython_Node_Reader_List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Node_Reader_List = {
  __pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_3__len__, /*mp_length*/
  __pyx_pw_6pytact_22graph_api_capnp_cython_16Node_Reader_List_1__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""Node_Reader_List", /*tp_name*/
  sizeof(struct __pyx_obj_6pytact_22graph_api_capnp_cython_Node_Reader_List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pytact_22graph_api_capnp_cython_Node_Reader_List, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Node_Reader_List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Node_Reader_List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pytact_22graph_api_capnp_cython_Node_Reader_List, /*tp_traverse*/
  __pyx_tp_clear_6pytact_22graph_api_capnp_cython_Node_Reader_List, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pytact_22graph_api_capnp_cython_Node_Reader_List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pytact_22graph_api_capnp_cython_Node_Reader_List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static int __pyx_tp_traverse___Pyx_EnumMeta(PyObject *o, visitproc v, void *a) {
  int e;
  if (!(&PyType_Type)->tp_traverse); else { e = (&PyType_Type)->tp_traverse(o,v,a); if (e) return e; }
  return 0;
}

static int __pyx_tp_clear___Pyx_EnumMeta(PyObject *o) {
  if (!(&PyType_Type)->tp_clear); else (&PyType_Type)->tp_clear(o);
  return 0;
}
static PyObject *__pyx_sq_item___Pyx_EnumMeta(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods___Pyx_EnumMeta[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __Pyx_EnumMeta_slots[] = {
  {Py_sq_item, (void *)__pyx_sq_item___Pyx_EnumMeta},
  {Py_mp_subscript, (void *)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse___Pyx_EnumMeta},
  {Py_tp_clear, (void *)__pyx_tp_clear___Pyx_EnumMeta},
  {Py_tp_iter, (void *)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__},
  {Py_tp_methods, (void *)__pyx_methods___Pyx_EnumMeta},
  {Py_tp_init, (void *)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__},
  {0, 0},
};
static PyType_Spec __Pyx_EnumMeta_spec = {
  "pytact.graph_api_capnp_cython.__Pyx_EnumMeta",
  sizeof(struct __pyx_obj___Pyx_EnumMeta),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __Pyx_EnumMeta_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence___Pyx_EnumMeta = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item___Pyx_EnumMeta, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping___Pyx_EnumMeta = {
  0, /*mp_length*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __Pyx_EnumMeta = {
  PyVarObject_HEAD_INIT(0, 0)
  "pytact.graph_api_capnp_cython.""__Pyx_EnumMeta", /*tp_name*/
  sizeof(struct __pyx_obj___Pyx_EnumMeta), /*tp_basicsize*/
  0, /*tp_itemsize*/
  0, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence___Pyx_EnumMeta, /*tp_as_sequence*/
  &__pyx_tp_as_mapping___Pyx_EnumMeta, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse___Pyx_EnumMeta, /*tp_traverse*/
  __pyx_tp_clear___Pyx_EnumMeta, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods___Pyx_EnumMeta, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  0, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  #if CYTHON_USE_MODULE_STATE
  {0, __pyx_k_ALIGNMENT, sizeof(__pyx_k_ALIGNMENT), 0, 0, 1, 1},
  {0, __pyx_k_APP, sizeof(__pyx_k_APP), 0, 0, 1, 1},
  {0, __pyx_k_APP_ARG, sizeof(__pyx_k_APP_ARG), 0, 0, 1, 1},
  {0, __pyx_k_APP_FUN, sizeof(__pyx_k_APP_FUN), 0, 0, 1, 1},
  {0, __pyx_k_AbstractTactic, sizeof(__pyx_k_AbstractTactic), 0, 0, 1, 1},
  {0, __pyx_k_AbstractTactic_Reader, sizeof(__pyx_k_AbstractTactic_Reader), 0, 0, 1, 1},
  {0, __pyx_k_AbstractTactic_Reader_List, sizeof(__pyx_k_AbstractTactic_Reader_List), 0, 0, 1, 1},
  {0, __pyx_k_AbstractTactic_Reader_List___red, sizeof(__pyx_k_AbstractTactic_Reader_List___red), 0, 0, 1, 1},
  {0, __pyx_k_AbstractTactic_Reader_List___set, sizeof(__pyx_k_AbstractTactic_Reader_List___set), 0, 0, 1, 1},
  {0, __pyx_k_AbstractTactic_Reader___reduce_c, sizeof(__pyx_k_AbstractTactic_Reader___reduce_c), 0, 0, 1, 1},
  {0, __pyx_k_AbstractTactic_Reader___setstate, sizeof(__pyx_k_AbstractTactic_Reader___setstate), 0, 0, 1, 1},
  {0, __pyx_k_Argument, sizeof(__pyx_k_Argument), 0, 0, 1, 1},
  {0, __pyx_k_Argument_Reader, sizeof(__pyx_k_Argument_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Argument_Reader_List, sizeof(__pyx_k_Argument_Reader_List), 0, 0, 1, 1},
  {0, __pyx_k_Argument_Reader_List___reduce_cy, sizeof(__pyx_k_Argument_Reader_List___reduce_cy), 0, 0, 1, 1},
  {0, __pyx_k_Argument_Reader_List___setstate, sizeof(__pyx_k_Argument_Reader_List___setstate), 0, 0, 1, 1},
  {0, __pyx_k_Argument_Reader___reduce_cython, sizeof(__pyx_k_Argument_Reader___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Argument_Reader___setstate_cytho, sizeof(__pyx_k_Argument_Reader___setstate_cytho), 0, 0, 1, 1},
  {0, __pyx_k_Argument_Term_Reader, sizeof(__pyx_k_Argument_Term_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Argument_Term_Reader___reduce_cy, sizeof(__pyx_k_Argument_Term_Reader___reduce_cy), 0, 0, 1, 1},
  {0, __pyx_k_Argument_Term_Reader___setstate, sizeof(__pyx_k_Argument_Term_Reader___setstate), 0, 0, 1, 1},
  {0, __pyx_k_Argument_Which, sizeof(__pyx_k_Argument_Which), 0, 0, 1, 1},
  {0, __pyx_k_CASE, sizeof(__pyx_k_CASE), 0, 0, 1, 1},
  {0, __pyx_k_CASE_BRANCH, sizeof(__pyx_k_CASE_BRANCH), 0, 0, 1, 1},
  {0, __pyx_k_CASE_BRANCH_POINTER, sizeof(__pyx_k_CASE_BRANCH_POINTER), 0, 0, 1, 1},
  {0, __pyx_k_CASE_IND, sizeof(__pyx_k_CASE_IND), 0, 0, 1, 1},
  {0, __pyx_k_CASE_RETURN, sizeof(__pyx_k_CASE_RETURN), 0, 0, 1, 1},
  {0, __pyx_k_CASE_TERM, sizeof(__pyx_k_CASE_TERM), 0, 0, 1, 1},
  {0, __pyx_k_CAST, sizeof(__pyx_k_CAST), 0, 0, 1, 1},
  {0, __pyx_k_CAST_TERM, sizeof(__pyx_k_CAST_TERM), 0, 0, 1, 1},
  {0, __pyx_k_CAST_TYPE, sizeof(__pyx_k_CAST_TYPE), 0, 0, 1, 1},
  {0, __pyx_k_CHECK_ALIGNMENT, sizeof(__pyx_k_CHECK_ALIGNMENT), 0, 0, 1, 1},
  {0, __pyx_k_COMPLETE, sizeof(__pyx_k_COMPLETE), 0, 0, 1, 1},
  {0, __pyx_k_CONSTRUCTOR, sizeof(__pyx_k_CONSTRUCTOR), 0, 0, 1, 1},
  {0, __pyx_k_CONSTRUCT_TERM, sizeof(__pyx_k_CONSTRUCT_TERM), 0, 0, 1, 1},
  {0, __pyx_k_CONST_DEF, sizeof(__pyx_k_CONST_DEF), 0, 0, 1, 1},
  {0, __pyx_k_CONST_EMPTY, sizeof(__pyx_k_CONST_EMPTY), 0, 0, 1, 1},
  {0, __pyx_k_CONST_OPAQUE_DEF, sizeof(__pyx_k_CONST_OPAQUE_DEF), 0, 0, 1, 1},
  {0, __pyx_k_CONST_PRIMITIVE, sizeof(__pyx_k_CONST_PRIMITIVE), 0, 0, 1, 1},
  {0, __pyx_k_CONST_TYPE, sizeof(__pyx_k_CONST_TYPE), 0, 0, 1, 1},
  {0, __pyx_k_CONST_UNDEF, sizeof(__pyx_k_CONST_UNDEF), 0, 0, 1, 1},
  {0, __pyx_k_CONTEXT_ASSUM, sizeof(__pyx_k_CONTEXT_ASSUM), 0, 0, 1, 1},
  {0, __pyx_k_CONTEXT_DEF, sizeof(__pyx_k_CONTEXT_DEF), 0, 0, 1, 1},
  {0, __pyx_k_CONTEXT_DEF_TERM, sizeof(__pyx_k_CONTEXT_DEF_TERM), 0, 0, 1, 1},
  {0, __pyx_k_CONTEXT_DEF_TYPE, sizeof(__pyx_k_CONTEXT_DEF_TYPE), 0, 0, 1, 1},
  {0, __pyx_k_CONTEXT_ELEM, sizeof(__pyx_k_CONTEXT_ELEM), 0, 0, 1, 1},
  {0, __pyx_k_CONTEXT_SUBJECT, sizeof(__pyx_k_CONTEXT_SUBJECT), 0, 0, 1, 1},
  {0, __pyx_k_CO_FIX, sizeof(__pyx_k_CO_FIX), 0, 0, 1, 1},
  {0, __pyx_k_CO_FIX_FUN, sizeof(__pyx_k_CO_FIX_FUN), 0, 0, 1, 1},
  {0, __pyx_k_CO_FIX_FUN_TERM, sizeof(__pyx_k_CO_FIX_FUN_TERM), 0, 0, 1, 1},
  {0, __pyx_k_CO_FIX_FUN_TYPE, sizeof(__pyx_k_CO_FIX_FUN_TYPE), 0, 0, 1, 1},
  {0, __pyx_k_CO_FIX_MUTUAL, sizeof(__pyx_k_CO_FIX_MUTUAL), 0, 0, 1, 1},
  {0, __pyx_k_CO_FIX_RETURN, sizeof(__pyx_k_CO_FIX_RETURN), 0, 0, 1, 1},
  {0, __pyx_k_C_B_CONSTRUCT, sizeof(__pyx_k_C_B_CONSTRUCT), 0, 0, 1, 1},
  {0, __pyx_k_C_B_TERM, sizeof(__pyx_k_C_B_TERM), 0, 0, 1, 1},
  {0, __pyx_k_ConflatableEdges, sizeof(__pyx_k_ConflatableEdges), 0, 0, 1, 1},
  {0, __pyx_k_ConflatableEdges_Reader, sizeof(__pyx_k_ConflatableEdges_Reader), 0, 0, 1, 1},
  {0, __pyx_k_ConflatableEdges_Reader___reduce, sizeof(__pyx_k_ConflatableEdges_Reader___reduce), 0, 0, 1, 1},
  {0, __pyx_k_ConflatableEdges_Reader___setsta, sizeof(__pyx_k_ConflatableEdges_Reader___setsta), 0, 0, 1, 1},
  {0, __pyx_k_DEFINITION, sizeof(__pyx_k_DEFINITION), 0, 0, 1, 1},
  {0, __pyx_k_DISCHARGED, sizeof(__pyx_k_DISCHARGED), 0, 0, 1, 1},
  {0, __pyx_k_DataVersion, sizeof(__pyx_k_DataVersion), 0, 0, 1, 1},
  {0, __pyx_k_DataVersion_Reader, sizeof(__pyx_k_DataVersion_Reader), 0, 0, 1, 1},
  {0, __pyx_k_DataVersion_Reader___reduce_cyth, sizeof(__pyx_k_DataVersion_Reader___reduce_cyth), 0, 0, 1, 1},
  {0, __pyx_k_DataVersion_Reader___setstate_cy, sizeof(__pyx_k_DataVersion_Reader___setstate_cy), 0, 0, 1, 1},
  {0, __pyx_k_Dataset, sizeof(__pyx_k_Dataset), 0, 0, 1, 1},
  {0, __pyx_k_Dataset_Reader, sizeof(__pyx_k_Dataset_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Dataset_Reader___reduce_cython, sizeof(__pyx_k_Dataset_Reader___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Dataset_Reader___setstate_cython, sizeof(__pyx_k_Dataset_Reader___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_Definition, sizeof(__pyx_k_Definition), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Reader, sizeof(__pyx_k_Definition_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Reader___reduce_cytho, sizeof(__pyx_k_Definition_Reader___reduce_cytho), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Reader___setstate_cyt, sizeof(__pyx_k_Definition_Reader___setstate_cyt), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Status_Reader, sizeof(__pyx_k_Definition_Status_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Status_Reader___reduc, sizeof(__pyx_k_Definition_Status_Reader___reduc), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Status_Reader___setst, sizeof(__pyx_k_Definition_Status_Reader___setst), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Status_Substituted_Re, sizeof(__pyx_k_Definition_Status_Substituted_Re), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Status_Substituted_Re_2, sizeof(__pyx_k_Definition_Status_Substituted_Re_2), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Status_Substituted_Re_3, sizeof(__pyx_k_Definition_Status_Substituted_Re_3), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Status_Which, sizeof(__pyx_k_Definition_Status_Which), 0, 0, 1, 1},
  {0, __pyx_k_Definition_Which, sizeof(__pyx_k_Definition_Which), 0, 0, 1, 1},
  {0, __pyx_k_EVAR, sizeof(__pyx_k_EVAR), 0, 0, 1, 1},
  {0, __pyx_k_EVAR_SUBJECT, sizeof(__pyx_k_EVAR_SUBJECT), 0, 0, 1, 1},
  {0, __pyx_k_EVAR_SUBST, sizeof(__pyx_k_EVAR_SUBST), 0, 0, 1, 1},
  {0, __pyx_k_EVAR_SUBST_POINTER, sizeof(__pyx_k_EVAR_SUBST_POINTER), 0, 0, 1, 1},
  {0, __pyx_k_EVAR_SUBST_TARGET, sizeof(__pyx_k_EVAR_SUBST_TARGET), 0, 0, 1, 1},
  {0, __pyx_k_EVAR_SUBST_TERM, sizeof(__pyx_k_EVAR_SUBST_TERM), 0, 0, 1, 1},
  {0, __pyx_k_EdgeClassification, sizeof(__pyx_k_EdgeClassification), 0, 0, 1, 1},
  {0, __pyx_k_EdgeTarget, sizeof(__pyx_k_EdgeTarget), 0, 0, 1, 1},
  {0, __pyx_k_EnumBase, sizeof(__pyx_k_EnumBase), 0, 0, 1, 1},
  {0, __pyx_k_Exception, sizeof(__pyx_k_Exception), 0, 0, 1, 1},
  {0, __pyx_k_Exception_Reader, sizeof(__pyx_k_Exception_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Exception_Reader___reduce_cython, sizeof(__pyx_k_Exception_Reader___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Exception_Reader___setstate_cyth, sizeof(__pyx_k_Exception_Reader___setstate_cyth), 0, 0, 1, 1},
  {0, __pyx_k_Exception_Which, sizeof(__pyx_k_Exception_Which), 0, 0, 1, 1},
  {0, __pyx_k_ExecutionResult, sizeof(__pyx_k_ExecutionResult), 0, 0, 1, 1},
  {0, __pyx_k_ExecutionResult_NewState_Reader, sizeof(__pyx_k_ExecutionResult_NewState_Reader), 0, 0, 1, 1},
  {0, __pyx_k_ExecutionResult_NewState_Reader_2, sizeof(__pyx_k_ExecutionResult_NewState_Reader_2), 0, 0, 1, 1},
  {0, __pyx_k_ExecutionResult_NewState_Reader_3, sizeof(__pyx_k_ExecutionResult_NewState_Reader_3), 0, 0, 1, 1},
  {0, __pyx_k_ExecutionResult_Reader, sizeof(__pyx_k_ExecutionResult_Reader), 0, 0, 1, 1},
  {0, __pyx_k_ExecutionResult_Reader___reduce, sizeof(__pyx_k_ExecutionResult_Reader___reduce), 0, 0, 1, 1},
  {0, __pyx_k_ExecutionResult_Reader___setstat, sizeof(__pyx_k_ExecutionResult_Reader___setstat), 0, 0, 1, 1},
  {0, __pyx_k_ExecutionResult_Which, sizeof(__pyx_k_ExecutionResult_Which), 0, 0, 1, 1},
  {0, __pyx_k_FAILURE, sizeof(__pyx_k_FAILURE), 0, 0, 1, 1},
  {0, __pyx_k_FIX, sizeof(__pyx_k_FIX), 0, 0, 1, 1},
  {0, __pyx_k_FIX_FUN, sizeof(__pyx_k_FIX_FUN), 0, 0, 1, 1},
  {0, __pyx_k_FIX_FUN_TERM, sizeof(__pyx_k_FIX_FUN_TERM), 0, 0, 1, 1},
  {0, __pyx_k_FIX_FUN_TYPE, sizeof(__pyx_k_FIX_FUN_TYPE), 0, 0, 1, 1},
  {0, __pyx_k_FIX_MUTUAL, sizeof(__pyx_k_FIX_MUTUAL), 0, 0, 1, 1},
  {0, __pyx_k_FIX_RETURN, sizeof(__pyx_k_FIX_RETURN), 0, 0, 1, 1},
  {0, __pyx_k_FLOAT, sizeof(__pyx_k_FLOAT), 0, 0, 1, 1},
  {0, __pyx_k_FloatP, sizeof(__pyx_k_FloatP), 0, 0, 1, 1},
  {0, __pyx_k_FloatP_Reader, sizeof(__pyx_k_FloatP_Reader), 0, 0, 1, 1},
  {0, __pyx_k_FloatP_Reader___reduce_cython, sizeof(__pyx_k_FloatP_Reader___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_FloatP_Reader___setstate_cython, sizeof(__pyx_k_FloatP_Reader___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_Graph, sizeof(__pyx_k_Graph), 0, 0, 1, 1},
  {0, __pyx_k_Graph_EdgeTarget_Reader, sizeof(__pyx_k_Graph_EdgeTarget_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Graph_EdgeTarget_Reader_List, sizeof(__pyx_k_Graph_EdgeTarget_Reader_List), 0, 0, 1, 1},
  {0, __pyx_k_Graph_EdgeTarget_Reader_List___r, sizeof(__pyx_k_Graph_EdgeTarget_Reader_List___r), 0, 0, 1, 1},
  {0, __pyx_k_Graph_EdgeTarget_Reader_List___s, sizeof(__pyx_k_Graph_EdgeTarget_Reader_List___s), 0, 0, 1, 1},
  {0, __pyx_k_Graph_EdgeTarget_Reader___reduce, sizeof(__pyx_k_Graph_EdgeTarget_Reader___reduce), 0, 0, 1, 1},
  {0, __pyx_k_Graph_EdgeTarget_Reader___setsta, sizeof(__pyx_k_Graph_EdgeTarget_Reader___setsta), 0, 0, 1, 1},
  {0, __pyx_k_Graph_EdgeTarget_Target_Reader, sizeof(__pyx_k_Graph_EdgeTarget_Target_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Graph_EdgeTarget_Target_Reader_2, sizeof(__pyx_k_Graph_EdgeTarget_Target_Reader_2), 0, 0, 1, 1},
  {0, __pyx_k_Graph_EdgeTarget_Target_Reader_3, sizeof(__pyx_k_Graph_EdgeTarget_Target_Reader_3), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Node_Label_Reader, sizeof(__pyx_k_Graph_Node_Label_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Node_Label_Reader___reduce, sizeof(__pyx_k_Graph_Node_Label_Reader___reduce), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Node_Label_Reader___setsta, sizeof(__pyx_k_Graph_Node_Label_Reader___setsta), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Node_Label_Which, sizeof(__pyx_k_Graph_Node_Label_Which), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Node_Reader, sizeof(__pyx_k_Graph_Node_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Node_Reader_List, sizeof(__pyx_k_Graph_Node_Reader_List), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Node_Reader_List___reduce, sizeof(__pyx_k_Graph_Node_Reader_List___reduce), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Node_Reader_List___setstat, sizeof(__pyx_k_Graph_Node_Reader_List___setstat), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Node_Reader___reduce_cytho, sizeof(__pyx_k_Graph_Node_Reader___reduce_cytho), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Node_Reader___setstate_cyt, sizeof(__pyx_k_Graph_Node_Reader___setstate_cyt), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Reader, sizeof(__pyx_k_Graph_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Reader___reduce_cython, sizeof(__pyx_k_Graph_Reader___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Graph_Reader___setstate_cython, sizeof(__pyx_k_Graph_Reader___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_ILLEGAL_ARGUMENT, sizeof(__pyx_k_ILLEGAL_ARGUMENT), 0, 0, 1, 1},
  {0, __pyx_k_INDUCTIVE, sizeof(__pyx_k_INDUCTIVE), 0, 0, 1, 1},
  {0, __pyx_k_IND_CONSTRUCT, sizeof(__pyx_k_IND_CONSTRUCT), 0, 0, 1, 1},
  {0, __pyx_k_IND_PROJECTION, sizeof(__pyx_k_IND_PROJECTION), 0, 0, 1, 1},
  {0, __pyx_k_IND_TYPE, sizeof(__pyx_k_IND_TYPE), 0, 0, 1, 1},
  {0, __pyx_k_INITIALIZE, sizeof(__pyx_k_INITIALIZE), 0, 0, 1, 1},
  {0, __pyx_k_INITIALIZED, sizeof(__pyx_k_INITIALIZED), 0, 0, 1, 1},
  {0, __pyx_k_INT, sizeof(__pyx_k_INT), 0, 0, 1, 1},
  {0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {0, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {0, __pyx_k_Int64_List, sizeof(__pyx_k_Int64_List), 0, 0, 1, 1},
  {0, __pyx_k_Int64_List___reduce_cython, sizeof(__pyx_k_Int64_List___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Int64_List___setstate_cython, sizeof(__pyx_k_Int64_List___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_IntEnum, sizeof(__pyx_k_IntEnum), 0, 0, 1, 1},
  {0, __pyx_k_IntP, sizeof(__pyx_k_IntP), 0, 0, 1, 1},
  {0, __pyx_k_IntP_Reader, sizeof(__pyx_k_IntP_Reader), 0, 0, 1, 1},
  {0, __pyx_k_IntP_Reader___reduce_cython, sizeof(__pyx_k_IntP_Reader___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_IntP_Reader___setstate_cython, sizeof(__pyx_k_IntP_Reader___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_KNOWN, sizeof(__pyx_k_KNOWN), 0, 0, 1, 1},
  {0, __pyx_k_LAMBDA, sizeof(__pyx_k_LAMBDA), 0, 0, 1, 1},
  {0, __pyx_k_LAMBDA_TERM, sizeof(__pyx_k_LAMBDA_TERM), 0, 0, 1, 1},
  {0, __pyx_k_LAMBDA_TYPE, sizeof(__pyx_k_LAMBDA_TYPE), 0, 0, 1, 1},
  {0, __pyx_k_LET_IN, sizeof(__pyx_k_LET_IN), 0, 0, 1, 1},
  {0, __pyx_k_LET_IN_DEF, sizeof(__pyx_k_LET_IN_DEF), 0, 0, 1, 1},
  {0, __pyx_k_LET_IN_TERM, sizeof(__pyx_k_LET_IN_TERM), 0, 0, 1, 1},
  {0, __pyx_k_LET_IN_TYPE, sizeof(__pyx_k_LET_IN_TYPE), 0, 0, 1, 1},
  {0, __pyx_k_MANUAL_CONSTANT, sizeof(__pyx_k_MANUAL_CONSTANT), 0, 0, 1, 1},
  {0, __pyx_k_MANUAL_SECTION_CONSTANT, sizeof(__pyx_k_MANUAL_SECTION_CONSTANT), 0, 0, 1, 1},
  {0, __pyx_k_MISMATCHED_ARGUMENTS, sizeof(__pyx_k_MISMATCHED_ARGUMENTS), 0, 0, 1, 1},
  {0, __pyx_k_NEW_STATE, sizeof(__pyx_k_NEW_STATE), 0, 0, 1, 1},
  {0, __pyx_k_NO_SUCH_TACTIC, sizeof(__pyx_k_NO_SUCH_TACTIC), 0, 0, 1, 1},
  {0, __pyx_k_Node, sizeof(__pyx_k_Node), 0, 0, 1, 1},
  {0, __pyx_k_Node_Reader, sizeof(__pyx_k_Node_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Node_Reader_List, sizeof(__pyx_k_Node_Reader_List), 0, 0, 1, 1},
  {0, __pyx_k_Node_Reader_List___reduce_cython, sizeof(__pyx_k_Node_Reader_List___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Node_Reader_List___setstate_cyth, sizeof(__pyx_k_Node_Reader_List___setstate_cyth), 0, 0, 1, 1},
  {0, __pyx_k_Node_Reader___reduce_cython, sizeof(__pyx_k_Node_Reader___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Node_Reader___setstate_cython, sizeof(__pyx_k_Node_Reader___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_ORIGINAL, sizeof(__pyx_k_ORIGINAL), 0, 0, 1, 1},
  {0, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
  {0, __pyx_k_Out_of_bounds, sizeof(__pyx_k_Out_of_bounds), 0, 1, 0, 0},
  {0, __pyx_k_Outcome, sizeof(__pyx_k_Outcome), 0, 0, 1, 1},
  {0, __pyx_k_Outcome_Reader, sizeof(__pyx_k_Outcome_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Outcome_Reader_List, sizeof(__pyx_k_Outcome_Reader_List), 0, 0, 1, 1},
  {0, __pyx_k_Outcome_Reader_List___reduce_cyt, sizeof(__pyx_k_Outcome_Reader_List___reduce_cyt), 0, 0, 1, 1},
  {0, __pyx_k_Outcome_Reader_List___setstate_c, sizeof(__pyx_k_Outcome_Reader_List___setstate_c), 0, 0, 1, 1},
  {0, __pyx_k_Outcome_Reader___reduce_cython, sizeof(__pyx_k_Outcome_Reader___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Outcome_Reader___setstate_cython, sizeof(__pyx_k_Outcome_Reader___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_Outcome_Term_Reader, sizeof(__pyx_k_Outcome_Term_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Outcome_Term_Reader___reduce_cyt, sizeof(__pyx_k_Outcome_Term_Reader___reduce_cyt), 0, 0, 1, 1},
  {0, __pyx_k_Outcome_Term_Reader___setstate_c, sizeof(__pyx_k_Outcome_Term_Reader___setstate_c), 0, 0, 1, 1},
  {0, __pyx_k_PARSE_ERROR, sizeof(__pyx_k_PARSE_ERROR), 0, 0, 1, 1},
  {0, __pyx_k_PREDICT, sizeof(__pyx_k_PREDICT), 0, 0, 1, 1},
  {0, __pyx_k_PREDICTION, sizeof(__pyx_k_PREDICTION), 0, 0, 1, 1},
  {0, __pyx_k_PRIMITIVE, sizeof(__pyx_k_PRIMITIVE), 0, 0, 1, 1},
  {0, __pyx_k_PROD, sizeof(__pyx_k_PROD), 0, 0, 1, 1},
  {0, __pyx_k_PROD_TERM, sizeof(__pyx_k_PROD_TERM), 0, 0, 1, 1},
  {0, __pyx_k_PROD_TYPE, sizeof(__pyx_k_PROD_TYPE), 0, 0, 1, 1},
  {0, __pyx_k_PROJECTION, sizeof(__pyx_k_PROJECTION), 0, 0, 1, 1},
  {0, __pyx_k_PROJ_TERM, sizeof(__pyx_k_PROJ_TERM), 0, 0, 1, 1},
  {0, __pyx_k_PROOF_STATE, sizeof(__pyx_k_PROOF_STATE), 0, 0, 1, 1},
  {0, __pyx_k_PROTOCOL_ERROR, sizeof(__pyx_k_PROTOCOL_ERROR), 0, 0, 1, 1},
  {0, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {0, __pyx_k_Prediction, sizeof(__pyx_k_Prediction), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol, sizeof(__pyx_k_PredictionProtocol), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Prediction_Re, sizeof(__pyx_k_PredictionProtocol_Prediction_Re), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Prediction_Re_2, sizeof(__pyx_k_PredictionProtocol_Prediction_Re_2), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Prediction_Re_3, sizeof(__pyx_k_PredictionProtocol_Prediction_Re_3), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Prediction_Re_4, sizeof(__pyx_k_PredictionProtocol_Prediction_Re_4), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Prediction_Re_5, sizeof(__pyx_k_PredictionProtocol_Prediction_Re_5), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Prediction_Re_6, sizeof(__pyx_k_PredictionProtocol_Prediction_Re_6), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Reader, sizeof(__pyx_k_PredictionProtocol_Reader), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Reader___redu, sizeof(__pyx_k_PredictionProtocol_Reader___redu), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Reader___sets, sizeof(__pyx_k_PredictionProtocol_Reader___sets), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Request_Initi, sizeof(__pyx_k_PredictionProtocol_Request_Initi), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Request_Initi_2, sizeof(__pyx_k_PredictionProtocol_Request_Initi_2), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Request_Initi_3, sizeof(__pyx_k_PredictionProtocol_Request_Initi_3), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Request_Predi, sizeof(__pyx_k_PredictionProtocol_Request_Predi), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Request_Predi_2, sizeof(__pyx_k_PredictionProtocol_Request_Predi_2), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Request_Predi_3, sizeof(__pyx_k_PredictionProtocol_Request_Predi_3), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Request_Reade, sizeof(__pyx_k_PredictionProtocol_Request_Reade), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Request_Reade_2, sizeof(__pyx_k_PredictionProtocol_Request_Reade_2), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Request_Reade_3, sizeof(__pyx_k_PredictionProtocol_Request_Reade_3), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Request_Which, sizeof(__pyx_k_PredictionProtocol_Request_Which), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Response_Alig, sizeof(__pyx_k_PredictionProtocol_Response_Alig), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Response_Alig_2, sizeof(__pyx_k_PredictionProtocol_Response_Alig_2), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Response_Alig_3, sizeof(__pyx_k_PredictionProtocol_Response_Alig_3), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Response_Read, sizeof(__pyx_k_PredictionProtocol_Response_Read), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Response_Read_2, sizeof(__pyx_k_PredictionProtocol_Response_Read_2), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Response_Read_3, sizeof(__pyx_k_PredictionProtocol_Response_Read_3), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_Response_Whic, sizeof(__pyx_k_PredictionProtocol_Response_Whic), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_TextPredictio, sizeof(__pyx_k_PredictionProtocol_TextPredictio), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_TextPredictio_2, sizeof(__pyx_k_PredictionProtocol_TextPredictio_2), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_TextPredictio_3, sizeof(__pyx_k_PredictionProtocol_TextPredictio_3), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_TextPredictio_4, sizeof(__pyx_k_PredictionProtocol_TextPredictio_4), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_TextPredictio_5, sizeof(__pyx_k_PredictionProtocol_TextPredictio_5), 0, 0, 1, 1},
  {0, __pyx_k_PredictionProtocol_TextPredictio_6, sizeof(__pyx_k_PredictionProtocol_TextPredictio_6), 0, 0, 1, 1},
  {0, __pyx_k_ProofState, sizeof(__pyx_k_ProofState), 0, 0, 1, 1},
  {0, __pyx_k_ProofState_Reader, sizeof(__pyx_k_ProofState_Reader), 0, 0, 1, 1},
  {0, __pyx_k_ProofState_Reader_List, sizeof(__pyx_k_ProofState_Reader_List), 0, 0, 1, 1},
  {0, __pyx_k_ProofState_Reader_List___reduce, sizeof(__pyx_k_ProofState_Reader_List___reduce), 0, 0, 1, 1},
  {0, __pyx_k_ProofState_Reader_List___setstat, sizeof(__pyx_k_ProofState_Reader_List___setstat), 0, 0, 1, 1},
  {0, __pyx_k_ProofState_Reader___reduce_cytho, sizeof(__pyx_k_ProofState_Reader___reduce_cytho), 0, 0, 1, 1},
  {0, __pyx_k_ProofState_Reader___setstate_cyt, sizeof(__pyx_k_ProofState_Reader___setstate_cyt), 0, 0, 1, 1},
  {0, __pyx_k_ProofState_Root_Reader, sizeof(__pyx_k_ProofState_Root_Reader), 0, 0, 1, 1},
  {0, __pyx_k_ProofState_Root_Reader___reduce, sizeof(__pyx_k_ProofState_Root_Reader___reduce), 0, 0, 1, 1},
  {0, __pyx_k_ProofState_Root_Reader___setstat, sizeof(__pyx_k_ProofState_Root_Reader___setstat), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep, sizeof(__pyx_k_ProofStep), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep_Reader, sizeof(__pyx_k_ProofStep_Reader), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep_Reader_List, sizeof(__pyx_k_ProofStep_Reader_List), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep_Reader_List___reduce_c, sizeof(__pyx_k_ProofStep_Reader_List___reduce_c), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep_Reader_List___setstate, sizeof(__pyx_k_ProofStep_Reader_List___setstate), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep_Reader___reduce_cython, sizeof(__pyx_k_ProofStep_Reader___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep_Reader___setstate_cyth, sizeof(__pyx_k_ProofStep_Reader___setstate_cyth), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep_Tactic_Reader, sizeof(__pyx_k_ProofStep_Tactic_Reader), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep_Tactic_Reader___reduce, sizeof(__pyx_k_ProofStep_Tactic_Reader___reduce), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep_Tactic_Reader___setsta, sizeof(__pyx_k_ProofStep_Tactic_Reader___setsta), 0, 0, 1, 1},
  {0, __pyx_k_ProofStep_Tactic_Which, sizeof(__pyx_k_ProofStep_Tactic_Which), 0, 0, 1, 1},
  {0, __pyx_k_Pyx_EnumBase, sizeof(__pyx_k_Pyx_EnumBase), 0, 0, 1, 1},
  {0, __pyx_k_Pyx_EnumBase___new, sizeof(__pyx_k_Pyx_EnumBase___new), 0, 0, 1, 1},
  {0, __pyx_k_Pyx_EnumBase___repr, sizeof(__pyx_k_Pyx_EnumBase___repr), 0, 0, 1, 1},
  {0, __pyx_k_Pyx_EnumBase___str, sizeof(__pyx_k_Pyx_EnumBase___str), 0, 0, 1, 1},
  {0, __pyx_k_Pyx_EnumMeta___reduce_cython, sizeof(__pyx_k_Pyx_EnumMeta___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Pyx_EnumMeta___setstate_cython, sizeof(__pyx_k_Pyx_EnumMeta___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_REL, sizeof(__pyx_k_REL), 0, 0, 1, 1},
  {0, __pyx_k_REL_POINTER, sizeof(__pyx_k_REL_POINTER), 0, 0, 1, 1},
  {0, __pyx_k_Request, sizeof(__pyx_k_Request), 0, 0, 1, 1},
  {0, __pyx_k_Response, sizeof(__pyx_k_Response), 0, 0, 1, 1},
  {0, __pyx_k_SORT_PROP, sizeof(__pyx_k_SORT_PROP), 0, 0, 1, 1},
  {0, __pyx_k_SORT_SET, sizeof(__pyx_k_SORT_SET), 0, 0, 1, 1},
  {0, __pyx_k_SORT_S_PROP, sizeof(__pyx_k_SORT_S_PROP), 0, 0, 1, 1},
  {0, __pyx_k_SORT_TYPE, sizeof(__pyx_k_SORT_TYPE), 0, 0, 1, 1},
  {0, __pyx_k_SUBSTITUTED, sizeof(__pyx_k_SUBSTITUTED), 0, 0, 1, 1},
  {0, __pyx_k_SYNCHRONIZE, sizeof(__pyx_k_SYNCHRONIZE), 0, 0, 1, 1},
  {0, __pyx_k_SYNCHRONIZED, sizeof(__pyx_k_SYNCHRONIZED), 0, 0, 1, 1},
  {0, __pyx_k_String_List, sizeof(__pyx_k_String_List), 0, 0, 1, 1},
  {0, __pyx_k_String_List___reduce_cython, sizeof(__pyx_k_String_List___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_String_List___setstate_cython, sizeof(__pyx_k_String_List___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_TACTICAL_CONSTANT, sizeof(__pyx_k_TACTICAL_CONSTANT), 0, 0, 1, 1},
  {0, __pyx_k_TACTICAL_SECTION_CONSTANT, sizeof(__pyx_k_TACTICAL_SECTION_CONSTANT), 0, 0, 1, 1},
  {0, __pyx_k_TERM, sizeof(__pyx_k_TERM), 0, 0, 1, 1},
  {0, __pyx_k_TEXT_PREDICTION, sizeof(__pyx_k_TEXT_PREDICTION), 0, 0, 1, 1},
  {0, __pyx_k_Tactic, sizeof(__pyx_k_Tactic), 0, 0, 1, 1},
  {0, __pyx_k_Tactic_Reader, sizeof(__pyx_k_Tactic_Reader), 0, 0, 1, 1},
  {0, __pyx_k_Tactic_Reader___reduce_cython, sizeof(__pyx_k_Tactic_Reader___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Tactic_Reader___setstate_cython, sizeof(__pyx_k_Tactic_Reader___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_TextPrediction, sizeof(__pyx_k_TextPrediction), 0, 0, 1, 1},
  {0, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {0, __pyx_k_UNKNOWN, sizeof(__pyx_k_UNKNOWN), 0, 0, 1, 1},
  {0, __pyx_k_UNRESOLVABLE, sizeof(__pyx_k_UNRESOLVABLE), 0, 0, 1, 1},
  {0, __pyx_k_Uint16_List, sizeof(__pyx_k_Uint16_List), 0, 0, 1, 1},
  {0, __pyx_k_Uint16_List___reduce_cython, sizeof(__pyx_k_Uint16_List___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Uint16_List___setstate_cython, sizeof(__pyx_k_Uint16_List___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_Uint32_List, sizeof(__pyx_k_Uint32_List), 0, 0, 1, 1},
  {0, __pyx_k_Uint32_List___reduce_cython, sizeof(__pyx_k_Uint32_List___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Uint32_List___setstate_cython, sizeof(__pyx_k_Uint32_List___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_Uint64_List, sizeof(__pyx_k_Uint64_List), 0, 0, 1, 1},
  {0, __pyx_k_Uint64_List___reduce_cython, sizeof(__pyx_k_Uint64_List___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Uint64_List___setstate_cython, sizeof(__pyx_k_Uint64_List___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_Unknown_enum_value_s, sizeof(__pyx_k_Unknown_enum_value_s), 0, 0, 1, 0},
  {0, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {0, __pyx_k__118, sizeof(__pyx_k__118), 0, 0, 1, 1},
  {0, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {0, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 1},
  {0, __pyx_k_as_struct, sizeof(__pyx_k_as_struct), 0, 0, 1, 1},
  {0, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
  {0, __pyx_k_capnp, sizeof(__pyx_k_capnp), 0, 0, 1, 1},
  {0, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {0, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
  {0, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {0, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {0, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {0, __pyx_k_dct, sizeof(__pyx_k_dct), 0, 0, 1, 1},
  {0, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {0, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
  {0, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
  {0, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {0, __pyx_k_dyn, sizeof(__pyx_k_dyn), 0, 0, 1, 1},
  {0, __pyx_k_dynamic, sizeof(__pyx_k_dynamic), 0, 0, 1, 1},
  {0, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
  {0, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
  {0, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
  {0, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {0, __pyx_k_graph_api_capnp, sizeof(__pyx_k_graph_api_capnp), 0, 0, 1, 1},
  {0, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {0, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {0, __pyx_k_init_subclass, sizeof(__pyx_k_init_subclass), 0, 0, 1, 1},
  {0, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
  {0, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
  {0, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
  {0, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {0, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
  {0, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {0, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {0, __pyx_k_mro_entries, sizeof(__pyx_k_mro_entries), 0, 0, 1, 1},
  {0, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {0, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {0, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {0, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {0, __pyx_k_parents, sizeof(__pyx_k_parents), 0, 0, 1, 1},
  {0, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {0, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {0, __pyx_k_pytact, sizeof(__pyx_k_pytact), 0, 0, 1, 1},
  {0, __pyx_k_pytact_graph_api_capnp, sizeof(__pyx_k_pytact_graph_api_capnp), 0, 0, 1, 1},
  {0, __pyx_k_pytact_graph_api_capnp_cython, sizeof(__pyx_k_pytact_graph_api_capnp_cython), 0, 0, 1, 1},
  {0, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {0, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {0, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {0, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {0, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle___Pyx_EnumMeta, sizeof(__pyx_k_pyx_unpickle___Pyx_EnumMeta), 0, 0, 1, 1},
  {0, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {0, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {0, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {0, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {0, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {0, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {0, __pyx_k_s_s, sizeof(__pyx_k_s_s), 0, 0, 1, 0},
  {0, __pyx_k_s_s_d, sizeof(__pyx_k_s_s_d), 0, 0, 1, 0},
  {0, __pyx_k_schema, sizeof(__pyx_k_schema), 0, 0, 1, 1},
  {0, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {0, __pyx_k_self_source_cannot_be_converted, sizeof(__pyx_k_self_source_cannot_be_converted), 0, 0, 1, 0},
  {0, __pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 0, 1, 1},
  {0, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {0, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
  {0, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {0, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {0, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {0, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {0, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {0, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
  {0, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {0, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {0, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  #else
  {&__pyx_n_s_ALIGNMENT, __pyx_k_ALIGNMENT, sizeof(__pyx_k_ALIGNMENT), 0, 0, 1, 1},
  {&__pyx_n_s_APP, __pyx_k_APP, sizeof(__pyx_k_APP), 0, 0, 1, 1},
  {&__pyx_n_s_APP_ARG, __pyx_k_APP_ARG, sizeof(__pyx_k_APP_ARG), 0, 0, 1, 1},
  {&__pyx_n_s_APP_FUN, __pyx_k_APP_FUN, sizeof(__pyx_k_APP_FUN), 0, 0, 1, 1},
  {&__pyx_n_s_AbstractTactic, __pyx_k_AbstractTactic, sizeof(__pyx_k_AbstractTactic), 0, 0, 1, 1},
  {&__pyx_n_s_AbstractTactic_Reader, __pyx_k_AbstractTactic_Reader, sizeof(__pyx_k_AbstractTactic_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_AbstractTactic_Reader_List, __pyx_k_AbstractTactic_Reader_List, sizeof(__pyx_k_AbstractTactic_Reader_List), 0, 0, 1, 1},
  {&__pyx_n_s_AbstractTactic_Reader_List___red, __pyx_k_AbstractTactic_Reader_List___red, sizeof(__pyx_k_AbstractTactic_Reader_List___red), 0, 0, 1, 1},
  {&__pyx_n_s_AbstractTactic_Reader_List___set, __pyx_k_AbstractTactic_Reader_List___set, sizeof(__pyx_k_AbstractTactic_Reader_List___set), 0, 0, 1, 1},
  {&__pyx_n_s_AbstractTactic_Reader___reduce_c, __pyx_k_AbstractTactic_Reader___reduce_c, sizeof(__pyx_k_AbstractTactic_Reader___reduce_c), 0, 0, 1, 1},
  {&__pyx_n_s_AbstractTactic_Reader___setstate, __pyx_k_AbstractTactic_Reader___setstate, sizeof(__pyx_k_AbstractTactic_Reader___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_Argument, __pyx_k_Argument, sizeof(__pyx_k_Argument), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_Reader, __pyx_k_Argument_Reader, sizeof(__pyx_k_Argument_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_Reader_List, __pyx_k_Argument_Reader_List, sizeof(__pyx_k_Argument_Reader_List), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_Reader_List___reduce_cy, __pyx_k_Argument_Reader_List___reduce_cy, sizeof(__pyx_k_Argument_Reader_List___reduce_cy), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_Reader_List___setstate, __pyx_k_Argument_Reader_List___setstate, sizeof(__pyx_k_Argument_Reader_List___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_Reader___reduce_cython, __pyx_k_Argument_Reader___reduce_cython, sizeof(__pyx_k_Argument_Reader___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_Reader___setstate_cytho, __pyx_k_Argument_Reader___setstate_cytho, sizeof(__pyx_k_Argument_Reader___setstate_cytho), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_Term_Reader, __pyx_k_Argument_Term_Reader, sizeof(__pyx_k_Argument_Term_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_Term_Reader___reduce_cy, __pyx_k_Argument_Term_Reader___reduce_cy, sizeof(__pyx_k_Argument_Term_Reader___reduce_cy), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_Term_Reader___setstate, __pyx_k_Argument_Term_Reader___setstate, sizeof(__pyx_k_Argument_Term_Reader___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_Which, __pyx_k_Argument_Which, sizeof(__pyx_k_Argument_Which), 0, 0, 1, 1},
  {&__pyx_n_s_CASE, __pyx_k_CASE, sizeof(__pyx_k_CASE), 0, 0, 1, 1},
  {&__pyx_n_s_CASE_BRANCH, __pyx_k_CASE_BRANCH, sizeof(__pyx_k_CASE_BRANCH), 0, 0, 1, 1},
  {&__pyx_n_s_CASE_BRANCH_POINTER, __pyx_k_CASE_BRANCH_POINTER, sizeof(__pyx_k_CASE_BRANCH_POINTER), 0, 0, 1, 1},
  {&__pyx_n_s_CASE_IND, __pyx_k_CASE_IND, sizeof(__pyx_k_CASE_IND), 0, 0, 1, 1},
  {&__pyx_n_s_CASE_RETURN, __pyx_k_CASE_RETURN, sizeof(__pyx_k_CASE_RETURN), 0, 0, 1, 1},
  {&__pyx_n_s_CASE_TERM, __pyx_k_CASE_TERM, sizeof(__pyx_k_CASE_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_CAST, __pyx_k_CAST, sizeof(__pyx_k_CAST), 0, 0, 1, 1},
  {&__pyx_n_s_CAST_TERM, __pyx_k_CAST_TERM, sizeof(__pyx_k_CAST_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_CAST_TYPE, __pyx_k_CAST_TYPE, sizeof(__pyx_k_CAST_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_CHECK_ALIGNMENT, __pyx_k_CHECK_ALIGNMENT, sizeof(__pyx_k_CHECK_ALIGNMENT), 0, 0, 1, 1},
  {&__pyx_n_s_COMPLETE, __pyx_k_COMPLETE, sizeof(__pyx_k_COMPLETE), 0, 0, 1, 1},
  {&__pyx_n_s_CONSTRUCTOR, __pyx_k_CONSTRUCTOR, sizeof(__pyx_k_CONSTRUCTOR), 0, 0, 1, 1},
  {&__pyx_n_s_CONSTRUCT_TERM, __pyx_k_CONSTRUCT_TERM, sizeof(__pyx_k_CONSTRUCT_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_CONST_DEF, __pyx_k_CONST_DEF, sizeof(__pyx_k_CONST_DEF), 0, 0, 1, 1},
  {&__pyx_n_s_CONST_EMPTY, __pyx_k_CONST_EMPTY, sizeof(__pyx_k_CONST_EMPTY), 0, 0, 1, 1},
  {&__pyx_n_s_CONST_OPAQUE_DEF, __pyx_k_CONST_OPAQUE_DEF, sizeof(__pyx_k_CONST_OPAQUE_DEF), 0, 0, 1, 1},
  {&__pyx_n_s_CONST_PRIMITIVE, __pyx_k_CONST_PRIMITIVE, sizeof(__pyx_k_CONST_PRIMITIVE), 0, 0, 1, 1},
  {&__pyx_n_s_CONST_TYPE, __pyx_k_CONST_TYPE, sizeof(__pyx_k_CONST_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_CONST_UNDEF, __pyx_k_CONST_UNDEF, sizeof(__pyx_k_CONST_UNDEF), 0, 0, 1, 1},
  {&__pyx_n_s_CONTEXT_ASSUM, __pyx_k_CONTEXT_ASSUM, sizeof(__pyx_k_CONTEXT_ASSUM), 0, 0, 1, 1},
  {&__pyx_n_s_CONTEXT_DEF, __pyx_k_CONTEXT_DEF, sizeof(__pyx_k_CONTEXT_DEF), 0, 0, 1, 1},
  {&__pyx_n_s_CONTEXT_DEF_TERM, __pyx_k_CONTEXT_DEF_TERM, sizeof(__pyx_k_CONTEXT_DEF_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_CONTEXT_DEF_TYPE, __pyx_k_CONTEXT_DEF_TYPE, sizeof(__pyx_k_CONTEXT_DEF_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_CONTEXT_ELEM, __pyx_k_CONTEXT_ELEM, sizeof(__pyx_k_CONTEXT_ELEM), 0, 0, 1, 1},
  {&__pyx_n_s_CONTEXT_SUBJECT, __pyx_k_CONTEXT_SUBJECT, sizeof(__pyx_k_CONTEXT_SUBJECT), 0, 0, 1, 1},
  {&__pyx_n_s_CO_FIX, __pyx_k_CO_FIX, sizeof(__pyx_k_CO_FIX), 0, 0, 1, 1},
  {&__pyx_n_s_CO_FIX_FUN, __pyx_k_CO_FIX_FUN, sizeof(__pyx_k_CO_FIX_FUN), 0, 0, 1, 1},
  {&__pyx_n_s_CO_FIX_FUN_TERM, __pyx_k_CO_FIX_FUN_TERM, sizeof(__pyx_k_CO_FIX_FUN_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_CO_FIX_FUN_TYPE, __pyx_k_CO_FIX_FUN_TYPE, sizeof(__pyx_k_CO_FIX_FUN_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_CO_FIX_MUTUAL, __pyx_k_CO_FIX_MUTUAL, sizeof(__pyx_k_CO_FIX_MUTUAL), 0, 0, 1, 1},
  {&__pyx_n_s_CO_FIX_RETURN, __pyx_k_CO_FIX_RETURN, sizeof(__pyx_k_CO_FIX_RETURN), 0, 0, 1, 1},
  {&__pyx_n_s_C_B_CONSTRUCT, __pyx_k_C_B_CONSTRUCT, sizeof(__pyx_k_C_B_CONSTRUCT), 0, 0, 1, 1},
  {&__pyx_n_s_C_B_TERM, __pyx_k_C_B_TERM, sizeof(__pyx_k_C_B_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_ConflatableEdges, __pyx_k_ConflatableEdges, sizeof(__pyx_k_ConflatableEdges), 0, 0, 1, 1},
  {&__pyx_n_s_ConflatableEdges_Reader, __pyx_k_ConflatableEdges_Reader, sizeof(__pyx_k_ConflatableEdges_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_ConflatableEdges_Reader___reduce, __pyx_k_ConflatableEdges_Reader___reduce, sizeof(__pyx_k_ConflatableEdges_Reader___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_ConflatableEdges_Reader___setsta, __pyx_k_ConflatableEdges_Reader___setsta, sizeof(__pyx_k_ConflatableEdges_Reader___setsta), 0, 0, 1, 1},
  {&__pyx_n_s_DEFINITION, __pyx_k_DEFINITION, sizeof(__pyx_k_DEFINITION), 0, 0, 1, 1},
  {&__pyx_n_s_DISCHARGED, __pyx_k_DISCHARGED, sizeof(__pyx_k_DISCHARGED), 0, 0, 1, 1},
  {&__pyx_n_s_DataVersion, __pyx_k_DataVersion, sizeof(__pyx_k_DataVersion), 0, 0, 1, 1},
  {&__pyx_n_s_DataVersion_Reader, __pyx_k_DataVersion_Reader, sizeof(__pyx_k_DataVersion_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_DataVersion_Reader___reduce_cyth, __pyx_k_DataVersion_Reader___reduce_cyth, sizeof(__pyx_k_DataVersion_Reader___reduce_cyth), 0, 0, 1, 1},
  {&__pyx_n_s_DataVersion_Reader___setstate_cy, __pyx_k_DataVersion_Reader___setstate_cy, sizeof(__pyx_k_DataVersion_Reader___setstate_cy), 0, 0, 1, 1},
  {&__pyx_n_s_Dataset, __pyx_k_Dataset, sizeof(__pyx_k_Dataset), 0, 0, 1, 1},
  {&__pyx_n_s_Dataset_Reader, __pyx_k_Dataset_Reader, sizeof(__pyx_k_Dataset_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Dataset_Reader___reduce_cython, __pyx_k_Dataset_Reader___reduce_cython, sizeof(__pyx_k_Dataset_Reader___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Dataset_Reader___setstate_cython, __pyx_k_Dataset_Reader___setstate_cython, sizeof(__pyx_k_Dataset_Reader___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Definition, __pyx_k_Definition, sizeof(__pyx_k_Definition), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Reader, __pyx_k_Definition_Reader, sizeof(__pyx_k_Definition_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Reader___reduce_cytho, __pyx_k_Definition_Reader___reduce_cytho, sizeof(__pyx_k_Definition_Reader___reduce_cytho), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Reader___setstate_cyt, __pyx_k_Definition_Reader___setstate_cyt, sizeof(__pyx_k_Definition_Reader___setstate_cyt), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Status_Reader, __pyx_k_Definition_Status_Reader, sizeof(__pyx_k_Definition_Status_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Status_Reader___reduc, __pyx_k_Definition_Status_Reader___reduc, sizeof(__pyx_k_Definition_Status_Reader___reduc), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Status_Reader___setst, __pyx_k_Definition_Status_Reader___setst, sizeof(__pyx_k_Definition_Status_Reader___setst), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Status_Substituted_Re, __pyx_k_Definition_Status_Substituted_Re, sizeof(__pyx_k_Definition_Status_Substituted_Re), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Status_Substituted_Re_2, __pyx_k_Definition_Status_Substituted_Re_2, sizeof(__pyx_k_Definition_Status_Substituted_Re_2), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Status_Substituted_Re_3, __pyx_k_Definition_Status_Substituted_Re_3, sizeof(__pyx_k_Definition_Status_Substituted_Re_3), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Status_Which, __pyx_k_Definition_Status_Which, sizeof(__pyx_k_Definition_Status_Which), 0, 0, 1, 1},
  {&__pyx_n_s_Definition_Which, __pyx_k_Definition_Which, sizeof(__pyx_k_Definition_Which), 0, 0, 1, 1},
  {&__pyx_n_s_EVAR, __pyx_k_EVAR, sizeof(__pyx_k_EVAR), 0, 0, 1, 1},
  {&__pyx_n_s_EVAR_SUBJECT, __pyx_k_EVAR_SUBJECT, sizeof(__pyx_k_EVAR_SUBJECT), 0, 0, 1, 1},
  {&__pyx_n_s_EVAR_SUBST, __pyx_k_EVAR_SUBST, sizeof(__pyx_k_EVAR_SUBST), 0, 0, 1, 1},
  {&__pyx_n_s_EVAR_SUBST_POINTER, __pyx_k_EVAR_SUBST_POINTER, sizeof(__pyx_k_EVAR_SUBST_POINTER), 0, 0, 1, 1},
  {&__pyx_n_s_EVAR_SUBST_TARGET, __pyx_k_EVAR_SUBST_TARGET, sizeof(__pyx_k_EVAR_SUBST_TARGET), 0, 0, 1, 1},
  {&__pyx_n_s_EVAR_SUBST_TERM, __pyx_k_EVAR_SUBST_TERM, sizeof(__pyx_k_EVAR_SUBST_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_EdgeClassification, __pyx_k_EdgeClassification, sizeof(__pyx_k_EdgeClassification), 0, 0, 1, 1},
  {&__pyx_n_s_EdgeTarget, __pyx_k_EdgeTarget, sizeof(__pyx_k_EdgeTarget), 0, 0, 1, 1},
  {&__pyx_n_s_EnumBase, __pyx_k_EnumBase, sizeof(__pyx_k_EnumBase), 0, 0, 1, 1},
  {&__pyx_n_s_Exception, __pyx_k_Exception, sizeof(__pyx_k_Exception), 0, 0, 1, 1},
  {&__pyx_n_s_Exception_Reader, __pyx_k_Exception_Reader, sizeof(__pyx_k_Exception_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Exception_Reader___reduce_cython, __pyx_k_Exception_Reader___reduce_cython, sizeof(__pyx_k_Exception_Reader___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Exception_Reader___setstate_cyth, __pyx_k_Exception_Reader___setstate_cyth, sizeof(__pyx_k_Exception_Reader___setstate_cyth), 0, 0, 1, 1},
  {&__pyx_n_s_Exception_Which, __pyx_k_Exception_Which, sizeof(__pyx_k_Exception_Which), 0, 0, 1, 1},
  {&__pyx_n_s_ExecutionResult, __pyx_k_ExecutionResult, sizeof(__pyx_k_ExecutionResult), 0, 0, 1, 1},
  {&__pyx_n_s_ExecutionResult_NewState_Reader, __pyx_k_ExecutionResult_NewState_Reader, sizeof(__pyx_k_ExecutionResult_NewState_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_ExecutionResult_NewState_Reader_2, __pyx_k_ExecutionResult_NewState_Reader_2, sizeof(__pyx_k_ExecutionResult_NewState_Reader_2), 0, 0, 1, 1},
  {&__pyx_n_s_ExecutionResult_NewState_Reader_3, __pyx_k_ExecutionResult_NewState_Reader_3, sizeof(__pyx_k_ExecutionResult_NewState_Reader_3), 0, 0, 1, 1},
  {&__pyx_n_s_ExecutionResult_Reader, __pyx_k_ExecutionResult_Reader, sizeof(__pyx_k_ExecutionResult_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_ExecutionResult_Reader___reduce, __pyx_k_ExecutionResult_Reader___reduce, sizeof(__pyx_k_ExecutionResult_Reader___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_ExecutionResult_Reader___setstat, __pyx_k_ExecutionResult_Reader___setstat, sizeof(__pyx_k_ExecutionResult_Reader___setstat), 0, 0, 1, 1},
  {&__pyx_n_s_ExecutionResult_Which, __pyx_k_ExecutionResult_Which, sizeof(__pyx_k_ExecutionResult_Which), 0, 0, 1, 1},
  {&__pyx_n_s_FAILURE, __pyx_k_FAILURE, sizeof(__pyx_k_FAILURE), 0, 0, 1, 1},
  {&__pyx_n_s_FIX, __pyx_k_FIX, sizeof(__pyx_k_FIX), 0, 0, 1, 1},
  {&__pyx_n_s_FIX_FUN, __pyx_k_FIX_FUN, sizeof(__pyx_k_FIX_FUN), 0, 0, 1, 1},
  {&__pyx_n_s_FIX_FUN_TERM, __pyx_k_FIX_FUN_TERM, sizeof(__pyx_k_FIX_FUN_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_FIX_FUN_TYPE, __pyx_k_FIX_FUN_TYPE, sizeof(__pyx_k_FIX_FUN_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_FIX_MUTUAL, __pyx_k_FIX_MUTUAL, sizeof(__pyx_k_FIX_MUTUAL), 0, 0, 1, 1},
  {&__pyx_n_s_FIX_RETURN, __pyx_k_FIX_RETURN, sizeof(__pyx_k_FIX_RETURN), 0, 0, 1, 1},
  {&__pyx_n_s_FLOAT, __pyx_k_FLOAT, sizeof(__pyx_k_FLOAT), 0, 0, 1, 1},
  {&__pyx_n_s_FloatP, __pyx_k_FloatP, sizeof(__pyx_k_FloatP), 0, 0, 1, 1},
  {&__pyx_n_s_FloatP_Reader, __pyx_k_FloatP_Reader, sizeof(__pyx_k_FloatP_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_FloatP_Reader___reduce_cython, __pyx_k_FloatP_Reader___reduce_cython, sizeof(__pyx_k_FloatP_Reader___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_FloatP_Reader___setstate_cython, __pyx_k_FloatP_Reader___setstate_cython, sizeof(__pyx_k_FloatP_Reader___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Graph, __pyx_k_Graph, sizeof(__pyx_k_Graph), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_EdgeTarget_Reader, __pyx_k_Graph_EdgeTarget_Reader, sizeof(__pyx_k_Graph_EdgeTarget_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_EdgeTarget_Reader_List, __pyx_k_Graph_EdgeTarget_Reader_List, sizeof(__pyx_k_Graph_EdgeTarget_Reader_List), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_EdgeTarget_Reader_List___r, __pyx_k_Graph_EdgeTarget_Reader_List___r, sizeof(__pyx_k_Graph_EdgeTarget_Reader_List___r), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_EdgeTarget_Reader_List___s, __pyx_k_Graph_EdgeTarget_Reader_List___s, sizeof(__pyx_k_Graph_EdgeTarget_Reader_List___s), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_EdgeTarget_Reader___reduce, __pyx_k_Graph_EdgeTarget_Reader___reduce, sizeof(__pyx_k_Graph_EdgeTarget_Reader___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_EdgeTarget_Reader___setsta, __pyx_k_Graph_EdgeTarget_Reader___setsta, sizeof(__pyx_k_Graph_EdgeTarget_Reader___setsta), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_EdgeTarget_Target_Reader, __pyx_k_Graph_EdgeTarget_Target_Reader, sizeof(__pyx_k_Graph_EdgeTarget_Target_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_EdgeTarget_Target_Reader_2, __pyx_k_Graph_EdgeTarget_Target_Reader_2, sizeof(__pyx_k_Graph_EdgeTarget_Target_Reader_2), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_EdgeTarget_Target_Reader_3, __pyx_k_Graph_EdgeTarget_Target_Reader_3, sizeof(__pyx_k_Graph_EdgeTarget_Target_Reader_3), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Node_Label_Reader, __pyx_k_Graph_Node_Label_Reader, sizeof(__pyx_k_Graph_Node_Label_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Node_Label_Reader___reduce, __pyx_k_Graph_Node_Label_Reader___reduce, sizeof(__pyx_k_Graph_Node_Label_Reader___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Node_Label_Reader___setsta, __pyx_k_Graph_Node_Label_Reader___setsta, sizeof(__pyx_k_Graph_Node_Label_Reader___setsta), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Node_Label_Which, __pyx_k_Graph_Node_Label_Which, sizeof(__pyx_k_Graph_Node_Label_Which), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Node_Reader, __pyx_k_Graph_Node_Reader, sizeof(__pyx_k_Graph_Node_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Node_Reader_List, __pyx_k_Graph_Node_Reader_List, sizeof(__pyx_k_Graph_Node_Reader_List), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Node_Reader_List___reduce, __pyx_k_Graph_Node_Reader_List___reduce, sizeof(__pyx_k_Graph_Node_Reader_List___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Node_Reader_List___setstat, __pyx_k_Graph_Node_Reader_List___setstat, sizeof(__pyx_k_Graph_Node_Reader_List___setstat), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Node_Reader___reduce_cytho, __pyx_k_Graph_Node_Reader___reduce_cytho, sizeof(__pyx_k_Graph_Node_Reader___reduce_cytho), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Node_Reader___setstate_cyt, __pyx_k_Graph_Node_Reader___setstate_cyt, sizeof(__pyx_k_Graph_Node_Reader___setstate_cyt), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Reader, __pyx_k_Graph_Reader, sizeof(__pyx_k_Graph_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Reader___reduce_cython, __pyx_k_Graph_Reader___reduce_cython, sizeof(__pyx_k_Graph_Reader___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Graph_Reader___setstate_cython, __pyx_k_Graph_Reader___setstate_cython, sizeof(__pyx_k_Graph_Reader___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_ILLEGAL_ARGUMENT, __pyx_k_ILLEGAL_ARGUMENT, sizeof(__pyx_k_ILLEGAL_ARGUMENT), 0, 0, 1, 1},
  {&__pyx_n_s_INDUCTIVE, __pyx_k_INDUCTIVE, sizeof(__pyx_k_INDUCTIVE), 0, 0, 1, 1},
  {&__pyx_n_s_IND_CONSTRUCT, __pyx_k_IND_CONSTRUCT, sizeof(__pyx_k_IND_CONSTRUCT), 0, 0, 1, 1},
  {&__pyx_n_s_IND_PROJECTION, __pyx_k_IND_PROJECTION, sizeof(__pyx_k_IND_PROJECTION), 0, 0, 1, 1},
  {&__pyx_n_s_IND_TYPE, __pyx_k_IND_TYPE, sizeof(__pyx_k_IND_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_INITIALIZE, __pyx_k_INITIALIZE, sizeof(__pyx_k_INITIALIZE), 0, 0, 1, 1},
  {&__pyx_n_s_INITIALIZED, __pyx_k_INITIALIZED, sizeof(__pyx_k_INITIALIZED), 0, 0, 1, 1},
  {&__pyx_n_s_INT, __pyx_k_INT, sizeof(__pyx_k_INT), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_Int64_List, __pyx_k_Int64_List, sizeof(__pyx_k_Int64_List), 0, 0, 1, 1},
  {&__pyx_n_s_Int64_List___reduce_cython, __pyx_k_Int64_List___reduce_cython, sizeof(__pyx_k_Int64_List___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Int64_List___setstate_cython, __pyx_k_Int64_List___setstate_cython, sizeof(__pyx_k_Int64_List___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_IntEnum, __pyx_k_IntEnum, sizeof(__pyx_k_IntEnum), 0, 0, 1, 1},
  {&__pyx_n_s_IntP, __pyx_k_IntP, sizeof(__pyx_k_IntP), 0, 0, 1, 1},
  {&__pyx_n_s_IntP_Reader, __pyx_k_IntP_Reader, sizeof(__pyx_k_IntP_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_IntP_Reader___reduce_cython, __pyx_k_IntP_Reader___reduce_cython, sizeof(__pyx_k_IntP_Reader___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_IntP_Reader___setstate_cython, __pyx_k_IntP_Reader___setstate_cython, sizeof(__pyx_k_IntP_Reader___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_KNOWN, __pyx_k_KNOWN, sizeof(__pyx_k_KNOWN), 0, 0, 1, 1},
  {&__pyx_n_s_LAMBDA, __pyx_k_LAMBDA, sizeof(__pyx_k_LAMBDA), 0, 0, 1, 1},
  {&__pyx_n_s_LAMBDA_TERM, __pyx_k_LAMBDA_TERM, sizeof(__pyx_k_LAMBDA_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_LAMBDA_TYPE, __pyx_k_LAMBDA_TYPE, sizeof(__pyx_k_LAMBDA_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_LET_IN, __pyx_k_LET_IN, sizeof(__pyx_k_LET_IN), 0, 0, 1, 1},
  {&__pyx_n_s_LET_IN_DEF, __pyx_k_LET_IN_DEF, sizeof(__pyx_k_LET_IN_DEF), 0, 0, 1, 1},
  {&__pyx_n_s_LET_IN_TERM, __pyx_k_LET_IN_TERM, sizeof(__pyx_k_LET_IN_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_LET_IN_TYPE, __pyx_k_LET_IN_TYPE, sizeof(__pyx_k_LET_IN_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_MANUAL_CONSTANT, __pyx_k_MANUAL_CONSTANT, sizeof(__pyx_k_MANUAL_CONSTANT), 0, 0, 1, 1},
  {&__pyx_n_s_MANUAL_SECTION_CONSTANT, __pyx_k_MANUAL_SECTION_CONSTANT, sizeof(__pyx_k_MANUAL_SECTION_CONSTANT), 0, 0, 1, 1},
  {&__pyx_n_s_MISMATCHED_ARGUMENTS, __pyx_k_MISMATCHED_ARGUMENTS, sizeof(__pyx_k_MISMATCHED_ARGUMENTS), 0, 0, 1, 1},
  {&__pyx_n_s_NEW_STATE, __pyx_k_NEW_STATE, sizeof(__pyx_k_NEW_STATE), 0, 0, 1, 1},
  {&__pyx_n_s_NO_SUCH_TACTIC, __pyx_k_NO_SUCH_TACTIC, sizeof(__pyx_k_NO_SUCH_TACTIC), 0, 0, 1, 1},
  {&__pyx_n_s_Node, __pyx_k_Node, sizeof(__pyx_k_Node), 0, 0, 1, 1},
  {&__pyx_n_s_Node_Reader, __pyx_k_Node_Reader, sizeof(__pyx_k_Node_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Node_Reader_List, __pyx_k_Node_Reader_List, sizeof(__pyx_k_Node_Reader_List), 0, 0, 1, 1},
  {&__pyx_n_s_Node_Reader_List___reduce_cython, __pyx_k_Node_Reader_List___reduce_cython, sizeof(__pyx_k_Node_Reader_List___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Node_Reader_List___setstate_cyth, __pyx_k_Node_Reader_List___setstate_cyth, sizeof(__pyx_k_Node_Reader_List___setstate_cyth), 0, 0, 1, 1},
  {&__pyx_n_s_Node_Reader___reduce_cython, __pyx_k_Node_Reader___reduce_cython, sizeof(__pyx_k_Node_Reader___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Node_Reader___setstate_cython, __pyx_k_Node_Reader___setstate_cython, sizeof(__pyx_k_Node_Reader___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_ORIGINAL, __pyx_k_ORIGINAL, sizeof(__pyx_k_ORIGINAL), 0, 0, 1, 1},
  {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
  {&__pyx_kp_u_Out_of_bounds, __pyx_k_Out_of_bounds, sizeof(__pyx_k_Out_of_bounds), 0, 1, 0, 0},
  {&__pyx_n_s_Outcome, __pyx_k_Outcome, sizeof(__pyx_k_Outcome), 0, 0, 1, 1},
  {&__pyx_n_s_Outcome_Reader, __pyx_k_Outcome_Reader, sizeof(__pyx_k_Outcome_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Outcome_Reader_List, __pyx_k_Outcome_Reader_List, sizeof(__pyx_k_Outcome_Reader_List), 0, 0, 1, 1},
  {&__pyx_n_s_Outcome_Reader_List___reduce_cyt, __pyx_k_Outcome_Reader_List___reduce_cyt, sizeof(__pyx_k_Outcome_Reader_List___reduce_cyt), 0, 0, 1, 1},
  {&__pyx_n_s_Outcome_Reader_List___setstate_c, __pyx_k_Outcome_Reader_List___setstate_c, sizeof(__pyx_k_Outcome_Reader_List___setstate_c), 0, 0, 1, 1},
  {&__pyx_n_s_Outcome_Reader___reduce_cython, __pyx_k_Outcome_Reader___reduce_cython, sizeof(__pyx_k_Outcome_Reader___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Outcome_Reader___setstate_cython, __pyx_k_Outcome_Reader___setstate_cython, sizeof(__pyx_k_Outcome_Reader___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Outcome_Term_Reader, __pyx_k_Outcome_Term_Reader, sizeof(__pyx_k_Outcome_Term_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Outcome_Term_Reader___reduce_cyt, __pyx_k_Outcome_Term_Reader___reduce_cyt, sizeof(__pyx_k_Outcome_Term_Reader___reduce_cyt), 0, 0, 1, 1},
  {&__pyx_n_s_Outcome_Term_Reader___setstate_c, __pyx_k_Outcome_Term_Reader___setstate_c, sizeof(__pyx_k_Outcome_Term_Reader___setstate_c), 0, 0, 1, 1},
  {&__pyx_n_s_PARSE_ERROR, __pyx_k_PARSE_ERROR, sizeof(__pyx_k_PARSE_ERROR), 0, 0, 1, 1},
  {&__pyx_n_s_PREDICT, __pyx_k_PREDICT, sizeof(__pyx_k_PREDICT), 0, 0, 1, 1},
  {&__pyx_n_s_PREDICTION, __pyx_k_PREDICTION, sizeof(__pyx_k_PREDICTION), 0, 0, 1, 1},
  {&__pyx_n_s_PRIMITIVE, __pyx_k_PRIMITIVE, sizeof(__pyx_k_PRIMITIVE), 0, 0, 1, 1},
  {&__pyx_n_s_PROD, __pyx_k_PROD, sizeof(__pyx_k_PROD), 0, 0, 1, 1},
  {&__pyx_n_s_PROD_TERM, __pyx_k_PROD_TERM, sizeof(__pyx_k_PROD_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_PROD_TYPE, __pyx_k_PROD_TYPE, sizeof(__pyx_k_PROD_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_PROJECTION, __pyx_k_PROJECTION, sizeof(__pyx_k_PROJECTION), 0, 0, 1, 1},
  {&__pyx_n_s_PROJ_TERM, __pyx_k_PROJ_TERM, sizeof(__pyx_k_PROJ_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_PROOF_STATE, __pyx_k_PROOF_STATE, sizeof(__pyx_k_PROOF_STATE), 0, 0, 1, 1},
  {&__pyx_n_s_PROTOCOL_ERROR, __pyx_k_PROTOCOL_ERROR, sizeof(__pyx_k_PROTOCOL_ERROR), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_Prediction, __pyx_k_Prediction, sizeof(__pyx_k_Prediction), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol, __pyx_k_PredictionProtocol, sizeof(__pyx_k_PredictionProtocol), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Prediction_Re, __pyx_k_PredictionProtocol_Prediction_Re, sizeof(__pyx_k_PredictionProtocol_Prediction_Re), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Prediction_Re_2, __pyx_k_PredictionProtocol_Prediction_Re_2, sizeof(__pyx_k_PredictionProtocol_Prediction_Re_2), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Prediction_Re_3, __pyx_k_PredictionProtocol_Prediction_Re_3, sizeof(__pyx_k_PredictionProtocol_Prediction_Re_3), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Prediction_Re_4, __pyx_k_PredictionProtocol_Prediction_Re_4, sizeof(__pyx_k_PredictionProtocol_Prediction_Re_4), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Prediction_Re_5, __pyx_k_PredictionProtocol_Prediction_Re_5, sizeof(__pyx_k_PredictionProtocol_Prediction_Re_5), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Prediction_Re_6, __pyx_k_PredictionProtocol_Prediction_Re_6, sizeof(__pyx_k_PredictionProtocol_Prediction_Re_6), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Reader, __pyx_k_PredictionProtocol_Reader, sizeof(__pyx_k_PredictionProtocol_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Reader___redu, __pyx_k_PredictionProtocol_Reader___redu, sizeof(__pyx_k_PredictionProtocol_Reader___redu), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Reader___sets, __pyx_k_PredictionProtocol_Reader___sets, sizeof(__pyx_k_PredictionProtocol_Reader___sets), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Request_Initi, __pyx_k_PredictionProtocol_Request_Initi, sizeof(__pyx_k_PredictionProtocol_Request_Initi), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Request_Initi_2, __pyx_k_PredictionProtocol_Request_Initi_2, sizeof(__pyx_k_PredictionProtocol_Request_Initi_2), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Request_Initi_3, __pyx_k_PredictionProtocol_Request_Initi_3, sizeof(__pyx_k_PredictionProtocol_Request_Initi_3), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Request_Predi, __pyx_k_PredictionProtocol_Request_Predi, sizeof(__pyx_k_PredictionProtocol_Request_Predi), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Request_Predi_2, __pyx_k_PredictionProtocol_Request_Predi_2, sizeof(__pyx_k_PredictionProtocol_Request_Predi_2), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Request_Predi_3, __pyx_k_PredictionProtocol_Request_Predi_3, sizeof(__pyx_k_PredictionProtocol_Request_Predi_3), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Request_Reade, __pyx_k_PredictionProtocol_Request_Reade, sizeof(__pyx_k_PredictionProtocol_Request_Reade), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Request_Reade_2, __pyx_k_PredictionProtocol_Request_Reade_2, sizeof(__pyx_k_PredictionProtocol_Request_Reade_2), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Request_Reade_3, __pyx_k_PredictionProtocol_Request_Reade_3, sizeof(__pyx_k_PredictionProtocol_Request_Reade_3), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Request_Which, __pyx_k_PredictionProtocol_Request_Which, sizeof(__pyx_k_PredictionProtocol_Request_Which), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Response_Alig, __pyx_k_PredictionProtocol_Response_Alig, sizeof(__pyx_k_PredictionProtocol_Response_Alig), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Response_Alig_2, __pyx_k_PredictionProtocol_Response_Alig_2, sizeof(__pyx_k_PredictionProtocol_Response_Alig_2), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Response_Alig_3, __pyx_k_PredictionProtocol_Response_Alig_3, sizeof(__pyx_k_PredictionProtocol_Response_Alig_3), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Response_Read, __pyx_k_PredictionProtocol_Response_Read, sizeof(__pyx_k_PredictionProtocol_Response_Read), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Response_Read_2, __pyx_k_PredictionProtocol_Response_Read_2, sizeof(__pyx_k_PredictionProtocol_Response_Read_2), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Response_Read_3, __pyx_k_PredictionProtocol_Response_Read_3, sizeof(__pyx_k_PredictionProtocol_Response_Read_3), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_Response_Whic, __pyx_k_PredictionProtocol_Response_Whic, sizeof(__pyx_k_PredictionProtocol_Response_Whic), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_TextPredictio, __pyx_k_PredictionProtocol_TextPredictio, sizeof(__pyx_k_PredictionProtocol_TextPredictio), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_TextPredictio_2, __pyx_k_PredictionProtocol_TextPredictio_2, sizeof(__pyx_k_PredictionProtocol_TextPredictio_2), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_TextPredictio_3, __pyx_k_PredictionProtocol_TextPredictio_3, sizeof(__pyx_k_PredictionProtocol_TextPredictio_3), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_TextPredictio_4, __pyx_k_PredictionProtocol_TextPredictio_4, sizeof(__pyx_k_PredictionProtocol_TextPredictio_4), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_TextPredictio_5, __pyx_k_PredictionProtocol_TextPredictio_5, sizeof(__pyx_k_PredictionProtocol_TextPredictio_5), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionProtocol_TextPredictio_6, __pyx_k_PredictionProtocol_TextPredictio_6, sizeof(__pyx_k_PredictionProtocol_TextPredictio_6), 0, 0, 1, 1},
  {&__pyx_n_s_ProofState, __pyx_k_ProofState, sizeof(__pyx_k_ProofState), 0, 0, 1, 1},
  {&__pyx_n_s_ProofState_Reader, __pyx_k_ProofState_Reader, sizeof(__pyx_k_ProofState_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_ProofState_Reader_List, __pyx_k_ProofState_Reader_List, sizeof(__pyx_k_ProofState_Reader_List), 0, 0, 1, 1},
  {&__pyx_n_s_ProofState_Reader_List___reduce, __pyx_k_ProofState_Reader_List___reduce, sizeof(__pyx_k_ProofState_Reader_List___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_ProofState_Reader_List___setstat, __pyx_k_ProofState_Reader_List___setstat, sizeof(__pyx_k_ProofState_Reader_List___setstat), 0, 0, 1, 1},
  {&__pyx_n_s_ProofState_Reader___reduce_cytho, __pyx_k_ProofState_Reader___reduce_cytho, sizeof(__pyx_k_ProofState_Reader___reduce_cytho), 0, 0, 1, 1},
  {&__pyx_n_s_ProofState_Reader___setstate_cyt, __pyx_k_ProofState_Reader___setstate_cyt, sizeof(__pyx_k_ProofState_Reader___setstate_cyt), 0, 0, 1, 1},
  {&__pyx_n_s_ProofState_Root_Reader, __pyx_k_ProofState_Root_Reader, sizeof(__pyx_k_ProofState_Root_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_ProofState_Root_Reader___reduce, __pyx_k_ProofState_Root_Reader___reduce, sizeof(__pyx_k_ProofState_Root_Reader___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_ProofState_Root_Reader___setstat, __pyx_k_ProofState_Root_Reader___setstat, sizeof(__pyx_k_ProofState_Root_Reader___setstat), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep, __pyx_k_ProofStep, sizeof(__pyx_k_ProofStep), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep_Reader, __pyx_k_ProofStep_Reader, sizeof(__pyx_k_ProofStep_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep_Reader_List, __pyx_k_ProofStep_Reader_List, sizeof(__pyx_k_ProofStep_Reader_List), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep_Reader_List___reduce_c, __pyx_k_ProofStep_Reader_List___reduce_c, sizeof(__pyx_k_ProofStep_Reader_List___reduce_c), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep_Reader_List___setstate, __pyx_k_ProofStep_Reader_List___setstate, sizeof(__pyx_k_ProofStep_Reader_List___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep_Reader___reduce_cython, __pyx_k_ProofStep_Reader___reduce_cython, sizeof(__pyx_k_ProofStep_Reader___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep_Reader___setstate_cyth, __pyx_k_ProofStep_Reader___setstate_cyth, sizeof(__pyx_k_ProofStep_Reader___setstate_cyth), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep_Tactic_Reader, __pyx_k_ProofStep_Tactic_Reader, sizeof(__pyx_k_ProofStep_Tactic_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep_Tactic_Reader___reduce, __pyx_k_ProofStep_Tactic_Reader___reduce, sizeof(__pyx_k_ProofStep_Tactic_Reader___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep_Tactic_Reader___setsta, __pyx_k_ProofStep_Tactic_Reader___setsta, sizeof(__pyx_k_ProofStep_Tactic_Reader___setsta), 0, 0, 1, 1},
  {&__pyx_n_s_ProofStep_Tactic_Which, __pyx_k_ProofStep_Tactic_Which, sizeof(__pyx_k_ProofStep_Tactic_Which), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase, __pyx_k_Pyx_EnumBase, sizeof(__pyx_k_Pyx_EnumBase), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___new, __pyx_k_Pyx_EnumBase___new, sizeof(__pyx_k_Pyx_EnumBase___new), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___repr, __pyx_k_Pyx_EnumBase___repr, sizeof(__pyx_k_Pyx_EnumBase___repr), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___str, __pyx_k_Pyx_EnumBase___str, sizeof(__pyx_k_Pyx_EnumBase___str), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumMeta___reduce_cython, __pyx_k_Pyx_EnumMeta___reduce_cython, sizeof(__pyx_k_Pyx_EnumMeta___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumMeta___setstate_cython, __pyx_k_Pyx_EnumMeta___setstate_cython, sizeof(__pyx_k_Pyx_EnumMeta___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_REL, __pyx_k_REL, sizeof(__pyx_k_REL), 0, 0, 1, 1},
  {&__pyx_n_s_REL_POINTER, __pyx_k_REL_POINTER, sizeof(__pyx_k_REL_POINTER), 0, 0, 1, 1},
  {&__pyx_n_s_Request, __pyx_k_Request, sizeof(__pyx_k_Request), 0, 0, 1, 1},
  {&__pyx_n_s_Response, __pyx_k_Response, sizeof(__pyx_k_Response), 0, 0, 1, 1},
  {&__pyx_n_s_SORT_PROP, __pyx_k_SORT_PROP, sizeof(__pyx_k_SORT_PROP), 0, 0, 1, 1},
  {&__pyx_n_s_SORT_SET, __pyx_k_SORT_SET, sizeof(__pyx_k_SORT_SET), 0, 0, 1, 1},
  {&__pyx_n_s_SORT_S_PROP, __pyx_k_SORT_S_PROP, sizeof(__pyx_k_SORT_S_PROP), 0, 0, 1, 1},
  {&__pyx_n_s_SORT_TYPE, __pyx_k_SORT_TYPE, sizeof(__pyx_k_SORT_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_SUBSTITUTED, __pyx_k_SUBSTITUTED, sizeof(__pyx_k_SUBSTITUTED), 0, 0, 1, 1},
  {&__pyx_n_s_SYNCHRONIZE, __pyx_k_SYNCHRONIZE, sizeof(__pyx_k_SYNCHRONIZE), 0, 0, 1, 1},
  {&__pyx_n_s_SYNCHRONIZED, __pyx_k_SYNCHRONIZED, sizeof(__pyx_k_SYNCHRONIZED), 0, 0, 1, 1},
  {&__pyx_n_s_String_List, __pyx_k_String_List, sizeof(__pyx_k_String_List), 0, 0, 1, 1},
  {&__pyx_n_s_String_List___reduce_cython, __pyx_k_String_List___reduce_cython, sizeof(__pyx_k_String_List___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_String_List___setstate_cython, __pyx_k_String_List___setstate_cython, sizeof(__pyx_k_String_List___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_TACTICAL_CONSTANT, __pyx_k_TACTICAL_CONSTANT, sizeof(__pyx_k_TACTICAL_CONSTANT), 0, 0, 1, 1},
  {&__pyx_n_s_TACTICAL_SECTION_CONSTANT, __pyx_k_TACTICAL_SECTION_CONSTANT, sizeof(__pyx_k_TACTICAL_SECTION_CONSTANT), 0, 0, 1, 1},
  {&__pyx_n_s_TERM, __pyx_k_TERM, sizeof(__pyx_k_TERM), 0, 0, 1, 1},
  {&__pyx_n_s_TEXT_PREDICTION, __pyx_k_TEXT_PREDICTION, sizeof(__pyx_k_TEXT_PREDICTION), 0, 0, 1, 1},
  {&__pyx_n_s_Tactic, __pyx_k_Tactic, sizeof(__pyx_k_Tactic), 0, 0, 1, 1},
  {&__pyx_n_s_Tactic_Reader, __pyx_k_Tactic_Reader, sizeof(__pyx_k_Tactic_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_Tactic_Reader___reduce_cython, __pyx_k_Tactic_Reader___reduce_cython, sizeof(__pyx_k_Tactic_Reader___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Tactic_Reader___setstate_cython, __pyx_k_Tactic_Reader___setstate_cython, sizeof(__pyx_k_Tactic_Reader___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_TextPrediction, __pyx_k_TextPrediction, sizeof(__pyx_k_TextPrediction), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN, __pyx_k_UNKNOWN, sizeof(__pyx_k_UNKNOWN), 0, 0, 1, 1},
  {&__pyx_n_s_UNRESOLVABLE, __pyx_k_UNRESOLVABLE, sizeof(__pyx_k_UNRESOLVABLE), 0, 0, 1, 1},
  {&__pyx_n_s_Uint16_List, __pyx_k_Uint16_List, sizeof(__pyx_k_Uint16_List), 0, 0, 1, 1},
  {&__pyx_n_s_Uint16_List___reduce_cython, __pyx_k_Uint16_List___reduce_cython, sizeof(__pyx_k_Uint16_List___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Uint16_List___setstate_cython, __pyx_k_Uint16_List___setstate_cython, sizeof(__pyx_k_Uint16_List___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Uint32_List, __pyx_k_Uint32_List, sizeof(__pyx_k_Uint32_List), 0, 0, 1, 1},
  {&__pyx_n_s_Uint32_List___reduce_cython, __pyx_k_Uint32_List___reduce_cython, sizeof(__pyx_k_Uint32_List___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Uint32_List___setstate_cython, __pyx_k_Uint32_List___setstate_cython, sizeof(__pyx_k_Uint32_List___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64_List, __pyx_k_Uint64_List, sizeof(__pyx_k_Uint64_List), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64_List___reduce_cython, __pyx_k_Uint64_List___reduce_cython, sizeof(__pyx_k_Uint64_List___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64_List___setstate_cython, __pyx_k_Uint64_List___setstate_cython, sizeof(__pyx_k_Uint64_List___setstate_cython), 0, 0, 1, 1},
  {&__pyx_kp_s_Unknown_enum_value_s, __pyx_k_Unknown_enum_value_s, sizeof(__pyx_k_Unknown_enum_value_s), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s__118, __pyx_k__118, sizeof(__pyx_k__118), 0, 0, 1, 1},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_n_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 1},
  {&__pyx_n_s_as_struct, __pyx_k_as_struct, sizeof(__pyx_k_as_struct), 0, 0, 1, 1},
  {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
  {&__pyx_n_s_capnp, __pyx_k_capnp, sizeof(__pyx_k_capnp), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_dct, __pyx_k_dct, sizeof(__pyx_k_dct), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
  {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dyn, __pyx_k_dyn, sizeof(__pyx_k_dyn), 0, 0, 1, 1},
  {&__pyx_n_s_dynamic, __pyx_k_dynamic, sizeof(__pyx_k_dynamic), 0, 0, 1, 1},
  {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
  {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
  {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_graph_api_capnp, __pyx_k_graph_api_capnp, sizeof(__pyx_k_graph_api_capnp), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_init_subclass, __pyx_k_init_subclass, sizeof(__pyx_k_init_subclass), 0, 0, 1, 1},
  {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
  {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
  {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_members, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_mro_entries, __pyx_k_mro_entries, sizeof(__pyx_k_mro_entries), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_parents, __pyx_k_parents, sizeof(__pyx_k_parents), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_pytact, __pyx_k_pytact, sizeof(__pyx_k_pytact), 0, 0, 1, 1},
  {&__pyx_n_s_pytact_graph_api_capnp, __pyx_k_pytact_graph_api_capnp, sizeof(__pyx_k_pytact_graph_api_capnp), 0, 0, 1, 1},
  {&__pyx_n_s_pytact_graph_api_capnp_cython, __pyx_k_pytact_graph_api_capnp_cython, sizeof(__pyx_k_pytact_graph_api_capnp_cython), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle___Pyx_EnumMeta, __pyx_k_pyx_unpickle___Pyx_EnumMeta, sizeof(__pyx_k_pyx_unpickle___Pyx_EnumMeta), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_kp_s_s_s, __pyx_k_s_s, sizeof(__pyx_k_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_d, __pyx_k_s_s_d, sizeof(__pyx_k_s_s_d), 0, 0, 1, 0},
  {&__pyx_n_s_schema, __pyx_k_schema, sizeof(__pyx_k_schema), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self_source_cannot_be_converted, __pyx_k_self_source_cannot_be_converted, sizeof(__pyx_k_self_source_cannot_be_converted), 0, 0, 1, 0},
  {&__pyx_n_s_set_name, __pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  #endif
  {0, 0, 0, 0, 0, 0, 0}
};
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(1, 2086, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 29, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pytact/graph_api_capnp_cython.pyx":2086
 *         source = self.source
 *         if index >= source.size():
 *             raise IndexError('Out of bounds')             # <<<<<<<<<<<<<<
 *         return source[index]
 * 
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_Out_of_bounds); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 2086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_tuple__2 = PyTuple_Pack(3, __pyx_int_238750788, __pyx_int_228825662, __pyx_int_222419149); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "pytact/graph_api_capnp_cython.pyx":11
 * import os
 * import capnp
 * import pytact.graph_api_capnp             # <<<<<<<<<<<<<<
 * 
 * cdef class Graph_Node_Reader:
 */
  __pyx_tuple__5 = PyTuple_Pack(2, __pyx_n_s_pytact, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__10 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__10)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__83 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__83)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__84 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__84)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__86 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__86)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__89 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__89)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__90 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__90)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__91 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__91)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__93 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__93)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__96 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__96)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__98 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__98)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__99 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__99)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__101 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__101)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__107 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__107)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__108 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__108, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */
  __pyx_codeobj__110 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__110)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "EnumBase":24
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int, metaclass=__Pyx_EnumMeta):
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */
  __pyx_tuple__111 = PyTuple_Pack(5, __pyx_n_s_cls, __pyx_n_s_value, __pyx_n_s_name, __pyx_n_s_v, __pyx_n_s_res); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);
  __pyx_codeobj__112 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__111, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_new, 24, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__112)) __PYX_ERR(0, 24, __pyx_L1_error)
  __pyx_tuple__113 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__113)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);

  /* "EnumBase":35
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_repr, 35, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(0, 35, __pyx_L1_error)

  /* "EnumBase":37
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */
  __pyx_codeobj__115 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_str, 37, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__115)) __PYX_ERR(0, 37, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__116 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);
  __pyx_codeobj__117 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__116, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__117)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  #if CYTHON_USE_MODULE_STATE
  if (__Pyx_InitString(__pyx_string_tab[0], &__pyx_n_s_ALIGNMENT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[1], &__pyx_n_s_APP) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[2], &__pyx_n_s_APP_ARG) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[3], &__pyx_n_s_APP_FUN) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[4], &__pyx_n_s_AbstractTactic) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[5], &__pyx_n_s_AbstractTactic_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[6], &__pyx_n_s_AbstractTactic_Reader_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[7], &__pyx_n_s_AbstractTactic_Reader_List___red) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[8], &__pyx_n_s_AbstractTactic_Reader_List___set) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[9], &__pyx_n_s_AbstractTactic_Reader___reduce_c) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[10], &__pyx_n_s_AbstractTactic_Reader___setstate) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[11], &__pyx_n_s_Argument) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[12], &__pyx_n_s_Argument_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[13], &__pyx_n_s_Argument_Reader_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[14], &__pyx_n_s_Argument_Reader_List___reduce_cy) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[15], &__pyx_n_s_Argument_Reader_List___setstate) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[16], &__pyx_n_s_Argument_Reader___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[17], &__pyx_n_s_Argument_Reader___setstate_cytho) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[18], &__pyx_n_s_Argument_Term_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[19], &__pyx_n_s_Argument_Term_Reader___reduce_cy) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[20], &__pyx_n_s_Argument_Term_Reader___setstate) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[21], &__pyx_n_s_Argument_Which) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[22], &__pyx_n_s_CASE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[23], &__pyx_n_s_CASE_BRANCH) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[24], &__pyx_n_s_CASE_BRANCH_POINTER) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[25], &__pyx_n_s_CASE_IND) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[26], &__pyx_n_s_CASE_RETURN) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[27], &__pyx_n_s_CASE_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[28], &__pyx_n_s_CAST) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[29], &__pyx_n_s_CAST_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[30], &__pyx_n_s_CAST_TYPE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[31], &__pyx_n_s_CHECK_ALIGNMENT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[32], &__pyx_n_s_COMPLETE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[33], &__pyx_n_s_CONSTRUCTOR) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[34], &__pyx_n_s_CONSTRUCT_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[35], &__pyx_n_s_CONST_DEF) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[36], &__pyx_n_s_CONST_EMPTY) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[37], &__pyx_n_s_CONST_OPAQUE_DEF) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[38], &__pyx_n_s_CONST_PRIMITIVE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[39], &__pyx_n_s_CONST_TYPE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[40], &__pyx_n_s_CONST_UNDEF) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[41], &__pyx_n_s_CONTEXT_ASSUM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[42], &__pyx_n_s_CONTEXT_DEF) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[43], &__pyx_n_s_CONTEXT_DEF_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[44], &__pyx_n_s_CONTEXT_DEF_TYPE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[45], &__pyx_n_s_CONTEXT_ELEM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[46], &__pyx_n_s_CONTEXT_SUBJECT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[47], &__pyx_n_s_CO_FIX) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[48], &__pyx_n_s_CO_FIX_FUN) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[49], &__pyx_n_s_CO_FIX_FUN_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[50], &__pyx_n_s_CO_FIX_FUN_TYPE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[51], &__pyx_n_s_CO_FIX_MUTUAL) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[52], &__pyx_n_s_CO_FIX_RETURN) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[53], &__pyx_n_s_C_B_CONSTRUCT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[54], &__pyx_n_s_C_B_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[55], &__pyx_n_s_ConflatableEdges) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[56], &__pyx_n_s_ConflatableEdges_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[57], &__pyx_n_s_ConflatableEdges_Reader___reduce) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[58], &__pyx_n_s_ConflatableEdges_Reader___setsta) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[59], &__pyx_n_s_DEFINITION) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[60], &__pyx_n_s_DISCHARGED) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[61], &__pyx_n_s_DataVersion) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[62], &__pyx_n_s_DataVersion_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[63], &__pyx_n_s_DataVersion_Reader___reduce_cyth) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[64], &__pyx_n_s_DataVersion_Reader___setstate_cy) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[65], &__pyx_n_s_Dataset) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[66], &__pyx_n_s_Dataset_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[67], &__pyx_n_s_Dataset_Reader___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[68], &__pyx_n_s_Dataset_Reader___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[69], &__pyx_n_s_Definition) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[70], &__pyx_n_s_Definition_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[71], &__pyx_n_s_Definition_Reader___reduce_cytho) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[72], &__pyx_n_s_Definition_Reader___setstate_cyt) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[73], &__pyx_n_s_Definition_Status_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[74], &__pyx_n_s_Definition_Status_Reader___reduc) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[75], &__pyx_n_s_Definition_Status_Reader___setst) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[76], &__pyx_n_s_Definition_Status_Substituted_Re) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[77], &__pyx_n_s_Definition_Status_Substituted_Re_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[78], &__pyx_n_s_Definition_Status_Substituted_Re_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[79], &__pyx_n_s_Definition_Status_Which) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[80], &__pyx_n_s_Definition_Which) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[81], &__pyx_n_s_EVAR) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[82], &__pyx_n_s_EVAR_SUBJECT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[83], &__pyx_n_s_EVAR_SUBST) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[84], &__pyx_n_s_EVAR_SUBST_POINTER) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[85], &__pyx_n_s_EVAR_SUBST_TARGET) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[86], &__pyx_n_s_EVAR_SUBST_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[87], &__pyx_n_s_EdgeClassification) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[88], &__pyx_n_s_EdgeTarget) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[89], &__pyx_n_s_EnumBase) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[90], &__pyx_n_s_Exception) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[91], &__pyx_n_s_Exception_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[92], &__pyx_n_s_Exception_Reader___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[93], &__pyx_n_s_Exception_Reader___setstate_cyth) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[94], &__pyx_n_s_Exception_Which) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[95], &__pyx_n_s_ExecutionResult) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[96], &__pyx_n_s_ExecutionResult_NewState_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[97], &__pyx_n_s_ExecutionResult_NewState_Reader_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[98], &__pyx_n_s_ExecutionResult_NewState_Reader_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[99], &__pyx_n_s_ExecutionResult_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[100], &__pyx_n_s_ExecutionResult_Reader___reduce) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[101], &__pyx_n_s_ExecutionResult_Reader___setstat) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[102], &__pyx_n_s_ExecutionResult_Which) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[103], &__pyx_n_s_FAILURE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[104], &__pyx_n_s_FIX) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[105], &__pyx_n_s_FIX_FUN) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[106], &__pyx_n_s_FIX_FUN_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[107], &__pyx_n_s_FIX_FUN_TYPE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[108], &__pyx_n_s_FIX_MUTUAL) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[109], &__pyx_n_s_FIX_RETURN) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[110], &__pyx_n_s_FLOAT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[111], &__pyx_n_s_FloatP) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[112], &__pyx_n_s_FloatP_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[113], &__pyx_n_s_FloatP_Reader___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[114], &__pyx_n_s_FloatP_Reader___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[115], &__pyx_n_s_Graph) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[116], &__pyx_n_s_Graph_EdgeTarget_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[117], &__pyx_n_s_Graph_EdgeTarget_Reader_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[118], &__pyx_n_s_Graph_EdgeTarget_Reader_List___r) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[119], &__pyx_n_s_Graph_EdgeTarget_Reader_List___s) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[120], &__pyx_n_s_Graph_EdgeTarget_Reader___reduce) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[121], &__pyx_n_s_Graph_EdgeTarget_Reader___setsta) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[122], &__pyx_n_s_Graph_EdgeTarget_Target_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[123], &__pyx_n_s_Graph_EdgeTarget_Target_Reader_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[124], &__pyx_n_s_Graph_EdgeTarget_Target_Reader_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[125], &__pyx_n_s_Graph_Node_Label_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[126], &__pyx_n_s_Graph_Node_Label_Reader___reduce) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[127], &__pyx_n_s_Graph_Node_Label_Reader___setsta) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[128], &__pyx_n_s_Graph_Node_Label_Which) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[129], &__pyx_n_s_Graph_Node_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[130], &__pyx_n_s_Graph_Node_Reader_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[131], &__pyx_n_s_Graph_Node_Reader_List___reduce) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[132], &__pyx_n_s_Graph_Node_Reader_List___setstat) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[133], &__pyx_n_s_Graph_Node_Reader___reduce_cytho) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[134], &__pyx_n_s_Graph_Node_Reader___setstate_cyt) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[135], &__pyx_n_s_Graph_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[136], &__pyx_n_s_Graph_Reader___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[137], &__pyx_n_s_Graph_Reader___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[138], &__pyx_n_s_ILLEGAL_ARGUMENT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[139], &__pyx_n_s_INDUCTIVE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[140], &__pyx_n_s_IND_CONSTRUCT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[141], &__pyx_n_s_IND_PROJECTION) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[142], &__pyx_n_s_IND_TYPE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[143], &__pyx_n_s_INITIALIZE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[144], &__pyx_n_s_INITIALIZED) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[145], &__pyx_n_s_INT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[146], &__pyx_kp_s_Incompatible_checksums_0x_x_vs_0) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[147], &__pyx_n_s_IndexError) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[148], &__pyx_n_s_Int64_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[149], &__pyx_n_s_Int64_List___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[150], &__pyx_n_s_Int64_List___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[151], &__pyx_n_s_IntEnum) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[152], &__pyx_n_s_IntP) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[153], &__pyx_n_s_IntP_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[154], &__pyx_n_s_IntP_Reader___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[155], &__pyx_n_s_IntP_Reader___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[156], &__pyx_n_s_KNOWN) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[157], &__pyx_n_s_LAMBDA) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[158], &__pyx_n_s_LAMBDA_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[159], &__pyx_n_s_LAMBDA_TYPE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[160], &__pyx_n_s_LET_IN) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[161], &__pyx_n_s_LET_IN_DEF) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[162], &__pyx_n_s_LET_IN_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[163], &__pyx_n_s_LET_IN_TYPE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[164], &__pyx_n_s_MANUAL_CONSTANT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[165], &__pyx_n_s_MANUAL_SECTION_CONSTANT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[166], &__pyx_n_s_MISMATCHED_ARGUMENTS) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[167], &__pyx_n_s_NEW_STATE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[168], &__pyx_n_s_NO_SUCH_TACTIC) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[169], &__pyx_n_s_Node) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[170], &__pyx_n_s_Node_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[171], &__pyx_n_s_Node_Reader_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[172], &__pyx_n_s_Node_Reader_List___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[173], &__pyx_n_s_Node_Reader_List___setstate_cyth) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[174], &__pyx_n_s_Node_Reader___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[175], &__pyx_n_s_Node_Reader___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[176], &__pyx_n_s_ORIGINAL) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[177], &__pyx_n_s_OrderedDict) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[178], &__pyx_kp_u_Out_of_bounds) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[179], &__pyx_n_s_Outcome) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[180], &__pyx_n_s_Outcome_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[181], &__pyx_n_s_Outcome_Reader_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[182], &__pyx_n_s_Outcome_Reader_List___reduce_cyt) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[183], &__pyx_n_s_Outcome_Reader_List___setstate_c) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[184], &__pyx_n_s_Outcome_Reader___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[185], &__pyx_n_s_Outcome_Reader___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[186], &__pyx_n_s_Outcome_Term_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[187], &__pyx_n_s_Outcome_Term_Reader___reduce_cyt) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[188], &__pyx_n_s_Outcome_Term_Reader___setstate_c) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[189], &__pyx_n_s_PARSE_ERROR) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[190], &__pyx_n_s_PREDICT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[191], &__pyx_n_s_PREDICTION) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[192], &__pyx_n_s_PRIMITIVE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[193], &__pyx_n_s_PROD) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[194], &__pyx_n_s_PROD_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[195], &__pyx_n_s_PROD_TYPE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[196], &__pyx_n_s_PROJECTION) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[197], &__pyx_n_s_PROJ_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[198], &__pyx_n_s_PROOF_STATE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[199], &__pyx_n_s_PROTOCOL_ERROR) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[200], &__pyx_n_s_PickleError) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[201], &__pyx_n_s_Prediction) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[202], &__pyx_n_s_PredictionProtocol) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[203], &__pyx_n_s_PredictionProtocol_Prediction_Re) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[204], &__pyx_n_s_PredictionProtocol_Prediction_Re_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[205], &__pyx_n_s_PredictionProtocol_Prediction_Re_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[206], &__pyx_n_s_PredictionProtocol_Prediction_Re_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[207], &__pyx_n_s_PredictionProtocol_Prediction_Re_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[208], &__pyx_n_s_PredictionProtocol_Prediction_Re_6) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[209], &__pyx_n_s_PredictionProtocol_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[210], &__pyx_n_s_PredictionProtocol_Reader___redu) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[211], &__pyx_n_s_PredictionProtocol_Reader___sets) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[212], &__pyx_n_s_PredictionProtocol_Request_Initi) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[213], &__pyx_n_s_PredictionProtocol_Request_Initi_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[214], &__pyx_n_s_PredictionProtocol_Request_Initi_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[215], &__pyx_n_s_PredictionProtocol_Request_Predi) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[216], &__pyx_n_s_PredictionProtocol_Request_Predi_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[217], &__pyx_n_s_PredictionProtocol_Request_Predi_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[218], &__pyx_n_s_PredictionProtocol_Request_Reade) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[219], &__pyx_n_s_PredictionProtocol_Request_Reade_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[220], &__pyx_n_s_PredictionProtocol_Request_Reade_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[221], &__pyx_n_s_PredictionProtocol_Request_Which) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[222], &__pyx_n_s_PredictionProtocol_Response_Alig) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[223], &__pyx_n_s_PredictionProtocol_Response_Alig_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[224], &__pyx_n_s_PredictionProtocol_Response_Alig_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[225], &__pyx_n_s_PredictionProtocol_Response_Read) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[226], &__pyx_n_s_PredictionProtocol_Response_Read_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[227], &__pyx_n_s_PredictionProtocol_Response_Read_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[228], &__pyx_n_s_PredictionProtocol_Response_Whic) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[229], &__pyx_n_s_PredictionProtocol_TextPredictio) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[230], &__pyx_n_s_PredictionProtocol_TextPredictio_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[231], &__pyx_n_s_PredictionProtocol_TextPredictio_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[232], &__pyx_n_s_PredictionProtocol_TextPredictio_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[233], &__pyx_n_s_PredictionProtocol_TextPredictio_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[234], &__pyx_n_s_PredictionProtocol_TextPredictio_6) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[235], &__pyx_n_s_ProofState) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[236], &__pyx_n_s_ProofState_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[237], &__pyx_n_s_ProofState_Reader_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[238], &__pyx_n_s_ProofState_Reader_List___reduce) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[239], &__pyx_n_s_ProofState_Reader_List___setstat) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[240], &__pyx_n_s_ProofState_Reader___reduce_cytho) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[241], &__pyx_n_s_ProofState_Reader___setstate_cyt) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[242], &__pyx_n_s_ProofState_Root_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[243], &__pyx_n_s_ProofState_Root_Reader___reduce) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[244], &__pyx_n_s_ProofState_Root_Reader___setstat) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[245], &__pyx_n_s_ProofStep) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[246], &__pyx_n_s_ProofStep_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[247], &__pyx_n_s_ProofStep_Reader_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[248], &__pyx_n_s_ProofStep_Reader_List___reduce_c) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[249], &__pyx_n_s_ProofStep_Reader_List___setstate) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[250], &__pyx_n_s_ProofStep_Reader___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[251], &__pyx_n_s_ProofStep_Reader___setstate_cyth) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[252], &__pyx_n_s_ProofStep_Tactic_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[253], &__pyx_n_s_ProofStep_Tactic_Reader___reduce) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[254], &__pyx_n_s_ProofStep_Tactic_Reader___setsta) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[255], &__pyx_n_s_ProofStep_Tactic_Which) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[256], &__pyx_n_s_Pyx_EnumBase) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[257], &__pyx_n_s_Pyx_EnumBase___new) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[258], &__pyx_n_s_Pyx_EnumBase___repr) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[259], &__pyx_n_s_Pyx_EnumBase___str) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[260], &__pyx_n_s_Pyx_EnumMeta___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[261], &__pyx_n_s_Pyx_EnumMeta___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[262], &__pyx_n_s_REL) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[263], &__pyx_n_s_REL_POINTER) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[264], &__pyx_n_s_Request) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[265], &__pyx_n_s_Response) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[266], &__pyx_n_s_SORT_PROP) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[267], &__pyx_n_s_SORT_SET) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[268], &__pyx_n_s_SORT_S_PROP) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[269], &__pyx_n_s_SORT_TYPE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[270], &__pyx_n_s_SUBSTITUTED) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[271], &__pyx_n_s_SYNCHRONIZE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[272], &__pyx_n_s_SYNCHRONIZED) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[273], &__pyx_n_s_String_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[274], &__pyx_n_s_String_List___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[275], &__pyx_n_s_String_List___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[276], &__pyx_n_s_TACTICAL_CONSTANT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[277], &__pyx_n_s_TACTICAL_SECTION_CONSTANT) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[278], &__pyx_n_s_TERM) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[279], &__pyx_n_s_TEXT_PREDICTION) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[280], &__pyx_n_s_Tactic) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[281], &__pyx_n_s_Tactic_Reader) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[282], &__pyx_n_s_Tactic_Reader___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[283], &__pyx_n_s_Tactic_Reader___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[284], &__pyx_n_s_TextPrediction) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[285], &__pyx_n_s_TypeError) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[286], &__pyx_n_s_UNKNOWN) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[287], &__pyx_n_s_UNRESOLVABLE) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[288], &__pyx_n_s_Uint16_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[289], &__pyx_n_s_Uint16_List___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[290], &__pyx_n_s_Uint16_List___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[291], &__pyx_n_s_Uint32_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[292], &__pyx_n_s_Uint32_List___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[293], &__pyx_n_s_Uint32_List___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[294], &__pyx_n_s_Uint64_List) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[295], &__pyx_n_s_Uint64_List___reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[296], &__pyx_n_s_Uint64_List___setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[297], &__pyx_kp_s_Unknown_enum_value_s) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[298], &__pyx_n_s_ValueError) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[299], &__pyx_n_s__118) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[300], &__pyx_kp_u__3) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[301], &__pyx_n_s__4) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[302], &__pyx_n_s_as_struct) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[303], &__pyx_n_s_asyncio_coroutines) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[304], &__pyx_n_s_capnp) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[305], &__pyx_n_s_class) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[306], &__pyx_n_s_class_getitem) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[307], &__pyx_n_s_cline_in_traceback) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[308], &__pyx_n_s_cls) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[309], &__pyx_n_s_collections) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[310], &__pyx_n_s_dct) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[311], &__pyx_n_s_dict) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[312], &__pyx_n_s_dict_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[313], &__pyx_kp_u_disable) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[314], &__pyx_n_s_doc) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[315], &__pyx_n_s_dyn) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[316], &__pyx_n_s_dynamic) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[317], &__pyx_kp_u_enable) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[318], &__pyx_n_s_enum) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[319], &__pyx_kp_u_gc) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[320], &__pyx_n_s_getstate) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[321], &__pyx_n_s_graph_api_capnp) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[322], &__pyx_n_s_import) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[323], &__pyx_n_s_init) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[324], &__pyx_n_s_init_subclass) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[325], &__pyx_n_s_initializing) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[326], &__pyx_n_s_is_coroutine) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[327], &__pyx_kp_u_isenabled) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[328], &__pyx_n_s_main) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[329], &__pyx_n_s_members) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[330], &__pyx_n_s_metaclass) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[331], &__pyx_n_s_module) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[332], &__pyx_n_s_mro_entries) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[333], &__pyx_n_s_name) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[334], &__pyx_n_s_name_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[335], &__pyx_n_s_new) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[336], &__pyx_n_s_os) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[337], &__pyx_n_s_parents) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[338], &__pyx_n_s_pickle) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[339], &__pyx_n_s_prepare) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[340], &__pyx_n_s_pytact) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[341], &__pyx_n_s_pytact_graph_api_capnp) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[342], &__pyx_n_s_pytact_graph_api_capnp_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[343], &__pyx_n_s_pyx_PickleError) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[344], &__pyx_n_s_pyx_checksum) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[345], &__pyx_n_s_pyx_result) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[346], &__pyx_n_s_pyx_state) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[347], &__pyx_n_s_pyx_type) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[348], &__pyx_n_s_pyx_unpickle___Pyx_EnumMeta) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[349], &__pyx_n_s_pyx_vtable) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[350], &__pyx_n_s_qualname) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[351], &__pyx_n_s_reduce) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[352], &__pyx_n_s_reduce_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[353], &__pyx_n_s_reduce_ex) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[354], &__pyx_n_s_repr) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[355], &__pyx_n_s_res) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[356], &__pyx_kp_s_s_s) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[357], &__pyx_kp_s_s_s_d) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[358], &__pyx_n_s_schema) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[359], &__pyx_n_s_self) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[360], &__pyx_kp_s_self_source_cannot_be_converted) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[361], &__pyx_n_s_set_name) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[362], &__pyx_n_s_setstate) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[363], &__pyx_n_s_setstate_cython) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[364], &__pyx_n_s_spec) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[365], &__pyx_n_s_state) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[366], &__pyx_n_s_str) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[367], &__pyx_kp_s_stringsource) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[368], &__pyx_n_s_super) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[369], &__pyx_n_s_test) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[370], &__pyx_n_s_update) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[371], &__pyx_n_s_use_setstate) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[372], &__pyx_n_s_v) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[373], &__pyx_n_s_value) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[374], &__pyx_n_s_values) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  #endif
  #if !CYTHON_USE_MODULE_STATE
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  #endif
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_228825662 = PyInt_FromLong(228825662L); if (unlikely(!__pyx_int_228825662)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_238750788 = PyInt_FromLong(238750788L); if (unlikely(!__pyx_int_238750788)) __PYX_ERR(1, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  return 0;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_OrderedDict = Py_None; Py_INCREF(Py_None);
  __Pyx_EnumBase = Py_None; Py_INCREF(Py_None);
  __Pyx_globals = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Node_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Node_Reader.init = (PyObject *(*)(Graph::Node::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader)) __PYX_ERR(1, 13, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Reader) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Graph_Node_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader.init = (PyObject *(*)(Graph::EdgeTarget::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader)) __PYX_ERR(1, 96, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader) < 0) __PYX_ERR(1, 96, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader) < 0) __PYX_ERR(1, 96, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader) < 0) __PYX_ERR(1, 96, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader) < 0) __PYX_ERR(1, 96, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Graph_EdgeTarget_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader) < 0) __PYX_ERR(1, 96, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader) < 0) __PYX_ERR(1, 96, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Reader.init = (PyObject *(*)(Graph::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_12Graph_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader)) __PYX_ERR(1, 128, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader) < 0) __PYX_ERR(1, 128, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader) < 0) __PYX_ERR(1, 128, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Reader) < 0) __PYX_ERR(1, 128, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader) < 0) __PYX_ERR(1, 128, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Graph_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader) < 0) __PYX_ERR(1, 128, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Reader) < 0) __PYX_ERR(1, 128, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_IntP_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_IntP_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_IntP_Reader.init = (PyObject *(*)(IntP::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_11IntP_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader)) __PYX_ERR(1, 178, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_IntP_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_IntP_Reader) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IntP_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_IntP_Reader) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Definition_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Definition_Reader.init = (PyObject *(*)(Definition::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_17Definition_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader)) __PYX_ERR(1, 202, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader) < 0) __PYX_ERR(1, 202, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader) < 0) __PYX_ERR(1, 202, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Reader) < 0) __PYX_ERR(1, 202, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader) < 0) __PYX_ERR(1, 202, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Definition_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader) < 0) __PYX_ERR(1, 202, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Reader) < 0) __PYX_ERR(1, 202, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_FloatP_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_FloatP_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_FloatP_Reader.init = (PyObject *(*)(FloatP::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_13FloatP_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader)) __PYX_ERR(1, 375, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader) < 0) __PYX_ERR(1, 375, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_FloatP_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader) < 0) __PYX_ERR(1, 375, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_FloatP_Reader) < 0) __PYX_ERR(1, 375, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader) < 0) __PYX_ERR(1, 375, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FloatP_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader) < 0) __PYX_ERR(1, 375, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_FloatP_Reader) < 0) __PYX_ERR(1, 375, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader.init = (PyObject *(*)(Graph::Node::Label::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader)) __PYX_ERR(1, 399, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader) < 0) __PYX_ERR(1, 399, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader) < 0) __PYX_ERR(1, 399, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader) < 0) __PYX_ERR(1, 399, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader) < 0) __PYX_ERR(1, 399, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Graph_Node_Label_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader) < 0) __PYX_ERR(1, 399, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Label_Reader) < 0) __PYX_ERR(1, 399, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader.init = (PyObject *(*)(Definition::Status::Substituted::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader)) __PYX_ERR(1, 631, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader) < 0) __PYX_ERR(1, 631, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader) < 0) __PYX_ERR(1, 631, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader) < 0) __PYX_ERR(1, 631, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader) < 0) __PYX_ERR(1, 631, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Definition_Status_Substituted_Re, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader) < 0) __PYX_ERR(1, 631, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Substituted_Reader) < 0) __PYX_ERR(1, 631, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Status_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Definition_Status_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Definition_Status_Reader.init = (PyObject *(*)(Definition::Status::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader)) __PYX_ERR(1, 669, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader) < 0) __PYX_ERR(1, 669, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Definition_Status_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader) < 0) __PYX_ERR(1, 669, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Definition_Status_Reader) < 0) __PYX_ERR(1, 669, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader) < 0) __PYX_ERR(1, 669, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Definition_Status_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader) < 0) __PYX_ERR(1, 669, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Definition_Status_Reader) < 0) __PYX_ERR(1, 669, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_ProofStep_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofStep_Reader.init = (PyObject *(*)(ProofStep::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader)) __PYX_ERR(1, 729, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader) < 0) __PYX_ERR(1, 729, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader) < 0) __PYX_ERR(1, 729, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Reader) < 0) __PYX_ERR(1, 729, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader) < 0) __PYX_ERR(1, 729, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ProofStep_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader) < 0) __PYX_ERR(1, 729, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader) < 0) __PYX_ERR(1, 729, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Tactic_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Tactic_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Tactic_Reader.init = (PyObject *(*)(Tactic::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_13Tactic_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader)) __PYX_ERR(1, 770, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader) < 0) __PYX_ERR(1, 770, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Tactic_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader) < 0) __PYX_ERR(1, 770, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Tactic_Reader) < 0) __PYX_ERR(1, 770, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader) < 0) __PYX_ERR(1, 770, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Tactic_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader) < 0) __PYX_ERR(1, 770, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Tactic_Reader) < 0) __PYX_ERR(1, 770, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader.init = (PyObject *(*)(ProofStep::Tactic::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader)) __PYX_ERR(1, 868, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader) < 0) __PYX_ERR(1, 868, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader) < 0) __PYX_ERR(1, 868, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader) < 0) __PYX_ERR(1, 868, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader) < 0) __PYX_ERR(1, 868, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ProofStep_Tactic_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader) < 0) __PYX_ERR(1, 868, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Tactic_Reader) < 0) __PYX_ERR(1, 868, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Outcome_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Outcome_Reader.init = (PyObject *(*)(Outcome::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_14Outcome_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader)) __PYX_ERR(1, 922, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader) < 0) __PYX_ERR(1, 922, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader) < 0) __PYX_ERR(1, 922, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Reader) < 0) __PYX_ERR(1, 922, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader) < 0) __PYX_ERR(1, 922, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Outcome_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader) < 0) __PYX_ERR(1, 922, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader) < 0) __PYX_ERR(1, 922, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader.init = (PyObject *(*)(Outcome::Term::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader)) __PYX_ERR(1, 996, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader) < 0) __PYX_ERR(1, 996, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader) < 0) __PYX_ERR(1, 996, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader) < 0) __PYX_ERR(1, 996, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader) < 0) __PYX_ERR(1, 996, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Outcome_Term_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader) < 0) __PYX_ERR(1, 996, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Term_Reader) < 0) __PYX_ERR(1, 996, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_ProofState_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofState_Reader.init = (PyObject *(*)(ProofState::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_17ProofState_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader)) __PYX_ERR(1, 1035, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader) < 0) __PYX_ERR(1, 1035, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader) < 0) __PYX_ERR(1, 1035, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Reader) < 0) __PYX_ERR(1, 1035, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader) < 0) __PYX_ERR(1, 1035, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ProofState_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader) < 0) __PYX_ERR(1, 1035, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader) < 0) __PYX_ERR(1, 1035, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Argument_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Argument_Reader.init = (PyObject *(*)(Argument::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_15Argument_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader)) __PYX_ERR(1, 1135, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader) < 0) __PYX_ERR(1, 1135, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader) < 0) __PYX_ERR(1, 1135, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Reader) < 0) __PYX_ERR(1, 1135, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader) < 0) __PYX_ERR(1, 1135, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Argument_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader) < 0) __PYX_ERR(1, 1135, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader) < 0) __PYX_ERR(1, 1135, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader.init = (PyObject *(*)(ProofState::Root::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader)) __PYX_ERR(1, 1184, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ProofState_Root_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Root_Reader) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Node_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Node_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Node_Reader.init = (PyObject *(*)(Node::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_11Node_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader)) __PYX_ERR(1, 1221, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader) < 0) __PYX_ERR(1, 1221, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader) < 0) __PYX_ERR(1, 1221, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Node_Reader) < 0) __PYX_ERR(1, 1221, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader) < 0) __PYX_ERR(1, 1221, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Node_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader) < 0) __PYX_ERR(1, 1221, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader) < 0) __PYX_ERR(1, 1221, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Term_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Argument_Term_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Argument_Term_Reader.init = (PyObject *(*)(Argument::Term::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader)) __PYX_ERR(1, 1259, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Term_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Term_Reader) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Argument_Term_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Term_Reader) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader.init = (PyObject *(*)(Graph::EdgeTarget::Target::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader)) __PYX_ERR(1, 1296, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader) < 0) __PYX_ERR(1, 1296, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader) < 0) __PYX_ERR(1, 1296, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader) < 0) __PYX_ERR(1, 1296, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader) < 0) __PYX_ERR(1, 1296, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Graph_EdgeTarget_Target_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader) < 0) __PYX_ERR(1, 1296, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Target_Reader) < 0) __PYX_ERR(1, 1296, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader.init = (PyObject *(*)(AbstractTactic::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader)) __PYX_ERR(1, 1330, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader) < 0) __PYX_ERR(1, 1330, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader) < 0) __PYX_ERR(1, 1330, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader) < 0) __PYX_ERR(1, 1330, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader) < 0) __PYX_ERR(1, 1330, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AbstractTactic_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader) < 0) __PYX_ERR(1, 1330, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader) < 0) __PYX_ERR(1, 1330, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_DataVersion_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_DataVersion_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_DataVersion_Reader.init = (PyObject *(*)(DataVersion::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader)) __PYX_ERR(1, 1364, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader) < 0) __PYX_ERR(1, 1364, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_DataVersion_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader) < 0) __PYX_ERR(1, 1364, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_DataVersion_Reader) < 0) __PYX_ERR(1, 1364, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader) < 0) __PYX_ERR(1, 1364, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DataVersion_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader) < 0) __PYX_ERR(1, 1364, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_DataVersion_Reader) < 0) __PYX_ERR(1, 1364, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Dataset_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Dataset_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Dataset_Reader.init = (PyObject *(*)(Dataset::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_14Dataset_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader)) __PYX_ERR(1, 1398, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader) < 0) __PYX_ERR(1, 1398, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Dataset_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader) < 0) __PYX_ERR(1, 1398, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Dataset_Reader) < 0) __PYX_ERR(1, 1398, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader) < 0) __PYX_ERR(1, 1398, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Dataset_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader) < 0) __PYX_ERR(1, 1398, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Dataset_Reader) < 0) __PYX_ERR(1, 1398, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Exception_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Exception_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Exception_Reader.init = (PyObject *(*)(Exception::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_16Exception_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader)) __PYX_ERR(1, 1486, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader) < 0) __PYX_ERR(1, 1486, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_Exception_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader) < 0) __PYX_ERR(1, 1486, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Exception_Reader) < 0) __PYX_ERR(1, 1486, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader) < 0) __PYX_ERR(1, 1486, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Exception_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader) < 0) __PYX_ERR(1, 1486, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Exception_Reader) < 0) __PYX_ERR(1, 1486, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader.init = (PyObject *(*)(ExecutionResult::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader)) __PYX_ERR(1, 1543, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader) < 0) __PYX_ERR(1, 1543, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader) < 0) __PYX_ERR(1, 1543, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader) < 0) __PYX_ERR(1, 1543, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader) < 0) __PYX_ERR(1, 1543, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ExecutionResult_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader) < 0) __PYX_ERR(1, 1543, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_Reader) < 0) __PYX_ERR(1, 1543, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader.init = (PyObject *(*)(ExecutionResult::NewState::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader)) __PYX_ERR(1, 1615, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader) < 0) __PYX_ERR(1, 1615, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader) < 0) __PYX_ERR(1, 1615, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader) < 0) __PYX_ERR(1, 1615, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader) < 0) __PYX_ERR(1, 1615, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ExecutionResult_NewState_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader) < 0) __PYX_ERR(1, 1615, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ExecutionResult_NewState_Reader) < 0) __PYX_ERR(1, 1615, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader.init = (PyObject *(*)(PredictionProtocol::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader)) __PYX_ERR(1, 1651, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader) < 0) __PYX_ERR(1, 1651, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader) < 0) __PYX_ERR(1, 1651, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader) < 0) __PYX_ERR(1, 1651, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader) < 0) __PYX_ERR(1, 1651, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PredictionProtocol_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader) < 0) __PYX_ERR(1, 1651, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Reader) < 0) __PYX_ERR(1, 1651, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader.init = (PyObject *(*)(PredictionProtocol::Request::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader)) __PYX_ERR(1, 1676, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader) < 0) __PYX_ERR(1, 1676, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader) < 0) __PYX_ERR(1, 1676, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader) < 0) __PYX_ERR(1, 1676, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader) < 0) __PYX_ERR(1, 1676, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PredictionProtocol_Request_Reade, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader) < 0) __PYX_ERR(1, 1676, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Reader) < 0) __PYX_ERR(1, 1676, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader.init = (PyObject *(*)(PredictionProtocol::Request::Initialize::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader)) __PYX_ERR(1, 1745, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader) < 0) __PYX_ERR(1, 1745, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader) < 0) __PYX_ERR(1, 1745, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader) < 0) __PYX_ERR(1, 1745, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader) < 0) __PYX_ERR(1, 1745, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PredictionProtocol_Request_Initi, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader) < 0) __PYX_ERR(1, 1745, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Initialize_Reader) < 0) __PYX_ERR(1, 1745, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader.init = (PyObject *(*)(PredictionProtocol::Request::Predict::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader)) __PYX_ERR(1, 1833, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader) < 0) __PYX_ERR(1, 1833, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader) < 0) __PYX_ERR(1, 1833, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader) < 0) __PYX_ERR(1, 1833, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader) < 0) __PYX_ERR(1, 1833, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PredictionProtocol_Request_Predi, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader) < 0) __PYX_ERR(1, 1833, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Request_Predict_Reader) < 0) __PYX_ERR(1, 1833, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader.init = (PyObject *(*)(PredictionProtocol::Prediction::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader)) __PYX_ERR(1, 1876, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader) < 0) __PYX_ERR(1, 1876, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader) < 0) __PYX_ERR(1, 1876, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader) < 0) __PYX_ERR(1, 1876, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader) < 0) __PYX_ERR(1, 1876, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PredictionProtocol_Prediction_Re, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader) < 0) __PYX_ERR(1, 1876, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader) < 0) __PYX_ERR(1, 1876, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader.init = (PyObject *(*)(PredictionProtocol::TextPrediction::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader)) __PYX_ERR(1, 1914, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader) < 0) __PYX_ERR(1, 1914, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader) < 0) __PYX_ERR(1, 1914, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader) < 0) __PYX_ERR(1, 1914, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader) < 0) __PYX_ERR(1, 1914, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PredictionProtocol_TextPredictio, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader) < 0) __PYX_ERR(1, 1914, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader) < 0) __PYX_ERR(1, 1914, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader.init = (PyObject *(*)(PredictionProtocol::Response::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader)) __PYX_ERR(1, 1946, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader) < 0) __PYX_ERR(1, 1946, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader) < 0) __PYX_ERR(1, 1946, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader) < 0) __PYX_ERR(1, 1946, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader) < 0) __PYX_ERR(1, 1946, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PredictionProtocol_Response_Read, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader) < 0) __PYX_ERR(1, 1946, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Reader) < 0) __PYX_ERR(1, 1946, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader.init = (PyObject *(*)(PredictionProtocol::Response::Alignment::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader)) __PYX_ERR(1, 2020, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader) < 0) __PYX_ERR(1, 2020, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader) < 0) __PYX_ERR(1, 2020, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader) < 0) __PYX_ERR(1, 2020, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader) < 0) __PYX_ERR(1, 2020, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PredictionProtocol_Response_Alig, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader) < 0) __PYX_ERR(1, 2020, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Response_Alignment_Reader) < 0) __PYX_ERR(1, 2020, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader = &__pyx_vtable_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader.init = (PyObject *(*)(ConflatableEdges::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader)) __PYX_ERR(1, 2053, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader) < 0) __PYX_ERR(1, 2053, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader = &__pyx_type_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader) < 0) __PYX_ERR(1, 2053, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader) < 0) __PYX_ERR(1, 2053, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader) < 0) __PYX_ERR(1, 2053, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ConflatableEdges_Reader, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader) < 0) __PYX_ERR(1, 2053, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ConflatableEdges_Reader) < 0) __PYX_ERR(1, 2053, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint16_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Uint16_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Uint16_List.init = (PyObject *(*)( ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_11Uint16_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List)) __PYX_ERR(1, 2074, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List) < 0) __PYX_ERR(1, 2074, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List = &__pyx_type_6pytact_22graph_api_capnp_cython_Uint16_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List) < 0) __PYX_ERR(1, 2074, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint16_List) < 0) __PYX_ERR(1, 2074, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List) < 0) __PYX_ERR(1, 2074, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Uint16_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List) < 0) __PYX_ERR(1, 2074, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint16_List) < 0) __PYX_ERR(1, 2074, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint32_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Uint32_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Uint32_List.init = (PyObject *(*)( ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_11Uint32_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List)) __PYX_ERR(1, 2092, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List) < 0) __PYX_ERR(1, 2092, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List = &__pyx_type_6pytact_22graph_api_capnp_cython_Uint32_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List) < 0) __PYX_ERR(1, 2092, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint32_List) < 0) __PYX_ERR(1, 2092, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List) < 0) __PYX_ERR(1, 2092, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Uint32_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List) < 0) __PYX_ERR(1, 2092, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint32_List) < 0) __PYX_ERR(1, 2092, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint64_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Uint64_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Uint64_List.init = (PyObject *(*)( ::capnp::List<uint64_t, ::capnp::Kind::PRIMITIVE>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_11Uint64_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List)) __PYX_ERR(1, 2110, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List) < 0) __PYX_ERR(1, 2110, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List = &__pyx_type_6pytact_22graph_api_capnp_cython_Uint64_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List) < 0) __PYX_ERR(1, 2110, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Uint64_List) < 0) __PYX_ERR(1, 2110, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List) < 0) __PYX_ERR(1, 2110, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Uint64_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List) < 0) __PYX_ERR(1, 2110, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Uint64_List) < 0) __PYX_ERR(1, 2110, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Int64_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Int64_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Int64_List.init = (PyObject *(*)( ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_10Int64_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Int64_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List)) __PYX_ERR(1, 2128, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Int64_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List) < 0) __PYX_ERR(1, 2128, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List = &__pyx_type_6pytact_22graph_api_capnp_cython_Int64_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List) < 0) __PYX_ERR(1, 2128, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Int64_List) < 0) __PYX_ERR(1, 2128, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List) < 0) __PYX_ERR(1, 2128, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Int64_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List) < 0) __PYX_ERR(1, 2128, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Int64_List) < 0) __PYX_ERR(1, 2128, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_String_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_String_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_String_List.init = (PyObject *(*)( ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_11String_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_String_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_String_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List)) __PYX_ERR(1, 2146, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_String_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_String_List) < 0) __PYX_ERR(1, 2146, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_String_List = &__pyx_type_6pytact_22graph_api_capnp_cython_String_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List) < 0) __PYX_ERR(1, 2146, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_String_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_String_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_String_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_String_List) < 0) __PYX_ERR(1, 2146, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_String_List) < 0) __PYX_ERR(1, 2146, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_String_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_String_List) < 0) __PYX_ERR(1, 2146, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_String_List) < 0) __PYX_ERR(1, 2146, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List.init = (PyObject *(*)( ::capnp::List<PredictionProtocol::Prediction, ::capnp::Kind::STRUCT>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List)) __PYX_ERR(1, 2164, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List) < 0) __PYX_ERR(1, 2164, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List = &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List) < 0) __PYX_ERR(1, 2164, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List) < 0) __PYX_ERR(1, 2164, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List) < 0) __PYX_ERR(1, 2164, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PredictionProtocol_Prediction_Re_2, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List) < 0) __PYX_ERR(1, 2164, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_Prediction_Reader_List) < 0) __PYX_ERR(1, 2164, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List.init = (PyObject *(*)( ::capnp::List<Graph::Node, ::capnp::Kind::STRUCT>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List)) __PYX_ERR(1, 2181, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List) < 0) __PYX_ERR(1, 2181, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List = &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List) < 0) __PYX_ERR(1, 2181, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List) < 0) __PYX_ERR(1, 2181, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List) < 0) __PYX_ERR(1, 2181, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Graph_Node_Reader_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List) < 0) __PYX_ERR(1, 2181, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_Node_Reader_List) < 0) __PYX_ERR(1, 2181, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List.init = (PyObject *(*)( ::capnp::List<AbstractTactic, ::capnp::Kind::STRUCT>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List)) __PYX_ERR(1, 2198, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List) < 0) __PYX_ERR(1, 2198, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List = &__pyx_type_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List) < 0) __PYX_ERR(1, 2198, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List) < 0) __PYX_ERR(1, 2198, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List) < 0) __PYX_ERR(1, 2198, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AbstractTactic_Reader_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List) < 0) __PYX_ERR(1, 2198, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_AbstractTactic_Reader_List) < 0) __PYX_ERR(1, 2198, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List.init = (PyObject *(*)( ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List)) __PYX_ERR(1, 2215, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List) < 0) __PYX_ERR(1, 2215, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List = &__pyx_type_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List) < 0) __PYX_ERR(1, 2215, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List) < 0) __PYX_ERR(1, 2215, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List) < 0) __PYX_ERR(1, 2215, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ProofStep_Reader_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List) < 0) __PYX_ERR(1, 2215, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofStep_Reader_List) < 0) __PYX_ERR(1, 2215, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List.init = (PyObject *(*)( ::capnp::List<PredictionProtocol::TextPrediction, ::capnp::Kind::STRUCT>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List)) __PYX_ERR(1, 2232, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List) < 0) __PYX_ERR(1, 2232, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List = &__pyx_type_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List) < 0) __PYX_ERR(1, 2232, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List) < 0) __PYX_ERR(1, 2232, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List) < 0) __PYX_ERR(1, 2232, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PredictionProtocol_TextPredictio_2, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List) < 0) __PYX_ERR(1, 2232, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_PredictionProtocol_TextPrediction_Reader_List) < 0) __PYX_ERR(1, 2232, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Reader_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_ProofState_Reader_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_ProofState_Reader_List.init = (PyObject *(*)( ::capnp::List<ProofState, ::capnp::Kind::STRUCT>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List)) __PYX_ERR(1, 2249, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List) < 0) __PYX_ERR(1, 2249, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List = &__pyx_type_6pytact_22graph_api_capnp_cython_ProofState_Reader_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List) < 0) __PYX_ERR(1, 2249, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_ProofState_Reader_List) < 0) __PYX_ERR(1, 2249, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List) < 0) __PYX_ERR(1, 2249, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ProofState_Reader_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List) < 0) __PYX_ERR(1, 2249, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_ProofState_Reader_List) < 0) __PYX_ERR(1, 2249, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List.init = (PyObject *(*)( ::capnp::List<Graph::EdgeTarget, ::capnp::Kind::STRUCT>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List)) __PYX_ERR(1, 2266, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List) < 0) __PYX_ERR(1, 2266, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List = &__pyx_type_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List) < 0) __PYX_ERR(1, 2266, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List) < 0) __PYX_ERR(1, 2266, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List) < 0) __PYX_ERR(1, 2266, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Graph_EdgeTarget_Reader_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List) < 0) __PYX_ERR(1, 2266, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Graph_EdgeTarget_Reader_List) < 0) __PYX_ERR(1, 2266, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Reader_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Argument_Reader_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Argument_Reader_List.init = (PyObject *(*)( ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List)) __PYX_ERR(1, 2283, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List) < 0) __PYX_ERR(1, 2283, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List = &__pyx_type_6pytact_22graph_api_capnp_cython_Argument_Reader_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List) < 0) __PYX_ERR(1, 2283, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Argument_Reader_List) < 0) __PYX_ERR(1, 2283, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List) < 0) __PYX_ERR(1, 2283, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Argument_Reader_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List) < 0) __PYX_ERR(1, 2283, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Argument_Reader_List) < 0) __PYX_ERR(1, 2283, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Reader_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Outcome_Reader_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Outcome_Reader_List.init = (PyObject *(*)( ::capnp::List<Outcome, ::capnp::Kind::STRUCT>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List)) __PYX_ERR(1, 2300, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List) < 0) __PYX_ERR(1, 2300, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List = &__pyx_type_6pytact_22graph_api_capnp_cython_Outcome_Reader_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List) < 0) __PYX_ERR(1, 2300, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Outcome_Reader_List) < 0) __PYX_ERR(1, 2300, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List) < 0) __PYX_ERR(1, 2300, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Outcome_Reader_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List) < 0) __PYX_ERR(1, 2300, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Outcome_Reader_List) < 0) __PYX_ERR(1, 2300, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Node_Reader_List = &__pyx_vtable_6pytact_22graph_api_capnp_cython_Node_Reader_List;
  __pyx_vtable_6pytact_22graph_api_capnp_cython_Node_Reader_List.init = (PyObject *(*)( ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader, PyObject *))__pyx_f_6pytact_22graph_api_capnp_cython_16Node_Reader_List_init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List_spec, NULL); if (unlikely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List)) __PYX_ERR(1, 2317, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List_spec, __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List) < 0) __PYX_ERR(1, 2317, __pyx_L1_error)
  #else
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List = &__pyx_type_6pytact_22graph_api_capnp_cython_Node_Reader_List;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List) < 0) __PYX_ERR(1, 2317, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List->tp_dictoffset && __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List, __pyx_vtabptr_6pytact_22graph_api_capnp_cython_Node_Reader_List) < 0) __PYX_ERR(1, 2317, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List) < 0) __PYX_ERR(1, 2317, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Node_Reader_List, (PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List) < 0) __PYX_ERR(1, 2317, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6pytact_22graph_api_capnp_cython_Node_Reader_List) < 0) __PYX_ERR(1, 2317, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)(&PyType_Type)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype___Pyx_EnumMeta = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__Pyx_EnumMeta_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype___Pyx_EnumMeta)) __PYX_ERR(0, 12, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__Pyx_EnumMeta_spec, __pyx_ptype___Pyx_EnumMeta) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  #else
  __pyx_ptype___Pyx_EnumMeta = &__Pyx_EnumMeta;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype___Pyx_EnumMeta->tp_dealloc = (&PyType_Type)->tp_dealloc;
  __pyx_ptype___Pyx_EnumMeta->tp_base = (&PyType_Type);
  __pyx_ptype___Pyx_EnumMeta->tp_new = (&PyType_Type)->tp_new;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype___Pyx_EnumMeta) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype___Pyx_EnumMeta->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype___Pyx_EnumMeta->tp_dictoffset && __pyx_ptype___Pyx_EnumMeta->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype___Pyx_EnumMeta->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype___Pyx_EnumMeta) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("capnp.lib.capnp"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5capnp_3lib_5capnp__StructSchemaField = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_StructSchemaField", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__StructSchemaField),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__StructSchemaField) __PYX_ERR(3, 23, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__StructSchemaField = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__StructSchemaField*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__StructSchemaField); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__StructSchemaField)) __PYX_ERR(3, 23, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp__StringArrayPtr = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_StringArrayPtr", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__StringArrayPtr), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__StringArrayPtr),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__StringArrayPtr) __PYX_ERR(3, 28, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__StringArrayPtr = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__StringArrayPtr*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__StringArrayPtr); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__StringArrayPtr)) __PYX_ERR(3, 28, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp_SchemaParser = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "SchemaParser", sizeof(struct __pyx_obj_5capnp_3lib_5capnp_SchemaParser), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp_SchemaParser),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp_SchemaParser) __PYX_ERR(3, 34, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp_SchemaParser = (struct __pyx_vtabstruct_5capnp_3lib_5capnp_SchemaParser*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp_SchemaParser); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp_SchemaParser)) __PYX_ERR(3, 34, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp__DynamicOrphan = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_DynamicOrphan", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__DynamicOrphan), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__DynamicOrphan),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__DynamicOrphan) __PYX_ERR(3, 41, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__DynamicOrphan = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicOrphan*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__DynamicOrphan); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__DynamicOrphan)) __PYX_ERR(3, 41, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_DynamicStructReader", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructReader),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader) __PYX_ERR(3, 51, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__DynamicStructReader = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicStructReader*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__DynamicStructReader); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__DynamicStructReader)) __PYX_ERR(3, 51, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp__DynamicStructBuilder = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_DynamicStructBuilder", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__DynamicStructBuilder),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__DynamicStructBuilder) __PYX_ERR(3, 70, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__DynamicStructBuilder = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicStructBuilder*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__DynamicStructBuilder); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__DynamicStructBuilder)) __PYX_ERR(3, 70, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp__DynamicEnumField = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_DynamicEnumField", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnumField), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnumField),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__DynamicEnumField) __PYX_ERR(3, 102, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__DynamicEnumField = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicEnumField*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__DynamicEnumField); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__DynamicEnumField)) __PYX_ERR(3, 102, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp__Schema = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_Schema", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__Schema), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__Schema),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__Schema) __PYX_ERR(3, 108, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__Schema = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__Schema); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__Schema)) __PYX_ERR(3, 108, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp__InterfaceSchema = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_InterfaceSchema", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__InterfaceSchema), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__InterfaceSchema),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__InterfaceSchema) __PYX_ERR(3, 119, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__InterfaceSchema = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__InterfaceSchema*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__InterfaceSchema); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__InterfaceSchema)) __PYX_ERR(3, 119, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp__DynamicEnum = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_DynamicEnum", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnum), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__DynamicEnum),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__DynamicEnum) __PYX_ERR(3, 124, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__DynamicEnum = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicEnum*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__DynamicEnum); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__DynamicEnum)) __PYX_ERR(3, 124, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp__DynamicListBuilder = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_DynamicListBuilder", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__DynamicListBuilder), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__DynamicListBuilder),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__DynamicListBuilder) __PYX_ERR(3, 131, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__DynamicListBuilder = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__DynamicListBuilder*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__DynamicListBuilder); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__DynamicListBuilder)) __PYX_ERR(3, 131, __pyx_L1_error)
  __pyx_ptype_5capnp_3lib_5capnp__MessageBuilder = __Pyx_ImportType(__pyx_t_1, "capnp.lib.capnp", "_MessageBuilder", sizeof(struct __pyx_obj_5capnp_3lib_5capnp__MessageBuilder), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_5capnp_3lib_5capnp__MessageBuilder),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5capnp_3lib_5capnp__MessageBuilder) __PYX_ERR(3, 144, __pyx_L1_error)
  __pyx_vtabptr_5capnp_3lib_5capnp__MessageBuilder = (struct __pyx_vtabstruct_5capnp_3lib_5capnp__MessageBuilder*)__Pyx_GetVtable(__pyx_ptype_5capnp_3lib_5capnp__MessageBuilder); if (unlikely(!__pyx_vtabptr_5capnp_3lib_5capnp__MessageBuilder)) __PYX_ERR(3, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("capnp.lib.capnp"); if (!__pyx_t_1) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "to_python_reader", (void (**)(void))&__pyx_f_5capnp_3lib_5capnp_to_python_reader, "PyObject *( ::capnp::DynamicValue::Reader, PyObject *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_graph_api_capnp_cython(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_graph_api_capnp_cython},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "graph_api_capnp_cython",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initgraph_api_capnp_cython(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initgraph_api_capnp_cython(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_graph_api_capnp_cython(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_graph_api_capnp_cython(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_graph_api_capnp_cython(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  PyObject *__pyx_t_38 = NULL;
  PyObject *__pyx_t_39 = NULL;
  PyObject *__pyx_t_40 = NULL;
  PyObject *__pyx_t_41 = NULL;
  PyObject *__pyx_t_42 = NULL;
  PyObject *__pyx_t_43 = NULL;
  PyObject *__pyx_t_44 = NULL;
  PyObject *__pyx_t_45 = NULL;
  PyObject *__pyx_t_46 = NULL;
  PyObject *__pyx_t_47 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'graph_api_capnp_cython' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("graph_api_capnp_cython", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #elif CYTHON_COMPILING_IN_LIMITED_API
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely((add_module_result < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_graph_api_capnp_cython(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pytact__graph_api_capnp_cython) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(1, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pytact.graph_api_capnp_cython")) {
      if (unlikely((PyDict_SetItemString(modules, "pytact.graph_api_capnp_cython", __pyx_m) < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely((__Pyx_modinit_function_import_code() < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif

  /* "pytact/graph_api_capnp_cython.pyx":9
 * # cython: language_level = 3
 * 
 * import os             # <<<<<<<<<<<<<<
 * import capnp
 * import pytact.graph_api_capnp
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_os, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_2) < 0) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":10
 * 
 * import os
 * import capnp             # <<<<<<<<<<<<<<
 * import pytact.graph_api_capnp
 * 
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_capnp, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_capnp, __pyx_t_2) < 0) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":11
 * import os
 * import capnp
 * import pytact.graph_api_capnp             # <<<<<<<<<<<<<<
 * 
 * cdef class Graph_Node_Reader:
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pytact_graph_api_capnp, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pytact, __pyx_t_2) < 0) __PYX_ERR(1, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_Node_Reader___reduce_cytho, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__7)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_17Graph_Node_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_Node_Reader___setstate_cyt, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__9)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":94
 *         """
 *         return self.source.getIdentity()
 * pytact.graph_api_capnp.Graph.Node.schema = _Schema()._init(getGraph_Node_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class Graph_EdgeTarget_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<Graph::Node>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Graph); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Node); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_EdgeTarget_Reader___reduce, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__10)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_23Graph_EdgeTarget_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_EdgeTarget_Reader___setsta, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__11)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":126
 *     def target(self):
 *         return Graph_EdgeTarget_Target_Reader.init(self.source.getTarget(), self.root)
 * pytact.graph_api_capnp.Graph.EdgeTarget.schema = _Schema()._init(getGraph_EdgeTarget_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class Graph_Reader:
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_4)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_4), capnp::Schema::from<Graph::EdgeTarget>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pytact); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Graph); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_EdgeTarget); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_12Graph_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_Reader___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__12)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_12Graph_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_Reader___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__13)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":176
 *     def has_edges(self):
 *         return self.source.hasEdges()
 * pytact.graph_api_capnp.Graph.schema = _Schema()._init(getGraph_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class IntP_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<Graph>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Graph); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11IntP_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntP_Reader___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__14)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11IntP_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntP_Reader___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__15)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":200
 *     def value(self):
 *         return self.source.getValue()
 * pytact.graph_api_capnp.IntP.schema = _Schema()._init(getIntP_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class Definition_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<IntP>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_IntP); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_17Definition_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Definition_Reader___reduce_cytho, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__16)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_17Definition_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Definition_Reader___setstate_cyt, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__17)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":373
 *     def has_term_text(self):
 *         return self.source.hasTermText()
 * pytact.graph_api_capnp.Definition.schema = _Schema()._init(getDefinition_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class FloatP_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<Definition>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Definition); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 373, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_13FloatP_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FloatP_Reader___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__18)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_13FloatP_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FloatP_Reader___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__19)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":397
 *     def value(self):
 *         return self.source.getValue()
 * pytact.graph_api_capnp.FloatP.schema = _Schema()._init(getFloatP_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class Graph_Node_Label_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<FloatP>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_FloatP); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 397, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_Node_Label_Reader___reduce, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_23Graph_Node_Label_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_Node_Label_Reader___setsta, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__21)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Definition_Status_Substituted_Re_2, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__22)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_36Definition_Status_Substituted_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Definition_Status_Substituted_Re_3, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Definition_Status_Reader___reduc, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_24Definition_Status_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Definition_Status_Reader___setst, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofStep_Reader___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__26)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_16ProofStep_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofStep_Reader___setstate_cyth, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":768
 *     def has_outcomes(self):
 *         return self.source.hasOutcomes()
 * pytact.graph_api_capnp.ProofStep.schema = _Schema()._init(getProofStep_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class Tactic_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<ProofStep>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ProofStep); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_13Tactic_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Tactic_Reader___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_13Tactic_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Tactic_Reader___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":866
 *     def has_text_non_anonymous(self):
 *         return self.source.hasTextNonAnonymous()
 * pytact.graph_api_capnp.Tactic.schema = _Schema()._init(getTactic_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class ProofStep_Tactic_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<Tactic>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Tactic); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 866, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofStep_Tactic_Reader___reduce, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_23ProofStep_Tactic_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofStep_Tactic_Reader___setsta, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_14Outcome_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Outcome_Reader___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_14Outcome_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Outcome_Reader___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":994
 *     def has_tactic_arguments(self):
 *         return self.source.hasTacticArguments()
 * pytact.graph_api_capnp.Outcome.schema = _Schema()._init(getOutcome_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class Outcome_Term_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<Outcome>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Outcome); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Outcome_Term_Reader___reduce_cyt, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_19Outcome_Term_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Outcome_Term_Reader___setstate_c, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_17ProofState_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofState_Reader___reduce_cytho, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_17ProofState_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofState_Reader___setstate_cyt, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1133
 *         """
 *         return self.source.getId()
 * pytact.graph_api_capnp.ProofState.schema = _Schema()._init(getProofState_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class Argument_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<ProofState>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ProofState); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_15Argument_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Argument_Reader___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_15Argument_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Argument_Reader___setstate_cytho, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1182
 *     def is_term(self):
 *         return self.source.isTerm()
 * pytact.graph_api_capnp.Argument.schema = _Schema()._init(getArgument_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class ProofState_Root_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<Argument>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Argument); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofState_Root_Reader___reduce, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_22ProofState_Root_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofState_Root_Reader___setstat, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11Node_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Node_Reader___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11Node_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Node_Reader___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1257
 *         """
 *         return self.source.getNodeIndex()
 * pytact.graph_api_capnp.Node.schema = _Schema()._init(getNode_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class Argument_Term_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<Node>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Node); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Argument_Term_Reader___reduce_cy, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_20Argument_Term_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Argument_Term_Reader___setstate, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_EdgeTarget_Target_Reader_2, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_30Graph_EdgeTarget_Target_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_EdgeTarget_Target_Reader_3, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__47)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AbstractTactic_Reader___reduce_c, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_21AbstractTactic_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AbstractTactic_Reader___setstate, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1362
 *         """
 *         return self.source.getParameters()
 * pytact.graph_api_capnp.AbstractTactic.schema = _Schema()._init(getAbstractTactic_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class DataVersion_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<AbstractTactic>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_AbstractTactic); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1362, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DataVersion_Reader___reduce_cyth, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_18DataVersion_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DataVersion_Reader___setstate_cy, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1396
 *         """
 *         return self.source.getMajor()
 * pytact.graph_api_capnp.DataVersion.schema = _Schema()._init(getDataVersion_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class Dataset_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<DataVersion>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_DataVersion); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_14Dataset_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Dataset_Reader___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_14Dataset_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Dataset_Reader___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1484
 *     def has_data_version(self):
 *         return self.source.hasDataVersion()
 * pytact.graph_api_capnp.Dataset.schema = _Schema()._init(getDataset_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class Exception_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<Dataset>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Dataset); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_16Exception_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Exception_Reader___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_16Exception_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Exception_Reader___setstate_cyth, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1541
 *     def is_illegal_argument(self):
 *         return self.source.isIllegalArgument()
 * pytact.graph_api_capnp.Exception.schema = _Schema()._init(getException_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class ExecutionResult_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<Exception>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1541, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ExecutionResult_Reader___reduce, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_22ExecutionResult_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ExecutionResult_Reader___setstat, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__57)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1613
 *     def has_protocol_error(self):
 *         return self.source.hasProtocolError()
 * pytact.graph_api_capnp.ExecutionResult.schema = _Schema()._init(getExecutionResult_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class ExecutionResult_NewState_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<ExecutionResult>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ExecutionResult); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1613, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ExecutionResult_NewState_Reader_2, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_31ExecutionResult_NewState_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ExecutionResult_NewState_Reader_3, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__59)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Reader___redu, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_25PredictionProtocol_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Reader___sets, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1674
 *     def __repr__(self):
 *         return repr(self.dynamic)
 * pytact.graph_api_capnp.PredictionProtocol.schema = _Schema()._init(getPredictionProtocol_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class PredictionProtocol_Request_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<PredictionProtocol>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_PredictionProtocol); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Request_Reade_2, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_33PredictionProtocol_Request_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Request_Reade_3, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1743
 *     def is_check_alignment(self):
 *         return self.source.isCheckAlignment()
 * pytact.graph_api_capnp.PredictionProtocol.Request.schema = _Schema()._init(getPredictionProtocol_Request_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class PredictionProtocol_Request_Initialize_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<PredictionProtocol::Request>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_PredictionProtocol); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Request); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1743, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Request_Initi_2, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Request_Initialize_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Request_Initi_3, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__65)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Request_Predi_2, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Request_Predict_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Request_Predi_3, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Prediction_Re_3, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__68)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_36PredictionProtocol_Prediction_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Prediction_Re_4, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__69)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1912
 *     def confidence(self):
 *         return self.source.getConfidence()
 * pytact.graph_api_capnp.PredictionProtocol.Prediction.schema = _Schema()._init(getPredictionProtocol_Prediction_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class PredictionProtocol_TextPrediction_Reader:
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_4)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_4), capnp::Schema::from<PredictionProtocol::Prediction>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pytact); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_PredictionProtocol); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Prediction); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1912, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_TextPredictio_3, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_40PredictionProtocol_TextPrediction_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_TextPredictio_4, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__71)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1944
 *     def confidence(self):
 *         return self.source.getConfidence()
 * pytact.graph_api_capnp.PredictionProtocol.TextPrediction.schema = _Schema()._init(getPredictionProtocol_TextPrediction_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class PredictionProtocol_Response_Reader:
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<PredictionProtocol::TextPrediction>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_PredictionProtocol); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TextPrediction); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 1944, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Response_Read_2, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_34PredictionProtocol_Response_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Response_Read_3, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2018
 *     def is_alignment(self):
 *         return self.source.isAlignment()
 * pytact.graph_api_capnp.PredictionProtocol.Response.schema = _Schema()._init(getPredictionProtocol_Response_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * cdef class PredictionProtocol_Response_Alignment_Reader:
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_4)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_4), capnp::Schema::from<PredictionProtocol::Response>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pytact); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_PredictionProtocol); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Response); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 2018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Response_Alig_2, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__74)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_44PredictionProtocol_Response_Alignment_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Response_Alig_3, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__75)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ConflatableEdges_Reader___reduce, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_23ConflatableEdges_Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ConflatableEdges_Reader___setsta, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":2071
 *     def __repr__(self):
 *         return repr(self.dynamic)
 * pytact.graph_api_capnp.ConflatableEdges.schema = _Schema()._init(getConflatableEdges_Schema()).as_struct()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5capnp_3lib_5capnp__Schema)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5capnp_3lib_5capnp__Schema *)((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2)->__pyx_vtab)->_init(((struct __pyx_obj_5capnp_3lib_5capnp__Schema *)__pyx_t_2), capnp::Schema::from<ConflatableEdges>()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_as_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pytact); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_graph_api_capnp); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ConflatableEdges); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_schema, __pyx_t_3) < 0) __PYX_ERR(1, 2071, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint16_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint16_List___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__78)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint16_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint16_List___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint32_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint32_List___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__80)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint32_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint32_List___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__81)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint64_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64_List___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11Uint64_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64_List___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__83)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_10Int64_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64_List___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__84)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_10Int64_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64_List___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11String_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_String_List___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__86)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_11String_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_String_List___setstate_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__87)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Prediction_Re_5, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_41PredictionProtocol_Prediction_Reader_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_Prediction_Re_6, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__89)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_Node_Reader_List___reduce, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__90)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_22Graph_Node_Reader_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_Node_Reader_List___setstat, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AbstractTactic_Reader_List___red, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_26AbstractTactic_Reader_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AbstractTactic_Reader_List___set, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__93)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofStep_Reader_List___reduce_c, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_21ProofStep_Reader_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofStep_Reader_List___setstate, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__95)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_TextPredictio_5, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__96)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_45PredictionProtocol_TextPrediction_Reader_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PredictionProtocol_TextPredictio_6, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofState_Reader_List___reduce, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__98)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_22ProofState_Reader_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProofState_Reader_List___setstat, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_EdgeTarget_Reader_List___r, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_28Graph_EdgeTarget_Reader_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Graph_EdgeTarget_Reader_List___s, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__101)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Argument_Reader_List___reduce_cy, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__102)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_20Argument_Reader_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Argument_Reader_List___setstate, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Outcome_Reader_List___reduce_cyt, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__104)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_19Outcome_Reader_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Outcome_Reader_List___setstate_c, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__105)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_16Node_Reader_List_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Node_Reader_List___reduce_cython, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.source cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pytact_22graph_api_capnp_cython_16Node_Reader_List_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Node_Reader_List___setstate_cyth, NULL, __pyx_n_s_pytact_graph_api_capnp_cython, __pyx_d, ((PyObject *)__pyx_codeobj__107)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pytact/graph_api_capnp_cython.pyx":1
 * # distutils: language = c++             # <<<<<<<<<<<<<<
 * # distutils: libraries = capnpc capnp capnp-rpc
 * # distutils: sources = pytact/graph_api.capnp.cpp
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":9
 * 
 * cdef object __Pyx_OrderedDict
 * from collections import OrderedDict as __Pyx_OrderedDict             # <<<<<<<<<<<<<<
 * 
 * @cython.internal
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_OrderedDict);
  __Pyx_GIVEREF(__pyx_n_s_OrderedDict);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_OrderedDict);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_XGOTREF(__Pyx_OrderedDict);
  __Pyx_DECREF_SET(__Pyx_OrderedDict, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Pyx_EnumMeta___reduce_cython, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype___Pyx_EnumMeta->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype___Pyx_EnumMeta);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Pyx_EnumMeta___setstate_cython, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__110)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype___Pyx_EnumMeta->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype___Pyx_EnumMeta);

  /* "EnumBase":23
 * 
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int, metaclass=__Pyx_EnumMeta):             # <<<<<<<<<<<<<<
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)(&PyInt_Type));
  __Pyx_GIVEREF((PyObject *)(&PyInt_Type));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyInt_Type)));
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(((PyObject *)__pyx_ptype___Pyx_EnumMeta), __pyx_t_2, __pyx_n_s_Pyx_EnumBase, __pyx_n_s_Pyx_EnumBase, __pyx_t_4, __pyx_n_s_EnumBase, (PyObject *) NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_2 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 23, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "EnumBase":24
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int, metaclass=__Pyx_EnumMeta):
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_1__new__, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_Pyx_EnumBase___new, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__113);
  if (__Pyx_SetNewInClass(__pyx_t_5, __pyx_n_s_new, __pyx_t_3) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "EnumBase":35
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_3__repr__, 0, __pyx_n_s_Pyx_EnumBase___repr, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__114)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_repr, __pyx_t_3) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "EnumBase":37
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_5__str__, 0, __pyx_n_s_Pyx_EnumBase___str, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__115)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_str, __pyx_t_3) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "EnumBase":23
 * 
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int, metaclass=__Pyx_EnumMeta):             # <<<<<<<<<<<<<<
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(((PyObject *)__pyx_ptype___Pyx_EnumMeta), __pyx_n_s_Pyx_EnumBase, __pyx_t_2, __pyx_t_5, __pyx_t_4, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(__Pyx_EnumBase);
  __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":40
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 *     from enum import IntEnum as __Pyx_EnumBase
 * 
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "EnumBase":41
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 *     from enum import IntEnum as __Pyx_EnumBase             # <<<<<<<<<<<<<<
 * 
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_IntEnum);
    __Pyx_GIVEREF(__pyx_n_s_IntEnum);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_IntEnum);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_enum, __pyx_t_2, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XGOTREF(__Pyx_EnumBase);
    __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumBase":40
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 *     from enum import IntEnum as __Pyx_EnumBase
 * 
 */
  }

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta, 0, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__117)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "CppScopedEnumType":70
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(PyDict_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_4))) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["Argument_Which"] = __Pyx_EnumBase('Argument_Which', __Pyx_OrderedDict([
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "CppScopedEnumType":75
 * 
 *     __Pyx_globals["Argument_Which"] = __Pyx_EnumBase('Argument_Which', __Pyx_OrderedDict([
 *         ('UNRESOLVABLE', <uint16_t>(Argument_Which.UNRESOLVABLE)),             # <<<<<<<<<<<<<<
 *         ('TERM', <uint16_t>(Argument_Which.TERM)),
 *     ]))
 */
    __pyx_t_4 = __Pyx_PyInt_From_uint16_t(((uint16_t)Argument::Which::UNRESOLVABLE)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_UNRESOLVABLE);
    __Pyx_GIVEREF(__pyx_n_s_UNRESOLVABLE);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_UNRESOLVABLE);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "CppScopedEnumType":76
 *     __Pyx_globals["Argument_Which"] = __Pyx_EnumBase('Argument_Which', __Pyx_OrderedDict([
 *         ('UNRESOLVABLE', <uint16_t>(Argument_Which.UNRESOLVABLE)),
 *         ('TERM', <uint16_t>(Argument_Which.TERM)),             # <<<<<<<<<<<<<<
 *     ]))
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_uint16_t(((uint16_t)Argument::Which::TERM)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_TERM);
    __Pyx_GIVEREF(__pyx_n_s_TERM);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_TERM);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "CppScopedEnumType":74
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     __Pyx_globals["Argument_Which"] = __Pyx_EnumBase('Argument_Which', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('UNRESOLVABLE', <uint16_t>(Argument_Which.UNRESOLVABLE)),
 *         ('TERM', <uint16_t>(Argument_Which.TERM)),
 */
    __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_2 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_Argument_Which);
    __Pyx_GIVEREF(__pyx_n_s_Argument_Which);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_Argument_Which);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_Argument_Which, __pyx_t_5) < 0))) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["Argument_Which"] = __Pyx_EnumBase('Argument_Which', __Pyx_OrderedDict([
 */
    goto __pyx_L3;
  }

  /* "CppScopedEnumType":80
 * 
 * else:
 *     __Pyx_globals["Argument_Which"] = type('Argument_Which', (__Pyx_EnumBase,), {})             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Argument_Which"](<uint16_t>(Argument_Which.UNRESOLVABLE), 'UNRESOLVABLE')
 *     __Pyx_globals["Argument_Which"](<uint16_t>(Argument_Which.TERM), 'TERM')
 */
  /*else*/ {
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __Pyx_EnumBase);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_Argument_Which);
    __Pyx_GIVEREF(__pyx_n_s_Argument_Which);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Argument_Which);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 80, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_Argument_Which, __pyx_t_4) < 0))) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "CppScopedEnumType":81
 * else:
 *     __Pyx_globals["Argument_Which"] = type('Argument_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["Argument_Which"](<uint16_t>(Argument_Which.UNRESOLVABLE), 'UNRESOLVABLE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Argument_Which"](<uint16_t>(Argument_Which.TERM), 'TERM')
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 81, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Argument_Which); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyInt_From_uint16_t(((uint16_t)Argument::Which::UNRESOLVABLE)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_UNRESOLVABLE);
    __Pyx_GIVEREF(__pyx_n_s_UNRESOLVABLE);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_n_s_UNRESOLVABLE);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "CppScopedEnumType":82
 *     __Pyx_globals["Argument_Which"] = type('Argument_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["Argument_Which"](<uint16_t>(Argument_Which.UNRESOLVABLE), 'UNRESOLVABLE')
 *     __Pyx_globals["Argument_Which"](<uint16_t>(Argument_Which.TERM), 'TERM')             # <<<<<<<<<<<<<<
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 82, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Argument_Which); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Argument::Which::TERM)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_TERM);
    __Pyx_GIVEREF(__pyx_n_s_TERM);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_TERM);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_L3:;

  /* "CppScopedEnumType":70
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __pyx_t_5 = __Pyx_Globals(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (!(likely(PyDict_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_5))) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_5));
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["Definition_Status_Which"] = __Pyx_EnumBase('Definition_Status_Which', __Pyx_OrderedDict([
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "CppScopedEnumType":75
 * 
 *     __Pyx_globals["Definition_Status_Which"] = __Pyx_EnumBase('Definition_Status_Which', __Pyx_OrderedDict([
 *         ('ORIGINAL', <uint16_t>(Definition_Status_Which.ORIGINAL)),             # <<<<<<<<<<<<<<
 *         ('DISCHARGED', <uint16_t>(Definition_Status_Which.DISCHARGED)),
 *         ('SUBSTITUTED', <uint16_t>(Definition_Status_Which.SUBSTITUTED)),
 */
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Status::Which::ORIGINAL)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_ORIGINAL);
    __Pyx_GIVEREF(__pyx_n_s_ORIGINAL);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_ORIGINAL);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "CppScopedEnumType":76
 *     __Pyx_globals["Definition_Status_Which"] = __Pyx_EnumBase('Definition_Status_Which', __Pyx_OrderedDict([
 *         ('ORIGINAL', <uint16_t>(Definition_Status_Which.ORIGINAL)),
 *         ('DISCHARGED', <uint16_t>(Definition_Status_Which.DISCHARGED)),             # <<<<<<<<<<<<<<
 *         ('SUBSTITUTED', <uint16_t>(Definition_Status_Which.SUBSTITUTED)),
 *     ]))
 */
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Status::Which::DISCHARGED)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_DISCHARGED);
    __Pyx_GIVEREF(__pyx_n_s_DISCHARGED);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_DISCHARGED);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "CppScopedEnumType":77
 *         ('ORIGINAL', <uint16_t>(Definition_Status_Which.ORIGINAL)),
 *         ('DISCHARGED', <uint16_t>(Definition_Status_Which.DISCHARGED)),
 *         ('SUBSTITUTED', <uint16_t>(Definition_Status_Which.SUBSTITUTED)),             # <<<<<<<<<<<<<<
 *     ]))
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Status::Which::SUBSTITUTED)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_n_s_SUBSTITUTED);
    __Pyx_GIVEREF(__pyx_n_s_SUBSTITUTED);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_SUBSTITUTED);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "CppScopedEnumType":74
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     __Pyx_globals["Definition_Status_Which"] = __Pyx_EnumBase('Definition_Status_Which', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('ORIGINAL', <uint16_t>(Definition_Status_Which.ORIGINAL)),
 *         ('DISCHARGED', <uint16_t>(Definition_Status_Which.DISCHARGED)),
 */
    __pyx_t_5 = PyList_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_Definition_Status_Which);
    __Pyx_GIVEREF(__pyx_n_s_Definition_Status_Which);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_Definition_Status_Which);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_Definition_Status_Which, __pyx_t_3) < 0))) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["Definition_Status_Which"] = __Pyx_EnumBase('Definition_Status_Which', __Pyx_OrderedDict([
 */
    goto __pyx_L4;
  }

  /* "CppScopedEnumType":81
 * 
 * else:
 *     __Pyx_globals["Definition_Status_Which"] = type('Definition_Status_Which', (__Pyx_EnumBase,), {})             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.ORIGINAL), 'ORIGINAL')
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.DISCHARGED), 'DISCHARGED')
 */
  /*else*/ {
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __Pyx_EnumBase);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_Definition_Status_Which);
    __Pyx_GIVEREF(__pyx_n_s_Definition_Status_Which);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Definition_Status_Which);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 81, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_Definition_Status_Which, __pyx_t_5) < 0))) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "CppScopedEnumType":82
 * else:
 *     __Pyx_globals["Definition_Status_Which"] = type('Definition_Status_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.ORIGINAL), 'ORIGINAL')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.DISCHARGED), 'DISCHARGED')
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.SUBSTITUTED), 'SUBSTITUTED')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 82, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Definition_Status_Which); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Status::Which::ORIGINAL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_ORIGINAL);
    __Pyx_GIVEREF(__pyx_n_s_ORIGINAL);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_ORIGINAL);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "CppScopedEnumType":83
 *     __Pyx_globals["Definition_Status_Which"] = type('Definition_Status_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.ORIGINAL), 'ORIGINAL')
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.DISCHARGED), 'DISCHARGED')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.SUBSTITUTED), 'SUBSTITUTED')
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 83, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Definition_Status_Which); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Status::Which::DISCHARGED)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_n_s_DISCHARGED);
    __Pyx_GIVEREF(__pyx_n_s_DISCHARGED);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_n_s_DISCHARGED);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "CppScopedEnumType":84
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.ORIGINAL), 'ORIGINAL')
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.DISCHARGED), 'DISCHARGED')
 *     __Pyx_globals["Definition_Status_Which"](<uint16_t>(Definition_Status_Which.SUBSTITUTED), 'SUBSTITUTED')             # <<<<<<<<<<<<<<
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 84, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Definition_Status_Which); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Status::Which::SUBSTITUTED)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_SUBSTITUTED);
    __Pyx_GIVEREF(__pyx_n_s_SUBSTITUTED);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_SUBSTITUTED);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_L4:;

  /* "CppScopedEnumType":70
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __pyx_t_5 = __Pyx_Globals(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (!(likely(PyDict_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_5))) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_5));
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["Definition_Which"] = __Pyx_EnumBase('Definition_Which', __Pyx_OrderedDict([
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "CppScopedEnumType":75
 * 
 *     __Pyx_globals["Definition_Which"] = __Pyx_EnumBase('Definition_Which', __Pyx_OrderedDict([
 *         ('INDUCTIVE', <uint16_t>(Definition_Which.INDUCTIVE)),             # <<<<<<<<<<<<<<
 *         ('CONSTRUCTOR', <uint16_t>(Definition_Which.CONSTRUCTOR)),
 *         ('PROJECTION', <uint16_t>(Definition_Which.PROJECTION)),
 */
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::INDUCTIVE)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_INDUCTIVE);
    __Pyx_GIVEREF(__pyx_n_s_INDUCTIVE);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_INDUCTIVE);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "CppScopedEnumType":76
 *     __Pyx_globals["Definition_Which"] = __Pyx_EnumBase('Definition_Which', __Pyx_OrderedDict([
 *         ('INDUCTIVE', <uint16_t>(Definition_Which.INDUCTIVE)),
 *         ('CONSTRUCTOR', <uint16_t>(Definition_Which.CONSTRUCTOR)),             # <<<<<<<<<<<<<<
 *         ('PROJECTION', <uint16_t>(Definition_Which.PROJECTION)),
 *         ('MANUAL_CONSTANT', <uint16_t>(Definition_Which.MANUAL_CONSTANT)),
 */
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::CONSTRUCTOR)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_n_s_CONSTRUCTOR);
    __Pyx_GIVEREF(__pyx_n_s_CONSTRUCTOR);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_CONSTRUCTOR);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "CppScopedEnumType":77
 *         ('INDUCTIVE', <uint16_t>(Definition_Which.INDUCTIVE)),
 *         ('CONSTRUCTOR', <uint16_t>(Definition_Which.CONSTRUCTOR)),
 *         ('PROJECTION', <uint16_t>(Definition_Which.PROJECTION)),             # <<<<<<<<<<<<<<
 *         ('MANUAL_CONSTANT', <uint16_t>(Definition_Which.MANUAL_CONSTANT)),
 *         ('TACTICAL_CONSTANT', <uint16_t>(Definition_Which.TACTICAL_CONSTANT)),
 */
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::PROJECTION)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_PROJECTION);
    __Pyx_GIVEREF(__pyx_n_s_PROJECTION);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_PROJECTION);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "CppScopedEnumType":78
 *         ('CONSTRUCTOR', <uint16_t>(Definition_Which.CONSTRUCTOR)),
 *         ('PROJECTION', <uint16_t>(Definition_Which.PROJECTION)),
 *         ('MANUAL_CONSTANT', <uint16_t>(Definition_Which.MANUAL_CONSTANT)),             # <<<<<<<<<<<<<<
 *         ('TACTICAL_CONSTANT', <uint16_t>(Definition_Which.TACTICAL_CONSTANT)),
 *         ('MANUAL_SECTION_CONSTANT', <uint16_t>(Definition_Which.MANUAL_SECTION_CONSTANT)),
 */
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::MANUAL_CONSTANT)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_n_s_MANUAL_CONSTANT);
    __Pyx_GIVEREF(__pyx_n_s_MANUAL_CONSTANT);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_MANUAL_CONSTANT);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "CppScopedEnumType":79
 *         ('PROJECTION', <uint16_t>(Definition_Which.PROJECTION)),
 *         ('MANUAL_CONSTANT', <uint16_t>(Definition_Which.MANUAL_CONSTANT)),
 *         ('TACTICAL_CONSTANT', <uint16_t>(Definition_Which.TACTICAL_CONSTANT)),             # <<<<<<<<<<<<<<
 *         ('MANUAL_SECTION_CONSTANT', <uint16_t>(Definition_Which.MANUAL_SECTION_CONSTANT)),
 *         ('TACTICAL_SECTION_CONSTANT', <uint16_t>(Definition_Which.TACTICAL_SECTION_CONSTANT)),
 */
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::TACTICAL_CONSTANT)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_TACTICAL_CONSTANT);
    __Pyx_GIVEREF(__pyx_n_s_TACTICAL_CONSTANT);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_TACTICAL_CONSTANT);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "CppScopedEnumType":80
 *         ('MANUAL_CONSTANT', <uint16_t>(Definition_Which.MANUAL_CONSTANT)),
 *         ('TACTICAL_CONSTANT', <uint16_t>(Definition_Which.TACTICAL_CONSTANT)),
 *         ('MANUAL_SECTION_CONSTANT', <uint16_t>(Definition_Which.MANUAL_SECTION_CONSTANT)),             # <<<<<<<<<<<<<<
 *         ('TACTICAL_SECTION_CONSTANT', <uint16_t>(Definition_Which.TACTICAL_SECTION_CONSTANT)),
 *     ]))
 */
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::MANUAL_SECTION_CONSTANT)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_MANUAL_SECTION_CONSTANT);
    __Pyx_GIVEREF(__pyx_n_s_MANUAL_SECTION_CONSTANT);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_MANUAL_SECTION_CONSTANT);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "CppScopedEnumType":81
 *         ('TACTICAL_CONSTANT', <uint16_t>(Definition_Which.TACTICAL_CONSTANT)),
 *         ('MANUAL_SECTION_CONSTANT', <uint16_t>(Definition_Which.MANUAL_SECTION_CONSTANT)),
 *         ('TACTICAL_SECTION_CONSTANT', <uint16_t>(Definition_Which.TACTICAL_SECTION_CONSTANT)),             # <<<<<<<<<<<<<<
 *     ]))
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::TACTICAL_SECTION_CONSTANT)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_n_s_TACTICAL_SECTION_CONSTANT);
    __Pyx_GIVEREF(__pyx_n_s_TACTICAL_SECTION_CONSTANT);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_TACTICAL_SECTION_CONSTANT);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "CppScopedEnumType":74
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     __Pyx_globals["Definition_Which"] = __Pyx_EnumBase('Definition_Which', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('INDUCTIVE', <uint16_t>(Definition_Which.INDUCTIVE)),
 *         ('CONSTRUCTOR', <uint16_t>(Definition_Which.CONSTRUCTOR)),
 */
    __pyx_t_5 = PyList_New(7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_7);
    PyList_SET_ITEM(__pyx_t_5, 3, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyList_SET_ITEM(__pyx_t_5, 4, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyList_SET_ITEM(__pyx_t_5, 5, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_10);
    PyList_SET_ITEM(__pyx_t_5, 6, __pyx_t_10);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_Definition_Which);
    __Pyx_GIVEREF(__pyx_n_s_Definition_Which);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_Definition_Which);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_5, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_Definition_Which, __pyx_t_10) < 0))) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["Definition_Which"] = __Pyx_EnumBase('Definition_Which', __Pyx_OrderedDict([
 */
    goto __pyx_L5;
  }

  /* "CppScopedEnumType":85
 * 
 * else:
 *     __Pyx_globals["Definition_Which"] = type('Definition_Which', (__Pyx_EnumBase,), {})             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.INDUCTIVE), 'INDUCTIVE')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.CONSTRUCTOR), 'CONSTRUCTOR')
 */
  /*else*/ {
    __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __Pyx_EnumBase);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_Definition_Which);
    __Pyx_GIVEREF(__pyx_n_s_Definition_Which);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_Definition_Which);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_5);
    __pyx_t_10 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 85, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_Definition_Which, __pyx_t_5) < 0))) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "CppScopedEnumType":86
 * else:
 *     __Pyx_globals["Definition_Which"] = type('Definition_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.INDUCTIVE), 'INDUCTIVE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.CONSTRUCTOR), 'CONSTRUCTOR')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.PROJECTION), 'PROJECTION')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 86, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Definition_Which); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::INDUCTIVE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_INDUCTIVE);
    __Pyx_GIVEREF(__pyx_n_s_INDUCTIVE);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_n_s_INDUCTIVE);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":87
 *     __Pyx_globals["Definition_Which"] = type('Definition_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.INDUCTIVE), 'INDUCTIVE')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.CONSTRUCTOR), 'CONSTRUCTOR')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.PROJECTION), 'PROJECTION')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.MANUAL_CONSTANT), 'MANUAL_CONSTANT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 87, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Definition_Which); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::CONSTRUCTOR)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_10);
    __Pyx_INCREF(__pyx_n_s_CONSTRUCTOR);
    __Pyx_GIVEREF(__pyx_n_s_CONSTRUCTOR);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_n_s_CONSTRUCTOR);
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_5, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "CppScopedEnumType":88
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.INDUCTIVE), 'INDUCTIVE')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.CONSTRUCTOR), 'CONSTRUCTOR')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.PROJECTION), 'PROJECTION')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.MANUAL_CONSTANT), 'MANUAL_CONSTANT')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.TACTICAL_CONSTANT), 'TACTICAL_CONSTANT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 88, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Definition_Which); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::PROJECTION)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_PROJECTION);
    __Pyx_GIVEREF(__pyx_n_s_PROJECTION);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_PROJECTION);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "CppScopedEnumType":89
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.CONSTRUCTOR), 'CONSTRUCTOR')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.PROJECTION), 'PROJECTION')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.MANUAL_CONSTANT), 'MANUAL_CONSTANT')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.TACTICAL_CONSTANT), 'TACTICAL_CONSTANT')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.MANUAL_SECTION_CONSTANT), 'MANUAL_SECTION_CONSTANT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 89, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Definition_Which); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::MANUAL_CONSTANT)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_MANUAL_CONSTANT);
    __Pyx_GIVEREF(__pyx_n_s_MANUAL_CONSTANT);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_n_s_MANUAL_CONSTANT);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":90
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.PROJECTION), 'PROJECTION')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.MANUAL_CONSTANT), 'MANUAL_CONSTANT')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.TACTICAL_CONSTANT), 'TACTICAL_CONSTANT')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.MANUAL_SECTION_CONSTANT), 'MANUAL_SECTION_CONSTANT')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.TACTICAL_SECTION_CONSTANT), 'TACTICAL_SECTION_CONSTANT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 90, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Definition_Which); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::TACTICAL_CONSTANT)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_10);
    __Pyx_INCREF(__pyx_n_s_TACTICAL_CONSTANT);
    __Pyx_GIVEREF(__pyx_n_s_TACTICAL_CONSTANT);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_n_s_TACTICAL_CONSTANT);
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_5, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "CppScopedEnumType":91
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.MANUAL_CONSTANT), 'MANUAL_CONSTANT')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.TACTICAL_CONSTANT), 'TACTICAL_CONSTANT')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.MANUAL_SECTION_CONSTANT), 'MANUAL_SECTION_CONSTANT')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.TACTICAL_SECTION_CONSTANT), 'TACTICAL_SECTION_CONSTANT')
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 91, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Definition_Which); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_5 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::MANUAL_SECTION_CONSTANT)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_MANUAL_SECTION_CONSTANT);
    __Pyx_GIVEREF(__pyx_n_s_MANUAL_SECTION_CONSTANT);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_MANUAL_SECTION_CONSTANT);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "CppScopedEnumType":92
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.TACTICAL_CONSTANT), 'TACTICAL_CONSTANT')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.MANUAL_SECTION_CONSTANT), 'MANUAL_SECTION_CONSTANT')
 *     __Pyx_globals["Definition_Which"](<uint16_t>(Definition_Which.TACTICAL_SECTION_CONSTANT), 'TACTICAL_SECTION_CONSTANT')             # <<<<<<<<<<<<<<
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 92, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Definition_Which); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)Definition::Which::TACTICAL_SECTION_CONSTANT)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_TACTICAL_SECTION_CONSTANT);
    __Pyx_GIVEREF(__pyx_n_s_TACTICAL_SECTION_CONSTANT);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_n_s_TACTICAL_SECTION_CONSTANT);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __pyx_L5:;

  /* "CppScopedEnumType":70
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __pyx_t_9 = __Pyx_Globals(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (!(likely(PyDict_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_9))) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_9));
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;

  /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["EdgeClassification"] = __Pyx_EnumBase('EdgeClassification', __Pyx_OrderedDict([
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "CppScopedEnumType":75
 * 
 *     __Pyx_globals["EdgeClassification"] = __Pyx_EnumBase('EdgeClassification', __Pyx_OrderedDict([
 *         ('CONTEXT_ELEM', <uint16_t>(EdgeClassification.CONTEXT_ELEM)),             # <<<<<<<<<<<<<<
 *         ('CONTEXT_SUBJECT', <uint16_t>(EdgeClassification.CONTEXT_SUBJECT)),
 *         ('CONTEXT_DEF_TYPE', <uint16_t>(EdgeClassification.CONTEXT_DEF_TYPE)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONTEXT_ELEM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_ELEM);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_ELEM);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_CONTEXT_ELEM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":76
 *     __Pyx_globals["EdgeClassification"] = __Pyx_EnumBase('EdgeClassification', __Pyx_OrderedDict([
 *         ('CONTEXT_ELEM', <uint16_t>(EdgeClassification.CONTEXT_ELEM)),
 *         ('CONTEXT_SUBJECT', <uint16_t>(EdgeClassification.CONTEXT_SUBJECT)),             # <<<<<<<<<<<<<<
 *         ('CONTEXT_DEF_TYPE', <uint16_t>(EdgeClassification.CONTEXT_DEF_TYPE)),
 *         ('CONTEXT_DEF_TERM', <uint16_t>(EdgeClassification.CONTEXT_DEF_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONTEXT_SUBJECT)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_SUBJECT);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_SUBJECT);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_CONTEXT_SUBJECT);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":77
 *         ('CONTEXT_ELEM', <uint16_t>(EdgeClassification.CONTEXT_ELEM)),
 *         ('CONTEXT_SUBJECT', <uint16_t>(EdgeClassification.CONTEXT_SUBJECT)),
 *         ('CONTEXT_DEF_TYPE', <uint16_t>(EdgeClassification.CONTEXT_DEF_TYPE)),             # <<<<<<<<<<<<<<
 *         ('CONTEXT_DEF_TERM', <uint16_t>(EdgeClassification.CONTEXT_DEF_TERM)),
 *         ('CONST_TYPE', <uint16_t>(EdgeClassification.CONST_TYPE)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONTEXT_DEF_TYPE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_DEF_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_DEF_TYPE);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_CONTEXT_DEF_TYPE);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":78
 *         ('CONTEXT_SUBJECT', <uint16_t>(EdgeClassification.CONTEXT_SUBJECT)),
 *         ('CONTEXT_DEF_TYPE', <uint16_t>(EdgeClassification.CONTEXT_DEF_TYPE)),
 *         ('CONTEXT_DEF_TERM', <uint16_t>(EdgeClassification.CONTEXT_DEF_TERM)),             # <<<<<<<<<<<<<<
 *         ('CONST_TYPE', <uint16_t>(EdgeClassification.CONST_TYPE)),
 *         ('CONST_UNDEF', <uint16_t>(EdgeClassification.CONST_UNDEF)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONTEXT_DEF_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_DEF_TERM);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_DEF_TERM);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_CONTEXT_DEF_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":79
 *         ('CONTEXT_DEF_TYPE', <uint16_t>(EdgeClassification.CONTEXT_DEF_TYPE)),
 *         ('CONTEXT_DEF_TERM', <uint16_t>(EdgeClassification.CONTEXT_DEF_TERM)),
 *         ('CONST_TYPE', <uint16_t>(EdgeClassification.CONST_TYPE)),             # <<<<<<<<<<<<<<
 *         ('CONST_UNDEF', <uint16_t>(EdgeClassification.CONST_UNDEF)),
 *         ('CONST_DEF', <uint16_t>(EdgeClassification.CONST_DEF)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONST_TYPE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_CONST_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_CONST_TYPE);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_CONST_TYPE);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":80
 *         ('CONTEXT_DEF_TERM', <uint16_t>(EdgeClassification.CONTEXT_DEF_TERM)),
 *         ('CONST_TYPE', <uint16_t>(EdgeClassification.CONST_TYPE)),
 *         ('CONST_UNDEF', <uint16_t>(EdgeClassification.CONST_UNDEF)),             # <<<<<<<<<<<<<<
 *         ('CONST_DEF', <uint16_t>(EdgeClassification.CONST_DEF)),
 *         ('CONST_OPAQUE_DEF', <uint16_t>(EdgeClassification.CONST_OPAQUE_DEF)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONST_UNDEF)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_n_s_CONST_UNDEF);
    __Pyx_GIVEREF(__pyx_n_s_CONST_UNDEF);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_CONST_UNDEF);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":81
 *         ('CONST_TYPE', <uint16_t>(EdgeClassification.CONST_TYPE)),
 *         ('CONST_UNDEF', <uint16_t>(EdgeClassification.CONST_UNDEF)),
 *         ('CONST_DEF', <uint16_t>(EdgeClassification.CONST_DEF)),             # <<<<<<<<<<<<<<
 *         ('CONST_OPAQUE_DEF', <uint16_t>(EdgeClassification.CONST_OPAQUE_DEF)),
 *         ('CONST_PRIMITIVE', <uint16_t>(EdgeClassification.CONST_PRIMITIVE)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONST_DEF)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_CONST_DEF);
    __Pyx_GIVEREF(__pyx_n_s_CONST_DEF);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_CONST_DEF);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":82
 *         ('CONST_UNDEF', <uint16_t>(EdgeClassification.CONST_UNDEF)),
 *         ('CONST_DEF', <uint16_t>(EdgeClassification.CONST_DEF)),
 *         ('CONST_OPAQUE_DEF', <uint16_t>(EdgeClassification.CONST_OPAQUE_DEF)),             # <<<<<<<<<<<<<<
 *         ('CONST_PRIMITIVE', <uint16_t>(EdgeClassification.CONST_PRIMITIVE)),
 *         ('IND_TYPE', <uint16_t>(EdgeClassification.IND_TYPE)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONST_OPAQUE_DEF)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_n_s_CONST_OPAQUE_DEF);
    __Pyx_GIVEREF(__pyx_n_s_CONST_OPAQUE_DEF);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_CONST_OPAQUE_DEF);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":83
 *         ('CONST_DEF', <uint16_t>(EdgeClassification.CONST_DEF)),
 *         ('CONST_OPAQUE_DEF', <uint16_t>(EdgeClassification.CONST_OPAQUE_DEF)),
 *         ('CONST_PRIMITIVE', <uint16_t>(EdgeClassification.CONST_PRIMITIVE)),             # <<<<<<<<<<<<<<
 *         ('IND_TYPE', <uint16_t>(EdgeClassification.IND_TYPE)),
 *         ('IND_CONSTRUCT', <uint16_t>(EdgeClassification.IND_CONSTRUCT)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONST_PRIMITIVE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_n_s_CONST_PRIMITIVE);
    __Pyx_GIVEREF(__pyx_n_s_CONST_PRIMITIVE);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_s_CONST_PRIMITIVE);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":84
 *         ('CONST_OPAQUE_DEF', <uint16_t>(EdgeClassification.CONST_OPAQUE_DEF)),
 *         ('CONST_PRIMITIVE', <uint16_t>(EdgeClassification.CONST_PRIMITIVE)),
 *         ('IND_TYPE', <uint16_t>(EdgeClassification.IND_TYPE)),             # <<<<<<<<<<<<<<
 *         ('IND_CONSTRUCT', <uint16_t>(EdgeClassification.IND_CONSTRUCT)),
 *         ('IND_PROJECTION', <uint16_t>(EdgeClassification.IND_PROJECTION)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::IND_TYPE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_n_s_IND_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_IND_TYPE);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_s_IND_TYPE);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":85
 *         ('CONST_PRIMITIVE', <uint16_t>(EdgeClassification.CONST_PRIMITIVE)),
 *         ('IND_TYPE', <uint16_t>(EdgeClassification.IND_TYPE)),
 *         ('IND_CONSTRUCT', <uint16_t>(EdgeClassification.IND_CONSTRUCT)),             # <<<<<<<<<<<<<<
 *         ('IND_PROJECTION', <uint16_t>(EdgeClassification.IND_PROJECTION)),
 *         ('PROJ_TERM', <uint16_t>(EdgeClassification.PROJ_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::IND_CONSTRUCT)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_IND_CONSTRUCT);
    __Pyx_GIVEREF(__pyx_n_s_IND_CONSTRUCT);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_s_IND_CONSTRUCT);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":86
 *         ('IND_TYPE', <uint16_t>(EdgeClassification.IND_TYPE)),
 *         ('IND_CONSTRUCT', <uint16_t>(EdgeClassification.IND_CONSTRUCT)),
 *         ('IND_PROJECTION', <uint16_t>(EdgeClassification.IND_PROJECTION)),             # <<<<<<<<<<<<<<
 *         ('PROJ_TERM', <uint16_t>(EdgeClassification.PROJ_TERM)),
 *         ('CONSTRUCT_TERM', <uint16_t>(EdgeClassification.CONSTRUCT_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::IND_PROJECTION)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_IND_PROJECTION);
    __Pyx_GIVEREF(__pyx_n_s_IND_PROJECTION);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_s_IND_PROJECTION);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":87
 *         ('IND_CONSTRUCT', <uint16_t>(EdgeClassification.IND_CONSTRUCT)),
 *         ('IND_PROJECTION', <uint16_t>(EdgeClassification.IND_PROJECTION)),
 *         ('PROJ_TERM', <uint16_t>(EdgeClassification.PROJ_TERM)),             # <<<<<<<<<<<<<<
 *         ('CONSTRUCT_TERM', <uint16_t>(EdgeClassification.CONSTRUCT_TERM)),
 *         ('CAST_TERM', <uint16_t>(EdgeClassification.CAST_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::PROJ_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_n_s_PROJ_TERM);
    __Pyx_GIVEREF(__pyx_n_s_PROJ_TERM);
    PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_n_s_PROJ_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":88
 *         ('IND_PROJECTION', <uint16_t>(EdgeClassification.IND_PROJECTION)),
 *         ('PROJ_TERM', <uint16_t>(EdgeClassification.PROJ_TERM)),
 *         ('CONSTRUCT_TERM', <uint16_t>(EdgeClassification.CONSTRUCT_TERM)),             # <<<<<<<<<<<<<<
 *         ('CAST_TERM', <uint16_t>(EdgeClassification.CAST_TERM)),
 *         ('CAST_TYPE', <uint16_t>(EdgeClassification.CAST_TYPE)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONSTRUCT_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_INCREF(__pyx_n_s_CONSTRUCT_TERM);
    __Pyx_GIVEREF(__pyx_n_s_CONSTRUCT_TERM);
    PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_n_s_CONSTRUCT_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":89
 *         ('PROJ_TERM', <uint16_t>(EdgeClassification.PROJ_TERM)),
 *         ('CONSTRUCT_TERM', <uint16_t>(EdgeClassification.CONSTRUCT_TERM)),
 *         ('CAST_TERM', <uint16_t>(EdgeClassification.CAST_TERM)),             # <<<<<<<<<<<<<<
 *         ('CAST_TYPE', <uint16_t>(EdgeClassification.CAST_TYPE)),
 *         ('PROD_TYPE', <uint16_t>(EdgeClassification.PROD_TYPE)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CAST_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_INCREF(__pyx_n_s_CAST_TERM);
    __Pyx_GIVEREF(__pyx_n_s_CAST_TERM);
    PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_n_s_CAST_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":90
 *         ('CONSTRUCT_TERM', <uint16_t>(EdgeClassification.CONSTRUCT_TERM)),
 *         ('CAST_TERM', <uint16_t>(EdgeClassification.CAST_TERM)),
 *         ('CAST_TYPE', <uint16_t>(EdgeClassification.CAST_TYPE)),             # <<<<<<<<<<<<<<
 *         ('PROD_TYPE', <uint16_t>(EdgeClassification.PROD_TYPE)),
 *         ('PROD_TERM', <uint16_t>(EdgeClassification.PROD_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CAST_TYPE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_INCREF(__pyx_n_s_CAST_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_CAST_TYPE);
    PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_n_s_CAST_TYPE);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":91
 *         ('CAST_TERM', <uint16_t>(EdgeClassification.CAST_TERM)),
 *         ('CAST_TYPE', <uint16_t>(EdgeClassification.CAST_TYPE)),
 *         ('PROD_TYPE', <uint16_t>(EdgeClassification.PROD_TYPE)),             # <<<<<<<<<<<<<<
 *         ('PROD_TERM', <uint16_t>(EdgeClassification.PROD_TERM)),
 *         ('LAMBDA_TYPE', <uint16_t>(EdgeClassification.LAMBDA_TYPE)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::PROD_TYPE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_INCREF(__pyx_n_s_PROD_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_PROD_TYPE);
    PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_n_s_PROD_TYPE);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":92
 *         ('CAST_TYPE', <uint16_t>(EdgeClassification.CAST_TYPE)),
 *         ('PROD_TYPE', <uint16_t>(EdgeClassification.PROD_TYPE)),
 *         ('PROD_TERM', <uint16_t>(EdgeClassification.PROD_TERM)),             # <<<<<<<<<<<<<<
 *         ('LAMBDA_TYPE', <uint16_t>(EdgeClassification.LAMBDA_TYPE)),
 *         ('LAMBDA_TERM', <uint16_t>(EdgeClassification.LAMBDA_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::PROD_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_INCREF(__pyx_n_s_PROD_TERM);
    __Pyx_GIVEREF(__pyx_n_s_PROD_TERM);
    PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_s_PROD_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":93
 *         ('PROD_TYPE', <uint16_t>(EdgeClassification.PROD_TYPE)),
 *         ('PROD_TERM', <uint16_t>(EdgeClassification.PROD_TERM)),
 *         ('LAMBDA_TYPE', <uint16_t>(EdgeClassification.LAMBDA_TYPE)),             # <<<<<<<<<<<<<<
 *         ('LAMBDA_TERM', <uint16_t>(EdgeClassification.LAMBDA_TERM)),
 *         ('LET_IN_DEF', <uint16_t>(EdgeClassification.LET_IN_DEF)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::LAMBDA_TYPE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_LAMBDA_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_LAMBDA_TYPE);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_s_LAMBDA_TYPE);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":94
 *         ('PROD_TERM', <uint16_t>(EdgeClassification.PROD_TERM)),
 *         ('LAMBDA_TYPE', <uint16_t>(EdgeClassification.LAMBDA_TYPE)),
 *         ('LAMBDA_TERM', <uint16_t>(EdgeClassification.LAMBDA_TERM)),             # <<<<<<<<<<<<<<
 *         ('LET_IN_DEF', <uint16_t>(EdgeClassification.LET_IN_DEF)),
 *         ('LET_IN_TYPE', <uint16_t>(EdgeClassification.LET_IN_TYPE)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::LAMBDA_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_LAMBDA_TERM);
    __Pyx_GIVEREF(__pyx_n_s_LAMBDA_TERM);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_LAMBDA_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":95
 *         ('LAMBDA_TYPE', <uint16_t>(EdgeClassification.LAMBDA_TYPE)),
 *         ('LAMBDA_TERM', <uint16_t>(EdgeClassification.LAMBDA_TERM)),
 *         ('LET_IN_DEF', <uint16_t>(EdgeClassification.LET_IN_DEF)),             # <<<<<<<<<<<<<<
 *         ('LET_IN_TYPE', <uint16_t>(EdgeClassification.LET_IN_TYPE)),
 *         ('LET_IN_TERM', <uint16_t>(EdgeClassification.LET_IN_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::LET_IN_DEF)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_LET_IN_DEF);
    __Pyx_GIVEREF(__pyx_n_s_LET_IN_DEF);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_LET_IN_DEF);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":96
 *         ('LAMBDA_TERM', <uint16_t>(EdgeClassification.LAMBDA_TERM)),
 *         ('LET_IN_DEF', <uint16_t>(EdgeClassification.LET_IN_DEF)),
 *         ('LET_IN_TYPE', <uint16_t>(EdgeClassification.LET_IN_TYPE)),             # <<<<<<<<<<<<<<
 *         ('LET_IN_TERM', <uint16_t>(EdgeClassification.LET_IN_TERM)),
 *         ('APP_FUN', <uint16_t>(EdgeClassification.APP_FUN)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::LET_IN_TYPE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_LET_IN_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_LET_IN_TYPE);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_LET_IN_TYPE);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":97
 *         ('LET_IN_DEF', <uint16_t>(EdgeClassification.LET_IN_DEF)),
 *         ('LET_IN_TYPE', <uint16_t>(EdgeClassification.LET_IN_TYPE)),
 *         ('LET_IN_TERM', <uint16_t>(EdgeClassification.LET_IN_TERM)),             # <<<<<<<<<<<<<<
 *         ('APP_FUN', <uint16_t>(EdgeClassification.APP_FUN)),
 *         ('APP_ARG', <uint16_t>(EdgeClassification.APP_ARG)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::LET_IN_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_LET_IN_TERM);
    __Pyx_GIVEREF(__pyx_n_s_LET_IN_TERM);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_LET_IN_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":98
 *         ('LET_IN_TYPE', <uint16_t>(EdgeClassification.LET_IN_TYPE)),
 *         ('LET_IN_TERM', <uint16_t>(EdgeClassification.LET_IN_TERM)),
 *         ('APP_FUN', <uint16_t>(EdgeClassification.APP_FUN)),             # <<<<<<<<<<<<<<
 *         ('APP_ARG', <uint16_t>(EdgeClassification.APP_ARG)),
 *         ('CASE_TERM', <uint16_t>(EdgeClassification.CASE_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::APP_FUN)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_APP_FUN);
    __Pyx_GIVEREF(__pyx_n_s_APP_FUN);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_s_APP_FUN);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":99
 *         ('LET_IN_TERM', <uint16_t>(EdgeClassification.LET_IN_TERM)),
 *         ('APP_FUN', <uint16_t>(EdgeClassification.APP_FUN)),
 *         ('APP_ARG', <uint16_t>(EdgeClassification.APP_ARG)),             # <<<<<<<<<<<<<<
 *         ('CASE_TERM', <uint16_t>(EdgeClassification.CASE_TERM)),
 *         ('CASE_RETURN', <uint16_t>(EdgeClassification.CASE_RETURN)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::APP_ARG)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_APP_ARG);
    __Pyx_GIVEREF(__pyx_n_s_APP_ARG);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_APP_ARG);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":100
 *         ('APP_FUN', <uint16_t>(EdgeClassification.APP_FUN)),
 *         ('APP_ARG', <uint16_t>(EdgeClassification.APP_ARG)),
 *         ('CASE_TERM', <uint16_t>(EdgeClassification.CASE_TERM)),             # <<<<<<<<<<<<<<
 *         ('CASE_RETURN', <uint16_t>(EdgeClassification.CASE_RETURN)),
 *         ('CASE_BRANCH_POINTER', <uint16_t>(EdgeClassification.CASE_BRANCH_POINTER)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CASE_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_INCREF(__pyx_n_s_CASE_TERM);
    __Pyx_GIVEREF(__pyx_n_s_CASE_TERM);
    PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_s_CASE_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":101
 *         ('APP_ARG', <uint16_t>(EdgeClassification.APP_ARG)),
 *         ('CASE_TERM', <uint16_t>(EdgeClassification.CASE_TERM)),
 *         ('CASE_RETURN', <uint16_t>(EdgeClassification.CASE_RETURN)),             # <<<<<<<<<<<<<<
 *         ('CASE_BRANCH_POINTER', <uint16_t>(EdgeClassification.CASE_BRANCH_POINTER)),
 *         ('CASE_IND', <uint16_t>(EdgeClassification.CASE_IND)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CASE_RETURN)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_INCREF(__pyx_n_s_CASE_RETURN);
    __Pyx_GIVEREF(__pyx_n_s_CASE_RETURN);
    PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_n_s_CASE_RETURN);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":102
 *         ('CASE_TERM', <uint16_t>(EdgeClassification.CASE_TERM)),
 *         ('CASE_RETURN', <uint16_t>(EdgeClassification.CASE_RETURN)),
 *         ('CASE_BRANCH_POINTER', <uint16_t>(EdgeClassification.CASE_BRANCH_POINTER)),             # <<<<<<<<<<<<<<
 *         ('CASE_IND', <uint16_t>(EdgeClassification.CASE_IND)),
 *         ('C_B_CONSTRUCT', <uint16_t>(EdgeClassification.C_B_CONSTRUCT)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CASE_BRANCH_POINTER)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_INCREF(__pyx_n_s_CASE_BRANCH_POINTER);
    __Pyx_GIVEREF(__pyx_n_s_CASE_BRANCH_POINTER);
    PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_n_s_CASE_BRANCH_POINTER);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":103
 *         ('CASE_RETURN', <uint16_t>(EdgeClassification.CASE_RETURN)),
 *         ('CASE_BRANCH_POINTER', <uint16_t>(EdgeClassification.CASE_BRANCH_POINTER)),
 *         ('CASE_IND', <uint16_t>(EdgeClassification.CASE_IND)),             # <<<<<<<<<<<<<<
 *         ('C_B_CONSTRUCT', <uint16_t>(EdgeClassification.C_B_CONSTRUCT)),
 *         ('C_B_TERM', <uint16_t>(EdgeClassification.C_B_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CASE_IND)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_INCREF(__pyx_n_s_CASE_IND);
    __Pyx_GIVEREF(__pyx_n_s_CASE_IND);
    PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_s_CASE_IND);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":104
 *         ('CASE_BRANCH_POINTER', <uint16_t>(EdgeClassification.CASE_BRANCH_POINTER)),
 *         ('CASE_IND', <uint16_t>(EdgeClassification.CASE_IND)),
 *         ('C_B_CONSTRUCT', <uint16_t>(EdgeClassification.C_B_CONSTRUCT)),             # <<<<<<<<<<<<<<
 *         ('C_B_TERM', <uint16_t>(EdgeClassification.C_B_TERM)),
 *         ('FIX_MUTUAL', <uint16_t>(EdgeClassification.FIX_MUTUAL)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::C_B_CONSTRUCT)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_33);
    __Pyx_INCREF(__pyx_n_s_C_B_CONSTRUCT);
    __Pyx_GIVEREF(__pyx_n_s_C_B_CONSTRUCT);
    PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_s_C_B_CONSTRUCT);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":105
 *         ('CASE_IND', <uint16_t>(EdgeClassification.CASE_IND)),
 *         ('C_B_CONSTRUCT', <uint16_t>(EdgeClassification.C_B_CONSTRUCT)),
 *         ('C_B_TERM', <uint16_t>(EdgeClassification.C_B_TERM)),             # <<<<<<<<<<<<<<
 *         ('FIX_MUTUAL', <uint16_t>(EdgeClassification.FIX_MUTUAL)),
 *         ('FIX_RETURN', <uint16_t>(EdgeClassification.FIX_RETURN)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::C_B_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_34);
    __Pyx_INCREF(__pyx_n_s_C_B_TERM);
    __Pyx_GIVEREF(__pyx_n_s_C_B_TERM);
    PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_s_C_B_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":106
 *         ('C_B_CONSTRUCT', <uint16_t>(EdgeClassification.C_B_CONSTRUCT)),
 *         ('C_B_TERM', <uint16_t>(EdgeClassification.C_B_TERM)),
 *         ('FIX_MUTUAL', <uint16_t>(EdgeClassification.FIX_MUTUAL)),             # <<<<<<<<<<<<<<
 *         ('FIX_RETURN', <uint16_t>(EdgeClassification.FIX_RETURN)),
 *         ('FIX_FUN_TYPE', <uint16_t>(EdgeClassification.FIX_FUN_TYPE)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::FIX_MUTUAL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_35);
    __Pyx_INCREF(__pyx_n_s_FIX_MUTUAL);
    __Pyx_GIVEREF(__pyx_n_s_FIX_MUTUAL);
    PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_s_FIX_MUTUAL);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":107
 *         ('C_B_TERM', <uint16_t>(EdgeClassification.C_B_TERM)),
 *         ('FIX_MUTUAL', <uint16_t>(EdgeClassification.FIX_MUTUAL)),
 *         ('FIX_RETURN', <uint16_t>(EdgeClassification.FIX_RETURN)),             # <<<<<<<<<<<<<<
 *         ('FIX_FUN_TYPE', <uint16_t>(EdgeClassification.FIX_FUN_TYPE)),
 *         ('FIX_FUN_TERM', <uint16_t>(EdgeClassification.FIX_FUN_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::FIX_RETURN)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_36);
    __Pyx_INCREF(__pyx_n_s_FIX_RETURN);
    __Pyx_GIVEREF(__pyx_n_s_FIX_RETURN);
    PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_s_FIX_RETURN);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":108
 *         ('FIX_MUTUAL', <uint16_t>(EdgeClassification.FIX_MUTUAL)),
 *         ('FIX_RETURN', <uint16_t>(EdgeClassification.FIX_RETURN)),
 *         ('FIX_FUN_TYPE', <uint16_t>(EdgeClassification.FIX_FUN_TYPE)),             # <<<<<<<<<<<<<<
 *         ('FIX_FUN_TERM', <uint16_t>(EdgeClassification.FIX_FUN_TERM)),
 *         ('CO_FIX_MUTUAL', <uint16_t>(EdgeClassification.CO_FIX_MUTUAL)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::FIX_FUN_TYPE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_37);
    __Pyx_INCREF(__pyx_n_s_FIX_FUN_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_FIX_FUN_TYPE);
    PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_s_FIX_FUN_TYPE);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":109
 *         ('FIX_RETURN', <uint16_t>(EdgeClassification.FIX_RETURN)),
 *         ('FIX_FUN_TYPE', <uint16_t>(EdgeClassification.FIX_FUN_TYPE)),
 *         ('FIX_FUN_TERM', <uint16_t>(EdgeClassification.FIX_FUN_TERM)),             # <<<<<<<<<<<<<<
 *         ('CO_FIX_MUTUAL', <uint16_t>(EdgeClassification.CO_FIX_MUTUAL)),
 *         ('CO_FIX_RETURN', <uint16_t>(EdgeClassification.CO_FIX_RETURN)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::FIX_FUN_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_38);
    __Pyx_INCREF(__pyx_n_s_FIX_FUN_TERM);
    __Pyx_GIVEREF(__pyx_n_s_FIX_FUN_TERM);
    PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_s_FIX_FUN_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":110
 *         ('FIX_FUN_TYPE', <uint16_t>(EdgeClassification.FIX_FUN_TYPE)),
 *         ('FIX_FUN_TERM', <uint16_t>(EdgeClassification.FIX_FUN_TERM)),
 *         ('CO_FIX_MUTUAL', <uint16_t>(EdgeClassification.CO_FIX_MUTUAL)),             # <<<<<<<<<<<<<<
 *         ('CO_FIX_RETURN', <uint16_t>(EdgeClassification.CO_FIX_RETURN)),
 *         ('CO_FIX_FUN_TYPE', <uint16_t>(EdgeClassification.CO_FIX_FUN_TYPE)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CO_FIX_MUTUAL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_39);
    __Pyx_INCREF(__pyx_n_s_CO_FIX_MUTUAL);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX_MUTUAL);
    PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_CO_FIX_MUTUAL);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":111
 *         ('FIX_FUN_TERM', <uint16_t>(EdgeClassification.FIX_FUN_TERM)),
 *         ('CO_FIX_MUTUAL', <uint16_t>(EdgeClassification.CO_FIX_MUTUAL)),
 *         ('CO_FIX_RETURN', <uint16_t>(EdgeClassification.CO_FIX_RETURN)),             # <<<<<<<<<<<<<<
 *         ('CO_FIX_FUN_TYPE', <uint16_t>(EdgeClassification.CO_FIX_FUN_TYPE)),
 *         ('CO_FIX_FUN_TERM', <uint16_t>(EdgeClassification.CO_FIX_FUN_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CO_FIX_RETURN)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_40);
    __Pyx_INCREF(__pyx_n_s_CO_FIX_RETURN);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX_RETURN);
    PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_CO_FIX_RETURN);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":112
 *         ('CO_FIX_MUTUAL', <uint16_t>(EdgeClassification.CO_FIX_MUTUAL)),
 *         ('CO_FIX_RETURN', <uint16_t>(EdgeClassification.CO_FIX_RETURN)),
 *         ('CO_FIX_FUN_TYPE', <uint16_t>(EdgeClassification.CO_FIX_FUN_TYPE)),             # <<<<<<<<<<<<<<
 *         ('CO_FIX_FUN_TERM', <uint16_t>(EdgeClassification.CO_FIX_FUN_TERM)),
 *         ('REL_POINTER', <uint16_t>(EdgeClassification.REL_POINTER)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CO_FIX_FUN_TYPE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_41);
    __Pyx_INCREF(__pyx_n_s_CO_FIX_FUN_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX_FUN_TYPE);
    PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_s_CO_FIX_FUN_TYPE);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":113
 *         ('CO_FIX_RETURN', <uint16_t>(EdgeClassification.CO_FIX_RETURN)),
 *         ('CO_FIX_FUN_TYPE', <uint16_t>(EdgeClassification.CO_FIX_FUN_TYPE)),
 *         ('CO_FIX_FUN_TERM', <uint16_t>(EdgeClassification.CO_FIX_FUN_TERM)),             # <<<<<<<<<<<<<<
 *         ('REL_POINTER', <uint16_t>(EdgeClassification.REL_POINTER)),
 *         ('EVAR_SUBST_POINTER', <uint16_t>(EdgeClassification.EVAR_SUBST_POINTER)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CO_FIX_FUN_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_42);
    __Pyx_INCREF(__pyx_n_s_CO_FIX_FUN_TERM);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX_FUN_TERM);
    PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_s_CO_FIX_FUN_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":114
 *         ('CO_FIX_FUN_TYPE', <uint16_t>(EdgeClassification.CO_FIX_FUN_TYPE)),
 *         ('CO_FIX_FUN_TERM', <uint16_t>(EdgeClassification.CO_FIX_FUN_TERM)),
 *         ('REL_POINTER', <uint16_t>(EdgeClassification.REL_POINTER)),             # <<<<<<<<<<<<<<
 *         ('EVAR_SUBST_POINTER', <uint16_t>(EdgeClassification.EVAR_SUBST_POINTER)),
 *         ('EVAR_SUBST_TERM', <uint16_t>(EdgeClassification.EVAR_SUBST_TERM)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::REL_POINTER)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_43);
    __Pyx_INCREF(__pyx_n_s_REL_POINTER);
    __Pyx_GIVEREF(__pyx_n_s_REL_POINTER);
    PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_s_REL_POINTER);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":115
 *         ('CO_FIX_FUN_TERM', <uint16_t>(EdgeClassification.CO_FIX_FUN_TERM)),
 *         ('REL_POINTER', <uint16_t>(EdgeClassification.REL_POINTER)),
 *         ('EVAR_SUBST_POINTER', <uint16_t>(EdgeClassification.EVAR_SUBST_POINTER)),             # <<<<<<<<<<<<<<
 *         ('EVAR_SUBST_TERM', <uint16_t>(EdgeClassification.EVAR_SUBST_TERM)),
 *         ('EVAR_SUBST_TARGET', <uint16_t>(EdgeClassification.EVAR_SUBST_TARGET)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::EVAR_SUBST_POINTER)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __Pyx_INCREF(__pyx_n_s_EVAR_SUBST_POINTER);
    __Pyx_GIVEREF(__pyx_n_s_EVAR_SUBST_POINTER);
    PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_EVAR_SUBST_POINTER);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":116
 *         ('REL_POINTER', <uint16_t>(EdgeClassification.REL_POINTER)),
 *         ('EVAR_SUBST_POINTER', <uint16_t>(EdgeClassification.EVAR_SUBST_POINTER)),
 *         ('EVAR_SUBST_TERM', <uint16_t>(EdgeClassification.EVAR_SUBST_TERM)),             # <<<<<<<<<<<<<<
 *         ('EVAR_SUBST_TARGET', <uint16_t>(EdgeClassification.EVAR_SUBST_TARGET)),
 *         ('EVAR_SUBJECT', <uint16_t>(EdgeClassification.EVAR_SUBJECT)),
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::EVAR_SUBST_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_INCREF(__pyx_n_s_EVAR_SUBST_TERM);
    __Pyx_GIVEREF(__pyx_n_s_EVAR_SUBST_TERM);
    PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_EVAR_SUBST_TERM);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":117
 *         ('EVAR_SUBST_POINTER', <uint16_t>(EdgeClassification.EVAR_SUBST_POINTER)),
 *         ('EVAR_SUBST_TERM', <uint16_t>(EdgeClassification.EVAR_SUBST_TERM)),
 *         ('EVAR_SUBST_TARGET', <uint16_t>(EdgeClassification.EVAR_SUBST_TARGET)),             # <<<<<<<<<<<<<<
 *         ('EVAR_SUBJECT', <uint16_t>(EdgeClassification.EVAR_SUBJECT)),
 *     ]))
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::EVAR_SUBST_TARGET)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_EVAR_SUBST_TARGET);
    __Pyx_GIVEREF(__pyx_n_s_EVAR_SUBST_TARGET);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_EVAR_SUBST_TARGET);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":118
 *         ('EVAR_SUBST_TERM', <uint16_t>(EdgeClassification.EVAR_SUBST_TERM)),
 *         ('EVAR_SUBST_TARGET', <uint16_t>(EdgeClassification.EVAR_SUBST_TARGET)),
 *         ('EVAR_SUBJECT', <uint16_t>(EdgeClassification.EVAR_SUBJECT)),             # <<<<<<<<<<<<<<
 *     ]))
 * 
 */
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::EVAR_SUBJECT)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_EVAR_SUBJECT);
    __Pyx_GIVEREF(__pyx_n_s_EVAR_SUBJECT);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_n_s_EVAR_SUBJECT);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "CppScopedEnumType":74
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     __Pyx_globals["EdgeClassification"] = __Pyx_EnumBase('EdgeClassification', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('CONTEXT_ELEM', <uint16_t>(EdgeClassification.CONTEXT_ELEM)),
 *         ('CONTEXT_SUBJECT', <uint16_t>(EdgeClassification.CONTEXT_SUBJECT)),
 */
    __pyx_t_9 = PyList_New(44); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_10);
    PyList_SET_ITEM(__pyx_t_9, 0, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_9, 1, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_8);
    PyList_SET_ITEM(__pyx_t_9, 2, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    PyList_SET_ITEM(__pyx_t_9, 3, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_9, 4, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_9, 5, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_9, 6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_11);
    PyList_SET_ITEM(__pyx_t_9, 7, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyList_SET_ITEM(__pyx_t_9, 8, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyList_SET_ITEM(__pyx_t_9, 9, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyList_SET_ITEM(__pyx_t_9, 10, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyList_SET_ITEM(__pyx_t_9, 11, __pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_16);
    PyList_SET_ITEM(__pyx_t_9, 12, __pyx_t_16);
    __Pyx_GIVEREF(__pyx_t_17);
    PyList_SET_ITEM(__pyx_t_9, 13, __pyx_t_17);
    __Pyx_GIVEREF(__pyx_t_18);
    PyList_SET_ITEM(__pyx_t_9, 14, __pyx_t_18);
    __Pyx_GIVEREF(__pyx_t_19);
    PyList_SET_ITEM(__pyx_t_9, 15, __pyx_t_19);
    __Pyx_GIVEREF(__pyx_t_20);
    PyList_SET_ITEM(__pyx_t_9, 16, __pyx_t_20);
    __Pyx_GIVEREF(__pyx_t_21);
    PyList_SET_ITEM(__pyx_t_9, 17, __pyx_t_21);
    __Pyx_GIVEREF(__pyx_t_22);
    PyList_SET_ITEM(__pyx_t_9, 18, __pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyList_SET_ITEM(__pyx_t_9, 19, __pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_9, 20, __pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_25);
    PyList_SET_ITEM(__pyx_t_9, 21, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_26);
    PyList_SET_ITEM(__pyx_t_9, 22, __pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_27);
    PyList_SET_ITEM(__pyx_t_9, 23, __pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyList_SET_ITEM(__pyx_t_9, 24, __pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_29);
    PyList_SET_ITEM(__pyx_t_9, 25, __pyx_t_29);
    __Pyx_GIVEREF(__pyx_t_30);
    PyList_SET_ITEM(__pyx_t_9, 26, __pyx_t_30);
    __Pyx_GIVEREF(__pyx_t_31);
    PyList_SET_ITEM(__pyx_t_9, 27, __pyx_t_31);
    __Pyx_GIVEREF(__pyx_t_32);
    PyList_SET_ITEM(__pyx_t_9, 28, __pyx_t_32);
    __Pyx_GIVEREF(__pyx_t_33);
    PyList_SET_ITEM(__pyx_t_9, 29, __pyx_t_33);
    __Pyx_GIVEREF(__pyx_t_34);
    PyList_SET_ITEM(__pyx_t_9, 30, __pyx_t_34);
    __Pyx_GIVEREF(__pyx_t_35);
    PyList_SET_ITEM(__pyx_t_9, 31, __pyx_t_35);
    __Pyx_GIVEREF(__pyx_t_36);
    PyList_SET_ITEM(__pyx_t_9, 32, __pyx_t_36);
    __Pyx_GIVEREF(__pyx_t_37);
    PyList_SET_ITEM(__pyx_t_9, 33, __pyx_t_37);
    __Pyx_GIVEREF(__pyx_t_38);
    PyList_SET_ITEM(__pyx_t_9, 34, __pyx_t_38);
    __Pyx_GIVEREF(__pyx_t_39);
    PyList_SET_ITEM(__pyx_t_9, 35, __pyx_t_39);
    __Pyx_GIVEREF(__pyx_t_40);
    PyList_SET_ITEM(__pyx_t_9, 36, __pyx_t_40);
    __Pyx_GIVEREF(__pyx_t_41);
    PyList_SET_ITEM(__pyx_t_9, 37, __pyx_t_41);
    __Pyx_GIVEREF(__pyx_t_42);
    PyList_SET_ITEM(__pyx_t_9, 38, __pyx_t_42);
    __Pyx_GIVEREF(__pyx_t_43);
    PyList_SET_ITEM(__pyx_t_9, 39, __pyx_t_43);
    __Pyx_GIVEREF(__pyx_t_44);
    PyList_SET_ITEM(__pyx_t_9, 40, __pyx_t_44);
    __Pyx_GIVEREF(__pyx_t_45);
    PyList_SET_ITEM(__pyx_t_9, 41, __pyx_t_45);
    __Pyx_GIVEREF(__pyx_t_46);
    PyList_SET_ITEM(__pyx_t_9, 42, __pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_47);
    PyList_SET_ITEM(__pyx_t_9, 43, __pyx_t_47);
    __pyx_t_10 = 0;
    __pyx_t_5 = 0;
    __pyx_t_8 = 0;
    __pyx_t_7 = 0;
    __pyx_t_4 = 0;
    __pyx_t_3 = 0;
    __pyx_t_2 = 0;
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_t_13 = 0;
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_t_16 = 0;
    __pyx_t_17 = 0;
    __pyx_t_18 = 0;
    __pyx_t_19 = 0;
    __pyx_t_20 = 0;
    __pyx_t_21 = 0;
    __pyx_t_22 = 0;
    __pyx_t_23 = 0;
    __pyx_t_24 = 0;
    __pyx_t_25 = 0;
    __pyx_t_26 = 0;
    __pyx_t_27 = 0;
    __pyx_t_28 = 0;
    __pyx_t_29 = 0;
    __pyx_t_30 = 0;
    __pyx_t_31 = 0;
    __pyx_t_32 = 0;
    __pyx_t_33 = 0;
    __pyx_t_34 = 0;
    __pyx_t_35 = 0;
    __pyx_t_36 = 0;
    __pyx_t_37 = 0;
    __pyx_t_38 = 0;
    __pyx_t_39 = 0;
    __pyx_t_40 = 0;
    __pyx_t_41 = 0;
    __pyx_t_42 = 0;
    __pyx_t_43 = 0;
    __pyx_t_44 = 0;
    __pyx_t_45 = 0;
    __pyx_t_46 = 0;
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_9); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_EdgeClassification);
    __Pyx_GIVEREF(__pyx_n_s_EdgeClassification);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_EdgeClassification);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_47);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_EdgeClassification, __pyx_t_47) < 0))) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["EdgeClassification"] = __Pyx_EnumBase('EdgeClassification', __Pyx_OrderedDict([
 */
    goto __pyx_L6;
  }

  /* "CppScopedEnumType":122
 * 
 * else:
 *     __Pyx_globals["EdgeClassification"] = type('EdgeClassification', (__Pyx_EnumBase,), {})             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_ELEM), 'CONTEXT_ELEM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_SUBJECT), 'CONTEXT_SUBJECT')
 */
  /*else*/ {
    __pyx_t_47 = PyTuple_New(1); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __Pyx_EnumBase);
    __pyx_t_9 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(3); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_EdgeClassification);
    __Pyx_GIVEREF(__pyx_n_s_EdgeClassification);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_EdgeClassification);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 2, __pyx_t_9);
    __pyx_t_47 = 0;
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 122, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_EdgeClassification, __pyx_t_9) < 0))) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":123
 * else:
 *     __Pyx_globals["EdgeClassification"] = type('EdgeClassification', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_ELEM), 'CONTEXT_ELEM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_SUBJECT), 'CONTEXT_SUBJECT')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_DEF_TYPE), 'CONTEXT_DEF_TYPE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 123, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONTEXT_ELEM)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_ELEM);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_ELEM);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_CONTEXT_ELEM);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":124
 *     __Pyx_globals["EdgeClassification"] = type('EdgeClassification', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_ELEM), 'CONTEXT_ELEM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_SUBJECT), 'CONTEXT_SUBJECT')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_DEF_TYPE), 'CONTEXT_DEF_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_DEF_TERM), 'CONTEXT_DEF_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 124, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONTEXT_SUBJECT)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_SUBJECT);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_SUBJECT);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_CONTEXT_SUBJECT);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":125
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_ELEM), 'CONTEXT_ELEM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_SUBJECT), 'CONTEXT_SUBJECT')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_DEF_TYPE), 'CONTEXT_DEF_TYPE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_DEF_TERM), 'CONTEXT_DEF_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_TYPE), 'CONST_TYPE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 125, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONTEXT_DEF_TYPE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_DEF_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_DEF_TYPE);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_CONTEXT_DEF_TYPE);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":126
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_SUBJECT), 'CONTEXT_SUBJECT')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_DEF_TYPE), 'CONTEXT_DEF_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_DEF_TERM), 'CONTEXT_DEF_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_TYPE), 'CONST_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_UNDEF), 'CONST_UNDEF')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 126, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONTEXT_DEF_TERM)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_DEF_TERM);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_DEF_TERM);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_CONTEXT_DEF_TERM);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":127
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_DEF_TYPE), 'CONTEXT_DEF_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_DEF_TERM), 'CONTEXT_DEF_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_TYPE), 'CONST_TYPE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_UNDEF), 'CONST_UNDEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_DEF), 'CONST_DEF')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 127, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONST_TYPE)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_CONST_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_CONST_TYPE);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_CONST_TYPE);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":128
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONTEXT_DEF_TERM), 'CONTEXT_DEF_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_TYPE), 'CONST_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_UNDEF), 'CONST_UNDEF')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_DEF), 'CONST_DEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_OPAQUE_DEF), 'CONST_OPAQUE_DEF')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 128, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONST_UNDEF)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_CONST_UNDEF);
    __Pyx_GIVEREF(__pyx_n_s_CONST_UNDEF);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_CONST_UNDEF);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":129
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_TYPE), 'CONST_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_UNDEF), 'CONST_UNDEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_DEF), 'CONST_DEF')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_OPAQUE_DEF), 'CONST_OPAQUE_DEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_PRIMITIVE), 'CONST_PRIMITIVE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 129, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONST_DEF)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_CONST_DEF);
    __Pyx_GIVEREF(__pyx_n_s_CONST_DEF);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_CONST_DEF);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":130
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_UNDEF), 'CONST_UNDEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_DEF), 'CONST_DEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_OPAQUE_DEF), 'CONST_OPAQUE_DEF')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_PRIMITIVE), 'CONST_PRIMITIVE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_TYPE), 'IND_TYPE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 130, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONST_OPAQUE_DEF)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_CONST_OPAQUE_DEF);
    __Pyx_GIVEREF(__pyx_n_s_CONST_OPAQUE_DEF);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_CONST_OPAQUE_DEF);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":131
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_DEF), 'CONST_DEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_OPAQUE_DEF), 'CONST_OPAQUE_DEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_PRIMITIVE), 'CONST_PRIMITIVE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_TYPE), 'IND_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_CONSTRUCT), 'IND_CONSTRUCT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 131, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONST_PRIMITIVE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_CONST_PRIMITIVE);
    __Pyx_GIVEREF(__pyx_n_s_CONST_PRIMITIVE);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_CONST_PRIMITIVE);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":132
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_OPAQUE_DEF), 'CONST_OPAQUE_DEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_PRIMITIVE), 'CONST_PRIMITIVE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_TYPE), 'IND_TYPE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_CONSTRUCT), 'IND_CONSTRUCT')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_PROJECTION), 'IND_PROJECTION')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 132, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::IND_TYPE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_IND_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_IND_TYPE);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_IND_TYPE);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":133
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONST_PRIMITIVE), 'CONST_PRIMITIVE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_TYPE), 'IND_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_CONSTRUCT), 'IND_CONSTRUCT')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_PROJECTION), 'IND_PROJECTION')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROJ_TERM), 'PROJ_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 133, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::IND_CONSTRUCT)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_IND_CONSTRUCT);
    __Pyx_GIVEREF(__pyx_n_s_IND_CONSTRUCT);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_IND_CONSTRUCT);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":134
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_TYPE), 'IND_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_CONSTRUCT), 'IND_CONSTRUCT')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_PROJECTION), 'IND_PROJECTION')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROJ_TERM), 'PROJ_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONSTRUCT_TERM), 'CONSTRUCT_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 134, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::IND_PROJECTION)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_IND_PROJECTION);
    __Pyx_GIVEREF(__pyx_n_s_IND_PROJECTION);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_IND_PROJECTION);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":135
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_CONSTRUCT), 'IND_CONSTRUCT')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_PROJECTION), 'IND_PROJECTION')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROJ_TERM), 'PROJ_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONSTRUCT_TERM), 'CONSTRUCT_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CAST_TERM), 'CAST_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 135, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::PROJ_TERM)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_PROJ_TERM);
    __Pyx_GIVEREF(__pyx_n_s_PROJ_TERM);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_PROJ_TERM);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":136
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.IND_PROJECTION), 'IND_PROJECTION')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROJ_TERM), 'PROJ_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONSTRUCT_TERM), 'CONSTRUCT_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CAST_TERM), 'CAST_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CAST_TYPE), 'CAST_TYPE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 136, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CONSTRUCT_TERM)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_CONSTRUCT_TERM);
    __Pyx_GIVEREF(__pyx_n_s_CONSTRUCT_TERM);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_CONSTRUCT_TERM);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":137
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROJ_TERM), 'PROJ_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONSTRUCT_TERM), 'CONSTRUCT_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CAST_TERM), 'CAST_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CAST_TYPE), 'CAST_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROD_TYPE), 'PROD_TYPE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 137, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CAST_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_CAST_TERM);
    __Pyx_GIVEREF(__pyx_n_s_CAST_TERM);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_CAST_TERM);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":138
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CONSTRUCT_TERM), 'CONSTRUCT_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CAST_TERM), 'CAST_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CAST_TYPE), 'CAST_TYPE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROD_TYPE), 'PROD_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROD_TERM), 'PROD_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 138, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CAST_TYPE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_CAST_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_CAST_TYPE);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_CAST_TYPE);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":139
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CAST_TERM), 'CAST_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CAST_TYPE), 'CAST_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROD_TYPE), 'PROD_TYPE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROD_TERM), 'PROD_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LAMBDA_TYPE), 'LAMBDA_TYPE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 139, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::PROD_TYPE)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_PROD_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_PROD_TYPE);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_PROD_TYPE);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":140
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CAST_TYPE), 'CAST_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROD_TYPE), 'PROD_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROD_TERM), 'PROD_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LAMBDA_TYPE), 'LAMBDA_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LAMBDA_TERM), 'LAMBDA_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 140, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::PROD_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_PROD_TERM);
    __Pyx_GIVEREF(__pyx_n_s_PROD_TERM);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_PROD_TERM);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":141
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROD_TYPE), 'PROD_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROD_TERM), 'PROD_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LAMBDA_TYPE), 'LAMBDA_TYPE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LAMBDA_TERM), 'LAMBDA_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_DEF), 'LET_IN_DEF')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 141, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::LAMBDA_TYPE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_LAMBDA_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_LAMBDA_TYPE);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_LAMBDA_TYPE);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":142
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.PROD_TERM), 'PROD_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LAMBDA_TYPE), 'LAMBDA_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LAMBDA_TERM), 'LAMBDA_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_DEF), 'LET_IN_DEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_TYPE), 'LET_IN_TYPE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 142, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::LAMBDA_TERM)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_LAMBDA_TERM);
    __Pyx_GIVEREF(__pyx_n_s_LAMBDA_TERM);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_LAMBDA_TERM);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":143
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LAMBDA_TYPE), 'LAMBDA_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LAMBDA_TERM), 'LAMBDA_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_DEF), 'LET_IN_DEF')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_TYPE), 'LET_IN_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_TERM), 'LET_IN_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 143, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::LET_IN_DEF)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_LET_IN_DEF);
    __Pyx_GIVEREF(__pyx_n_s_LET_IN_DEF);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_LET_IN_DEF);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":144
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LAMBDA_TERM), 'LAMBDA_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_DEF), 'LET_IN_DEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_TYPE), 'LET_IN_TYPE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_TERM), 'LET_IN_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.APP_FUN), 'APP_FUN')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 144, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::LET_IN_TYPE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_LET_IN_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_LET_IN_TYPE);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_LET_IN_TYPE);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":145
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_DEF), 'LET_IN_DEF')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_TYPE), 'LET_IN_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_TERM), 'LET_IN_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.APP_FUN), 'APP_FUN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.APP_ARG), 'APP_ARG')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 145, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::LET_IN_TERM)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_LET_IN_TERM);
    __Pyx_GIVEREF(__pyx_n_s_LET_IN_TERM);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_LET_IN_TERM);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":146
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_TYPE), 'LET_IN_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_TERM), 'LET_IN_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.APP_FUN), 'APP_FUN')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.APP_ARG), 'APP_ARG')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_TERM), 'CASE_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 146, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::APP_FUN)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_APP_FUN);
    __Pyx_GIVEREF(__pyx_n_s_APP_FUN);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_APP_FUN);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":147
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.LET_IN_TERM), 'LET_IN_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.APP_FUN), 'APP_FUN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.APP_ARG), 'APP_ARG')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_TERM), 'CASE_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_RETURN), 'CASE_RETURN')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 147, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::APP_ARG)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_APP_ARG);
    __Pyx_GIVEREF(__pyx_n_s_APP_ARG);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_APP_ARG);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":148
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.APP_FUN), 'APP_FUN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.APP_ARG), 'APP_ARG')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_TERM), 'CASE_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_RETURN), 'CASE_RETURN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_BRANCH_POINTER), 'CASE_BRANCH_POINTER')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 148, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CASE_TERM)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_CASE_TERM);
    __Pyx_GIVEREF(__pyx_n_s_CASE_TERM);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_CASE_TERM);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":149
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.APP_ARG), 'APP_ARG')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_TERM), 'CASE_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_RETURN), 'CASE_RETURN')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_BRANCH_POINTER), 'CASE_BRANCH_POINTER')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_IND), 'CASE_IND')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 149, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CASE_RETURN)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_CASE_RETURN);
    __Pyx_GIVEREF(__pyx_n_s_CASE_RETURN);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_CASE_RETURN);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":150
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_TERM), 'CASE_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_RETURN), 'CASE_RETURN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_BRANCH_POINTER), 'CASE_BRANCH_POINTER')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_IND), 'CASE_IND')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.C_B_CONSTRUCT), 'C_B_CONSTRUCT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 150, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CASE_BRANCH_POINTER)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_CASE_BRANCH_POINTER);
    __Pyx_GIVEREF(__pyx_n_s_CASE_BRANCH_POINTER);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_CASE_BRANCH_POINTER);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":151
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_RETURN), 'CASE_RETURN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_BRANCH_POINTER), 'CASE_BRANCH_POINTER')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_IND), 'CASE_IND')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.C_B_CONSTRUCT), 'C_B_CONSTRUCT')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.C_B_TERM), 'C_B_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 151, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CASE_IND)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_CASE_IND);
    __Pyx_GIVEREF(__pyx_n_s_CASE_IND);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_CASE_IND);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":152
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_BRANCH_POINTER), 'CASE_BRANCH_POINTER')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_IND), 'CASE_IND')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.C_B_CONSTRUCT), 'C_B_CONSTRUCT')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.C_B_TERM), 'C_B_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_MUTUAL), 'FIX_MUTUAL')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 152, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::C_B_CONSTRUCT)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_C_B_CONSTRUCT);
    __Pyx_GIVEREF(__pyx_n_s_C_B_CONSTRUCT);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_C_B_CONSTRUCT);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":153
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CASE_IND), 'CASE_IND')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.C_B_CONSTRUCT), 'C_B_CONSTRUCT')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.C_B_TERM), 'C_B_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_MUTUAL), 'FIX_MUTUAL')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_RETURN), 'FIX_RETURN')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 153, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::C_B_TERM)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_C_B_TERM);
    __Pyx_GIVEREF(__pyx_n_s_C_B_TERM);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_C_B_TERM);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":154
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.C_B_CONSTRUCT), 'C_B_CONSTRUCT')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.C_B_TERM), 'C_B_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_MUTUAL), 'FIX_MUTUAL')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_RETURN), 'FIX_RETURN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_FUN_TYPE), 'FIX_FUN_TYPE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 154, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::FIX_MUTUAL)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_FIX_MUTUAL);
    __Pyx_GIVEREF(__pyx_n_s_FIX_MUTUAL);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_FIX_MUTUAL);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":155
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.C_B_TERM), 'C_B_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_MUTUAL), 'FIX_MUTUAL')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_RETURN), 'FIX_RETURN')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_FUN_TYPE), 'FIX_FUN_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_FUN_TERM), 'FIX_FUN_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 155, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::FIX_RETURN)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_FIX_RETURN);
    __Pyx_GIVEREF(__pyx_n_s_FIX_RETURN);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_FIX_RETURN);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":156
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_MUTUAL), 'FIX_MUTUAL')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_RETURN), 'FIX_RETURN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_FUN_TYPE), 'FIX_FUN_TYPE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_FUN_TERM), 'FIX_FUN_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_MUTUAL), 'CO_FIX_MUTUAL')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 156, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::FIX_FUN_TYPE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_FIX_FUN_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_FIX_FUN_TYPE);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_FIX_FUN_TYPE);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":157
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_RETURN), 'FIX_RETURN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_FUN_TYPE), 'FIX_FUN_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_FUN_TERM), 'FIX_FUN_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_MUTUAL), 'CO_FIX_MUTUAL')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_RETURN), 'CO_FIX_RETURN')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 157, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::FIX_FUN_TERM)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_FIX_FUN_TERM);
    __Pyx_GIVEREF(__pyx_n_s_FIX_FUN_TERM);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_FIX_FUN_TERM);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":158
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_FUN_TYPE), 'FIX_FUN_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_FUN_TERM), 'FIX_FUN_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_MUTUAL), 'CO_FIX_MUTUAL')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_RETURN), 'CO_FIX_RETURN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_FUN_TYPE), 'CO_FIX_FUN_TYPE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 158, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CO_FIX_MUTUAL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_CO_FIX_MUTUAL);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX_MUTUAL);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_CO_FIX_MUTUAL);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":159
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.FIX_FUN_TERM), 'FIX_FUN_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_MUTUAL), 'CO_FIX_MUTUAL')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_RETURN), 'CO_FIX_RETURN')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_FUN_TYPE), 'CO_FIX_FUN_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_FUN_TERM), 'CO_FIX_FUN_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 159, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CO_FIX_RETURN)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_CO_FIX_RETURN);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX_RETURN);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_CO_FIX_RETURN);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":160
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_MUTUAL), 'CO_FIX_MUTUAL')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_RETURN), 'CO_FIX_RETURN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_FUN_TYPE), 'CO_FIX_FUN_TYPE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_FUN_TERM), 'CO_FIX_FUN_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.REL_POINTER), 'REL_POINTER')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 160, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CO_FIX_FUN_TYPE)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_CO_FIX_FUN_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX_FUN_TYPE);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_CO_FIX_FUN_TYPE);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":161
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_RETURN), 'CO_FIX_RETURN')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_FUN_TYPE), 'CO_FIX_FUN_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_FUN_TERM), 'CO_FIX_FUN_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.REL_POINTER), 'REL_POINTER')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_POINTER), 'EVAR_SUBST_POINTER')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 161, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::CO_FIX_FUN_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_CO_FIX_FUN_TERM);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX_FUN_TERM);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_CO_FIX_FUN_TERM);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":162
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_FUN_TYPE), 'CO_FIX_FUN_TYPE')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_FUN_TERM), 'CO_FIX_FUN_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.REL_POINTER), 'REL_POINTER')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_POINTER), 'EVAR_SUBST_POINTER')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_TERM), 'EVAR_SUBST_TERM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 162, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::REL_POINTER)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_REL_POINTER);
    __Pyx_GIVEREF(__pyx_n_s_REL_POINTER);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_REL_POINTER);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":163
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.CO_FIX_FUN_TERM), 'CO_FIX_FUN_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.REL_POINTER), 'REL_POINTER')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_POINTER), 'EVAR_SUBST_POINTER')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_TERM), 'EVAR_SUBST_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_TARGET), 'EVAR_SUBST_TARGET')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 163, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::EVAR_SUBST_POINTER)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_EVAR_SUBST_POINTER);
    __Pyx_GIVEREF(__pyx_n_s_EVAR_SUBST_POINTER);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_EVAR_SUBST_POINTER);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":164
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.REL_POINTER), 'REL_POINTER')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_POINTER), 'EVAR_SUBST_POINTER')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_TERM), 'EVAR_SUBST_TERM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_TARGET), 'EVAR_SUBST_TARGET')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBJECT), 'EVAR_SUBJECT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 164, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::EVAR_SUBST_TERM)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_EVAR_SUBST_TERM);
    __Pyx_GIVEREF(__pyx_n_s_EVAR_SUBST_TERM);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_EVAR_SUBST_TERM);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":165
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_POINTER), 'EVAR_SUBST_POINTER')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_TERM), 'EVAR_SUBST_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_TARGET), 'EVAR_SUBST_TARGET')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBJECT), 'EVAR_SUBJECT')
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 165, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::EVAR_SUBST_TARGET)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_EVAR_SUBST_TARGET);
    __Pyx_GIVEREF(__pyx_n_s_EVAR_SUBST_TARGET);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_EVAR_SUBST_TARGET);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_47, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":166
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_TERM), 'EVAR_SUBST_TERM')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBST_TARGET), 'EVAR_SUBST_TARGET')
 *     __Pyx_globals["EdgeClassification"](<uint16_t>(EdgeClassification.EVAR_SUBJECT), 'EVAR_SUBJECT')             # <<<<<<<<<<<<<<
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 166, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_EdgeClassification); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)capnp::schemas::EdgeClassification_e40ecefe7a595f8e::EVAR_SUBJECT)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_EVAR_SUBJECT);
    __Pyx_GIVEREF(__pyx_n_s_EVAR_SUBJECT);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_EVAR_SUBJECT);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_9, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  }
  __pyx_L6:;

  /* "CppScopedEnumType":70
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __pyx_t_47 = __Pyx_Globals(); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  if (!(likely(PyDict_CheckExact(__pyx_t_47))||((__pyx_t_47) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_47))) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_47));
  __Pyx_GIVEREF(__pyx_t_47);
  __pyx_t_47 = 0;

  /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["Exception_Which"] = __Pyx_EnumBase('Exception_Which', __Pyx_OrderedDict([
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "CppScopedEnumType":75
 * 
 *     __Pyx_globals["Exception_Which"] = __Pyx_EnumBase('Exception_Which', __Pyx_OrderedDict([
 *         ('NO_SUCH_TACTIC', <uint16_t>(Exception_Which.NO_SUCH_TACTIC)),             # <<<<<<<<<<<<<<
 *         ('MISMATCHED_ARGUMENTS', <uint16_t>(Exception_Which.MISMATCHED_ARGUMENTS)),
 *         ('PARSE_ERROR', <uint16_t>(Exception_Which.PARSE_ERROR)),
 */
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)Exception::Which::NO_SUCH_TACTIC)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_NO_SUCH_TACTIC);
    __Pyx_GIVEREF(__pyx_n_s_NO_SUCH_TACTIC);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_NO_SUCH_TACTIC);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_47);
    __pyx_t_47 = 0;

    /* "CppScopedEnumType":76
 *     __Pyx_globals["Exception_Which"] = __Pyx_EnumBase('Exception_Which', __Pyx_OrderedDict([
 *         ('NO_SUCH_TACTIC', <uint16_t>(Exception_Which.NO_SUCH_TACTIC)),
 *         ('MISMATCHED_ARGUMENTS', <uint16_t>(Exception_Which.MISMATCHED_ARGUMENTS)),             # <<<<<<<<<<<<<<
 *         ('PARSE_ERROR', <uint16_t>(Exception_Which.PARSE_ERROR)),
 *         ('ILLEGAL_ARGUMENT', <uint16_t>(Exception_Which.ILLEGAL_ARGUMENT)),
 */
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)Exception::Which::MISMATCHED_ARGUMENTS)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_MISMATCHED_ARGUMENTS);
    __Pyx_GIVEREF(__pyx_n_s_MISMATCHED_ARGUMENTS);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_MISMATCHED_ARGUMENTS);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_47);
    __pyx_t_47 = 0;

    /* "CppScopedEnumType":77
 *         ('NO_SUCH_TACTIC', <uint16_t>(Exception_Which.NO_SUCH_TACTIC)),
 *         ('MISMATCHED_ARGUMENTS', <uint16_t>(Exception_Which.MISMATCHED_ARGUMENTS)),
 *         ('PARSE_ERROR', <uint16_t>(Exception_Which.PARSE_ERROR)),             # <<<<<<<<<<<<<<
 *         ('ILLEGAL_ARGUMENT', <uint16_t>(Exception_Which.ILLEGAL_ARGUMENT)),
 *     ]))
 */
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)Exception::Which::PARSE_ERROR)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_INCREF(__pyx_n_s_PARSE_ERROR);
    __Pyx_GIVEREF(__pyx_n_s_PARSE_ERROR);
    PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_PARSE_ERROR);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_47);
    __pyx_t_47 = 0;

    /* "CppScopedEnumType":78
 *         ('MISMATCHED_ARGUMENTS', <uint16_t>(Exception_Which.MISMATCHED_ARGUMENTS)),
 *         ('PARSE_ERROR', <uint16_t>(Exception_Which.PARSE_ERROR)),
 *         ('ILLEGAL_ARGUMENT', <uint16_t>(Exception_Which.ILLEGAL_ARGUMENT)),             # <<<<<<<<<<<<<<
 *     ]))
 * 
 */
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)Exception::Which::ILLEGAL_ARGUMENT)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __Pyx_INCREF(__pyx_n_s_ILLEGAL_ARGUMENT);
    __Pyx_GIVEREF(__pyx_n_s_ILLEGAL_ARGUMENT);
    PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_ILLEGAL_ARGUMENT);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_47);
    __pyx_t_47 = 0;

    /* "CppScopedEnumType":74
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     __Pyx_globals["Exception_Which"] = __Pyx_EnumBase('Exception_Which', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('NO_SUCH_TACTIC', <uint16_t>(Exception_Which.NO_SUCH_TACTIC)),
 *         ('MISMATCHED_ARGUMENTS', <uint16_t>(Exception_Which.MISMATCHED_ARGUMENTS)),
 */
    __pyx_t_47 = PyList_New(4); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_9);
    PyList_SET_ITEM(__pyx_t_47, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_46);
    PyList_SET_ITEM(__pyx_t_47, 1, __pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_45);
    PyList_SET_ITEM(__pyx_t_47, 2, __pyx_t_45);
    __Pyx_GIVEREF(__pyx_t_44);
    PyList_SET_ITEM(__pyx_t_47, 3, __pyx_t_44);
    __pyx_t_9 = 0;
    __pyx_t_46 = 0;
    __pyx_t_45 = 0;
    __pyx_t_44 = 0;
    __pyx_t_44 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_47); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_Exception_Which);
    __Pyx_GIVEREF(__pyx_n_s_Exception_Which);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_n_s_Exception_Which);
    __Pyx_GIVEREF(__pyx_t_44);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_t_44);
    __pyx_t_44 = 0;
    __pyx_t_44 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_47, NULL); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_Exception_Which, __pyx_t_44) < 0))) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;

    /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["Exception_Which"] = __Pyx_EnumBase('Exception_Which', __Pyx_OrderedDict([
 */
    goto __pyx_L7;
  }

  /* "CppScopedEnumType":82
 * 
 * else:
 *     __Pyx_globals["Exception_Which"] = type('Exception_Which', (__Pyx_EnumBase,), {})             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.NO_SUCH_TACTIC), 'NO_SUCH_TACTIC')
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.MISMATCHED_ARGUMENTS), 'MISMATCHED_ARGUMENTS')
 */
  /*else*/ {
    __pyx_t_44 = PyTuple_New(1); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_44, 0, __Pyx_EnumBase);
    __pyx_t_47 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_45 = PyTuple_New(3); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_INCREF(__pyx_n_s_Exception_Which);
    __Pyx_GIVEREF(__pyx_n_s_Exception_Which);
    PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_Exception_Which);
    __Pyx_GIVEREF(__pyx_t_44);
    PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_44);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_45, 2, __pyx_t_47);
    __pyx_t_44 = 0;
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_45, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 82, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_Exception_Which, __pyx_t_47) < 0))) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":83
 * else:
 *     __Pyx_globals["Exception_Which"] = type('Exception_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.NO_SUCH_TACTIC), 'NO_SUCH_TACTIC')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.MISMATCHED_ARGUMENTS), 'MISMATCHED_ARGUMENTS')
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.PARSE_ERROR), 'PARSE_ERROR')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 83, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Exception_Which); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_45 = __Pyx_PyInt_From_uint16_t(((uint16_t)Exception::Which::NO_SUCH_TACTIC)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __Pyx_GIVEREF(__pyx_t_45);
    PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_t_45);
    __Pyx_INCREF(__pyx_n_s_NO_SUCH_TACTIC);
    __Pyx_GIVEREF(__pyx_n_s_NO_SUCH_TACTIC);
    PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_n_s_NO_SUCH_TACTIC);
    __pyx_t_45 = 0;
    __pyx_t_45 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_44, NULL); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;

    /* "CppScopedEnumType":84
 *     __Pyx_globals["Exception_Which"] = type('Exception_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.NO_SUCH_TACTIC), 'NO_SUCH_TACTIC')
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.MISMATCHED_ARGUMENTS), 'MISMATCHED_ARGUMENTS')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.PARSE_ERROR), 'PARSE_ERROR')
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.ILLEGAL_ARGUMENT), 'ILLEGAL_ARGUMENT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 84, __pyx_L1_error)
    }
    __pyx_t_45 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Exception_Which); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_44 = __Pyx_PyInt_From_uint16_t(((uint16_t)Exception::Which::MISMATCHED_ARGUMENTS)); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_44);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_t_44);
    __Pyx_INCREF(__pyx_n_s_MISMATCHED_ARGUMENTS);
    __Pyx_GIVEREF(__pyx_n_s_MISMATCHED_ARGUMENTS);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_n_s_MISMATCHED_ARGUMENTS);
    __pyx_t_44 = 0;
    __pyx_t_44 = __Pyx_PyObject_Call(__pyx_t_45, __pyx_t_47, NULL); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;

    /* "CppScopedEnumType":85
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.NO_SUCH_TACTIC), 'NO_SUCH_TACTIC')
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.MISMATCHED_ARGUMENTS), 'MISMATCHED_ARGUMENTS')
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.PARSE_ERROR), 'PARSE_ERROR')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.ILLEGAL_ARGUMENT), 'ILLEGAL_ARGUMENT')
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 85, __pyx_L1_error)
    }
    __pyx_t_44 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Exception_Which); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __pyx_t_47 = __Pyx_PyInt_From_uint16_t(((uint16_t)Exception::Which::PARSE_ERROR)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_GIVEREF(__pyx_t_47);
    PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_PARSE_ERROR);
    __Pyx_GIVEREF(__pyx_n_s_PARSE_ERROR);
    PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_n_s_PARSE_ERROR);
    __pyx_t_47 = 0;
    __pyx_t_47 = __Pyx_PyObject_Call(__pyx_t_44, __pyx_t_45, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

    /* "CppScopedEnumType":86
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.MISMATCHED_ARGUMENTS), 'MISMATCHED_ARGUMENTS')
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.PARSE_ERROR), 'PARSE_ERROR')
 *     __Pyx_globals["Exception_Which"](<uint16_t>(Exception_Which.ILLEGAL_ARGUMENT), 'ILLEGAL_ARGUMENT')             # <<<<<<<<<<<<<<
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 86, __pyx_L1_error)
    }
    __pyx_t_47 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Exception_Which); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __pyx_t_45 = __Pyx_PyInt_From_uint16_t(((uint16_t)Exception::Which::ILLEGAL_ARGUMENT)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __Pyx_GIVEREF(__pyx_t_45);
    PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_t_45);
    __Pyx_INCREF(__pyx_n_s_ILLEGAL_ARGUMENT);
    __Pyx_GIVEREF(__pyx_n_s_ILLEGAL_ARGUMENT);
    PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_n_s_ILLEGAL_ARGUMENT);
    __pyx_t_45 = 0;
    __pyx_t_45 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_44, NULL); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
    __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  }
  __pyx_L7:;

  /* "CppScopedEnumType":70
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __pyx_t_45 = __Pyx_Globals(); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  if (!(likely(PyDict_CheckExact(__pyx_t_45))||((__pyx_t_45) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_45))) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_45));
  __Pyx_GIVEREF(__pyx_t_45);
  __pyx_t_45 = 0;

  /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["ExecutionResult_Which"] = __Pyx_EnumBase('ExecutionResult_Which', __Pyx_OrderedDict([
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "CppScopedEnumType":75
 * 
 *     __Pyx_globals["ExecutionResult_Which"] = __Pyx_EnumBase('ExecutionResult_Which', __Pyx_OrderedDict([
 *         ('FAILURE', <uint16_t>(ExecutionResult_Which.FAILURE)),             # <<<<<<<<<<<<<<
 *         ('COMPLETE', <uint16_t>(ExecutionResult_Which.COMPLETE)),
 *         ('NEW_STATE', <uint16_t>(ExecutionResult_Which.NEW_STATE)),
 */
    __pyx_t_45 = __Pyx_PyInt_From_uint16_t(((uint16_t)ExecutionResult::Which::FAILURE)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __Pyx_INCREF(__pyx_n_s_FAILURE);
    __Pyx_GIVEREF(__pyx_n_s_FAILURE);
    PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_FAILURE);
    __Pyx_GIVEREF(__pyx_t_45);
    PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_45);
    __pyx_t_45 = 0;

    /* "CppScopedEnumType":76
 *     __Pyx_globals["ExecutionResult_Which"] = __Pyx_EnumBase('ExecutionResult_Which', __Pyx_OrderedDict([
 *         ('FAILURE', <uint16_t>(ExecutionResult_Which.FAILURE)),
 *         ('COMPLETE', <uint16_t>(ExecutionResult_Which.COMPLETE)),             # <<<<<<<<<<<<<<
 *         ('NEW_STATE', <uint16_t>(ExecutionResult_Which.NEW_STATE)),
 *         ('PROTOCOL_ERROR', <uint16_t>(ExecutionResult_Which.PROTOCOL_ERROR)),
 */
    __pyx_t_45 = __Pyx_PyInt_From_uint16_t(((uint16_t)ExecutionResult::Which::COMPLETE)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_COMPLETE);
    __Pyx_GIVEREF(__pyx_n_s_COMPLETE);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_n_s_COMPLETE);
    __Pyx_GIVEREF(__pyx_t_45);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_t_45);
    __pyx_t_45 = 0;

    /* "CppScopedEnumType":77
 *         ('FAILURE', <uint16_t>(ExecutionResult_Which.FAILURE)),
 *         ('COMPLETE', <uint16_t>(ExecutionResult_Which.COMPLETE)),
 *         ('NEW_STATE', <uint16_t>(ExecutionResult_Which.NEW_STATE)),             # <<<<<<<<<<<<<<
 *         ('PROTOCOL_ERROR', <uint16_t>(ExecutionResult_Which.PROTOCOL_ERROR)),
 *     ]))
 */
    __pyx_t_45 = __Pyx_PyInt_From_uint16_t(((uint16_t)ExecutionResult::Which::NEW_STATE)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_NEW_STATE);
    __Pyx_GIVEREF(__pyx_n_s_NEW_STATE);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_NEW_STATE);
    __Pyx_GIVEREF(__pyx_t_45);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_45);
    __pyx_t_45 = 0;

    /* "CppScopedEnumType":78
 *         ('COMPLETE', <uint16_t>(ExecutionResult_Which.COMPLETE)),
 *         ('NEW_STATE', <uint16_t>(ExecutionResult_Which.NEW_STATE)),
 *         ('PROTOCOL_ERROR', <uint16_t>(ExecutionResult_Which.PROTOCOL_ERROR)),             # <<<<<<<<<<<<<<
 *     ]))
 * 
 */
    __pyx_t_45 = __Pyx_PyInt_From_uint16_t(((uint16_t)ExecutionResult::Which::PROTOCOL_ERROR)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_PROTOCOL_ERROR);
    __Pyx_GIVEREF(__pyx_n_s_PROTOCOL_ERROR);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_PROTOCOL_ERROR);
    __Pyx_GIVEREF(__pyx_t_45);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_45);
    __pyx_t_45 = 0;

    /* "CppScopedEnumType":74
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     __Pyx_globals["ExecutionResult_Which"] = __Pyx_EnumBase('ExecutionResult_Which', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('FAILURE', <uint16_t>(ExecutionResult_Which.FAILURE)),
 *         ('COMPLETE', <uint16_t>(ExecutionResult_Which.COMPLETE)),
 */
    __pyx_t_45 = PyList_New(4); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_GIVEREF(__pyx_t_44);
    PyList_SET_ITEM(__pyx_t_45, 0, __pyx_t_44);
    __Pyx_GIVEREF(__pyx_t_47);
    PyList_SET_ITEM(__pyx_t_45, 1, __pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_46);
    PyList_SET_ITEM(__pyx_t_45, 2, __pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyList_SET_ITEM(__pyx_t_45, 3, __pyx_t_9);
    __pyx_t_44 = 0;
    __pyx_t_47 = 0;
    __pyx_t_46 = 0;
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_45); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
    __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_INCREF(__pyx_n_s_ExecutionResult_Which);
    __Pyx_GIVEREF(__pyx_n_s_ExecutionResult_Which);
    PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_ExecutionResult_Which);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_45, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_ExecutionResult_Which, __pyx_t_9) < 0))) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["ExecutionResult_Which"] = __Pyx_EnumBase('ExecutionResult_Which', __Pyx_OrderedDict([
 */
    goto __pyx_L8;
  }

  /* "CppScopedEnumType":82
 * 
 * else:
 *     __Pyx_globals["ExecutionResult_Which"] = type('ExecutionResult_Which', (__Pyx_EnumBase,), {})             # <<<<<<<<<<<<<<
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.FAILURE), 'FAILURE')
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.COMPLETE), 'COMPLETE')
 */
  /*else*/ {
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __Pyx_EnumBase);
    __pyx_t_45 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_46 = PyTuple_New(3); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_ExecutionResult_Which);
    __Pyx_GIVEREF(__pyx_n_s_ExecutionResult_Which);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_ExecutionResult_Which);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_45);
    PyTuple_SET_ITEM(__pyx_t_46, 2, __pyx_t_45);
    __pyx_t_9 = 0;
    __pyx_t_45 = 0;
    __pyx_t_45 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_46, NULL); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 82, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_ExecutionResult_Which, __pyx_t_45) < 0))) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;

    /* "CppScopedEnumType":83
 * else:
 *     __Pyx_globals["ExecutionResult_Which"] = type('ExecutionResult_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.FAILURE), 'FAILURE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.COMPLETE), 'COMPLETE')
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.NEW_STATE), 'NEW_STATE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 83, __pyx_L1_error)
    }
    __pyx_t_45 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_ExecutionResult_Which); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)ExecutionResult::Which::FAILURE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_FAILURE);
    __Pyx_GIVEREF(__pyx_n_s_FAILURE);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_FAILURE);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_45, __pyx_t_9, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":84
 *     __Pyx_globals["ExecutionResult_Which"] = type('ExecutionResult_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.FAILURE), 'FAILURE')
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.COMPLETE), 'COMPLETE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.NEW_STATE), 'NEW_STATE')
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.PROTOCOL_ERROR), 'PROTOCOL_ERROR')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 84, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_ExecutionResult_Which); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_9 = __Pyx_PyInt_From_uint16_t(((uint16_t)ExecutionResult::Which::COMPLETE)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_COMPLETE);
    __Pyx_GIVEREF(__pyx_n_s_COMPLETE);
    PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_n_s_COMPLETE);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_45, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "CppScopedEnumType":85
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.FAILURE), 'FAILURE')
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.COMPLETE), 'COMPLETE')
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.NEW_STATE), 'NEW_STATE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.PROTOCOL_ERROR), 'PROTOCOL_ERROR')
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 85, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_ExecutionResult_Which); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_45 = __Pyx_PyInt_From_uint16_t(((uint16_t)ExecutionResult::Which::NEW_STATE)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_45);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_45);
    __Pyx_INCREF(__pyx_n_s_NEW_STATE);
    __Pyx_GIVEREF(__pyx_n_s_NEW_STATE);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_NEW_STATE);
    __pyx_t_45 = 0;
    __pyx_t_45 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_46, NULL); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;

    /* "CppScopedEnumType":86
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.COMPLETE), 'COMPLETE')
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.NEW_STATE), 'NEW_STATE')
 *     __Pyx_globals["ExecutionResult_Which"](<uint16_t>(ExecutionResult_Which.PROTOCOL_ERROR), 'PROTOCOL_ERROR')             # <<<<<<<<<<<<<<
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 86, __pyx_L1_error)
    }
    __pyx_t_45 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_ExecutionResult_Which); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)ExecutionResult::Which::PROTOCOL_ERROR)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_PROTOCOL_ERROR);
    __Pyx_GIVEREF(__pyx_n_s_PROTOCOL_ERROR);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_PROTOCOL_ERROR);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_45, __pyx_t_9, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  }
  __pyx_L8:;

  /* "CppScopedEnumType":70
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __pyx_t_46 = __Pyx_Globals(); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  if (!(likely(PyDict_CheckExact(__pyx_t_46))||((__pyx_t_46) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_46))) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_46));
  __Pyx_GIVEREF(__pyx_t_46);
  __pyx_t_46 = 0;

  /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["Graph_Node_Label_Which"] = __Pyx_EnumBase('Graph_Node_Label_Which', __Pyx_OrderedDict([
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "CppScopedEnumType":75
 * 
 *     __Pyx_globals["Graph_Node_Label_Which"] = __Pyx_EnumBase('Graph_Node_Label_Which', __Pyx_OrderedDict([
 *         ('PROOF_STATE', <uint16_t>(Graph_Node_Label_Which.PROOF_STATE)),             # <<<<<<<<<<<<<<
 *         ('CONTEXT_DEF', <uint16_t>(Graph_Node_Label_Which.CONTEXT_DEF)),
 *         ('CONTEXT_ASSUM', <uint16_t>(Graph_Node_Label_Which.CONTEXT_ASSUM)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::PROOF_STATE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_PROOF_STATE);
    __Pyx_GIVEREF(__pyx_n_s_PROOF_STATE);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_PROOF_STATE);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":76
 *     __Pyx_globals["Graph_Node_Label_Which"] = __Pyx_EnumBase('Graph_Node_Label_Which', __Pyx_OrderedDict([
 *         ('PROOF_STATE', <uint16_t>(Graph_Node_Label_Which.PROOF_STATE)),
 *         ('CONTEXT_DEF', <uint16_t>(Graph_Node_Label_Which.CONTEXT_DEF)),             # <<<<<<<<<<<<<<
 *         ('CONTEXT_ASSUM', <uint16_t>(Graph_Node_Label_Which.CONTEXT_ASSUM)),
 *         ('DEFINITION', <uint16_t>(Graph_Node_Label_Which.DEFINITION)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CONTEXT_DEF)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_DEF);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_DEF);
    PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_CONTEXT_DEF);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":77
 *         ('PROOF_STATE', <uint16_t>(Graph_Node_Label_Which.PROOF_STATE)),
 *         ('CONTEXT_DEF', <uint16_t>(Graph_Node_Label_Which.CONTEXT_DEF)),
 *         ('CONTEXT_ASSUM', <uint16_t>(Graph_Node_Label_Which.CONTEXT_ASSUM)),             # <<<<<<<<<<<<<<
 *         ('DEFINITION', <uint16_t>(Graph_Node_Label_Which.DEFINITION)),
 *         ('CONST_EMPTY', <uint16_t>(Graph_Node_Label_Which.CONST_EMPTY)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CONTEXT_ASSUM)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_47);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_ASSUM);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_ASSUM);
    PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_n_s_CONTEXT_ASSUM);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":78
 *         ('CONTEXT_DEF', <uint16_t>(Graph_Node_Label_Which.CONTEXT_DEF)),
 *         ('CONTEXT_ASSUM', <uint16_t>(Graph_Node_Label_Which.CONTEXT_ASSUM)),
 *         ('DEFINITION', <uint16_t>(Graph_Node_Label_Which.DEFINITION)),             # <<<<<<<<<<<<<<
 *         ('CONST_EMPTY', <uint16_t>(Graph_Node_Label_Which.CONST_EMPTY)),
 *         ('SORT_S_PROP', <uint16_t>(Graph_Node_Label_Which.SORT_S_PROP)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::DEFINITION)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_44);
    __Pyx_INCREF(__pyx_n_s_DEFINITION);
    __Pyx_GIVEREF(__pyx_n_s_DEFINITION);
    PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_DEFINITION);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":79
 *         ('CONTEXT_ASSUM', <uint16_t>(Graph_Node_Label_Which.CONTEXT_ASSUM)),
 *         ('DEFINITION', <uint16_t>(Graph_Node_Label_Which.DEFINITION)),
 *         ('CONST_EMPTY', <uint16_t>(Graph_Node_Label_Which.CONST_EMPTY)),             # <<<<<<<<<<<<<<
 *         ('SORT_S_PROP', <uint16_t>(Graph_Node_Label_Which.SORT_S_PROP)),
 *         ('SORT_PROP', <uint16_t>(Graph_Node_Label_Which.SORT_PROP)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CONST_EMPTY)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_43);
    __Pyx_INCREF(__pyx_n_s_CONST_EMPTY);
    __Pyx_GIVEREF(__pyx_n_s_CONST_EMPTY);
    PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_s_CONST_EMPTY);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":80
 *         ('DEFINITION', <uint16_t>(Graph_Node_Label_Which.DEFINITION)),
 *         ('CONST_EMPTY', <uint16_t>(Graph_Node_Label_Which.CONST_EMPTY)),
 *         ('SORT_S_PROP', <uint16_t>(Graph_Node_Label_Which.SORT_S_PROP)),             # <<<<<<<<<<<<<<
 *         ('SORT_PROP', <uint16_t>(Graph_Node_Label_Which.SORT_PROP)),
 *         ('SORT_SET', <uint16_t>(Graph_Node_Label_Which.SORT_SET)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::SORT_S_PROP)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_42);
    __Pyx_INCREF(__pyx_n_s_SORT_S_PROP);
    __Pyx_GIVEREF(__pyx_n_s_SORT_S_PROP);
    PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_s_SORT_S_PROP);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":81
 *         ('CONST_EMPTY', <uint16_t>(Graph_Node_Label_Which.CONST_EMPTY)),
 *         ('SORT_S_PROP', <uint16_t>(Graph_Node_Label_Which.SORT_S_PROP)),
 *         ('SORT_PROP', <uint16_t>(Graph_Node_Label_Which.SORT_PROP)),             # <<<<<<<<<<<<<<
 *         ('SORT_SET', <uint16_t>(Graph_Node_Label_Which.SORT_SET)),
 *         ('SORT_TYPE', <uint16_t>(Graph_Node_Label_Which.SORT_TYPE)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::SORT_PROP)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_41);
    __Pyx_INCREF(__pyx_n_s_SORT_PROP);
    __Pyx_GIVEREF(__pyx_n_s_SORT_PROP);
    PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_s_SORT_PROP);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":82
 *         ('SORT_S_PROP', <uint16_t>(Graph_Node_Label_Which.SORT_S_PROP)),
 *         ('SORT_PROP', <uint16_t>(Graph_Node_Label_Which.SORT_PROP)),
 *         ('SORT_SET', <uint16_t>(Graph_Node_Label_Which.SORT_SET)),             # <<<<<<<<<<<<<<
 *         ('SORT_TYPE', <uint16_t>(Graph_Node_Label_Which.SORT_TYPE)),
 *         ('REL', <uint16_t>(Graph_Node_Label_Which.REL)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::SORT_SET)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_40);
    __Pyx_INCREF(__pyx_n_s_SORT_SET);
    __Pyx_GIVEREF(__pyx_n_s_SORT_SET);
    PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_SORT_SET);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":83
 *         ('SORT_PROP', <uint16_t>(Graph_Node_Label_Which.SORT_PROP)),
 *         ('SORT_SET', <uint16_t>(Graph_Node_Label_Which.SORT_SET)),
 *         ('SORT_TYPE', <uint16_t>(Graph_Node_Label_Which.SORT_TYPE)),             # <<<<<<<<<<<<<<
 *         ('REL', <uint16_t>(Graph_Node_Label_Which.REL)),
 *         ('EVAR', <uint16_t>(Graph_Node_Label_Which.EVAR)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::SORT_TYPE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_39);
    __Pyx_INCREF(__pyx_n_s_SORT_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_SORT_TYPE);
    PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_SORT_TYPE);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":84
 *         ('SORT_SET', <uint16_t>(Graph_Node_Label_Which.SORT_SET)),
 *         ('SORT_TYPE', <uint16_t>(Graph_Node_Label_Which.SORT_TYPE)),
 *         ('REL', <uint16_t>(Graph_Node_Label_Which.REL)),             # <<<<<<<<<<<<<<
 *         ('EVAR', <uint16_t>(Graph_Node_Label_Which.EVAR)),
 *         ('EVAR_SUBST', <uint16_t>(Graph_Node_Label_Which.EVAR_SUBST)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::REL)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_38);
    __Pyx_INCREF(__pyx_n_s_REL);
    __Pyx_GIVEREF(__pyx_n_s_REL);
    PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_s_REL);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":85
 *         ('SORT_TYPE', <uint16_t>(Graph_Node_Label_Which.SORT_TYPE)),
 *         ('REL', <uint16_t>(Graph_Node_Label_Which.REL)),
 *         ('EVAR', <uint16_t>(Graph_Node_Label_Which.EVAR)),             # <<<<<<<<<<<<<<
 *         ('EVAR_SUBST', <uint16_t>(Graph_Node_Label_Which.EVAR_SUBST)),
 *         ('CAST', <uint16_t>(Graph_Node_Label_Which.CAST)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::EVAR)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_37);
    __Pyx_INCREF(__pyx_n_s_EVAR);
    __Pyx_GIVEREF(__pyx_n_s_EVAR);
    PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_s_EVAR);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":86
 *         ('REL', <uint16_t>(Graph_Node_Label_Which.REL)),
 *         ('EVAR', <uint16_t>(Graph_Node_Label_Which.EVAR)),
 *         ('EVAR_SUBST', <uint16_t>(Graph_Node_Label_Which.EVAR_SUBST)),             # <<<<<<<<<<<<<<
 *         ('CAST', <uint16_t>(Graph_Node_Label_Which.CAST)),
 *         ('PROD', <uint16_t>(Graph_Node_Label_Which.PROD)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::EVAR_SUBST)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_36);
    __Pyx_INCREF(__pyx_n_s_EVAR_SUBST);
    __Pyx_GIVEREF(__pyx_n_s_EVAR_SUBST);
    PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_s_EVAR_SUBST);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":87
 *         ('EVAR', <uint16_t>(Graph_Node_Label_Which.EVAR)),
 *         ('EVAR_SUBST', <uint16_t>(Graph_Node_Label_Which.EVAR_SUBST)),
 *         ('CAST', <uint16_t>(Graph_Node_Label_Which.CAST)),             # <<<<<<<<<<<<<<
 *         ('PROD', <uint16_t>(Graph_Node_Label_Which.PROD)),
 *         ('LAMBDA', <uint16_t>(Graph_Node_Label_Which.LAMBDA)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CAST)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_35);
    __Pyx_INCREF(__pyx_n_s_CAST);
    __Pyx_GIVEREF(__pyx_n_s_CAST);
    PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_s_CAST);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":88
 *         ('EVAR_SUBST', <uint16_t>(Graph_Node_Label_Which.EVAR_SUBST)),
 *         ('CAST', <uint16_t>(Graph_Node_Label_Which.CAST)),
 *         ('PROD', <uint16_t>(Graph_Node_Label_Which.PROD)),             # <<<<<<<<<<<<<<
 *         ('LAMBDA', <uint16_t>(Graph_Node_Label_Which.LAMBDA)),
 *         ('LET_IN', <uint16_t>(Graph_Node_Label_Which.LET_IN)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::PROD)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_34);
    __Pyx_INCREF(__pyx_n_s_PROD);
    __Pyx_GIVEREF(__pyx_n_s_PROD);
    PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_s_PROD);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":89
 *         ('CAST', <uint16_t>(Graph_Node_Label_Which.CAST)),
 *         ('PROD', <uint16_t>(Graph_Node_Label_Which.PROD)),
 *         ('LAMBDA', <uint16_t>(Graph_Node_Label_Which.LAMBDA)),             # <<<<<<<<<<<<<<
 *         ('LET_IN', <uint16_t>(Graph_Node_Label_Which.LET_IN)),
 *         ('APP', <uint16_t>(Graph_Node_Label_Which.APP)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::LAMBDA)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_33);
    __Pyx_INCREF(__pyx_n_s_LAMBDA);
    __Pyx_GIVEREF(__pyx_n_s_LAMBDA);
    PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_s_LAMBDA);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":90
 *         ('PROD', <uint16_t>(Graph_Node_Label_Which.PROD)),
 *         ('LAMBDA', <uint16_t>(Graph_Node_Label_Which.LAMBDA)),
 *         ('LET_IN', <uint16_t>(Graph_Node_Label_Which.LET_IN)),             # <<<<<<<<<<<<<<
 *         ('APP', <uint16_t>(Graph_Node_Label_Which.APP)),
 *         ('CASE', <uint16_t>(Graph_Node_Label_Which.CASE)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::LET_IN)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_INCREF(__pyx_n_s_LET_IN);
    __Pyx_GIVEREF(__pyx_n_s_LET_IN);
    PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_s_LET_IN);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":91
 *         ('LAMBDA', <uint16_t>(Graph_Node_Label_Which.LAMBDA)),
 *         ('LET_IN', <uint16_t>(Graph_Node_Label_Which.LET_IN)),
 *         ('APP', <uint16_t>(Graph_Node_Label_Which.APP)),             # <<<<<<<<<<<<<<
 *         ('CASE', <uint16_t>(Graph_Node_Label_Which.CASE)),
 *         ('CASE_BRANCH', <uint16_t>(Graph_Node_Label_Which.CASE_BRANCH)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::APP)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_INCREF(__pyx_n_s_APP);
    __Pyx_GIVEREF(__pyx_n_s_APP);
    PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_n_s_APP);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":92
 *         ('LET_IN', <uint16_t>(Graph_Node_Label_Which.LET_IN)),
 *         ('APP', <uint16_t>(Graph_Node_Label_Which.APP)),
 *         ('CASE', <uint16_t>(Graph_Node_Label_Which.CASE)),             # <<<<<<<<<<<<<<
 *         ('CASE_BRANCH', <uint16_t>(Graph_Node_Label_Which.CASE_BRANCH)),
 *         ('FIX', <uint16_t>(Graph_Node_Label_Which.FIX)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CASE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_INCREF(__pyx_n_s_CASE);
    __Pyx_GIVEREF(__pyx_n_s_CASE);
    PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_n_s_CASE);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":93
 *         ('APP', <uint16_t>(Graph_Node_Label_Which.APP)),
 *         ('CASE', <uint16_t>(Graph_Node_Label_Which.CASE)),
 *         ('CASE_BRANCH', <uint16_t>(Graph_Node_Label_Which.CASE_BRANCH)),             # <<<<<<<<<<<<<<
 *         ('FIX', <uint16_t>(Graph_Node_Label_Which.FIX)),
 *         ('FIX_FUN', <uint16_t>(Graph_Node_Label_Which.FIX_FUN)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CASE_BRANCH)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_INCREF(__pyx_n_s_CASE_BRANCH);
    __Pyx_GIVEREF(__pyx_n_s_CASE_BRANCH);
    PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_s_CASE_BRANCH);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":94
 *         ('CASE', <uint16_t>(Graph_Node_Label_Which.CASE)),
 *         ('CASE_BRANCH', <uint16_t>(Graph_Node_Label_Which.CASE_BRANCH)),
 *         ('FIX', <uint16_t>(Graph_Node_Label_Which.FIX)),             # <<<<<<<<<<<<<<
 *         ('FIX_FUN', <uint16_t>(Graph_Node_Label_Which.FIX_FUN)),
 *         ('CO_FIX', <uint16_t>(Graph_Node_Label_Which.CO_FIX)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::FIX)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_FIX);
    __Pyx_GIVEREF(__pyx_n_s_FIX);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_FIX);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":95
 *         ('CASE_BRANCH', <uint16_t>(Graph_Node_Label_Which.CASE_BRANCH)),
 *         ('FIX', <uint16_t>(Graph_Node_Label_Which.FIX)),
 *         ('FIX_FUN', <uint16_t>(Graph_Node_Label_Which.FIX_FUN)),             # <<<<<<<<<<<<<<
 *         ('CO_FIX', <uint16_t>(Graph_Node_Label_Which.CO_FIX)),
 *         ('CO_FIX_FUN', <uint16_t>(Graph_Node_Label_Which.CO_FIX_FUN)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::FIX_FUN)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_FIX_FUN);
    __Pyx_GIVEREF(__pyx_n_s_FIX_FUN);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_s_FIX_FUN);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":96
 *         ('FIX', <uint16_t>(Graph_Node_Label_Which.FIX)),
 *         ('FIX_FUN', <uint16_t>(Graph_Node_Label_Which.FIX_FUN)),
 *         ('CO_FIX', <uint16_t>(Graph_Node_Label_Which.CO_FIX)),             # <<<<<<<<<<<<<<
 *         ('CO_FIX_FUN', <uint16_t>(Graph_Node_Label_Which.CO_FIX_FUN)),
 *         ('INT', <uint16_t>(Graph_Node_Label_Which.INT)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CO_FIX)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_CO_FIX);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_CO_FIX);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":97
 *         ('FIX_FUN', <uint16_t>(Graph_Node_Label_Which.FIX_FUN)),
 *         ('CO_FIX', <uint16_t>(Graph_Node_Label_Which.CO_FIX)),
 *         ('CO_FIX_FUN', <uint16_t>(Graph_Node_Label_Which.CO_FIX_FUN)),             # <<<<<<<<<<<<<<
 *         ('INT', <uint16_t>(Graph_Node_Label_Which.INT)),
 *         ('FLOAT', <uint16_t>(Graph_Node_Label_Which.FLOAT)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CO_FIX_FUN)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_CO_FIX_FUN);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX_FUN);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_CO_FIX_FUN);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":98
 *         ('CO_FIX', <uint16_t>(Graph_Node_Label_Which.CO_FIX)),
 *         ('CO_FIX_FUN', <uint16_t>(Graph_Node_Label_Which.CO_FIX_FUN)),
 *         ('INT', <uint16_t>(Graph_Node_Label_Which.INT)),             # <<<<<<<<<<<<<<
 *         ('FLOAT', <uint16_t>(Graph_Node_Label_Which.FLOAT)),
 *         ('PRIMITIVE', <uint16_t>(Graph_Node_Label_Which.PRIMITIVE)),
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::INT)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_INT);
    __Pyx_GIVEREF(__pyx_n_s_INT);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_INT);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":99
 *         ('CO_FIX_FUN', <uint16_t>(Graph_Node_Label_Which.CO_FIX_FUN)),
 *         ('INT', <uint16_t>(Graph_Node_Label_Which.INT)),
 *         ('FLOAT', <uint16_t>(Graph_Node_Label_Which.FLOAT)),             # <<<<<<<<<<<<<<
 *         ('PRIMITIVE', <uint16_t>(Graph_Node_Label_Which.PRIMITIVE)),
 *     ]))
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::FLOAT)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_FLOAT);
    __Pyx_GIVEREF(__pyx_n_s_FLOAT);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_FLOAT);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":100
 *         ('INT', <uint16_t>(Graph_Node_Label_Which.INT)),
 *         ('FLOAT', <uint16_t>(Graph_Node_Label_Which.FLOAT)),
 *         ('PRIMITIVE', <uint16_t>(Graph_Node_Label_Which.PRIMITIVE)),             # <<<<<<<<<<<<<<
 *     ]))
 * 
 */
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::PRIMITIVE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_PRIMITIVE);
    __Pyx_GIVEREF(__pyx_n_s_PRIMITIVE);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_s_PRIMITIVE);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_46);
    __pyx_t_46 = 0;

    /* "CppScopedEnumType":74
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     __Pyx_globals["Graph_Node_Label_Which"] = __Pyx_EnumBase('Graph_Node_Label_Which', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('PROOF_STATE', <uint16_t>(Graph_Node_Label_Which.PROOF_STATE)),
 *         ('CONTEXT_DEF', <uint16_t>(Graph_Node_Label_Which.CONTEXT_DEF)),
 */
    __pyx_t_46 = PyList_New(26); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_9);
    PyList_SET_ITEM(__pyx_t_46, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_45);
    PyList_SET_ITEM(__pyx_t_46, 1, __pyx_t_45);
    __Pyx_GIVEREF(__pyx_t_47);
    PyList_SET_ITEM(__pyx_t_46, 2, __pyx_t_47);
    __Pyx_GIVEREF(__pyx_t_44);
    PyList_SET_ITEM(__pyx_t_46, 3, __pyx_t_44);
    __Pyx_GIVEREF(__pyx_t_43);
    PyList_SET_ITEM(__pyx_t_46, 4, __pyx_t_43);
    __Pyx_GIVEREF(__pyx_t_42);
    PyList_SET_ITEM(__pyx_t_46, 5, __pyx_t_42);
    __Pyx_GIVEREF(__pyx_t_41);
    PyList_SET_ITEM(__pyx_t_46, 6, __pyx_t_41);
    __Pyx_GIVEREF(__pyx_t_40);
    PyList_SET_ITEM(__pyx_t_46, 7, __pyx_t_40);
    __Pyx_GIVEREF(__pyx_t_39);
    PyList_SET_ITEM(__pyx_t_46, 8, __pyx_t_39);
    __Pyx_GIVEREF(__pyx_t_38);
    PyList_SET_ITEM(__pyx_t_46, 9, __pyx_t_38);
    __Pyx_GIVEREF(__pyx_t_37);
    PyList_SET_ITEM(__pyx_t_46, 10, __pyx_t_37);
    __Pyx_GIVEREF(__pyx_t_36);
    PyList_SET_ITEM(__pyx_t_46, 11, __pyx_t_36);
    __Pyx_GIVEREF(__pyx_t_35);
    PyList_SET_ITEM(__pyx_t_46, 12, __pyx_t_35);
    __Pyx_GIVEREF(__pyx_t_34);
    PyList_SET_ITEM(__pyx_t_46, 13, __pyx_t_34);
    __Pyx_GIVEREF(__pyx_t_33);
    PyList_SET_ITEM(__pyx_t_46, 14, __pyx_t_33);
    __Pyx_GIVEREF(__pyx_t_32);
    PyList_SET_ITEM(__pyx_t_46, 15, __pyx_t_32);
    __Pyx_GIVEREF(__pyx_t_31);
    PyList_SET_ITEM(__pyx_t_46, 16, __pyx_t_31);
    __Pyx_GIVEREF(__pyx_t_30);
    PyList_SET_ITEM(__pyx_t_46, 17, __pyx_t_30);
    __Pyx_GIVEREF(__pyx_t_29);
    PyList_SET_ITEM(__pyx_t_46, 18, __pyx_t_29);
    __Pyx_GIVEREF(__pyx_t_28);
    PyList_SET_ITEM(__pyx_t_46, 19, __pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_27);
    PyList_SET_ITEM(__pyx_t_46, 20, __pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_26);
    PyList_SET_ITEM(__pyx_t_46, 21, __pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_25);
    PyList_SET_ITEM(__pyx_t_46, 22, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_46, 23, __pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_23);
    PyList_SET_ITEM(__pyx_t_46, 24, __pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_22);
    PyList_SET_ITEM(__pyx_t_46, 25, __pyx_t_22);
    __pyx_t_9 = 0;
    __pyx_t_45 = 0;
    __pyx_t_47 = 0;
    __pyx_t_44 = 0;
    __pyx_t_43 = 0;
    __pyx_t_42 = 0;
    __pyx_t_41 = 0;
    __pyx_t_40 = 0;
    __pyx_t_39 = 0;
    __pyx_t_38 = 0;
    __pyx_t_37 = 0;
    __pyx_t_36 = 0;
    __pyx_t_35 = 0;
    __pyx_t_34 = 0;
    __pyx_t_33 = 0;
    __pyx_t_32 = 0;
    __pyx_t_31 = 0;
    __pyx_t_30 = 0;
    __pyx_t_29 = 0;
    __pyx_t_28 = 0;
    __pyx_t_27 = 0;
    __pyx_t_26 = 0;
    __pyx_t_25 = 0;
    __pyx_t_24 = 0;
    __pyx_t_23 = 0;
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_46); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_Graph_Node_Label_Which);
    __Pyx_GIVEREF(__pyx_n_s_Graph_Node_Label_Which);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_Graph_Node_Label_Which);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_22);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_46, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which, __pyx_t_22) < 0))) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["Graph_Node_Label_Which"] = __Pyx_EnumBase('Graph_Node_Label_Which', __Pyx_OrderedDict([
 */
    goto __pyx_L9;
  }

  /* "CppScopedEnumType":104
 * 
 * else:
 *     __Pyx_globals["Graph_Node_Label_Which"] = type('Graph_Node_Label_Which', (__Pyx_EnumBase,), {})             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PROOF_STATE), 'PROOF_STATE')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONTEXT_DEF), 'CONTEXT_DEF')
 */
  /*else*/ {
    __pyx_t_22 = PyTuple_New(1); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __Pyx_EnumBase);
    __pyx_t_46 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = PyTuple_New(3); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_Graph_Node_Label_Which);
    __Pyx_GIVEREF(__pyx_n_s_Graph_Node_Label_Which);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_Graph_Node_Label_Which);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_23, 2, __pyx_t_46);
    __pyx_t_22 = 0;
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_23, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 104, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which, __pyx_t_46) < 0))) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":105
 * else:
 *     __Pyx_globals["Graph_Node_Label_Which"] = type('Graph_Node_Label_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PROOF_STATE), 'PROOF_STATE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONTEXT_DEF), 'CONTEXT_DEF')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONTEXT_ASSUM), 'CONTEXT_ASSUM')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 105, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::PROOF_STATE)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_PROOF_STATE);
    __Pyx_GIVEREF(__pyx_n_s_PROOF_STATE);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_n_s_PROOF_STATE);
    __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_22, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "CppScopedEnumType":106
 *     __Pyx_globals["Graph_Node_Label_Which"] = type('Graph_Node_Label_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PROOF_STATE), 'PROOF_STATE')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONTEXT_DEF), 'CONTEXT_DEF')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONTEXT_ASSUM), 'CONTEXT_ASSUM')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.DEFINITION), 'DEFINITION')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 106, __pyx_L1_error)
    }
    __pyx_t_23 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CONTEXT_DEF)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_DEF);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_DEF);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_CONTEXT_DEF);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_46, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":107
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PROOF_STATE), 'PROOF_STATE')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONTEXT_DEF), 'CONTEXT_DEF')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONTEXT_ASSUM), 'CONTEXT_ASSUM')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.DEFINITION), 'DEFINITION')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONST_EMPTY), 'CONST_EMPTY')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 107, __pyx_L1_error)
    }
    __pyx_t_22 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CONTEXT_ASSUM)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_CONTEXT_ASSUM);
    __Pyx_GIVEREF(__pyx_n_s_CONTEXT_ASSUM);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_n_s_CONTEXT_ASSUM);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_23, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":108
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONTEXT_DEF), 'CONTEXT_DEF')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONTEXT_ASSUM), 'CONTEXT_ASSUM')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.DEFINITION), 'DEFINITION')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONST_EMPTY), 'CONST_EMPTY')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_S_PROP), 'SORT_S_PROP')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 108, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::DEFINITION)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_DEFINITION);
    __Pyx_GIVEREF(__pyx_n_s_DEFINITION);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_n_s_DEFINITION);
    __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_22, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "CppScopedEnumType":109
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONTEXT_ASSUM), 'CONTEXT_ASSUM')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.DEFINITION), 'DEFINITION')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONST_EMPTY), 'CONST_EMPTY')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_S_PROP), 'SORT_S_PROP')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_PROP), 'SORT_PROP')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 109, __pyx_L1_error)
    }
    __pyx_t_23 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CONST_EMPTY)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_CONST_EMPTY);
    __Pyx_GIVEREF(__pyx_n_s_CONST_EMPTY);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_CONST_EMPTY);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_46, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":110
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.DEFINITION), 'DEFINITION')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONST_EMPTY), 'CONST_EMPTY')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_S_PROP), 'SORT_S_PROP')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_PROP), 'SORT_PROP')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_SET), 'SORT_SET')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 110, __pyx_L1_error)
    }
    __pyx_t_22 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::SORT_S_PROP)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_SORT_S_PROP);
    __Pyx_GIVEREF(__pyx_n_s_SORT_S_PROP);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_n_s_SORT_S_PROP);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_23, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":111
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CONST_EMPTY), 'CONST_EMPTY')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_S_PROP), 'SORT_S_PROP')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_PROP), 'SORT_PROP')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_SET), 'SORT_SET')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_TYPE), 'SORT_TYPE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 111, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::SORT_PROP)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_SORT_PROP);
    __Pyx_GIVEREF(__pyx_n_s_SORT_PROP);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_n_s_SORT_PROP);
    __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_22, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "CppScopedEnumType":112
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_S_PROP), 'SORT_S_PROP')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_PROP), 'SORT_PROP')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_SET), 'SORT_SET')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_TYPE), 'SORT_TYPE')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.REL), 'REL')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 112, __pyx_L1_error)
    }
    __pyx_t_23 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::SORT_SET)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_SORT_SET);
    __Pyx_GIVEREF(__pyx_n_s_SORT_SET);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_SORT_SET);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_46, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":113
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_PROP), 'SORT_PROP')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_SET), 'SORT_SET')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_TYPE), 'SORT_TYPE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.REL), 'REL')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.EVAR), 'EVAR')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 113, __pyx_L1_error)
    }
    __pyx_t_22 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::SORT_TYPE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_SORT_TYPE);
    __Pyx_GIVEREF(__pyx_n_s_SORT_TYPE);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_n_s_SORT_TYPE);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_23, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":114
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_SET), 'SORT_SET')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_TYPE), 'SORT_TYPE')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.REL), 'REL')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.EVAR), 'EVAR')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.EVAR_SUBST), 'EVAR_SUBST')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 114, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::REL)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_REL);
    __Pyx_GIVEREF(__pyx_n_s_REL);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_n_s_REL);
    __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_22, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "CppScopedEnumType":115
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.SORT_TYPE), 'SORT_TYPE')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.REL), 'REL')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.EVAR), 'EVAR')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.EVAR_SUBST), 'EVAR_SUBST')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CAST), 'CAST')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 115, __pyx_L1_error)
    }
    __pyx_t_23 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::EVAR)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_EVAR);
    __Pyx_GIVEREF(__pyx_n_s_EVAR);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_EVAR);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_46, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":116
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.REL), 'REL')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.EVAR), 'EVAR')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.EVAR_SUBST), 'EVAR_SUBST')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CAST), 'CAST')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PROD), 'PROD')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 116, __pyx_L1_error)
    }
    __pyx_t_22 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::EVAR_SUBST)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_EVAR_SUBST);
    __Pyx_GIVEREF(__pyx_n_s_EVAR_SUBST);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_n_s_EVAR_SUBST);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_23, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":117
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.EVAR), 'EVAR')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.EVAR_SUBST), 'EVAR_SUBST')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CAST), 'CAST')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PROD), 'PROD')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.LAMBDA), 'LAMBDA')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 117, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CAST)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_CAST);
    __Pyx_GIVEREF(__pyx_n_s_CAST);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_n_s_CAST);
    __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_22, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "CppScopedEnumType":118
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.EVAR_SUBST), 'EVAR_SUBST')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CAST), 'CAST')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PROD), 'PROD')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.LAMBDA), 'LAMBDA')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.LET_IN), 'LET_IN')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 118, __pyx_L1_error)
    }
    __pyx_t_23 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::PROD)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_PROD);
    __Pyx_GIVEREF(__pyx_n_s_PROD);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_PROD);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_46, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":119
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CAST), 'CAST')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PROD), 'PROD')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.LAMBDA), 'LAMBDA')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.LET_IN), 'LET_IN')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.APP), 'APP')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 119, __pyx_L1_error)
    }
    __pyx_t_22 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::LAMBDA)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_LAMBDA);
    __Pyx_GIVEREF(__pyx_n_s_LAMBDA);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_n_s_LAMBDA);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_23, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":120
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PROD), 'PROD')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.LAMBDA), 'LAMBDA')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.LET_IN), 'LET_IN')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.APP), 'APP')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CASE), 'CASE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 120, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::LET_IN)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_LET_IN);
    __Pyx_GIVEREF(__pyx_n_s_LET_IN);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_n_s_LET_IN);
    __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_22, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "CppScopedEnumType":121
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.LAMBDA), 'LAMBDA')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.LET_IN), 'LET_IN')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.APP), 'APP')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CASE), 'CASE')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CASE_BRANCH), 'CASE_BRANCH')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 121, __pyx_L1_error)
    }
    __pyx_t_23 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::APP)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_APP);
    __Pyx_GIVEREF(__pyx_n_s_APP);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_APP);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_46, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":122
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.LET_IN), 'LET_IN')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.APP), 'APP')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CASE), 'CASE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CASE_BRANCH), 'CASE_BRANCH')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FIX), 'FIX')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 122, __pyx_L1_error)
    }
    __pyx_t_22 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CASE)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_CASE);
    __Pyx_GIVEREF(__pyx_n_s_CASE);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_n_s_CASE);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_23, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":123
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.APP), 'APP')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CASE), 'CASE')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CASE_BRANCH), 'CASE_BRANCH')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FIX), 'FIX')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FIX_FUN), 'FIX_FUN')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 123, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CASE_BRANCH)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_CASE_BRANCH);
    __Pyx_GIVEREF(__pyx_n_s_CASE_BRANCH);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_n_s_CASE_BRANCH);
    __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_22, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "CppScopedEnumType":124
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CASE), 'CASE')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CASE_BRANCH), 'CASE_BRANCH')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FIX), 'FIX')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FIX_FUN), 'FIX_FUN')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CO_FIX), 'CO_FIX')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 124, __pyx_L1_error)
    }
    __pyx_t_23 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::FIX)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_FIX);
    __Pyx_GIVEREF(__pyx_n_s_FIX);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_FIX);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_46, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":125
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CASE_BRANCH), 'CASE_BRANCH')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FIX), 'FIX')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FIX_FUN), 'FIX_FUN')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CO_FIX), 'CO_FIX')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CO_FIX_FUN), 'CO_FIX_FUN')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 125, __pyx_L1_error)
    }
    __pyx_t_22 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::FIX_FUN)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_FIX_FUN);
    __Pyx_GIVEREF(__pyx_n_s_FIX_FUN);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_n_s_FIX_FUN);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_23, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":126
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FIX), 'FIX')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FIX_FUN), 'FIX_FUN')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CO_FIX), 'CO_FIX')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CO_FIX_FUN), 'CO_FIX_FUN')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.INT), 'INT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 126, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CO_FIX)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_CO_FIX);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_n_s_CO_FIX);
    __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_22, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "CppScopedEnumType":127
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FIX_FUN), 'FIX_FUN')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CO_FIX), 'CO_FIX')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CO_FIX_FUN), 'CO_FIX_FUN')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.INT), 'INT')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FLOAT), 'FLOAT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 127, __pyx_L1_error)
    }
    __pyx_t_23 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::CO_FIX_FUN)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_CO_FIX_FUN);
    __Pyx_GIVEREF(__pyx_n_s_CO_FIX_FUN);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_CO_FIX_FUN);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_46, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":128
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CO_FIX), 'CO_FIX')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CO_FIX_FUN), 'CO_FIX_FUN')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.INT), 'INT')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FLOAT), 'FLOAT')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PRIMITIVE), 'PRIMITIVE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 128, __pyx_L1_error)
    }
    __pyx_t_22 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::INT)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_INT);
    __Pyx_GIVEREF(__pyx_n_s_INT);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_n_s_INT);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_23, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":129
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.CO_FIX_FUN), 'CO_FIX_FUN')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.INT), 'INT')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FLOAT), 'FLOAT')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PRIMITIVE), 'PRIMITIVE')
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 129, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_23 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::FLOAT)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_FLOAT);
    __Pyx_GIVEREF(__pyx_n_s_FLOAT);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_n_s_FLOAT);
    __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_22, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "CppScopedEnumType":130
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.INT), 'INT')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.FLOAT), 'FLOAT')
 *     __Pyx_globals["Graph_Node_Label_Which"](<uint16_t>(Graph_Node_Label_Which.PRIMITIVE), 'PRIMITIVE')             # <<<<<<<<<<<<<<
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 130, __pyx_L1_error)
    }
    __pyx_t_23 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_Graph_Node_Label_Which); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)Graph::Node::Label::Which::PRIMITIVE)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_PRIMITIVE);
    __Pyx_GIVEREF(__pyx_n_s_PRIMITIVE);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_PRIMITIVE);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_46, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  }
  __pyx_L9:;

  /* "CppScopedEnumType":70
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __pyx_t_22 = __Pyx_Globals(); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  if (!(likely(PyDict_CheckExact(__pyx_t_22))||((__pyx_t_22) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_22))) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_22));
  __Pyx_GIVEREF(__pyx_t_22);
  __pyx_t_22 = 0;

  /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["PredictionProtocol_Request_Which"] = __Pyx_EnumBase('PredictionProtocol_Request_Which', __Pyx_OrderedDict([
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "CppScopedEnumType":75
 * 
 *     __Pyx_globals["PredictionProtocol_Request_Which"] = __Pyx_EnumBase('PredictionProtocol_Request_Which', __Pyx_OrderedDict([
 *         ('INITIALIZE', <uint16_t>(PredictionProtocol_Request_Which.INITIALIZE)),             # <<<<<<<<<<<<<<
 *         ('PREDICT', <uint16_t>(PredictionProtocol_Request_Which.PREDICT)),
 *         ('SYNCHRONIZE', <uint16_t>(PredictionProtocol_Request_Which.SYNCHRONIZE)),
 */
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Request::Which::INITIALIZE)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_INITIALIZE);
    __Pyx_GIVEREF(__pyx_n_s_INITIALIZE);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_INITIALIZE);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_22);
    __pyx_t_22 = 0;

    /* "CppScopedEnumType":76
 *     __Pyx_globals["PredictionProtocol_Request_Which"] = __Pyx_EnumBase('PredictionProtocol_Request_Which', __Pyx_OrderedDict([
 *         ('INITIALIZE', <uint16_t>(PredictionProtocol_Request_Which.INITIALIZE)),
 *         ('PREDICT', <uint16_t>(PredictionProtocol_Request_Which.PREDICT)),             # <<<<<<<<<<<<<<
 *         ('SYNCHRONIZE', <uint16_t>(PredictionProtocol_Request_Which.SYNCHRONIZE)),
 *         ('CHECK_ALIGNMENT', <uint16_t>(PredictionProtocol_Request_Which.CHECK_ALIGNMENT)),
 */
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Request::Which::PREDICT)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_PREDICT);
    __Pyx_GIVEREF(__pyx_n_s_PREDICT);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_PREDICT);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_22);
    __pyx_t_22 = 0;

    /* "CppScopedEnumType":77
 *         ('INITIALIZE', <uint16_t>(PredictionProtocol_Request_Which.INITIALIZE)),
 *         ('PREDICT', <uint16_t>(PredictionProtocol_Request_Which.PREDICT)),
 *         ('SYNCHRONIZE', <uint16_t>(PredictionProtocol_Request_Which.SYNCHRONIZE)),             # <<<<<<<<<<<<<<
 *         ('CHECK_ALIGNMENT', <uint16_t>(PredictionProtocol_Request_Which.CHECK_ALIGNMENT)),
 *     ]))
 */
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Request::Which::SYNCHRONIZE)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_SYNCHRONIZE);
    __Pyx_GIVEREF(__pyx_n_s_SYNCHRONIZE);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_SYNCHRONIZE);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_22);
    __pyx_t_22 = 0;

    /* "CppScopedEnumType":78
 *         ('PREDICT', <uint16_t>(PredictionProtocol_Request_Which.PREDICT)),
 *         ('SYNCHRONIZE', <uint16_t>(PredictionProtocol_Request_Which.SYNCHRONIZE)),
 *         ('CHECK_ALIGNMENT', <uint16_t>(PredictionProtocol_Request_Which.CHECK_ALIGNMENT)),             # <<<<<<<<<<<<<<
 *     ]))
 * 
 */
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Request::Which::CHECK_ALIGNMENT)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_CHECK_ALIGNMENT);
    __Pyx_GIVEREF(__pyx_n_s_CHECK_ALIGNMENT);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_CHECK_ALIGNMENT);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_22);
    __pyx_t_22 = 0;

    /* "CppScopedEnumType":74
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     __Pyx_globals["PredictionProtocol_Request_Which"] = __Pyx_EnumBase('PredictionProtocol_Request_Which', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('INITIALIZE', <uint16_t>(PredictionProtocol_Request_Which.INITIALIZE)),
 *         ('PREDICT', <uint16_t>(PredictionProtocol_Request_Which.PREDICT)),
 */
    __pyx_t_22 = PyList_New(4); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_46);
    PyList_SET_ITEM(__pyx_t_22, 0, __pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_23);
    PyList_SET_ITEM(__pyx_t_22, 1, __pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_22, 2, __pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_25);
    PyList_SET_ITEM(__pyx_t_22, 3, __pyx_t_25);
    __pyx_t_46 = 0;
    __pyx_t_23 = 0;
    __pyx_t_24 = 0;
    __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_22); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_PredictionProtocol_Request_Which);
    __Pyx_GIVEREF(__pyx_n_s_PredictionProtocol_Request_Which);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_s_PredictionProtocol_Request_Which);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_25);
    __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_22, NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Request_Which, __pyx_t_25) < 0))) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["PredictionProtocol_Request_Which"] = __Pyx_EnumBase('PredictionProtocol_Request_Which', __Pyx_OrderedDict([
 */
    goto __pyx_L10;
  }

  /* "CppScopedEnumType":82
 * 
 * else:
 *     __Pyx_globals["PredictionProtocol_Request_Which"] = type('PredictionProtocol_Request_Which', (__Pyx_EnumBase,), {})             # <<<<<<<<<<<<<<
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.INITIALIZE), 'INITIALIZE')
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.PREDICT), 'PREDICT')
 */
  /*else*/ {
    __pyx_t_25 = PyTuple_New(1); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __Pyx_EnumBase);
    __pyx_t_22 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_24 = PyTuple_New(3); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_PredictionProtocol_Request_Which);
    __Pyx_GIVEREF(__pyx_n_s_PredictionProtocol_Request_Which);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_PredictionProtocol_Request_Which);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_24, 2, __pyx_t_22);
    __pyx_t_25 = 0;
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_24, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 82, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Request_Which, __pyx_t_22) < 0))) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":83
 * else:
 *     __Pyx_globals["PredictionProtocol_Request_Which"] = type('PredictionProtocol_Request_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.INITIALIZE), 'INITIALIZE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.PREDICT), 'PREDICT')
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.SYNCHRONIZE), 'SYNCHRONIZE')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 83, __pyx_L1_error)
    }
    __pyx_t_22 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Request_Which); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_24 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Request::Which::INITIALIZE)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_INITIALIZE);
    __Pyx_GIVEREF(__pyx_n_s_INITIALIZE);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_n_s_INITIALIZE);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_25, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "CppScopedEnumType":84
 *     __Pyx_globals["PredictionProtocol_Request_Which"] = type('PredictionProtocol_Request_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.INITIALIZE), 'INITIALIZE')
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.PREDICT), 'PREDICT')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.SYNCHRONIZE), 'SYNCHRONIZE')
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.CHECK_ALIGNMENT), 'CHECK_ALIGNMENT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 84, __pyx_L1_error)
    }
    __pyx_t_24 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Request_Which); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Request::Which::PREDICT)); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_PREDICT);
    __Pyx_GIVEREF(__pyx_n_s_PREDICT);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_n_s_PREDICT);
    __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_22, NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "CppScopedEnumType":85
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.INITIALIZE), 'INITIALIZE')
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.PREDICT), 'PREDICT')
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.SYNCHRONIZE), 'SYNCHRONIZE')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.CHECK_ALIGNMENT), 'CHECK_ALIGNMENT')
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 85, __pyx_L1_error)
    }
    __pyx_t_25 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Request_Which); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_22 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Request::Which::SYNCHRONIZE)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_SYNCHRONIZE);
    __Pyx_GIVEREF(__pyx_n_s_SYNCHRONIZE);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_SYNCHRONIZE);
    __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_25, __pyx_t_24, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "CppScopedEnumType":86
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.PREDICT), 'PREDICT')
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.SYNCHRONIZE), 'SYNCHRONIZE')
 *     __Pyx_globals["PredictionProtocol_Request_Which"](<uint16_t>(PredictionProtocol_Request_Which.CHECK_ALIGNMENT), 'CHECK_ALIGNMENT')             # <<<<<<<<<<<<<<
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 86, __pyx_L1_error)
    }
    __pyx_t_22 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Request_Which); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_24 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Request::Which::CHECK_ALIGNMENT)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_CHECK_ALIGNMENT);
    __Pyx_GIVEREF(__pyx_n_s_CHECK_ALIGNMENT);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_n_s_CHECK_ALIGNMENT);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_25, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  }
  __pyx_L10:;

  /* "CppScopedEnumType":70
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __pyx_t_24 = __Pyx_Globals(); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (!(likely(PyDict_CheckExact(__pyx_t_24))||((__pyx_t_24) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_24))) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_24));
  __Pyx_GIVEREF(__pyx_t_24);
  __pyx_t_24 = 0;

  /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["PredictionProtocol_Response_Which"] = __Pyx_EnumBase('PredictionProtocol_Response_Which', __Pyx_OrderedDict([
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "CppScopedEnumType":75
 * 
 *     __Pyx_globals["PredictionProtocol_Response_Which"] = __Pyx_EnumBase('PredictionProtocol_Response_Which', __Pyx_OrderedDict([
 *         ('INITIALIZED', <uint16_t>(PredictionProtocol_Response_Which.INITIALIZED)),             # <<<<<<<<<<<<<<
 *         ('PREDICTION', <uint16_t>(PredictionProtocol_Response_Which.PREDICTION)),
 *         ('TEXT_PREDICTION', <uint16_t>(PredictionProtocol_Response_Which.TEXT_PREDICTION)),
 */
    __pyx_t_24 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Response::Which::INITIALIZED)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_INITIALIZED);
    __Pyx_GIVEREF(__pyx_n_s_INITIALIZED);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_INITIALIZED);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "CppScopedEnumType":76
 *     __Pyx_globals["PredictionProtocol_Response_Which"] = __Pyx_EnumBase('PredictionProtocol_Response_Which', __Pyx_OrderedDict([
 *         ('INITIALIZED', <uint16_t>(PredictionProtocol_Response_Which.INITIALIZED)),
 *         ('PREDICTION', <uint16_t>(PredictionProtocol_Response_Which.PREDICTION)),             # <<<<<<<<<<<<<<
 *         ('TEXT_PREDICTION', <uint16_t>(PredictionProtocol_Response_Which.TEXT_PREDICTION)),
 *         ('SYNCHRONIZED', <uint16_t>(PredictionProtocol_Response_Which.SYNCHRONIZED)),
 */
    __pyx_t_24 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Response::Which::PREDICTION)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_PREDICTION);
    __Pyx_GIVEREF(__pyx_n_s_PREDICTION);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_s_PREDICTION);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "CppScopedEnumType":77
 *         ('INITIALIZED', <uint16_t>(PredictionProtocol_Response_Which.INITIALIZED)),
 *         ('PREDICTION', <uint16_t>(PredictionProtocol_Response_Which.PREDICTION)),
 *         ('TEXT_PREDICTION', <uint16_t>(PredictionProtocol_Response_Which.TEXT_PREDICTION)),             # <<<<<<<<<<<<<<
 *         ('SYNCHRONIZED', <uint16_t>(PredictionProtocol_Response_Which.SYNCHRONIZED)),
 *         ('ALIGNMENT', <uint16_t>(PredictionProtocol_Response_Which.ALIGNMENT)),
 */
    __pyx_t_24 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Response::Which::TEXT_PREDICTION)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_TEXT_PREDICTION);
    __Pyx_GIVEREF(__pyx_n_s_TEXT_PREDICTION);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_TEXT_PREDICTION);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "CppScopedEnumType":78
 *         ('PREDICTION', <uint16_t>(PredictionProtocol_Response_Which.PREDICTION)),
 *         ('TEXT_PREDICTION', <uint16_t>(PredictionProtocol_Response_Which.TEXT_PREDICTION)),
 *         ('SYNCHRONIZED', <uint16_t>(PredictionProtocol_Response_Which.SYNCHRONIZED)),             # <<<<<<<<<<<<<<
 *         ('ALIGNMENT', <uint16_t>(PredictionProtocol_Response_Which.ALIGNMENT)),
 *     ]))
 */
    __pyx_t_24 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Response::Which::SYNCHRONIZED)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_SYNCHRONIZED);
    __Pyx_GIVEREF(__pyx_n_s_SYNCHRONIZED);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_SYNCHRONIZED);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "CppScopedEnumType":79
 *         ('TEXT_PREDICTION', <uint16_t>(PredictionProtocol_Response_Which.TEXT_PREDICTION)),
 *         ('SYNCHRONIZED', <uint16_t>(PredictionProtocol_Response_Which.SYNCHRONIZED)),
 *         ('ALIGNMENT', <uint16_t>(PredictionProtocol_Response_Which.ALIGNMENT)),             # <<<<<<<<<<<<<<
 *     ]))
 * 
 */
    __pyx_t_24 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Response::Which::ALIGNMENT)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_ALIGNMENT);
    __Pyx_GIVEREF(__pyx_n_s_ALIGNMENT);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_ALIGNMENT);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "CppScopedEnumType":74
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     __Pyx_globals["PredictionProtocol_Response_Which"] = __Pyx_EnumBase('PredictionProtocol_Response_Which', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('INITIALIZED', <uint16_t>(PredictionProtocol_Response_Which.INITIALIZED)),
 *         ('PREDICTION', <uint16_t>(PredictionProtocol_Response_Which.PREDICTION)),
 */
    __pyx_t_24 = PyList_New(5); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_25);
    PyList_SET_ITEM(__pyx_t_24, 0, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_22);
    PyList_SET_ITEM(__pyx_t_24, 1, __pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyList_SET_ITEM(__pyx_t_24, 2, __pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_46);
    PyList_SET_ITEM(__pyx_t_24, 3, __pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_26);
    PyList_SET_ITEM(__pyx_t_24, 4, __pyx_t_26);
    __pyx_t_25 = 0;
    __pyx_t_22 = 0;
    __pyx_t_23 = 0;
    __pyx_t_46 = 0;
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_24); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_PredictionProtocol_Response_Whic);
    __Pyx_GIVEREF(__pyx_n_s_PredictionProtocol_Response_Whic);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_PredictionProtocol_Response_Whic);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_26);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_24, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Response_Whic, __pyx_t_26) < 0))) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["PredictionProtocol_Response_Which"] = __Pyx_EnumBase('PredictionProtocol_Response_Which', __Pyx_OrderedDict([
 */
    goto __pyx_L11;
  }

  /* "CppScopedEnumType":83
 * 
 * else:
 *     __Pyx_globals["PredictionProtocol_Response_Which"] = type('PredictionProtocol_Response_Which', (__Pyx_EnumBase,), {})             # <<<<<<<<<<<<<<
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.INITIALIZED), 'INITIALIZED')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.PREDICTION), 'PREDICTION')
 */
  /*else*/ {
    __pyx_t_26 = PyTuple_New(1); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __Pyx_EnumBase);
    __pyx_t_24 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_46 = PyTuple_New(3); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_PredictionProtocol_Response_Whic);
    __Pyx_GIVEREF(__pyx_n_s_PredictionProtocol_Response_Whic);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_PredictionProtocol_Response_Whic);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_46, 2, __pyx_t_24);
    __pyx_t_26 = 0;
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_46, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 83, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Response_Whic, __pyx_t_24) < 0))) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "CppScopedEnumType":84
 * else:
 *     __Pyx_globals["PredictionProtocol_Response_Which"] = type('PredictionProtocol_Response_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.INITIALIZED), 'INITIALIZED')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.PREDICTION), 'PREDICTION')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.TEXT_PREDICTION), 'TEXT_PREDICTION')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 84, __pyx_L1_error)
    }
    __pyx_t_24 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Response_Whic); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Response::Which::INITIALIZED)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_INITIALIZED);
    __Pyx_GIVEREF(__pyx_n_s_INITIALIZED);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_n_s_INITIALIZED);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_26, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":85
 *     __Pyx_globals["PredictionProtocol_Response_Which"] = type('PredictionProtocol_Response_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.INITIALIZED), 'INITIALIZED')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.PREDICTION), 'PREDICTION')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.TEXT_PREDICTION), 'TEXT_PREDICTION')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.SYNCHRONIZED), 'SYNCHRONIZED')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 85, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Response_Whic); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_26 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Response::Which::PREDICTION)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_PREDICTION);
    __Pyx_GIVEREF(__pyx_n_s_PREDICTION);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_PREDICTION);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_24, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "CppScopedEnumType":86
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.INITIALIZED), 'INITIALIZED')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.PREDICTION), 'PREDICTION')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.TEXT_PREDICTION), 'TEXT_PREDICTION')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.SYNCHRONIZED), 'SYNCHRONIZED')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.ALIGNMENT), 'ALIGNMENT')
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 86, __pyx_L1_error)
    }
    __pyx_t_26 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Response_Whic); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Response::Which::TEXT_PREDICTION)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_TEXT_PREDICTION);
    __Pyx_GIVEREF(__pyx_n_s_TEXT_PREDICTION);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_TEXT_PREDICTION);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_26, __pyx_t_46, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "CppScopedEnumType":87
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.PREDICTION), 'PREDICTION')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.TEXT_PREDICTION), 'TEXT_PREDICTION')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.SYNCHRONIZED), 'SYNCHRONIZED')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.ALIGNMENT), 'ALIGNMENT')
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 87, __pyx_L1_error)
    }
    __pyx_t_24 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Response_Whic); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Response::Which::SYNCHRONIZED)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_SYNCHRONIZED);
    __Pyx_GIVEREF(__pyx_n_s_SYNCHRONIZED);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_n_s_SYNCHRONIZED);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_26, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":88
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.TEXT_PREDICTION), 'TEXT_PREDICTION')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.SYNCHRONIZED), 'SYNCHRONIZED')
 *     __Pyx_globals["PredictionProtocol_Response_Which"](<uint16_t>(PredictionProtocol_Response_Which.ALIGNMENT), 'ALIGNMENT')             # <<<<<<<<<<<<<<
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 88, __pyx_L1_error)
    }
    __pyx_t_46 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_PredictionProtocol_Response_Whic); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_26 = __Pyx_PyInt_From_uint16_t(((uint16_t)PredictionProtocol::Response::Which::ALIGNMENT)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_ALIGNMENT);
    __Pyx_GIVEREF(__pyx_n_s_ALIGNMENT);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_ALIGNMENT);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_46, __pyx_t_24, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  }
  __pyx_L11:;

  /* "CppScopedEnumType":70
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __pyx_t_26 = __Pyx_Globals(); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (!(likely(PyDict_CheckExact(__pyx_t_26))||((__pyx_t_26) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_26))) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_26));
  __Pyx_GIVEREF(__pyx_t_26);
  __pyx_t_26 = 0;

  /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["ProofStep_Tactic_Which"] = __Pyx_EnumBase('ProofStep_Tactic_Which', __Pyx_OrderedDict([
 */
  __pyx_t_6 = (PY_VERSION_HEX >= 0x03040000);
  if (__pyx_t_6) {

    /* "CppScopedEnumType":75
 * 
 *     __Pyx_globals["ProofStep_Tactic_Which"] = __Pyx_EnumBase('ProofStep_Tactic_Which', __Pyx_OrderedDict([
 *         ('UNKNOWN', <uint16_t>(ProofStep_Tactic_Which.UNKNOWN)),             # <<<<<<<<<<<<<<
 *         ('KNOWN', <uint16_t>(ProofStep_Tactic_Which.KNOWN)),
 *     ]))
 */
    __pyx_t_26 = __Pyx_PyInt_From_uint16_t(((uint16_t)ProofStep::Tactic::Which::UNKNOWN)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_UNKNOWN);
    __Pyx_GIVEREF(__pyx_n_s_UNKNOWN);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_UNKNOWN);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "CppScopedEnumType":76
 *     __Pyx_globals["ProofStep_Tactic_Which"] = __Pyx_EnumBase('ProofStep_Tactic_Which', __Pyx_OrderedDict([
 *         ('UNKNOWN', <uint16_t>(ProofStep_Tactic_Which.UNKNOWN)),
 *         ('KNOWN', <uint16_t>(ProofStep_Tactic_Which.KNOWN)),             # <<<<<<<<<<<<<<
 *     ]))
 * 
 */
    __pyx_t_26 = __Pyx_PyInt_From_uint16_t(((uint16_t)ProofStep::Tactic::Which::KNOWN)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_KNOWN);
    __Pyx_GIVEREF(__pyx_n_s_KNOWN);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_KNOWN);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "CppScopedEnumType":74
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     __Pyx_globals["ProofStep_Tactic_Which"] = __Pyx_EnumBase('ProofStep_Tactic_Which', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('UNKNOWN', <uint16_t>(ProofStep_Tactic_Which.UNKNOWN)),
 *         ('KNOWN', <uint16_t>(ProofStep_Tactic_Which.KNOWN)),
 */
    __pyx_t_26 = PyList_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_26, 0, __pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_46);
    PyList_SET_ITEM(__pyx_t_26, 1, __pyx_t_46);
    __pyx_t_24 = 0;
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_26); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_ProofStep_Tactic_Which);
    __Pyx_GIVEREF(__pyx_n_s_ProofStep_Tactic_Which);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_ProofStep_Tactic_Which);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_46);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_26, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_ProofStep_Tactic_Which, __pyx_t_46) < 0))) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

    /* "CppScopedEnumType":72
 * cdef dict __Pyx_globals = globals()
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     __Pyx_globals["ProofStep_Tactic_Which"] = __Pyx_EnumBase('ProofStep_Tactic_Which', __Pyx_OrderedDict([
 */
    goto __pyx_L12;
  }

  /* "CppScopedEnumType":80
 * 
 * else:
 *     __Pyx_globals["ProofStep_Tactic_Which"] = type('ProofStep_Tactic_Which', (__Pyx_EnumBase,), {})             # <<<<<<<<<<<<<<
 *     __Pyx_globals["ProofStep_Tactic_Which"](<uint16_t>(ProofStep_Tactic_Which.UNKNOWN), 'UNKNOWN')
 *     __Pyx_globals["ProofStep_Tactic_Which"](<uint16_t>(ProofStep_Tactic_Which.KNOWN), 'KNOWN')
 */
  /*else*/ {
    __pyx_t_46 = PyTuple_New(1); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __Pyx_EnumBase);
    __pyx_t_26 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(3); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_ProofStep_Tactic_Which);
    __Pyx_GIVEREF(__pyx_n_s_ProofStep_Tactic_Which);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_ProofStep_Tactic_Which);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 2, __pyx_t_26);
    __pyx_t_46 = 0;
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_24, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 80, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__Pyx_globals, __pyx_n_s_ProofStep_Tactic_Which, __pyx_t_26) < 0))) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "CppScopedEnumType":81
 * else:
 *     __Pyx_globals["ProofStep_Tactic_Which"] = type('ProofStep_Tactic_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["ProofStep_Tactic_Which"](<uint16_t>(ProofStep_Tactic_Which.UNKNOWN), 'UNKNOWN')             # <<<<<<<<<<<<<<
 *     __Pyx_globals["ProofStep_Tactic_Which"](<uint16_t>(ProofStep_Tactic_Which.KNOWN), 'KNOWN')
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 81, __pyx_L1_error)
    }
    __pyx_t_26 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_ProofStep_Tactic_Which); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_PyInt_From_uint16_t(((uint16_t)ProofStep::Tactic::Which::UNKNOWN)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_UNKNOWN);
    __Pyx_GIVEREF(__pyx_n_s_UNKNOWN);
    PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_n_s_UNKNOWN);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_26, __pyx_t_46, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "CppScopedEnumType":82
 *     __Pyx_globals["ProofStep_Tactic_Which"] = type('ProofStep_Tactic_Which', (__Pyx_EnumBase,), {})
 *     __Pyx_globals["ProofStep_Tactic_Which"](<uint16_t>(ProofStep_Tactic_Which.UNKNOWN), 'UNKNOWN')
 *     __Pyx_globals["ProofStep_Tactic_Which"](<uint16_t>(ProofStep_Tactic_Which.KNOWN), 'KNOWN')             # <<<<<<<<<<<<<<
 * 
 */
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 82, __pyx_L1_error)
    }
    __pyx_t_24 = __Pyx_PyDict_GetItem(__Pyx_globals, __pyx_n_s_ProofStep_Tactic_Which); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_46 = __Pyx_PyInt_From_uint16_t(((uint16_t)ProofStep::Tactic::Which::KNOWN)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_46);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_46);
    __Pyx_INCREF(__pyx_n_s_KNOWN);
    __Pyx_GIVEREF(__pyx_n_s_KNOWN);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_n_s_KNOWN);
    __pyx_t_46 = 0;
    __pyx_t_46 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_26, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_46);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  }
  __pyx_L12:;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_XDECREF(__pyx_t_38);
  __Pyx_XDECREF(__pyx_t_39);
  __Pyx_XDECREF(__pyx_t_40);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_XDECREF(__pyx_t_42);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_XDECREF(__pyx_t_44);
  __Pyx_XDECREF(__pyx_t_45);
  __Pyx_XDECREF(__pyx_t_46);
  __Pyx_XDECREF(__pyx_t_47);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pytact.graph_api_capnp_cython", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pytact.graph_api_capnp_cython");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#if _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;  // error
            return kwvalues[i];
        }
    }
    return NULL;  // not found (no exception set)
}
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        if (kwds_is_tuple) {
            if (pos >= PyTuple_GET_SIZE(kwds)) break;
            key = PyTuple_GET_ITEM(kwds, pos);
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        if (unlikely(PyTuple_GET_SIZE(kw) == 0))
            return 1;
        if (!kw_allowed) {
            key = PyTuple_GET_ITEM(kw, 0);
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < PyTuple_GET_SIZE(kw); pos++) {
            key = PyTuple_GET_ITEM(kw, pos);
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY ||  CYTHON_COMPILING_IN_LIMITED_API
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]);
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
    Py_DECREF(argstuple);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
#ifdef __Pyx_CyFunction_USED
        if (__Pyx_IsCyOrPyCFunction(func))
#else
        if (PyCFunction_Check(func))
#endif
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
                return __Pyx_PyObject_CallMethO(func, NULL);
            }
        }
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (PyCFunction_Check(func))
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
                return __Pyx_PyObject_CallMethO(func, args[0]);
            }
        }
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    #if CYTHON_VECTORCALL
    vectorcallfunc f = _PyVectorcall_Function(func);
    if (f) {
        return f(func, args, (size_t)nargs, kwargs);
    }
    #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
    if (__Pyx_CyFunction_CheckExact(func)) {
        __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
        if (f) return f(func, args, (size_t)nargs, kwargs);
    }
    #endif
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (meth) {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
#endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                #if CYTHON_COMPILING_IN_LIMITED_API
                module = PyImport_ImportModuleLevelObject(
                    name, empty_dict, empty_dict, from_list, 1);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                #endif
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            #if CYTHON_COMPILING_IN_LIMITED_API
            module = PyImport_ImportModuleLevelObject(
                name, empty_dict, empty_dict, from_list, level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__3);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    (void) spec;
    (void) type;
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg = NULL;
    return __Pyx_PyObject_FastCall(func, (&arg)+1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n = PyTuple_GET_SIZE(bases);
    for (i = 1; i < n; i++)
    {
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
            return -1;
        }
        if (dictoffset == 0 && b->tp_dictoffset)
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "extension type '%.200s' has no __dict__ slot, "
                "but base type '" __Pyx_FMT_TYPENAME "' has: "
                "either add 'cdef dict __dict__' to the extension type "
                "or add '__slots__ = [...]' to the base type",
                type_name, b_name);
            __Pyx_DECREF_TypeName(b_name);
            return -1;
        }
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* SetupReduce */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__4;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    Py_ssize_t i, nparts;
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (likely(module))
        return module;
    return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    PyObject *abi_module = PyImport_AddModule((char*) __PYX_ABI_MODULE_NAME);
    if (unlikely(!abi_module)) return NULL;
    Py_INCREF(abi_module);
    return abi_module;
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    (void) module;
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
        PyList_SET_ITEM(fromlist, 0, marker);
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    cf->m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#ifdef _Py_TPFLAGS_HAVE_VECTORCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    (void) module;
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* Py3UpdateBases */
static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base, *meth, *new_base, *result, *new_bases = NULL;
    size_bases = PyTuple_GET_SIZE(bases);
    for (i = 0; i < size_bases; i++) {
        base  = PyTuple_GET_ITEM(bases, i);
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_n_s_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                base = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base);
                Py_INCREF(base);
            }
        }
        j = PyList_GET_SIZE(new_bases);
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
    return result;
error:
    Py_XDECREF(new_bases);
    return NULL;
}

/* SetNewInClass */
static int __Pyx_SetNewInClass(PyObject *ns, PyObject *name, PyObject *value) {
#ifdef __Pyx_CyFunction_USED
    int ret;
    if (__Pyx_CyFunction_Check(value)) {
        PyObject *staticnew = PyStaticMethod_New(value);
        if (unlikely(!staticnew)) return -1;
        ret = __Pyx_SetNameInClass(ns, name, staticnew);
        Py_DECREF(staticnew);
        return ret;
    }
#endif
    return __Pyx_SetNameInClass(ns, name, value);
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectLookupSpecial */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return with_error ? __Pyx_PyObject_GetAttrStr(obj, attr_name) : __Pyx_PyObject_GetAttrStrNoError(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
#if PY_VERSION_HEX >= 0x03030000
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
#else
    CYTHON_MAYBE_UNUSED_VAR(qualname);
#endif
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
static int __Pyx_SetNamesPEP487(PyObject *type_obj) {
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *names_to_set, *key, *value, *set_name, *tmp;
    Py_ssize_t i = 0;
#if CYTHON_USE_TYPE_SLOTS
    names_to_set = PyDict_Copy(type->tp_dict);
#else
    {
        PyObject *d = PyObject_GetAttr(type_obj, __pyx_n_s_dict);
        names_to_set = NULL;
        if (likely(d)) {
            PyObject *names_to_set = PyDict_New();
            int ret = likely(names_to_set) ? PyDict_Update(names_to_set, d) : -1;
            Py_DECREF(d);
            if (unlikely(ret < 0))
                Py_CLEAR(names_to_set);
        }
    }
#endif
    if (unlikely(names_to_set == NULL))
        goto bad;
    while (PyDict_Next(names_to_set, &i, &key, &value)) {
        set_name = __Pyx_PyObject_LookupSpecialNoError(value, __pyx_n_s_set_name);
        if (unlikely(set_name != NULL)) {
            tmp = __Pyx_PyObject_Call2Args(set_name, type_obj, key);
            Py_DECREF(set_name);
            if (unlikely(tmp == NULL)) {
                __Pyx_TypeName value_type_name =
                    __Pyx_PyType_GetName(Py_TYPE(value));
                __Pyx_TypeName type_name = __Pyx_PyType_GetName(type);
                PyErr_Format(PyExc_RuntimeError,
#if PY_MAJOR_VERSION >= 3
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %R " "in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name, key, type_name);
#else
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %.100s in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name,
                    PyString_Check(key) ? PyString_AS_STRING(key) : "?",
                    type_name);
#endif
                goto bad;
            } else {
                Py_DECREF(tmp);
            }
        }
        else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
    Py_DECREF(names_to_set);
    return 0;
bad:
    Py_XDECREF(names_to_set);
    return -1;
}
static PyObject *__Pyx_InitSubclassPEP487(PyObject *type_obj, PyObject *mkw) {
#if CYTHON_USE_TYPE_SLOTS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *mro = type->tp_mro;
    Py_ssize_t i, nbases;
    if (unlikely(!mro)) goto done;
    (void) &__Pyx_GetBuiltinName;
    Py_INCREF(mro);
    nbases = PyTuple_GET_SIZE(mro);
    assert(PyTuple_GET_ITEM(mro, 0) == type_obj);
    for (i = 1; i < nbases-1; i++) {
        PyObject *base, *dict, *meth;
        base = PyTuple_GET_ITEM(mro, i);
        dict = ((PyTypeObject *)base)->tp_dict;
        meth = __Pyx_PyDict_GetItemStrWithError(dict, __pyx_n_s_init_subclass);
        if (unlikely(meth)) {
            descrgetfunc f = Py_TYPE(meth)->tp_descr_get;
            PyObject *res;
            Py_INCREF(meth);
            if (likely(f)) {
                res = f(meth, NULL, type_obj);
                Py_DECREF(meth);
                if (unlikely(!res)) goto bad;
                meth = res;
            }
            res = __Pyx_PyObject_FastCallDict(meth, NULL, 0, mkw);
            Py_DECREF(meth);
            if (unlikely(!res)) goto bad;
            Py_DECREF(res);
            goto done;
        } else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
done:
    Py_XDECREF(mro);
    return type_obj;
bad:
    Py_XDECREF(mro);
    Py_DECREF(type_obj);
    return NULL;
#else
    PyObject *super_type, *super, *func, *res;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    super_type = __Pyx_GetBuiltinName(__pyx_n_s_super);
#else
    super_type = (PyObject*) &PySuper_Type;
    (void) &__Pyx_GetBuiltinName;
#endif
    super = likely(super_type) ? __Pyx_PyObject_Call2Args(super_type, type_obj, type_obj) : NULL;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    Py_XDECREF(super_type);
#endif
    if (unlikely(!super)) {
        Py_CLEAR(type_obj);
        goto done;
    }
    func = __Pyx_PyObject_GetAttrStrNoError(super, __pyx_n_s_init_subclass);
    Py_DECREF(super);
    if (likely(!func)) {
        if (unlikely(PyErr_Occurred()))
            Py_CLEAR(type_obj);
        goto done;
    }
    res = __Pyx_PyObject_FastCallDict(func, NULL, 0, mkw);
    Py_DECREF(func);
    if (unlikely(!res))
        Py_CLEAR(type_obj);
    Py_XDECREF(res);
done:
    return type_obj;
#endif
}
#endif
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET,
#if PY_VERSION_HEX < 0x030600A4
        (metaclass == (PyObject*)&PyType_Type) ? NULL : mkw
#else
        mkw
#endif
    );
    Py_XDECREF(owned_metaclass);
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
    if (likely(result) && likely(PyType_Check(result))) {
        if (unlikely(__Pyx_SetNamesPEP487(result) < 0)) {
            Py_CLEAR(result);
        } else {
            result = __Pyx_InitSubclassPEP487(result, mkw);
        }
    }
#else
    (void) &__Pyx_GetBuiltinName;
#endif
    return result;
}

/* Globals */
static PyObject* __Pyx_Globals(void) {
    return __Pyx_NewRef(__pyx_d);
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    _PyTraceback_Add(funcname, filename, py_line);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(unsigned int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int, digit, digits[0])
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(unsigned int) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int,  digit, +digits[0])
                case -2:
                    if ((8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if ((sizeof(unsigned int) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(unsigned int) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            unsigned int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned int) -1;
        }
    } else {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint16_t(uint16_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint16_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint16_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint16_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint16_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint16_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(long) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if ((sizeof(long) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name_2);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XSETREF(name, __Pyx_NewRef(__pyx_n_s__118));
    }
    return name;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if ((sizeof(int) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compile time version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}
#endif

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#if _MSV_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
